from typing import NamedTuple


def compute_code_checksum(source: str) -> bytes:
    """
    Compute the checksum of the code of a contract.
    """
    from Crypto.Hash import keccak

    from wake.regex_parser.solidity_parser import SoliditySourceParser

    code = bytearray(source.encode("utf-8"))

    parser = SoliditySourceParser()
    parser.strip_comments(code)

    code_str = code.decode("utf-8").replace("\r\n", "\n")
    normalized_code = b"\n".join(
        b" ".join(part.encode("utf-8") for part in line.split() if part)
        for line in code_str.splitlines()
        if line.strip()
    )
    return keccak.new(data=normalized_code, digest_bits=256).digest()


class KnownContract(NamedTuple):
    project: str
    path: str
    name: str
    version: str


# fmt: off

KNOWN_CONTRACTS = {
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x00\x05\xb7\xd7\x8e\x67\x71\xa9\x43\x73\x12\x82\xb9\x99\x43\x7f\x1c\x32\xeb\x14\x89\xbf\xcf\xf1\x58\x40\xbe\x51\x61\x5f\xaa\x16": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7
    b"\x00\x3e\xca\x30\x7d\xd3\xed\x33\x36\x3a\x94\x54\xb0\x8a\x36\x33\x36\xa0\x1b\x5a\xbc\x69\x0b\xed\xaa\xb7\x82\x22\x8d\x91\x1a\xbf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.1.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.8
    b"\x00\x46\x0a\xdb\x37\xbe\xd5\xcb\xce\xd8\xa9\xb0\xfe\x55\xec\x79\xaf\xea\x5e\x51\xd3\x73\x12\xe6\x09\xd9\x52\x27\xf1\x3d\x88\xd3": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.8",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x00\x9d\x7c\xe4\x35\xe5\xbd\xc8\x2b\x99\x10\xcd\xe9\x21\x0b\xdf\x15\xbc\x11\x8d\x50\xbc\xce\x52\x83\x66\x3f\xd4\x17\xa1\xdf\x3f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingFractionalUpgradeable.sol",
            name="GovernorCountingFractionalUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x00\x9f\xfc\x6f\xf9\x7e\x48\x16\xb3\x54\xba\x16\x4c\x65\x6e\x73\x07\x16\xf0\xfe\x21\x01\x33\x2e\x6e\x71\x73\xae\x4d\x97\x9c\xa1": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/HexStrings.sol",
            name="HexStrings",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x00\xb8\x91\x64\x66\x67\x2f\x1a\xfb\x5f\xfe\x11\x19\x18\x83\x4a\xac\xee\xfb\x08\xd1\x85\x41\x56\xd2\xc8\xb2\xf4\xc3\x71\x0c\x68": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.152, v0.0.153
    b"\x00\xe3\x4b\x97\x7a\xcf\xf6\x63\x5f\xf0\x1c\xed\x8c\x47\x2a\xf2\x23\x2e\x85\x2b\xfb\x3a\x6d\xc1\xdd\x96\x1e\x21\x2c\xea\x1e\xbe": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.153",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x01\x25\x1a\x92\xb5\x6b\x01\xe6\xca\x93\x75\x27\x53\x10\xfa\x1c\x38\xda\xd0\x84\x99\xb7\xc1\x06\xbe\xce\x72\xb7\x33\xe3\x6e\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.144, v0.0.145, v0.0.146, v0.0.147
    b"\x01\x2b\x2d\x80\xb4\x84\x48\x4d\xcb\xf8\x62\xc0\x1f\xef\x07\x8c\xec\x2f\x71\xbc\xc8\x16\x50\x24\x5b\x6c\x55\x19\x06\x8e\x22\x36": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.147",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x01\x31\x51\xa5\xbf\x27\x3b\xbc\x35\x85\x90\xba\xa1\x23\x71\xe4\x68\x53\xe7\x90\x78\xc0\xc9\xfb\xdc\x21\x20\xc9\x42\xd7\x1c\x28": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x01\x6e\x46\x87\x59\x52\xa6\x09\xbc\x7f\xd8\x3b\x30\x86\x56\x3a\x69\x6c\x65\xe7\x8e\x67\x49\xc2\xf3\x7c\x4c\xf9\x6b\x77\xaf\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x01\x78\x50\x39\xd1\xf1\xc6\x30\xe1\x3d\xb7\x58\x92\x22\xa8\x9c\x52\x08\xd9\x87\x9f\xdd\x7f\xc0\x56\x16\x1d\xfe\xdc\x15\x4a\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\x01\x95\x27\xce\x88\xae\xda\x60\x3f\xa3\x4d\x96\xc0\x6c\x50\x5c\x68\x0b\xf1\x8a\x35\xf5\x77\x78\x70\x08\x03\x98\x06\xb3\x53\xfb": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoterV2.sol",
            name="IQuoterV2",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoterV2.sol",
            name="IQuoterV2",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoterV2.sol",
            name="IQuoterV2",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x01\xa2\x27\x65\x7e\x01\x23\xad\x70\x66\xcf\xaa\x9a\x0b\xd4\x44\x97\x6e\x5c\xf2\x4b\x25\xc9\x16\xf7\x5f\x49\xda\x59\xa8\x8f\x65": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113
    b"\x01\xa9\x80\xb6\xec\xb2\x97\xff\xa5\x30\x88\x37\x0c\x31\x4a\x76\xc5\x4d\x4f\x58\x51\x55\xff\x23\x2b\xf9\x7c\x93\xbb\xee\xa3\xb4": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.113",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.2
    b"\x01\xe7\x84\xa3\x8f\xe5\xe6\xf6\xe2\x90\x74\x89\xf7\xb9\xce\x0b\xd9\x1f\x35\xf6\x72\x1c\x73\x3b\x0f\xa7\x7e\x42\x4a\xd4\x6b\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.195
    b"\x01\xe8\xc8\x7e\x8f\x0b\xa9\xd6\x84\x22\xbe\xf6\x22\x79\x3c\xd9\x6d\x70\xa0\x90\x06\xf5\xa7\xd0\xb2\xee\xb4\x8c\x87\x0c\xb6\xa0": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.195",
        ),
    ],
    # Hash appears in versions: v0.0.262
    b"\x01\xff\x91\xbd\x11\xcb\x18\xfe\x4a\x7a\xae\x19\xab\xf6\x59\x24\xf2\xf2\xf8\x46\x68\x7f\x5f\xe4\xc9\x30\x13\x12\xdd\x3b\xeb\x54": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.262",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x02\x06\xae\x66\xb6\x41\xbf\x8b\x3b\x44\xc7\x07\xad\xe8\x1c\x75\xcc\xc8\x3e\xd2\x0b\x19\x85\xc5\xa2\x21\xaa\x90\xad\x73\x57\xe8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x02\x1b\x7a\xf5\x26\xbd\x4c\x07\x31\xf4\x0b\xd9\xfc\x32\x5c\x6c\xa0\xc5\xc7\x57\x80\xbe\xf5\x77\xc3\x03\x7f\x9e\x31\x54\x5e\xcc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x02\x27\x64\x16\xf6\xa5\x42\x67\xd6\x2f\x58\x73\xfd\xab\x34\xdb\xc5\xb6\x87\x1a\x21\x11\x18\x3c\xcb\xcd\xff\x8c\x19\x33\xb9\x7d": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionKey.sol",
            name="PositionKey",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionKey.sol",
            name="PositionKey",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionKey.sol",
            name="PositionKey",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionKey.sol",
            name="PositionKey",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionKey.sol",
            name="PositionKey",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x02\x4b\x2c\x79\xba\x35\x37\xb1\xe0\x0a\x16\x6f\xc2\xfb\x0e\xc3\xf5\xba\xe4\x10\x28\x33\xaa\x76\xa3\xdd\x15\x4a\xa8\xbf\x6b\xc7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x02\x4c\xb2\xd2\xac\x46\x9e\x25\xb3\x31\xf9\x23\x87\x90\xa3\x0e\x1e\xf1\x80\xa7\xfd\xe5\x50\x51\xa4\xbd\x82\x79\x87\x6c\x34\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipient.sol",
            name="IRelayRecipient",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x02\x4d\xf4\xe3\x11\x33\xde\xc7\x92\x67\xf3\x3e\x76\x7e\xd7\x15\x08\x0e\x2b\x64\xaf\xc6\x2f\x0d\x4c\x25\xc8\x1c\x0d\xd6\x5b\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.2
    b"\x02\x56\x09\x79\xa9\x52\x96\x7a\xf1\xf8\xad\x69\xb0\x0f\x49\x44\x9c\xd1\xf7\xa8\x46\xda\x17\x11\x26\x39\x78\xd5\x18\xc2\x29\x95": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x02\x6c\x2d\x94\x93\x9c\x97\x90\xef\x4c\xfa\x05\x92\x55\x15\xa8\x9d\x51\x7f\x3b\x74\x07\x7c\xaf\xe8\x17\x40\x86\x7d\xec\x95\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x02\x88\xac\x8e\x6b\x51\x02\xa2\xee\xf6\x92\x99\x49\xea\x1f\x48\x0b\x63\xc6\x66\xd4\x87\x53\x25\x15\x27\xcf\xcd\xf8\xcc\x9e\x2f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x02\xa2\x1a\x7d\x96\xfc\x7e\xaf\xdd\xe2\xff\xfe\xaa\x80\x4e\x6b\x2b\x7b\xfa\xe2\x4f\xf3\x0a\xc9\x5a\x38\x84\xa1\x25\x23\x24\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33
    b"\x02\xa5\xe0\x47\x3c\x33\xce\x0f\x62\x88\x49\x7a\xc1\xc9\x58\xe9\x2f\xa1\xd8\xbf\xb9\xcc\x88\x5d\x60\xbc\x7e\x67\xe4\x4d\x41\x45": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.33",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x02\xa9\x11\xec\x70\x58\xcd\x16\x71\x2f\xe8\x54\x46\x75\xc2\x31\xa3\x5b\x71\x27\x2f\x46\xf8\xe0\x45\xef\x27\x57\x99\x55\x66\x63": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x02\xb7\xde\xa6\xa3\x18\xa3\x5f\xd2\x1f\xb7\x58\xff\x13\xd9\x7e\xe0\x9e\xbb\x1c\x59\x56\x9a\x88\x04\xbf\xcc\x8d\x54\xd2\x9b\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x02\xc8\x6c\xcd\x49\x75\xef\x10\xe7\xfe\x44\x04\x6d\xeb\x14\x24\xbf\xfb\x44\xaf\xe3\x5f\x40\xf7\x10\x08\x86\xae\xe5\x49\x51\x61": [
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC1155.sol",
            name="ERC1155",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x02\xe5\xab\x94\x23\xe0\x90\xaa\x8d\x95\x7a\x95\xab\xcd\x69\x12\xe0\xd0\xe5\x44\x19\x8f\x85\x53\xe4\x24\x92\x1e\x05\x1d\xbc\x98": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x02\xf9\x95\xb1\x7f\xa7\x84\xa4\xf3\x43\xcb\x0a\xbd\x05\x16\xe5\x3e\x85\xd3\x48\xeb\x40\xc2\x8d\xd6\x87\x8b\x44\xb0\xd4\x69\x20": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/ProtocolFees.sol",
            name="ProtocolFees",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x03\x08\xbe\xda\x7c\x29\x3c\x02\x03\x95\x76\xba\x5b\x01\xdf\x11\x2b\x61\x0d\xed\x2f\x3a\x33\xf8\x29\x59\x72\xa7\xed\x9d\x04\x19": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x03\x0f\x90\xe5\xa6\x6b\x4b\x8d\x8c\x26\x8d\x79\x66\xa6\x66\xa5\xf4\xff\xc7\xeb\xe2\x5e\xad\x6f\xb7\x74\xb5\x97\x49\xdb\x86\x15": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x03\x10\xea\xba\x85\xc3\xd0\x16\x18\x97\xb4\xd8\xa8\x56\xa3\x95\x75\x63\x5b\xfd\x1a\x54\x58\x06\xcf\xc8\x41\x7f\x26\x6c\x0b\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x03\x34\xc8\x0a\x63\x78\x37\x8a\x64\x24\x8f\xfe\xd7\x14\x90\xd6\xbd\xe3\x42\xa7\xb9\xad\x00\x14\xc6\xe1\xab\xd3\x77\x0f\x10\x79": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.8.0
    b"\x03\xc4\x5a\x67\x2e\xa2\x34\x79\x8d\x87\x31\xc9\xcf\xba\xa0\xec\x9a\x2e\xc6\x6a\xc2\xa3\xbf\x1c\x43\x7c\x77\x41\x70\x1b\x3c\xd9": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x03\xcd\x64\x42\xc2\xf1\x47\x90\xae\x30\xd5\xa9\xf4\x56\x7d\x45\x1f\x0d\xcf\xa0\x3e\x6d\x37\x6e\x3e\xc8\x16\xbd\xc8\x4b\xaf\x5b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x03\xe8\xda\xa6\x3a\xd6\xb0\xfd\x04\xa6\x58\xca\xd7\xed\x3f\xec\x2b\xd7\x3b\x86\x00\x84\x06\x2b\x76\x52\x22\xe2\x2a\xa4\xf9\x29": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x04\x0b\x10\x35\x9c\xcf\xb0\xdf\xbe\xc9\x32\x71\x27\x26\xcc\x39\xcd\xd3\x6b\x9e\xb5\xf4\xa8\x95\x13\x8f\xa6\x17\x8d\x91\x7a\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.3.0
    b"\x04\x1b\xa0\xca\xf8\xe0\x78\xaa\x39\x6b\x21\x79\x71\xa6\x75\xf9\x40\xac\xde\xaa\x86\x97\x5d\xc1\x95\x0f\x40\x10\xe5\x95\xf8\x7e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\x04\x3c\x14\xf0\x86\xea\xd8\xd4\x12\x8e\xca\x96\x19\x45\x12\x36\x8f\x6f\xc0\xe9\xdd\x14\x81\x4d\xb1\x4f\xc7\xd5\x61\xf8\x81\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0-rc.0, v4.3.0-rc.0
    b"\x04\x40\x92\xe5\x1c\xe1\x3e\x5f\x66\x99\xc3\x00\x23\xea\x39\xfc\xd7\xcf\x9d\xf0\x21\xa9\x71\x3d\xa1\xc1\x6f\xb3\xb3\x5f\x1c\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.2, v3.4.2-solc-0.7
    b"\x04\x49\xee\x57\x85\xf2\xf4\x22\xd6\xc6\xbb\xb2\x43\x27\x82\xd3\xea\x00\x69\x61\x6d\xd3\x8d\xed\x5e\x1d\x94\xab\x60\x84\xfb\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x04\x94\x6f\x4d\xce\x6a\xe5\x4c\xe0\x60\x6a\xb7\x42\xc7\x8b\x3b\x3d\x19\x97\xdf\xc4\xa2\x5c\x78\x94\xc8\x42\x8d\x04\xab\x33\xed": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\x04\xac\xfc\x1e\xd5\xc1\x0d\x26\x95\x18\x24\x4c\x8f\xb5\x2d\x42\xda\xbd\xf3\xb3\xbe\x4a\xb1\x50\x66\xfe\x57\xf7\xa5\x56\x7e\x4e": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/UniswapInterfaceMulticall.sol",
            name="UniswapInterfaceMultical l",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/UniswapInterfaceMulticall.sol",
            name="UniswapInterfaceMultical l",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/UniswapInterfaceMulticall.sol",
            name="UniswapInterfaceMultical l",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x04\xe4\x76\x18\xd3\x6b\x9d\x68\x01\x3a\x9c\x40\xc5\x0c\x26\x19\x7e\x3b\x4c\x38\x34\x33\x58\x25\x39\xf1\xfd\xba\xd4\x1d\x25\x21": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x04\xfb\xe3\x46\x4b\x1c\xee\xf3\xcb\x03\xd1\xfd\xf5\xea\x1e\x83\xf9\xa7\x1b\x59\x1e\xf7\x49\x61\x80\xf3\xa0\x71\x16\x0b\xf7\xb2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.6.0-rc.0
    b"\x04\xfd\x9c\x86\x28\xdb\x3d\x1e\x61\x03\x23\x04\x6b\xe6\xb2\xd0\x51\xad\x17\x3f\xe2\xa7\xbf\x70\x58\x8f\x70\x90\xe3\x56\x8c\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133
    b"\x05\x02\xf0\x36\xbd\x91\xb9\xb6\xbe\x04\x78\x43\x49\x8d\x2a\xc0\x45\xd2\x94\xe8\x92\x03\x83\x26\xaa\x38\x39\xc0\x1c\xff\xbf\x87": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.133",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x05\x12\x8e\xbf\x82\x68\x7c\x03\x26\x04\xf0\x12\x77\xf5\x3f\xb7\x9a\x64\x8f\x79\x0a\x9c\xd9\xae\x69\xf3\x36\x32\x35\x7a\xdb\xd9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x05\x5b\x16\xcf\xb3\x27\x44\x1e\x85\x91\xaa\x65\xd2\x18\x5b\x2a\x72\x2a\xab\x0b\x3d\xef\xf7\xdd\x6c\x03\x63\x39\xcc\xc1\xcc\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.160
    b"\x05\x72\xe3\xa9\xdf\xc3\xef\xf5\x4a\x4e\x12\x9d\xc6\x09\x73\xd5\x7f\xa9\x58\x3a\xa9\x26\xb7\xa5\x75\x89\xc3\x14\x6e\x42\x3b\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.160",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x05\x86\x43\x54\x52\xfd\x2b\x3b\x1d\x99\xba\xae\xa1\x99\xbb\xc4\x54\x5a\xb5\xb3\x24\x84\x6a\x1d\x5a\x8e\xbe\xf9\xb3\xa6\x0d\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x05\xbb\x49\x43\x3f\xed\x3e\x37\x78\x11\x94\xc6\xe1\x72\xd0\x3c\xa7\xb0\xfd\x4b\xc1\xcc\x41\x14\x37\xc3\x64\xe7\x55\x10\x50\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x05\xdd\x4f\x6f\xf7\xc6\x63\x33\xb0\xa6\xa2\x17\xb6\x4f\xb5\x5d\xc9\x3e\x16\x5a\xe4\x2f\x1a\xdb\x6f\x5c\xef\x73\xd6\x59\x55\x89": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x05\xe6\xca\x6b\x4c\x2a\x08\x5c\xae\x5e\x6b\x0b\x4a\x09\x6f\x10\xd6\x0d\x92\x38\x8d\x30\x80\x67\x50\x69\x3d\xea\x7a\xeb\xe2\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2
    b"\x05\xef\xbb\xf6\x38\x7d\x63\x29\x24\x1c\x33\x1b\x2a\x38\xc6\xce\xf1\xa1\x72\xcb\x3d\x7c\x6e\x87\xf5\xba\xa7\xbb\x8e\xe5\x48\x00": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.2",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x06\x05\x43\x73\x9e\x77\xd4\x8a\xd2\x89\xe5\x55\x11\xf3\x27\xa3\x23\x22\x97\x27\xb6\xdf\xfb\x95\x6a\xbe\xe9\xb5\xc4\x0b\x00\x5b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x06\x35\x84\x22\x4e\xc6\x5f\x71\x34\xc2\x41\xdb\x49\x78\x54\xa8\x73\x52\x61\xdd\x1d\xfb\x3f\x2e\x3a\x60\xb2\x8e\xcf\x40\x0a\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x06\x4e\x70\x91\x09\x62\x47\x6f\x9a\xb5\x2b\x6e\x8a\xcc\xb7\x2f\x79\xb9\x38\xa1\xe9\xd8\x2e\x6e\x7d\x74\x75\x76\xcb\x16\x2b\x6c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x06\x57\xf9\x91\x80\x0d\x9d\xd1\x6d\xcf\x56\x64\x43\x92\x96\x54\xa2\x38\xd6\x4a\x1b\x8b\xc7\xd8\x2b\x19\x0b\x7a\x08\xf9\x8c\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x06\x60\x29\xcb\x2d\x15\xa7\xad\xee\x95\xb4\x63\xf3\x54\x9c\x90\x5b\x01\x81\xc6\xc2\xb2\xf6\x65\x3d\xcf\x29\xbe\x6c\x74\x71\xd4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2
    b"\x06\x73\x93\xe9\xc3\x8a\xd4\x1a\x69\x05\x2a\x38\xb2\xe2\x10\x30\x76\x6f\xb8\x3d\xa9\x8d\xa2\x57\xd5\xc1\x22\xa2\x69\xb2\x39\xbe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.2",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x06\x76\xbe\x7d\xa5\xdd\x84\x2f\x7b\xf9\x00\x68\x7f\xd8\x53\xed\x6d\x57\x0f\xec\x39\x76\x19\x70\x67\x8f\x5b\xa7\xaf\x8f\xf6\x25": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x06\x78\x31\xd3\xfc\x96\x98\x83\x97\x45\x02\xf9\xcb\x98\xfa\x77\x4c\x21\xd2\x02\xb1\x38\xca\xcd\x95\x59\xad\xb1\xb0\x18\x9b\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226
    b"\x06\x98\xaf\x0e\xe8\x7c\x62\x3e\x51\xa5\xb5\xd3\xe3\x03\xb3\x92\x2f\x4f\xfe\x19\xfc\x99\x69\x78\xec\x9a\x07\xbb\x95\xa8\xdd\x76": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.226",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x06\x9e\xee\x1e\x1e\x7f\xe4\x8e\xee\x55\x56\x64\x4c\x5f\xef\xf3\x98\x37\x53\x66\xc2\x7e\xdb\x82\x20\x7f\xd5\xd8\x36\xce\xd1\x4e": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x06\xa4\x31\xd1\x3a\xc6\x6d\x07\x14\xbf\xa2\xf9\xa2\xc3\xd5\x03\x88\x60\x20\x2d\xf1\x06\x01\x5a\x0a\x99\x9b\xb2\x2e\xae\xdf\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x07\x08\x70\xf7\x36\x55\x97\x07\x03\xd5\x49\x0a\x99\x25\x8e\x5e\x73\x6c\xa6\x61\xd6\x3c\x5e\x18\xf5\x70\xc5\x33\x8c\x91\x57\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x07\x31\x1b\x9d\x29\x65\xd4\x72\x8e\xce\x63\xde\xba\xa5\x74\x89\x3f\x53\x75\x19\x82\xe2\x69\xc5\xe2\x57\x5d\xa4\x2e\xb5\xe7\x8c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x07\x32\x3c\xa1\xcb\x3d\x12\xa6\x6f\x27\x48\x19\x73\x91\x7a\x93\x7f\x3d\x6f\x7a\x69\x5e\x3b\xb1\x05\x55\x73\xf0\xdb\xa2\x9f\xd4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x07\x75\xf0\xcf\xa3\x46\x3e\xf2\xe9\x53\x49\xff\x73\x02\x79\x4d\x0e\x9b\x70\x62\xa5\x44\xee\x89\xbd\x9c\xfd\x7c\x6e\xb6\x24\x32": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrowUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x07\xd2\x68\x74\x08\x6e\x91\x78\x95\x57\x7a\xfa\xf4\x04\xf7\x13\x88\xd2\xe3\xdb\x4b\xac\x67\x45\xd6\xa0\xac\x3b\xff\xea\x67\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x07\xeb\x7e\x54\x68\x9d\x87\x39\xe2\x3f\xcb\xce\x7c\x3b\xad\xd9\x36\xed\xd9\x20\xaf\x31\x48\x12\x78\x83\xa8\x72\x94\x86\xeb\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x08\x0e\xc2\x51\x02\xb5\xd2\x60\x41\xd1\xd5\x79\x08\x14\x48\x61\xea\x6f\xb4\x12\x26\x86\x29\xea\x2e\xba\xd0\x26\xfd\x34\x8b\x48": [
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x08\x3d\x88\x2e\x36\x26\x12\xab\x4a\x9d\xf5\xdd\xcf\xf3\x05\xd4\x89\x47\x52\x9c\xf6\x61\x63\xa2\xad\x99\x19\x21\x99\xf1\x8d\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x08\x48\xdd\x7c\xa8\x21\xc8\x89\x1c\x2e\x4d\x1e\x8b\x98\x91\x74\xb7\xb6\x4c\x8c\x7e\x3e\xdd\x7a\xc4\x10\x15\xd2\x19\x6a\x1d\xb6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x08\x59\x43\xec\x32\x3a\x5d\xa8\x80\x42\xba\x1b\x48\x14\x45\xbc\x14\x98\x22\x4c\xc0\x8c\xe8\xf7\x28\x45\xad\x05\xa3\x61\x5b\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2
    b"\x08\x59\xdb\x79\xfc\x1d\xab\x1d\x1b\x4e\x9e\xef\xad\x9a\xc8\x74\x17\x3b\xe0\xa1\x5c\xdb\x06\x7f\xa3\xff\x8b\x07\x76\x0b\x64\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.7.2",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x08\x7f\x88\xea\xb3\xdc\x6d\x2a\xea\xab\x8a\xe0\x73\x2e\x9b\x33\x62\x38\xea\x37\xcb\x87\x16\x6c\x1f\xcd\x9d\xb9\x5d\xab\x7e\xde": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x08\xa5\x64\x7f\x14\x98\x8a\x11\x4f\x4c\x1f\x13\x03\x41\x23\x08\x12\xad\xd0\xde\x60\x59\x79\x3f\xd2\x20\x3b\x9d\x1e\xd9\xbb\x0f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Metadata.sol",
            name="ERC721Metadata",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Metadata.sol",
            name="ERC721Metadata",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x08\xaf\xae\x58\xb7\xf9\xc1\x67\x80\x71\x1c\x17\x24\x37\xfc\xdf\xcc\x21\x24\x68\xfb\x02\x2c\xfc\x2e\x75\x92\xb3\xad\xb2\xd7\xb7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.16
    b"\x08\xb6\xd1\xdc\x20\x67\x69\xf2\x3f\x41\x36\x13\xdd\x69\x74\x8f\x3e\x8f\xbb\xcd\x5d\x6b\x1d\x5d\x6c\xcf\x49\x38\xc3\x5e\x86\xc1": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.16",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.9
    b"\x08\xba\x85\x9e\x3e\xfa\xd1\xbc\x46\xb7\x68\x72\x10\x0a\x4c\x6e\x6a\xa2\x27\xd7\x07\x91\x92\x5f\x36\x50\xf2\x1f\x73\x83\x17\xce": [
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x08\xbb\xa2\x0a\x40\xd3\x72\xa5\xe6\x0c\x23\x95\xe8\x36\xa3\xe5\xca\x17\xdc\x01\x79\x57\xf6\x91\x99\x25\x17\x02\x4c\x23\x15\xba": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x08\xdd\x96\xd7\xcf\x3b\xc3\x1c\x6d\x3f\x2d\x7f\xe1\x67\x6b\x08\xb8\x6d\xdf\xb1\xb3\x8b\x75\xb7\x37\x16\xcb\x04\xf5\xd1\xde\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207
    b"\x08\xdf\xe7\x3e\x29\x9b\xf0\x55\xe8\xdc\x7a\x64\x38\x6b\x6f\x84\x29\xa7\x83\xc9\x75\x02\x44\xea\x52\xff\x08\xb3\xf1\x72\xa6\x58": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.207",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x09\x0b\x03\xc0\x9e\x2e\xb5\x53\x60\x52\x90\x94\x6b\xcd\x20\x17\x4d\xb0\xbf\x78\x6b\xd4\x66\x31\x43\x53\x2b\xe5\x47\x59\xd3\x0b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.1.1
    b"\x09\x6d\x2a\xfa\xfc\x6d\x28\xac\x0f\xc4\xe2\xfc\x04\xfd\x81\x5a\x66\x8b\xb3\xa8\x93\x73\xae\xff\xb2\x4c\x06\x6b\xe7\xb8\x91\x26": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/OracleLibrary.sol",
            name="OracleLibrary",
            version="v1.1.1",
        ),
    ],
    # Hash appears in versions: v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286
    b"\x09\x93\x56\xa7\x35\x6b\xaa\x07\x3c\x8e\x18\xba\x9c\xa5\x66\x2c\xcb\xce\x2d\xf6\x55\xf0\x7f\x9f\xbd\x0a\x31\x2f\xe9\x72\x67\x4d": [
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.286",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x09\xa9\xe3\x7e\xfa\x78\xce\x98\x45\x29\x4e\xfd\xa9\x6b\xa0\xd5\xc4\xb0\xd3\x35\xed\x9c\xfc\x10\x54\x6d\x2c\xbc\xa3\x06\x24\x6b": [
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x09\xb0\x42\xf2\x0f\xd9\x45\xd7\xc0\x98\x32\x6c\x03\x29\x87\x02\xb9\x32\x32\x26\xe8\x1e\x19\x64\xc0\x5b\xf3\x6b\x64\x16\x72\xd7": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/callback/IUnlockCallback.sol",
            name="IUnlockCallback",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0
    b"\x09\xb7\xa6\x54\x1a\xe1\x84\x50\x1c\x7a\x1f\xec\xba\x39\xec\xca\x4a\x30\x4f\x3a\x89\x30\x63\x0f\x37\x73\x59\x72\xd1\x50\xdb\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x09\xbc\xd2\x5a\x04\xa6\x2e\x7e\x83\x21\x29\xd9\x12\x2a\xbe\xce\xe0\x71\x30\x63\x69\x40\x1e\xb5\xd1\x15\xf6\xa6\x38\x3e\xa4\x21": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x09\xcb\x6d\x48\xb4\xa3\xaf\x02\x6c\x0c\xab\xc8\xb3\xa4\x39\x9d\x81\xb3\x94\xb8\x8e\xa5\x62\x48\xfe\xee\x8c\x70\x5b\xf4\x39\xd0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x09\xd1\xfc\x72\x85\x63\x96\x4a\x25\x82\x63\x9f\x34\xa2\x97\x0f\x00\xce\x6b\x00\xb5\x99\xf5\xee\xa3\xf3\x01\x73\x2c\x9b\xc9\xf5": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/FullMath.sol",
            name="FullMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\x09\xd5\x53\xa4\xe9\xdb\xae\x3c\xc8\x4d\xe4\x5c\xf6\xcc\x99\x46\x6a\xf3\xa7\x89\x82\x5c\xf4\xd8\xba\x96\x24\xe1\xeb\x05\x62\x4e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x09\xd8\xe5\x79\xbe\xbf\x75\x34\xab\x8f\x8e\x26\x45\x08\xd3\x5f\x57\x3f\x96\x66\x02\xc6\xac\x98\x0b\x5d\xc9\xe6\x49\xdb\x4b\x38": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC1046/ERC20Metadata.sol",
            name="ERC20Metadata",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC1046/ERC20Metadata.sol",
            name="ERC20Metadata",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x0a\x03\x34\x48\x26\xd7\x90\x90\x6d\x5d\x4d\xd9\x49\xfd\xaa\xb4\xd0\x96\x8f\x74\x1f\x34\x09\xfc\x0a\x30\xfd\xd6\xa3\x4c\x33\x23": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x0a\x0b\x3e\x77\x05\x15\xeb\x75\xe7\x43\x30\xdd\x2d\x28\x97\xd5\x95\x36\xfe\xf3\xec\xf4\x1d\x6b\x45\x94\x7a\x8d\xd6\x58\xff\xf7": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/Lock.sol",
            name="Lock",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0a\x12\xbd\x20\x2a\x8d\xb1\xb6\xc5\x93\x77\x57\x53\x45\x2f\x6a\x36\x46\x7c\xb9\x20\x9a\xf6\x8d\x2e\xad\x17\x6a\x45\xf9\x16\x27": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x0a\x15\x46\x3a\x7d\xac\x11\x77\x78\x92\xbb\x2b\xc0\x2e\x43\x5d\x7a\x50\xdf\x21\xa1\xf0\xad\x57\x40\xba\x7e\x67\x36\xb7\x7e\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177
    b"\x0a\x23\x37\x87\xaf\xf2\x39\xfd\xea\xb6\xac\x56\xb0\x49\xa1\x65\x02\xb1\x81\x45\xdd\x72\x55\x54\xb4\x9f\x67\x30\x92\xf9\xb6\x12": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.177",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x0a\x30\xb1\x67\xe5\xc6\x92\xd7\x6d\x8a\xc4\x86\xc9\xbf\x4f\xbe\xea\xd6\xb0\xf6\x3b\x4b\x08\x37\x88\x16\x9b\xff\x9f\xff\x7c\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0a\x3a\x45\x5a\xd6\x7c\x88\xc5\x18\x18\x01\x1a\x8c\x05\x7a\x81\xc5\xd3\x2c\xf5\x2a\xc1\xfc\x6e\x75\xad\x50\xb9\x5f\x8c\x59\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x0a\x40\x31\x0a\xba\x2f\xf6\x05\x11\xb1\xc0\xb7\x59\x52\xc6\x78\x95\x56\xe0\x43\xbf\xe4\xfe\xc9\x3c\x3b\xfd\xd7\x0c\xfc\x84\xef": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x0a\x45\xd1\x6f\x1e\x15\xcd\x1a\xf0\xbe\x2a\x22\x5c\xe3\x23\x1b\xa3\x9f\xe7\x4b\x4b\x3e\xb6\x32\x37\xd0\x33\x42\x16\xef\xa9\xcd": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/LiquidityMath.sol",
            name="LiquidityMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x0a\x94\x60\xe9\xa6\x3d\x74\x25\xf6\xb6\xe5\x03\xcd\xe1\x50\x37\x94\x93\xb9\xd3\x2e\x6c\x60\xfa\x33\x8a\x66\x63\x9c\xca\x7b\xb8": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.230, v0.0.231, v0.0.232, v0.0.233
    b"\x0a\xa9\xd6\x30\x87\xd1\xbd\x05\x68\xfa\x38\x36\x32\xbe\x5a\xcf\x51\x57\x25\x50\xc6\xcb\xd2\x43\x0d\xe4\xf0\x48\xa4\x16\x21\x21": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.233",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x0a\xbf\xc1\x62\xbc\xfe\x22\x4c\x9a\x4c\x56\x1c\x23\x66\xde\xba\xa6\x70\x40\x8e\xfd\x0b\x68\xf3\x39\xb7\x05\xe4\x2d\x71\x08\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0a\xdd\x40\xe4\x3b\xf7\x2e\xff\x05\xee\x4b\x71\x68\xba\x45\x81\x33\xb9\x87\x11\xb5\x5c\x90\x7f\x6f\x8b\x47\x7f\x38\x40\xf5\xe2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0b\x0c\xab\xc1\x44\x6f\x18\xad\x48\x96\x5e\x34\x84\x88\x17\xb3\xc1\x01\x58\x7c\xda\xfe\x9a\x38\x4e\x1e\x81\x4c\xc7\x43\xa9\xc0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66
    b"\x0b\x3e\xa5\xbe\x58\xa2\xbe\xbd\x81\x3b\x8e\x73\xd8\xc0\xd3\x35\x07\xa6\xae\xd5\x74\x7a\x08\xa5\x92\x5a\x5f\xb7\x8f\x4a\x7a\x6c": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.55",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.66",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x0b\x63\x8e\xbe\xe1\xbe\xa5\xe2\x8c\xe1\x82\x59\x87\xa9\x7d\x7c\x8c\xdb\x5e\x90\x2d\x5e\xef\x1c\x56\x12\x6b\xb0\xba\x07\xfa\xf1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x0b\xa1\xd4\xea\xf2\xd5\xb2\xfc\x03\x35\x92\x00\x0c\x31\x0f\x24\x1f\xdd\x38\x65\x19\x7b\x79\x14\xcc\xf1\xb0\x3f\x98\x13\x06\x4b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.5, v0.0.6, v0.0.7, v0.0.8, v0.0.9
    b"\x0b\xa8\x1a\x98\x48\xb3\xce\x80\x2e\xd4\xb5\x88\xe4\x94\xc9\x10\x0b\xef\x5f\x32\x9d\xe0\xab\x64\x16\x86\x4d\xe5\x8b\x4e\x66\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v1.3.0
    b"\x0b\xb4\x05\x78\xc8\x21\x82\xc7\xf6\x4d\x5e\x89\x7b\xca\xbf\x73\x69\x80\x59\x6e\x49\x73\x22\x68\xb1\x0f\xf3\x78\x2d\x68\x54\xe3": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/examples/PairFlash.sol",
            name="PairFlash",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0
    b"\x0b\xcb\xee\xab\x00\xf1\x2d\x65\xc7\xd4\x8e\xeb\xd5\x15\x08\x55\x19\xee\xda\xc7\xd2\xa9\x5d\x12\x4e\x25\xd3\x6b\x52\x1c\x1a\xd1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x0b\xcc\x1b\xd8\x86\x05\xf7\xac\x0e\x12\xb0\x8c\x57\x3e\xb8\x4b\x72\xd2\xfb\x8c\xc9\x5f\xb7\x29\xd9\xa9\xd1\x36\x20\xf6\xe4\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x0b\xd1\xf6\x2f\x89\xc7\xa1\xaa\xc6\xa9\xe2\xa2\x12\xd9\xbd\x2b\x42\x0a\x69\x68\x3e\xbe\x55\x3d\x13\x2f\x6c\x98\x84\x3f\xdc\x16": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0b\xe2\xa7\x5d\x79\xd3\x83\x8e\x97\xd0\x45\x18\x7f\xdd\x6a\xf7\x87\xac\x47\x57\xed\xa4\x98\x84\x17\x97\x22\x1f\x59\x22\x7d\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x0b\xe6\xc1\xcb\xc1\xa3\xaf\xf5\x49\x6e\x44\x95\x81\x88\x73\x36\x19\xa4\xa0\xc0\x36\xbf\x9e\xb8\x08\x87\x3a\xf9\xa2\x07\xbc\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.151
    b"\x0b\xf5\x4d\xe4\xf8\x24\x15\xcb\xd7\x37\x50\x51\x26\x17\x35\xb1\x06\xba\x44\x82\x21\x11\xbf\x75\x09\x14\x2d\x5d\x8b\x01\x35\x82": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.151",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0c\x17\xf4\xf6\xa4\xc7\x86\x69\x04\x67\x22\xe9\x47\xe7\x91\x86\x61\x77\xd7\x3c\xe5\x65\xf9\x09\xf6\x2e\xf8\xf4\xb7\x12\x3d\xd7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/CrossChainEnabled.sol",
            name="CrossChainEnabled",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0c\x23\x88\x9e\xfd\x19\x0a\x5c\xde\x5c\xf7\x7a\x17\xf1\x85\x90\x3b\x51\x08\x25\xa4\x78\x20\x4d\xa6\x41\x70\xdd\x5c\x8a\x26\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x0c\x24\xe9\x34\xb5\x25\x65\x2d\xf4\x03\x00\xe7\x24\xf3\xb9\x3f\xd8\x8b\x71\xf1\xff\xe0\x25\xcd\x65\x32\xbc\xfe\xd5\xb7\x98\x7c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x0c\x2e\xda\x40\xaf\x16\x3b\x1a\xa6\xe6\x0e\x59\x01\xe4\x8a\xb3\x1f\xf5\x2e\xfd\xd9\xaa\x1b\x6b\x10\x9b\x64\xff\xb5\x9f\x1b\x13": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAuthority.sol",
            name="IAuthority",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x0c\x2f\x4a\x41\x23\xb3\x04\xea\xe3\xb3\xda\x95\x3c\xa2\x59\x50\x88\xfc\x37\xab\x03\xaf\xd0\x53\x0a\x3d\xb4\xf8\xd6\x0d\xcb\x72": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x0c\x3a\x09\xba\xc9\xe9\xc2\x72\xbf\xa1\xc9\xb8\x83\x16\x13\xd7\x6e\x73\xd8\xa7\x63\x33\x0e\xb5\xd7\x80\x8f\x3b\x96\x0d\xef\xfb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x0c\x42\x44\xb7\xe1\x71\x7c\x05\x88\xc8\x17\x88\xde\xe7\x68\x71\x5e\x81\xf4\x50\x7a\xed\x9b\x62\x66\xb1\xc0\x85\xed\x5d\x31\x58": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/LiquidityMath.sol",
            name="LiquidityMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x0c\x4a\x70\x38\xeb\x0b\xa2\xa8\x9c\xfa\x03\x0a\xc4\xe8\xe6\x4c\xfa\xbb\x98\x7a\xb6\x96\x89\x1c\x4e\x54\xbb\xea\x80\x0b\xf3\x96": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x0c\x4b\x12\x51\x47\x0e\xe0\x33\x64\xe0\xf4\xe5\x3d\x32\x5d\x66\x36\xb8\xe6\x4a\x7a\x5b\xb5\x65\x0a\xa4\x1b\x5b\xde\xfc\xb4\xc1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43
    b"\x0c\xc6\xf6\xfa\xa5\xaf\x6a\xaf\x1b\xf5\x92\x52\x90\x46\x74\x15\xee\x11\x1f\xbd\x05\xaf\x21\x19\x6f\xb2\x53\x66\x0b\x26\xa2\x87": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.43",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0d\x00\x5a\x9f\x45\x3c\x55\x1f\x47\x49\x2a\x8b\x66\x53\x97\xc8\x85\x0e\x0d\xf6\xe1\x30\x02\x29\xd2\xdd\x74\x69\xc8\x3e\xe7\x16": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.3.0, v3.4.0, v3.4.2, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x0d\x21\x8f\x55\xa8\x60\xb4\x78\xb5\xbe\x15\x87\x71\x3d\xb1\x9e\xee\x90\x42\x03\xaa\x29\xaa\x12\x4c\xba\xcb\xb0\x9b\xad\x4c\x2d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181
    b"\x0d\x27\xd6\x0c\x0a\xe6\xfe\x0b\x06\xa5\xe6\xd5\xba\x58\x85\x05\x23\xe5\xfc\x4d\xb7\xce\x53\xe6\xdd\xad\x21\xc1\xfc\xbe\x60\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.181",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x0d\x2e\x69\x22\x9a\x97\xf9\x72\xeb\x5d\x6e\xcd\x55\xfa\x38\xbc\x50\x8e\xc4\x88\x46\x6b\x3a\x52\x34\x3b\x93\x47\x58\x58\x66\xc0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0d\x4c\xeb\xc1\x3b\x39\x5b\x6e\x15\x6c\x0d\x78\xfd\x39\xfa\x53\x64\x7d\xf2\xfd\x2a\x84\x7d\x60\xb6\x5c\xeb\x7f\x1e\xf1\x3a\x11": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x0d\x55\x68\xb1\x2b\x76\x53\x72\xca\x1a\x89\xf8\x7d\xf1\xe1\x3a\x0b\x79\xf7\xf8\x65\x58\x3c\x9b\xe6\x9c\x27\x94\x0a\x9f\x16\x99": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0d\x5b\xd6\xa3\x6a\xc5\x17\xe3\x00\xee\xfd\x19\x6d\x93\x1e\xdf\x58\xb2\x69\xfb\x7e\xfc\xaa\xa5\x1b\x85\x98\x57\xc8\x19\x90\xa6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x0d\x68\x92\xec\xd0\x34\xd8\x36\x39\xbd\x97\x71\x9a\xa3\xa5\xad\x0f\xdf\x18\xd3\xe6\x09\x79\x71\xef\xc1\xdd\xb7\xff\xaf\x7b\x53": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x0d\x90\x47\x49\x6a\x36\x00\x27\xd5\x5d\x3f\x80\x6f\xa7\xa1\x04\x9c\x0a\xc9\x16\xaa\x03\x3f\xc1\x11\x48\xa1\xa2\x6b\xb4\xf9\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0d\x92\x2b\x6c\x75\x60\x60\x9f\x7b\x24\x64\x69\x42\x6a\xd3\x73\x75\xf4\xfc\x13\x31\x53\x8f\x85\x00\xf0\x53\x38\xd1\xe9\xfd\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154
    b"\x0d\xa2\xea\x1c\xbb\xbd\xdf\x62\x2f\xa1\x58\x60\x25\x3a\x01\x82\x0e\x77\x4d\x9a\xb6\x53\x37\x90\xe2\xed\x6c\xe5\x27\x66\xc7\xec": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.154",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x0d\xc3\xc8\xdd\x9e\x61\xe0\x9c\xfd\x3a\xd7\x4d\x5a\xa5\x3e\x0e\xe4\x4f\x60\xa1\xbd\x13\xff\xb5\xb1\x32\x8d\x2e\x41\x75\x2f\xc0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\x0d\xc5\xc5\x5d\xea\x59\xa7\xd4\x22\xee\x84\x4b\x81\xd0\xaf\x35\x40\x63\xa1\xcf\xd6\xfa\xca\x94\x50\x8f\x25\xc1\xa5\xe2\x86\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/deploy-ready/ERC20MinterPauser.sol",
            name="ERC20MinterPauser",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/deploy-ready/ERC20MinterPauser.sol",
            name="ERC20MinterPauser",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.25
    b"\x0d\xc8\x63\x18\xd0\xb7\x0a\xee\xbf\xcb\x5a\x8c\x87\x18\x74\x2e\xcd\x53\x5b\xa0\xa4\x38\xa8\xc5\xfb\x88\x0e\xf7\xb2\xc7\x6e\x85": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.25",
        ),
    ],
    # Hash appears in versions: v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86
    b"\x0d\xf4\x5a\x3a\x8e\x8b\x4a\x7e\xe5\x8c\xbe\x63\x88\xc3\xd2\xe8\x27\x4b\x6b\x13\xd3\xbd\x6b\x0f\xfd\xef\x33\x6f\x0e\x1c\xf6\x7b": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.86",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x0e\x05\xd6\x9d\xf4\x95\xad\x8a\xd0\x7d\xdc\x27\x83\x75\xf6\xe1\x5b\x1a\x79\x02\x16\x76\x3a\x7f\xf3\x3b\xd2\x93\xab\x94\x0a\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesExtended.sol",
            name="VotesExtended",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesExtended.sol",
            name="VotesExtended",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesExtended.sol",
            name="VotesExtended",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesExtended.sol",
            name="VotesExtended",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesExtended.sol",
            name="VotesExtended",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0e\x18\xe2\x4c\x92\x71\x61\x45\x22\x63\x22\xb3\x11\xa4\x17\xfd\x67\x2f\x7c\x91\x00\xb2\x92\xea\xd9\x5e\xed\xe5\x2a\x4f\xaf\xd5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0e\x1a\x7e\xf7\x2c\x47\x6d\x48\xc4\xfc\x3a\x95\xaf\x4a\x5a\x62\xce\xad\x3c\x74\x82\xfb\xe2\xb3\x49\xe7\xe7\x43\x91\x58\x7c\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x0e\x27\xac\x71\xe9\xfc\xcf\xf2\x8c\xaf\x99\xba\x5c\x3a\xb7\x7c\xea\x9d\x7e\x2f\xb3\xa7\xcb\x19\xc0\x86\xb6\x0c\x6c\xeb\xe7\x96": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2
    b"\x0e\x8b\xe2\xc8\xc7\xa6\xfe\xa0\x5f\xf9\xdd\x25\x87\x57\xce\x69\xae\x6a\x95\xe3\x83\x12\x6d\xff\x97\x41\x44\x80\x95\x41\xe6\xb8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x0e\x99\x02\x2a\xf5\xe9\xe0\x83\x65\xca\x9f\x64\xa9\xf1\x4e\x31\xa9\x59\x94\xdd\x3d\x83\x7c\xe1\x32\x57\x9f\x02\xed\x35\x74\x4a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x0e\xa4\x39\x58\x1c\x41\x80\xcb\x89\xab\x0b\x09\xcc\xc0\x3a\x8c\x0a\xcb\x7f\xef\x39\x30\xd3\x91\xe5\xf4\xcf\x05\xfb\xc2\xe1\xc6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x0f\x12\xf6\xdf\x91\xe1\xff\x91\x01\x51\x1a\x8e\x61\x5e\x77\x2a\x3a\x1a\x12\xac\xfd\xbb\x07\xf6\x97\x75\x2c\xf1\xdf\x55\x4f\x1a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlCrossChain.sol",
            name="AccessControlCrossChain",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x0f\x22\xeb\x73\xbe\x24\x01\x42\x08\x51\x7d\x53\xda\xa1\xd3\xeb\xc9\xff\xa6\x5e\x9f\xe6\xdf\xa4\x12\x44\x5f\x2a\x93\x32\xae\xed": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x0f\x2b\xd5\x46\xd6\xa5\x98\x5c\xe9\x51\x05\xd3\x77\x6c\x81\xbf\xe6\x6b\x93\x03\x55\x52\x46\xdc\xcc\x4e\xc1\xb2\x6f\xe1\xab\x85": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x0f\x3d\x93\x4a\xea\x66\x59\x40\xad\x21\xc7\xce\x44\x2f\xde\x4b\x54\x3e\xca\xbd\xc0\x21\xc0\x0b\xb0\x39\x48\xed\xd9\x17\x26\x56": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x0f\x40\xbe\xe4\xe1\x25\x53\x22\x5d\xe8\xb7\x78\x59\x1d\xd0\x89\xd7\x85\xa0\x15\x5d\x49\x28\x8d\x20\x4e\x9e\x4f\xa4\xa5\x2a\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x0f\x98\x4e\xbb\x92\x2f\xa4\x46\x86\x2e\x3f\xe8\xe1\xd1\x27\xbe\x8c\xa0\x84\xfa\xf0\x41\x90\x0a\x69\xc7\xb5\x11\x6f\x86\xeb\xa4": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/TickBitmap.sol",
            name="TickBitmap",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x0f\xa3\x32\x9d\x23\xac\x12\x60\x2c\x2d\xd6\x06\x7d\x32\x0e\xf3\x8d\x13\x33\x0c\xba\xce\x08\x9b\xf3\x95\xa8\x10\x83\xb3\x49\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x0f\xbb\xb4\x31\xdf\xa0\x7e\x48\x50\x98\xd1\x1c\x2a\xdb\xab\x2d\x76\x83\x68\xbb\x53\xa2\xb5\x2d\x8d\xbb\x1d\x5b\x45\xeb\x9f\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x0f\xbf\x8f\x58\x8d\x67\x02\xae\x1c\xa9\x18\xeb\xed\xe5\xa9\x14\x26\x25\xee\x88\xfe\xea\x94\x72\xe3\x1e\x21\x57\xfe\xe9\x4e\x2f": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/UniswapV3Pool.sol",
            name="UniswapV3Pool",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x0f\xde\x4f\x80\x72\xbf\x60\x73\x4a\x77\x2d\xa6\x25\x0b\x74\xf0\xfa\x65\xf3\xb3\x2f\x62\x04\x81\x49\xf9\x4f\x92\xac\x01\x70\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x0f\xf5\x50\xf7\x4b\x13\x8b\x8d\x8e\x82\x99\x53\x2e\x0d\x28\x6a\x6b\x18\xf5\x78\x1b\x88\xff\x21\x5f\x2d\x9a\xdb\x3a\x7d\xa0\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.17, v0.0.18, v0.0.19
    b"\x10\x07\xcc\xa5\xfa\xb6\x48\x09\x40\x4d\x0f\x0a\x94\x69\x2f\x7b\xf3\x47\xb3\xff\x57\x20\x96\x88\x04\xa2\x56\x50\x50\xcc\xb9\xbc": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x10\x09\xcf\xe7\x5c\xae\x46\x7e\x0b\x04\xa5\x14\xf5\x5a\xa3\x94\xe4\xfd\xd8\xf0\x12\xe6\x0c\xf8\xce\x32\x56\xd6\x2a\xb4\xe7\x34": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControlUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.1.0
    b"\x10\x41\xca\xe6\xb8\xe5\xcb\xed\x8d\x4c\xf5\xbd\xee\x1c\x98\x78\xfb\x0d\x6a\x0b\x29\xc8\xec\xf9\xd4\x89\xdb\x9b\xab\xac\x42\x2a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.1.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x10\x49\xc3\x45\x10\x3f\x1f\xab\x14\x50\x5a\xa7\x41\xcf\x02\x90\xe5\x10\x6a\xa7\x2d\x6e\xd8\x8a\xf8\x11\x24\x88\xb8\x45\x20\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x10\x4a\xba\x65\x58\x37\xa1\xc3\x6b\x10\xaa\xfc\x25\xbe\x1c\xac\x4f\x47\xbd\x08\x13\x70\xc7\xca\x06\x4e\x11\x95\x1a\x85\x88\xe3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2
    b"\x10\x4b\x20\x27\x1a\x05\xad\x36\x90\xaf\x85\x24\x38\x7f\xe8\x9f\x22\xad\x3d\x94\x3e\x2e\x75\xae\x99\x4c\x59\xea\x25\xb2\x9e\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x10\x66\x43\x5b\x12\xe1\xa0\x09\xd4\x91\xd8\xe3\x93\xb7\xec\x6a\x0d\x81\x29\x6f\x7c\x23\xf9\x98\xde\x78\x59\x50\xd3\x17\xf5\x0e": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/NoDelegateCall.sol",
            name="NoDelegateCall",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x10\xbd\x4e\x2a\x41\xab\x87\xa1\x04\x8d\x08\x4f\xd5\x86\xeb\x61\x29\x03\x08\x3d\xfe\xb0\xa9\xcd\x30\x3c\xd3\x47\x14\x6b\xae\x96": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x10\xc3\xd3\x46\x71\x93\x10\xb6\x67\xd4\x97\x05\x2a\xca\x31\xe4\xfd\xe1\x9f\x84\xa5\xdf\x70\xf2\xa6\xdb\x70\xa8\x3b\x6a\x09\xd9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x10\xc9\xad\xd5\x68\x72\xbb\xf1\x04\x87\x58\x1c\xea\x48\xde\x0d\xd9\x93\x05\x33\x51\x69\xb4\xc9\x9f\x9e\xb5\x22\x18\xc4\x25\x6b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x10\xdc\x98\x79\x11\x18\xc2\x69\xa4\x25\x66\x87\xc1\x05\xd3\xac\x5c\x50\xa9\x48\xd0\x5c\xe3\xe4\x34\xe6\x69\x09\xd6\x6e\xac\x97": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59
    b"\x11\x40\xa4\x05\x9b\xa6\xcb\xf2\xcf\x7b\xe5\x9b\x7c\x3e\x02\xb3\xc9\x05\x37\xa5\x63\xe6\xdf\x08\x70\x65\x9d\x78\x95\xe6\x89\xeb": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.59",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2
    b"\x11\x63\xab\x1d\xf8\xfa\xf5\xcc\x2c\x52\xcc\x56\x6f\x3e\xc9\xb6\x7d\x64\xb9\x70\x84\xef\xdf\xe8\xd6\xd2\x19\x75\x07\x75\xdd\x75": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x11\x81\xb8\x66\xd6\xbd\x7a\x85\x2e\x63\x50\x6f\x91\xd7\xa6\x2b\x20\xa1\xf7\xcc\x5f\x3c\xe4\xf4\x37\x6b\xc6\xc6\x0e\xfd\xf8\xc4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x11\xa6\xed\xf2\xa6\x9e\x03\xa6\xd8\xd0\xca\xd7\x7e\x46\xd2\xb7\x06\x85\x1c\xf6\x38\x31\xbe\x46\xd2\xc6\x3f\x93\x72\xc7\xf8\x2c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199
    b"\x11\xae\xf6\xd3\x76\x76\x3d\x06\x11\x04\x21\x9a\xe8\x2a\x55\x79\x83\x44\xe2\xd6\xce\x70\x1a\x68\x31\x39\xba\x35\x34\xf3\xf9\x8b": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.199",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x11\xfa\x9c\x23\xe0\x57\xc8\xa1\xb1\x36\xcb\x4e\xa4\x61\x07\x61\xc8\xd4\x21\x1e\xa6\xa7\x87\xb4\xa2\x6c\xb1\x74\x48\xb3\x53\xd1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x12\x1b\xe1\x48\x1d\x2e\xf9\x3d\x4e\x8e\xd1\x98\xdc\xbf\x58\xb5\x9f\x14\xa1\xd4\xf2\xe9\x59\x5f\xbe\xa3\xec\x78\x8a\x60\xbb\xc4": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungibleTokenPositionDescriptor.sol",
            name="INonfung ibleTokenPositionDescriptor",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x12\x43\xdc\x7a\x06\xea\xd3\x03\x36\x21\x5b\xcf\x72\x8f\x12\xc7\xd3\xfd\x31\x79\x42\xa0\x50\x29\x86\xa2\xdb\xcf\x09\x17\xb7\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/IAuthorityUpgradeable.sol",
            name="IAuthorityUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x12\x5b\x0e\x33\x61\xb2\x75\x86\xf6\xc6\xec\x98\xd2\xfe\x34\xa3\x34\x58\x62\x32\x97\xbb\xe3\xa9\xea\x5b\xbe\xfa\x30\xf7\x22\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2
    b"\x12\x9e\xdc\x64\x23\xe6\x38\x05\xeb\x5c\x44\xb2\xd0\xad\xa6\xdc\x43\x6d\x0a\x91\x8d\x28\xa6\x64\xb2\x86\x49\x86\x70\x0d\xbc\x59": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.3.2",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x12\xad\xf8\xc6\x46\xfc\xdf\xa2\xcb\x0a\xbe\x13\x0e\x63\xc9\x53\x9e\x0e\x62\xbe\x69\xbc\x9d\xa0\x45\x2a\x05\x20\x0d\x5a\x89\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111
    b"\x12\xae\x36\x95\xf1\xb2\x90\xe2\x00\xe1\x31\x41\x3d\xe2\xad\x65\xac\xcb\x36\x37\xf9\x02\xdb\xf3\x52\x42\xaa\x3d\x14\x5e\xb8\x29": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.111",
        ),
    ],
    # Hash appears in versions: v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49
    b"\x12\xc6\x55\xfd\x39\xac\x63\xd7\x08\x1f\x89\x28\x7d\x2e\x9c\x4d\x6d\x47\x9f\xee\x6e\x74\xc8\x43\x9a\xc9\x2b\x4c\x2e\xd5\xe4\xc1": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.49",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x12\xcf\xb0\xae\x57\x43\xcb\xd0\xa5\x7f\xc6\x76\xb6\x63\x1c\x95\x25\xed\x22\x94\xc1\x09\xc1\x85\x31\x07\x83\xb5\x8b\x79\xba\x65": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\x12\xdf\x63\x8f\x40\x37\xeb\x89\x16\x5a\xed\xdb\x7a\x87\x29\xc3\xb9\x09\xaa\x95\x3b\xfd\xce\x29\x67\x0e\xde\xa4\xb2\x86\xc0\x8d": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x12\xe2\x0d\x3d\x80\x4c\x60\xe2\x0e\x43\xd6\xdb\xdc\x6a\x7d\x4a\x03\x4a\x73\xbf\x53\xb7\x01\x35\xc0\x56\xca\x1a\x39\x1a\xbc\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x13\x11\x4a\x7a\xc3\xed\x6e\x74\x54\x49\x84\x03\x83\xec\xc5\x31\xea\x6c\xd6\xfc\xc6\x73\x5b\xbc\x94\x55\x35\x97\x97\xac\x8a\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x13\x20\x67\xfa\x85\xf0\x84\x10\x12\x8f\x75\x3a\x35\x06\x4a\xe5\x4f\xb4\x3c\xc6\x58\x5a\x9d\x06\x58\xe8\x9a\x36\x2a\xf3\x32\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesSuperQuorumFraction.sol",
            name="GovernorVotesSuperQuorumFraction",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesSuperQuorumFraction.sol",
            name="GovernorVotesSuperQuorumFraction",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.1, v5.0.2
    b"\x13\x2a\x4d\xcc\xd5\x5b\xf3\xff\x28\xbc\x0b\xd5\xec\x69\xe9\x81\x22\xba\x3c\x78\xa5\xc5\x3e\xd3\x24\x96\xea\x7b\x39\x5b\xac\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x13\x2b\x4e\xeb\x47\xcb\xcd\x0b\xd9\xcc\x99\x63\x75\x7b\x0a\xf9\x29\x96\xd1\xf3\x1d\x73\x7f\x85\xdf\xb8\x08\xba\x81\x8f\x6b\xf7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x13\x41\x7b\xfe\xcc\xb7\xe0\x8b\x86\x08\x15\x68\xee\x10\x0c\xf1\x73\x54\x00\x77\x06\x57\x21\x88\x04\xa0\x1a\x1a\xa7\xd7\x39\xb1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236
    b"\x13\x6e\x19\x08\x44\x42\x47\xa7\x76\x17\x48\x9f\x02\x46\xcc\x9c\xab\x5d\xb2\x45\x99\xd2\x44\x41\xab\x03\xb9\x70\xd3\xbd\xa7\x0d": [
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.236",
        ),
    ],
    # Hash appears in versions: v0.0.93
    b"\x13\x83\xf4\x30\xc1\x1d\x70\x5b\xa5\xcb\x69\xab\x9d\x5c\xa6\xfc\x1d\x48\x3e\x23\xbd\x7e\xec\x8a\x70\xa6\xd2\x57\x9f\x6c\x13\x9d": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.93",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x13\x99\x8f\x8e\xd8\xad\xf9\x0f\x8e\x1a\xb7\xd4\x14\x5e\x76\xb6\x0a\x0d\xec\xbc\x72\x5a\x3b\xf2\x0d\x7f\x41\x58\x7a\x4e\xb1\xcc": [
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x13\xb1\x69\x4f\xa3\x69\x61\xc9\xc5\x5f\x5f\x17\x17\xc9\xff\x88\xd7\x40\xdf\x34\x87\xc0\x60\x14\x98\xe8\x2f\xd8\x97\xf0\x3b\xca": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x13\xb3\x17\x49\xac\x91\xa3\x7e\xe7\xb8\x7c\xfc\xf1\xf5\x33\xa4\x03\xb5\x09\x17\x28\x67\xd6\x50\x35\x7b\xd9\x69\x0d\x39\x4a\x54": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x14\x0a\xdd\x25\x51\x0e\x96\x9d\x95\x5c\x35\xf0\xfd\x35\x5a\xca\x68\x29\x7f\x68\x05\x9a\x51\xa5\x6d\xbe\xca\xe9\xec\x1e\xd6\x58": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x14\x0b\xa7\xf0\xaf\xb6\xa4\x89\x68\xfa\x07\xbd\xa5\x19\xb5\xc0\xe9\xc9\xb6\xd2\x48\x29\xcb\xf4\x0f\x0f\xa1\xb5\x7f\xec\x76\x73": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x14\x4f\xd0\x75\xd0\x16\x7b\xcf\x1c\xda\x5b\x0d\x1e\x19\x1f\x3c\xfa\xb4\x0b\x86\xae\x17\x57\x71\x24\x5f\x51\xd1\x03\x2b\x42\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x14\x50\x8a\xf1\x7b\x22\xd2\xdc\x46\x2f\x0e\x95\x06\x08\x4b\x2e\xe3\x98\xa7\xa2\x0c\x03\xb3\x0b\xae\x1d\x9f\x23\x6b\x1d\xf3\x11": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.48
    b"\x14\x60\xf2\x4f\xc3\xdf\x19\xb7\x35\x0a\x49\x50\x84\xf5\x1f\x12\xd0\x5e\x5f\x42\x25\xc0\x99\xc7\x8d\xd1\xe9\x42\x9f\xc5\x5a\xad": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.48",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x14\x82\x79\x0a\xed\x2a\xdb\x0f\xe2\xe4\x92\x47\x42\x91\x47\xd1\xb3\xcc\x27\x8c\xa6\xef\xae\x73\x9a\xb7\xbd\x6d\x00\x67\x79\xaf": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/SwapMath.sol",
            name="SwapMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x14\x8f\x6b\x1c\x99\x07\x92\xd3\x6f\xdb\xfb\x5b\x1c\x4d\xc4\x42\x2a\x1e\x44\x0e\x6b\xac\x62\xc2\xc8\x1f\x59\x30\xa7\x63\x32\x9d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x14\xa0\x04\xdd\x25\x65\x31\xd1\x73\x23\xdd\x34\xcd\xa0\x29\x30\x4e\xc7\xad\xcb\x02\xa8\xbf\xb3\xc2\xfe\x88\xb4\xee\x41\x3c\xae": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\x14\xa0\xb6\x41\x20\x72\x49\x9e\x67\x29\x37\xa2\x7a\xc9\xc3\x08\x0d\xf4\xee\x14\x74\x8a\x79\x79\x8d\x1b\x22\x87\x3f\x4f\x50\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x14\xa5\xfa\x59\xba\xd6\x7b\x18\x2f\x39\x43\x9e\x2f\x3f\xf1\x3b\xe3\xbe\xde\x68\x4d\x69\x7b\x3d\x6e\x36\x66\xfb\xd6\xcc\xa0\x36": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x14\xa6\x73\x1b\xd2\x1e\x49\x8c\xee\x09\x9e\xde\x71\x9b\x4f\xc0\x38\xc1\x61\x49\x88\x75\x34\xf3\xf2\x0d\x82\xa2\xda\x21\x83\xe3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x14\xc1\xbf\xb3\xf3\x61\x5b\x2b\xe1\x16\x3f\xfb\x10\xe4\x56\x86\x42\xfe\x16\x70\x4d\x8e\x9e\x5c\x26\x62\x83\x10\xf9\xc1\x9f\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x14\xc2\x4f\x46\xcf\x68\xaa\xbc\x14\x37\x4d\x89\x16\x35\x91\xb0\x7a\x05\xed\xe5\x5e\x5d\x3c\xbd\x35\xd0\xb4\xbb\x43\xc1\x7a\x16": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x14\xc9\x13\x10\xd9\x07\xbe\xa0\x76\x6b\x5e\xdd\xf5\xb9\x1f\xd0\x43\x24\xe5\xa5\x53\x82\xf2\x2d\x03\x88\x49\x73\x97\x0d\x4c\x4b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IPaymaster",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IPaymaster",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IPaymaster",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IPaymaster",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IPaymaster",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x14\xe2\x1d\x4e\x81\xa0\x1f\x58\x62\x88\xf0\xfa\xff\x4e\x3c\xf4\xb4\x9b\xdf\xa9\xed\xc7\xce\xcc\xe6\x50\x1e\xc9\x00\xfc\xc0\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x14\xe8\xea\x62\x13\x44\xa9\xeb\xee\x3b\xc2\xb8\x18\x35\xb9\x59\x36\x1b\x8b\x63\xc4\x46\xc9\x66\x67\x49\xd4\x41\x8e\x92\xa5\xa1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x14\xed\x4b\x54\x4a\x6a\x1c\xe8\x2a\xb1\xbb\x82\x54\xa5\xbc\xa0\xf5\x35\x2f\x85\x33\xe3\x59\x54\x3c\x90\x36\xb0\x4a\x4c\x99\xd1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x14\xf1\x43\xc1\x33\xe4\x90\x4e\x7e\x87\x08\x03\xc1\x6e\x4a\x18\xca\x17\x9b\x70\xec\xb7\x61\x5d\x8a\x45\xae\xd6\x81\x38\x3b\x01": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x15\x08\x0e\x9b\x90\x17\x7e\x8e\x53\xd6\x81\x53\x9b\x88\x4e\x1e\xb6\x8a\xd6\x65\xe0\xac\x9b\xad\x42\x98\xbd\x72\xc9\x6c\x5d\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.0.1, v3.0.2, v3.0.1, v3.0.2
    b"\x15\x17\xe3\x53\x4f\x4f\x81\x6f\x72\xf0\xfb\xa7\x12\xe2\x99\x61\xa5\x66\xb6\xdb\xd0\x7a\x47\xf8\x2e\x62\x01\x58\xcb\xd6\x59\x77": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x15\x2c\x76\xd7\x9a\x07\x0e\x00\x7c\x61\xb0\xcf\xed\x25\x6b\xcd\x6b\x0e\x43\xea\x38\xbb\xaa\xb1\xc4\xf6\x63\x50\x72\x5f\x24\xd7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x15\x3f\x44\x17\x67\x3e\xe1\xd0\xf0\xa0\xc7\xfd\xd5\xc1\xbf\x5c\x3e\x49\xf4\x68\x44\xbd\xec\x0e\x7d\xe1\x11\xc9\x7f\xbb\x93\x8c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x15\x64\xd1\x21\x3c\xd6\xcd\x2e\x72\x7c\xdd\xb6\xb1\xbe\x8b\x22\x97\x11\x9e\x6e\x7e\xc9\x48\x26\xd2\x71\x1a\xd5\xfc\x19\x74\x13": [
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/auth/TimedRoles.sol",
            name="TimedRoles",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123
    b"\x15\x78\xb7\x66\x9e\xac\x15\x84\xd5\xeb\x32\x28\xa4\x18\x5a\x15\x83\x02\x52\x1a\x2f\x81\xef\x3b\x62\x67\xd5\x7e\x54\x24\xab\xe4": [
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.123",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x15\x78\xf2\x7a\x61\x7e\x1f\x86\xb3\x37\x51\x41\x9a\x62\x82\x72\xb0\x50\xeb\xc8\xb6\xa4\xe4\x1f\xe6\x08\xcc\x8b\x02\x42\x1f\xdf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x15\x79\x36\x99\x5e\xf6\xe4\xf2\xbf\xbd\xa6\xae\xea\x79\x32\x2a\xcb\x91\xce\x32\xfc\x28\xca\xcd\xc4\xca\x8f\x7a\x6c\x59\x31\xee": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\x15\xa3\x9d\x30\x89\x60\x16\x7f\x4b\xd4\x23\xa8\xdf\xe4\x96\xc6\x8d\x15\x86\x27\xd4\x5d\x22\x9d\x7d\x01\x86\xbe\x11\xa9\xf9\x12": [
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x15\xac\x52\xe6\x10\x99\xf8\x21\x31\x80\x5f\xe2\x5a\xb5\x24\xdb\xd3\xa4\x6e\x72\xa7\xca\xfd\x98\x39\x4a\xf6\x59\x7d\xd6\xa2\x55": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x15\xf2\xfd\xf8\x1e\x3a\xe7\xb7\xec\xc6\x4d\xcf\xac\x2b\x92\x55\xf0\xb3\x9c\xd6\xec\x23\xee\xd4\x5e\x78\xa0\xe8\x04\xc1\xe2\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x15\xf5\x17\x30\x8e\x8d\x06\xcc\xa8\x5d\xf4\xb0\xca\x22\x8b\x96\x4e\x69\x2c\x45\x68\x89\x65\xa7\x2f\x6d\x70\x86\x43\xa1\x9f\xdd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x15\xfe\x69\xfa\x3e\x14\x9b\xf8\x31\x9e\x93\xdb\x38\xcc\xee\x1d\x9b\x2c\x7b\x45\x67\x56\x7b\x0a\x49\x6c\xc9\x4d\xe1\x68\x47\x4f": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC20PermitAllowed.sol",
            name="IERC20PermitAll owed",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC20PermitAllowed.sol",
            name="IERC20PermitAll owed",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC20PermitAllowed.sol",
            name="IERC20PermitAll owed",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC20PermitAllowed.sol",
            name="IERC20PermitAll owed",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC20PermitAllowed.sol",
            name="IERC20PermitAll owed",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.133, v0.0.134
    b"\x16\x00\x97\x2d\x04\x1a\xe6\xe2\xaa\xb6\x03\xad\xf6\x07\x0a\xd8\x6f\x7b\xbb\xb4\xbc\x5c\x37\xe2\xec\x79\x2b\xab\x59\xe2\xda\x23": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.134",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x16\x10\x97\x29\x2f\xff\x9c\x14\x6c\xeb\x69\x51\xbd\x96\x57\xf7\x3b\xfc\xff\x4c\x1d\x65\xda\xf4\xab\xa9\xdc\xfd\x51\x90\x5a\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x16\x20\x38\x76\x01\xce\xbe\xce\x39\x4e\x6d\xae\xe6\x7e\x1a\x7b\xf5\x0c\x12\x50\xf6\x56\xe1\x5b\xb6\x0a\xb7\xe0\xc4\xfc\x19\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x16\x7a\x79\xde\xc2\x1b\xd1\x78\x43\xf9\xc8\x7b\x04\x2a\xd3\xae\x02\xed\x7b\x1f\x04\x60\x6a\x54\x30\x6a\x75\x75\x6c\x9b\x6d\xf0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x16\xb6\x8c\x83\x06\xbc\xf5\x3d\x0a\xff\xa9\x8e\xc9\x62\x73\x34\x17\x15\x8f\x5b\x53\xe2\xc8\xc6\x03\x31\x26\xa2\x8d\xa8\xbe\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x16\xd9\x4d\x4e\xb7\x83\xa2\x8c\x3f\xcf\x55\x7a\xf2\x97\x6b\x22\x2b\x39\x0a\x64\xac\x99\xe9\x84\x41\xe5\x8c\x1b\x7d\x8d\xe1\xb5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x16\xe2\x10\xa6\x12\x96\x44\x04\xab\xda\x22\x6a\x53\xc9\x6c\x84\x96\x4f\x72\xf5\x13\x0d\x0f\x61\xcb\xaf\xe4\x8f\xa7\xae\xb8\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x17\x38\xf9\x8b\xac\x77\xa1\x20\x1e\xc2\x66\x9d\x9c\xa8\x4d\xd6\x71\xe5\x2d\x89\xcd\xb0\x97\x0d\x6f\x9a\x1c\x99\xdd\xc8\x93\x74": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.20
    b"\x17\x3e\x18\xad\x0f\xbe\x93\x0f\x07\xf0\xc9\xde\x07\xa0\x82\x73\xe5\x81\x7a\x4b\xd5\xc8\x20\x8f\x71\x60\x7d\xdf\x77\x1e\x87\xcf": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.20",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x17\x48\x5d\xaf\x96\x8f\x91\x48\x74\xe9\x3a\xd4\x91\xc1\x83\x97\x8a\xf9\xf7\xc5\xef\x4b\x01\x02\xdd\x49\xf8\x35\xd7\xe8\x25\xfa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x17\x54\xa4\x3b\xe6\x8b\x67\x95\xf6\x2a\x1b\x26\xd5\xc8\x06\x50\xa8\xaa\x6d\x99\xbc\x2d\xc1\x80\x4e\x91\x94\xa6\x4e\xad\x11\x08": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x17\x58\x12\x1b\x2d\xbe\x93\xe5\x5d\x5e\x27\xe6\x4d\x24\xc2\xc7\x7b\x50\xaf\xb6\x06\x1b\x48\xc3\xdf\x2f\x65\x28\xb8\x9f\xe0\xf1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x17\x5e\xe7\x3a\xca\xb4\x25\xb8\x2e\x33\x88\xf8\x9d\x23\x82\x90\x43\x5e\xd6\x52\xa6\xae\xb5\xfb\x57\x3e\x40\x44\x4d\xfc\x8a\x4c": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/SelfPermit.sol",
            name="SelfPermit",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/SelfPermit.sol",
            name="SelfPermit",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/SelfPermit.sol",
            name="SelfPermit",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/SelfPermit.sol",
            name="SelfPermit",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/SelfPermit.sol",
            name="SelfPermit",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x17\x80\xc7\x46\xbc\xd9\xbc\xc0\x9f\x49\x89\xfd\xa4\x28\xf8\x22\xae\xad\x3f\x61\x1d\x05\x08\x28\xba\x4d\xc5\xd1\x05\x8e\x07\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x17\x82\x21\xc0\x57\x4e\x0e\x89\x5f\x70\x29\x02\x5d\x6e\xb4\xd5\x3c\xb4\xa8\xb5\x4b\x29\x74\x61\x9f\xd5\x21\xe5\xc0\x5e\x30\x85": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x17\xa2\x59\xa8\xbc\x24\xdf\x54\xd3\xc2\xcf\x81\x3e\xc4\x6c\xdc\x30\x00\x24\x63\x54\xc8\xaf\x46\x92\x17\xe3\x19\x7f\x68\x97\xf3": [
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.132
    b"\x17\xa8\xae\xdd\x35\x7e\xc3\xff\x00\xa8\x31\x8a\x95\xcb\x30\x16\xed\x75\x72\xf6\xff\xca\x72\x5a\x29\x44\x0e\x32\xc1\xf8\x18\xfe": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.132",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x17\xab\x69\x19\x0c\x71\x95\xcc\xaa\xa6\x9b\x45\x30\xed\x29\x12\x53\xc7\xfb\x1d\x84\xe0\xb8\x98\x99\xd8\x4b\x92\xad\x4a\x74\x42": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/Exttload.sol",
            name="Exttload",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207
    b"\x17\xd5\x88\xf9\xe1\xae\x04\xfc\x21\x98\xf5\xf6\x86\x29\xba\x5f\xc1\xcd\x2c\x17\x66\x9d\x46\xe0\xc9\x77\x0e\x3f\x17\x10\x4c\xd6": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.207",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x17\xf7\xdb\x5b\x8b\xb2\xff\x7d\xb6\x94\x1f\x1c\x2a\x9d\x8b\x9a\xe9\xe5\x48\x2f\x91\xa4\xef\x36\x10\x3f\x9e\x0c\x19\xa5\xbe\x44": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x18\x12\x58\xe5\x26\x64\x05\xc5\x96\x7c\x36\xbc\xb9\xa1\x3a\x03\x9c\xb6\x86\xee\x46\x2a\x33\x77\xba\x2e\x06\x54\x55\x54\xae\xe9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x18\x14\x99\x79\xd9\x43\xf9\x21\x4b\x34\xec\x97\x38\x4f\xa1\x89\xa8\x37\x1f\x85\x9c\x11\x6b\x60\x63\x68\x6e\x43\x42\x20\xd8\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x18\x1c\x2c\x06\xe7\x66\x04\x7b\x3d\x3d\x33\xb3\xf0\x11\x50\x21\x11\xf3\x9c\xcd\x94\x40\xb7\x25\x4d\x48\x8b\x4f\xdd\x41\xf3\x82": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.130
    b"\x18\x30\x7b\x78\x74\x5b\x45\x98\x8c\x5c\xdc\xd2\x8d\x4a\x2a\x92\x17\x32\x89\x2d\x07\x37\xf9\xeb\xb3\xf0\x43\x99\x1c\x26\xca\x96": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.130",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14
    b"\x18\x6e\xa2\xff\x14\xa2\xe1\x65\xe4\x53\xa4\xae\xc3\xba\xf4\x85\x83\x05\x4d\x5b\xd3\xd5\x96\x8f\x45\xb7\x3a\x93\x6c\x48\xb9\x82": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.14",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x18\x98\x36\xbc\x72\x5b\xdb\x0d\x83\xfa\xee\x33\xa6\x18\xf5\xd9\x38\xaa\x43\x3c\x51\x76\x2f\x03\x17\x98\x2e\x4a\xf8\x9a\x00\x10": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x18\xa6\x2b\xff\x2b\x3e\xfd\xb2\x29\x22\x72\x1e\xd5\x7d\xaa\x1b\x54\x7d\x01\x6d\xda\x36\x2b\x56\x5f\x92\x41\xb3\xba\x74\x91\xdc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180
    b"\x18\xae\x15\x2d\x83\x87\x54\x59\x67\xa1\x49\xb1\x48\x6d\x40\xd6\x52\xf0\xe0\xe6\x40\x52\x9f\x3c\x6a\x81\xb3\x75\x55\x09\xd7\x57": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.180",
        ),
    ],
    # Hash appears in versions: v6.6.0, v6.6.1, v6.6.2, v6.7.0
    b"\x18\xf5\x58\x31\x91\x87\x7f\x44\xca\xa5\xe2\x88\xf3\xff\xb7\x20\x56\x04\x3f\xf6\x46\xe3\xa2\x4d\xa7\x41\x72\x5b\xf1\x39\x93\xbd": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.7.0",
        ),
    ],
    # Hash appears in versions: v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79
    b"\x19\x27\x4f\x6d\xe5\xe9\x50\x2a\xea\x94\x2c\xf7\x66\xf3\x16\xc5\x91\xfc\xa5\x9d\xee\x8e\xdb\x79\x86\xd6\x1a\x86\x8d\x91\x5d\x45": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.79",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x19\x3a\x4d\xfa\x8c\x37\xe8\x16\x16\x86\x0c\x49\x65\x8a\x68\xd7\x43\x7d\x72\x65\x67\xb1\x51\xf8\xb8\x11\xef\x8c\xad\x67\x26\x07": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/LiquidityManagement.sol",
            name="LiquidityManagement",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/LiquidityManagement.sol",
            name="LiquidityManagement",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/LiquidityManagement.sol",
            name="LiquidityManagement",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/LiquidityManagement.sol",
            name="LiquidityManagement",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/LiquidityManagement.sol",
            name="LiquidityManagement",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26
    b"\x19\x4a\x8b\x7d\x9c\xe1\x5a\x83\xdb\x08\x11\x7b\xa8\x9e\xf4\xcd\xf8\xe0\xc0\x42\x55\x29\x23\x9d\x6c\x1a\x66\x8e\x9b\xb2\xa2\xca": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.26",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x19\x62\x24\x08\xaa\x6d\xf9\xea\x61\x5a\x00\x2a\x23\xeb\x07\x75\x06\x90\x40\x98\xdf\x92\x0d\xba\xa2\xb9\x44\x15\xef\xf1\x4d\x69": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x19\x7e\x19\x73\x90\x8c\xaa\xff\x2d\xe7\x55\x11\x5b\x41\x41\x2d\x39\x04\xf2\xa2\x16\x4d\xf0\xf5\x79\xf5\xd2\xae\xbc\xd4\x3f\xe4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x19\x90\x25\x2b\x39\xa3\x1e\x51\xcd\x51\x94\x94\xdd\x48\x21\xfc\xf1\x18\xb4\x38\x57\x6d\x71\x70\xb5\x71\x2d\x67\xa4\x62\xda\xbc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/types/TimeUpgradeable.sol",
            name="TimeUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.106
    b"\x19\x98\x5e\xf7\xd2\xb4\x3d\x6c\x77\x75\x0f\x53\x95\x8a\x17\x94\x0a\x7f\x0a\xc7\x17\xf2\xb9\xd5\x10\x71\x82\x5c\x3f\x25\x30\xf8": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.106",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x19\xa0\x80\xd6\x52\xdf\x69\x6e\x74\xde\x73\x9a\x50\xb9\xec\x3b\x82\x88\xaf\x32\x70\xd8\x6f\x35\x0e\x4d\x69\xf5\xa7\x6d\xc6\xb3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURIUpgradeable.sol",
            name="IERC1155MetadataURIUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x19\xa4\xb5\x05\x06\x84\xb8\x69\x35\xf8\x26\x8c\xcb\xbd\x88\x50\x03\x7a\xdb\x34\x4a\x44\x61\x1f\x3e\xa0\xd3\xaa\x6d\x50\x1c\x54": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909TokenSupply",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909TokenSupply",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295
    b"\x19\xbf\x55\x9a\x6e\x2a\xe8\x3c\xe3\x57\xa7\x60\x01\x89\xbf\x07\x19\x8f\x7d\x32\xe0\x84\x4e\x2a\x9a\x46\x73\x1c\x1d\xd8\x9b\xf1": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.295",
        ),
    ],
    # Hash appears in versions: v3.4.2, v3.4.2-solc-0.7
    b"\x19\xd3\xbc\xdb\x65\xfc\xa3\xaf\x77\x58\x55\xe6\x41\x19\x4a\xb5\x34\xb0\x7e\x65\xb7\x6c\xcd\x51\x63\xf6\x00\x76\x50\x8d\x8e\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201
    b"\x19\xed\x74\xb0\x02\x38\x57\x35\xc9\x89\x47\x77\xa1\x88\xcf\x6b\xf9\x26\xa9\xfe\xc7\xaf\xbe\x76\xde\x72\xee\x21\x05\xba\xbc\x54": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.201",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.9
    b"\x19\xfe\x23\x1f\x67\x58\x37\x09\x08\xb0\xac\x2c\x9a\x21\xdd\xc5\x08\x97\x40\xdd\x4b\x98\xf7\x15\xad\x6d\x4a\x30\xa6\x26\x55\x83": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108
    b"\x1a\x02\x45\x43\x86\x84\x4c\x41\x9e\x20\xd2\x27\x17\xc2\xd3\x53\xc2\xf1\x07\xf6\xbf\x74\x7e\x24\x31\x11\x98\xce\x6f\x24\x77\x54": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.108",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7
    b"\x1a\x39\xeb\x1b\x5d\x47\xb5\xca\xc9\xa6\xbb\x95\x01\x12\x58\x35\x15\xb8\x07\xff\x6c\x60\x7c\x71\x23\xd5\xb3\xb5\xfb\x9f\x43\xe1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x1a\x4a\xfb\x30\xb8\x0d\x14\x01\xaa\x85\xd3\xab\x02\x9d\x12\x2e\x2e\x75\xd1\x65\xee\x05\xb6\x33\x4c\x58\x49\xe4\xcb\x39\x42\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x1a\x62\x94\x9b\xb3\x0e\xe2\xe6\x71\xc8\xec\xf6\x0e\x29\xf3\x1c\x99\x8c\xa0\x1c\x1a\x99\xd7\x3e\x06\x72\xb9\xdf\x6d\xb1\x6a\xc6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1a\x73\xef\x67\x1c\x66\x35\x41\x21\x49\x7e\xba\x10\xf5\xa7\xfd\x7a\xb8\x8e\xbf\x81\x1c\x67\xe7\x1f\x34\x16\x71\x6d\x48\x46\xc9": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x1a\xa6\x0e\xe6\x1c\x40\x0e\x1c\xd0\x41\xcd\x4b\x2e\x02\x9d\xec\x24\x82\xea\xaf\x04\xb1\x96\x06\x30\x34\xb1\xe5\xf5\x9a\xb6\x5a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.37
    b"\x1a\xbb\x6a\xf4\x7e\x05\xcd\xfa\xcd\x53\x7a\x03\x81\xd0\xa0\x24\x3a\xfa\xa3\x3e\x40\x4e\x65\xac\x75\x42\x64\xbe\xce\x35\xa6\x25": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.37",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x1a\xef\x7a\xfe\xde\xf4\x51\x3d\xc7\x7a\xa3\xb3\xd3\x1b\x80\xde\xf1\x42\x06\x97\x50\xe2\xc3\xe0\x14\x46\x34\x4e\x1d\xfc\x84\xf0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x1b\x05\x84\xaa\xd5\xb0\x1c\x65\x28\x61\x27\xea\xf2\x76\xba\x4e\x7f\x6f\x8b\xb3\x55\x4a\xb2\x9e\xb9\xb8\x54\xf3\x73\x5d\x56\x9b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/VanityAddressLib.sol",
            name="VanityAddressLib",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x1b\x0e\x61\x76\x1e\x4f\x74\x13\x1b\xbc\xe1\x77\x01\x57\x40\x66\xb6\x2c\x00\x77\x2b\xd9\x7d\xdd\xc2\x01\x18\xcc\x57\x92\xe0\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1b\x32\x8c\xea\xa3\x71\x46\x64\x5f\x8d\x14\x57\x65\x8e\xb3\xe9\xb4\x43\x48\x40\xe0\x8e\xa1\xcb\x87\x28\x18\x5c\xbd\x3a\xbc\x48": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.42
    b"\x1b\x60\xa3\xa0\x69\xc0\x5a\xa8\xc2\x78\x3b\xf7\x1f\x70\x5f\x35\x1e\x81\x21\x28\x1c\xb7\xfb\x65\x73\x2b\x55\xd8\x56\xd4\x75\x76": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.42",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x1b\x64\x15\xd9\xd8\x1e\x7f\x1d\x0b\x9d\x1f\xb5\x16\x05\x8a\x38\xaf\x95\xe4\x79\x00\x19\x81\xae\xaa\x24\xc4\xf2\x30\x60\x6c\xb2": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x1b\x9c\x5c\xac\x65\xf3\x8a\x23\x87\x22\xd9\x4c\x96\xf0\x3b\x06\x57\xfc\xf4\x58\x18\x3d\xfb\xe5\x70\xfd\xc3\xa1\xfa\x21\x93\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v0.0.191
    b"\x1b\xae\x49\x07\xc9\x10\x72\x54\x72\x7e\xcb\x87\x8a\xab\xd8\x71\x1a\x8f\xa9\xc2\x73\x23\x6f\xe9\x39\x55\xc8\xec\x86\x2c\x8d\x1f": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.191",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x1b\xc3\x9b\xdc\xd7\x46\x21\x56\xac\x2f\x16\xa6\xab\x69\x2e\xc9\xd9\x36\xc5\xc4\xff\x33\x6d\x4c\x15\x20\xb1\x0d\x46\x96\x00\x9c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285
    b"\x1b\xd1\x38\x2b\x40\x09\x3b\x2b\xf9\x53\xe5\xee\xa2\xd3\xb7\x1c\x94\x8e\xdb\x1d\x23\xc2\x67\x51\xcf\x12\x30\xdb\x9e\xc1\x52\x93": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.285",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x1b\xf1\x65\x09\x51\x17\x7a\x7e\x32\x99\x2e\xb5\x3e\xfe\x48\x32\x22\x17\x45\xb9\x19\x73\x98\xa0\xd7\xd8\x6f\xf9\xa3\x0c\x5f\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotes.sol",
            name="IVotes",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.3.0
    b"\x1b\xff\xa5\xef\x16\x6d\xe3\x9e\x91\x9e\x9c\x22\x3b\x07\x7a\x80\x16\x0d\x81\x65\x3e\xf2\x30\x05\xc0\x3e\xd1\x4c\x92\x35\x8c\x56": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="ERC4337Utils",
            version="v5.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1c\x13\xe5\x49\x73\xd8\x5a\x9c\x38\x8e\xf8\x64\x3b\x36\x5c\xd3\x94\x02\x2f\xaa\x5f\x41\xb6\x79\x6b\x09\x72\x84\xc5\x38\x2d\xbc": [
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x1c\x14\xb6\x22\x78\x42\x2d\x44\x42\x28\xa5\x4b\x99\x66\xf6\x18\xd2\xc7\x39\xb5\xef\x6a\x9f\x69\xef\x10\xa7\xe6\x15\xff\x2f\x26": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.199, v0.0.200, v0.0.201
    b"\x1c\x1b\x0b\xc0\x16\x2d\xfd\x05\x1b\xc2\x3f\xf8\x41\x5b\x4e\xa5\xf3\x5b\x4e\x06\x2b\x73\x93\x39\x6e\x02\x76\x8d\x88\x1a\x32\x21": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.201",
        ),
    ],
    # Hash appears in versions: v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160
    b"\x1c\x36\x6f\x18\xf9\x35\xe8\xc5\x6a\xa7\x26\x41\xd0\x7c\x29\x49\xe6\x55\x82\x0f\x72\xa4\x4a\x24\xd6\x67\x8e\xf4\xa7\x6d\x80\x96": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.160",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x1c\x45\x29\x33\xa5\x34\xfd\x07\x43\x89\x8b\x79\xfe\xa3\x11\xbd\xb1\x47\xa1\x50\x66\xac\x35\xbe\xe3\x48\xb9\x13\xa5\x15\x66\x0a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x1c\x51\x8d\xd7\x1a\x37\x30\xa5\x91\xa5\x87\x55\x40\x7f\x05\x04\x64\xab\x79\x57\x0b\xb3\xd6\xc6\xaa\xa0\x01\x89\xca\x35\xc0\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x1c\x5a\x7c\x7d\x84\xc3\x11\x61\xea\x2a\xa2\x6c\x95\x6a\xf5\xc0\xa9\x99\xdd\x79\x21\x26\x5e\x8c\x4b\xcf\xef\x6b\x8f\x2c\xca\x23": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISwapRouter.sol",
            name="ISwapRouter",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISwapRouter.sol",
            name="ISwapRouter",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISwapRouter.sol",
            name="ISwapRouter",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISwapRouter.sol",
            name="ISwapRouter",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISwapRouter.sol",
            name="ISwapRouter",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x1c\x66\xbd\x57\x74\x07\xf4\x6a\xf5\xda\x68\x94\xbd\x99\x49\x00\x9e\xab\x6a\x2e\x1a\x29\xb1\x55\x68\xad\x5c\xc3\xcf\x88\xe3\x2e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.180, v0.0.181, v0.0.182, v0.0.183
    b"\x1c\x72\x7c\xfe\xfb\xe2\xf0\xce\xb3\xec\xc1\xd1\xd5\x94\xf9\x81\xa4\x90\xaa\x7e\xb6\xa4\xb8\xfb\xc9\x5e\xa2\x4d\x2e\xde\xc7\x22": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.183",
        ),
    ],
    # Hash appears in versions: v0.0.127
    b"\x1c\x79\x60\x7a\xfa\xa6\x4d\xf0\xf8\x86\x6d\xde\xaf\xfa\x7f\xd2\xe4\x47\x61\x2c\x9d\x81\xd9\x07\xa1\x78\x41\x09\x39\x31\xde\x9a": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.127",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x1c\x88\x4c\x85\xb6\x71\xa0\xd8\xf0\x8b\x27\xb6\xe0\x9d\xb0\x57\x28\xe9\xd2\xf9\x41\x26\x50\x3e\xb3\x4f\xfc\xd0\x91\x63\x7e\xb4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68
    b"\x1c\x98\xce\x63\xe4\x61\xe7\xa4\x61\xfa\x10\xa4\xd4\x40\x19\x5b\xa4\xfc\x71\xbb\x09\xf5\x95\x90\xf9\x36\x14\xb4\xfd\x80\xc9\x96": [
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x1c\xa9\xb0\x31\x2e\x76\xbc\x84\x73\x9e\xc7\x3a\x56\x9e\xa7\xf1\x94\x5c\xbd\xf0\x47\x53\x32\xf5\x49\x6c\x4b\xe8\xc7\x39\xc2\xa2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x1c\xb2\x48\xdb\x1f\x71\x03\x97\x13\xa4\x82\xc5\x90\x32\xff\xa3\xd6\x1b\x73\x17\x63\x25\xb8\x7a\xfa\x92\xb7\x96\xde\xce\xb8\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.4.2-solc-0.7
    b"\x1c\xc0\x4b\xf1\x75\x97\x73\x39\xbe\xa9\x92\xe5\x75\x04\x38\xad\xde\x57\x2b\x55\xee\x4b\xa5\x6d\xc8\xe4\xe4\x9a\x98\xca\x2b\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x1c\xc2\x14\xa6\xdf\x54\x93\x3e\x35\xb6\x80\x4b\x7c\x09\x3c\xa9\xad\x88\x1f\x07\xa5\x57\x33\xbc\x92\x96\x85\x72\x85\x02\x0d\xdc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1c\xee\xb1\x84\xbe\x66\xce\x3a\x45\x4f\xbe\xd7\xf6\x5a\x3e\x12\xd8\x71\x41\x3c\x7f\xb9\x8c\x1c\x7c\x19\x7a\xc8\x65\x2c\x95\x52": [
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1c\xef\x11\xb2\x42\x0c\x43\x24\x37\x1e\x14\x72\x94\x2d\x62\xe9\x2f\xea\xe3\x84\x6c\xb0\x9f\x81\x9a\xd8\xf3\x14\x34\x0d\xf5\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1c\xfa\x63\x2c\xba\xf1\x76\x2c\x3f\xe3\xcf\xb7\x59\xfe\x14\xef\x7b\x0e\xa3\xf3\x0e\x31\x01\xa9\xc4\xa7\x2e\x3f\xf6\xc1\xc8\xd6": [
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1c\xfc\x40\x34\x49\xf5\x10\xf7\xe3\x65\x8d\x65\x1e\xca\x21\xc1\xe1\x11\x88\x85\x6b\x8e\xc6\x62\x2b\x55\xc6\xe0\xbc\x6f\x49\x7c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1d\x08\x3f\x35\x1a\x81\x7f\x64\xf2\x38\x67\xe9\x19\x67\x25\x49\x45\xba\x3a\xaa\xc4\x6a\x0e\x5f\x4c\x13\x0d\xac\x8a\x61\x9c\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129
    b"\x1d\x0c\x0b\x52\xfe\xa0\x4a\x55\xda\x68\x02\xd1\xa1\x17\x57\xa8\x80\x6f\x10\x6b\x70\x87\x56\x15\x24\x7a\xc7\x56\x17\x42\x40\x4a": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.129",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x1d\x11\x9a\x2b\x4b\x52\xd7\xae\x31\xea\x3c\x6c\x38\xe9\xc4\xcf\x8e\xe7\x20\x15\xa4\xa5\x3c\xf2\xbe\x48\x5f\xba\xbd\x10\x0d\xad": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x1d\x58\x28\xd1\xb8\xde\x69\x31\xcc\x08\x85\xe9\xd8\x70\xe3\xe4\x9d\x8b\x65\xed\x59\xef\xcf\xbd\xb0\x62\xfa\xf1\x8a\xd4\x5d\x44": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x1d\x76\x77\x57\xf7\x5d\x51\x29\x01\x6f\x1d\x52\x1d\x0d\xde\x6e\xfb\xd5\x09\x0a\xd8\x93\x4d\x51\xe9\x5a\x5b\xd6\x30\xeb\xa3\x3b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x1d\x84\xf9\x94\x32\x69\x3e\xe5\xb8\xb4\xed\x73\xe7\x6d\xa5\xc8\x71\x48\xf6\xcd\x7c\x86\x60\x30\x7e\x2c\xa3\xbb\x23\x56\x50\x71": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.157
    b"\x1d\x8b\xce\x15\x7e\x8f\x6a\x55\x12\xda\x20\xe9\x0a\xf9\x71\x51\x41\xca\x67\x6d\x54\xf3\x41\x91\x08\x58\x38\x2a\xf0\x5b\x64\x53": [
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.157",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285
    b"\x1d\x96\xc4\x54\xde\xb8\x23\x6e\x8a\x11\xca\x2a\x7c\x9c\xa4\x9c\x26\x86\x88\xe3\xb4\x34\x4f\xf6\x67\xcf\xea\x15\x5c\x37\x61\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.285",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x1d\x9b\xba\x57\x06\x39\xb9\x48\x13\x0f\xd8\x4a\x02\x9c\xd1\xa9\xb7\x1c\xf9\x88\xf6\x07\xee\xd4\x29\x82\x77\xce\xd4\xa2\x86\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1d\xc5\x20\x2c\xde\x20\x80\x87\x48\x72\xa8\x3f\xec\x98\xdb\x60\xed\x62\x3e\x94\x3c\x5e\x55\xcc\x2c\xcf\xad\x09\x26\x58\x7c\x40": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1d\xcc\x53\xcf\x05\x0a\xf9\x89\x6a\x37\x92\xe0\x6b\x1a\x0e\xc7\xb0\x72\xbc\x75\x58\x2e\x05\xb5\xf3\xd8\xde\x24\x9b\xb4\x99\x3c": [
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1d\xdc\xd2\x16\xde\x91\x49\x7c\xfb\x76\x42\x99\x79\xb2\xf0\x2b\x0f\x39\x70\xdd\xb4\x2c\x1d\x5b\x3a\x11\xe9\xda\x84\x86\x74\xa9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268
    b"\x1d\xe4\xfb\xd6\x1a\x42\x0c\xa3\xde\xeb\xd2\xda\xd1\x7b\xcd\x66\xb1\x27\x08\xbc\x29\xd9\x24\xb6\x71\xa6\xe9\xa6\xed\x19\x5e\x21": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.268",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x1d\xf9\xb2\x6b\xc5\x1e\xa4\xfd\x63\x30\xfd\xe9\xe5\x3a\x5d\x9e\x77\xdd\xb7\x42\x62\x9a\x92\x86\x79\x02\x73\x8e\xf1\xa7\x80\x82": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x1e\x08\xa3\xcd\xd8\x61\x08\xe7\x3e\xed\x71\xa8\x6a\x32\x31\x7c\x5d\xe6\xe8\xd2\x93\x3d\xde\xdb\xb8\xf5\x8c\x70\x2e\xbd\x09\x78": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5313Upgradeable.sol",
            name="IERC5313Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x1e\x0c\x27\x65\x85\xb8\xd7\x88\xa4\x67\xe3\x3e\x57\x60\x9d\xfa\x06\xcf\x32\xf2\x93\xe5\x13\x61\xfc\x7a\xfb\x38\x73\xdb\xf2\x00": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x1e\x18\xeb\xfb\xbb\xb8\x6c\x8c\xbb\xf6\x68\xf7\xb5\xca\xc9\xc5\x36\x4b\x9c\x26\x88\xda\x0d\x63\xaa\xc1\xf9\xe8\x56\xeb\x0f\x26": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-rc.0
    b"\x1e\x1a\xbd\xe7\xce\x06\x5c\x46\xb5\xd1\x27\x31\x01\xca\x44\xe9\x84\x5b\xe2\x5f\x87\xb6\xb2\x6e\xc3\x74\xdc\x8a\x44\xf4\x54\x06": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x1e\x2f\x44\x41\x3c\x8c\xe3\x40\x0f\xba\x10\xfd\x1c\x44\x5e\xe6\xe8\xbc\x33\xa7\x69\xe3\xa3\x80\x43\x98\xdf\x27\x02\x82\x39\x81": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPaymentsWithFee.sol",
            name="PeripheryPaymentsWithFee",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPaymentsWithFee.sol",
            name="PeripheryPaymentsWithFee",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPaymentsWithFee.sol",
            name="PeripheryPaymentsWithFee",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPaymentsWithFee.sol",
            name="PeripheryPaymentsWithFee",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPaymentsWithFee.sol",
            name="PeripheryPaymentsWithFee",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x1e\x4e\x35\x55\x6e\xca\xe6\xb1\x81\x40\xd6\xc6\x67\x63\xba\x03\xbc\xa7\xe9\x4a\x40\x4a\xb9\xf8\x3c\xdf\x40\x4a\xf0\x81\x06\xe1": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150
    b"\x1e\x60\x1f\x41\x5c\xe4\x63\xd4\xd9\xe9\x65\xa4\x85\x90\xe0\x20\x93\xf4\x04\x42\xfb\x67\x96\x6c\xf6\x35\x97\xd3\x23\xdf\xc6\xd8": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.150",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\x1e\x8e\x9f\xd4\x98\x86\x9e\x7f\x89\x71\x5f\x6b\x85\x20\xee\x67\xf6\x5f\xe9\xfd\xa9\xb8\x3e\x2c\x8c\x3a\x9f\x29\x88\xec\xc2\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.111
    b"\x1e\xa7\x3e\x02\x93\xee\x88\x44\x21\xdd\xa6\x7a\x7f\x57\xfc\xce\xec\xe5\x73\x0a\x94\x26\xce\xa8\xb1\x2a\x87\x1b\x90\x59\xea\x91": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.111",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x1e\xb0\xfe\x6f\xee\x65\x36\x76\xee\x9d\xeb\x52\x84\xee\x44\x56\xb7\xc2\xb5\x9a\x9d\x83\xeb\xca\x27\xcb\x4c\x24\xa8\x76\x40\x8f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.214
    b"\x1e\xc9\xb0\xbc\x27\x7d\x7e\x87\x06\xa7\x7b\x6b\xf6\xec\xb3\x6e\x68\x9f\x08\x2d\xbc\xb3\x2e\xd1\x49\x24\x78\x30\x96\x54\x30\xbb": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.214",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x1e\xdf\x95\xf3\x22\x07\xe8\x62\x89\xe5\x3f\xe1\x51\x67\x55\x4b\x08\x6a\x27\xc8\x0d\x9c\x2a\x33\x12\x88\x47\x98\x2a\xdd\x72\xd0": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x1e\xeb\xb3\xcd\x1d\x03\x1c\xe3\x58\x90\xa0\xf0\xe6\xa1\x2a\xc6\x4b\x80\xda\x3f\x56\x06\xc4\x5b\x0d\x63\x7b\xf0\x94\xfd\xaa\x8b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x1e\xf6\x3b\x0f\xa7\x02\x8b\x98\xdb\xc1\x12\xf8\x07\x2d\xfe\x2d\xa6\x5c\x74\x85\x89\x9a\xdc\xf8\x8e\x27\xf2\xf2\x4c\xa2\x4e\xd1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x1e\xfe\xc2\xcc\xba\xb6\xe1\xa8\x3d\x55\x5b\x45\xf6\xec\xa6\xba\xa8\xae\x77\xd7\x5a\x86\x67\x40\x06\xc7\x1a\x62\x8b\xa5\x03\xc7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x1f\x18\x0e\x16\x69\x18\xcc\xcf\xab\xcb\xcd\xdf\x1b\xff\x22\xfe\x41\x59\x78\x58\x70\x25\x87\x2a\x6f\xc0\x07\x3f\xcb\x41\xf7\xe0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x1f\x1d\x3c\x2e\x74\x53\x27\xda\x61\x16\x70\xb6\x5d\x25\x08\x28\x10\xf5\x4c\xff\x5e\x8a\x40\xbb\xd3\x3c\x92\xef\x7a\x73\x20\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x1f\x1e\x0e\x21\xba\xf5\x23\xc9\x11\x7d\xf2\x34\x40\x2e\x12\x79\xe9\x8b\x9c\x92\x71\x4f\xb9\xcf\x56\x25\x1e\x97\x69\x2a\x2e\x81": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.38
    b"\x1f\x25\x7a\xa3\xcd\x3a\x60\x2f\xf3\xfd\xba\x92\xe2\x3d\xb3\x4c\x4c\x21\x19\x48\xce\x4c\x4c\x07\x09\xdc\x4c\xd4\x9a\x0c\x8f\x13": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.38",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x1f\x3e\xfe\x8b\xbd\xee\x2a\x6a\xef\x43\xc8\xdd\x82\x8a\xb4\x82\x7d\x75\xfe\xed\xa6\x8f\xe8\xef\x8f\x50\x18\x58\xc7\x52\xeb\xa4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x1f\x5f\x82\x0b\x93\xd8\x04\x5f\xba\xc1\x7e\x5c\xff\x65\xc5\xa4\x0a\x2d\xab\xe2\xc9\xf2\x78\x04\x1b\x8f\x9e\x6f\xb3\xab\x28\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Validator",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Validator",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Validator",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Validator",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Validator",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.7, v0.1.8, v0.1.9
    b"\x1f\x83\x36\xac\xed\x9f\x37\x16\x22\xd3\xc1\xe6\x83\xf0\x83\x42\x0f\x93\x75\x78\xf7\x1c\xe1\x95\x60\x3a\xb6\x52\xfb\xf0\x92\xf7": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x1f\x83\x6b\x80\x98\xa0\x31\xb6\x75\xf5\xe2\x36\x68\x0a\xd7\x88\xc8\xc6\x14\x3c\xb2\x77\x1f\xe5\xff\xcf\x9a\x2b\x56\xcc\x97\x81": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x1f\x85\x77\x62\x4f\xa2\xe2\x71\x91\x53\x8b\x45\xe5\xc4\x57\x97\x2b\xf8\x10\x79\xe4\x77\x6e\x3e\xde\x4a\x68\x43\x30\xf0\xaa\xc6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265
    b"\x1f\x89\x48\xa2\xbc\xeb\xe8\xdf\x62\xa1\x7e\xc7\x7c\xeb\xd7\x98\x76\xf9\xa2\xe6\x4d\xb4\x2b\xed\x97\x01\x1e\xf2\xa8\xab\x53\xb6": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.265",
        ),
    ],
    # Hash appears in versions: v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262
    b"\x1f\x8b\x2d\xf2\xde\xe7\xc3\x4a\xa3\x74\x5d\xcf\xd3\xc9\x50\x2c\x15\xed\xb8\xc3\x17\xf9\x3e\x82\xcc\x44\xf1\x0e\xee\x87\x9a\x93": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.262",
        ),
    ],
    # Hash appears in versions: v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274
    b"\x1f\x9b\x09\x85\xad\x1f\xd4\x01\x7a\xb9\x3e\x38\x16\x95\x91\x68\x0d\x8e\x40\x7b\xeb\x15\xf9\x60\xae\xde\x3e\xb3\xb6\xfa\x16\xfe": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.274",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x1f\xb0\x72\x3b\x3e\x33\x6c\x06\xb0\x2c\x0a\x00\xa9\x84\x94\xd9\x76\x59\x29\x44\x1f\xe8\x77\x13\xff\x22\x73\x38\xc7\x82\xd6\x04": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/Quoter.sol",
            name="Quoter",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/Quoter.sol",
            name="Quoter",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/Quoter.sol",
            name="Quoter",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/Quoter.sol",
            name="Quoter",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/Quoter.sol",
            name="Quoter",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x1f\xb4\xae\xf3\x71\x38\xb3\xca\x57\xce\xc5\x7d\x78\xdc\xdc\xf9\x78\x51\xc6\x07\x94\xbe\xbe\x41\x93\x23\x40\x47\x48\x2e\x5e\x54": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x1f\xf6\x27\x41\x03\x09\xb1\xde\x99\x42\x40\x51\xfa\x96\x97\xa9\x55\x2a\x92\xb1\xf6\x40\xde\x91\x55\x72\x07\x22\xba\xb5\x63\x23": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721BurnableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x20\x11\xc6\x44\x7f\x52\xf3\x03\x50\xdf\x4f\xca\x4d\xf2\x34\x40\x5c\x28\xb1\x23\x8f\x91\x71\x87\xc0\x76\xa3\x9d\xaf\x2c\x76\xee": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x20\x35\xeb\x13\xfa\x86\xde\x40\xda\x01\x78\xdc\x8b\xfe\x3e\x17\x6f\x06\x43\x04\xf1\x5c\xa5\xc4\x32\xd4\xce\xea\xe2\x26\xe0\x09": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x20\x3a\x2c\x5c\x5f\xdc\x7f\x11\x56\xa3\x7b\x2e\xa8\x03\x59\x1f\xae\xa6\xd7\xb1\x29\x76\xc7\x14\xd6\x7e\x9a\xb5\xf9\x2f\x36\x36": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x20\x43\x61\x89\x36\x91\x3e\x46\xe6\xb6\x8a\xac\xe6\x15\x20\x2b\x2d\x49\x9a\x76\x12\x40\xdc\xe2\x25\x22\xd0\x3b\x3d\xc5\x1f\x22": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Permit.sol",
            name="ERC20Permit",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.55
    b"\x20\x57\xf1\xb9\x6f\xa8\x35\x99\x83\x9d\xd4\x27\x42\xfb\x27\xdb\x68\xd8\x1c\xcf\x49\x7b\x70\xb5\xcc\x23\x6a\x82\x19\xbb\x60\x0d": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v0.0.183, v0.0.184, v0.0.185
    b"\x20\x7b\xbf\x0a\xd9\xcb\xb0\x26\x31\x33\x7a\x01\x75\xd5\x46\x05\x60\x54\x01\x3a\xaa\x7c\x2a\xb6\x34\xad\x0c\x12\x30\xca\x04\xe1": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.185",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x20\x88\x0c\x3b\x33\x90\xfc\x03\x05\xba\x72\xd8\x63\x66\xf1\x37\x6a\x84\x32\x9b\xd5\x02\x60\x32\x7e\x6f\x6f\xef\x71\xb3\x38\xb4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86
    b"\x20\xaa\x93\xa8\x06\x40\x87\x2c\xea\xed\x87\x79\xf0\xd4\x85\x46\x85\xd9\xa1\x7e\x71\xf6\xf9\x5f\xa5\x04\x8c\x82\xca\x97\x38\x74": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.86",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x20\xbc\xba\xc2\x2c\x30\x1d\xf9\x2c\x76\x23\x71\x16\xfa\xe0\x9e\x4c\xde\xc5\x5c\xe8\x3e\x63\xb4\x61\xd4\xe8\x0c\xe2\x7f\x08\x66": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x20\xc6\x2a\x6d\xa8\x48\x2e\x9b\xe5\x35\x3d\xdb\xa2\xf4\x22\x96\x48\x06\xe3\xeb\x87\xac\x2d\x81\xf5\xf9\xa0\xb8\xc0\xaf\x5b\xe8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x20\xc9\x3b\x3a\x6a\xd2\x6d\x15\x25\xbf\xfc\x7f\x9b\x79\xfe\x20\xfb\xf4\x23\xa7\x51\xc6\x9a\x36\xbb\x9a\x9b\xf3\x6d\x6d\x22\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x20\xcd\xe6\xaf\x24\x2d\x3e\x3b\xb5\xa5\xa9\xfc\x12\x3f\x89\xa7\x6f\x16\x84\x77\x2d\x45\xaf\x37\x61\xe8\xe7\x1f\xff\x0f\xec\x59": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x20\xea\xb3\xb0\x30\xe9\x41\xd9\x3c\x0d\xdc\xc1\x94\x68\xfc\xdf\xe4\x69\xdc\xbc\xd4\xa2\x74\x7a\x84\xce\xc2\x3d\x4a\x79\xfc\x4e": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/DeltaResolver.sol",
            name="DeltaResolver",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.3.0-rc.0, v3.3.0-rc.1, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x20\xee\x48\x82\xfd\xfd\xc8\x1b\x1f\x41\x96\x54\x7a\x3b\xae\xaf\x8e\xfe\x84\x63\x03\x8e\x1a\x64\x73\x1c\x28\x25\x0e\x8d\x88\xa2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v6.8.0
    b"\x20\xf8\x45\x0b\x4d\xf2\x0f\xd7\x23\xc1\xb7\x1b\x8f\x01\xf5\x58\x9e\x5c\xa9\x97\x77\xd2\xb7\x44\xe3\xd3\xd2\x6b\x19\x08\x4b\x32": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.86
    b"\x21\x0c\x78\x05\xbb\xeb\x26\x26\xb6\x08\xca\xcd\x09\xec\x87\x35\x5b\x40\x85\x57\x8e\xcd\x33\x7f\x80\x33\x2e\x86\xf8\x85\xd4\xff": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.86",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x21\x12\x18\x49\x64\xd1\xb3\x7d\xfc\xa8\x6b\x71\xc6\x26\xe8\x03\x19\xae\xcc\xea\xa6\x0d\xaf\x5b\x68\x63\xe0\xf7\xc3\xa1\x54\xe8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC1155Errors",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x21\x66\xda\xe4\x29\xb3\xc3\x84\xac\x90\xfd\x90\xbc\x83\xf5\x68\xeb\x7a\x5a\x8b\x6a\x11\x42\xb4\x80\xbb\x9f\x97\xd9\xbc\xb0\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x21\x7d\x5b\xa8\x7e\xe4\x9b\x38\xfa\xae\xb4\xc8\xed\xe0\xed\x90\x41\xfa\x04\xeb\x88\x40\x9d\xeb\x16\x57\x2c\x6e\xf3\xd1\xeb\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721MetadataUpgradeable.sol",
            name="IERC721MetadataUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x21\x99\x7d\xba\xcf\x97\x75\x09\xf2\xbf\xbc\xe4\x3d\x0f\xc9\x25\xe1\x57\xf1\xfd\x84\xd5\x9c\xd3\xe0\x39\x43\x50\x90\x49\x57\x4e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x21\xa6\xfc\x7f\x7d\x6f\x70\xeb\xb8\x7f\xa0\xbc\x3e\x3a\x98\x80\x78\xa1\xba\x99\x9d\x28\xf5\x1d\x0a\x29\x24\xe5\x13\xf7\xd0\xc5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.7, v0.1.8, v0.1.9
    b"\x21\xaf\xd9\x9d\x99\xaf\xc4\x80\xc7\x71\x24\x2b\x21\x8b\x73\x34\xa2\xf0\x65\x3a\x71\xaa\x1d\xf2\x75\x74\xd0\x12\x40\x1f\xef\xb9": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x21\xcc\x8c\xb6\x1d\xfc\x94\x1d\x6e\x03\x97\x5c\xd6\xe9\x99\x80\x10\x02\x32\x8c\x3f\xf5\x96\x7c\xbb\xae\x98\xaf\x4c\x0b\x59\x2b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x22\x00\x6a\x5a\xba\x44\xb8\x3a\xb3\xa2\x9e\xd5\x99\xc3\x4e\xef\xda\x14\x93\xe4\x7d\xfd\xb0\xee\xdc\x70\x48\xf0\x6d\xc5\x6b\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x22\x02\xdc\x1e\xd9\xbb\x81\x61\x6a\x8f\x18\xcb\xd1\xa9\x57\x33\xf5\x5c\x9e\x6d\xbc\x40\xce\x11\xa8\x93\xbe\xc9\x4d\x91\x51\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.254, v0.0.255
    b"\x22\x20\x44\xed\x2e\x75\xe5\x2c\xbb\x7b\xd3\x7f\x2b\xe2\xa4\xf2\x49\xcb\x91\x0e\xcb\xc6\xb6\xf5\x4c\xce\x08\x21\x36\x03\xf7\x0a": [
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.255",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x22\x22\xf7\xc0\x72\xad\x2c\xf7\xf1\x58\xcc\x0e\xbf\xef\xc3\x60\xb0\xb6\x80\xb8\x27\x07\x78\xbe\x55\x8a\xe4\xc4\xdf\xb6\xca\xd3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87
    b"\x22\x41\x30\x46\x61\xb5\xa1\xcd\x65\xba\x70\x2b\x3c\x53\x69\xaf\x4b\x8c\xdb\xd6\xf4\x2b\xa3\x11\xf5\xd6\x9b\xd7\x49\xf5\x9a\xaf": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.87",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x22\x44\x52\x4b\xe6\xc3\x53\x36\x5a\x13\xdc\xa6\xe0\xb7\x7f\x22\xef\x8b\x8b\xbd\x22\x9b\xe5\x17\xed\xfa\xeb\xbd\x09\x45\x3e\xb5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.3.0
    b"\x22\x45\x7a\x33\x26\x1b\x95\x1b\xea\xad\xe0\x85\x30\x6e\xf9\x5e\x9e\xab\x6f\x87\xd0\x9a\x45\xf1\x87\xa2\x2f\x84\x54\xcc\xdb\x48": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPayments.sol",
            name="PeripheryPayments",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x22\x4c\xa7\x39\xd6\x56\x17\xc8\xff\xfc\x84\xd8\x22\xbc\x45\x06\x21\x2d\x43\xfd\x7e\xba\x1d\x81\xb7\x8c\x80\x61\x06\x58\xa6\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x22\x60\x6d\x33\x7a\x3d\x5f\x59\x60\x21\xd5\x3c\xe7\x8a\x54\x25\xdf\xfa\x71\x98\x10\x70\xe5\xdb\x7b\xeb\x52\x00\xad\xba\xdd\xd2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x22\x70\x08\xa6\x2f\x1a\x97\xe6\x9c\x35\x63\x64\x47\x50\x3b\xdb\xfb\xc6\x60\x72\x91\x2b\x63\x7b\x7a\xc3\xbc\xb1\x6c\x96\xf1\x7b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20FeeUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x22\x7e\xff\x4b\x6b\xc7\xc3\xc4\x25\x99\x85\x0d\x39\x3c\xe2\xfb\x55\xe1\x55\x31\x74\xb1\xf3\xdf\x53\x36\xd2\x66\x32\xb5\xb9\xbe": [
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v0.0.24
    b"\x22\xba\x66\x0c\xcf\x0e\x96\xd9\x2a\xe0\x94\xc5\x04\x13\xc3\x3f\x5c\xa6\xe8\xf6\xb2\x46\xdd\x79\xea\xea\xea\xbe\x6c\x0c\xb9\xee": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.24",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x22\xc1\xf7\x3e\x31\xac\xc8\x8a\x59\x24\x96\x12\xc5\x1b\x49\x35\xc2\x28\x51\x4d\x29\x5e\x97\xab\xb0\xd3\x10\x87\x92\x4a\x41\xbe": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/ChainId.sol",
            name="ChainId",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/ChainId.sol",
            name="ChainId",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/ChainId.sol",
            name="ChainId",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/ChainId.sol",
            name="ChainId",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/ChainId.sol",
            name="ChainId",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x22\xc5\xd1\x69\x71\xdc\xcd\x74\xd7\x03\xb2\x39\xde\x0d\xdb\xb1\xdc\xb4\x50\x51\x9f\xad\x8f\xe1\xed\x3e\x60\x08\x5b\x3c\x2e\x97": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112
    b"\x23\x1a\xe2\x4a\xc2\x81\x16\xdf\x6a\x6e\x88\xd2\x8a\x88\xf7\x1e\x74\xbd\x96\xd3\xe6\x4b\x5d\xa8\x8e\x96\x64\xa2\x97\xcb\xf8\xdd": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.112",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x23\x30\x10\x32\xab\x8a\x6a\x7b\xf9\xca\x80\x45\x09\x85\xc2\xba\x44\xa3\x0a\x91\x68\x5f\x05\x7f\xfd\x42\x00\x3e\x31\x84\x63\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165.sol",
            name="IERC165",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165.sol",
            name="IERC165",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181
    b"\x23\x40\xde\xbb\xb6\x77\x02\x6e\x16\xf4\x5a\x4d\xee\x87\x87\x4a\x3b\xeb\x98\xe5\xc6\xab\x43\x77\x89\xbe\xf9\x94\x7c\x58\x2b\x03": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.181",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124
    b"\x23\x50\xb5\xab\x47\xc6\xd6\x41\x77\x71\xcf\x2c\x04\x91\x0f\x9a\xf3\xfc\x6d\x3a\x98\x42\x18\xb1\x07\x19\xd0\x20\xea\x85\x97\x85": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x23\x69\x88\xba\xc6\xdd\x4a\x2a\xc7\xc0\x7d\xfa\xe9\xf0\x21\x7c\x98\xc7\xf0\x95\x7b\x81\xa0\xa5\x79\x87\x01\x55\x00\x4a\xd7\x28": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x23\x76\x74\xd8\x0b\xba\x74\xf5\xbc\x20\xab\xad\x6e\xcb\x4c\xc7\x4c\x16\x55\x82\x83\xde\x81\xc9\x7d\x6e\x8c\xc1\x74\x5b\xa8\x21": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.157, v0.0.158, v0.0.159, v0.0.160
    b"\x23\x7a\x31\x78\x93\x7d\xd7\x52\x76\xd4\xd3\x90\xc2\x48\xdd\x16\xce\x2a\x0e\xab\x77\x34\xc5\x80\xd0\x3d\x57\x6f\xc4\xcc\x1b\x2e": [
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.160",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x23\xa3\xf2\x6c\x0a\xa0\x6d\x56\x8a\xfa\x5d\x04\x41\x87\xa5\x18\xb4\x5c\x15\x00\xaa\x6b\x77\xf3\xae\x6c\x7a\x89\x3b\x41\xd0\x3b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC6093Upgradeable.sol",
            name="IERC1155ErrorsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192
    b"\x23\xc7\x61\x5f\xe6\x62\x1f\x90\x4d\xa9\x52\x07\x0f\x7a\xff\x3d\x28\x1b\x7c\x71\xd4\xc0\x1a\x6f\xaf\xef\xbd\xe9\x9d\xa1\x0a\x05": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.192",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x23\xdb\xb4\x28\xa6\x8b\xc2\x89\x22\x81\xd4\x12\xbc\x93\x93\xa1\x7e\x5c\x31\x35\x4b\xb9\x8d\xaa\x84\x5e\xd7\x1d\xa6\x54\x90\x03": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.72, v0.0.73, v0.0.74
    b"\x24\x26\x3c\xf2\x49\x6b\x87\x08\x71\xa1\xf1\x21\x7c\xd0\x4b\x6d\xe8\x0b\xe3\x83\x88\x55\x5d\xc9\xd8\x47\xc6\x15\x59\x7c\x79\xe1": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.74",
        ),
    ],
    # Hash appears in versions: main
    b"\x24\x44\x1a\xdb\x48\x60\x9f\xcd\xc0\xa0\x34\x7b\x64\xc6\xbe\x97\x2d\x5d\xbd\x6c\xb6\xc0\xd5\x3b\x8a\x61\x19\xf0\xe8\xa4\xb1\xe2": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/PositionManager.sol",
            name="PositionManager",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236
    b"\x24\x52\x19\xe3\x4c\xf3\xfc\xb7\x69\xcf\xad\x96\xc1\x09\xef\xd3\x2f\x00\xfa\xbe\x68\xbf\xb7\x02\xfd\xdc\x4e\x5b\x58\x36\xe1\xe7": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.236",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x24\x5f\xb3\x69\x3c\xf1\xa2\x99\x82\x1d\xb5\x4b\xca\xbd\x6f\x69\x2e\xdd\x4b\x58\x13\x5a\x29\xa1\x37\x58\x55\x2c\xd4\x30\xaa\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x24\x65\x97\x1b\xcf\x22\xa3\x60\x7a\xcd\x96\x59\x6e\xc3\xa9\x97\xeb\xe7\x54\x93\x1a\x7f\x8a\x9f\xa5\x54\xd4\x96\xa2\xd6\xf5\x1f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MessageHashUtilsUpgradeable.sol",
            name="MessageHashUtilsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x24\x74\x4f\x59\x90\xaa\x18\xd2\xe1\x16\xc8\x2f\x46\x20\xea\xc9\x00\x96\xef\x84\x43\xdc\x38\xa2\x34\x0c\xc2\x4e\x68\x08\xd8\x26": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.104, v0.0.105
    b"\x24\x85\x11\x34\x4d\x20\x33\x6e\x6b\x63\x5f\xa7\xd0\x0c\x46\xe4\x3d\xdd\xb3\xae\x2c\xe4\xc2\x24\x50\x30\x62\x2a\xe7\x57\x8f\x8e": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.105",
        ),
    ],
    # Hash appears in versions: v4.4.2, v4.5.0, v4.5.0-rc.0
    b"\x24\x8a\x6b\x9c\x45\x52\xea\x83\x8f\x9f\xae\x38\x19\x81\xe1\x72\x67\x69\xfa\x71\x8c\x94\x72\x24\xf2\x5b\x6c\x78\xa4\x90\xbf\x68": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x24\x9c\x3e\x76\xe1\x7e\xa9\x9a\x6e\xde\xdf\x86\x5a\x44\x83\x8f\x03\x37\x3d\x46\x11\x6a\xf7\x03\xa1\x0f\x71\x27\x52\x4c\xf7\x90": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.21, v0.0.22
    b"\x24\xad\xa6\x95\xf5\x50\xaa\xa7\x02\x9b\x6b\x75\x47\x56\x23\x73\x96\x76\x90\x44\x55\xa9\x6a\xa6\x03\x38\x47\xe3\x9b\x36\x36\x7c": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.22",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x24\xde\x57\x7d\x85\xac\xb5\xb5\x20\x49\xe8\x49\x79\xb5\x0b\xd7\xb2\xe0\x8b\x1d\x18\x2b\x2b\x9b\x78\xfc\xed\x86\xc5\x20\xde\xf8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x24\xeb\x18\xa8\xc8\x87\x06\x02\x76\x2e\x74\x55\x1f\x21\x60\xf0\x07\x2e\xf2\x9e\x4c\x6f\xbe\x00\xf8\x80\x44\xae\x87\xce\xf9\xf4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x25\x05\xd7\x51\x80\x18\x7e\xd4\x1d\x76\x9c\x80\xf8\xbe\xa6\x2a\xe3\x28\xf8\x4c\x37\x1b\x8b\xf6\x8b\xd6\xef\x07\x66\xed\x04\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x25\x14\x3e\x5d\x1f\x5d\x4e\x3f\x6c\x58\x50\xf1\x97\x7e\x90\x15\x33\x5d\xe7\x38\x0e\x52\x26\xe8\x37\x16\xe8\xf3\x35\x64\xf0\x6b": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x25\x2b\x88\x04\xe8\xd6\xef\xcc\xdd\x03\xac\xf2\x13\xa5\xa9\xb5\x27\xea\x27\x7e\x47\x62\x09\x70\x53\x2b\xc1\xf5\x3f\xbf\xc0\x69": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x25\x2b\xdf\x19\x67\xee\x51\x06\x94\xe4\x71\x74\xa4\xbf\x17\x4a\xe3\x87\xd9\x8c\x77\xa7\xc8\x48\xbb\xac\xe4\xef\x72\xe1\x84\x8f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0
    b"\x25\x40\x1c\x29\xf6\xd2\x64\x70\x7c\xc8\xfa\x92\x59\x3a\xed\x2f\xfe\x3e\x58\xb0\xeb\xa6\x4c\x6f\x1d\x65\x70\x62\xdf\x5e\xce\x02": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.7.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x25\x44\x32\x80\xbd\x8f\xf9\x1b\x8f\x14\x21\xab\xec\xf3\x10\xae\x42\xd0\xc3\xc3\xae\xd8\x94\x67\xdb\xb9\xe0\xa6\xfe\xb1\x5b\x1a": [
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.99",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x25\x55\xb7\xcf\xc1\x6b\xcd\xaf\x85\x20\x40\x36\xc1\x47\x5d\x64\x1f\xf6\xf1\x02\x3c\xf7\x00\x5e\x11\xe9\x1d\xf2\x83\x8d\x74\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x25\x9b\xf0\x72\xe9\x24\xcb\xed\xa0\xab\xc2\x64\x6f\x57\x81\x0b\x71\xf9\x01\x49\x13\x1d\x5a\x5b\x67\x65\x4b\x45\x60\xdb\xf4\x5e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC721Errors",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.72
    b"\x25\xa2\x09\x5f\x09\x11\x1e\xb5\xb2\x3f\x29\xd6\x8e\x57\x11\x2e\x66\x2c\x61\xcf\xe0\x2f\xa3\x17\xed\x4e\xb6\x01\xcf\xed\xb2\x34": [
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.72",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x25\xab\x97\x69\x62\xe5\x96\x9c\xb1\x4d\x42\x21\xef\x1a\xe7\x1a\xb0\x7e\xf5\x45\x61\xec\xf0\x41\x0a\x69\x24\x37\x2e\x5d\xd7\xc4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x25\xb8\x97\xc1\x58\xf8\x94\x89\x36\xf5\x98\x94\xdb\x53\x40\x4e\x07\x39\x56\x75\x1c\x52\x02\xa0\xbf\x82\xb6\xb5\x0a\xe5\xa1\x08": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/Pool.sol",
            name="Pool",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x25\xbc\xc3\x4d\x52\x20\x5d\xf2\x5d\x31\xc3\x67\x9d\xd1\xda\xb6\xe8\x6a\x6b\x76\x07\x17\x7e\xfb\x0b\x8b\x6c\x4e\x64\xc3\x9a\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: main
    b"\x25\xe8\xf4\xb1\x71\xee\xba\x77\xad\xb8\xd3\xa1\x19\xd6\x80\x6b\xc6\x4d\x8b\x68\x8e\x4a\xdd\x6c\x3a\x69\x11\x3d\xaa\xcb\xf0\x34": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/PositionConfig.sol",
            name="PositionConfigLibrary",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x26\x08\xe9\xa8\xe3\xc4\x7b\x82\x11\x7b\x84\xa0\x21\x8d\xef\x14\x93\xcc\x33\xa2\x4f\x22\x7e\xb7\x44\xe6\x72\x81\xda\x46\xde\xbc": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/ProtocolFeeLibrary.sol",
            name="ProtocolFeeLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x26\x0c\xf7\x5c\x13\x49\xfd\x31\xa7\x1f\xf1\xb2\x24\x2d\x61\x3b\x00\xfa\xa8\xa9\x14\x24\x38\x5e\xaa\xea\x12\x34\xff\x38\x23\xcb": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.8.3
    b"\x26\x39\xa1\x08\xc2\x14\x60\x13\xd4\x7c\x6d\xf6\x17\x73\xd7\x67\x37\x7e\x9c\x4e\xa2\xee\x2f\x8b\x50\x9c\x28\xd2\xec\xf8\x34\xd8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172
    b"\x26\x39\xfd\xfd\x7f\xf7\xd6\xf1\x6d\xe4\x0d\x75\xa8\xe3\xa9\x60\xad\x3f\x9f\x19\x97\xa7\x82\xf9\x74\x3f\x08\x4f\xa1\x12\xa4\xa1": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.172",
        ),
    ],
    # Hash appears in versions: v0.0.134
    b"\x26\x41\xf2\x27\xeb\x2d\xa1\x85\x39\xd5\x5c\xd5\x17\x93\x7a\x18\x3a\xc9\xe5\x39\xa5\xf9\xef\x5c\x1e\x3c\x4d\x5a\x74\xc3\x45\x4c": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.134",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x26\x54\x26\xe2\xe4\x14\xa9\x54\x0c\xd5\x3e\xae\xf0\x0e\xcc\x42\x28\x8b\x31\xae\xeb\xe5\xc6\xdd\xed\x36\x67\xb1\xbe\x49\xc8\xf0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.0-rc.0, v4.4.0-rc.1, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x26\x5c\xe1\x1a\x63\x0f\x2c\xca\xd4\x92\xa4\xb1\xad\x9e\xff\xf1\xd4\xfd\xe4\x08\xd6\x8a\xaf\x12\xd0\x69\x57\x73\xda\xa9\x3e\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x26\x85\xbd\x04\xe3\x12\x97\x35\x4f\x94\x22\x83\xd1\xaa\x1c\xfa\x8d\xac\xf4\x2d\x49\x28\x85\x97\x13\xbd\x93\x9e\xbe\xd2\x57\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x26\xab\x4c\x67\x36\x4c\x34\x14\x50\xd0\x62\xdf\x2c\xbb\x16\x3f\xd2\xac\xbf\x3e\x93\xe2\x21\x5e\x80\xbd\x44\xeb\x6a\x42\xab\x07": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/BlockTimestamp.sol",
            name="BlockTimestamp",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/BlockTimestamp.sol",
            name="BlockTimestamp",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/BlockTimestamp.sol",
            name="BlockTimestamp",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/BlockTimestamp.sol",
            name="BlockTimestamp",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/BlockTimestamp.sol",
            name="BlockTimestamp",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x26\xd7\xbd\x15\x71\x77\xe4\x1d\x20\x52\x12\x19\x18\x19\x89\xff\x75\x31\x10\xfb\x80\x0a\xed\xd5\xc0\x3e\x6c\xc3\x8d\x64\xcd\xe2": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x26\xe3\x24\x2e\x86\x6a\xea\xe8\xc5\x7b\x33\x87\x0a\x94\xd1\x91\xb3\x67\x37\xae\x55\x55\xbf\x67\xe8\x26\xeb\x4f\xc3\xff\x8f\xac": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0-rc.0
    b"\x26\xe9\x0c\xf3\xd8\x65\xff\x11\x08\x1c\x06\x82\xd6\x28\xa1\x0c\x00\x8d\x00\xe8\x27\x54\x22\x8f\x5f\x9b\x77\x8a\x0c\xf7\x74\xe8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296
    b"\x27\x00\xdd\x5b\x44\x32\x4a\x6d\xa7\x2d\x53\x10\xfc\x84\x1c\x1a\x54\x69\x72\x31\xf2\x47\x86\x6a\xb1\xb4\x26\x13\xb8\xe3\xfe\x57": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.296",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0
    b"\x27\x2e\xf2\xa3\x35\xdb\x30\xc5\xf6\xcc\xcb\xa8\x1a\x80\x5a\xe8\x5e\x61\xa8\x89\x4c\x9e\x38\x50\xef\x24\xf0\x99\xa8\x29\xb8\xfa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.2.0, v1.3.0
    b"\x27\x58\x01\xe9\x23\x5f\xea\x08\x03\xb2\x95\x29\xe7\x32\x3a\x21\x9c\xef\x97\x28\x47\x34\xb2\x2d\x22\xb8\x25\x62\x0e\x13\x33\x96": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionValue.sol",
            name="PositionValue",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PositionValue.sol",
            name="PositionValue",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207
    b"\x27\x6a\x36\x55\x18\xb2\xd2\x8d\x3f\x49\x02\x0c\x35\x73\xe6\x57\x5c\xd3\x52\x89\x8b\xe1\x5a\xab\x7a\x0d\x8e\xdf\x61\x84\xa1\xea": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.207",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x27\x7d\x47\xc1\x61\xca\xa2\xe2\xb5\x4c\xe7\xd0\xfc\x7f\x94\x72\xfa\x58\x2c\x24\xf9\x4e\x2b\x31\x28\x08\xd5\x9c\xb0\x7e\xe5\xd7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x27\x7f\x1a\xb4\x79\x4e\xf1\x24\x05\xba\xd2\xaa\x70\xe3\x1d\x98\x7c\xf8\xb6\x24\x95\x7d\x67\x41\xd1\xa1\x3d\x59\x43\xe5\x2d\xd7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="OwnableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x27\x8c\xb6\xc8\x87\x3b\xd9\xfb\xed\xff\xf8\x6c\xa4\xe7\x83\x7f\x3e\x5e\xc1\x78\x5d\xed\x0c\x23\xa4\xc8\x2a\x72\x82\xd2\x75\x17": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x27\xe2\x46\xdd\xb1\x25\xcd\xa8\xd6\x97\x06\x39\x31\x1f\x05\x1e\x07\x85\x02\x47\xa0\x62\x90\x18\x68\xeb\x2d\x17\x2d\xeb\x6c\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.172, v0.0.173, v0.0.174, v0.0.175
    b"\x27\xee\xc4\x30\xf1\x32\x5d\xcb\x60\xd7\x2a\x82\x44\x9f\xcd\x2b\x0c\x74\x51\xe9\x02\x82\x63\x24\x56\x7e\xf6\x4d\x58\x8b\xd3\x21": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.175",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2
    b"\x28\x5d\x15\x9c\x5c\xc0\xce\x42\xf7\x70\xe0\x82\xa8\xf4\x9b\x21\x6a\xc6\xb8\x06\x31\xa2\x91\x21\xaa\xf5\xe7\x68\x7f\xab\xff\x22": [
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.6.2",
        ),
    ],
    # Hash appears in versions: v0.0.152
    b"\x28\x6a\x1b\x36\xed\xfd\xe5\xd4\xea\x24\x57\xb1\x55\x6f\x6b\x8b\x11\x90\x17\x23\x9f\x92\xbe\x00\x44\x77\x8a\x32\x05\xec\xfc\x9e": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.152",
        ),
    ],
    # Hash appears in versions: v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273
    b"\x28\x6f\x07\x3e\xde\x13\x3e\x9e\xcc\x1c\x66\x10\xb3\xd4\xeb\x60\xf5\x23\xe0\x5e\xc0\x30\x1f\x93\x2a\xf5\x41\x0a\x74\x44\x12\x15": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x28\x7c\xd0\x09\x54\x12\x08\x64\x6a\xb1\x89\xde\x09\x74\x03\xc7\x8c\xa7\x8d\x54\xfe\x46\xc3\xfe\x50\x7d\xcb\xd4\xe1\x88\x69\x02": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.133
    b"\x28\x8f\x1f\xc0\xcb\x50\xe8\xf8\xb7\x98\x5c\xe3\x37\x7e\xe5\x89\x62\x18\x40\x32\x0f\x9b\x78\xac\x10\x43\x5e\x54\xc8\x3c\x62\xd4": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.133",
        ),
    ],
    # Hash appears in versions: main
    b"\x28\x8f\x4e\x96\xdd\xe2\x87\xd9\xff\x4d\x58\x45\x42\x92\x64\xed\x8f\xfa\x80\x2e\xd4\x25\xc0\x2e\x10\x08\x23\x0b\xcc\x27\x47\xa1": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IPoolInitializer_v4.sol",
            name="IPoolInitializer_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7-2
    b"\x28\x93\x1b\xad\xf4\xb7\xb5\x11\xff\xa4\xcf\x50\x8d\x98\x1f\x32\x9e\xa2\x9a\x60\xd7\x25\x69\xb1\xad\x29\xcd\x4f\x0f\x12\x7b\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.242
    b"\x28\x93\x1d\xbc\x68\xd1\xb8\xe6\xf0\x4b\x7d\xcf\x8b\xea\x4b\xcf\x61\x20\xf9\x53\x4f\x4e\xb9\x84\x06\xea\x78\x16\xaf\x9d\xe4\xec": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.242",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x28\xae\xdc\xe5\x24\xcd\x61\xe9\x9a\x4b\x95\xc7\xb8\xb2\xbb\xb1\xe5\x3f\xa9\xaa\xf1\x27\xbc\xe8\xfb\x7e\x9e\x99\x9d\xd9\x5a\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x28\xb9\x4d\xb1\x93\xc6\x54\xe9\xac\xaf\x90\x30\x37\x0d\x30\x1d\x65\xc6\xca\xc8\xab\x31\x49\xd6\x4f\x55\x6d\x47\xd0\xb5\xa6\xcb": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IUniswapV4DeployerCompetition.sol",
            name="IUniswapV4DeployerCompetition",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x28\xbe\x29\x83\x49\xe0\x90\xbe\x08\x03\x71\x3d\x50\xd5\x5f\x38\xd7\x28\x3a\xa0\x79\x49\x38\x56\x0b\xf5\xd0\xc1\xeb\xc9\xbc\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x28\xc1\xfc\xc4\x28\x6e\xbd\x68\xe1\xa5\xd5\x6d\x07\xb1\x9a\x36\xb9\x3b\x6a\x80\xad\x19\x95\x87\xdd\xe7\x46\xbb\x2d\x43\x25\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x28\xfa\xe4\x01\x90\x9f\x06\x8e\x51\xa3\xff\x77\xa1\x9b\x01\xdf\x7e\xd2\xa0\x77\x6e\x4b\xec\x27\x68\x53\x5a\x5d\x42\x7f\x8e\x62": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesKeyedUpgradeable.sol",
            name="NoncesKeyedUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesKeyedUpgradeable.sol",
            name="NoncesKeyedUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesKeyedUpgradeable.sol",
            name="NoncesKeyedUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesKeyedUpgradeable.sol",
            name="NoncesKeyedUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesKeyedUpgradeable.sol",
            name="NoncesKeyedUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.266
    b"\x29\x19\x18\x53\xfb\x6a\x9c\x75\xc4\xf2\x11\xcc\x1d\x4d\x4d\xbe\xa4\x60\x3e\x99\x95\xa2\xff\xe7\x3e\x70\xe5\x04\x8f\xf3\xb5\x06": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.266",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x29\x25\xa2\x25\x8b\x8b\xd2\x9a\x2c\x8c\xce\x1a\xf8\x80\x06\x18\x51\x90\xfd\x6b\x62\x00\xb0\x14\x4d\xd1\x14\x39\x24\xdb\x0f\x8f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73
    b"\x29\x30\xea\xd2\xe9\x9a\xbb\x22\x0a\x17\x5d\x5d\x55\x7b\x6b\xfa\x4f\xc0\xc9\xb3\xbb\xc2\x52\xf2\x4c\x76\xc9\x91\x03\x12\x75\x60": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.73",
        ),
    ],
    # Hash appears in versions: v4.7.3
    b"\x29\x37\x10\x82\x51\x6d\xa3\x51\xbd\x5f\x40\x80\x78\x02\x30\x94\xc4\x45\x38\x67\x66\xb5\xd3\x4d\x19\x92\xf2\x75\x25\x26\xcf\xd1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\x29\x40\x6e\xa7\xbb\x79\x66\x73\x36\xd6\x76\xc0\xc9\xc9\x3c\xb0\xc0\xf2\x99\x7f\xbd\xb7\x74\x17\xeb\x46\xd2\xcd\x00\xb1\xb8\x4f": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x29\x7e\xe5\x10\xd1\x0d\x61\xeb\xd8\xe0\x57\xd4\x5c\x7e\xba\xb9\x0a\x4a\xfc\x06\x78\xb2\x38\x25\x43\x7c\x3b\xf5\x96\x7c\x8b\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSequentialProposalId.sol",
            name="GovernorSequentialProposalId",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSequentialProposalId.sol",
            name="GovernorSequentialProposalId",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x29\xa2\x64\xda\x19\x2a\xc3\xc1\x5e\xee\x2c\x30\xac\xb5\x17\xb1\x4e\xa1\x8e\xd7\x7d\x3e\x59\xc5\x41\x94\x21\xba\xa7\x71\x31\xbb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="PausableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x29\xa2\xe6\x4f\x86\xf6\xed\xb0\x6a\x70\x2a\xf8\x44\x7f\x22\x53\x11\x9e\xbd\x76\x8d\x72\x2c\x97\xbf\x98\xac\x85\xe7\x80\xf7\xe3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x29\xe6\xd1\xf9\x34\x4e\x54\x0e\x4e\xa1\x04\x07\x8d\xf5\xfb\x56\x04\xef\xde\x57\x80\x6a\x09\xbc\x80\x59\xad\x65\xc7\x3f\x20\x04": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1
    b"\x29\xe9\x30\x30\xe1\x30\x86\xb6\x7c\x45\xe5\xaa\x95\xcc\xa2\xb4\x22\xea\x76\x00\x3e\x65\x06\x4a\x0f\xfc\x88\x33\x2f\x5d\x63\xcd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/Strings.sol",
            name="Strings",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/Strings.sol",
            name="Strings",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87
    b"\x29\xeb\x72\xfd\xd8\x41\xc9\xed\x66\x10\xf6\x2a\xe9\xb5\x3a\xba\xb7\x10\xdd\xe0\x94\xa5\x96\x73\xe3\xb4\xbc\x66\xab\xf5\x0a\xf0": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.87",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x29\xf0\x53\x69\x27\x31\xe8\xf9\x1c\x9b\xe3\x35\xb5\xf8\xdc\x95\x9e\xc0\x86\x55\xfb\xfa\x33\x8c\x50\xb1\xea\xfd\x1b\x75\x78\x66": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2
    b"\x29\xf2\xf3\xd9\x87\x09\xb0\xf6\x4b\x27\xdc\xab\xc2\x29\x5d\x3c\x55\xa2\xed\x0b\x8f\x5b\xb9\x31\x97\x0f\xa7\x46\x24\x48\x72\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x29\xfc\x3b\x8d\x9b\x46\x7c\x3c\x9d\xd4\x51\x31\xc6\x5a\x67\xe9\x5a\x82\xdb\xfc\xba\x67\xb1\x8b\xc5\xec\xe8\xbc\x7e\x0e\xee\x91": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x2a\x13\x6f\x21\x7d\x8b\x27\x84\x13\xda\x72\x1b\x10\x03\x8c\x9b\xb2\xb5\x0c\xf9\x93\xf4\x59\x72\xde\xb3\x78\xe3\x24\xb9\xe3\x00": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/compound/ICompoundTimelockUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166
    b"\x2a\x18\x80\xf8\x41\x91\xdb\xd4\x2f\xac\xe4\xf1\x05\x46\x2b\x5d\xbb\x03\xe4\x1d\xbd\xb0\x00\x91\x25\x45\x7f\xa3\xd2\x8b\x44\xd5": [
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.166",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x2a\x1c\x12\x98\x85\xe4\x8d\x82\x50\x08\x61\xa4\x86\x65\x83\x7e\x80\xd3\xd3\x55\x1b\x24\xfb\xaa\x09\x3b\x4f\xd9\x74\x70\xe1\x38": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/CallbackValidation.sol",
            name="CallbackValidation",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/CallbackValidation.sol",
            name="CallbackValidation",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/CallbackValidation.sol",
            name="CallbackValidation",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/CallbackValidation.sol",
            name="CallbackValidation",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/CallbackValidation.sol",
            name="CallbackValidation",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x2a\x21\x27\x93\x44\xa1\xda\x32\x0f\xe1\x41\x44\x3f\x05\x72\xe6\x93\x1e\x04\xd0\x47\xc8\xfe\x57\xfe\xc5\x94\x63\x28\x50\xbf\x6c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x2a\x3d\x9a\x64\x50\x00\xa8\x4d\xe5\xac\xb5\x2e\x47\xd9\xc7\xb8\xef\xf0\xa1\x8e\x7d\xfd\x9f\x91\x68\x05\xd0\x16\x86\x96\xc7\xe5": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x2a\x50\xe8\x36\x4c\x51\x81\xc0\xd7\xd6\x1f\xb4\x4e\xd6\x65\xe2\x8f\x2d\x45\x38\xab\x64\xd2\xc2\x87\x1b\xef\xe6\x4b\x0c\x8b\x8b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x2a\x5c\x22\xb6\x66\xe3\xd3\x73\x95\xbf\xb0\x8a\x40\xe1\x5b\x8d\x3a\x7c\x8b\xe6\x8f\x2b\x53\x03\xf4\x5f\x8c\xe7\x73\x53\xbc\x45": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967BeaconProxy.sol",
            name="ERC1967BeaconProxy",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.219
    b"\x2a\x61\xd7\x50\xee\x4c\x5c\x19\xa9\x41\x0b\xab\x31\xc7\xb7\x23\x06\xc9\x4e\xb9\x49\xc8\x46\x77\x1a\x71\x12\xfb\x17\x46\x53\x1e": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.219",
        ),
    ],
    # Hash appears in versions: v4.8.2, v4.8.3
    b"\x2a\x63\x9f\x30\xeb\x06\xcb\xd6\x50\xb2\x7c\x87\x04\xc9\x23\x92\xa3\xe1\x03\x2e\x6c\x27\x07\x86\xb2\x96\x37\x6b\x16\xf1\xa4\xc6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.1
    b"\x2a\x92\x46\xf3\xad\xef\x1a\x2a\xef\x6f\x20\x7f\xec\x8d\xf6\x23\xec\x2a\x0f\x5e\x91\xb3\x88\x60\xb8\x7b\x65\x20\xe3\x2b\x8c\x3e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.1-solc-0.7, v3.2.1-solc-0.7
    b"\x2a\x94\x98\xbb\xb1\x6f\x47\x9d\xb5\x55\x03\xea\x7b\xa0\x8d\xc2\x62\xee\x28\x80\xad\x44\xc9\xdc\xc0\xea\x1d\xe5\xe9\xb3\x63\xf1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\x2a\x95\xe2\x17\x0c\x1b\x81\x0b\x57\x3b\xbd\xfc\x59\x2d\x02\x87\x7b\x47\x7d\x37\x76\x0a\x48\xac\xe9\xd3\x93\x03\xb0\x4e\xcf\x67": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237
    b"\x2a\xa9\x11\x69\xf9\x2d\xa1\xe1\x51\x4a\x13\xb0\xa2\xa6\xab\x7d\xfc\xd2\x3b\x5e\x53\x09\x73\x1a\x92\x7e\xfe\x11\xc1\x63\x8e\xfa": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.237",
        ),
    ],
    # Hash appears in versions: v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x2a\xb2\xc8\xb2\x13\x1a\x2b\xfb\xc6\x24\xea\x88\x94\xd0\x58\xc5\x8b\x21\x63\x83\x36\xb8\x35\x2f\xdc\x45\xd4\x59\xb3\x6a\x3e\x90": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x2a\xb5\x28\xf9\xee\x81\x3c\xac\xc0\xe6\x7b\xf3\x58\x26\xdd\x58\x1a\x05\x33\x66\x8a\x69\x1b\xf0\x8c\xb7\x11\xaf\xcb\x2c\x8e\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0
    b"\x2a\xba\xcb\x7d\xf1\x10\xb7\xb6\x74\xdb\xaf\x94\x9d\xe9\x11\x31\x15\xbd\x00\xaa\xbe\xbe\x71\x65\xa6\x51\x4b\xcc\x67\x93\xbe\x91": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.261, v0.0.262, v0.0.263, v0.0.264
    b"\x2a\xd2\xec\xa9\x46\x3a\x0a\x83\x90\xca\x99\x93\x45\x16\xf8\xb5\xc5\x66\x94\xc2\xd9\x07\x36\xff\x14\x2a\x8d\xe2\xcb\xdd\x1e\xed": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.264",
        ),
    ],
    # Hash appears in versions: v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256
    b"\x2a\xdc\xc9\x3c\x64\xc4\xa5\x9c\xde\x46\x51\xff\x97\xc2\xe1\x45\xd6\x92\xd0\x63\xa3\x8e\xe2\x81\x92\x6a\x3c\xa7\x6b\x73\x87\xd9": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.256",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x2a\xfe\x8e\xf6\x9c\x26\xd5\x30\x68\x22\x39\x70\xb1\xc3\xab\x85\xaa\x1a\xcb\x0f\xab\x4d\x4d\xc8\x6a\x88\xf3\x9f\x46\xb7\x01\xa4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x2b\x16\xd6\x4b\x18\xb1\x05\x69\xc4\x6a\x35\x20\x8e\x9c\xaa\x69\x60\xe6\xc9\x3b\x18\x13\x67\x67\x86\x6d\xbe\x78\xbc\xb4\x93\xe8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x2b\x3b\xf8\x8a\x64\x79\xbc\x8e\x3b\x5e\x14\xad\x89\x03\x19\xef\x4c\x92\x1c\xf1\x7c\x48\x65\x56\xfb\x72\xa0\xeb\x22\x39\x4f\xf4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x2b\x96\x3a\x55\xcf\x76\xaf\xef\x5e\xaa\x58\x4a\xa7\xc3\x6c\xca\x04\xb8\x80\x1b\xf6\x36\x2c\x52\xb7\x7f\x68\x92\xb0\xb2\x8b\x89": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x2b\xba\x7d\x4f\x79\xe4\x5e\x7f\x84\x93\x4c\xc5\x2d\x5e\x19\x66\x23\xc6\xca\x6d\x96\x46\xec\xab\x8d\x96\x03\x61\x00\x51\x7a\x3f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\x2b\xcc\x9d\x1f\xfe\x53\x04\xec\x76\xec\x21\x3d\x69\x22\x28\xbb\x64\x5c\x32\xce\xa0\x6d\x8a\xdb\x2b\x81\x17\x1b\x53\x1c\xd6\xed": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.3.0, v5.3.0-rc.0
    b"\x2b\xe2\xc7\x6f\xdb\x3f\x2e\x92\x17\xda\x23\x8e\xb1\x87\x1a\xc8\x7e\xbe\x36\x69\x10\x42\x71\xb5\x2a\x80\x21\xe3\xf9\x29\x99\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Execution",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Execution",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Execution",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x2c\x12\x6f\x83\xc3\xf3\xdb\x68\xcd\x78\x79\x2d\xff\x63\xc6\xad\xfa\x47\x42\x3b\xcc\x9e\x76\xca\x4d\xf7\xf6\xd3\x5e\xbe\x99\x5a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.248, v0.0.249
    b"\x2c\x1d\xd8\x47\x07\xa1\xa6\xb8\x9e\xea\x91\x5e\x33\x9c\x1c\xb2\xc8\x05\xd1\x83\xa8\xfb\xe8\x68\x14\xf0\x66\x78\xa4\xde\xd2\xec": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.249",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\x2c\x23\xa5\x6a\x41\x91\x68\xd2\x02\xd0\x23\x37\x2f\xb4\x53\xcd\xd7\x2b\x55\x17\x35\x7f\x0c\x82\x76\xff\x1f\xdf\x12\xca\x49\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x2c\x28\x7e\x8f\x3e\x6b\x4c\x21\xd3\x0b\x17\xf2\xc3\x8e\x60\x81\x72\xd9\x6a\xcf\x6f\x14\xf5\x19\xd0\xd6\x4a\x33\xe9\xf7\x01\xed": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x2c\x48\x43\x24\x18\x3d\xcb\x3c\x84\xbc\x11\x43\x58\x1f\xd4\x3d\x7f\x5f\x40\xbb\x36\xce\x2b\x72\xa2\x30\x35\xd1\xa1\xaa\x24\x3c": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/lens/StateView.sol",
            name="StateView",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98
    b"\x2c\x5a\xb3\x4c\x55\x2c\x01\x0d\x8e\x11\x03\x5c\x61\x90\x2c\x54\x61\x02\x14\x8b\x40\xf7\x7b\xec\x29\xcd\x52\x1b\xde\x82\x66\x14": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.98",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x2c\x5d\x03\x56\xfe\x78\x72\xe9\x15\xd1\xfd\x37\xf6\x43\x7d\x5d\x57\x73\x5c\x2b\x95\x37\x2d\xe6\x8f\x40\xdd\x66\x38\x1e\xee\x2f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280
    b"\x2c\x77\x37\x8b\xb3\x91\xfb\x60\xd1\x8a\x64\x56\xb4\x7d\xfb\x42\xc7\x70\x29\xb6\x3d\x74\xe9\x71\xf9\xb5\xe9\xd1\xcf\x6e\x60\x25": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.280",
        ),
    ],
    # Hash appears in versions: v4.3.3, v4.4.0
    b"\x2c\x7f\xc8\x9c\x67\x65\x27\xc7\xff\x05\x59\x5f\xec\x68\x13\x75\x67\xcf\x91\x18\x53\xc5\x4b\x4c\xc1\x9a\x8b\x72\xc3\x15\xe1\x1a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x2c\x8d\x7b\x98\x37\x99\xde\x26\xe1\x39\x2d\xe1\x59\xaa\x90\x56\x54\xdd\x4b\x71\x5f\x3c\x1d\xad\x0a\x64\x9f\x57\xfd\xe8\x4d\xc5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x2c\x90\x9a\xe8\xb6\x0a\x01\xa0\x6d\x45\xe2\x8f\x58\xcf\x4f\xec\xb7\xa1\xe3\x46\x7d\xf9\x3a\xf2\xae\x85\xe0\x8b\x06\x30\x18\xca": [
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/ithaca/BLS.sol",
            name="BLS",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: main
    b"\x2c\xb3\x04\x8d\xcc\xd1\x17\x59\x82\x00\x39\x64\xe9\xc3\xca\x07\xfc\xac\x8b\xce\x2a\x25\x96\xb1\xea\x1f\x3d\x21\x08\xa3\x4b\x78": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IStateView.sol",
            name="IStateView",
            version="main",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x2c\xfd\x44\xbe\x82\x36\x30\x9a\xb5\x5f\xe5\x55\xab\x82\x40\xd1\x1a\xae\x82\xae\xcc\x60\x3a\x9b\x36\xc5\xb1\x67\x50\x8b\xcd\x61": [
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x2d\x08\x00\x8e\x73\x05\x94\x1a\x69\x8b\xae\xa1\xa1\xf5\x35\x68\x9a\xb7\x37\xb0\x3c\xba\x33\x52\xc0\x69\x56\x9e\x1a\x80\x81\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x2d\x23\x63\x3e\x17\x01\x99\xe4\x02\x7e\x26\x9a\xd6\x80\x4e\x6f\x78\xdb\x96\xca\xff\xc8\x75\x14\xee\x1a\x42\x7f\x25\xc1\x02\xd9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\x2d\x39\xa1\xc5\x0d\x81\x63\xb7\x06\x29\x1f\xf2\xbc\x50\x14\xac\x40\xe8\x40\x89\x2c\x18\xa4\x8a\xf6\x4a\x77\x67\xfb\x0b\x8f\x9f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7
    b"\x2d\x48\xac\x48\x45\x52\x16\x32\xaa\x28\xc0\x2f\xff\x87\xf5\x19\x1f\xde\x58\xa5\xbd\x9d\x9e\xc6\x81\xb7\x69\xba\x2e\x33\x13\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.9.4
    b"\x2d\x56\xfb\xa2\x89\x43\x8a\x7a\x61\x8d\x2e\xb1\x9e\x56\x51\x30\x36\xce\x59\xeb\x26\x0a\xac\x24\xe8\x6d\x2f\x29\x7e\x0d\x36\x7b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.4",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x2d\x5e\x00\xc0\x27\xd8\x96\x62\x56\x8a\xf0\x20\x45\x13\xa4\xfe\x25\x37\xb5\xc8\x28\xf1\xa0\x35\x71\xa4\xc1\x49\x05\xa4\xed\x60": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.1.0-rc.0
    b"\x2d\x75\x8b\x5b\xaa\xfb\xf3\xb2\x6e\x38\x47\xaa\x6c\xb2\x29\x1c\x09\x9f\x40\xb8\x81\x53\x02\xa4\xde\x12\x94\x7b\x63\xc9\xde\xd9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.3
    b"\x2d\x99\xf6\x70\x09\x78\x82\x26\x2a\x01\xba\x05\xf9\x76\x3f\x9d\xff\xbb\x53\x1b\x0b\x46\x72\x64\x06\x81\x74\xa5\x46\x1e\x7e\x3b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x2d\xe6\x65\x03\xae\x09\x46\x6c\x40\xd6\x8b\x96\x4c\x9a\x5a\xf4\x99\x1c\x94\x98\x5c\x59\xa2\x7c\x5b\x16\x52\x19\xbe\x7c\x74\x4b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/ISubscriber.sol",
            name="ISubscriber",
            version="main",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x2d\xe6\xd4\x41\xd6\xc7\x3f\x28\x97\x28\x22\x6b\x5b\x20\xa2\xa3\x03\x79\xe6\xdf\x96\x6e\xc5\x3b\xa0\xd2\x50\xb3\xbc\x03\xf7\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x2d\xea\x55\xcf\xdd\x6a\x06\xa5\xb6\xb2\xdb\x7b\x31\xb0\xc5\xd3\xe6\x34\x1f\xe2\x9c\xf6\x9d\xbe\x69\x07\xc1\x2a\xd0\x21\xb5\xc1": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/TransientStateLibrary.sol",
            name="TransientStateLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-beta.1
    b"\x2e\x13\x79\x04\x57\xb8\xa2\xc5\xd6\xe7\x96\x7f\xdd\x68\xa6\x8f\x62\x61\x14\xfa\x6c\x4e\xc2\xf4\xc3\x48\xa2\x3e\x77\xbd\x69\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x2e\x17\x42\xa6\x42\xfe\xcf\x51\x57\x01\x94\x74\x0d\xe6\x82\xd4\x9e\xdf\x85\x14\xb0\x18\x47\x19\xf1\x85\x8d\xaa\xc7\x5f\x8a\x46": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124
    b"\x2e\x5d\xe6\x6e\xf1\xbd\x46\x9a\x91\xf4\x3a\x6a\xed\xea\xdf\x36\xb7\x56\x4c\x84\xd4\xee\xeb\x43\x86\xdc\x36\xc2\x4d\x6e\x27\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x2e\x74\x97\x28\x7a\x7f\x13\x69\xe4\xbf\xcf\x5e\xb2\x8f\x33\x24\x3d\x4c\x6a\x12\xc5\x1d\x83\x16\x01\xec\x1f\x78\xb1\x29\xdd\xa9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165
    b"\x2e\x78\x5a\x2f\xd1\x6a\x62\x3f\xf0\x48\xc8\x78\xec\xca\x4b\xde\x7d\xe0\x09\x4a\x93\x01\xec\xd3\xad\xb4\x45\xc2\xa3\x04\x23\x48": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.165",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x2e\xde\x31\x48\x2b\x98\x69\xdc\x38\x0d\x21\xae\xf2\x87\xb4\x2a\x86\x03\xd3\xa0\x50\x7f\x1d\x95\xca\xfd\x88\x7c\x51\xf9\x1e\xcf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x2f\x2f\x9b\x71\xcd\x9a\xd9\x2a\x70\xce\x11\xcb\x4c\x5d\xea\x73\x42\x45\xf3\x90\xa4\x39\xa2\x4d\x76\x7d\x3c\x9b\x66\xd8\xbb\x3f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x2f\x44\x1e\x44\x4b\xf9\x53\x9c\x6a\x93\x08\x10\x85\x5c\x4c\x9e\xa5\x54\xa9\x47\x85\xe0\x9c\x92\x46\xe4\xc8\x20\x1a\xd1\x5c\xa8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x2f\x54\x80\x9e\x57\x95\x36\xa5\x96\x3a\xda\x8f\xff\xde\x3c\x5b\x51\xac\xe3\x30\x8c\x39\x05\x38\x02\x4b\x4d\xb6\x1e\x12\xa7\x84": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTDescriptor.sol",
            name="NFTDescriptor",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTDescriptor.sol",
            name="NFTDescriptor",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTDescriptor.sol",
            name="NFTDescriptor",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTDescriptor.sol",
            name="NFTDescriptor",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTDescriptor.sol",
            name="NFTDescriptor",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\x2f\x65\x17\x26\x0f\x53\xe9\xc8\xe4\xa9\xbc\x47\xa9\x98\xe5\x13\x4b\x8b\xc0\xfa\x8b\x47\x29\xa1\xbb\x90\xf9\x59\x2d\x13\xb5\x9a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x2f\x6f\xc9\xc8\x2b\xc3\x99\x90\xdc\x47\xed\xf3\xda\x70\x28\x2a\x05\x72\x91\x7f\x49\xa0\x64\x60\xad\xb9\x77\xb4\x68\xd0\xff\x16": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x2f\x87\x99\xe0\x4e\xe4\x8a\x4b\x9a\x1f\x01\x7d\x1b\x7f\x3d\x05\x69\xd6\xd5\xf7\x62\x4b\x8f\x71\xae\x0a\xf4\x25\x5d\xaa\x72\x79": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrowUpgradeable.sol",
            name="ConditionalEscrowUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x2f\x8e\xda\x1e\x35\x86\x9c\x2e\xfc\x04\xac\x6b\x99\x23\xfa\x73\x07\x41\xd5\x34\xb6\x0e\x47\xb2\xc2\x28\xcb\xa0\x2b\xdd\x2a\xa6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x2f\xa4\xde\xd6\x25\x5d\xb8\x61\xe8\xf3\x3e\x3c\xfb\x63\xf7\x30\xdd\xa2\x82\x27\x17\x26\x00\xd2\xac\x2f\x44\xc8\x74\xc2\xce\x8e": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/IUniswapV3Factory.sol",
            name="IUniswapV3Factory",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x2f\xb9\xfb\x01\x35\x23\x4f\xb7\xf0\x64\x84\x35\x42\xb5\x28\xbb\x53\x12\x09\x82\xb4\xea\x98\x1b\x43\xe4\x31\x4c\xe9\x8a\x47\x26": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x2f\xe7\x3b\x48\xc5\xa1\x07\x24\x7d\xc6\x9c\x82\x58\x2b\x4e\xb8\x94\xb1\x9e\x9c\x0b\x3d\x77\xc2\x29\x82\x2d\x8a\xc2\x8d\x3c\x8a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x30\x16\xd7\xfa\x1b\xec\xcc\x9e\x28\xf2\x70\xda\x38\xb7\xa4\xd9\x98\xcb\xca\xb1\x49\x4b\xaa\xb2\xdb\x3b\x3c\xad\x22\x29\x01\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295
    b"\x30\x1e\x22\x3a\x63\xa4\xf1\x3a\x1b\x18\xc8\x06\x10\xb8\x8c\x10\x71\x0b\xcb\x21\x38\x97\x0b\x9b\x71\x0f\x89\x89\x94\x80\x00\x1f": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.295",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x30\x38\xeb\x4e\x83\x87\x77\x6c\x6c\x97\xd3\x08\x3f\xb0\x07\xe9\xdf\x92\xda\x94\x54\xdf\xf2\xc5\xa4\x21\x09\x2b\xf7\xd1\x36\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x30\x41\x9b\x95\xc0\x53\x09\x9c\x5a\xdd\x49\xab\xe7\x8c\xdf\x2f\xc4\x78\x41\x8d\x06\x57\xf6\xd6\x33\x0d\xce\x86\x6c\xd8\x60\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x30\x61\xe8\xfa\x7b\x27\x5e\xf5\xbc\x49\xd2\x50\x5a\x61\x34\xc1\x2a\x7f\x4b\xec\x9b\xf9\x1d\x6b\x74\x38\xea\x79\x9b\x26\x96\x2f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.5, v0.0.6
    b"\x30\x78\xd1\xe3\x5d\x43\x3e\xa1\x2e\x21\xd8\xd0\xe1\xf8\xf7\x68\x3b\x6b\x4b\x0f\x5e\x0a\x15\x19\x27\x8c\x40\x26\x57\x7d\xdb\x5b": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x30\xc6\x6b\xe1\x5f\x58\x59\x6b\xdc\x02\x09\xc0\xd8\xe2\x4b\xb9\x7a\x4a\xc9\x1b\x35\xef\x77\xf7\x93\x3f\xb9\xbf\x63\xf9\x35\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.3
    b"\x30\xd1\x58\xa8\x25\xbc\x30\x6c\x8d\x02\x18\x46\xaa\x9a\x89\x3e\x02\xcd\x26\x15\x66\x26\x31\x85\x6b\xd6\xcd\xe6\x08\xa3\x63\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x30\xee\x13\xed\x48\x74\xa9\x0b\xb2\xc0\x06\x14\xba\x44\x9f\x9e\xcd\xd2\x79\x55\xba\xda\x97\x5b\xf0\x09\x69\x27\x00\x5b\x31\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x30\xf1\xa9\x77\x83\xc4\x3f\xc2\x8c\xd2\x64\x43\x23\x47\xb5\xdb\xa7\xd8\xe2\x1b\x2a\x1a\x2a\xe2\xc2\x8f\xad\xd3\xd9\x57\xa6\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x30\xfa\x6b\x1e\x21\x35\x24\x43\x36\xce\x46\x34\x91\x97\x73\x4c\xec\x43\x1f\x06\xd8\xad\x68\x83\x98\xc8\xf2\xf7\x58\xd9\xa6\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x31\x08\x83\xa2\x00\xe3\xa1\xf4\xe0\xe6\x65\x54\x7e\x5f\xea\x56\x5c\x5c\xdb\xd1\x83\xdf\x41\x9d\x60\x17\x61\x07\xe0\xd0\x6a\xe6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/ConditionalEscrow.sol",
            name="ConditionalEscrow",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x31\x11\x4d\xd1\x0b\xab\x73\xf5\xc9\x58\x5b\x26\xe7\xf3\xfe\xa0\x75\x04\xf7\x6b\x28\x78\x27\xd9\xa2\xd4\xb4\x38\x35\x2a\x53\xa6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x31\x14\xdb\x19\x1f\x6c\x67\x56\x26\xf5\x8e\xd1\x31\xf4\x9d\x01\x14\x27\xd4\x45\x98\xc3\x40\x7e\xb6\x1d\x0f\xbe\x0f\x21\x73\x69": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v6.6.0, v6.6.1, v6.6.2, v6.7.0
    b"\x31\x37\x0f\x41\xb9\xf5\x5a\x2e\x35\x61\x6c\x08\x7d\x51\x0c\xf9\xf2\x83\x93\x3f\xd5\xaa\x11\xf6\x60\x5d\xfa\xcb\x5b\x6d\xb3\x77": [
        KnownContract(
            project="solmate",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v6.7.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x31\x63\x77\x12\x6b\x09\x8d\x9d\x23\xbf\xb2\x1e\xd9\x35\x86\xb8\xce\x39\xf8\x1e\x7d\xf7\x2d\x34\xdd\x8b\xf1\x00\xfc\x40\x6c\x11": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1
    b"\x31\x6a\x31\x53\x09\x5b\x8b\xc5\xe6\xbc\xab\xdc\x5e\xdd\xc9\x03\x95\xf7\xa0\x2c\x4a\x5e\xa6\xff\xca\x95\x83\x34\x3c\x83\xca\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.1",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x31\x7e\x78\xb5\x17\x31\xa0\x7c\x96\x30\xd5\x07\x11\x49\xe2\xd8\xe2\x3f\x82\x09\xb8\x28\xdd\xa3\xbf\x4f\xaf\x59\x6c\xa4\x00\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC1363.sol",
            name="ERC1363",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1
    b"\x31\xa3\x58\x2a\x83\x0f\x07\xa7\x19\xdc\x82\x70\x16\xff\x39\x87\x2b\x66\x9a\x56\x0f\xd6\x61\x13\x6d\x91\x50\x2e\xde\x38\x7c\xf6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x31\xb1\xbd\x02\xac\x94\xb5\xae\x95\x0d\x20\x45\xa2\x92\xab\xed\x94\xc2\xda\xc2\x1b\x3d\x43\x77\xf6\x14\xa1\x72\xc8\x49\x7a\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/Counters.sol",
            name="Counters",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/Counters.sol",
            name="Counters",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x31\xb1\xf9\x73\x4a\xb9\x26\x80\x21\x8a\x21\x11\x12\x94\xaf\x3f\xa8\x71\x6f\xdc\xed\xb4\x82\x91\x3e\xf6\x92\x7a\xec\xd8\xe2\xf8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20UpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4
    b"\x31\xb6\x5e\x1f\x34\xd0\x55\xab\x29\xdd\x52\x58\x41\x3e\x07\xe2\x89\x73\xad\x42\xbd\xb8\x6f\x58\x53\x96\x2b\x80\x4d\x16\xe1\xc0": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.4",
        ),
    ],
    # Hash appears in versions: v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x31\xc4\x51\xf9\xbb\xe9\xf3\xcf\x1c\xe8\x2d\x6f\x93\x0e\xa1\x10\x74\xb0\xd5\xc3\x85\x15\x19\x8d\xb9\x9f\xd4\xe8\xd3\x11\xfe\xba": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164
    b"\x31\xcc\xc9\x31\x06\xa8\x5d\xe6\x39\x11\x8a\x19\xa1\x80\xd5\x88\x9a\xba\x0d\x19\xb1\xee\x7c\xba\x08\xeb\x65\xf2\x3a\x37\xfc\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.164",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x31\xdc\x3f\xdf\xfe\x2f\xb4\x66\x19\x14\x65\xd5\x43\x2c\xd0\x93\x02\x62\xb1\x0c\xcc\x79\x51\x5c\x3a\x8c\x2e\xe5\xd5\x84\x84\x13": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x31\xe1\xd5\x90\x95\xcc\xd6\x15\x96\x7a\x92\x73\x2d\x6e\xc3\x9f\x5f\x1b\x8b\x95\xe8\x14\x12\xff\x8c\x0b\x6e\x8d\x5b\x5c\xe6\x83": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/PositionDescriptor.sol",
            name="PositionDescriptor",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170
    b"\x31\xe5\x37\xc4\x2f\xde\xc2\xa4\xae\xf2\x69\xf2\x27\x78\xdc\x55\xe3\x35\xf5\xd0\x98\xf0\xf7\xfe\x70\x85\xab\xa3\xf5\x2c\xf8\x0a": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.170",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x31\xe8\xc6\x00\xb3\xc7\x6c\x15\xcd\x06\x9c\xd4\xfa\x7f\xa2\x56\x00\x96\xd5\xfd\x5d\x17\x28\xb9\xe1\xb4\xed\xec\x53\x50\xd9\x89": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x31\xfb\x88\x5d\x46\x28\x98\xcd\x47\x42\x59\x05\x43\xa3\x34\xe6\xdf\x4c\x96\x07\xf5\x12\x5e\x5d\x66\xdf\x3e\xc8\xbe\x74\xb4\x20": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x32\x0c\x78\x8b\x43\x80\x6c\x12\x77\x63\x89\x5f\x75\xd2\xe6\x0a\x18\x07\xf1\x0d\x61\xe8\xe1\xfa\xf4\x92\xb2\x71\x64\xe2\xc6\x41": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMathUpgradeable.sol",
            name="SignedSafeMathUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x32\x18\xc6\xca\x4a\xb5\x19\x4d\x32\x02\xb0\xab\x7d\x2c\xf7\xfb\x02\x8e\xac\x5c\x15\xf6\xdf\x9b\x4e\xab\xc5\xd3\x6e\x1e\x94\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x32\x2f\x1c\xdf\x46\xf2\x0b\x01\x63\x2d\x8f\xfa\x0f\x47\x1c\xce\x6b\x42\x2f\xff\x32\x36\x21\x86\xe6\x7a\xb7\x0b\x89\x87\x6f\x63": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x32\x35\x09\x85\xeb\x05\xa9\x1c\x2e\xec\x78\x87\x0b\x6f\x1e\xcc\x08\xe9\xbb\x48\x69\xd3\x8e\xbd\xbd\x35\xfb\x51\xf3\x7e\x6c\xaa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.1.0-rc.0
    b"\x32\x48\x7e\x48\x61\xa0\x9e\xdc\xb9\xaa\x9e\x91\x3b\xff\xee\x88\x7e\xa8\x93\x92\xb4\xa1\x6c\x39\x0c\x97\x5c\x7b\x7c\xdd\x52\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Storage.sol",
            name="ERC1967Storage",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x32\x53\xe1\x69\x64\xc0\x69\xf9\xd6\x4c\x34\x35\x7e\x75\xb7\x31\xa1\xcc\x85\x89\x99\xe4\x32\x92\x44\x31\xa5\x4a\x07\xd0\x61\x32": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/UniswapV3Factory.sol",
            name="UniswapV3Factory",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x32\x85\x4f\x13\x5e\xea\x18\x66\xc7\xfa\xf9\xbc\xe7\xa3\x69\x56\xf6\x88\xfd\x34\xf2\x05\x1d\xf5\xd4\x25\x7e\x95\x22\x01\x57\x81": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/IUniswapV3Pool.sol",
            name="IUniswapV3Pool",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x32\x86\x6e\x84\x48\x4e\x76\x8b\xbe\xfc\xc3\x8e\xa0\x7a\x61\xdf\xf6\x48\x01\x86\x33\xf9\x8d\x64\x92\x72\x03\x2e\x13\x0c\xd0\x1c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IMessageProviderUpgradeable.sol",
            name="IMessageProviderUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.261
    b"\x32\xa0\x2a\xe9\x3e\x12\x5e\xf0\xb4\xd3\x0c\xa1\xa1\xf0\x5e\xfa\xad\x6d\x53\x09\x6f\x07\x50\x8f\x48\xe1\x33\x57\xe6\x35\x99\xa6": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.261",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x32\xf6\xef\xb5\xca\x53\x35\xd5\x69\xee\x45\xbf\xf7\xae\x5e\xe4\xc0\xfa\x1d\x9d\xaa\x8b\x50\x78\xcc\x7e\x8e\x5e\x84\xbf\x83\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/RSA.sol",
            name="RSA",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x33\x21\x97\xff\x26\x1d\x44\xf5\x36\x5d\x4b\x2e\x2b\xc0\xba\xc0\x74\x8a\x41\xce\x66\x43\x93\x23\x63\x0a\x54\xa4\x32\x4b\x35\xaa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x33\x29\xcc\x09\xaf\x5d\x6d\x07\xa7\xa6\x25\x61\x07\x70\xc8\x50\x41\x2d\xad\x04\xc1\x58\x06\x65\x52\x05\x04\xf0\x47\xac\x6b\x1c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x33\x51\x29\x20\xf2\xf6\x95\x53\x74\x26\x02\xc6\x1a\x03\xae\xea\x7b\x40\xf9\x01\xa2\x34\x06\x9e\x3b\xe4\x19\x22\xb1\x15\x27\x4c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x33\x55\xd0\x1e\x18\xf6\xbe\x2a\xb1\xd6\x20\x53\x90\xe8\x47\x4e\x4a\x3f\x65\x3e\x35\x66\x80\x7f\x34\x43\x0e\x78\x4b\xd7\x12\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75
    b"\x33\x5e\xfa\x04\xf0\x05\x5a\x65\x86\x37\xb5\x49\xb7\x0c\xd0\x86\xd2\x86\x31\x98\x15\x26\xa8\x23\xf8\xf1\x42\xdd\xd6\x9b\x9d\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.75",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x33\x8d\x6a\xf9\xd1\x72\x56\x92\x32\x34\x76\x2f\xac\xef\x75\x47\x8c\xb3\xb0\x2d\xcf\xc1\x11\xb9\x38\xa7\xe9\xc3\x50\x00\x8e\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x33\x8e\x16\x98\x3f\xbc\x3c\x56\x51\xb5\x1b\x50\xab\x60\x8a\xbc\xc2\x9c\x6a\x00\x2a\x15\x0b\x92\xa4\xfb\x0d\x6b\x3e\x72\xa9\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x33\x9e\x01\x4b\x63\x50\x5a\x67\x56\xfb\x39\x08\x39\x48\xb8\x4f\xcb\xba\x23\x90\x27\x0e\x75\x3e\xbd\x08\x6d\x49\x3f\x61\x07\xe4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupply.sol",
            name="ERC20PresetFixedSupply",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x33\xc3\x6b\x26\x89\x63\xb3\xdb\xc4\xa2\x08\x24\xba\x6e\x42\xb3\xe0\x7e\x1c\xe2\x24\xae\x51\xc8\xea\x9b\xd4\x23\xcd\x8b\xb0\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7
    b"\x33\xd5\x91\x47\x9e\x15\xf3\x0c\x15\xba\x2d\x9f\x28\x13\xa2\xb7\xc2\xd8\x0d\xf1\xc5\xff\x4b\x1e\x8c\xd0\x4e\x1b\xcb\xe0\xda\xc9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62
    b"\x33\xd9\xe0\x99\xb2\xfc\xc6\xf0\x0f\x4b\x4b\x05\x0f\xb6\x3b\xc5\x22\x41\xc0\x40\x9e\x7f\xb4\x8d\x73\xfc\xf2\x83\x1a\x15\x82\x2d": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.62",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x34\x01\x72\xe8\x92\x8c\x78\x71\xf5\x79\x68\x5e\xa5\x91\xa1\x7e\x69\x7f\xbd\x6e\x7b\x3f\x30\x72\x97\xfe\x17\xa7\x2f\x98\xdb\x25": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218
    b"\x34\x08\x53\x7e\x90\x37\x8f\x60\x1f\x42\x65\xa0\xb2\xdf\x51\x2e\x15\xc6\x68\x13\x34\x96\xe8\xeb\x75\x9d\xb3\x67\x0d\x34\xe5\x44": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.218",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x34\x0a\x34\x51\xd4\xe8\xbc\xd5\x75\xfc\xdb\x16\xd0\xf6\x34\x6b\x23\x95\x35\xe0\x2e\x4b\x16\x1b\x38\x06\xfd\xc0\xae\xa7\x3c\xd1": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x34\x30\xc7\x54\xc6\x0d\x65\x02\x24\x5e\xc1\x5e\x35\xb2\xc5\x09\x7f\x50\x26\x7f\x13\xe4\xac\x6a\xc6\xc1\x30\x9a\x59\x94\xdb\x94": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.52
    b"\x34\x3b\xc4\xf3\x05\xc3\x46\x7b\x19\xf6\x71\x1e\x23\xe8\xec\x9c\x52\x3b\xbb\x76\xe8\xa1\x46\x34\x7a\x91\x62\xc8\xd7\x2b\x4e\xec": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.52",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x34\x3b\xf6\x4a\xb2\x8b\xe0\xb5\x7d\x7d\x7c\x0d\xf9\xa8\x94\xb4\xea\x1a\xa7\xc7\x58\x10\x8a\x81\x4d\x21\x0f\xba\x5d\xed\xba\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x34\x59\x03\x51\x23\x29\x48\x8a\x4e\x62\x8f\x60\x77\xb8\xfb\x5b\xce\x34\xd7\x80\xf1\xe3\x87\x06\x56\xb0\xdb\x6e\x3a\xa0\x48\x8b": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/SqrtPriceMath.sol",
            name="SqrtPriceMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\x34\x62\xe5\xd8\x97\x49\xc7\x4a\xc7\x44\x5b\xc0\x28\x7e\xf9\xf9\x57\x68\xc7\x7c\x9d\x84\x78\xc0\x90\x65\x69\xe1\x09\x54\xd7\x6f": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/Path.sol",
            name="Path",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/Path.sol",
            name="Path",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/Path.sol",
            name="Path",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4
    b"\x34\x6a\x15\x39\x6f\xb7\x09\x27\x30\x1d\x00\x5d\xc3\x3c\xc9\x2a\x94\xc9\xc8\x3a\xb6\x1a\x26\x53\xd7\x2e\x0e\x0d\x34\xc5\x8c\x3f": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.4",
        ),
    ],
    # Hash appears in versions: v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x34\xc6\xab\x76\xa5\x69\x62\xf3\x55\xe4\xbe\xd6\xfd\xb2\xea\x7e\xd3\xbf\xa3\x30\x15\xe4\xbb\x09\x03\xb7\x9a\xa6\xd5\x07\x27\x33": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x34\xda\xf5\xa0\x4a\xdf\xb5\xea\x32\x8e\x68\xe0\x97\x4c\x36\x42\x74\x84\x7c\x93\x4b\xcb\x12\xdd\x81\x81\x26\xdf\xe4\xea\x1f\x3e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0
    b"\x35\x17\x9d\x4e\xf8\x93\xcd\x5b\x52\xed\xc8\x6a\xbb\x5b\x1b\xb2\x6e\x31\xab\xfa\x6a\x7f\x70\x88\xe4\x15\xd5\x28\x99\x58\x9f\x56": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191
    b"\x35\x1a\x67\x5f\x27\x01\x7f\x5b\xff\x59\xb5\x2f\x00\x7b\x50\x3c\x04\x71\x68\xa1\xd6\x4f\xdd\x16\x0f\xb8\xd7\xe3\x99\xd8\x7a\xe4": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.191",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x35\x62\xbc\x93\x42\x68\xe9\xbf\x9d\x9d\x4a\x3e\x70\xac\xf2\x7e\x00\x32\xbd\xff\x62\xbe\x1f\x0a\xa8\xde\x97\xfb\xbe\xd9\xc7\x5a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x35\x8c\xf1\x5a\x18\x21\xda\x6c\xd4\x0d\x86\xee\x13\x06\x27\x72\x17\xd3\x5b\xce\x41\xbb\x01\x30\x1a\x67\x15\xac\xcc\x2c\x40\xfe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79
    b"\x35\xb8\xfb\x3d\xf5\x66\x7e\x94\xba\x06\xc6\x36\x0b\x76\x65\xd2\xf1\x10\x90\x66\x8f\x2f\x92\x7f\xdd\x1a\xd7\x65\x83\xc6\x7f\x47": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.79",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x35\xdc\x24\x89\xe0\xaf\x22\x94\xd4\x0b\x51\xfb\x2e\x5a\x3e\xcd\xf5\x74\x06\x01\x4a\x68\x91\xef\x0a\x95\xca\x3f\xce\xfc\xa8\xb0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60
    b"\x35\xe4\x7b\x0f\xcf\x42\x84\xe5\x86\x73\x30\xfc\xbb\x11\xe4\x65\xf3\xb4\x0a\xa6\x4d\x22\xd2\x27\x45\xca\x6a\x71\x85\x58\x65\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.60",
        ),
    ],
    # Hash appears in versions: v0.0.186, v0.0.187
    b"\x36\x0f\x90\x45\xf3\xa0\x07\x35\xba\xce\x95\xcc\xcb\xf3\x4b\x71\x6d\xf5\x72\x38\xa2\xb1\xc4\x7a\xe6\xda\xe0\x51\x8f\xb5\x4c\x61": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.187",
        ),
    ],
    # Hash appears in versions: v0.0.203
    b"\x36\x1b\x64\x94\x4d\x95\x68\xc1\x19\x10\xab\x32\x92\xf0\xda\xc6\x6f\xd2\x15\x49\x9d\xac\x6d\xb1\x6a\xc3\x3c\xaa\x2a\x61\x70\x36": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.203",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2
    b"\x36\x2e\xfc\x6a\x2f\x8b\xb5\xf4\x76\xdb\x20\xf8\x9d\x6d\xe0\xf0\x46\xec\xdb\x3a\xb4\x85\x88\xf7\xa1\xa8\x24\xd9\x13\xd5\xb7\xe9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x36\x40\x62\x53\x1b\x32\xb6\x1d\x94\xe8\x30\x7b\xb6\x40\xa6\x63\x28\x19\x9c\x70\x00\xfe\xd6\x75\x66\x9a\x6d\xba\x94\x24\xab\xbb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27
    b"\x36\x4e\x2e\x3f\x0c\x94\xc1\xf8\x99\x8c\x37\xb7\xb7\xe2\x20\x66\xf0\x6f\x90\x47\x78\xd6\x7b\x9e\xd7\x51\x72\xfb\x4f\x56\xbb\x60": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.27",
        ),
    ],
    # Hash appears in versions: v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246
    b"\x36\x77\xdc\x9f\xa8\x07\xdb\x05\xd0\xdf\x08\xc2\x8a\x4e\x92\x24\xf2\x7d\x93\x38\xce\x29\x9b\x17\xa3\xde\x12\x22\x40\xda\xf4\x01": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.246",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x36\x89\x75\xf1\x74\x27\xfa\xc5\x9d\x7c\x87\x71\x6d\x15\xf3\xe7\x0d\x8a\x4e\x4a\x51\xc2\x45\x7b\x43\xd0\xc0\xb1\x2d\x5a\x8b\x28": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x36\xb3\x36\xae\x97\xc0\x4e\x01\x85\x37\x98\x62\xce\x49\x32\x71\xdd\x73\x98\x2b\x68\x6d\x89\xfe\xca\xb0\x73\xe2\xb6\x9a\xa5\x69": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x36\xc8\xed\x16\xc0\x72\x23\x12\x3f\x89\x6a\x6b\x94\xc3\x31\xcb\x21\x73\x9a\x8a\x87\x9c\x64\x06\xd2\xcd\xa9\x17\xa5\x7b\xa2\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x36\xd3\x80\xf2\x98\xcc\x1d\xe4\x2e\x59\xe1\x20\xff\x25\xa9\xc2\x28\xc3\xc3\xa2\xcb\x9f\x87\xc5\xce\x33\x85\x95\x41\x79\x9f\xb0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.124
    b"\x36\xd5\xb5\xd2\xcc\xf7\xd3\xa0\x53\x93\x75\x98\xdb\xde\xff\x37\xd9\x11\x69\x26\x4e\xc0\x42\x94\x1b\xcb\xa8\x68\xb2\x2d\x9d\x97": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x36\xda\x83\xe2\x4c\xd8\x4f\x91\xb4\xf4\xef\x6c\x4c\x35\x4a\x8a\x79\x49\x72\xfe\x28\xf2\x63\xb4\x16\x86\xa8\x60\xf0\x31\xfb\x48": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x36\xe6\x8b\x6a\x52\x80\x1c\xd3\x9a\x56\xb3\x17\x01\x54\xed\x96\x58\x01\x9d\xbc\x44\x64\x6b\x31\xb3\x42\x7e\xa8\xa9\x05\x1e\x7d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Pausable.sol",
            name="ERC721PausableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x36\xeb\x7e\x10\x06\x1e\x4d\x85\xeb\x86\x98\x71\x9a\xe7\x76\x79\x6c\x54\xfd\xf8\x7b\x5a\x09\x01\xad\xc6\x6c\x81\x5e\x7b\xa5\xb5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x36\xed\xc8\xf4\x62\x19\x05\xb0\x52\x43\x0a\x73\x59\xaf\xe3\x1a\x8d\x71\x55\x31\xaa\x27\x86\xcf\xcb\xdd\x9d\xc6\x5a\xe8\x77\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0-rc.0
    b"\x36\xfc\x2d\x70\xfa\xa2\x8b\x82\x88\xd9\x07\x53\x7c\x73\xeb\x41\x2f\x65\x92\x2d\x31\xf4\x50\x49\x18\x3a\x07\x9d\x9b\x4b\xd4\x1b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x37\x05\x20\x04\x96\xf6\xff\x6d\x57\x2b\xe5\x73\x73\x16\x84\x00\x6a\xe0\xb9\xfa\x6d\x8d\x53\x2a\x94\x8d\xad\x8a\xdb\x3f\xd8\x49": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\x37\x14\xe9\x2d\xda\x0b\x84\xc2\x37\xba\xe9\xcf\x74\x4d\x44\xdb\xf3\x94\x39\xaa\x10\x38\x22\xee\x92\x64\x54\x70\xf2\xf5\xf3\x4b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/INotifier.sol",
            name="INotifier",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.154
    b"\x37\x15\x0f\x41\x25\x53\xb4\xba\x49\xe9\x87\xd5\xc2\x17\x85\x01\x36\x92\xce\x32\x29\x45\x1b\x31\x63\x39\xef\x7b\x19\xa7\x14\x0a": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.154",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0-rc.0
    b"\x37\x42\x6a\xf6\xd0\x33\x96\xab\x74\xe0\x30\xd5\x6e\x07\x31\x5c\xba\xf3\xa0\x3c\x30\x1d\x16\x1d\x71\x13\x57\x33\x2d\xf8\x88\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x37\x44\x54\x5b\x3d\x89\x1f\xd0\x2d\xbb\x76\x08\xa0\x5d\xaa\x62\xf4\xdc\xc1\x94\x62\xa9\x7b\x4a\xec\xbb\xf5\x95\x5a\x82\xa8\xbc": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x37\x4b\x09\x43\x2c\xcd\x5f\xb3\x26\x05\x88\x50\xb3\x16\x59\x0a\xb1\xf7\x33\xeb\x13\x2a\x20\x95\x1a\x8b\x91\x03\x89\x70\x09\xea": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x37\xb5\x21\xbf\xe4\x0c\x9c\xb3\x0c\x1b\x3c\x79\xa7\x64\x51\x5e\x68\x10\x28\xe9\xdf\xef\x8f\x0f\x01\x31\x44\x17\x25\x71\xc8\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v4.0.0
    b"\x37\xe1\x90\x4a\x5d\xff\xd5\x1b\xbe\x46\x34\x70\x04\x12\xb5\x46\x3c\x6e\x46\x3b\x72\x24\x33\x84\x8d\xd6\x35\x28\x0f\x6c\x84\x2e": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/FixedPoint96.sol",
            name="FixedPoint96",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/FixedPoint96.sol",
            name="FixedPoint96",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x38\x15\xd6\xa3\x71\x1d\xc1\xaf\x7c\x4d\xe3\x83\x20\x63\xc8\xa5\x0f\xc5\xe0\x6e\x9e\x79\x1c\xa8\x2f\xbb\x49\xf0\x22\xed\x07\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x38\x2c\x8e\x02\x78\x5f\xaf\x40\xb6\x48\x3e\x40\xb7\x2f\xb5\xa6\x3a\x17\x6c\xde\x06\xad\x91\x95\x90\xad\xdd\x28\x8b\x2e\x00\x4e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.139
    b"\x38\x46\x71\x62\xc2\x87\x93\x15\xcd\xc7\x79\x2c\x93\x5c\xec\xdc\xd3\xe3\xe9\xfb\x75\x87\xec\x56\x5f\x91\x2f\x1d\xbf\x72\x35\xe6": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.139",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x38\x6b\x2b\xf0\x33\x21\x03\x12\x10\xff\xb9\x6f\x72\x40\xbf\x17\xb1\x66\xd3\x78\x36\x1c\x23\x8f\x60\x5d\x0f\xec\xa0\xc1\x4d\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0-beta.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x38\x6e\x05\x8f\x42\x06\xae\xe6\x83\x0d\xc1\xb4\x19\x9f\xc5\x91\x24\x2f\x1b\x49\xbb\xad\x29\x64\x1c\x76\x21\x14\xd5\x6b\x93\xe4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x38\x77\xef\x86\x8d\x03\x99\xa3\xbe\xb9\x91\x83\xc6\x08\x7d\x74\x21\xf1\xf6\x9e\x35\xdb\xaa\x99\xc5\x99\x29\x20\xff\xc6\x2d\xaf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletCliffUpgradeable.sol",
            name="VestingWalletCliffUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x38\x79\xa4\x28\xc9\x37\x9c\x3d\x3c\x8b\x4d\xad\xe8\x15\x6b\xcc\x26\x60\xc7\xe3\xf3\x6e\x57\xd4\x8d\xec\x35\x33\x64\x72\x94\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-beta.1
    b"\x38\x7f\x8f\x2b\x16\x87\x71\x3a\xd4\x96\xa2\xf4\xed\xab\xd0\x40\xee\x75\x92\x7c\xbb\x88\xe3\x21\x4e\x38\x83\x4e\x32\xb2\x4b\xd1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x38\x97\xea\x8d\x07\x77\xac\x86\x48\x1f\x27\x96\x2d\xcb\x21\x52\x6f\xd0\xa1\x9a\xf3\x14\xb7\x88\x02\x2d\x54\x25\xda\xde\xb4\x06": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x38\xa8\x71\xdc\xc0\xd9\x6c\x55\x3a\x93\x77\xc7\xc0\x6b\xeb\xe4\x47\xa8\x7c\x05\x41\x00\x48\x0f\xe9\x81\xbb\x08\x75\x00\x6f\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92
    b"\x38\xa8\xef\x0a\x28\x2f\x63\x58\xc2\xa9\x7f\xe7\x25\x06\x8e\xf3\x21\xe8\xd4\x02\x5b\xaf\x8a\xa0\x89\xd5\xbf\xdd\x61\xa0\xf7\xb0": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.92",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x38\xb5\x21\x51\x7d\xe2\x7c\xc2\x3f\xbe\x78\xd2\xc9\x62\xed\xde\x82\xf6\xb0\xde\x02\xf9\xde\xca\xcf\x50\x0f\xce\x48\x39\x66\x20": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/AddressStringUtil.sol",
            name="AddressStringU til",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/AddressStringUtil.sol",
            name="AddressStringU til",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/AddressStringUtil.sol",
            name="AddressStringU til",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/AddressStringUtil.sol",
            name="AddressStringU til",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/AddressStringUtil.sol",
            name="AddressStringU til",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x38\xc1\xbe\x34\x9a\x69\x7a\x1f\xe7\xda\x33\x5c\x3b\x0d\x05\xcd\x6e\x44\x4b\x26\x9a\xf4\x98\x19\xaa\x9e\x6e\x41\xd3\x80\xd4\x12": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x38\xce\x30\x1f\xa2\x5c\x9d\xe3\x5c\x63\x8e\x94\xae\xac\x65\xb5\xc9\x33\x83\xef\xbc\xdf\x07\x4c\xb5\x9d\xd5\x7c\xd3\x86\xb8\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x38\xd1\x41\xd1\x03\x0b\x41\x75\x61\x93\xda\xf4\xa0\xd5\x4a\xc3\xf0\x5d\xf8\x57\x51\xd1\xaa\xe0\x9c\x19\xf1\x2d\x66\x8b\xcf\x47": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265
    b"\x39\x16\x68\x2f\x10\x42\xcf\x01\x43\x24\x0e\x28\x37\x04\xac\x3f\x62\x25\x5d\x22\xf0\x0d\x6d\x36\x5f\x93\x1e\x9b\x79\x25\x5b\xc6": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.265",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x39\x34\x8e\xfa\xcf\x8d\x1a\x5e\x89\x04\x84\x03\x71\xa2\xea\x8a\x11\x0f\x35\xc8\xd8\xd2\xb1\x20\x3a\x6f\xd4\xca\x59\x06\x96\xc6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x39\x57\x88\xe7\x62\xe8\x7a\x9f\x39\xa0\xd4\xda\x5e\xdc\x7e\xb3\x18\x30\x40\xc7\x26\xcd\x38\x12\x13\x2c\x90\xf5\x0b\x98\x54\x86": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC6093Upgradeable.sol",
            name="IERC721ErrorsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x39\x66\xaf\x68\x6b\x9c\xbe\x6f\xa8\xcd\x02\x66\x12\xb0\xdb\x14\xcb\xdc\xfe\xbc\x90\xd4\x6c\xf7\x8c\xe6\x43\xcc\x78\x8f\x05\xc9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x39\xa1\x6d\x93\xa9\x71\x1d\x97\x70\x72\x55\xc0\x8e\x12\x71\xae\xb7\x97\x40\xd7\xb4\x82\xb8\xd7\xf2\xfc\x7f\x2b\x36\xb3\x82\xf2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168
    b"\x39\xc2\xed\x23\x57\xb2\x9a\x80\x52\xc4\xcb\xdb\xfb\x9a\xcc\xbb\x6f\xdf\x27\x6e\xbc\x6b\xa6\x6d\xa5\x40\xae\x89\x71\x5e\x79\xb0": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.168",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74
    b"\x39\xe5\x54\x3a\x0f\x83\x5d\xb4\x29\x6f\xfb\x19\x65\xfb\x8a\x8c\x3b\x60\x77\x3d\x65\xd1\x3b\xa9\x0d\x5e\x83\xd6\x1e\xd3\xf2\xaf": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.74",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x39\xff\xe0\x76\x00\xaf\xc7\xe8\x11\xbf\x86\x2e\x9c\x72\x78\x6a\x5e\x6f\x8d\xec\xde\x8d\xec\x66\x4d\x9a\x43\x26\x8e\x77\x7b\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashLender",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashLender",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x3a\x23\x87\xe2\xa9\x80\x49\x16\x34\x2c\xb3\x68\xc9\x60\xa9\x00\x66\x1c\x0a\x22\x6b\xfe\x4b\xf8\x48\x96\x75\x61\xc0\x55\xb7\xe3": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/Position.sol",
            name="Position",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x3a\x58\x5f\x29\x17\xb2\xc1\x34\x75\x51\x38\xb1\x75\x04\x54\x87\x70\x4c\xd7\x80\xfe\xbe\x73\x65\x81\xdc\xb9\x25\xc4\x1b\x6f\x56": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IPositionManager.sol",
            name="IPositionManager",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3a\x5a\x5b\xc7\xcd\x33\xea\xbb\x3a\x65\xef\x50\x1d\x03\x28\xa7\x8a\x80\x36\x7c\x62\x9f\x6a\x69\x81\xd4\x84\xf0\x6e\x50\xeb\xad": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x3a\x6b\xf1\x87\x13\x77\x5c\x55\xc9\x12\x9e\x6c\xe6\x73\xc7\xa2\x0e\x89\xc8\x68\x71\x25\x47\x66\x3c\x31\x1c\xba\x54\x50\x91\x44": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0, v3.1.0-rc.0
    b"\x3a\x72\x05\xce\xb8\x29\x2e\x1e\xb4\xd9\xfa\x48\x61\x9e\x7c\xaf\x9d\xde\x64\xeb\x46\x72\x70\x89\xcd\x03\xde\x58\x54\xdb\x3e\x68": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3a\x73\x2a\xe8\xf1\x83\xbf\xa5\xc9\x6a\xf9\xb7\x1d\x12\xa1\xdf\x30\x35\xc4\xfd\xe5\x1d\xcd\x50\x59\x77\xfc\x05\x8f\xea\xda\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822.sol",
            name="IERC1822Proxiable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141
    b"\x3a\x9a\x4c\x6d\xd6\x6a\xf6\xa9\xd3\x6f\x08\xa0\x7e\x67\x07\xa6\x31\x8f\xe9\x43\xa0\x09\x50\x4c\x69\x9d\x6f\x30\xea\xf9\x5d\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.141",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x3a\xa7\x0c\x52\xa8\xb5\x3d\x5d\xb5\x2a\x5d\xc1\xbc\x79\xe9\xb2\x17\xeb\xf8\x57\xc6\x9a\xb7\x42\xe7\x67\x32\x37\x40\xbb\xc5\xd0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x3a\xae\xd0\xbb\x51\x57\x67\xc7\x5c\x82\xad\xc4\xae\xa1\x5f\xab\x9a\xbf\x44\x6f\xf6\xaa\x12\xc0\x65\xc9\x97\x19\xe9\xab\xbd\x24": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3a\xbb\xda\x0e\xa7\x6b\x00\x79\x2f\x5f\xf5\x8e\xc9\x8b\xd0\xd6\xfb\x71\xf1\x4e\x69\xe9\x11\x25\xc1\xdc\x14\xc4\x4d\x33\x09\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3a\xcd\x1a\x09\xd1\x1f\x12\x05\xa5\x14\xbe\xf1\x65\x61\x2a\x69\xe0\x1d\xf2\x2c\x17\x36\x05\x07\x65\x66\xc8\x92\x44\xa8\x03\x1b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3a\xe7\x49\xff\xa7\x7b\x16\xc4\xf8\x51\xac\x7e\xa0\xbc\x1b\x3e\xc0\x43\xf1\x67\x1a\x69\xf7\x57\x3a\x0d\x56\xa8\xb1\xf1\xd8\x2e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x3b\x1f\x4a\xd9\x84\x35\x2c\xa0\xdb\x26\xfd\xff\x05\xb7\x01\x78\x83\x24\x89\x7f\x62\x12\x96\xaf\x69\x26\xf3\x2e\xab\x12\x9b\x06": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x3b\x25\x3a\x00\x23\x97\x80\xfa\x11\x3b\xc2\xa5\x5b\x0e\x72\x6b\x01\xa4\x13\x8b\x13\x25\x89\x9d\xd1\xef\x3e\xa6\xb0\x12\x87\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.1.0-rc.0
    b"\x3b\x2c\xad\x25\x24\x45\xee\x86\x0a\x28\x3a\xa9\x26\xbb\x8f\x8c\xe3\x00\xcc\x37\x24\xfb\xbd\xbd\x6e\xfb\xb5\x5b\x60\x70\x89\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x3b\x55\x97\xe4\x84\x0d\xc7\x12\xa1\x14\xd3\x76\x38\xf4\x19\x95\xaa\xa4\xc5\xc2\xb8\xfd\xc8\xf1\xd8\xc4\xe9\xf9\x5e\xa7\xea\x99": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x3b\x6b\x76\xe7\x66\x31\x90\xcb\x9e\xd5\x07\x5a\x3e\xec\xdd\x0b\x72\x7e\xfd\x43\x0a\x0f\xaa\x0f\xed\xcb\xb5\x8a\x42\xd9\x9d\x8b": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: main
    b"\x3b\x79\x26\x48\x14\x19\xa4\x1d\xfc\x6f\x95\x2a\x5f\xe2\x1d\xb5\x09\x8b\x16\xf4\x7b\xc4\x2b\x8a\xc1\x99\x74\xfa\x67\xdb\x5c\xde": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IUnorderedNonce.sol",
            name="IUnorderedNonce",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3b\x79\x2f\x06\xb4\xb2\x93\x93\x50\x07\x0f\x8a\xe2\x87\x8e\x9b\x46\xcf\xb7\xb5\xbb\x6b\xd2\x13\xb8\xad\x12\xef\x7e\x51\x34\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\x3b\x7f\x5c\x16\xf4\xe0\xbc\x2c\x65\xab\xdf\xb7\x50\x2b\xad\xf4\xa0\x89\x5c\x6a\x0f\xf0\x9e\x71\x05\xaf\x02\x15\x24\x05\x49\xab": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/Descriptor.sol",
            name="Descriptor",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274
    b"\x3b\x8a\x80\xcb\xef\xa9\x27\x3e\xf3\x48\x1c\x8e\xba\x92\x43\x79\xa7\x35\xd0\x92\x28\xd4\x77\xda\xa4\x7f\x3b\x92\xdc\x21\x4b\x52": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.274",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2
    b"\x3b\xb4\xfd\x3f\xf2\xc3\xce\xb5\xd6\xdc\x2a\xb2\x9f\xa6\xfc\x7e\xa9\x25\x94\xb4\x2f\x17\x88\xe7\x82\xf0\x5a\x7e\x74\xfa\x98\x96": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x3b\xb8\xd4\xb5\xf4\x2d\xd9\xf1\xc5\x14\x60\x2f\x02\x90\x3d\x9c\xf9\xad\x79\xf7\x8a\xd2\xe7\x45\xd6\x09\x87\x17\x34\xd5\x97\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x3b\xc5\xc6\xad\x24\xb3\x02\xac\x93\xa9\x51\xc1\x0a\xea\xa0\xe0\x04\x08\xae\x6e\xf1\x1d\xfa\xfd\x5c\x52\xd6\x97\xfd\x28\xc2\x7c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.246, v0.0.247, v0.0.248
    b"\x3b\xce\xcb\x00\x49\xd3\xa2\x79\x46\xab\x04\x9b\x27\x85\x64\xdd\xf4\x66\xe8\xef\x1b\x2c\x51\x62\x02\xc1\x37\xa1\x3e\xe9\x32\x57": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.248",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x3b\xd2\x5d\xa3\xe1\x68\xdb\xa1\x0d\xdb\xd1\x86\x23\xed\x83\x41\xf5\x65\x22\xef\x75\x69\xf1\x5c\xc2\x67\x5e\x83\x84\x70\x1d\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3b\xd4\xb3\xa5\xba\x38\x88\x86\xa3\x23\x14\xf5\xeb\x00\x8f\x5f\x07\x28\x8b\x42\x87\xec\x4f\xd8\xfb\xcd\xdd\x3c\xf3\xa2\x0a\x8e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x3b\xeb\x29\x44\x6b\x98\x07\x84\xd8\x21\x7a\x0e\xe3\xf3\x98\xa0\x60\xd5\x3b\xba\xd7\xe7\xce\x71\x02\x17\x48\xe8\x82\xd0\x05\x70": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x3c\x1b\x1d\x99\x32\xab\xd2\xbd\x3c\x9a\x90\xbf\x6d\x50\x2e\x50\xc7\x2b\x08\x18\xb8\x0a\xd2\xc7\x91\x41\x21\x1d\x7d\x47\x0b\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x3c\x24\x9e\xda\x25\x39\xfe\x58\x03\x0b\x6d\x0c\x24\xb7\x28\xde\x6d\xf5\x8a\xe9\xbd\xf9\xd4\x29\x6a\x70\xc8\xd2\xf4\xe8\x94\xe6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x3c\x26\xfc\xfd\x41\x61\x7a\x5d\x4c\x7f\x54\x96\x8c\xb9\xe1\x29\x66\x3f\x2c\xb1\xc4\x7e\xaf\x92\x16\x08\xbf\x1b\xa5\x73\x8d\xaa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.6, v0.1.7
    b"\x3c\x2d\x3c\x37\x8e\xc8\xe6\x60\xd6\x62\x7e\x83\x44\x8f\xac\xc3\x84\xb5\x74\xd8\x99\xc2\x3a\xd1\x37\xe7\xd1\x61\x91\xa3\x02\x4b": [
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3c\x3b\x64\xd9\x81\x7d\x7a\x3c\x37\x6e\x72\x2f\xdb\x3b\x1e\x34\x19\x29\xfa\x59\x27\xeb\xbb\xea\x58\x39\x93\x14\x0a\x02\x8a\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x3c\x40\x61\x20\xda\x04\x02\xb6\x6d\x68\xaf\x7b\xb2\x39\x22\x8e\xcb\x8b\x56\x13\x72\x17\xa6\xa8\xc2\xb5\xea\xd1\x0f\xa9\xa9\xfe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3c\x5d\xd8\x52\x89\x2e\x11\xd0\x0e\xdd\x3a\x91\x70\xe9\x08\xde\x25\x02\x7b\x31\x16\x32\x93\xed\x4e\x67\x43\x06\xec\xb5\x56\x81": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.1.15, v0.1.16
    b"\x3c\x7c\x54\x34\xeb\x8c\xff\x6a\x04\x7c\x03\xf6\x21\x30\x77\x71\xf2\xf0\x0d\x2e\xf8\xe4\xdc\xe5\x8c\x19\xff\x6d\x07\x3f\x48\x21": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x3c\x90\x9b\xe8\xdf\x7d\x24\xc9\x4f\xd1\x41\x7d\x50\x17\x60\x7e\xe2\x18\xd9\xd4\x16\x19\xa5\x50\xd9\x1a\x22\x6b\x5e\x1d\xf4\x7c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x3c\xa4\xa3\x81\xc5\xb7\x94\x13\xec\x4a\x36\xd7\x55\x60\xb4\x58\x70\xce\xe1\xa1\xc0\x18\x83\x78\x0d\xff\x8d\x09\xd4\x25\x03\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3c\xac\xc7\x89\x90\x14\xdd\x6e\xb6\xfd\xcb\x3e\x05\x46\xf7\x89\xd1\x81\x53\x3c\x14\x4e\xe6\x87\x75\xab\xcd\x05\xc1\xc2\x83\x11": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccount",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccount",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccount",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccount",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccount",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x3c\xb2\x7b\xe7\x3b\xc6\xa4\xdd\x69\xfa\x16\x94\x56\x33\xc1\xdb\x70\xf6\x7b\xd7\x09\x00\xf9\xfe\xc6\x26\x69\x8c\x38\x87\xfb\x24": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20BurnableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x3c\xc2\x4c\xfe\x47\x38\xe5\x09\x09\x69\xbb\xda\xf3\xae\x4e\xcd\x21\x3b\x3e\x63\xb7\xdf\x18\xc6\x57\x39\xee\x7e\x00\xac\x4b\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x3c\xc7\x90\x54\xae\x39\xd1\xac\xd1\xa2\xe3\xa9\x89\x98\xed\xf1\xa2\xf2\x2b\x1e\xfb\x97\x08\x6a\xbf\xdf\x72\xf8\x78\xf4\x99\x17": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74
    b"\x3c\xe5\x42\xee\xdc\xd2\x2e\xdd\x54\xb8\x70\xa3\x3b\x4b\x19\xa0\xa9\x16\x9a\x25\x36\x22\xff\x63\x27\xfd\xd6\xa7\xc8\x18\xd6\xc0": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.74",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3c\xfa\x2e\x83\x65\xb7\xda\x59\x6a\x7c\xf4\xa8\x84\x30\x0b\xd1\x7f\x7c\xc8\xbe\x3c\x98\x4b\x4b\x7a\xf9\xe1\xd5\x4f\xd9\x9a\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x3d\x5b\xaa\x15\xce\x99\x1b\xde\xd8\x34\x6d\xce\xdd\xd1\xfb\xb6\xd9\xcb\x39\xb4\xfe\x33\x00\x8c\x32\x47\xb7\xf5\xfe\xd9\xe2\xdb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x3d\x64\xe0\x99\x79\xd0\xf1\x1d\x64\x62\x94\x18\x5f\xc4\xb1\x83\x0d\x7c\x5b\x30\xac\x29\x79\x9f\x69\xf5\x8a\x90\xa9\x12\x84\x5f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0-solc-0.7
    b"\x3d\x66\x45\x6c\x0d\x81\xca\x60\x08\x3c\xfa\x13\x31\x81\x9e\x8b\x55\x20\x77\x5a\x8c\x08\x98\x1b\x78\x4d\xf9\x11\x2d\x93\x12\xc0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.1.0-rc.0
    b"\x3d\x66\xeb\xbe\xf2\xd4\x41\x4e\xfa\xd9\xf5\x10\xf8\x3c\x8f\x4d\x32\x8e\xe3\x36\xfe\x2b\xd1\x75\xe2\x81\x0e\xec\x2d\x27\x60\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x3d\x7b\x8d\xed\x47\xa8\x13\xf9\x71\x1b\xad\xa2\x7b\xc2\x51\x39\x7b\x45\xfd\x24\xb2\x46\xc1\xe9\x8e\x95\x66\x05\xba\x76\xff\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x3d\xbb\xbd\x5a\x13\xfd\xbb\xe4\x4a\x67\x5e\x86\x33\xb6\x51\x14\x6c\x93\x22\x30\x30\xd6\x6d\xc0\x21\x05\x5a\x42\x7d\x0b\xa3\x2e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.87, v0.0.88, v0.0.89, v0.0.90
    b"\x3d\xce\xb0\x58\x40\x1a\xe7\xdb\x12\xc4\x67\x21\xba\xa8\x35\x1f\x8a\xaf\x31\xb8\xa1\x54\x4f\x66\x62\xcb\xb8\x09\x80\x22\x94\x4a": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.90",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x3d\xd4\x02\x15\x8d\xf8\x10\xb5\x14\x29\xa3\x61\xf6\xc7\x99\x71\x55\x99\xf6\x55\x76\x8a\x94\xd2\xba\xa5\xbc\x10\x37\x69\xb1\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x3d\xd4\xba\x21\x10\xc5\xd9\x5b\xc6\x6c\x06\x98\x8c\x0a\xc1\xc4\x55\x97\x3b\x59\xbe\xb5\x86\x77\xed\xe7\x86\x61\x0e\x4c\xa6\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x3d\xdc\x98\x05\x2b\x0a\xaf\xa7\xe5\x90\x93\x12\x94\x24\x06\x6e\x06\xd1\x66\x1b\x79\xca\xee\x2b\xae\x9e\x7a\x90\xfd\x3a\x55\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x3e\x04\xde\x04\x90\x15\xe2\x72\x62\xb7\x8a\x72\x19\xe4\xba\xf1\xfc\xd1\x93\x75\x39\xa2\xde\x23\x84\x16\x14\x60\xd2\x2a\xb2\x2a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3e\x21\x53\xe7\x66\xf4\xc1\xc3\xf0\x07\x80\x2b\x14\xd9\x37\xae\x1c\xdb\xd8\x84\xae\xaf\xf0\x36\x5d\xb6\x97\x13\xfb\x72\xba\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x3e\x3e\xa3\xf0\x22\x49\xa2\xc1\x28\x2b\x67\x2d\x1f\xaa\xda\x77\x9b\x49\x7b\x63\x6c\xd3\xe9\x3f\xda\x3e\x5f\x7b\xe5\xea\x7b\x96": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296
    b"\x3e\x46\x00\x98\xa3\x9e\x6a\x92\x44\x2d\x84\x25\x35\x8f\x53\x0d\x05\x46\xa3\x26\x8c\x46\xed\xd4\xb0\xa0\xcf\x12\x27\x34\x88\x8d": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.296",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x3e\x46\x1f\x19\x9a\xd3\x93\x52\x15\xb7\xea\xd1\x50\xa6\xeb\x48\xba\xa5\x6f\x51\xf7\xa3\x65\xa6\x4f\x1d\x47\xa2\x1f\x19\xd2\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x3e\x74\x67\xd7\xe6\xc8\xce\x7e\xa7\x4e\xb1\xe9\xc7\x89\xc8\x0e\x86\x29\xbe\x8f\x4c\x1e\x92\x64\xf3\xb5\x7a\x30\xa0\x84\xa7\x41": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x3e\x9b\x9f\xa8\x99\x9f\x98\x53\x2c\x6c\x88\xe5\x6d\x8a\xa0\x66\x01\xf6\xea\x97\x94\x5e\x51\x15\x33\x81\x88\xc9\x9c\xa0\xaf\x59": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x3e\xcd\x04\x4f\x82\xdb\xb4\x8e\x1f\xc0\xc0\x27\x29\x33\x61\x52\xf8\x31\x85\x7b\xbb\x9d\x42\x74\xd8\xae\x52\x36\xbb\xb4\xb5\x5f": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/external/IERC20Minimal.sol",
            name="IERC20Minimal",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x3e\xda\xc0\x29\x65\x14\xec\xfd\x4c\x33\x15\x82\xbd\x30\xee\x23\x26\x26\xa9\xbe\x6e\x82\x95\x3a\x5c\xe4\x61\xf0\x08\xa2\x00\x99": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43
    b"\x3e\xf1\xac\x80\x07\xa0\xcc\xff\x19\x08\xc8\xed\x48\xd8\xac\xca\x57\xf5\x10\x47\xd6\x46\x77\x31\xe7\xc0\x00\x06\xdc\xe9\xb8\x6f": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.43",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\x3e\xfd\x20\x2e\xa7\xdc\xcb\x67\xca\xc2\xe1\x77\xd7\x74\xed\x23\xa8\x13\x05\xfe\xc3\x97\x69\x83\x6e\x20\x89\xff\x4a\x38\x87\xc6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: main
    b"\x3e\xfd\xb4\xf1\x23\x6d\xd5\x7e\x29\x7b\x3f\xac\x24\x7c\x3e\x62\x2f\x95\xcb\xab\x0e\xe6\xd5\xed\x65\xc7\x87\xc4\x30\x87\x0f\xf6": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/QuoterRevert.sol",
            name="QuoterRevert",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.3.0-rc.2
    b"\x3f\x07\x8c\x5f\x1a\xd7\xf0\x77\x8f\x13\xbf\x96\x69\x07\x2d\x23\xd0\xb0\x26\xf7\x8f\xbc\xc3\x99\xd1\x3c\x99\x83\x3e\xf7\x47\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x3f\x24\x4f\x9f\xdb\x24\x92\x4a\xea\x61\x22\x6a\x5f\xc2\x65\x1b\x81\xd3\xb5\x28\xae\xff\x6f\x98\xe0\x9f\xc4\x6d\x94\xb4\x7d\xa8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3f\x4b\x6d\x9a\xce\xfa\xb8\x97\x16\x4e\x4a\x81\xcf\x3a\x3a\xa7\x34\x57\xd8\x51\x2f\x3c\xb8\xd3\xdc\x66\x10\x62\x40\x0a\x30\x12": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SlotDerivation.sol",
            name="SlotDerivation",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x3f\x69\x68\x94\x5b\x9b\xfb\x19\x30\xae\x29\x44\x1c\x5c\x51\x19\xce\x5f\x51\x0e\x6d\x7d\xf4\x17\xbd\x46\x7d\x53\xa4\x7f\x45\xd7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.241
    b"\x3f\xb6\x48\x8d\x9f\xd0\x20\x9e\x1e\xf4\x45\x49\xca\xc2\x8a\x0b\x3f\x9a\x70\x68\x06\x87\xe0\x35\x9a\x23\xa6\x45\x5a\x5e\x0c\x21": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.241",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x3f\xba\x05\xa0\x3e\x64\xac\x9f\x65\xfd\x9a\xfa\xdc\x98\x8c\x64\xd0\x21\xa8\x05\xc2\x95\x76\x95\x47\xba\x7a\x64\x21\x23\x06\x49": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90
    b"\x3f\xc8\x18\x7c\x76\x11\xa3\x0a\x24\xb8\x1a\xd7\xb2\x26\x2e\x51\x9d\xa1\x01\x0d\xd3\xd1\x65\x59\x8b\x31\x52\xa9\xb2\x29\x58\x08": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.90",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17
    b"\x3f\xc8\x7e\xc2\x89\xa3\x46\x7f\xcf\xae\x5c\x02\x1d\x68\xc1\x37\x19\x3c\xbb\x44\xcd\x9a\x48\x8c\x3d\xe3\xe9\xcb\xdb\xa5\x12\x3b": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.17",
        ),
    ],
    # Hash appears in versions: main
    b"\x3f\xc9\x74\x91\x24\x4e\x90\x8a\x51\x8b\x31\x7c\x8a\xd9\x10\x20\xdf\x7d\xef\x94\x17\xf1\x93\x24\x37\x36\x5f\x45\xa2\x64\xd3\xc7": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/EIP712_v4.sol",
            name="EIP712_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x40\x17\x5c\x28\xc8\x8a\x60\x8f\x1b\x35\x06\xdb\x07\xda\x2f\xf9\x58\xf1\x5e\xf8\x75\x9a\x68\x68\x03\xc6\xdd\xb2\xae\x3e\x2f\x89": [
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x40\x2b\xae\x47\x4f\xe1\x0f\xf1\x72\xa7\xbd\x41\x02\xe5\xc0\x31\x72\x71\x36\x88\x8b\xf3\x39\x40\xc8\x7c\x28\x3a\xb0\xa7\xbf\x57": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Full.sol",
            name="IERC721Full",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Full.sol",
            name="IERC721Full",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Full.sol",
            name="IERC721Full",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Full.sol",
            name="IERC721Full",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x40\x3a\x4a\xa7\x9b\x73\x90\xf8\x66\x1b\x70\x03\xb7\x88\xdd\x72\x13\x9f\x19\x68\x7f\x84\x60\x82\x19\xe3\xe2\x45\xb0\x7f\x72\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909ContentURIUpgradeable.sol",
            name="ERC6909ContentURIUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909ContentURIUpgradeable.sol",
            name="ERC6909ContentURIUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.34
    b"\x40\x3f\x28\xac\x97\xb0\x47\xde\x8d\x32\x91\x35\xf8\x39\x0e\x7e\xf6\xc9\x26\x40\x27\xbb\x33\xda\x59\x85\x13\x82\x8e\x7b\x53\x6e": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.34",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x40\x80\x2a\xd7\xd1\x23\x33\x1e\x15\x53\x09\x1f\xff\x1f\x3c\x73\x02\x04\x30\x38\x31\x4f\x86\xa8\x5a\xdb\x0f\xf9\x1c\xa3\xc9\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Errors.sol",
            name="Errors",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x40\x9f\x7d\xf3\x7f\x5a\xf4\x42\x4b\xa0\x64\xb6\x38\xcf\x52\x2c\x20\xc0\x68\xce\x2c\x3a\x2b\xd7\x06\x5c\x4d\x24\x2d\x7d\x56\x2c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.137, v0.0.138
    b"\x40\xa7\xe9\xc1\x2b\x07\x7d\xb6\x19\x0b\x3f\xfe\x28\x9c\xa5\x0f\x02\xaf\x68\x4a\xb4\x76\x1a\xff\x58\x18\xc4\x0d\xff\xfa\x83\x19": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.138",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x40\xb0\x75\x20\x6d\x9f\x7f\x08\x26\x05\x26\x61\xc9\xa8\xff\x38\x08\x4d\x33\x94\x44\x91\xb1\x30\x16\x06\xe6\x87\x02\xc7\xfe\xeb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x40\xbc\x0e\x64\x5e\x55\x9f\xba\xee\x4f\xaf\x33\x0e\x30\x06\x54\x46\x2f\xa6\x3b\x02\x6d\xdd\xce\xbf\xd0\x29\x22\x7f\x80\xd2\xdf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x40\xd5\xae\x7e\x5b\xd5\x5a\x52\x3e\xd4\x38\xf2\xb3\xd3\x0c\x71\xab\xd2\x59\xe0\xf3\x63\xab\x8f\x49\xfe\xc0\x81\x01\xde\xbc\x03": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/optimism/ICrossDomainMessengerUpgradeable.sol",
            name="ICrossDomainMessengerUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x40\xe8\x96\x98\x4d\xcf\xc3\x53\x69\x28\xd9\x90\x43\xa2\x83\xcd\x88\x7f\xb2\xb0\x4c\x83\xa7\xf2\x7b\xa1\xf7\x6a\x4a\x2d\x22\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52
    b"\x41\x14\xa1\x55\x10\x20\xf8\xf6\x27\x33\x7e\x1a\xdd\x47\xcd\xac\xa0\xa2\xb1\x48\x31\x7b\x5a\x14\x49\x84\x3c\x07\x2e\xe1\x1a\xef": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.52",
        ),
    ],
    # Hash appears in versions: v4.5.0
    b"\x41\x19\x82\xaa\x42\x56\x93\x76\xcd\xa4\x23\x37\x93\x3b\x7f\xc7\x46\x01\xf8\xeb\xb8\x50\xea\x9d\x67\xf4\x50\xd7\xf3\x4b\x2f\x8c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.5.0",
        ),
    ],
    # Hash appears in versions: v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x41\x2a\x28\x12\xe9\x49\xf0\xe8\x00\x9f\x70\x6e\xbd\x5f\xbf\x96\xe3\xc4\x49\x08\xf3\x4d\x5c\xc3\x4e\xbd\xf4\xef\x76\xe0\x93\x06": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124
    b"\x41\x2b\x3c\x9e\xfb\xbb\x8b\x35\x52\x8d\x05\x35\xc9\xf8\x87\x3c\x36\x66\x92\xc9\x64\xdd\xbf\xd2\x76\xe5\x59\x8a\x70\x67\x24\x2f": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v0.0.220, v0.0.221
    b"\x41\x4e\x45\xb7\x8f\x7d\xb7\x7b\x43\x1e\x3f\x91\x77\xa7\xf1\x75\x35\x7f\x28\x88\xb5\xdb\x21\x6e\xc0\x3f\xc0\xd0\x02\x07\xfe\x4b": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.221",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x41\x5c\x31\x97\x65\x1c\x69\xde\x8d\xd6\x1c\xf2\x06\xf5\x26\xfe\x8f\x52\x4b\x94\x56\xc9\xba\xf9\x42\x83\xf0\xb5\x75\x40\xfc\x7b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x41\x5c\xf1\xa3\x94\xe0\x93\x91\xb9\xce\xfc\xbe\x04\x25\x1d\xb3\x4f\x78\xff\xdf\xab\xf2\xc9\x19\xab\x91\x00\x59\x70\x6e\x18\xcb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x41\x61\x96\x2b\x5e\x8f\xb2\x54\x38\x59\x71\xff\x3f\x72\x2b\xe3\xe4\xee\x9a\xbb\x64\xac\x19\x3d\xde\x50\x43\x4f\x96\xe7\xcd\x2d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x41\x73\x43\x8c\xb6\xfc\x55\x4c\x83\xd2\xde\xa7\xaf\xf4\x4a\xe4\x8e\xb1\x21\xba\x73\xe9\x17\x7e\x98\x2f\xc6\xdc\x4b\xc2\xc1\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Utils.sol",
            name="ERC1155Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260
    b"\x41\x76\xcb\x5a\x91\xac\xb7\x5b\x6b\x75\x36\x53\xbb\x3e\xb7\xf1\x86\x09\x54\x9d\xe7\x2a\x50\xd2\x9d\xea\x88\x04\x48\xb6\x4c\xf0": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.260",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x41\x7a\x71\x60\xaa\x51\x7c\x36\xb6\x9a\xe5\xfd\x6e\xc0\x96\xb2\x46\xeb\xf5\xc6\xd7\x64\xc4\x83\xe0\x3a\x48\x29\x86\x60\x2c\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x41\x87\xc9\xcd\x40\xdd\xbe\xe8\xe2\x37\x2f\x0b\xb9\x63\xd5\x54\x6b\xc1\x06\x71\xb1\xbe\xb5\x6e\xdc\x53\xda\xa5\xd3\x0f\x39\xf8": [
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35
    b"\x41\xa7\x9a\x77\x45\xab\x5a\x2a\x49\x44\x3d\x2c\x26\x68\x95\x86\x3c\x79\x4b\xbd\xa6\xf9\x8c\x37\xb5\x49\xab\xb2\xf9\x42\xf9\x84": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.35",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x41\xad\xd0\xc6\x77\x40\x11\x39\xb3\xe0\x4f\x6e\x0a\x1a\x61\xc8\xbc\xe1\x41\xf2\x47\xb2\x06\x80\xf2\x08\x6c\x52\xac\x51\x5d\xb6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0
    b"\x41\xcc\x84\x20\xf4\xae\x00\xb7\xec\x28\xe6\x11\x86\x0e\xa1\x32\xc4\xd0\xae\xcf\x99\x77\x08\xcb\x1a\x45\x4f\xf9\x49\x58\xfc\x3f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x41\xe0\x3a\x5d\x9b\xfe\xcf\x6d\x64\x44\xdc\x9c\xb0\x34\x43\xd6\x39\xae\xe8\xd6\xe7\x35\xc5\xb4\x96\x0b\x28\xdd\xba\xe5\xb8\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.143
    b"\x41\xe9\xf2\xf8\x9f\xb8\x29\x41\x9e\x30\xeb\xa4\x17\x7f\xac\xb9\x12\x89\xcb\xa2\x12\x0d\xe9\x19\xf1\x48\xaf\xf8\x75\x32\x65\x5a": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.143",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x42\x13\x58\x20\x60\x0a\xe5\x97\xc5\xf5\xa0\x98\x6a\xb5\x59\xae\xf7\xac\x32\xe3\xce\x26\x63\x45\x44\x53\xbe\x2b\x6b\x5f\x68\x41": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x42\x49\x3d\x3b\x0b\x6a\x7c\x75\x92\xf3\x3f\x84\xff\x2f\x0a\x89\xf5\x88\x91\xee\x1a\x7d\x61\xdf\xb1\x40\xd8\x21\x29\xdd\x67\xfc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x42\x49\x98\x52\x8c\x1e\x32\x5d\x54\xdf\x64\xc5\xe8\xdd\xa9\xe6\x4d\x2f\x7c\x9c\xcd\xe9\x53\x36\x11\x1c\xc2\x11\x6e\xcb\x82\x74": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x42\x7e\x55\x15\xbd\x4f\x23\xe2\xc8\x68\xd4\x2a\x69\xc3\x62\x61\x0b\xcd\x8c\xab\xc7\x5b\xbe\x01\xc2\x49\xa6\x52\x0b\xe9\xfb\xc1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x42\xa8\xdd\x91\x72\x0d\xad\xfa\xb6\xb1\x76\x9e\x67\x9d\x42\x15\xc3\x29\xe5\xa4\x09\xd4\x28\x75\xc9\x6e\xdd\x29\x93\x5e\x58\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x42\xbe\x44\x56\xd9\xd3\x69\x60\x2d\xf7\x0f\x3d\x38\xc3\x70\x4a\x42\xbb\x95\xc3\xd5\xff\x92\xec\x80\xc8\x5e\x57\xe9\x4e\x89\x9a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x43\x2b\x4e\x87\x70\x69\x0b\xa5\x12\xa2\xe8\x15\xe6\x91\xd7\x7f\x91\x08\x3a\x79\x1c\x58\x26\xe6\x34\x05\x6d\xd1\x6c\x95\xf7\x79": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelock.sol",
            name="TokenTimelock",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.7.1
    b"\x43\x7e\xde\x1a\xb1\x87\xb8\x8d\x97\x77\x7b\x6f\x56\xd4\x75\xc7\x13\x40\x0c\x61\x94\x3b\xcf\x57\x28\xec\x56\x4c\x13\xfa\xb6\xc7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.7.1",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x43\x83\x6c\x7e\x08\x4b\xf6\xab\x39\xb3\xd4\x08\x32\x0c\x4c\xf5\x69\x28\x18\x68\x72\x02\x1d\xe8\xea\x9f\xcf\x5a\xe8\xe7\xcf\xe3": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/callback/IUniswapV3MintCallback.sol",
            name="IUniswapV3Mi ntCallback",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x43\x8e\x1b\xa7\xd5\x90\x1e\xce\x16\x0f\xc5\xff\x8e\xe1\x3d\x17\xd2\xe5\xbc\x97\x2e\x6c\x43\x47\x78\xe7\xc5\x77\xbb\x05\xfc\x80": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/UnorderedNonce.sol",
            name="UnorderedNonce",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27
    b"\x43\xa2\x0f\xa9\x43\x6c\xc9\xe7\xb6\xb0\xae\x10\x8b\x11\x5c\x91\xe8\xeb\xf1\xd6\xf9\x97\xe9\xe9\x0f\x65\xb6\x56\x5d\x05\x36\x5e": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.27",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x43\xb3\xa9\x40\x91\x7c\x41\xb8\x24\xdd\x91\x09\xdc\x4f\x43\x8b\xd2\x4f\xb8\xec\x2c\xc7\x7d\x32\xfc\xb0\x54\x3c\x0a\x31\xce\xb3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC6372.sol",
            name="IERC6372",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x43\xd4\xe8\x3a\xe4\x8e\x18\xa7\xc5\xd8\xad\xb6\xe1\x01\xb0\xb4\xa5\x8d\x88\x9d\x2c\x03\x51\x7c\xa1\xbd\x35\x84\x5b\x00\x9f\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/ownership/Ownable.sol",
            name="Ownable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/ownership/Ownable.sol",
            name="Ownable",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x43\xf1\xcf\xde\x49\x0c\x76\x72\x24\x20\x4b\xf1\x5e\xf2\xab\xb1\x74\x04\xc9\x05\xe8\xc2\x58\x68\x96\x13\xab\x9b\x5c\x8d\x41\xe6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Checkpoints.sol",
            name="Checkpoints",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.2
    b"\x44\x56\xb1\x70\x6b\x6d\x34\x0f\x06\x56\xf9\xef\x4e\x65\x08\x9f\x63\x35\xfd\x61\xdb\x4e\xb4\x4c\xc8\x39\xdf\xe5\x75\xb3\x27\x97": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v3.4.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x44\x72\x47\xa8\x3b\x8e\x51\xa5\x55\xcf\x47\x0d\xee\x71\x4e\x35\x40\x3f\x86\xa9\x08\x9e\x70\x4e\x84\x52\xf0\x23\x69\x85\xcd\x37": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x44\x7d\x03\xbd\xb1\xd2\x20\x1b\xe0\x8d\x58\x6f\x65\xa2\x76\x6a\xc7\xca\x46\xc3\xc6\x56\x18\x57\x81\x55\x4f\xe5\xc3\xf8\xa3\xd4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x44\x94\x4e\xf2\x13\xd2\xba\x75\x4d\x8c\x50\xc1\x64\x52\x88\xa9\xb0\x33\xac\x93\x2f\xb4\xa9\xa7\xda\xce\xae\xca\xd3\xf4\xbd\x10": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.269
    b"\x44\xae\xd5\x23\xa3\x7e\x4d\x48\x1b\x81\x20\xae\x53\x1d\x65\x4a\x83\x11\x69\x58\x73\xdb\x04\xaa\x94\xc3\x1e\x8d\xc2\x51\x50\x71": [
        KnownContract(
            project="solady",
            path="src/accounts/MinimalBatchExecutor.sol",
            name="MinimalBatchExecutor",
            version="v0.0.269",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0
    b"\x44\xc6\x13\xf3\xae\x38\xa3\x18\x15\xfa\x6f\xde\x47\x4e\x81\x83\x65\xb4\x12\xec\x2a\x28\xa9\xec\xbf\x55\x3a\xf3\x13\xda\x0b\x6a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243
    b"\x44\xdb\x2f\x6d\xc9\x17\x21\x49\xcf\xe9\xfd\xbc\xd0\xf0\x25\x03\xac\xca\x61\x40\x2f\xb5\xec\xb5\x8c\x29\x93\x67\xad\x42\x8b\x95": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.243",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x44\xf2\x94\xa3\x5a\xe6\x2c\x87\x39\xbe\xa3\x10\x89\xbf\xf1\x90\x9d\x80\xe6\xc6\xa7\x0f\x41\xd2\x48\x7a\x20\x98\x62\xf1\x91\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x44\xff\xa5\xd1\xe5\x5d\xa3\x97\xdd\x69\xbd\x61\xf6\x39\x48\xe7\x52\xc3\x60\x9b\x33\x55\x4b\x2c\x6f\xa3\xe9\x8f\x74\x75\x64\x94": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x45\x1a\x54\x46\x91\x5e\x7d\x06\xfa\x44\xf1\x92\x0f\x62\x35\x53\x68\x45\xc8\x4a\x0a\x1f\xad\xde\x8a\x33\xc1\xb5\xec\xd7\x6c\x66": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x45\x2b\x26\xf3\x6f\x17\xd7\xd1\x34\xa9\xe2\x9f\xab\xa9\xc8\xb7\x31\x77\x61\x9c\x49\x22\x00\x3e\xae\x93\x2d\xa2\xb9\x9a\xda\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x45\x55\x1e\xe2\x6f\x2d\x49\xff\x60\xee\xb6\xe4\x29\xd3\x5e\x84\xc8\x52\x75\xb0\x48\x72\xfb\x1b\x82\x8f\xa8\xab\x5f\xc7\x04\x14": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x45\x5c\xd3\x20\x11\x88\x3a\x12\x20\x76\xde\x2c\x2c\x01\xf5\x2d\xef\x9f\x28\xac\x0a\xe1\xc0\x74\x58\xb8\xa8\x5c\xbf\x70\xbf\xcf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248
    b"\x45\x6e\x74\xa6\xa2\xa4\x26\x73\x23\x86\x03\xef\x6c\x0c\xec\xb3\xb9\xea\xaa\xe4\x3e\xd1\xfe\xd2\xac\xa8\x32\x91\x9a\x3e\x38\x48": [
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.248",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x45\x74\xf0\x22\xda\x6e\x84\x20\x5f\x59\x2a\x87\xe0\xf2\x53\x47\x2d\x96\x73\x23\x1b\x42\x48\x3d\xce\x5c\x38\x15\x2f\x87\x3b\x90": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="ICompoundTimelock",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x45\xac\xb0\xe9\x11\x8b\x91\xaa\xcb\xed\x8e\x4d\x95\xc5\x9a\xf1\x09\x72\xa6\xb8\x67\x1a\xcf\x6c\x0e\x10\xa7\x7f\xdd\x63\x3c\x48": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x45\xb0\x49\xfc\xa5\x6e\xbd\x9d\x7d\x63\xdd\x60\x0d\xd8\xcf\xfa\xe1\x95\xef\x23\x44\x34\xed\x68\x1c\x6d\x23\x60\xfc\x5e\x47\xa9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x45\xc3\xf1\x79\x0f\xa0\xec\x90\x98\xb6\x1a\x1f\xdf\x2a\x11\x95\x0d\xb8\x00\x07\xd8\x0c\xd4\xc5\xab\xed\xa4\x08\x57\x52\x2c\xb1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x45\xcb\xb9\x01\x08\x93\x2b\x3c\x38\x86\x5b\xcd\xa2\x8d\x1a\xa1\x74\xbe\xf5\x1f\x89\x63\x84\x9b\xe0\x77\x5a\x27\x16\xb7\x29\x8d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1
    b"\x45\xd0\x91\xbf\xa7\x2b\x95\xa1\x5a\x74\x6e\x41\xd1\x29\xea\x0f\x82\xd3\x71\x5b\x89\x46\x7e\x84\x0b\x86\x6f\xe4\xdb\xc2\xfb\x57": [
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.1.1",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.97, v0.0.98, v0.0.99
    b"\x46\x23\xf3\x31\x84\xc9\xc6\x72\x06\x06\x69\x74\xc0\x40\xb7\x05\x0c\x4d\xd9\xb4\xa5\x0d\x37\xed\x45\x62\x45\x49\xbf\x74\xaf\xc4": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x46\x25\x67\x51\x78\x5d\xe6\x54\x96\xa9\x10\xfd\x96\xe1\xe0\x55\x5c\xa6\x8a\x0b\x0e\x47\x42\x77\x14\x6c\x87\x89\xb3\x00\x9f\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Checkpoints.sol",
            name="Checkpoints",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x46\x6e\x8b\x65\x6b\x24\xb9\xd8\xd3\xb2\xbe\x02\x24\x07\x65\xf7\xf8\x3e\x54\xf2\x0d\xa9\x4b\x30\x8d\xfe\x5d\x8b\xad\xc6\x08\xe1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x46\x77\x49\x09\x67\xe6\x94\xc2\xaa\x84\xb5\xef\x4b\xd3\xc3\xa2\x5c\xd4\x74\xca\x0b\xd9\x61\x58\xd6\xa6\x4e\x4f\xd6\x57\xf1\x9c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/Escrow.sol",
            name="Escrow",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: main
    b"\x46\xcc\xf1\xd1\xb3\xf8\x91\x52\x2e\xc8\x24\x9a\x93\xa3\x2b\xf2\x23\x44\xa2\xff\xda\xee\x35\x81\x65\xc4\xf9\x57\x3e\xa7\x7b\xd1": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/ERC721Permit_v4.sol",
            name="ERC721Permit_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x46\xe3\x27\x82\x54\xa9\xa8\x0a\x45\x15\x4d\x1b\x32\x8b\x13\xa0\x83\x43\x29\x90\xb9\xcf\x3e\x62\xd3\xb1\xe8\x95\x02\xda\x7a\x85": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x47\x19\x86\x8a\xd5\x91\x52\xc3\xb9\x5e\x80\xe1\x52\x70\x34\xe8\x34\xff\xfc\x01\xcc\x40\x7d\xb4\x0d\x73\x0c\xe2\x63\xa8\xeb\xeb": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.59, v0.0.60, v0.0.61, v0.0.62
    b"\x47\x22\x2a\x87\x80\x68\x9d\xd1\x85\x6d\x1d\xad\xe7\xe4\xd8\xf3\x66\x72\x72\x38\x7f\x8f\xb0\x96\xe5\x09\xbb\x05\xd7\xfe\x0b\x4d": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.62",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x47\x83\x83\x85\x40\xda\xf6\x22\x4f\xf9\xbd\x11\x22\x27\xf0\x1a\x89\x25\x2f\x8e\xd6\xa8\xd4\xf6\xa6\x21\xa7\x29\xf3\x21\x36\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296
    b"\x47\x8a\xa5\x0f\x5b\x36\x20\x6a\x5c\x5c\x25\xb6\x0d\xa6\x94\x22\x83\x86\xbb\xef\x17\x44\x82\x95\x4b\xa2\x8d\x7c\x3f\xbc\xdd\x52": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.296",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x47\xa5\x76\xaf\x0d\xf4\xf1\x7d\xce\xe9\x9a\x30\x08\xe0\x59\xe4\x0a\x9f\x58\xc7\x29\x81\x00\xdf\x81\x19\xfb\x19\x41\x43\x4f\xa6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.285
    b"\x47\xab\x9e\x6f\x86\x78\x74\xec\x77\xb6\x3a\xc3\x85\x1f\x29\xc4\xa3\xcc\x02\x08\x80\x6e\xa5\x8e\x4c\xc8\xf4\xa8\x23\xf2\xda\xfe": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.285",
        ),
    ],
    # Hash appears in versions: v0.0.243, v0.0.244, v0.0.245, v0.0.246
    b"\x47\xb6\xa9\xfb\xe2\xdd\x13\x49\x9d\x23\x01\xce\x34\x63\x26\x96\x78\xc2\xc1\x08\x72\x7a\x6b\x9c\x2c\xf5\xf7\x00\x3f\x67\xd3\xc3": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.246",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x47\xb8\x18\x71\x20\xf2\x04\x37\x9f\xec\xef\x44\xd4\x6e\x29\x9e\x41\x42\xc7\xc5\xa6\x05\x0f\x8c\xf5\x7a\x9e\xad\x2a\x58\x0a\x0c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x47\xc9\x46\x59\x79\xbe\x9c\x83\x64\x07\x5a\x25\x03\x78\x11\x05\xc8\xbf\x75\xf2\x5c\xf5\xf6\xdb\xd3\xc3\x16\xac\x6b\x5f\x4e\x4b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalGuardian.sol",
            name="GovernorProposalGuardian",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalGuardian.sol",
            name="GovernorProposalGuardian",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0
    b"\x47\xce\x75\xd7\xd0\x70\xa0\x8c\xaf\x4c\xa7\xe0\x67\xdc\xb3\xb7\x38\x54\xb5\x44\x18\xa1\x66\x08\xff\x69\xf2\xeb\xdc\x9c\xe6\x8a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.8.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268
    b"\x47\xdc\x6e\x2c\x86\x77\xe1\xc6\xd5\xdf\xdf\xe4\x7b\xd6\xad\x63\x5b\xa2\x14\xfa\x83\xd3\x8c\x97\xaa\x66\x74\xb0\xc0\x66\x79\xe9": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.268",
        ),
    ],
    # Hash appears in versions: v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68
    b"\x48\x5b\x2c\xa8\x0a\x9a\x2b\xa7\xce\x6c\x4d\x84\x44\x86\xb9\xea\x0e\xa7\xbf\x92\x3f\xd7\xcb\xb7\x44\xa7\x43\x67\xf7\xa3\x60\xcb": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1
    b"\x48\x5c\x97\x6e\xc8\x96\xe2\x89\x0e\x61\xf0\xbb\xa1\x23\x17\x57\x01\xd4\x63\x7d\xc2\x60\x1a\xd9\x2b\x4d\xb3\xc1\x4c\x2b\x82\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.1, v5.0.2
    b"\x48\x7b\xb0\x8c\x3d\x19\x32\xe2\x12\xc4\xbf\x39\xa9\x11\x19\x72\x57\x35\xe0\x23\x15\x85\x4c\x9c\x31\x9d\x03\x00\x7b\x5b\x10\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x48\x82\x0f\x55\x79\x71\x69\x31\x95\x70\xea\xaf\x8d\x79\x89\x07\x8a\x88\xe2\x95\xc8\x53\x9f\x5d\x1a\x51\x90\xd5\x2b\xb7\x58\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x48\xbd\xce\x2c\x0e\x97\x90\x45\xef\xcd\x57\xb4\xe7\x4d\x8a\xab\x9b\x7f\x01\x54\x8e\x6f\x25\xd1\xd8\x50\xf4\xa8\x77\x56\x5f\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x48\xc1\xc5\xe4\x94\xe0\xc9\x20\x7d\x25\x8e\x85\x1b\xb4\x03\xe2\x45\x25\xac\x0c\x63\x48\x6c\x2b\x7f\xff\xa4\x7e\x5a\xe4\x9b\x47": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20SnapshotUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x48\xd7\xca\x15\x1c\xd3\xf5\x9e\x15\x49\xb9\x99\xa4\x79\xd0\x56\x68\xa2\x1b\x48\xa1\x0b\x39\x14\x00\x51\x4e\x8e\xd0\xe0\x4a\xdd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x48\xf8\xf5\xc7\xc0\xb0\x7e\x48\x66\x95\x34\x6e\x52\x36\xf7\x3b\x7f\xa4\xc0\x5c\xa8\x78\x64\x53\x90\x63\x46\xc3\xfd\xb7\x1e\xc2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.247
    b"\x49\x53\x82\x73\x49\x27\xcd\x8e\x1a\x19\x8c\x0f\x98\x8e\x28\xca\xe5\x45\x93\x8f\xc6\x4c\x39\x25\x93\xa4\xa4\xc5\xd8\x22\x05\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.247",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x49\x5c\x82\x77\x69\xd3\x82\x85\x90\x4f\x3c\xa5\x2f\x81\xe0\xfe\xb0\x12\x0d\xe3\xf3\xbf\x41\xd4\x47\x4f\xa9\xff\x8f\x0e\xc6\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x49\x5e\x5d\xc8\xb3\x72\xa0\x20\xc8\x3e\x5a\x7c\xe5\x3a\x4f\xe0\x75\x08\x52\x76\x0a\xac\x3c\x6c\x83\x6d\x2a\x06\x10\xb6\x7b\x5b": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551Proxy.sol",
            name="ERC6551Proxy",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x49\x71\xd9\x02\x13\xa3\x3a\xfe\xfc\x7c\x62\x9b\x28\x2f\xf5\x1c\x56\x65\x1e\xa3\x9d\x45\x80\x7e\x0f\xad\x65\xb8\x84\xd7\x4f\xf1": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x49\xa4\x5d\xd7\x9d\x73\xae\x87\x05\x1c\x04\x41\x52\x9d\xb8\x7a\x39\xb6\x20\xfc\x35\xdd\x4c\x41\x08\xb0\x40\x63\x47\xdc\xb5\xd1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x49\xe1\x9f\x73\x74\x30\x60\xac\x56\x8d\x44\x7a\xa5\x42\xab\xb2\x33\x07\x13\xb2\xaa\x1d\x15\x02\x0f\x6c\x3c\x18\x91\x6e\x66\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x49\xe1\xd3\xf7\x2c\x5d\x10\x53\xe0\x4b\x7b\xd1\x9c\x00\x79\x1f\x8e\x75\x9a\x48\x77\x88\x69\x64\x53\x7b\xc0\x7e\x14\xac\x40\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v1.2.0
    b"\x49\xe2\xaf\x79\xd8\xaf\xaa\x69\xa8\xf3\xa8\x25\x23\x5b\xcb\x1e\x15\x5a\x2c\x23\x71\x14\x56\x57\xa6\xdb\x40\xa6\x51\x13\xb2\x36": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/OracleLibrary.sol",
            name="OracleLibrary",
            version="v1.2.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x49\xec\x95\x3f\x76\xe4\x0b\x03\x08\x86\xd5\x13\x27\xe7\x02\x90\x38\x12\x83\xd2\x0d\xe5\xa8\x7d\x85\xe7\xbd\x6b\xe0\x4f\x50\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlDefaultAdminRules.sol",
            name="IAccessControlDefaultAdminRules",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x4a\x05\x5d\x8c\x7b\x8c\x0d\xdc\x73\xe8\x18\x50\x84\x02\x8a\xa7\xab\xe5\xd2\xed\x27\xe4\x34\xe6\xa2\x82\x3f\x5c\xbd\x04\xfd\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManaged.sol",
            name="AccessManaged",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x4a\x64\x09\x05\xa3\x59\x34\xa7\x19\xac\x05\x56\x8f\xc6\x18\xd7\x8b\xaf\x42\x26\x10\x55\xac\xd2\x3f\x80\x94\x17\x3b\x48\xf1\xf6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.41, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49
    b"\x4a\x69\xa4\x0b\x7a\xac\xbf\xeb\xc4\x4f\xe9\x47\x91\x77\xb5\x82\x94\x4b\xb3\x14\xf7\x12\xe3\x50\x5e\x1b\x0b\x47\x92\x6b\x43\x0b": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.49",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x4a\x79\xcf\xd5\x8c\xf0\xa9\xe0\x24\x18\x43\xeb\xe6\x7d\x9d\xb8\x33\x32\x31\x09\xec\xc3\x3a\x42\x42\x98\x7d\xfc\x02\x52\x7d\xd3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4a\x83\x0e\x67\x44\xdb\x2a\xdc\x7d\x16\xc4\x89\xfe\x1c\x91\x1c\xaf\x3c\x59\xc3\x33\xdd\xca\x53\x9a\xdc\x9d\xf6\xc9\x21\x67\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x4a\x85\x9b\xbb\x00\xe6\x37\x2f\x35\xe0\xb6\xe7\x51\x86\xa4\xba\xaa\xcc\x62\x7d\x71\xb3\x74\xa4\x8f\x2d\xa1\x89\x1d\x46\x00\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275
    b"\x4a\xb7\xe1\x9a\x28\xd2\xa0\x06\x47\x70\xfd\x2a\xd0\xc4\x11\x8c\xe8\x53\x09\x02\x30\x78\xcd\xa8\x5b\xd4\x7f\x75\x4b\xe3\x6a\xc3": [
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Receiver.sol",
            name="Receiver",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v0.0.68, v0.0.69
    b"\x4b\x14\xee\x58\x43\x76\xc8\x4f\xcc\xbb\x5a\x7f\x3d\x21\x8b\x8a\x5a\x4b\x7d\x52\xfa\xdd\x2f\x5f\xff\xda\x93\x26\xc2\x05\x4f\x1e": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.69",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x4b\x1f\x5f\x6e\xa5\x06\xf6\xbe\x60\xb2\xbc\x49\x02\x4c\xa0\xba\xe3\x65\x96\x61\xbf\xf3\x8e\xd9\x91\x9d\x12\xc8\x1b\x78\x2d\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4b\x20\x0a\x22\x49\x20\xad\x7d\xad\x94\xe6\xb4\xd1\xbd\x81\x49\x5a\x84\xb0\x4f\x84\x80\xfa\x65\xaf\xce\x44\x0c\x1f\xa4\x8d\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.54, v0.0.55
    b"\x4b\x25\xb1\x9f\xf4\xd4\xd5\x06\x9f\xb2\xb5\x8c\x3c\xf7\x4e\xfb\xd4\x8b\x07\x29\x9b\x45\x7d\xf2\x8c\x06\xa0\x9c\x08\x58\x6e\xb2": [
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x4b\x25\xf1\x22\x20\xf1\x9a\xef\xe5\x25\x3c\xdb\x83\x26\x0f\x12\x2c\x36\xb9\x2a\xf2\xc1\xed\x74\x98\x7b\x37\x71\x1f\x5d\xfa\x20": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.16, v0.0.17, v0.0.18, v0.0.19
    b"\x4b\x5a\x31\x44\x6e\x94\xaf\x25\x7b\x2e\x0c\x1e\x24\xbe\x84\x13\xb1\x1f\x75\x95\x92\x4e\xa1\xc6\x33\xd4\x70\x4c\x40\xc7\x68\xfd": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v0.0.165
    b"\x4b\x7a\x3a\xba\x1b\x23\x18\xf0\x2d\xd4\xc5\x81\x4d\x41\xfe\x49\x8a\xd4\xa7\x49\x8a\x9f\x73\xd8\x41\x84\x3f\xef\x44\x97\x10\xc2": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.165",
        ),
    ],
    # Hash appears in versions: v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x4b\x7d\xae\x5c\xca\x8e\xa3\xfa\x6c\x2b\xd5\x7a\x54\xaa\x0e\x68\x97\xac\x25\xa6\xc9\xd2\xc3\xbc\x54\xaa\x85\x30\xee\x94\x40\x2c": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4b\x7e\x75\x92\x43\x8c\x94\x27\xaf\x1d\x85\x0a\xec\x5d\xcf\x81\x5f\xd2\x83\xe5\x44\xcd\x11\x17\x1b\xac\x85\x3b\x2f\x3c\x35\x52": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x4b\x86\x5a\xcc\x12\xe4\x83\x72\xe8\xb7\x9b\x72\x40\x05\x97\x96\xc6\x39\xb0\x87\x16\x0f\xf7\xc2\xcc\x2d\x4e\xb7\x55\xf5\x60\x2d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4b\xa1\xa7\x67\xd1\x05\x8b\xb8\x81\xf3\xf8\x9d\x23\x2f\xa9\xa9\x2d\x7b\x6a\x0e\x7f\xaf\xe6\x3e\x47\xad\x61\xad\x3d\xe1\x21\x7c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2
    b"\x4b\xa8\x0e\x89\x22\x79\xc7\x65\x9c\xb6\x01\x06\xc0\x46\x3d\x3b\x04\xaf\x86\xf7\x48\x70\x53\xa6\xc6\x9c\x54\x28\x5a\xa1\x71\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x4b\xaf\xd1\xaf\x4e\x14\xc0\x45\xa0\x67\x52\xd2\x66\xc1\xea\x0e\x61\x28\xf1\xea\x8b\x3d\x79\xda\x2a\xd7\xe4\xf3\x68\x97\x9d\x2b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x4b\xba\x0b\x92\xcf\xcc\x54\xc9\xa0\x81\xb1\xde\x32\xf2\x61\xe5\x7c\x2c\x59\x57\x95\xe9\x64\x58\x74\x70\x85\xf5\xc3\x13\xa9\x70": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="ERC4337Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="ERC4337Utils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="ERC4337Utils",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x4b\xcc\x24\x57\xbb\x68\x91\x3c\x36\x7a\x90\xbb\x29\x83\x94\xd4\xe0\x9c\xa9\x3b\xce\xb6\xdf\xaf\x64\xd6\x3f\xea\x5d\x46\x32\x0b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50
    b"\x4b\xd9\x7c\x2d\x95\x10\x9a\x48\xba\x6a\x2a\x75\x31\x56\x59\x95\x92\x26\x1d\xcd\x2d\x28\x1e\xe1\xff\x2e\xfc\x69\x6e\xda\x0f\xba": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.50",
        ),
    ],
    # Hash appears in versions: main
    b"\x4c\x07\xde\xb1\xaa\x18\x14\x44\x24\xdb\xe9\xe7\xb3\xe3\xf9\x2f\xdc\x33\x48\xae\x7e\x02\x75\x94\x41\xbd\x87\xde\x49\x2d\x54\xc5": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/hooks/WstETHHook.sol",
            name="WstETHHook",
            version="main",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x4c\x0a\xb5\x1d\x93\x58\x89\xc8\xc1\xda\x8d\x16\x45\xd3\xb8\xa6\x93\x6a\x24\xc6\x2c\x47\x0e\x95\x8b\x8b\x54\xc6\xd5\x95\x0c\x74": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="base64-sol/base64.sol",
            name="Base64",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="base64-sol/base64.sol",
            name="Base64",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="base64-sol/base64.sol",
            name="Base64",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="base64-sol/base64.sol",
            name="Base64",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="base64-sol/base64.sol",
            name="Base64",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\x4c\x12\x28\x05\xc0\x87\x0c\xd0\xdf\x4f\x19\x34\x5c\x47\x01\x20\xd1\xb4\xab\x65\x5a\x81\x6d\xeb\x1e\x6e\x02\xda\x8c\xd3\xee\x4e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x4c\x45\xbf\x50\x7a\x52\xb4\x3e\xec\x14\xc0\xbc\x4d\x0c\xe6\x73\xc0\xcb\x4a\x97\xff\xdf\x36\x0f\xcd\x72\x1b\xfc\xaf\x34\x6e\xe1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x4c\x47\xb8\xf4\x65\x66\xfe\x92\x21\x79\x45\x1d\xd8\x71\x13\x2f\x5a\x0f\xd5\x93\xa3\x04\x7e\x23\xab\x20\x81\xc7\x59\x25\xc0\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.87
    b"\x4c\x56\x14\x35\x8b\xc9\xb7\x2f\xad\xc9\x8b\xe1\x09\x57\x56\xdb\x7a\xd9\xe6\x5a\xa2\xb6\x62\x5a\x25\xd9\x5d\xbd\xfd\x6b\x58\xa0": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.87",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x4c\x56\xdc\x14\xdd\x59\x2f\xe9\xd4\xce\x3a\x1a\x69\xf3\x78\x64\x6d\x7a\xae\xb7\xc7\x21\xee\x10\x7c\xb7\xb6\xf1\xf5\x71\xec\xc7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalGuardianUpgradeable.sol",
            name="GovernorProposalGuardianUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalGuardianUpgradeable.sol",
            name="GovernorProposalGuardianUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.211, v0.0.212
    b"\x4c\x5a\xb5\x65\xc1\xba\x26\x1b\x4e\x61\xa5\xa1\xd2\xee\x99\x54\x5a\x19\x9a\x58\x87\xa9\x68\xaf\x0e\xe5\xd4\x35\x03\x03\x2f\x3f": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.212",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x4c\x60\x46\x32\xcc\x80\x7e\x9b\x8e\x2d\xc6\x75\xd1\x0f\x05\xb0\x3e\xd6\x5a\x6a\xff\x48\x36\x63\xa3\xac\x9c\xec\x5c\x06\xd8\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x4c\x8c\xac\x80\xff\xa0\x8d\x06\xe7\x29\x4f\xe9\x50\xe4\x13\x44\x91\x53\x34\x25\x93\xb3\xef\x2c\xb7\x42\x56\x5c\x03\xb1\xb5\x23": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x4c\x8d\x68\x96\x09\xe6\x5b\x49\xf3\xcd\x11\xdb\xd8\xd9\x07\x9a\x7e\xd4\xb8\x8e\x71\x3a\xdc\x68\xfc\xac\xad\x5c\x9d\x10\xd5\x39": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x4c\x94\x02\x91\x66\xf7\x98\x6a\x84\xea\x1f\xa5\xba\x78\x92\xc8\x95\x81\xfa\x61\x38\x7b\x04\xda\x2d\xab\xa7\xb6\xf0\x2f\x4d\x20": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.273, v0.0.274, v0.0.275
    b"\x4c\xa1\xa2\x6e\x58\xe4\x4e\xef\x86\x1b\x00\x11\x22\xd0\x24\xe1\x39\x0a\x67\x5f\x2c\xe2\x88\x07\x74\x8e\xf1\x15\x9f\xe5\xaf\x87": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4c\xae\x39\x85\x56\xbb\x14\x87\xda\x3a\x73\xa6\x9b\xd7\x03\xe2\x32\xb3\x4e\x57\x56\x0b\x96\x2e\x56\xb5\xd6\xec\x7d\xac\xf7\xe8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0
    b"\x4c\xbf\x73\x17\xd4\x7b\x7f\xc6\x24\xeb\x6d\x56\xf8\x21\x34\x38\xe9\xd3\xc5\x9b\xa5\x76\xe3\xc0\x0f\xb5\x3d\xf6\x2f\xb8\xb9\x0f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.5.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x4c\xec\xbe\xbb\x62\x0e\xb3\x96\x86\x7d\xba\xe2\xbf\x93\x3a\x50\xcc\x02\xa1\xe9\x3c\x9e\xbc\x2b\xa1\x23\x3e\xdd\xef\x81\xde\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4c\xf0\xc9\xf4\x61\x34\x9b\x33\xf9\xbf\xef\x2b\x3e\xff\x05\x5d\x01\x47\xd8\x0b\xb2\x10\x67\x60\x12\x69\xe2\x8e\xfb\x11\x4b\x9f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x4c\xfe\xfc\x68\xf4\xbc\x41\x68\x57\x2f\xa8\x9e\x18\xa4\x01\x11\xf5\xff\xc2\xae\x15\x43\x5c\x26\x8a\xd2\x96\x51\xfe\xb0\x92\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70
    b"\x4d\x15\xdd\xac\xb7\xbd\x1d\xca\x46\x93\xc5\xc0\x27\x2c\x5b\x9f\x70\xec\x99\x44\x8f\x3b\x41\x31\x37\xa7\xf2\x6f\xc2\x3f\xd5\xac": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.70",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4d\x25\xbb\xa7\x11\xb4\xb5\x06\x3d\x53\x45\x5a\x23\x1c\x73\x59\x1b\x61\xc2\x4c\xec\x58\xa5\x3d\x1e\x85\xb8\xfc\xb2\xd5\x70\x11": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/NoncesKeyed.sol",
            name="NoncesKeyed",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/NoncesKeyed.sol",
            name="NoncesKeyed",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/NoncesKeyed.sol",
            name="NoncesKeyed",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/NoncesKeyed.sol",
            name="NoncesKeyed",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/NoncesKeyed.sol",
            name="NoncesKeyed",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.182
    b"\x4d\x3e\xf1\xf6\x60\xba\x47\x32\x4e\x50\xb6\x1e\xc3\x31\xd8\x34\xa7\x86\x08\x12\x54\xe1\x0a\x83\x6f\xe6\x21\x4d\x02\xb5\x61\x5b": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.182",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283
    b"\x4d\x43\xc9\xbe\xb6\xb1\xe8\xb7\xa9\xa0\xfd\x1b\x81\xf7\x56\xfc\xa9\x65\x6a\x0e\x66\x0e\x40\x86\x73\x78\x2e\x74\xbc\xd3\x6b\xcb": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.283",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4d\x52\xb7\x55\x8c\xed\x6b\xcd\x1a\x6f\xd1\xa4\x57\x51\x7f\xa2\x87\xfb\x66\x69\xf8\xea\x34\xce\xd0\x3c\xb9\x41\x75\x12\x6c\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x4d\x58\xe9\x2a\xdf\x6e\xe5\x83\x4e\xd7\xaf\x34\xb9\xd7\xd7\xca\x99\x58\x1d\x3a\x0e\xfd\x1b\xe2\x03\xcd\x98\xdd\x95\x7f\x56\x9a": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/CurrencyRatioSortOrder.sol",
            name="CurrencyRatioSortOrder",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x4d\x76\xd1\x07\x5d\xec\x9f\x22\xb4\xe3\x89\xdf\x20\x4f\x7e\x34\xe1\x27\x25\xf4\x6e\xc4\x58\x2d\xe8\x95\x9a\x9e\x3a\x45\xb4\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x4d\x81\x6a\x84\xe0\x97\x23\x5d\x8e\x48\xf0\xab\x59\x98\x36\x8f\xa0\x04\x11\xe3\xd8\x71\xe1\xe1\xa4\x21\x82\xe9\x62\x40\x6d\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v6.7.0, v6.8.0
    b"\x4d\x86\x12\x0c\x95\x01\xba\x64\x3f\x73\x7e\xf6\x81\x00\xbe\x68\x89\x5f\xc3\xec\xa5\xb3\x63\xd0\x4c\x68\x9a\x2e\x53\xdf\x3b\x00": [
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x4d\x9f\x5b\xa4\xcb\x2e\x0a\xb3\xef\x64\x72\xe0\x90\x02\x85\xca\x1d\xdc\x9d\xd2\x1e\x59\x3e\x41\xe2\x6c\x17\x16\x30\x8c\xa5\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.1.15
    b"\x4d\xa6\x05\x44\x80\x71\xce\x15\x56\xae\x17\x07\x86\x07\xa5\xda\x84\xfc\xd3\xe8\xf0\x6d\x5c\x23\x24\x99\x36\x60\x52\x55\x76\x29": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.15",
        ),
    ],
    # Hash appears in versions: v0.0.166, v0.0.167
    b"\x4d\xd1\x28\x97\x9e\x4b\xcc\x72\x96\x37\x64\x05\x2f\xd6\x5c\x3d\x5f\x70\x1a\xa3\xbd\xd9\xe7\x10\x18\x7e\xe1\xf8\x07\x5d\x64\x86": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.167",
        ),
    ],
    # Hash appears in versions: v0.0.261
    b"\x4d\xe0\x0f\x59\x13\xf5\x6a\x87\x3c\x79\x8a\x11\xdd\xec\x54\x6b\xa4\xfd\xb3\x73\x54\x2b\x25\x8e\x61\x79\x8a\x80\x4b\xda\x1d\x91": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.261",
        ),
    ],
    # Hash appears in versions: v0.0.217
    b"\x4e\x10\x59\x22\x60\x8c\x26\xa3\x2f\xb8\x32\x2e\xcb\x95\xf9\x88\xa7\xf1\x33\x97\x41\x04\xd0\x31\xa9\x83\x3f\xc7\x91\x35\xe7\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/ext/delegatexyz/DelegateCheckerLib.sol",
            name="DelegateCheckerLib",
            version="v0.0.217",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4e\x13\x96\x4d\xcd\xa7\x0a\xc1\xa1\x7b\x05\xea\xa5\xa6\xc9\x30\x6e\xfc\xb4\x1b\xed\x86\x0d\x09\xd6\x60\x77\x26\x8c\x95\x9c\xb6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7674.sol",
            name="IERC7674",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4e\x36\x9e\x24\x39\x12\xe3\x8b\x8d\xb1\xdf\xa4\x71\x17\xbf\xb3\x40\x01\x68\x77\xb3\x4a\x6f\x8b\xf0\xd5\xbd\x6d\x5c\x29\xda\x71": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x4e\x3a\xf3\x47\x8f\x5b\x06\xf5\x29\x65\xc3\x49\x30\xa1\xfe\x48\x15\x2a\x13\x72\x4b\x43\x51\x11\xae\xb3\x71\x60\x89\xe7\xd6\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x4e\x3e\x73\xea\x46\xda\x9a\x09\x45\xf6\xe7\x7a\x5f\xb6\xb5\x93\x9b\xdf\xfb\xbe\x52\xb2\xe8\x5b\x88\xb0\x5c\x8d\xc1\x01\x30\x1f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x4e\x70\x3b\x1f\x3a\x55\xfa\x5f\xe6\xd7\xbe\x47\xf0\x6e\x83\x6d\xdb\x85\x28\x32\x67\x5e\x89\xed\x4e\x11\x1e\xa2\xb7\xd2\x7c\x13": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x4e\x9c\xf8\xdf\x93\xba\xb3\x1f\x31\xc8\xb0\x0b\x08\x7a\xb6\x80\x19\x93\x8a\xa8\x55\xb4\xdb\x31\xb1\x1a\x10\x8a\xa5\x57\x17\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4e\xcb\x74\xce\x44\x27\x04\xa2\x2d\xa0\x07\x75\x33\x27\xca\x9a\x14\xf2\xe4\x9a\x6d\x1a\xf4\x56\x6c\x8c\x62\x25\x18\x11\x0f\x9a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x4e\xe0\x52\x7b\xde\xf1\x45\x7c\x50\xb8\x8d\x38\x52\xe6\x56\x53\xda\x57\xbb\x53\x77\xc1\xaf\x8b\xb2\x98\x37\x3c\x0f\x58\x04\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.224, v0.0.225
    b"\x4e\xeb\x80\x17\x4b\x09\x5f\x20\xf2\x1d\x39\xeb\x31\xba\x13\x50\xe7\x89\x9a\xd4\x84\x51\x73\x14\xc5\x0b\x7b\x97\x31\xea\x46\xf5": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.225",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x4f\x19\x75\xc4\xd1\xe5\xd5\xe7\xb3\x2f\x62\xb0\x17\xea\x22\x70\x24\xde\x2a\xbf\xcf\xd8\x18\xba\x48\xff\x6a\x07\x7b\x3b\x87\xba": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Implementer.sol",
            name="IERC1820Implementer",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x4f\x38\x02\xbb\x1b\xfd\x2d\x8f\x40\x1b\xf5\xc3\xe5\xad\x62\x34\xe4\x61\x50\x8c\x45\x85\xef\x6b\xa6\xdd\x01\xe2\x2c\x44\x03\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x4f\x40\xe8\x6a\x6e\xcf\x61\x92\x03\x4e\x59\x4c\xc8\xfb\xef\xb2\x8d\x0d\x75\x6c\x3d\x22\xd7\x7c\x66\x1a\xf7\xfe\xd0\x7a\xa4\x45": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.21
    b"\x4f\x4b\xd7\x4f\x2d\xf6\x90\xfa\x92\x3b\xcc\xc5\xb8\x6b\x7d\x1b\x0d\x56\x22\xfa\xcf\x6c\x1a\xfc\x18\x47\xd9\x51\x21\xb4\x03\xef": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.21",
        ),
    ],
    # Hash appears in versions: main
    b"\x4f\x57\xae\xf7\x56\xc9\x27\xbc\x40\x9f\x71\x45\x72\xda\xe2\x9c\xa7\xba\x38\x66\xf0\x75\xca\x73\xe7\xf0\x66\x68\xb8\x1b\xde\x0a": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/UniswapV4DeployerCompetition.sol",
            name="UniswapV4DeployerCompetition",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.292
    b"\x4f\x57\xca\x6f\xae\x35\x82\x41\xdf\xfc\x34\xe1\x82\x6a\x6f\x18\x61\xb7\xe6\xde\xdb\x58\xe8\x4e\xcc\xfa\xe6\x2e\xd9\x8b\x1f\xbc": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143
    b"\x4f\x5f\xb2\x39\x26\xb8\x60\x00\x4c\x1a\x56\xf0\x0a\x58\xbe\xa2\x3b\xfd\xac\xc0\x88\xba\x63\xaf\x78\x30\xa0\xbd\x0b\x6c\xc0\x38": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.143",
        ),
    ],
    # Hash appears in versions: v1.0.0, v4.0.0
    b"\x4f\x7e\x0a\x38\xb8\xb5\x29\x5c\x52\xcb\x89\x9d\xa4\x6b\x0e\xa5\x67\x0e\x60\xe5\x0c\x9b\x6a\xcd\xec\x5b\xe1\xe4\x49\x5e\x84\x76": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/FixedPoint128.sol",
            name="FixedPoint128",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/FixedPoint128.sol",
            name="FixedPoint128",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x4f\x90\x22\xa5\x22\xc9\x4b\x9c\xa8\xd7\xda\x58\xf7\x3e\xa0\x0d\xef\x36\xde\xa8\xe1\x09\x64\x9f\x9d\x47\x95\x32\x2f\xa6\x58\xd0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.275, v0.0.276
    b"\x4f\x99\xf6\x52\x00\x65\x56\xb1\xf8\xf4\x76\xa1\x17\x5e\xfa\xfd\x4c\xfd\xec\xad\x20\xb1\x34\x88\x93\xb6\xaa\xb3\x6f\xb1\x2a\xe2": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.276",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x4f\xa0\xbc\x47\x88\x12\x69\xf8\x98\x13\xd6\x3d\x94\x1c\x2d\xa5\x1a\xd3\x97\x5d\x43\x71\x6e\x8a\x98\xf4\xd1\x34\x4e\x11\x1e\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x4f\xb5\xb7\xc3\xb0\x34\x6a\xa5\x07\x4c\xb6\xc8\x12\xa6\xb6\x68\x45\xef\x1b\x51\x89\x5d\x9d\xd6\xc7\x3f\x80\xe5\x52\x36\xb5\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.262, v0.0.263
    b"\x4f\xbd\x42\xc5\x72\xe2\x29\x33\x06\xb2\x41\x06\xfe\x24\xe9\x98\x1f\x80\x1b\xb1\x81\x02\x51\x3c\x42\xf8\x6c\x2c\xab\xed\x82\xd8": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.263",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x4f\xc1\x3c\xe8\xdf\x24\x90\x03\xfd\xe6\x62\x8e\x19\x2f\x6e\xdd\xac\x2f\xbd\x19\xb8\x43\x67\x8d\x73\xb3\x4b\x6e\x7e\x70\x71\xbb": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TransferHelper.sol",
            name="TransferHelper",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x4f\xe2\xb7\x0e\x5b\x46\xed\xf3\xda\x96\xf8\x4e\xae\x17\x2b\x96\x53\xe3\xe3\x85\x49\xd5\x6b\x4f\xf0\x42\x17\xee\x41\xd2\x5f\xf5": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x4f\xe6\xef\x9a\xe3\x9c\xf5\x21\x5f\x88\xe1\x7b\x94\xc5\x60\x61\x0e\xdb\x32\xb2\xac\x1f\xb6\x2f\xc6\x9b\x8c\x87\x14\x7a\xe7\x6c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x50\x1e\x36\x09\x6c\xdb\xb9\x25\xd7\x37\xaf\x60\x87\x67\x6b\xa4\x8f\x07\x0e\xd6\xed\x9a\x81\x78\x99\x9c\xc1\xfd\x4c\xf8\xf6\xb2": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IV4Quoter.sol",
            name="IV4Quoter",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x50\x1e\xe6\xe8\xc4\x38\xbb\xc3\x20\x19\x93\x52\xcc\xea\x1d\xcd\x4f\xfd\x3d\xbb\x29\xc9\x1b\xa2\x00\x5e\x07\xec\xb8\x6a\x3e\xc6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.115
    b"\x50\x24\x80\x99\xcf\xdf\xa4\xae\x2b\x5c\xf8\x5e\x7a\x92\x92\xd3\x26\x39\x3a\x8a\x33\x52\x14\xfb\xbb\x5a\x20\xae\x8e\x84\xc5\x01": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.115",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x50\x27\x7c\x78\x24\x9c\xda\xc7\xdf\xc2\xae\xad\x3d\xcb\x0c\x2d\xfa\x18\x30\x78\x2d\xee\x39\x83\x63\x1c\xa8\xf4\x9c\xf8\xa1\x91": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165UpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x50\x35\xa4\x7f\x5e\x46\x0a\xc7\x50\x08\x72\xeb\x6e\xca\x15\xee\x2b\x78\x2a\xec\xd7\x60\xa5\x0a\x36\xd3\xc9\xa2\xa9\x40\xd8\xbf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.4.2
    b"\x50\x36\x6e\x3e\xbf\x5f\x0e\x74\x22\xef\x65\xd0\xad\x6e\x87\xe5\xd2\x19\x19\xc1\x27\xd5\x1d\x9f\x09\xf3\x4b\xb8\x3f\xbd\x0f\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x50\x43\xa3\x17\x61\x0b\xcb\x4c\xad\x2e\x08\x44\x97\xdb\xca\xfc\x58\x05\x11\xbe\xd9\xff\x35\x16\x77\xb9\x95\x83\x55\x4c\x81\x02": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\x50\x77\x4e\x90\xe4\x3b\x6f\x5c\x96\x6c\x7a\x66\xe7\xdb\x91\x58\x9c\xaf\x9f\xf0\x7d\x21\x84\x1b\xcc\x5f\xaf\x16\xbb\x1d\x9e\x1a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.190, v0.0.191, v0.0.192
    b"\x50\x8c\xec\x01\x8b\x87\xdb\xfa\xf8\x3b\x39\x92\xd5\xbd\x91\xc7\x8c\xd9\x64\x69\xb5\xe1\xd5\x56\x8a\x14\x77\xf5\xac\x66\x2e\x1f": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.192",
        ),
    ],
    # Hash appears in versions: v4.9.0-rc.0
    b"\x50\x90\x90\xe2\x80\x83\x19\x61\xa2\xa3\x7c\xb7\x32\xa5\xf6\x35\xdb\xd4\xc2\x2b\x1a\x25\x17\xb9\x13\x44\xc6\x58\x04\x20\x58\xde": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x50\x9f\xb5\xa2\x1b\x52\x3c\x31\x85\xdd\xb7\x21\x5c\x38\x24\x72\x37\xfc\xb0\xc9\x0f\x61\x6f\x70\x71\xce\xef\x09\x15\xc4\x3c\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x50\xae\x27\x04\x9c\x78\xde\xa7\x4b\x4a\xcd\x34\x52\x04\x76\xe4\x5f\x0c\xf2\x35\x63\xce\xa1\x23\x5c\x58\x2d\x50\x3c\xf1\x1c\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115
    b"\x50\xb3\x55\xbe\xb2\xde\xc1\x10\x7b\xa3\x76\xa8\xb2\xb9\xa7\x5a\xb7\x45\xa5\xe2\x62\xfd\x70\x1f\x4c\x0f\x8f\x12\x6d\x53\x8a\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.115",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x50\xba\xfa\x12\x90\xd8\xdf\xc7\x0b\xca\x6b\xfc\xc2\xd9\x14\x0c\x69\x69\x51\xca\x06\x37\x0e\x54\x1d\x7c\xc2\x6d\xcf\xca\x6a\xbc": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/UniswapV3PoolDeployer.sol",
            name="UniswapV3PoolDeployer",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x50\xe3\x47\x83\x4c\x6f\x1e\x10\xd5\x17\xff\x58\x46\xc5\xa2\x46\xfb\x33\x13\x4d\xd5\xf1\x4f\x05\x3c\x5e\x09\x4e\xb0\x45\x8b\x05": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.16
    b"\x50\xeb\xf7\x16\xa0\xb4\xa1\xfa\x38\xc4\x94\xc6\x29\xf8\x0b\xfb\x0e\x15\xa9\xdb\x0e\x5e\x79\x03\xcf\x3d\x4e\xf0\x82\x5d\x4d\xc1": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x51\x10\xe9\x9e\xb5\xf4\x7c\x10\x97\xa2\x1f\xda\xa5\xfe\x2b\xcc\xf9\xb5\x2b\x9a\xc7\x55\xfd\xa0\x9b\x53\x20\xaf\xdb\x03\x80\x4e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x51\x1c\x30\xb7\x20\x52\xdb\x85\x8a\x9a\x04\x03\x86\x7c\x72\xeb\xd7\xa9\xdc\xc0\x18\x3e\x88\x96\x7b\x4e\x1e\xcc\x0b\x09\x79\xca": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StorageSlotUpgradeable.sol",
            name="StorageSlotUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x51\x34\x11\x8e\xa6\x06\x73\x2a\x26\x67\x9c\xf8\x3f\x44\x8e\x67\xa3\xf6\x5b\x14\x32\xfd\x7c\xf5\x41\x96\x7c\x59\xdb\xbc\x6c\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x51\x3d\x2d\xfa\xeb\xf5\x33\x1f\xe1\x57\x96\xfa\x4e\xe4\x34\xd5\xc3\xde\xad\x7f\x89\xf5\xa5\xaf\xf1\x8b\xea\xa4\xe8\xb8\x8e\x9c": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/types/PoolId.sol",
            name="PoolIdLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x51\x40\x71\x25\x01\x35\x3f\xe6\xac\xff\x5b\xb9\x1c\xc4\xba\xe6\xce\x29\xd7\xe7\x0f\xa1\x31\x05\x15\x85\x49\xfa\x92\x7a\x24\xa8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x51\x43\x2f\x75\x14\x0c\xd6\xce\xad\x7d\x38\xfb\x7b\x2b\xdc\xd4\xe0\x75\x50\xd2\x5f\x03\x55\xee\xe8\x04\x6d\xa4\x73\xad\x30\x45": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x51\x45\xc9\xb2\xad\xa1\x76\x2b\x0f\x5c\xe1\xdd\xbc\x39\x30\x97\x2e\x35\xa5\xb4\x24\xfa\xd5\x0b\xd1\xc0\x45\xdf\x93\x05\x77\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x51\x46\x1a\x9e\x14\xaf\xb7\x92\xd7\x62\x8f\x62\x5e\x11\x83\x49\x02\x7c\xaa\x9b\x9d\x8e\xaf\x6f\xcb\x14\x09\x0e\x76\x7d\xba\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x51\x68\xb3\xff\xfd\x63\xaa\x9c\xaf\xa2\x2f\x5b\xd6\x6c\xe3\xcf\x18\xfb\xc1\xdd\xcd\x2a\x42\x9d\xcc\xbf\xc7\x2b\x4c\x4f\x39\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\x51\x73\x85\x62\xb3\xc0\x01\x43\xed\x92\xbd\x8b\x8d\x44\x64\x78\x8a\xcf\x44\xf4\x0c\x0d\xee\xd3\x98\x4c\x2f\x98\x74\x81\xf4\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x51\x7c\x13\x0a\x91\x0e\x35\x80\x95\xdd\xb6\xc6\xb4\x4c\x79\x32\x2f\xf7\xe5\x9b\x70\x81\x02\x5e\x47\x1c\x9d\xc0\xdb\xf8\x95\x79": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x51\x7f\x08\x9d\x87\x94\x1a\x3a\xfd\x77\xfb\xd7\x65\x7b\xfa\xe4\x6d\xcf\xb9\xc0\x56\x8b\x07\x96\x58\x01\x1a\xe5\x54\xc1\xca\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x51\x81\x48\xd2\xc4\x20\x1b\xe1\x54\x8c\x62\x33\xb3\x9d\xc6\xb8\xcc\xfe\x74\x00\x77\x19\x98\x87\x2f\x9c\x10\x5d\x7f\x19\x15\x06": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x51\x99\x41\xbc\x8e\x81\x9e\x31\xd2\xde\x66\x16\x95\x2e\xbc\xad\xc8\xe0\x7e\x24\x40\x3b\x90\x53\x44\x30\x8f\xb2\x4c\x88\xc9\x43": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: main
    b"\x51\xaa\x76\x15\x69\xea\x3d\xcc\xee\x09\xab\x7c\x0b\x23\x97\x8a\x10\xbc\xd8\x56\xd8\x38\x44\x60\x8d\xef\x4f\xd1\x03\x6e\xf8\xea": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/Multicall_v4.sol",
            name="Multicall_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.249, v0.0.250
    b"\x51\xc8\xac\xea\x8f\xa1\x0f\x53\x46\x66\xfb\x81\x79\xbc\x00\x1d\xf6\x76\xe3\xb6\xa3\xc6\xed\x29\xba\x83\xc7\x71\x97\x18\x29\x43": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.250",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x51\xea\x12\x19\xe7\xca\x96\x8f\x62\xfc\xfb\x0d\xa5\xe5\x05\x43\x3b\xad\x80\x15\x53\x14\xf5\x45\x67\x1f\x23\x5e\xa3\xb3\x3a\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimismUpgradeable.sol",
            name="CrossChainEnabledOptimismUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x51\xf7\x3b\xd3\x71\x00\x22\x0b\x28\xca\x73\x42\x66\x86\xca\xd5\xf3\xad\x0e\xc4\xf1\x41\xc1\x69\x9b\x31\xbd\x09\x47\x37\xef\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232
    b"\x51\xfa\x55\x91\xbd\xa4\x1f\xa8\x3b\x08\x0e\xe7\xd9\x5a\xe2\xc6\x10\x4a\xc4\x01\xbf\x87\xe1\x94\xe5\x1f\x8a\x61\x93\xa8\x4b\x01": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.232",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1
    b"\x51\xfa\x58\x64\x73\x5e\xdf\x4f\x15\xd8\xb6\xa4\xcd\x0c\x8f\x70\xa9\xec\x4e\xc3\xa6\x8a\x84\xf2\x82\x3f\x77\xe8\x66\x28\xd8\x99": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.1",
        ),
    ],
    # Hash appears in versions: v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134
    b"\x52\x0e\x46\x0f\xaf\x22\xe0\xec\x8b\x57\x57\x7b\xd2\xf3\x1e\xc7\x32\x43\xd7\xf5\x3a\x7c\xb9\xf4\x1c\x5f\x7e\xea\x4e\x5e\xd3\xb8": [
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.134",
        ),
    ],
    # Hash appears in versions: main
    b"\x52\x20\x39\x93\x0b\xc7\x17\x03\x25\x6b\xad\x8d\x71\x90\xbc\xef\x8b\x03\xc2\x86\xf5\x35\xcd\xc1\xbc\x8e\x1e\x21\x73\xf8\x26\x18": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/PathKey.sol",
            name="PathKeyLibrary",
            version="main",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1
    b"\x52\x20\xb6\xb9\xd9\xb6\xe8\x24\xf4\xdc\xd9\x14\x24\xb7\x39\x2d\xa8\x04\x39\xd2\x05\xe8\x64\xaf\x01\xae\xdc\x60\x6d\x98\x4c\x4f": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/Path.sol",
            name="Path",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/Path.sol",
            name="Path",
            version="v1.0.1",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x52\x24\x8d\xe4\x19\xad\x00\x74\x61\xbb\x76\x96\xb7\x1c\xc2\xce\x07\xa8\x5e\x22\x0d\x75\x76\x2c\x88\x4d\xf8\x1d\x1b\x96\x42\x2f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7
    b"\x52\x4e\xdc\xe5\x3c\x49\xdb\x4a\x9e\x17\xc5\x6e\xbe\x48\x96\x81\x18\xd0\xdf\x04\x21\xe9\x81\xaa\x76\x55\x1d\xdb\xe0\x74\x4f\x33": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x52\x7d\xe4\xa4\xa5\xaf\xf6\xf4\x6e\x10\xba\xce\x46\x04\xda\xb5\x3f\x07\x0e\xc4\x13\xd2\x6b\x48\xb4\x09\xb3\xfb\x9f\xda\x90\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x52\xe6\xc8\xd1\x98\xa8\xe2\x02\xfc\x77\xf1\xcd\xda\xa1\x8f\x43\xfd\x42\x6d\x86\xdf\x85\xb0\x55\x4c\x31\x2b\xfb\x40\xe6\xdc\x25": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85
    b"\x52\xf1\x77\xfe\x50\x3f\x0d\xb8\x46\x13\xd4\x0f\x0e\x82\x6d\xb3\xe5\x35\xc6\x3f\xe0\x2d\xe9\xfa\x16\x7e\xc5\x13\xfb\xc1\x3c\xdd": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.85",
        ),
    ],
    # Hash appears in versions: v0.0.144, v0.0.145, v0.0.146, v0.0.147
    b"\x53\x2c\xcf\x98\xcf\xc1\x66\xf6\x5f\xf1\x38\x1a\x02\xf3\x2d\x13\xb7\x7f\x29\x12\xdc\x5d\x8d\x11\xad\x5a\x45\xd7\x0a\x88\x5e\xcb": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.147",
        ),
    ],
    # Hash appears in versions: v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117
    b"\x53\x37\x3c\x6c\x44\x95\x54\xe6\xbe\xfa\x9f\xbd\x04\x64\x4d\xf8\x41\x6f\x75\xe1\x3f\x2e\xed\xde\x3f\xfa\x59\x5b\x4f\x55\xb3\xec": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.117",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x53\x49\x9c\xa2\xde\xcf\xeb\x7a\x78\x5b\x3f\xc3\xd0\x9e\x39\x0b\xce\x95\xf4\x51\x45\x7c\x51\x88\x25\xff\xad\xf5\x38\xb5\x56\x60": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPaymentsWithFee.sol",
            name="IPeripheryPayments WithFee",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPaymentsWithFee.sol",
            name="IPeripheryPayments WithFee",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPaymentsWithFee.sol",
            name="IPeripheryPayments WithFee",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPaymentsWithFee.sol",
            name="IPeripheryPayments WithFee",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPaymentsWithFee.sol",
            name="IPeripheryPayments WithFee",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x53\x60\xaf\xc7\x8d\xd6\x95\xc2\xd6\x66\x96\x7b\x55\x13\x65\x9f\xed\xb8\xba\xed\xce\x1b\x3d\x7b\xa4\x3e\x91\xb3\xcd\x14\x9a\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x53\x69\xa7\xae\xbf\x03\x62\x52\xdd\x3c\xf1\x0f\xb5\x03\x38\xe2\xe2\x40\x08\xe7\x20\xc6\xf9\xf9\x97\x0a\x14\x8a\x5f\x40\xfb\x94": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\x53\x89\x82\x31\xe1\xbd\xcb\x4e\x27\x58\x3a\xf5\xd3\xa7\x18\x58\xa7\x8e\x7c\x23\x45\x87\x5a\x80\xb9\x2e\x93\x2a\x11\x63\xc5\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x53\xad\xda\xf4\x90\x63\x5a\x73\x59\x9a\x63\x3a\xb2\xd5\x06\x98\x6e\x37\x6a\x09\xf3\x8e\x38\x1c\x19\x90\xed\xd0\x25\x5d\x11\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x53\xb1\x48\x27\x1e\x8c\xf3\x7c\x7a\x50\xe8\xa8\xba\x9e\x08\xbd\xfe\x82\x49\x93\x80\x43\xfd\x62\xdf\x2d\xf1\xea\xd0\x23\xc0\xfe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x53\xc3\x62\x7e\xb5\x15\x04\xe0\x8d\x4c\xcc\x09\x05\xee\xf0\x3a\x0a\x42\xbb\xb6\x2e\x47\x69\x9e\xa7\x8d\x9f\xae\x84\xaa\x88\x1a": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/StateLibrary.sol",
            name="StateLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\x53\xdb\xa9\xdb\x99\xa8\x68\xa8\xd4\x22\x43\x56\x88\x33\x2d\xd1\x65\x69\x03\x32\x28\xec\xee\xb9\xc7\x05\xae\x5f\x37\xdd\xc0\xc5": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293
    b"\x53\xe6\x37\x19\x02\x7e\x0c\xdd\xed\x66\x81\x9e\x5f\xf6\x15\x60\xc9\xd8\xa0\x64\xea\xcf\x68\x03\xc5\x1d\xa6\xcd\x6c\x47\x5b\x57": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.293",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x53\xe6\xa2\xe4\x16\xf7\x95\xee\x4a\x65\x8e\xdb\xec\x4d\x09\x3c\x72\x78\x80\xb1\x60\xea\x17\xca\x1d\x44\x8a\x82\xd7\xbb\x2b\x87": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x54\x01\xf8\xe9\x02\x82\x32\x06\xf1\x2f\x3f\x83\x60\x81\x6e\x69\x53\xab\x0e\xf4\xd2\x00\x1b\xce\xb5\xcb\x28\x82\x55\xe9\x3f\xcf": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x54\x05\x0c\xcf\xd6\x20\xf0\x2d\xae\x3e\x5e\xa4\x71\x5e\xa4\x43\x79\xa7\x0a\xf4\xfd\x23\x72\xfb\x90\x97\x62\x42\x40\xa4\x5c\x8b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97
    b"\x54\x16\x17\xb1\x06\x67\x78\x3b\x30\xc4\x31\x10\x40\x83\x92\xcd\x64\xe9\xae\x77\x7f\x5e\xb9\x08\x34\xc9\x20\xb6\x3c\x4c\x51\xfa": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.97",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x54\x1c\xaa\xcc\x36\xf5\x23\x60\xe9\xa6\x19\xd2\x7f\xf0\xc8\x04\xc4\x26\x19\xc8\x75\x0a\x8e\xc0\x53\x28\xd3\xa9\x02\xa7\x94\x1e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286
    b"\x54\x44\x91\x66\xff\x67\x47\xdb\x29\xd6\xd3\x7f\x35\x3e\x91\x2b\x3d\x07\x80\xd3\xaf\x80\xed\xbc\xb4\x4c\xdc\x2d\x40\xbf\x17\xf4": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.286",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x54\x48\x15\xeb\x93\x18\x4a\xb9\xce\xb0\xb9\xd2\xbf\x4e\x1f\x05\x8a\x85\x39\xc6\xbf\x9f\x55\x9a\x1a\x5f\x62\x3c\xb3\xf2\x3d\x5d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x54\x4e\x41\xa6\x59\x4c\x47\xd4\x50\x5d\xef\xc9\xb0\x6c\xe7\xf9\x01\x04\x07\x95\x17\xef\x4e\x9a\xa1\x53\x8f\x21\xcc\x87\xc9\xd4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.206
    b"\x54\x55\x9e\x84\x3c\xa1\x12\x12\x5d\xf4\x3f\x8f\xb3\x90\x21\xb5\xf2\x29\x51\x99\xe0\xde\x01\x41\xa3\x8a\xe9\x6a\xb6\x40\xd6\x14": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.206",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x54\x5b\x36\x25\x72\xe0\x99\x5e\x90\x9e\x4d\x23\xc7\x7f\x2c\x35\x8d\x39\x72\x84\x80\xed\x0f\xac\xa2\x4a\x28\xbf\xdb\xf8\xff\xbf": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/Oracle.sol",
            name="Oracle",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228
    b"\x54\x62\x2e\x7e\x9c\x50\x2c\xfd\xc3\xe0\xa0\xde\x0b\x0b\x5a\xc8\xdd\xcf\x61\x09\xe2\x66\x07\x16\x98\xad\x2b\xd3\x75\x09\x47\x37": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.228",
        ),
    ],
    # Hash appears in versions: v0.0.171
    b"\x54\x63\xbf\xdb\x6b\x16\xf2\xd2\x90\x6b\x64\xe2\xac\x96\x84\xfc\xeb\xed\x79\x09\x6c\xd4\x9a\x44\x1f\x3d\x21\x43\xe3\x6e\x22\xae": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.171",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x54\x73\x29\x36\x54\xe7\x82\xfc\xc2\x8f\xee\x18\x97\x67\x5d\x1d\xf0\xa3\x41\x76\x81\x35\x06\x5b\x38\x89\x75\x3a\x8f\xf1\xc4\xb8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x54\xa9\xe7\xc8\x0f\x17\x84\x2d\x49\xa3\x1b\x56\x4c\xaa\x0a\x25\x37\x30\x04\x47\x7b\x90\x9d\x1e\x01\xc8\xa3\x75\x14\xb9\x17\x77": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259
    b"\x54\xe1\x6c\xd1\x31\x9c\xd7\x6c\x94\x64\xf2\x9c\xbf\xe8\x4d\x7f\xb6\x16\xf6\xb8\x92\xa4\x9b\x5c\x90\xc0\x5a\xbe\x8d\x58\xe2\x8e": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.259",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x54\xe5\xa9\x0c\x12\xb5\x6b\xe3\x20\x18\xd0\xce\x2a\xab\xad\xa1\xac\x01\x4c\xb1\x5b\x24\x5b\x0a\x8c\xc6\x61\x55\xa2\xc4\x8c\x8d": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x54\xed\x5f\x9e\xbe\x1a\xd8\xeb\xc4\x21\x7c\x4b\x59\xbd\x26\xd2\xc6\x00\xf2\x5d\x2b\x6e\xc5\x00\xdf\x5c\x33\x7a\x2e\x5a\xe3\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254
    b"\x55\x31\x07\xa0\x2a\xa2\xec\xc7\x5c\x24\xdc\x46\x7f\x71\x47\x3c\x3b\x30\x69\x79\xaf\x98\xcf\x80\xe8\x16\x08\xdd\xfc\xba\x21\x26": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.254",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x55\x38\x9e\xce\xae\x66\xaa\xb4\x08\xae\x8d\x38\xab\x2e\x94\x73\x7c\xce\x21\x5b\xa7\xeb\x29\xd7\x25\x24\xd5\x0e\x4a\x54\x1a\xf9": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x55\x71\xf8\x6f\x40\xc7\xf9\xf6\xd5\x37\xb6\x5a\xab\x4b\x82\xb2\xbb\xae\xf2\xbf\xa6\xef\xfe\x57\x85\x3c\xed\x8f\x13\x00\x07\xa1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.16
    b"\x55\x87\xdc\xcd\x35\xcb\xbb\xb3\x38\xd1\x4c\x25\xa6\xfc\x08\xac\xd2\xb9\xdb\xa0\x0a\x1b\x2a\x49\x3d\x6e\x20\xe6\xe9\xc9\x07\xcd": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.16",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x55\xb9\xfb\x51\xdf\x7c\x2f\xb9\xd1\x75\x1b\x75\xf6\xa8\x43\xd2\xa3\x9f\x8c\x0b\x03\x44\xec\x25\xb7\x8c\x6d\x5e\x39\x9c\x07\x74": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Proxy.sol",
            name="ERC1967Proxy",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\x55\xd6\x3a\xa0\x53\x42\xe8\x68\x49\xc4\x52\x07\x57\x56\x55\xa5\x20\xaf\x0b\x08\x0d\x69\xf0\x8f\xc6\x97\x75\x4e\xee\xec\x34\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.185
    b"\x55\xd9\x48\xb7\x37\xf7\x03\x99\x2a\xeb\x00\xe6\x39\x88\xd5\x6b\xae\x0d\x30\x49\xf2\xcb\x67\x32\x63\xfc\x01\x75\x2d\xeb\x0f\xff": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.185",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x55\xe6\x10\x18\xa1\xb1\xd3\xe7\x19\x2d\x4e\xad\x25\x72\x35\xe8\xe2\x35\x0b\xed\xb8\xa0\x37\x6e\xa6\x80\x81\xf6\x07\xd5\x19\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x56\x11\xdf\x4d\xc1\x9d\x7f\xfa\xdf\x73\xa3\x33\x53\xe6\x2b\xfb\x3e\x96\x63\xfe\x52\xf6\x13\x17\x89\xb8\xae\x97\x06\x2c\x5a\x17": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/NoDelegateCall.sol",
            name="NoDelegateCall",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x56\x43\x8b\xfa\xb6\x6c\x70\x3e\x3c\x9c\xa9\xa7\xc5\xe6\x2f\xb7\x9d\x61\xc4\x84\x0f\x16\xb2\x69\xf2\x0d\xab\x37\xc2\x4c\x16\x3d": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.260
    b"\x56\x62\x38\x0d\x4f\xd8\xd5\x7d\xf0\x8e\x59\x07\x23\xfc\x25\xce\x35\x06\xb2\xbe\x93\x6e\x44\x73\x08\xf5\x0d\x5b\xb6\xa5\x66\x07": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.260",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19
    b"\x56\x70\xab\x7c\x8b\x3e\x05\xa2\x87\xed\xcd\xed\xc2\x23\x7c\x1d\x84\xef\xf1\xd2\xf0\xf9\x6b\xde\xec\x9c\x43\x1a\x2a\x7e\xd7\x2a": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x56\x87\x68\xaf\xdb\x29\x3a\xeb\xa1\x52\x63\xb0\x03\x46\xc1\x13\x6d\x75\x07\x2b\x48\xd5\xf6\x62\x2c\xfc\x28\x7e\x1c\x8f\x7e\xe5": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/TickMath.sol",
            name="TickMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1, v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x56\xb8\xd2\x83\xb5\x98\x58\x21\x78\xbd\xe3\xc3\x05\xc9\x73\x79\xd2\x8d\x26\xe0\xe3\x22\xd8\x0c\x26\x2c\xd6\x8f\xd8\x2d\xd6\x17": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271.sol",
            name="IERC1271",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC1271.sol",
            name="IERC1271",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC1271.sol",
            name="IERC1271",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC1271.sol",
            name="IERC1271",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC1271.sol",
            name="IERC1271",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IERC1271.sol",
            name="IERC1271",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121
    b"\x56\xd0\x3a\x04\xce\xab\xaf\x13\x38\x78\x39\xb7\x9a\x60\x08\x42\x40\xea\xbd\xa1\x26\x64\xe4\x11\x2e\xf9\x22\x4e\x1e\xb9\xe0\xb9": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x56\xde\xa4\xc0\xfb\xff\x0a\xcc\x45\x75\x50\x5f\xfc\x55\xc7\x4d\x32\xf6\x5b\x1a\x39\x88\x8d\x1a\x32\xe0\x35\x2a\x64\x80\xae\xea": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x56\xe8\x5d\x14\xdc\x24\x31\x3e\x1d\xd2\xd8\xc0\x9c\x19\xb3\x2b\x3c\xc0\x36\xa3\xd2\x08\xdd\xe2\x51\xe7\xfd\x1f\x25\x28\x38\x6a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x56\xe8\xe0\x9b\xe2\x24\x5a\x9c\xf4\xda\x92\x4f\xbb\x6d\xce\x1e\x1a\x2f\x27\x8a\x86\x69\xc5\x3e\xc4\x16\x06\x2a\x50\xce\x52\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x57\x1b\xbd\x57\x8d\x6e\x26\x9e\x9d\x49\x51\x4f\x5b\xca\x29\x2e\xaf\xc9\xca\xdb\x32\x53\x8c\x8f\xa4\xb2\x50\xff\x78\x5f\xed\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.1
    b"\x57\x29\x03\x20\x92\x4d\x16\x83\xf9\xe3\xb5\x51\xa3\xdd\xc2\x0f\xe2\x46\x56\xf4\x77\xab\x11\x67\x55\x7f\x33\x05\x55\x48\x63\xaf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x57\x73\xe3\xc0\xf5\x99\xb2\xc9\x31\x70\x25\x67\x54\x56\xb1\x24\x19\x51\x58\xdb\xcb\xfc\xef\xfe\xe4\x45\xa7\x23\x3a\xaa\x2e\x1f": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/BytesLib.sol",
            name="BytesLib",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/BytesLib.sol",
            name="BytesLib",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/BytesLib.sol",
            name="BytesLib",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/BytesLib.sol",
            name="BytesLib",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/BytesLib.sol",
            name="BytesLib",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x57\x81\x45\xf5\xfb\x8e\x10\xdc\x3a\x9c\x37\x1a\xca\x7d\x74\xbe\x79\xe8\xfb\xcb\x87\x62\x83\xf4\x67\xd3\xa2\x2d\x0f\x22\xbc\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x57\x99\x94\xd7\x80\x4d\x73\x8e\xcb\xa5\xa9\xbe\x50\x01\x0d\x5d\xa3\x20\x5a\xb1\x99\x80\x04\x70\x12\x74\xad\x90\x85\x42\xb7\x86": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x57\x9e\x77\x3c\x26\x7d\xfc\x03\xd6\x1d\x1a\x3c\x1b\x37\x0a\x6e\x90\x22\x2f\x27\x4c\xaa\x83\xac\x27\x9e\xc1\xa9\x7e\xc8\xa3\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0-rc.0
    b"\x57\xad\xbf\xba\xad\x65\xee\x11\x96\x3a\x51\x7a\x40\x85\x1e\x4d\x44\x9e\x27\xef\xbf\x74\xe6\x97\x2f\xa5\x0b\x0f\x97\x46\xeb\x2c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x57\xbe\x1b\xc0\x01\x2f\x1e\x4c\x1b\x73\xc6\x13\x5d\xb0\x34\x02\xb8\x34\x1f\x17\x56\xa6\xdb\x37\x49\xc3\xb6\x74\xb0\x1a\x47\x83": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLenderUpgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x57\xeb\x7a\xbe\x93\xe0\x05\xd4\x0d\x7e\xc8\x11\xb0\xf6\xfd\x60\x9e\x7e\x77\xcc\x88\x98\x61\x8c\xb6\xd1\x57\x5e\x2a\xc4\x3d\x83": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x57\xf8\xe9\xa2\xba\x2e\xe0\x39\x1c\x69\x82\x40\x51\x0c\x9c\xd7\xa3\x20\xa4\x31\x2a\xb0\x94\x26\x03\x99\xe8\x22\x6a\xd7\xf3\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x58\x3d\x2e\xe1\x65\x27\x8b\x0f\xcc\xe3\xc4\xae\x8d\xda\x74\x76\x7f\x3f\x2f\xd3\x55\x59\x2b\xef\x98\xd9\x0c\x27\xe2\x74\x17\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.47
    b"\x58\x48\x6e\xfb\x70\x2e\x6a\xc6\x9b\xb7\x7a\x93\xb7\x39\x06\x5e\xea\x89\x54\xe4\x0b\x1b\x84\xb5\xf8\xe6\x90\x1e\xd1\xe1\xe6\x40": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.47",
        ),
    ],
    # Hash appears in versions: v0.0.176
    b"\x58\x6d\x87\xd5\x25\xf5\x1f\x4c\x53\xdb\xc0\x36\xbc\xa9\xcb\xf3\x49\x32\x85\xbb\x49\x85\x30\x8a\xa2\x33\xfa\x76\x93\xff\x16\xb4": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.176",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x58\x78\xdb\x3a\xaa\x7e\xa9\x4a\x0a\xa2\xa8\xc8\xf9\x95\xb0\x98\x13\x47\x84\xd5\x38\x96\x4a\xd9\x04\x38\xa2\x44\xde\x94\x0a\x51": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolImmutables.sol",
            name="IUniswapV3Pool Immutables",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.222
    b"\x58\x8b\x09\x0c\x91\xf1\x26\x18\x1f\x2a\xff\x05\xfe\x57\x33\xc3\x3b\x10\x1e\x24\xf9\x16\x66\xa5\x4c\xd0\xb6\x66\x8b\x27\x64\xf6": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.222",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x58\xb3\x49\xb2\xfa\x09\x7e\xbc\xfa\xaf\xf1\x9c\xcf\xa2\x27\xd8\x77\xb9\xdb\x1c\xa4\x3d\x86\xf6\xe5\x7e\xc8\x4f\xc1\xa7\x49\x62": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.8, v0.0.9
    b"\x58\xb4\x26\x36\x04\xbd\x5a\xeb\x8b\x1a\x2a\xbf\xb1\x65\xeb\x75\x32\x18\x90\x7e\xbb\xae\x98\x52\x94\xce\xd9\x32\x13\xa4\x5c\x75": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x58\xbf\xf1\xf1\x9d\x22\x11\x72\xa8\x83\x87\xe9\x53\x57\x57\x3a\xa4\x9f\x38\x75\xb5\x7e\xbf\xde\xc9\x96\x3f\xd0\x60\x81\x1b\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x58\xc6\xd3\xf3\xcc\x9e\x98\xf6\xa8\xa8\xb4\xe6\xa0\x95\x47\x1d\xaf\xd7\x8a\xae\x42\x1a\x54\x4f\x5a\x7b\xa6\xc8\xbc\x4b\x48\x32": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1967Upgradeable.sol",
            name="IERC1967Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x58\xe3\x4b\x1c\xc7\xc9\xe5\x59\x8e\x4c\xe1\xd4\x91\x44\x6e\xa3\x05\xb8\x73\xe5\xe5\xad\xf1\x99\x4d\x0f\x37\x87\xfe\x04\x39\x6e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x59\x04\xb2\x82\xa7\xcc\xa1\x09\x4e\xda\x07\xe7\x00\x6c\x63\x9c\x7d\xf7\x07\x67\xf6\x53\x63\x20\x1a\x70\xcb\x11\x16\x2a\x1f\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC1822Upgradeable.sol",
            name="IERC1822ProxiableUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x59\x23\x58\xb7\x89\xc0\x81\xc7\x07\x7f\x5a\xb3\x27\x87\x82\xd1\xe5\xe9\x14\xce\x51\x9b\xa3\x65\xc3\x89\xac\x6c\x19\x1c\xc3\x6f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x59\x25\xe2\x85\x7c\xd7\x8d\x35\xbe\x08\x89\xe7\xaa\x59\x09\xb4\x81\x00\x9f\xc8\x7a\xaf\x6b\x18\xac\xa5\x1c\x31\xd0\x20\x6f\xb1": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/ERC6909.sol",
            name="ERC6909",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x59\x39\xb9\x5a\x30\x4d\xda\x95\xf1\x24\x12\x18\xef\x8d\xe3\x4f\x0c\x77\x95\x0b\x0a\x4e\x80\x6b\x2b\x0f\x3c\x68\xe4\x7f\xae\x6d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66
    b"\x59\x70\xfe\xb1\xb2\xff\x43\x4a\x72\x92\xcd\xff\x49\x51\xf9\x2a\x27\x73\xbf\x6b\xbe\x8c\x7f\x3e\x3f\xa0\x77\x6c\x59\x04\x56\x9a": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.66",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x59\x79\xff\x63\x10\xd3\xcb\x0f\x82\x34\x34\xfa\x6f\x46\xe4\xc4\x78\xc2\x07\xd0\x79\xa2\xe4\x64\x11\x8b\x5e\x7f\x91\x83\x4f\x44": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820ImplementerUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x59\x7a\x4f\xaf\x5e\xd6\xe3\xfc\x09\x39\x82\x40\x20\x5f\xe4\xfa\xa9\x10\x6e\x5b\x3a\xe8\xbc\x85\xaf\xe9\xad\x4f\xd6\x28\x36\x39": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x59\x9b\x01\x6e\x0a\x0b\xf0\x71\x92\x66\x01\x14\x20\x9e\x7e\xc1\x6c\x10\x1c\xec\x30\x46\x52\x4c\xfa\x75\x30\x44\x43\x36\x8b\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.278, v0.0.279, v0.0.280
    b"\x59\xab\xd6\x3c\x8c\x91\xaa\x6a\xb7\x5f\x4c\x21\x27\x3b\xac\x9b\xf8\x58\x39\x1c\xde\xaa\x20\xcc\x9a\xad\xa7\x4b\x28\x59\x78\x50": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.280",
        ),
    ],
    # Hash appears in versions: v0.0.5, v0.0.6
    b"\x59\xb9\x23\x42\xb5\xb7\x6f\xc7\x8d\xfa\x58\xf9\x84\xa9\xd6\x63\x0b\x93\xd1\xa0\xe0\xce\xde\x3a\xd6\x78\x90\x6b\x55\xc7\x1c\xb7": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProof.sol",
            name="MerkleProof",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProof.sol",
            name="MerkleProof",
            version="v0.0.6",
        ),
    ],
    # Hash appears in versions: v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x59\xde\xc2\x4b\x52\xd1\xfc\xcb\x5c\xf5\x39\xba\xba\x67\xc4\x6a\xd0\x20\xae\x70\x94\x6f\x6c\x98\x3b\xf3\xa9\xe2\x21\x44\x42\x7d": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173
    b"\x59\xf9\x32\xf4\xe6\x37\x6f\x81\xe3\x92\xd3\xc8\xd7\x42\xa4\x63\x23\x59\xe1\x1d\x30\xb0\xda\xb2\x5f\x76\x3c\x2c\x54\xf2\x28\x91": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.173",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x5a\x40\x5f\x2f\x3b\xf0\x1d\x58\x23\x88\xd1\x13\xf3\x3a\x3d\x3e\x87\x9c\x04\x47\xae\x3c\x7b\xb7\xed\x87\x85\xd8\xdc\xd0\x37\xc2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x5a\x43\x65\x97\xd2\x66\xe1\xa4\x1f\x5f\xe6\xa4\x92\x25\xed\xe9\x31\x7e\xcc\xb0\xea\x4c\xb8\x99\xdb\x05\xc9\x10\x04\x04\xb9\x2a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x5a\x55\x32\x8e\x74\x95\xe7\x16\xd4\x08\xea\x41\xad\x04\x90\xb0\x43\x04\xf7\xa1\x43\x85\x55\x4f\xf2\x10\x34\xff\xb4\x8a\x0e\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x5a\x55\xc8\xda\x8c\x8b\x5d\x12\x09\x71\x2e\x0c\x55\xfb\x07\x8c\xf1\x1d\x07\x3b\x5c\xbb\x56\x1f\x2b\x77\x06\xcd\x4b\x1c\xb7\x4b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x5a\x5a\x6f\xf9\xa7\xcc\xc3\x41\xef\x49\x98\x4d\xed\x01\x67\xbd\x8c\x26\x36\xb5\xfc\x8a\x7c\xc6\x38\x61\x6a\xeb\x9c\x17\x4b\xc1": [
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x5a\x75\x69\xd1\x35\xc1\xfc\x5e\xe1\xa1\x8c\x93\x4d\x31\x11\x19\xcd\x53\x3a\x35\xd2\x7f\x64\x47\x88\x4f\x79\x21\xbf\x63\xf8\xc5": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x5a\x90\xe2\xbc\x8c\x20\xb0\xbc\x97\xc6\x03\x43\x82\x3c\x55\x09\x37\x0f\xd7\x94\xc5\x58\x13\x2a\xce\xfe\x36\x18\xe2\xef\x28\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.3.0-rc.0, v3.3.0-rc.0
    b"\x5a\xc8\x79\xcb\x71\xc3\x5a\x03\x61\x86\x86\x5d\x5b\x7a\x8f\x8d\x40\xbc\x2c\xdd\x78\x7a\x8a\x00\x2e\xe5\x71\x67\xc4\x4a\x79\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x5a\xef\x82\xde\xff\x7f\xcb\x5c\x88\x66\xe0\xef\xc1\xf8\x7e\x7d\x00\x9a\x96\xc7\x95\xd9\xb5\x28\x41\x43\xef\x14\x0f\x2f\xf7\x16": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibMap.sol",
            name="LibMap",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x5b\x22\xf8\xb7\x79\xca\x9f\xb6\x41\x10\xb0\x91\xeb\x4b\x18\xad\xd5\x8f\xa6\xf3\x8e\x09\xec\x29\x5b\x1f\x45\x1c\xd1\xe1\x5b\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.50
    b"\x5b\x25\xc5\x9c\xaa\x94\x27\x85\xf9\x37\x6d\x03\x92\xba\xe4\x53\x0c\x2d\x13\x9e\xed\xf4\xca\x8b\xa4\x22\xf6\xde\x99\x71\xed\xe0": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.50",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x5b\x6b\x27\x03\x2a\x10\xf0\xab\xd3\x3b\x55\xe8\x59\xf2\x4c\x88\xe0\xb4\x9b\xd8\xd6\x05\x4c\xda\xd6\x63\xf9\x34\x6a\xc5\x5c\xb3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelockUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x5b\x6d\x4d\x6d\xe4\xed\x70\xa4\x93\xa7\x5e\xa1\x74\xd1\x3d\x22\x3c\xa9\xea\xcf\x81\xc2\xc3\xb1\xea\x5f\x42\xc8\x76\xd7\x07\x6e": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74
    b"\x5b\xae\x93\xe7\x7c\x32\x38\x37\x8f\x9e\x04\x05\xb2\x2c\xd8\xb0\x2f\xcf\x72\xf0\x94\x95\x5e\xac\xf0\x54\x53\x2d\xf3\xcc\xf7\xcb": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.74",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x5b\xbe\x50\x57\x4a\xe6\x71\x75\x2f\x0d\xed\x28\x13\xc6\xf3\x8b\x24\x4d\x93\xcb\x64\xaf\xe5\xb7\xa0\x29\xff\x18\xe8\x0f\x73\x4f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5b\xd6\xe2\x06\xa6\x23\x68\xb4\x8b\x79\x25\x6b\xf0\xac\x7b\x8c\x43\x6e\xf4\xa1\xe8\xff\x98\x09\xd5\xe4\x6c\xc5\x2a\x37\xd7\xa0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x5b\xf7\x38\xc3\xb2\x95\x76\xe0\xed\x6a\xe9\x4e\x4b\x23\x91\xea\xfc\x77\x3f\x95\x5d\x53\x95\x27\x3d\xfa\x2b\xee\xe6\x8b\xa4\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x5c\x06\xb7\x4c\xd8\x31\xb7\x55\x90\xd2\x48\xa1\x24\xc9\xcf\x1d\x5a\x8a\x42\xe9\xda\x2c\x9c\xa5\x9f\xbf\xaf\x2b\x6b\x11\xf9\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7
    b"\x5c\x1a\xe7\x43\xe8\x7a\xb3\xd1\x2c\x36\x08\xc5\x46\x89\x2b\x68\x2e\x6f\x88\xdb\xbf\xd8\x2f\x06\xf8\x26\x37\xa8\x49\x8f\xdd\xff": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x5c\x2f\x8a\x4f\xa6\xef\x82\xd8\x4f\x6c\x3d\xe0\xfd\xdb\xac\x1c\xf1\xd0\x79\xc3\x53\x58\xb8\x92\xa8\xdd\x3b\xfb\x0c\x52\xe5\xd6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IArbSysUpgradeable.sol",
            name="IArbSysUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5c\x31\x52\xab\x6b\x08\x17\xe5\x63\xf7\x3e\xe0\x3c\xe1\x22\x84\xd3\x1f\xe6\x0b\x83\xa2\x01\x12\x12\xd2\x2a\x14\x38\x2c\x30\xd7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x5c\x3d\x10\x7a\xcc\xa9\x77\x33\x90\x5a\x12\x53\x14\xef\x15\x08\x0a\x5b\x54\x24\xf7\xbc\x9f\x28\x58\x44\xfc\x0a\x63\xbc\x47\x74": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x5c\x55\x68\x78\xd0\xc1\x83\x58\xc7\xb9\xd5\xa3\x22\x8d\x0e\x72\xa4\x01\xcc\x2d\x23\xf5\x80\xc0\x02\x87\x21\x00\x89\x59\xa6\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x5c\x57\x94\x64\x24\xd8\x31\xc4\xf3\x9f\xa6\x05\x76\x09\x29\xa4\xdc\xc5\xde\x6d\x31\x02\x5e\x78\x81\xb6\x7d\xb3\x37\xf2\xfe\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x5c\x58\xb4\xb3\x85\x3f\xf2\xf9\xdc\x46\x90\x4d\x8e\x7f\x58\x31\x51\x2c\x0b\xc3\x24\xdb\x48\x99\x7b\x74\xbd\xfc\x66\x8b\x23\x47": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x5c\x68\x5e\x27\xdb\x31\xd1\x21\x7d\xb6\xd5\x48\xeb\x0a\x92\x39\xd9\x67\xfc\xfa\xf3\x7a\x9f\x35\xb3\xf0\xaf\x96\xf0\x5d\x57\xe1": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x5c\x6e\xe5\xfb\x52\x38\x4d\x29\xb9\xee\x6e\xed\xde\xae\x68\xc6\x91\x4c\xab\xef\x14\x94\xb4\x82\x60\x7f\x63\xb2\x1c\xd5\xff\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.275, v0.0.276
    b"\x5c\x7b\xf9\x3d\xf2\x06\xa7\xa6\x44\x77\xbc\x6d\x05\x54\x2c\xa8\xa2\xee\xb7\x4b\xf4\x41\xd9\x8a\xc6\x3d\x28\x30\x88\x39\xbb\xe6": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.276",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x5c\x7e\x75\x80\x87\x3b\x87\xb1\x4e\x5c\xeb\xe2\xd4\x33\x0d\x03\x80\x2f\x9a\xa1\xca\x7e\xd4\x2d\xf5\xe1\x1e\x03\x12\xae\x5c\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x5c\xb8\xad\x27\xb2\xfa\xae\xdc\x1e\xf2\x5a\x92\xfa\xac\x58\xe2\x58\x6c\x12\xe9\x3e\xfb\xcc\xa4\x28\x98\xae\x6c\x21\xea\xfd\x83": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x5c\xd2\x56\xe6\xc9\xb6\x04\xf0\x0c\x36\xe4\x93\xed\xbf\xeb\x77\xbb\x0a\xb0\xc9\x21\x7c\xc5\x9e\x22\x0a\xbd\x6a\x1f\xc8\xb2\xd9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x5c\xd7\xd5\x6c\xae\x20\xe0\x5e\x20\xa5\xf0\x24\x14\x14\x23\xc0\x1d\x5b\x7d\x45\x04\x4b\xc8\x1b\xef\x38\x3c\xfd\x32\x47\xc4\xa4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x5d\x23\x68\xd4\xba\xa9\x97\x9d\x62\x63\xd5\x89\x11\xc0\x6a\xab\xd2\xed\x9d\xc1\x1f\xb3\x6c\x2d\x56\x31\xfa\x4b\xa2\x04\xbd\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35
    b"\x5d\x29\x9f\xb9\x5e\x62\x2c\xce\x28\xd7\x4c\x74\x1e\xa4\x5b\x5e\xce\xc7\x3d\xb9\xcb\x98\x41\x54\xeb\x42\xa8\x58\x0e\x7f\x32\x94": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.35",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.1, v4.4.1, v4.4.0-rc.1
    b"\x5d\x2e\x98\x43\x96\x95\x38\xe9\x8a\x43\x26\xaa\xed\x7b\x8f\x8f\xee\xf3\x5d\x2b\x3f\xe0\xa7\xee\x14\xd8\x14\x3a\x64\x37\x30\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x5d\x37\xf1\xd7\x2d\x21\x10\x71\xc0\x28\x41\x80\x15\x6d\xea\x09\xd3\xc8\x43\x55\xaa\xa3\xd6\x90\xe9\x1a\x8d\xb5\xf6\x64\x48\x44": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x5d\x3b\xa8\x6a\xc3\x97\x6b\x76\xe1\xed\x81\x57\x4a\x89\x0b\xd4\x6b\x01\x06\x0e\x25\xcf\xfb\x0e\x3a\xd7\x38\xdd\xdb\x66\x24\xdd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.6
    b"\x5d\x45\xd4\xb3\x93\x69\x5d\xc6\xa4\xbf\xb3\x5a\x27\x3c\x86\x13\xf9\x46\xbe\xca\x8c\x55\x53\x9c\x8f\xc9\xb9\x8f\xad\x99\xfc\x31": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.6",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x5d\x45\xdf\xa7\xe3\xea\xa0\xe1\x50\xda\x74\xd9\x49\xaa\xe8\xeb\x63\xff\xed\xad\x4f\xa2\x4d\x5e\x02\xb8\xdb\x74\x71\xfd\xe6\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x5d\x46\xa4\x40\xbf\xcb\xac\x4f\xbe\x20\x5f\xb5\xe5\xef\x2b\x3e\x61\xc3\x7d\xcd\xaa\x38\xb9\x26\x4e\x26\x64\xe5\x31\xa6\x65\x70": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x5d\x53\x62\x59\xee\x15\x90\x41\x7b\x78\x5e\x77\x62\x57\x84\x9f\x1b\xa6\x74\x10\x6b\x80\xb6\x8a\xc5\xea\x39\xc3\xfb\x36\xc1\x31": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/BaseActionsRouter.sol",
            name="BaseActionsRouter",
            version="main",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x5d\x5a\x94\x37\xf5\xcc\xa5\xb9\x97\x72\xed\x79\xf1\x60\xaf\x66\x1b\x95\x81\xfc\x62\xac\x23\x2b\x3f\x2d\x5e\xb3\x0f\xf9\x51\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x5d\x79\x81\xa1\xa3\x27\x43\x14\x79\x7f\xa1\xed\x7f\x52\xcc\x5d\xf8\xbc\x2d\x69\x01\x2a\xb1\x57\x75\x39\x3e\xa1\x3b\x90\x52\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.51
    b"\x5d\x7d\x38\x33\xab\x2f\x4d\xd7\xf6\xbf\x6b\xc6\x49\xf3\x1c\x21\xf8\xc3\xe6\x15\xe5\xd5\x7b\xfb\x47\xf9\x38\x9a\x96\xa1\xe5\xea": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.51",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x5d\x85\x83\x07\xf4\x1c\xa3\x80\xc2\x14\xe3\x3f\x2e\x17\x55\x34\x30\xe2\x1d\xc8\xd0\x1d\xc4\x13\x25\xa8\x4a\xac\xee\x09\x68\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x5d\x9e\xa4\xc1\x0a\xc3\xf6\xa0\x80\xac\x07\x72\xa2\xab\x86\x8e\x68\xbf\xbc\x71\xb7\x11\xad\x01\xba\xf2\x36\x31\x06\xee\x4e\x39": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x5d\xad\xfb\x12\x4e\x0f\xca\x07\x5a\x9a\x25\x63\x9c\x16\x88\x86\x0c\x9b\xd4\xc0\x91\x4b\x6e\xa5\x8f\x83\xb4\x83\x7c\xa1\x4d\x36": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x5d\xb4\x02\x2b\x8e\xaa\x84\x48\x4b\x6d\x3e\x62\xd9\x5a\x7f\x16\x0b\x55\x65\xc5\x62\x37\xdf\x8a\xfd\x44\xb0\xcf\x86\x5d\x44\x0a": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISelfPermit.sol",
            name="ISelfPermit",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISelfPermit.sol",
            name="ISelfPermit",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISelfPermit.sol",
            name="ISelfPermit",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISelfPermit.sol",
            name="ISelfPermit",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ISelfPermit.sol",
            name="ISelfPermit",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x5d\xb8\xe1\x77\x13\xe6\x1c\xc5\x33\x0e\x5a\x4a\x3b\xa3\x80\x0c\x6d\x67\x86\xab\x42\x67\x2d\xf6\x21\x81\x8d\x3a\x69\x8d\x22\xa8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.194
    b"\x5d\xc3\xb0\x20\xaa\x44\x25\x1b\xcd\xd9\xc8\x3e\xf1\x53\x89\x01\xee\xb8\x52\xa0\xc4\x13\x72\x9f\x49\x17\x0f\x09\x60\x9e\x72\xa3": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.194",
        ),
    ],
    # Hash appears in versions: v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x5d\xc7\x77\xd0\x87\x3e\x3c\x14\xc6\x1c\x96\x67\x58\x97\xaf\x61\x25\xf6\x48\xe4\x3d\x24\x7e\x1e\x81\x06\xe0\x80\x34\x97\xc9\x71": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x5d\xcf\x6b\xff\x4c\xc9\x94\xaa\xca\xaf\x58\xbb\xc3\x77\xb1\xd4\x3f\x0a\xc8\x57\x5b\xd5\x27\x2e\xc8\x21\x01\x75\xa4\x80\x59\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x5d\xd7\xc0\xe8\xdb\xb8\x76\xe9\x68\xe1\x36\x40\x8a\x54\x66\x30\xa6\xef\x15\x67\xa8\x74\x66\x72\x53\x03\xec\x32\x60\xf6\xa1\xfe": [
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.99",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x5d\xf0\x03\x3a\x04\x6a\x03\x2d\xf1\xc7\xd3\x1d\x87\x72\x0a\x85\x2c\x97\x6a\x98\x17\xff\x01\x25\x22\x34\x76\x19\x0e\x8e\xbb\x1e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x5e\x01\xb9\x92\xba\xd6\xf5\x00\x96\x3f\xf3\x4e\xda\xa7\xf3\x00\x6d\xc9\x35\xd9\xf2\xc0\xac\x21\xb5\xfc\x05\x14\x49\xff\xb8\x26": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5e\x10\xcc\x36\x5d\x77\x24\xeb\x4e\x74\x0c\x79\x0d\x64\xb2\x28\xef\x29\x97\x3d\x37\xfa\x4c\x31\x77\x9c\x0b\x6a\x32\xc4\x67\x4f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP2.sol",
            name="CAIP2",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP2.sol",
            name="CAIP2",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP2.sol",
            name="CAIP2",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP2.sol",
            name="CAIP2",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP2.sol",
            name="CAIP2",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x5e\x45\x62\x68\xea\x72\x87\x23\x4b\xcb\xe4\xc7\xe5\x06\x4f\x54\x99\x52\x18\xc8\x23\xed\xf5\x82\x42\x29\x45\x04\xed\xc4\x0b\x9c": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5e\x5a\x88\x39\xe5\xc4\xec\x95\x89\x65\xe2\xcb\x07\x6b\x2b\x88\x9b\x71\x1d\xb7\x11\x6b\x15\xe4\xef\x4a\x1a\x24\x91\x11\x28\xe4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x5e\x7e\x5b\x77\x9e\x93\xbb\xa7\x36\xee\x11\xb2\x88\xa2\x35\x28\xfa\x10\x6b\xb9\x79\xb3\xff\x2f\xa6\xc9\x08\x78\xb6\x66\xba\x98": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0
    b"\x5e\x8e\xc2\x1b\x90\x3b\x51\x2e\xaf\x57\x2f\x52\x7d\xcc\x2f\xa9\x7f\x69\x40\xb3\x49\xb9\xa1\xf1\x24\x11\xc4\x50\x90\x76\xed\x38": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.190
    b"\x5e\x90\xea\x24\xaa\x83\x55\xce\x9d\x51\x8f\xac\xe4\xa4\x01\xbc\xba\x84\x1a\x46\xbb\xcb\x98\xa2\x9a\x99\x98\xa6\xd1\x1f\xa6\xbd": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.190",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x5e\x96\xf2\xf3\xc3\x58\x3c\x33\x40\x5b\x23\xa9\xb4\xed\xc0\x76\x84\x0e\x43\xb3\xe0\x99\xac\x39\x32\xa7\x70\x54\x38\x68\x24\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/LibAMBUpgradeable.sol",
            name="LibAMBUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.0.0-beta.1
    b"\x5e\xdf\x36\xf2\x2b\xbc\x21\x21\x7a\x75\x84\x51\xd2\xa8\xcf\x03\x9d\x87\x14\x49\xae\x43\xe0\xaf\xc3\xff\x18\xe2\xc2\x6c\xba\x72": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x5e\xe3\xe1\xaf\xf1\x1a\x1b\x24\xe6\xbc\x95\x64\xe0\x3d\x6e\x33\xe7\xc1\xf0\xa2\x61\x7c\x2a\x02\x88\x61\xe7\x93\x72\x87\x3b\x0f": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/types/BeforeSwapDelta.sol",
            name="BeforeSwapDeltaLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.7
    b"\x5e\xf6\x43\x48\x63\xff\xbe\xca\x9c\xf2\x56\x0d\x94\xe9\x5f\x67\x26\x35\x2e\x5c\x5b\x7e\x98\xeb\x49\xc1\x12\x3c\x4f\x5a\x04\xd8": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.7",
        ),
    ],
    # Hash appears in versions: main
    b"\x5f\x03\xd0\xec\x05\x53\x89\xc3\xf3\x2b\x81\x6d\xc2\x01\xda\x7a\x47\xd6\xa0\xc0\x96\xa1\xc1\xe1\xef\x8f\x14\xb7\x86\xaa\x65\x1b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/hooks/BaseTokenWrapperHook.sol",
            name="BaseTokenWrapperHook",
            version="main",
        ),
    ],
    # Hash appears in versions: main
    b"\x5f\x33\x18\xae\xd3\x69\x52\x75\x95\xd6\x52\x4c\x0f\x33\x8b\x5d\xff\xf3\xc7\x41\xab\xef\xf0\x75\x0c\x35\x0e\xa8\x9e\xe7\x7e\x71": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/SafeCallback.sol",
            name="SafeCallback",
            version="main",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5f\x55\xd6\x52\x72\x9b\x76\x4c\x38\x17\xd2\x0f\xce\xa8\x98\xdd\xf5\x37\x4d\xe6\x01\x39\xd8\x81\x2e\xd6\xe0\xfb\x1b\x2d\x88\x8e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockAccess.sol",
            name="GovernorTimelockAccess",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x5f\x8b\xd4\xab\x44\xdb\xe1\x60\x3a\x1a\xcf\x46\xf4\x2e\xa6\xef\x47\xd1\xaa\xaa\xc5\xcb\xfd\xf9\x92\x73\x51\x9b\xaa\x34\xfc\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerable.sol",
            name="IAccessControlEnumerable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285
    b"\x5f\x8f\xe8\xac\x61\xfa\xeb\x3a\x76\x6d\x1d\xf1\x90\x07\x22\x1f\x20\xdd\xc6\x78\x70\xe6\x9f\x93\x0b\xa9\x98\x17\x84\x5c\x23\xd5": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.285",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x5f\xa5\x6c\x94\xf2\xce\x9e\xc4\xf7\xa5\x0c\x1c\xd7\xfd\x22\x25\x89\x01\xaf\xc7\xee\xb0\x19\x4e\x7c\x1a\xd7\x76\x8d\x2f\xcd\x8d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.68
    b"\x5f\xb0\xaa\x78\x18\xfe\x72\xa3\xad\x02\x1e\xa3\x31\x1a\xfb\x67\x4c\xfe\xf3\x30\x4e\xab\x40\x17\x01\x72\xde\x8e\x6a\x8f\xf5\xdf": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x5f\xcc\x94\xf6\xd5\xcd\x79\x97\xeb\x74\xb2\x8e\x8b\x00\x56\x5f\xda\x27\xa4\x66\x67\x8b\xc1\x73\x04\x69\xce\x31\x8c\xaf\x9a\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.0.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7
    b"\x5f\xee\xff\xe7\xca\x8f\xf1\xec\x6e\x19\x07\x66\x51\x74\x48\xb8\x7c\x34\x6d\x1e\x1c\x28\xcb\xc8\x41\xa2\xe8\x29\xee\xdc\xa7\x32": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x5f\xf0\x1f\x17\x9c\xec\x95\x19\xd7\x3d\x3a\xd0\xa9\x5e\x4c\x5f\x0b\x87\x50\x72\x21\xbe\x19\xce\xb7\x61\x0b\x21\x67\x6d\x68\xfe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPoint",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPoint",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPoint",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPoint",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPoint",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x60\x02\xc6\x72\x09\x95\x16\x0b\x7e\xe2\x10\x68\xd7\x1b\xef\xfa\x7d\xc2\x84\x67\x11\xc9\x8f\x8e\xca\xd1\x3d\x28\x41\x1e\x28\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x60\x2e\x58\x4f\x38\x0d\x67\x83\x92\x27\x67\xfe\x03\xd0\x0c\x0b\x1f\xa4\x89\xac\x02\x6e\xe2\x39\x14\x68\x18\x8b\x1d\x3e\x14\x3e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeaconUpgradeable.sol",
            name="IBeaconUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x60\x42\xd2\x2f\x44\x8a\x88\xa8\x5e\x03\x05\x95\x72\x3c\xc6\xf3\xa6\xcf\x51\x41\x1a\xe1\x5e\x94\x17\x82\x83\xf0\x92\x02\x9f\x17": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x60\xc6\x97\x3e\x99\xd7\xfd\x90\xf0\x2b\x4c\xfa\xf5\x67\x2e\x27\x1b\xa1\x16\xfa\xed\xf8\xec\x18\xf1\x77\xb8\x2a\xdf\x7c\x64\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160
    b"\x60\xcb\x43\x01\xe8\x5c\xe3\x16\x42\xc2\xd6\xa2\xd3\x8c\xd1\xdd\xee\xf6\xe7\x5d\x1e\x1a\xae\x8d\xc9\x79\x40\x4e\xe5\xc1\xc7\xa2": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.160",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x60\xeb\x9b\x55\x70\xaa\x7e\xd8\xfd\x58\xf3\x1d\x2f\xa7\x28\x53\xa0\x2e\x34\xf9\xe2\xdc\x9b\x4a\xdb\x20\xb1\x29\x7c\xfa\xe5\x4d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.3.0
    b"\x60\xf7\xfe\x96\x46\x75\x55\x0a\x4d\x24\x3c\x47\x74\x0d\x4e\x71\x47\x9d\xa4\x31\x19\x11\x09\x2b\x96\x93\xa8\x9b\x4d\x7c\x32\xcd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/P256.sol",
            name="P256",
            version="v5.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x61\x09\xab\x3e\x5e\x3b\x85\xeb\x1a\x92\x5c\xd6\x35\x94\x6c\xd5\xfb\x9e\x24\xce\xab\x6d\xa1\x4a\x1c\x0f\x77\x31\xb8\xae\x84\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC6093Upgradeable.sol",
            name="IERC20ErrorsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0
    b"\x61\x0b\xb6\xb0\xdd\x4a\xcc\x80\x6b\x3f\x8f\xc3\xc2\xfa\xeb\xbc\x84\xf5\x91\x12\xdd\xc2\x1d\x42\x29\x20\xab\x97\x1a\x9b\x6b\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x61\x29\x63\x59\x13\xcb\xca\x74\xdb\x03\xe2\x1b\x7b\xd3\xc8\xb7\xd0\x57\x01\x80\x6a\xb5\x4f\xbf\x13\x3d\xc5\xca\x14\xd0\x94\x51": [
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/Math.sol",
            name="Math",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x61\x32\xd4\x17\xec\xe5\x4a\xca\xc8\x3a\xa4\xd7\x45\xac\x98\x95\xb5\x92\x80\x5f\x1f\x92\x6a\x30\x87\x77\x77\x66\x37\x09\xeb\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.5, v4.9.6, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x61\x6c\xf5\x29\x20\xcf\x29\xfa\x06\xf5\x5f\x46\xc0\xab\x27\xf1\xe1\x34\x23\xef\x9c\x9d\xa5\xde\xad\x78\x7d\x34\x0c\xb5\x0e\x38": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x61\x7c\x57\x75\xc4\x6a\x21\x90\xe0\xe5\xee\x69\xb8\x2d\xd6\x40\x5d\xa6\x7d\x1c\xb1\x47\xe6\xd0\x0b\x00\x25\x0f\x81\xb4\x0d\xe9": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/CurrencyDelta.sol",
            name="CurrencyDelta",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x61\xbb\x94\x45\x88\x5b\x37\x5a\xa4\xea\x75\x3a\xaf\x55\x6e\xca\x81\x39\xea\xd6\x81\xa6\x33\xf6\x78\x94\x5d\x4e\x28\xe2\x1b\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273
    b"\x61\xbb\xdd\x7c\x6d\xf1\x07\xf1\xf5\xac\x67\xb3\x5b\xfb\x72\xa7\x33\x79\xdf\x2e\x7b\x37\x44\x9e\x4f\xcd\xbc\x89\x9e\xe7\x41\x24": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x61\xcb\x64\x3c\x80\xdd\xa1\xee\x3d\xcf\xea\xb3\x34\xee\x9d\x4a\xa8\xd4\x0c\x17\x0a\x74\x23\x31\x68\x72\x8f\x00\x37\x74\x7a\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x61\xd0\x4a\xbd\x56\x32\xc2\x61\x61\xe8\xc5\x9a\x16\x7d\x95\xea\xd0\x3c\xec\x41\xb2\x4d\xd5\x61\x28\x7f\x88\x28\xe0\x23\xde\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x61\xd3\x5a\x2a\xf2\x72\x2b\x3b\x5c\xc9\xd9\x45\x4b\x91\x69\xae\x6e\xf2\xe3\xd2\x3c\xe5\x9c\x73\xe8\xab\x2b\x0d\xf1\x33\x58\xf5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x61\xd4\x3c\xbc\xea\xd7\x74\x3b\x34\xcc\x9e\x71\x95\x42\xb5\x51\x0d\x32\x1f\x5a\xee\x8e\xc7\x01\x69\x4a\x87\xfd\x9c\x89\x8c\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20Permit.sol",
            name="ERC20Permit",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x61\xdc\xee\xd1\x24\xa9\x8a\x0e\xe0\x93\x7c\x0c\x20\x72\x11\xe0\xa0\x33\xbf\xf9\xcc\xeb\x1a\xbf\x70\xc4\x30\x16\xaa\x9e\x31\xfd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x61\xf8\x56\x91\xa2\x9c\x71\x00\xe3\x90\x40\x15\x6c\x5f\xb0\x94\x4f\xa1\x0b\x93\x42\xd0\x9a\x78\x53\x33\x3b\x5d\xbf\xe7\xcd\xd4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x62\x2e\x3b\xa8\xfe\xd2\xcc\x5a\x80\x9b\x7d\xa6\x74\xf1\xd9\xb4\x9e\x92\x12\xf8\x84\x32\xe8\x5e\x99\x54\x14\x1e\x47\x33\x84\x65": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132
    b"\x62\x5b\xea\x63\x8b\x91\x50\x61\x3f\x69\x98\x26\xb5\xb7\x1a\xe0\x24\xfe\x67\xa1\xbb\x9f\xc7\x64\x6b\xe0\x39\x4f\x08\x4b\x71\x68": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.132",
        ),
    ],
    # Hash appears in versions: v0.0.284
    b"\x62\x66\x8a\x38\xe8\x55\x5b\xef\xae\x59\xe7\xc2\x3a\x2d\x27\xf2\xe1\xc4\xd2\x41\x79\x54\x79\x7f\xac\xc2\xd3\x27\xb7\xe0\x92\x99": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.284",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x62\x69\x82\x94\x65\x41\xcb\x80\xd0\x1e\x65\xbc\xf4\x9b\x5e\xd9\xe7\x13\x9e\x5b\x12\xa9\x3e\xc6\x24\xbf\x78\x9f\x5b\xbd\x37\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.277
    b"\x62\x8d\x57\xd0\x57\xa0\x29\x5e\x0c\x25\x17\x2b\x74\x63\xbe\x50\xb1\x5e\x00\x17\x45\x95\xca\x84\x76\x15\xee\xd7\x31\x19\xa6\xdd": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.277",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x62\xa0\xa0\x86\x93\x61\xc2\xb9\x76\xb5\xe9\xd2\x68\xc4\xae\x2a\xa4\x1f\xbf\xa9\x6e\x8b\x73\x1e\xda\x2a\xa8\x0d\x7c\x0e\x4c\x44": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Recipient.sol",
            name="IERC777Recipient",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x62\xc3\x79\x76\x45\xf4\x5e\xab\xf6\x25\x7a\xc4\x2c\xbb\xe4\x51\x98\xde\xfd\x11\x9d\x6d\xfb\x8e\xc2\x4f\xa6\x09\x75\xd9\x7d\xb0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x62\xd2\x24\x0c\xcc\x07\x38\x74\x41\x59\xad\x67\x57\x18\xc2\xc5\xf4\xb1\xf1\x72\xbb\x2d\x5c\x60\xf7\x9d\x5b\xd6\x29\xcc\x74\x70": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/external/IERC6909Claims.sol",
            name="IERC6909Claims",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x62\xf8\x58\xff\xf8\xdc\x9a\x8f\x71\xce\xf7\xc5\xc6\xf6\x43\x46\xc1\xa7\xdd\x17\xa3\x67\x70\xdf\xe8\x71\x86\x68\x95\x8b\xc2\x05": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x63\x08\xd8\x74\xb0\xd6\x97\x8e\xdd\x71\x5a\xc4\xec\x2e\x87\xe0\x47\xa4\x41\x5c\xfb\x36\x2d\x7b\xd0\x56\x9a\x7d\x8c\x31\x26\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49
    b"\x63\x11\xcd\x7c\xd3\xdd\x6d\xb2\x9d\xdb\x00\x74\x1a\x50\x43\x5a\x77\x7d\x44\xe8\x74\xa7\xa0\x58\x67\x69\xd3\x1f\x0c\x74\x81\x50": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.49",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\x63\x19\x08\x21\xca\xfe\x91\xfd\x14\x11\x4d\x64\xb2\xee\x92\x80\x18\x17\x1b\xd2\x89\xb0\xc2\xa7\x69\x5a\xac\xc9\x08\x22\x63\xe3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x63\x21\xc8\x1b\x13\x90\xc5\xbc\x37\xc9\xeb\x01\x76\xe0\x78\x17\x97\x5c\x23\xe0\x00\xb3\x73\x85\x67\xd3\x73\x6d\x44\x9b\x66\xd8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x63\x36\x8b\x37\xfe\x3e\x1c\x78\xf8\x25\xb4\x7e\x6d\xad\x0d\x8f\x5e\x39\xe2\x45\xe2\xf9\xbc\x2f\x66\x4c\x87\xcf\x79\x23\x5c\x52": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\x63\x67\xab\xe6\x65\x1d\xb3\xd7\xf0\xdf\xd1\x1b\xe1\x54\x39\x3f\xd4\xc2\x52\x4e\xed\x83\x55\xa9\xbb\x31\x5c\x46\x3f\x50\x02\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.0.1, v3.0.2
    b"\x63\x80\x00\x89\x8a\x3d\x77\x8b\x32\xdd\xb7\x47\xa7\x01\x5c\x66\x1c\x47\x5e\x43\xbb\xcf\xa6\x41\x25\x79\xed\xd5\x55\x52\xf3\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v5.1.0-rc.0
    b"\x63\x84\xca\x3f\xf5\x8a\x8f\xee\xbd\xc5\x87\x2f\x2a\xd7\x8a\xf5\x8c\x58\xf6\x83\x2f\x75\x8f\xf3\xce\xb9\xe8\x18\x3f\x0c\x9d\x6a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x63\x98\xf2\xf3\xd0\xc3\xee\x7d\x44\xfd\x84\x28\xd5\x35\x49\x2f\xe3\x6c\x6c\x95\x1a\x94\xe0\xa6\xf2\xcd\xe5\x13\x06\xd9\x13\xf3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x63\xaf\xcb\x77\x85\xd3\x01\xf1\x68\xf3\x73\xf7\x76\x91\xb6\x06\x36\xd0\x1c\xe3\xa3\x3d\x5a\x0f\x09\x4a\xec\xfa\xe6\x6d\x4e\x98": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/IAccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\x63\xde\x69\xfc\xf8\x25\xc3\x67\xee\x70\xd4\x54\xf3\x41\xea\xc2\xef\x6a\x47\x01\x41\x83\x68\xc4\x2d\x53\x7b\xbb\x7f\x6e\x59\xee": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.1
    b"\x64\x00\x10\x9f\x36\x7d\xce\x7f\xb5\x5d\x73\x29\x3f\x64\x51\x6d\xeb\xdf\x68\x54\x80\xb3\x10\x0e\x70\x14\x7f\xbd\xe3\x47\xf9\x8c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x64\x0b\x56\x32\x02\x07\x96\x96\xf4\x3f\x03\xfc\xe4\x7d\xaf\xc5\x31\x8f\xfb\xd3\x0b\x85\xee\x24\x4f\x05\x8f\x2a\xa4\x78\x8b\x04": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x64\x0e\x36\xc3\x3d\xa8\x0c\x16\xdf\x8e\x41\xb4\x34\xef\x35\xab\x37\xd5\xc5\x00\x84\xca\x59\xdc\x20\x3c\x18\x4d\x0d\xe6\x79\x57": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.205
    b"\x64\x14\xcb\x12\x35\xe9\x3b\xde\xed\x57\x8b\xfe\x1f\xa8\xe9\xf8\x72\x40\x19\x84\x48\xc3\x22\xa0\xa2\x49\x64\xf1\xcc\x0a\x28\x1c": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.205",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x64\x4a\xa0\xb7\x26\xe7\x63\x20\x13\x18\x2b\x5d\xd7\xe0\xc4\xbd\x0d\xc1\x0c\xf4\xac\xa8\x70\x07\x6a\x67\x19\x88\x87\x35\xc7\x9e": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/FullMath.sol",
            name="FullMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x64\x55\x10\x67\xee\x88\xcc\xc8\xba\xe9\x5f\xa3\xd3\x90\x47\xfb\x49\x53\xc2\xaa\x28\xdf\xd7\x50\xb3\xa7\xf1\xbb\x28\xba\xa9\xbb": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x64\x76\x23\x74\xec\x60\x9b\x29\xe7\xeb\xc2\xb3\x35\x45\xde\x70\xe4\xd9\xfe\xdc\x31\x39\x20\x7c\xde\x9c\x1b\xec\xf2\x49\x57\x31": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.8.3
    b"\x64\x78\xfc\xb5\x6b\x9a\xd2\x18\x63\xb3\x3f\x4a\x0b\x62\x7b\xda\xb9\x32\x6f\x32\xcf\xc0\x85\xdc\x32\xaf\x9e\x53\x05\x72\x7b\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x64\x85\x8c\x9f\x90\x4e\xc2\xf4\x68\xa5\xbe\x66\x7e\xf2\x93\xf8\xa2\x41\xfe\x8d\x4b\x07\x2d\xd6\xd8\x6f\xc5\x49\x2f\x39\x68\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1
    b"\x64\x9e\xa0\x10\x37\x51\x28\xa5\xdd\x98\x9c\xb2\xc0\xaa\xbd\xb0\xc4\xea\xc0\x8a\x47\xac\x1d\xf2\x9a\xf0\x1d\x2c\x22\x2f\xc7\x75": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.1",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x64\xbb\xfe\x55\xe1\xd0\x11\x5d\xed\xe4\x70\x6e\x2d\x5c\x3e\xf3\x2f\x28\xbf\xa3\x09\x63\xb6\x06\xc9\x59\x56\xfd\x31\xf7\xcb\x66": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/callback/IUniswapV3FlashCallback.sol",
            name="IUniswapV3F lashCallback",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.210
    b"\x64\xce\xaf\xfb\xd8\x0f\x28\xf2\x00\xb3\xe1\x43\xa3\x2c\xd1\x83\x13\x5a\xb0\x13\x11\x6e\x96\xfd\x58\xcd\xaa\x4a\xdd\xcf\xe9\x9d": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.210",
        ),
    ],
    # Hash appears in versions: v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x64\xd0\xd4\x12\xce\x63\x73\xaa\x62\xbd\x60\x85\xcb\xba\xe6\x95\xfa\xb5\x7a\x1c\x6a\x19\xf0\xb0\xcf\xbd\x3e\x06\x84\x3c\x3a\x10": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x64\xdb\x70\x1b\x96\x33\xa2\x90\x87\x8b\x6d\xa1\xb8\xa3\xec\x03\xc3\x9c\x21\x7d\x1a\x62\x17\x5c\xaf\xe3\x71\x3e\x74\x1e\xbb\x30": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x64\xff\x68\x43\xe3\x31\x18\x22\x1d\x06\x10\x10\x58\x88\x42\xb0\x46\xb1\x40\x3e\xb2\x35\xdd\x91\x0b\x63\x51\x91\x9a\x90\x74\xf9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172
    b"\x65\x03\x7f\xe0\x8c\xac\x5a\x74\x6f\xc5\x35\x7e\xe9\xd5\xe3\x07\x3f\x99\x4f\x32\x22\xfb\x52\x94\x12\x9c\xa6\x9e\x43\xc4\x6c\xd5": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.172",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\x65\x29\x77\xc8\x46\x4a\x8d\xc2\xd3\x5a\xce\xbd\x86\x95\x30\x2e\x19\x08\xe9\xf4\xc8\x9b\xf9\x21\x42\x03\xaa\xd3\xba\x71\xc5\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x65\x66\xcf\x6b\x60\x8f\xe9\xbe\x65\x02\xa6\x6b\xcb\xb0\x33\x26\x64\xdb\x13\xc0\x6a\x31\x8c\x8b\xfa\xa2\xa3\xc6\x18\xcc\x87\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x65\x6d\x4f\x0e\xcb\x89\x55\x3a\xc3\x88\xc0\xde\x7a\xf4\xbc\xc7\xba\x1d\xb7\xdf\xed\xb3\x6a\xe1\x5b\x7e\x75\x86\xa7\x99\x8e\xf5": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.8, v0.0.9
    b"\x65\xa2\x0b\x52\xee\x99\x8d\xe7\x75\x28\xd5\xc7\x2e\xb1\x63\x52\x4b\x4b\xc9\x2f\xd6\x72\xf5\x33\x07\x55\xce\x79\x64\x37\xe0\x3c": [
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x65\xad\xb1\xe2\x99\xaf\xc1\x75\x9d\xb8\xc9\xfe\xf5\x2d\x38\xbf\x9b\x8a\x81\x23\xf0\x8b\xac\x88\x93\x3a\x38\xf7\xdc\x35\xe0\x93": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2
    b"\x65\xd8\x43\x92\xf0\x5e\xe6\x7e\x1d\xa2\x98\x7a\xf5\x5e\xad\x9d\x48\xea\xf9\x1c\x0a\x2c\xfc\x57\xd7\x67\xc8\x27\x6a\x2b\xa9\x85": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.8.2",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.6, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.7, v0.0.8, v0.0.9
    b"\x65\xdb\xb2\x7b\xed\x71\xbd\xda\xd0\x81\x95\xcc\xf1\x6a\x3f\xea\xc7\xcb\x16\x20\x87\xae\xda\x80\xec\x77\x99\x02\xfd\xeb\x0f\x08": [
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.55",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.7",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/Milady.sol",
            name="Milady",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x65\xde\x5d\xc9\x34\x20\x33\x57\x0d\xfa\x8f\x85\x43\x2c\x9a\xf2\x76\xed\x2f\x85\x5d\xfb\x92\xef\x8a\x6e\xfe\x14\xbe\x3c\x9b\x8c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x65\xde\xc4\x18\xb8\xb4\xd8\x62\x20\x0a\x1d\xcc\x6e\xb9\xac\xb1\x69\x29\xeb\xdc\x04\x7d\xee\x90\x45\xf1\x1f\xdf\xd9\xa0\x40\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x65\xe9\x63\xff\x35\x97\xf9\xcb\x95\x68\x41\x96\xb3\xde\x78\xa0\x2b\x96\x0b\x44\x9d\xba\x9d\x79\xd4\xad\x87\x50\x41\x69\x55\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x66\x05\x59\x01\x15\x8a\xf1\xd9\x83\x19\x56\x29\x7b\xf4\x7e\xff\xe3\x84\x87\xb2\x52\x46\x39\x5d\xbb\xcd\x66\x27\xdd\x4f\xa6\xba": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x66\x17\x26\xe3\xda\xcf\x1e\x78\xbf\xa9\xc8\xcf\xe8\x85\x4c\x22\x7c\x08\xda\x7d\x07\x58\xd3\x63\x32\x5a\x4e\xae\x9b\xd2\xd9\x58": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5267.sol",
            name="IERC5267",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x66\x3c\xee\x1d\x29\x40\x42\x61\x8b\x46\x4a\x50\x26\xde\x42\x36\x92\x3b\xf5\x07\xc7\xcd\xf5\xa5\x78\x1c\xb4\xe2\xe1\x2c\xb3\x83": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x66\x47\xed\x72\xc9\xba\x09\xcc\x8e\x3d\x95\xe2\xf3\xa4\x98\x39\x34\x3b\x81\xc6\xf0\x90\x75\x24\x22\x73\x8a\xee\xc5\x3f\xc3\x97": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\x66\x55\xf7\x9c\xaf\xda\x66\x58\x59\x03\x9d\x73\xca\x85\x58\x86\xcd\x5a\xb0\xa2\x16\x92\xb6\x68\xa3\x05\x55\x25\x14\x52\xe8\x24": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/ReentrancyLock.sol",
            name="ReentrancyLock",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x66\x87\x4f\xa5\x20\x36\xa4\x35\x88\xe9\xac\xfb\xb9\x2d\x24\x2d\x21\x22\x6a\x06\xba\xe5\xc8\xa6\x4b\xce\x85\xa9\xd2\xb1\xac\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x66\x88\x79\x63\xa4\x9b\x86\xe9\xb0\xb2\x34\x3d\x91\x2e\x99\xbb\xae\x4b\xb8\xc5\xc6\x64\xd0\x01\xc5\x97\x71\xa3\x7a\x50\x7d\x69": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x66\x92\x62\x0d\xab\xa8\xc6\x00\xa2\xbd\x1d\xab\xa2\x82\x87\xb0\x76\xa2\xbf\x63\xf2\x84\x47\x10\xca\x27\x9e\x36\xe2\x2f\xd2\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x66\xa0\x77\xe6\xdd\x59\x78\xc4\xfa\x26\xe1\x82\x42\x4b\xd6\x78\x24\xa5\x80\x97\x5d\x5c\xe1\x96\xa0\x0c\x6f\xb5\xe3\x18\x3f\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC165Upgradeable.sol",
            name="IERC165Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.8.0
    b"\x66\xa0\x7d\xf4\x54\xff\x88\x05\x33\x2f\x58\xf2\x52\xee\xe9\x12\xb7\xa8\x56\x68\xab\xeb\xfb\xfd\x73\xc2\x74\x03\x57\xd0\xa5\xb3": [
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x66\xb3\x9c\x43\x19\x1e\x42\x59\x9a\xc1\x9a\x49\x9a\x2e\xc2\xae\xf4\xbf\xc7\xc3\x81\x9a\x32\x6a\x9f\x6d\x99\xa8\x75\x74\x03\xff": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1271Upgradeable.sol",
            name="IERC1271Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x66\xf5\x01\x6a\x14\x83\xcc\x60\x70\xfd\xa8\xd6\x95\xd6\x86\xa8\x49\xc7\x9e\xab\x1b\xf5\xb7\xf3\x83\x0a\x0f\xc7\xae\x03\x33\x58": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSuperQuorumUpgradeable.sol",
            name="GovernorSuperQuorumUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSuperQuorumUpgradeable.sol",
            name="GovernorSuperQuorumUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.250
    b"\x66\xf5\x81\xf0\x5f\xc2\x0c\x01\xf3\x5d\xf9\x65\x87\x8e\x36\xe6\x76\xf0\x07\x2d\x81\xb4\x99\x62\x61\x73\x87\x0a\x51\x06\x38\xa6": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.250",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x67\x04\xba\xbc\xdc\xa6\xbc\x61\x67\xf3\x7d\x4e\x16\xfd\x13\xfa\x67\xa5\x9f\xad\x27\x67\x19\x02\x86\x05\x04\x3e\xca\xf0\x08\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.233, v0.0.234, v0.0.235
    b"\x67\x0d\x06\xb6\x3c\x99\x60\x28\x16\x29\x52\x77\xf7\x88\x35\xdd\x8b\x2c\xd0\xc7\x8b\x21\x80\x6e\xc8\x4f\xd1\xd1\x42\x3c\x68\xae": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.235",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x67\x0d\xc9\xd9\x4a\x5b\x97\xd3\x2b\xe7\xea\xe5\x92\xf8\xe7\x78\x53\x24\x09\xe9\x27\x2e\x82\x44\xda\x50\x4b\x41\x45\xe4\xe3\xa0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x67\x0e\xae\xb8\x3a\x08\xe1\x7e\xc4\x7a\xc8\xf7\xf9\xce\x15\x72\xa3\x33\x41\x76\xd6\x1d\xf4\xb0\x57\x85\xc3\x12\xc5\x2b\xd3\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x67\x10\x13\xc4\xa5\x91\x03\x3f\x82\x37\xf4\x5d\x1c\xc0\x77\x23\x2b\x7e\xa4\x23\x0d\x31\x30\xfb\xde\x36\x8e\x00\x39\xe4\xbc\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\x67\x16\xb2\x1d\xec\x4e\x6c\xf3\x86\x71\x66\x5c\x63\x99\x7f\xa4\x57\xb0\xe5\xf2\x5f\x28\x92\xea\x56\x63\xe9\x92\xfd\xe5\x49\x3e": [
        KnownContract(
            project="solmate",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.233
    b"\x67\x21\x45\x6e\xc8\xbd\x03\xd8\xe6\xbc\x36\x74\xff\xff\x35\x36\x3e\x1a\x94\x11\x81\xb8\x44\x99\xbf\xee\xc7\x2d\x7c\x3b\xb4\x7f": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.233",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x67\x21\xed\xff\x64\x66\x08\x35\x96\xf5\x31\xb4\xc5\xc2\x31\xc4\xb5\x21\x8c\xbf\xaa\x80\xc0\x46\xaf\xe2\x52\xf5\x2e\x85\xc1\x8b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5267Upgradeable.sol",
            name="IERC5267Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.166
    b"\x67\x2a\xa3\x9c\xd9\xaa\x56\x2c\x33\x5d\x80\xf1\xcf\x1b\x2d\xae\x13\xfd\x75\x84\xbc\xb5\x9d\x1a\x09\x49\xd1\x23\xa5\x8f\x53\x32": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.166",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x67\x34\x96\x01\x5c\xa5\x3e\x7d\x2d\x54\x94\x3a\x4a\x37\x6d\x67\xc2\x12\x05\xf8\x2a\x72\x4b\xc4\xd9\x1a\x42\x02\x01\x06\x89\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x67\x48\x38\xad\x7c\x84\x56\xfa\x1b\x19\x85\x90\x57\x71\x5f\xb0\xec\x57\xc5\x7b\x35\x8b\xae\xb3\x8b\x16\x8b\x48\x28\x97\xdc\xa6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x67\x4e\x6b\x86\x17\x71\x68\x82\x91\xeb\xf3\xc9\x68\x42\x16\x21\xe0\xd7\xc9\x8f\x59\x81\x05\xf0\x55\x30\x40\x34\xd9\x38\xd7\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x67\x53\x2f\x26\x94\xd8\xc1\x3b\xef\x8a\xb1\x1c\x64\xa2\x05\x4a\x6b\x8d\x6e\x40\x9a\x7e\x70\xf7\x31\x14\x05\x9a\x06\x99\xb2\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Panic.sol",
            name="Panic",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x67\x54\x47\xb7\xf3\x4b\xa5\xc8\xe6\xca\x8b\xd3\x90\x63\x95\x24\x96\x04\x40\x89\x0c\xb6\x88\x00\xdb\x49\x88\xb0\xda\x38\xe1\x87": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x67\x8a\xbc\xa4\xaa\x92\x95\x3c\xcb\x20\xc9\x43\x52\xa2\xf5\x08\xac\xe7\x86\x1b\x01\xcd\xb7\xb2\x04\x71\x57\x9f\xc2\x4e\x82\x88": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x67\x99\x6a\x32\xbb\xdf\x76\xf3\xa2\x1a\x20\x9e\x7b\x03\x93\xc7\x68\xc0\xb8\x7e\x11\xae\x23\x79\x75\xfe\x6a\x1f\xe9\xe1\x38\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/ownership/Secondary.sol",
            name="Secondary",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/ownership/Secondary.sol",
            name="Secondary",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x67\xa0\x4b\xf8\x24\x28\x67\xff\x98\x63\x36\xe7\x81\xfa\xde\x79\x91\x1f\x91\x00\xb5\x78\x47\xd0\x41\x1d\x87\xb7\x4c\x5e\x74\x0e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x67\xca\x75\xc8\x7a\xa0\x47\xd3\x39\xf6\xce\x43\x16\x60\x65\x21\x1b\x57\x7c\x0d\x8d\x78\x47\x6a\x68\xfe\x2e\x21\x30\x8c\x74\xf4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.112, v0.0.113
    b"\x67\xea\xb0\x34\x4f\xdf\xfb\x7c\xd5\xdd\xf9\x20\x6a\x96\xad\x02\x0a\x69\x14\x6f\x83\x96\x2f\x2d\x53\xee\xab\x41\x4d\xbd\xf8\xe0": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.113",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x67\xf2\xd0\xbe\x06\xc6\xde\x6b\x99\x68\xa8\xbe\x40\x66\x1c\xf1\x0d\x87\xae\x62\x31\x6c\x86\x81\xf3\xec\xa0\x18\x11\x4f\x41\x2e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14
    b"\x68\x0f\x8d\xd0\xca\x69\x5b\x5a\x3b\x8d\xb0\x92\x28\xd6\x93\x64\xfa\x7c\xf6\x7a\x00\xb2\x02\x1b\x15\x6e\x37\x48\xda\x62\xc8\xda": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.14",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x68\x3e\xe7\x88\x44\xcb\xcc\x3a\xdf\x95\x01\xd6\xf7\xa3\x2d\xd0\xec\x86\x36\x0d\x1e\x33\xe7\x39\xa6\x9b\x43\x00\x04\xb1\x46\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136
    b"\x68\x43\xe1\x64\x13\x5f\x58\x1e\x86\x1e\x05\x34\x91\x2d\x7a\xf3\x99\xa3\x33\x51\x18\x3a\x84\x4e\x6f\xdf\xb0\xb8\xcf\x53\x02\xf7": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.136",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7
    b"\x68\x46\x27\x80\xc6\xdd\x20\x4c\xee\x76\x0c\x8e\x10\xae\x61\x42\xf7\x26\xa2\x3d\x7a\x18\xde\xf5\x14\xed\x2f\xf7\xd5\x88\xc1\x66": [
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x68\x66\xed\x10\xb1\x30\x2a\x38\x2d\x19\x11\x59\x7e\x9a\xab\xfb\x43\x86\x8c\x8b\x50\x99\x11\x71\x27\xcb\x0e\x22\x9e\x8d\x04\x1d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\x68\x7b\x22\x45\x01\x25\xa6\x30\xca\x0a\xc5\x2a\x54\x8a\x81\xc8\x80\x17\x9c\x9f\xa4\xdf\xe7\x03\xdc\x82\x8b\x70\x76\x03\xe2\xc9": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2
    b"\x68\x86\x45\x9e\xe0\x2d\x81\x88\xfb\x47\x92\x8e\x3d\x08\x5d\xcd\xa6\xf7\x4f\xb0\x59\xdb\x0f\xdd\x85\x97\x28\x56\x9d\x2c\xf9\x48": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.2",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x68\x90\x2e\xf9\x63\x4f\x27\x07\x45\x8c\x0c\xca\x67\xd7\x5a\x10\x71\x03\x36\x52\xf0\xe0\x99\x2d\x7d\xa1\x31\x35\xc5\xe9\x23\x79": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x68\xbc\x81\x6d\x64\xfa\xa3\xe1\x82\x09\x06\x08\xe3\xb4\x5f\x05\xfa\xda\x9f\x7c\x6d\xcb\xa1\x39\xef\xed\x18\x04\x80\xed\x66\x6d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x68\xce\xcc\xc3\x8d\x73\xf9\x1c\x69\x9d\x99\x59\x01\x35\xe3\x06\xa6\x81\x29\x59\x1a\xf3\x6a\xfb\x38\xcf\x96\x29\xb9\x1f\xeb\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.2.0-rc.0
    b"\x68\xf2\x6a\x46\x11\x93\xea\x0a\x41\xbc\xd6\x1a\x91\xdd\xbf\x4f\x06\xc3\x31\x5e\x68\x67\xe4\x60\x8e\xdf\xcb\x65\xf0\x7e\x62\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x69\x12\xb8\xb0\xa5\xbb\x72\x8c\x3b\x4c\xfc\xd8\x15\x46\x12\xc2\xcc\x34\x09\x0e\xb8\xaf\x79\x63\xaa\x11\xaf\xca\x7a\x13\x6c\x7e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.8
    b"\x69\x2a\xa9\xf9\xca\x80\xeb\x0e\x41\xe8\x0c\x7b\x9d\x43\xce\x6a\x3b\x60\xcd\xbc\x3b\x3d\xed\xc1\xc0\xee\x43\xa3\x8c\x54\x18\x3a": [
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.8",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x69\x2d\xf2\xa8\xd0\x60\x76\xbd\xce\x4b\x9f\xf9\x18\xb1\xb6\x84\x84\x37\xd4\x33\x17\x2a\xe8\x4f\x27\x44\x10\xdf\xe6\x8c\xf3\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7
    b"\x69\x6e\xf6\x64\x78\xb9\x69\x37\x79\xd4\xc5\xb8\x6c\x82\xf8\x7b\x01\x8e\x30\xe7\xce\x96\x62\x9c\x1a\x65\x0a\xb9\x3c\x2f\xa6\xb7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.2
    b"\x69\x7c\xa2\x79\x97\xf6\xa4\x01\x87\xfb\xef\xb2\x88\xc0\x4b\xa8\x59\x29\xbf\x75\x2f\x02\x12\x7c\x9f\x81\xba\xf3\x0f\xe9\x2d\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v3.4.2",
        ),
    ],
    # Hash appears in versions: v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x69\x99\x6d\x10\x53\x0b\x96\x70\x73\xc7\xba\x46\x7d\x6d\x37\xb7\x1d\xad\xee\xfd\x93\x57\xb1\x98\xfe\xa8\xc3\x1e\x4f\xa7\xee\xda": [
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x69\xdd\x9b\xd1\x92\x97\x1f\x39\x13\xa6\x52\x33\x70\x8d\x32\x8f\xeb\x89\x03\x34\x06\x3f\xd6\x5f\xce\x73\x8d\x29\xfd\x72\x92\xb6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x69\xe2\x74\x58\x75\x25\x39\x53\x9f\x8e\x03\x3d\x8f\xc6\x8e\x08\x54\xb2\xe9\x3c\x33\x60\xe4\x5c\x6c\x08\xf8\x2e\x93\x09\x72\x34": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.1.6, v0.1.7, v0.1.8
    b"\x69\xe6\x32\x72\x8a\xee\x39\x0d\xf6\xa9\x42\x35\x34\x59\xf9\x9f\x49\xc2\x02\xcd\x70\x74\x10\x8c\xa7\xec\x9c\x3b\x9c\xf2\x81\xe4": [
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.8",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x69\xee\xe5\xd8\x71\xbc\xa3\x7f\xfa\xb0\x14\xf5\x5c\x4a\xb6\xd4\x3d\xf0\x23\x0f\xfd\x7b\x51\x2e\xda\x17\x91\x49\xf9\xd9\xd9\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/optimism/LibOptimismUpgradeable.sol",
            name="LibOptimismUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x6a\x07\x95\x83\x71\xb4\xd4\x8e\x83\xfb\xcd\xdc\xe2\xc6\x78\x80\x53\x4e\x94\x37\x0c\x40\xb3\x21\x9c\x2f\x88\x99\x2c\x85\xde\xc0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162
    b"\x6a\x15\x11\x13\xab\xd9\x23\x40\xbb\x6f\x62\x7c\x52\xb2\x48\x74\xc0\x6d\x8f\xfc\x77\x32\xf4\x8a\xa6\x27\x98\xef\x0c\xce\x6b\x8b": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.162",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x6a\x33\x9e\x1e\xe0\xe1\x84\x4b\x1b\x17\xf6\xdc\x6a\x65\x3f\x8a\x6c\xc6\x30\xa4\xef\xeb\x49\x9e\xa2\x7d\xf6\x3b\x9b\xd5\x6c\x9a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x6a\x66\xfd\x60\x40\x04\xca\xac\x0a\x4d\x2b\x8f\xed\xa8\xb0\x83\xe5\x59\xa6\x2b\xb2\x81\x6e\xc6\x26\xeb\x2b\x76\xde\x37\x27\x3c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/compound/ICompoundTimelock.sol",
            name="ICompoundTimelock",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.215
    b"\x6a\xa3\x0a\x5b\x94\xcb\xcd\x44\x1d\x44\x62\xd7\x51\xe1\x21\xf7\xa0\xdc\xeb\x96\x50\x00\xe5\x87\xb0\x8b\x13\xc6\x9c\x01\xd5\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.215",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x6a\xed\xc5\xc4\x9f\xd2\xe9\x97\xc2\x57\x3a\x2a\x0c\x71\xcf\xe2\xeb\xd3\x0a\x42\xeb\x15\x1e\x6f\xa3\xac\x02\x11\x2b\xfb\xcd\x7c": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolEvents.sol",
            name="IUniswapV3PoolEven ts",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x6a\xfc\x7f\xda\x95\x09\x7d\x92\x45\x76\xf6\x37\x97\x12\xdd\x6e\x69\xa5\xec\x58\xb1\xc6\xc7\x51\x29\x08\x57\x4c\xd9\xfe\x32\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x6b\x01\x97\x49\x78\x1c\x6b\x47\x54\x31\x59\x90\x8c\xaf\xcf\xbd\x3f\xa5\xee\x69\x69\x83\x8e\xbf\xb0\xdb\xe8\x88\xa0\x21\x36\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x6b\x10\xc9\x45\x39\xa4\x45\x43\x0e\xbb\xdb\x75\x35\xa6\x88\x84\x2b\x7b\x5f\xef\x76\xa9\xc1\x1c\xf9\x34\x98\xaa\xcc\x14\x3e\xc2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x6b\x20\xf6\x39\x43\xfd\xfd\x90\x13\xeb\x3c\xba\x31\xfc\xf8\x6d\xc2\xca\x93\xe4\x0a\x37\xa6\xc9\xe5\x4b\x90\x1e\x99\xcf\x38\x77": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3
    b"\x6b\x21\xb8\xcb\xe4\xbb\x1a\x7b\x59\x23\x87\x76\x82\x9d\xe4\x52\x31\x4d\x61\x91\x55\xcd\xb9\x8f\x8c\x06\x61\x79\xbd\x53\x63\x4a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x6b\x22\x11\xd8\xc7\xbe\x0a\x7c\x86\xe2\xb7\x7b\x97\xa4\xee\x9d\x2f\x55\x4a\x3b\x14\x52\xbf\xc9\xa8\x9f\x34\x13\x74\x15\xc4\xb6": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/NonzeroDeltaCount.sol",
            name="NonzeroDeltaCount",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x6b\x2d\x4f\x3d\x28\x1e\x7f\x32\xc0\x0a\x89\xee\x37\x36\x7c\x09\x5d\x69\xce\x40\x9a\x31\x06\xa5\x48\x6b\x48\xfd\x1c\x70\x4f\xbe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x6b\x33\xa2\x2f\xc6\x29\x88\x24\x00\xa5\xa4\xb8\xa7\xc3\x0a\x85\x78\x0d\xf1\xb4\xb2\xd5\x40\x66\xa6\x59\xc7\x9a\xca\x25\x2b\x05": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x6b\x4e\x4e\xa6\x7b\x56\xf9\xf4\xe2\x35\xc4\xf8\x2a\xb5\x85\x4f\x49\x3f\x05\xc8\xc7\xee\x8f\x31\x31\x4b\x1d\xcf\xb6\x86\x50\xbd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x6b\x55\xa3\x33\x60\xdf\x0d\xba\xf9\x8e\x7b\x24\xc6\xf0\x0a\xe3\xb8\xd3\xda\xb5\xe5\xa3\xe5\x6b\x39\x36\x41\x03\xcf\xef\x08\x85": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashBorrower",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashBorrower",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashBorrower",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashBorrower",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashBorrower",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrower.sol",
            name="IERC3156FlashBorrower",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x6b\xb9\x62\xd4\x3a\x21\xde\x46\x69\x6a\x46\xb8\x8c\x46\xfd\x54\x45\x89\x82\xf8\xeb\xdc\xd3\x34\x92\x31\x46\x8f\x7e\xe7\x31\x9d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x6b\xba\xa5\x43\xd4\x30\xe9\xbd\x55\x2b\xc9\x98\x15\x02\xad\x78\x3b\x88\x47\x5c\xb5\x8b\x5e\x41\x0f\x4f\x3f\x71\x22\x6a\x31\x5c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70
    b"\x6b\xd0\xf0\xba\x49\x94\x44\x9b\xb8\x95\x73\x34\x22\xd7\xb0\x68\xaf\x68\xf8\x20\xdc\xce\x2a\xe7\x3f\x94\x54\xb2\x0e\xd5\xfc\x4b": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.70",
        ),
    ],
    # Hash appears in versions: main
    b"\x6c\x15\x64\x90\x05\x7e\x0d\xbe\x53\xe3\x1a\x14\x1f\xf2\x98\x50\xb2\xb6\x7d\x13\x27\x3d\x2a\x57\x26\x53\x92\xaf\x0d\xfd\x2c\x90": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x6c\x2b\x8a\x96\x4e\xc0\xf5\x8b\x53\xfb\x56\x31\x61\xb9\xee\x5f\x9d\xf1\x70\x43\x2c\xec\xdb\xae\xfc\xc8\x05\x5d\xed\x92\x38\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x6c\x32\x79\x3a\xbe\xdc\xc0\x4c\x53\x56\xe7\x40\x8a\x6b\x33\x04\x0f\x70\x52\x43\x75\x89\x42\x66\x37\xc3\x25\x63\x27\x99\x0f\xa1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x6c\x47\xb2\xb2\xb2\x44\x22\xe5\x25\xa4\x79\x86\x18\x7d\xbd\xab\x00\x17\x81\xec\x37\x7e\x34\xbd\x0b\xbf\xef\xbf\x7b\x6a\x6c\xc3": [
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.270
    b"\x6c\x65\x30\x7f\x50\xf2\xe9\x44\xa6\x97\x37\xca\xeb\x8c\x3d\x02\x90\x13\x58\x70\x81\xb2\x3d\x2c\xc1\x67\x27\xec\x12\x88\x9b\x65": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.270",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x6c\x6b\x59\xe0\x3c\x3b\xfe\x6e\x9e\xc0\x17\xd2\x22\x04\xb3\xac\x23\x1e\x3f\xae\x61\xcf\xdd\x03\x94\x14\xe2\x05\xca\xb8\xd7\xeb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/ERC1363Utils.sol",
            name="ERC1363Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6c\x77\xcc\x52\x04\x53\xce\xca\x19\x9b\xa8\x48\x7e\x4d\x7c\x7c\x28\x19\xee\xcb\x06\x76\xe0\x8d\x94\xec\x4c\x96\x72\x41\x25\x5f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IDelayedMessageProvider.sol",
            name="IDelayedMessageProvider",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245
    b"\x6c\x79\xae\xe1\x4a\xf5\x23\xc6\xd5\x17\x61\xb1\xd4\x05\xd8\x78\x69\x2f\xeb\xac\x93\x0f\xbf\x06\xc1\x95\x5e\x2a\x07\xaa\x01\x0b": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.245",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6c\x7b\x33\x2c\x48\xaf\x92\xda\x0b\xf0\x6a\xb2\xed\x5c\x99\xfa\x41\x36\xac\x6b\x5e\xa2\x61\x43\x77\x59\x06\x27\xf8\x3c\x33\x82": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x6c\x87\x17\x40\x77\xbd\xa2\xd9\x3a\x24\x9c\x84\xa5\xaf\x38\x92\x24\x67\x59\x7e\x00\x24\x4f\xd7\xe7\xe6\x27\x73\x7d\xe5\x58\xeb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x6c\x9e\x43\x34\xb1\xdd\x4e\x6e\xa5\xc3\x9c\x9e\x87\x65\x52\xec\x6f\x0e\x8c\x74\x15\xb7\x81\x90\x9f\xa0\xa1\xf3\x1a\xeb\x7a\x74": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x6c\xa5\xf7\x1f\x78\xe0\x2e\x84\xc5\x7f\x00\xba\x04\xa8\xf9\x5d\x89\x53\xbe\x71\xa8\x40\x95\xd7\x92\x60\x7d\x5c\xd9\xdc\xff\xe4": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x6c\xc2\xcf\x8e\x58\xa5\x22\x66\xdb\x33\x29\x73\x54\xec\xed\xfa\x98\xaa\x94\x04\x63\xb8\xa6\x07\x04\x51\x2e\xa9\x81\x41\x43\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6c\xe6\x62\x27\x64\x8b\xea\x08\x34\xf7\x18\x27\x30\x41\xe9\xf6\x2f\x4d\x2e\x23\x3a\x80\x17\xb5\xef\x54\xe4\xb6\xcf\x15\x38\x97": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x6c\xec\xf6\xdb\x9e\x27\x75\x7d\xea\x33\x87\xbe\x54\xb2\x5d\x36\x64\xbd\x45\x1c\x8d\x29\xbe\x14\xf1\xb7\xce\x65\x82\xc0\x8f\xed": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/Position.sol",
            name="Position",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x6d\x0a\xa1\xe6\x8d\xb2\xb5\x08\x01\xbc\x84\xb5\x27\xca\x90\x48\x49\xa8\x0b\x48\x57\xa3\x2c\x03\x4a\x1c\xd0\x46\xed\x4b\x7c\x90": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v1.0.1, v1.1.1, v1.2.0
    b"\x6d\x1e\x26\x37\x88\x56\xaf\x9e\xbb\x9c\x7b\xe6\xae\x2e\xcf\xc9\xfc\xe3\xac\x50\x99\xca\xe5\x88\xb6\x0e\xa9\x1d\xf6\xf2\xc4\xa2": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungibleTokenPositionDescriptor.sol",
            name="NonfungibleTokenPosi tionDescriptor",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungibleTokenPositionDescriptor.sol",
            name="NonfungibleTokenPosi tionDescriptor",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungibleTokenPositionDescriptor.sol",
            name="NonfungibleTokenPosi tionDescriptor",
            version="v1.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291
    b"\x6d\x36\xf7\xff\x16\x71\xaf\x43\x98\xc7\x7c\xb2\x9d\x92\x5d\x0b\xd3\xda\x63\xb1\x23\xd9\x1b\x4b\x8e\xc8\x9e\x81\xb8\x30\xfb\x9f": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.291",
        ),
    ],
    # Hash appears in versions: v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6d\x57\x03\x5b\xbd\xcd\x62\xd6\x6e\x6b\xf9\x18\x75\xcb\x16\x16\x69\x4c\x38\xea\x32\x30\x4a\x18\xd0\x7e\xa0\x80\x2c\xea\xcd\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281
    b"\x6d\x61\x14\x9d\xa8\x1d\xd6\x48\xe0\x06\xaf\x8b\x62\x67\x8b\xa5\xb9\x90\x50\x92\xfd\x21\x43\x84\x8b\x27\xd6\x8c\xa5\x1c\xe0\x79": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263
    b"\x6d\x7a\x23\x24\xdf\xe8\x22\x68\x6e\xf1\x7d\xbe\x33\x4d\x01\x3e\xd4\xa6\x8f\x5e\x98\xd7\xff\x17\x91\x32\x40\x51\xfc\x56\x71\x1e": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.263",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x6d\x7a\x51\x34\xe3\x51\x02\x66\xc6\xfc\xa5\xec\x60\x61\x9a\x73\x5f\xea\x5c\x76\xa0\x38\xba\xbf\xac\xba\xca\x88\x73\x84\xb6\x1c": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/Tick.sol",
            name="Tick",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1
    b"\x6d\xac\x08\x89\x0b\x05\x63\x47\x0d\xbd\xcc\x72\x02\x4f\xa4\x34\x0e\x80\xc2\xf5\x95\xea\x2d\x43\xa0\xa3\x36\x9b\x51\xf0\x45\x2d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.1",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x6d\xad\x94\x60\x38\x51\x2e\xf4\x64\x8b\x82\x42\x74\x55\xb8\x65\xa1\x94\x5c\x94\xeb\xcc\x60\x09\x4b\x1b\x23\x25\xdf\xaf\xe6\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCast.sol",
            name="SafeCast",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6d\xcb\x3a\xdf\x15\x73\x2d\x96\xcc\xc4\xf6\x99\x06\x51\xb2\x4a\xb9\x40\x4f\xfc\x04\xd9\x43\x73\x19\x45\x3f\xda\xd6\xb5\x76\xa1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x6d\xce\x01\x8f\xe9\xc2\xf2\x9b\x3d\x70\xde\x12\xf8\xdb\x13\x90\x43\x11\x1d\x61\xa4\x06\x1b\x69\xa4\x42\x61\x7b\x07\xe2\xeb\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: main
    b"\x6e\x3b\xee\xa9\x33\x28\x60\x5e\x32\xfb\xe0\x06\x88\xc0\x5a\xe4\x01\x9e\x1e\x30\xa9\x48\xd9\x26\x16\xb8\xf3\x63\xb0\xa4\x54\xf3": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/PoolInitializer_v4.sol",
            name="PoolInitializer_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x6e\x4b\x6e\x0e\xfa\xa6\xa7\x14\x30\x1c\xe5\x37\x58\xb5\x4b\x3e\xf8\x36\x22\x61\xf1\x71\xc0\x0d\x94\x89\x17\x3d\x53\xf2\xc2\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x6e\x55\x5b\x75\x27\xf0\x07\xb6\xec\x8e\xb2\x86\x5a\x30\x72\xb5\xc7\x86\xb6\x23\xfe\x7d\x7d\x41\xe5\xcb\x03\xcb\x64\x5c\x8b\x3e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x6e\x56\x39\xa1\x9d\x28\xb0\xec\xd9\x35\x2d\x1f\xc1\x22\xa8\x6b\xf7\x9d\xda\x70\x88\xae\xa1\x12\xd6\x2a\xe6\x06\x9f\x4b\x66\x94": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x6e\x5d\x98\xe0\x42\xe2\xc4\x67\x9d\x16\xf3\xaf\xbf\x93\x9b\xf6\x44\x9a\x5e\x39\x91\x2f\xe5\x4f\xc2\xd0\xdd\x65\xfe\x80\x68\xa7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x6e\x71\x64\x36\x84\xba\x17\x00\x7a\xdc\x19\xae\xaa\x6a\x9a\xd0\x5b\xf5\xbf\x9b\x81\x39\xf4\x1e\xa6\x68\xf0\x65\xe3\x81\x97\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.148
    b"\x6e\x87\x3d\xce\xb0\x17\x6a\x86\xae\xd1\x9b\x4f\x52\xb9\xb1\x2d\xd9\xeb\x71\x1e\x94\x4d\x3f\xe8\x8b\xba\xb3\xfd\x77\x17\xef\xd3": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.148",
        ),
    ],
    # Hash appears in versions: v0.0.80
    b"\x6e\xe4\x6f\xa0\xdc\x04\xa2\xdf\x2e\xd5\x10\x65\x14\xb8\x83\x60\x65\x33\x50\x60\x8e\xe8\x9d\xca\xaa\xdd\xc1\x33\x8c\x51\x81\x27": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.80",
        ),
    ],
    # Hash appears in versions: v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x6e\xea\xca\xb4\x7d\xb8\x5c\x02\x5e\x33\xa4\xf0\xaf\x7f\xc3\x19\x86\x59\x7b\x60\x6a\x8f\x4d\x6d\x4a\x6e\x7c\x6d\x7b\xee\x7e\x38": [
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.240, v0.0.241, v0.0.242
    b"\x6e\xf1\x81\xfc\xf4\x52\x5e\x1b\xb5\xe0\xa9\x9f\xc6\x42\x31\x2c\xef\xce\xb4\x99\x4c\xa4\x83\xc4\x6a\x1f\xac\x32\xfd\xb1\xbb\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.242",
        ),
    ],
    # Hash appears in versions: main
    b"\x6e\xf4\x74\x66\x96\xe8\x7b\x51\x8d\x90\x56\xad\x5d\x74\x1d\xaf\xc7\xfa\x1e\x69\x95\xb1\x7e\x7a\xfa\x33\x2e\xc0\xbc\x31\xc1\x53": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IImmutableState.sol",
            name="IImmutableState",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x6e\xf7\xd4\x1e\x7c\x35\xea\xc0\x0c\x76\x7b\xff\xf8\xfa\xc6\x9b\x95\x9d\xb1\x24\xf3\x6b\xad\x2e\x75\xd0\xf0\x3c\x13\x22\xad\x7e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x6f\x03\xe1\x86\x67\x72\x81\xf8\xbb\xb4\xae\x0c\xcc\xd7\xf0\xeb\x60\xfc\x10\x0d\x9f\xd7\xe6\xb7\x7e\x1b\xe4\x1e\x8f\x70\x95\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x6f\x05\xf2\x09\xdd\x54\x63\xa2\x37\x7a\xde\x32\xd1\x5a\xfe\xa8\x21\x10\xba\x4f\xcd\x18\xf1\xe9\xab\xde\x3e\x80\xe5\xbb\x94\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.1, v3.0.0-rc.0, v3.0.0-rc.1
    b"\x6f\x20\xf1\x7a\x4f\x8f\x45\xb9\xc1\x00\xe8\x5a\xf5\x8f\x20\x91\xd6\xce\x51\x9a\xc9\x19\x5c\x45\x71\x8b\xae\xb3\x05\xae\x7a\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x6f\x24\xbb\x8d\x6a\x5e\xbe\xa9\xf7\xdb\xca\xc8\xd6\x41\x23\x50\xfe\xa2\x5f\x31\x79\x34\x95\x8c\x91\xc0\xca\xc2\x99\x9a\xec\x64": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x6f\x34\x29\x3a\xcd\x18\x4d\xf4\x63\xb4\x27\xf8\x42\x66\x59\x9a\x20\x5c\x5e\x00\x38\xb6\xa2\x9a\x4b\x1c\x87\xa8\x44\xa7\x1d\x43": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFraction.sol",
            name="GovernorVotesQuorumFraction",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x6f\x54\x68\x03\x4d\x7a\xe2\x3f\xa4\x22\xf9\x02\x17\x1f\x83\x62\x37\xb7\xff\x22\x0c\x3b\x2d\x68\x8e\x7e\x22\xb1\xb9\x17\x46\x76": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x6f\x5a\xe8\x9d\xb5\x6e\xb8\x80\xa0\x8a\xbe\x6e\xdc\xda\xcd\xc8\xf3\x90\x22\xe9\x0a\x3e\xdf\x28\x00\x42\x7e\x5c\x27\x93\x13\x5b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x6f\x74\x2f\x3f\x7f\xec\xcb\x3b\xa5\x89\x39\x9b\x2e\xde\x3a\x1c\xe0\xc3\x98\xdd\xe2\x23\x7b\x85\x98\xbb\x59\x08\x23\x20\xaf\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.5, v0.0.6, v0.0.7, v0.0.8, v0.0.9
    b"\x6f\x87\xd4\x34\x17\x5b\x02\x5a\xa3\x59\x23\x1f\x30\xc1\x0d\xab\x16\x4d\x18\xca\x68\x93\x60\x44\x90\x7d\x26\xa0\x25\x47\x41\x64": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.7, v0.0.8, v0.0.9
    b"\x6f\x9a\xbc\x0b\x61\x84\x9d\xe3\xa2\x18\x8e\x99\xb2\x2d\xee\xe5\xe8\x07\x08\x44\x4b\xe7\xa8\xbb\x8c\x21\x9c\x8b\xe2\xd2\x4e\x31": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.55",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x6f\x9d\xdb\x1c\x2a\x55\x85\xb6\x39\x2c\x15\x77\x90\xf8\xdd\x44\x4f\x06\x78\x5d\xe4\xe3\x71\xcb\x4a\x7a\x3e\x63\x5e\x65\xd1\x22": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/SqrtPriceMath.sol",
            name="SqrtPriceMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x6f\xba\xc0\x65\x03\x44\x9a\x67\x6e\x2a\xa8\xc7\x8c\xff\x90\x20\x60\xc2\x50\x8c\xac\xb3\x2c\x37\xa7\xdd\x86\xb3\x50\x07\x47\x0e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.129, v0.0.130
    b"\x6f\xc9\xaf\xcc\x12\xb8\x7e\xac\xd3\xeb\x90\xed\xa8\x0b\x3b\x04\x1a\xf0\xef\xec\xec\x69\x81\x10\xba\x48\xa4\xc7\xe3\x51\x90\xe2": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.130",
        ),
    ],
    # Hash appears in versions: v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132
    b"\x6f\xcc\x3d\xf9\xf1\x0b\x8d\xd7\xdf\xf8\xd6\x23\xb9\x5f\x74\xb7\x9c\xef\xe0\xcd\x0c\x53\x7e\xad\x1c\x66\xa9\x3f\x7e\x88\xfc\x43": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.132",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x6f\xe5\x5c\xb8\xee\xe1\x1f\x67\xc7\xaa\xbf\xdc\xbc\x00\xc9\x2a\x2e\x8e\xd4\x56\xc1\x51\xed\x84\x0f\x20\x3b\xed\x4f\x61\x40\x1c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x6f\xe8\x29\x62\xa7\x26\x40\x35\xd8\x99\x24\xf1\xc8\x8d\x8a\x92\x86\x3d\x64\xa0\x69\x5d\xaf\xd1\xe8\x90\x8c\x3f\x45\x87\x62\x24": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\x6f\xfc\xb1\xa0\xdf\xd0\x49\x39\xa9\x36\xc2\x82\x30\x26\xaa\x5f\x0a\x11\xf8\x60\xfc\x04\xb5\x2a\x2d\xa3\x0c\xd3\x07\xd0\x67\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.260
    b"\x70\x03\x6b\xd0\x08\x91\x07\xe2\x34\x8e\xc6\x5d\x26\x40\x9d\xd9\x2f\x28\xc5\xb1\x32\xce\x61\x4a\x4b\x47\x71\xb3\x27\xe8\x8e\x19": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.260",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x70\x05\x9f\x13\xe7\xab\x1b\xe9\xf3\xdc\xf5\xc8\x0d\xfd\x2a\xef\x27\x84\x08\x4f\x10\xba\x43\x88\x8c\x25\x88\xfc\xb7\xd2\xb2\x07": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x70\x0a\x31\xc8\xd0\xbf\x96\x48\x81\xea\x1b\x2e\x21\x8e\xcd\x4e\x92\x61\x6d\x29\x64\xa5\x22\x36\x28\x59\xb5\x5e\xd8\x7c\xf0\x44": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x70\x10\xe8\x47\x60\xbc\xa9\x82\xd0\x39\xdb\x51\x3c\x79\x91\xe1\x33\x48\x01\x06\xbc\x7f\xbd\xc7\x3c\x33\xde\xeb\x43\x83\x30\xf4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x70\x25\xfe\x36\x42\x9a\xf5\x48\x41\x00\x03\x64\x30\x3a\x6a\x1e\x12\x6f\xff\xab\xd5\x34\x47\x28\x48\xe4\x9d\xf2\xa2\xa3\x24\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AuthorityUtilsUpgradeable.sol",
            name="AuthorityUtilsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x70\x3c\xa3\x51\x4f\xde\x72\x76\xef\x92\xb1\x47\x68\xae\xb4\x03\x49\xc1\x8c\x5e\x18\xe3\x20\xbe\x83\xc5\xcf\x28\xbd\xbf\xe3\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x70\x63\x3c\x49\x99\x36\xe9\x63\xa7\xce\x69\x09\x30\x5a\x15\xae\xf7\xfb\x01\xc2\xfb\xe5\x3d\x8c\x6a\xf5\xaf\x61\x7d\x41\x78\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x70\x67\xd6\x98\x98\xb2\x7e\x58\x8e\xce\xf5\xbc\xb3\x09\x6e\x87\x17\xd6\x03\x65\x5c\x3d\x97\x8e\x51\xfd\x68\x4e\x15\xe4\x06\x3b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IBridgeUpgradeable.sol",
            name="IBridgeUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x70\x98\x1e\x87\x00\xec\x22\x52\x79\x5a\xfa\x55\x54\x0b\x8c\xe5\x19\x33\x43\x71\x77\x54\xd9\x30\xea\x30\xc1\x25\x31\x56\xfc\x49": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v4.5.0
    b"\x70\xa5\x1d\x8d\xaf\xf3\xc7\xcd\x12\x7a\xe0\xbc\x9e\x33\x7f\x6f\x27\x57\x16\x9a\x3e\x37\xea\x60\x01\x80\x6c\x36\x8d\x52\xd6\xe0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.5.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x70\xab\x0d\xd9\x55\x74\xc7\xf8\xb3\xbc\x2b\x2b\xa7\xfb\x78\xdf\x9c\x77\x5e\xd3\x84\xf3\xcb\xe2\x99\xc5\x13\x21\xa1\xb9\x3e\xb4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x70\xd9\xb8\x30\x78\x45\x6a\x6f\xc7\xd7\xa8\xec\x55\xdd\xfa\xe3\xe8\x3c\x54\x38\xfe\x72\x0d\x6d\x45\x12\xa2\x23\x11\x04\xf4\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x70\xf0\xe6\x2a\x7a\xb0\x23\x36\xc0\x19\xa8\x07\x77\xec\xbc\xf3\xda\x51\xbf\xb5\x52\xe7\x30\xa0\x60\x0b\xcd\xdc\x48\xbf\xfa\x7d": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x71\x03\x0d\x26\x4e\x72\x95\x1a\x47\xc9\xa8\x73\x37\x25\x19\x56\x6e\xfa\x90\x1c\xda\x81\x33\x36\xf5\x6d\x69\xa1\x28\xbf\x8d\xa6": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/BitMath.sol",
            name="BitMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x71\x1f\x72\x1f\x68\xc7\x2d\xa7\xba\xaa\x73\x65\x49\x26\x92\xe1\xc8\x54\xe0\x9d\xd9\x3e\x18\x37\xe1\x38\x11\xd8\x85\x1b\x70\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.51, v0.0.52
    b"\x71\x29\xc9\x43\x57\x56\x0b\xc9\x6b\xc7\x81\xcd\x07\xb6\xe0\x69\x63\xc2\xe8\x15\x0b\x50\x8a\xf9\x57\xa4\x3f\xb0\xa1\x7c\x4f\x26": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.52",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x71\x8f\xdd\x6c\x85\x38\x23\x15\x7d\xfa\x6f\x85\xcf\xe5\x87\x4b\xaf\x82\x80\x15\xd9\xe0\xdd\x8c\xaa\xd3\x7c\x1b\x15\xe1\x80\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x71\xbc\xe8\x26\xde\x5a\x32\x2e\x46\x9f\xd9\x20\xa7\x76\x1a\xb9\xd3\x6c\xba\x8d\x3a\x97\xea\x3d\x08\xf3\x9a\xaf\x4f\xb0\x18\xd3": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IV3Migrator.sol",
            name="IV3Migrator",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IV3Migrator.sol",
            name="IV3Migrator",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IV3Migrator.sol",
            name="IV3Migrator",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IV3Migrator.sol",
            name="IV3Migrator",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IV3Migrator.sol",
            name="IV3Migrator",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x71\xc5\x1e\x14\x7f\xd0\x96\x76\x99\xa7\x02\x4f\x81\x97\xad\x23\xcb\xc3\x5a\xe5\x6b\x13\x59\x89\x6a\xb7\x7d\x37\xae\x6e\x18\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x71\xd4\xe1\x6b\x17\x98\x34\x4d\x5b\x70\x1a\xd6\x82\xdb\xc9\x0a\x16\x6b\x18\xae\x02\x67\xeb\x9a\xba\xa7\xe5\x39\xd2\x86\x17\x1e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\x72\x12\x49\x0c\xb3\x60\x20\xfc\x8a\x96\x63\xf4\xdf\x96\xdc\xa4\x34\x45\x77\xd4\x43\x8c\x9d\x5a\x11\x3e\xcf\xb0\x5a\x2e\x30\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x72\x30\xb7\x6a\xa4\x61\x10\xf0\x49\x22\x31\x89\x14\x25\xa0\x2d\xc5\x4f\x90\x7c\xe5\xa6\x8b\xbe\xf1\x30\x08\xa5\x0c\x7b\x84\x85": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x72\x46\x5e\xa5\x8a\x8a\x09\x92\xef\x3f\x15\xb0\x76\x6d\xa8\x10\x11\xca\xf1\x94\x53\xfd\x93\xf3\xc1\x27\x8d\x8d\xaa\xf6\x5a\x43": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x72\x4d\xbf\xfc\xf7\x4d\xc4\xc5\x5b\x48\xcf\x9d\x4f\x76\x01\xdd\x6e\xd7\x0b\x00\xb7\x09\x55\xaa\xed\x8e\x77\x31\x4f\xbf\x8b\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x72\x61\x79\xee\x58\xb5\x40\xe2\x25\x26\x00\x6f\x48\x36\x10\xb6\xc9\x74\x36\xf5\x1d\x65\x9f\xd1\x5b\x16\xf3\xb1\x4d\x59\x4a\xab": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/ActionConstants.sol",
            name="ActionConstants",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x72\x70\xc1\x09\x48\x94\xd7\xd7\xde\xe0\x20\xf8\x13\x43\xfb\xea\x7e\x86\xfe\x00\xa9\xdb\xea\xfc\xb7\x4c\xcd\x57\x91\x19\x43\xcd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Pausable.sol",
            name="ERC721Pausable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0
    b"\x72\x89\x69\x83\xdc\xa7\x52\x9c\x8a\xc9\x0a\xa3\x58\xf3\x4e\xf8\x25\x9c\x47\xb4\x38\xfe\xdc\x0a\x1b\xa4\x64\x2a\x16\x12\x56\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x72\xa5\xf8\xdb\x93\xd5\x87\xa8\x09\x2e\xd8\x4c\x83\xdd\x13\x7f\xf4\x47\x99\xc8\x45\x83\x56\xbd\x62\xf6\x89\x10\x71\x39\x1d\x61": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\x72\xa9\x90\xc4\xc4\xc8\xbd\xf2\x2a\xb8\x2a\x8e\xf5\x61\xc3\xe6\x31\xb5\xb9\xf9\x5a\xfe\xc3\xf8\x9c\x23\x46\xfe\xb1\xa0\x63\x0a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x72\xaa\xf9\x25\xc8\x98\x87\xf4\xbf\x9e\xdb\x86\xde\xe9\xb5\x6f\x97\x85\xaa\xb4\x80\xb7\xae\x65\x10\x13\x5a\x41\xa2\x41\x93\x55": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Receiver.sol",
            name="IERC1155Receiver",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.135
    b"\x72\xbd\x78\x49\xfa\xef\xf0\x98\x1d\x92\x85\xd7\xfc\x34\x50\x0d\x64\x1d\xa7\x19\xca\x67\xaa\xba\x41\xb9\x29\x75\x1b\xb2\x3f\xd3": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.135",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x73\x01\xc3\x9c\x8d\xcd\x40\xbd\xa8\x4d\x76\xbd\xb3\xf6\x64\xdb\x46\x5b\x36\x16\x6a\x38\x7d\x77\x25\x24\x51\x52\x0c\x61\x27\x3b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x73\x06\x18\x73\xef\x0e\xa0\x27\x9b\xa9\xbf\x14\x4b\x42\x77\x30\xf2\xe4\x18\x83\xb7\xf8\xe3\x3f\x63\x4c\x4d\x27\x1d\xb3\x54\x1e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x73\x0c\xdf\x93\x09\x9b\xac\x22\x50\x83\x53\x4f\x0a\x7a\xec\x49\x91\x97\xe6\xe4\x0a\xfa\xe2\x1b\x77\x35\xfd\x54\xda\x3a\xa2\x8c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauserUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30
    b"\x73\x17\x80\x9e\x54\x0e\x67\x2a\x41\x74\x9f\xe4\xa1\x05\xad\xec\x73\x8a\xa9\x27\x6f\x71\x44\xc5\xc6\x17\xe2\x7d\x9e\xb4\xd7\x83": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.30",
        ),
    ],
    # Hash appears in versions: v0.0.261, v0.0.262, v0.0.263, v0.0.264
    b"\x73\x3b\x48\x04\x1a\xa7\xcc\x80\x73\x69\xfd\xfe\x2f\xb5\x77\xce\x5f\x79\xb8\x4e\xe6\x94\x54\xed\x91\x7e\x04\x5d\xb1\x8a\xc8\x51": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.264",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x73\x67\x1c\xfb\x2a\xed\xc8\x85\x0a\xa1\xb1\x67\xf5\xe8\xc8\x28\x07\xc5\xd0\x6f\x00\x99\x30\x4b\xba\xbd\x3b\x8b\x74\x65\xb7\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0
    b"\x73\x6a\x0d\xab\xa2\xc5\xa3\x40\x84\x18\x85\x5a\x09\x9f\xe5\xee\x7b\x6f\x56\x66\xc4\x7a\xf3\xd7\x67\xb0\xc2\xb5\x7d\xf8\xf2\x07": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPayments.sol",
            name="PeripheryPayments",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPayments.sol",
            name="PeripheryPayments",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPayments.sol",
            name="PeripheryPayments",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryPayments.sol",
            name="PeripheryPayments",
            version="v1.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x73\x70\x23\x40\x1a\x3b\x49\xe5\x64\x2e\x29\xb6\xb9\xa4\x19\x0a\xa2\xb9\x69\x2e\x32\x65\x2c\x61\x5c\x50\xc6\x95\x23\x0e\xd7\x26": [
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26
    b"\x73\x7c\x14\xff\xae\x38\xd2\xc5\x51\x7d\x4e\x40\xc5\x37\x08\x58\x00\xd7\x80\xb4\xb3\xf4\xcd\x23\xc0\x1e\x7d\xbb\xd9\xfa\xbb\x6f": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.26",
        ),
    ],
    # Hash appears in versions: v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201
    b"\x73\x87\x74\x54\xf5\xf4\x55\xdd\x51\xfc\xe6\x3b\xad\x61\xfa\x6e\x78\x93\x19\xe5\xd6\xb0\x79\xa8\x2a\x4e\x9b\x04\xc6\x68\xad\x9f": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.201",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x73\x8a\x02\x08\xc3\x59\x1e\xdb\xfe\xbb\x56\x0e\xbc\x91\xf7\x88\x85\xb0\x9c\x2b\xc4\xaf\xbf\x14\x53\x1e\x09\xdf\x54\x88\xaa\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x73\x8d\x0f\xa1\xc7\xf2\x49\xa4\x8a\xf9\xab\xe9\xed\xd6\x94\x00\xca\xd1\x53\xa0\x81\x48\x73\xd7\x37\x7d\xa9\xb4\xe5\xff\x14\x6d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.169
    b"\x73\xa6\xe7\x3a\x27\xe2\x5c\x65\xa8\xeb\xb3\xe1\xc6\x67\x79\xdd\x2f\x2d\xa5\x94\x02\xdd\x9b\xdd\xbf\xb8\x6f\xc0\xad\x7c\x27\x3a": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.169",
        ),
    ],
    # Hash appears in versions: v0.0.114
    b"\x73\xbc\xff\xa8\x56\xfc\xfc\x6e\xac\x76\xda\xfe\x99\x12\x0c\x6d\xc0\xef\x59\x8c\x71\x7e\x65\xc8\xfb\xe3\x00\x6d\x2d\x11\x74\x62": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.114",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x74\x00\x84\x6b\x27\x95\x2f\x62\x7a\x8e\xf0\xff\x62\x38\x24\xb4\xbd\x24\x58\xee\x51\xd3\x29\x9b\xe3\x79\x06\x8e\x19\xa1\x94\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.17, v0.0.18, v0.0.19
    b"\x74\x5d\xf3\xd7\x12\xc2\x28\x47\x27\xbd\x70\xa0\x8f\x22\x7f\x07\x5d\x51\x29\x64\xcf\x61\xad\xb4\xab\xc4\x37\x8d\x67\x96\xc9\x92": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4
    b"\x74\x73\x9d\xba\x06\x24\x2e\x08\xdf\x68\x4f\xf6\xa3\x30\x96\x6a\x5d\x1b\xfa\xbe\x34\x31\x7e\x8b\x2f\xa2\x2a\x15\x84\x98\x51\x6a": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.4",
        ),
    ],
    # Hash appears in versions: v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181
    b"\x74\xa3\x3e\xf5\x2a\x40\xde\x4c\x38\x16\x8a\xa7\xc9\x9c\x57\xee\x7d\x52\x40\x45\xc2\x71\xc5\xbc\xca\x05\x77\x30\xd8\xbc\x93\xd9": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.181",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x75\x30\x5b\x44\xe1\x0f\xab\x32\x42\xf9\x4c\x8f\x08\x0f\x25\xc0\xdd\x28\x28\x54\x60\xe4\x6a\x0f\xb6\x8a\x4a\x20\x37\xc2\x6e\xf5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x75\x47\xa2\x71\x19\xe1\x01\x16\x87\xe0\x79\x04\x72\xd3\xe0\xe1\x14\x10\xf5\x66\xef\xc4\x6a\xee\x1e\x2d\x69\xe4\xaf\xf8\xac\xdc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.84
    b"\x75\x78\xdb\x8e\x9c\xaf\x34\x2b\x87\xa8\xe0\x34\x92\xf1\x5b\xa3\x46\x3f\xb8\xe8\x81\xa4\x80\x20\x7c\x40\x77\x32\x09\xf3\xa8\x79": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.84",
        ),
    ],
    # Hash appears in versions: v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123
    b"\x75\x8c\x86\xc2\xed\xe0\x4f\x97\x76\xbd\x31\xf6\xa6\x58\x52\x9e\xb2\x77\x74\x7d\xb2\xd7\xdc\xed\x45\xbe\x54\x90\x29\x8b\xac\x21": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.123",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x75\x9e\x2b\xd5\xd2\x7e\x4e\x4d\xa3\xc6\x1a\x66\x8f\xc0\x65\xee\xf2\x31\x57\xed\x71\x62\xf4\xf3\x06\x69\x94\x6b\xc7\xe6\x8a\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.127
    b"\x75\xa6\x94\x9d\xc7\x0b\x76\x92\xd4\x5d\x25\xfc\x5d\xac\x11\xab\xc2\xc3\xa8\xd1\x8b\x9e\x0f\x33\x1c\x29\x0f\xbb\x5a\xd9\x56\x06": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.127",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x75\xb3\x52\xcc\x1d\x04\xfa\x26\x27\x8d\x4c\x33\xee\x43\x40\xaf\x9c\x1c\x2d\x98\x60\x7d\x49\xc0\x29\x9d\x2a\x44\x28\x03\xbb\xa5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x75\xe0\xde\x6c\x8e\x12\x1d\x32\xe1\x15\x2f\x90\x83\xb1\xcb\xa6\xd9\xbe\xc3\xf5\x5f\x16\x39\xbb\x88\xe2\x7b\x16\x8f\x39\x57\xc1": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolOwnerActions.sol",
            name="IUniswapV3Po olOwnerActions",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.273
    b"\x75\xe5\x71\xf9\x9a\x6f\x98\x49\x4d\xaa\x42\x86\x1a\xb1\xbe\x59\xcf\x89\x3e\x39\x6a\x87\x6d\x04\xfe\x42\xcf\x7c\x6b\x48\x16\x18": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x76\x05\xed\xe9\x86\xac\x08\xb8\x95\x55\x29\x93\xe5\x67\x28\x62\x27\x41\xdb\xd0\x98\xce\x47\x79\xcd\xc5\x0c\x04\xc5\xb3\x7c\x8e": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x76\x06\x5b\x5d\xd3\xb1\x4e\xf1\xf7\x0d\xcd\x58\x41\xad\x52\xd4\xdb\xd6\x70\x32\x9e\xf6\x51\xca\xc5\x03\x67\xda\xc0\x01\x97\x4d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.272, v0.0.273
    b"\x76\x08\x7f\x8f\x80\xf2\xe4\x35\xf6\x8b\x18\x39\x7c\x23\x14\xbe\xf1\x6b\x80\x18\xe8\x89\xb5\x9d\x2e\x12\xfc\x85\xd0\xc2\xdb\x39": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x76\x0c\xb1\xa1\xfb\x07\x1e\x4d\x56\x7f\x9e\x35\x43\x8e\x9c\xaf\x78\x22\x57\x0a\x13\x3e\x1f\x51\x8e\x5b\xd7\xf0\xd0\xf0\x36\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x76\x25\x36\xd6\xe8\x25\x0c\x9d\xd7\xd7\x78\x47\xe6\xcf\xdf\xf9\x84\xd6\x34\x3c\x64\x1a\xda\x84\xe8\xef\xfa\x61\x2f\x52\x3b\xb9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63
    b"\x76\x29\x9e\x7f\xf2\xa0\x81\xeb\x3b\x3f\xc5\xf5\xcd\x2d\xd8\x4c\x65\x9f\x1c\x34\xf6\xb4\x25\x80\x00\x3d\x25\x75\x16\xa1\xb1\xef": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.63",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x76\x40\xb0\x61\xb7\xc0\xc1\x10\x2e\xb2\x13\xdb\x12\x0c\x66\xba\x35\x46\xa5\xb8\xcb\xaf\xa7\xc3\x11\x22\x62\x06\x2f\x19\xf6\xca": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\x76\x44\x5f\x1e\x78\x22\xec\x47\x43\x27\x0b\xf8\x19\x2b\x16\x70\x14\xf9\x96\x5a\x5e\x20\x05\xc6\xb7\x7c\x34\x5b\x21\xec\x2e\x9e": [
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162
    b"\x76\x62\x93\x9d\x5d\x4b\x3a\x65\xe4\x2b\x9d\xa9\x45\x0b\x5a\x06\x8d\xcd\x12\x44\x52\x38\xb0\x07\xa7\x30\xe8\xd0\x0e\x47\x47\x8f": [
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MetadataReaderLib.sol",
            name="MetadataReaderLib",
            version="v0.0.162",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x76\x71\x69\xfa\xdc\xbd\x89\x2f\xad\xb8\x11\x23\x37\xd4\xab\x9e\x4e\x98\x7b\xd1\x99\xe1\x87\xf3\x5a\x3e\x96\xed\x0f\xa7\x6e\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x76\x82\x75\xbd\x9e\xa4\xa2\xdc\x08\x2c\x97\xfc\x77\x65\x0d\x85\xf9\x81\x48\xb2\x48\x69\x0f\x62\x6d\xe1\x87\xb5\x18\x3a\x06\x1a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x76\x83\xc3\xcd\x54\x95\x29\xde\x0d\xd1\x2b\x50\xa4\x20\x2e\xd1\x96\xf6\x91\x4f\x2d\xc6\xe5\x98\x08\xae\x18\xe1\xa5\x06\x45\x88": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x76\x8d\x31\x26\xff\x93\x30\xc2\x6a\xbe\xc1\x8e\xf0\xdc\xcf\x80\x87\x26\x79\x9b\x9e\x19\x19\xec\xa2\x11\x30\x95\x91\x8e\xc2\xb4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x76\x9c\x44\xb2\xe9\xa7\x8b\x12\x42\x5d\xe9\x7f\x12\xd5\x35\x72\xec\x2b\xbd\x2e\xe5\x54\x6b\x0a\x5b\x02\x56\xba\x7d\x03\xb3\x18": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x76\xf3\x1b\x61\x02\xe3\x1c\xfd\x25\xf9\xfc\x82\xa3\x92\x96\x14\x30\xda\x3d\x67\x60\x6a\x38\xbc\xd2\x31\x18\x26\xcf\xfa\x22\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x77\x13\x55\xe2\xe8\x72\xb5\x6a\x92\x38\x02\x1e\x2c\xb1\x5e\x30\x85\x38\x18\xe2\xb1\xd5\x8c\x60\x9e\xcd\xab\x01\xa3\xe6\xdb\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x77\x1f\x32\xe1\x20\xe5\x7a\x8f\xe4\x29\x10\xab\x5f\x48\x6e\x08\xd4\xe3\x11\x7e\x4b\x90\x0e\x5c\xb8\xc4\xd3\x90\x71\x4a\x60\x6a": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/ParseBytes.sol",
            name="ParseBytes",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x77\x26\x92\x0c\xdf\xa6\x16\x6f\x00\x71\xa1\x29\xfb\xb2\xa4\xac\xa1\x41\x79\x68\xc8\x99\x9c\x10\xc9\xed\x29\x9f\x22\xaa\xa7\xf0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x77\x33\x2b\xb3\x0c\xc2\x21\x5b\xec\x2a\xb1\x0d\x81\x40\xc9\xe7\x8c\x30\xff\xfa\xea\x5c\x36\xf7\x0d\x38\x75\x1c\x3b\x31\xfb\x58": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x77\x62\x61\xac\xf6\x72\xde\x2f\x63\x68\xa1\x41\xd4\x91\x0d\xd1\x69\x75\x7b\x90\x91\xf3\x21\xd4\x2b\x79\xa5\xfc\xae\x97\x4c\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1820ImplementerUpgradeable.sol",
            name="IERC1820ImplementerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x77\x8b\xb1\x5e\xcd\x04\xf9\x93\xc1\xeb\xcd\x93\x3b\xab\xdd\x11\xe8\xc8\xe2\x3d\xbe\x09\x28\x47\x95\x90\xe8\xd4\xb4\x49\xa9\xf4": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x77\xb8\x02\x67\xb4\x1e\x68\xc5\xac\xb0\x75\xcc\x38\xe3\xa2\x12\xb3\xee\x24\x66\x70\x99\xf8\x98\x80\x80\x8f\x11\xdd\x14\xcd\x97": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/LiquidityAmounts.sol",
            name="LiquidityAmounts",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x77\xc4\xcb\x1e\x42\xf5\x0d\x23\x2a\x5b\x33\x8f\x31\x4f\x46\x17\xb3\x22\xa0\xe1\x9d\xb9\xd6\xb3\xed\x7d\xfe\xdf\x82\x6b\x0c\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x78\x08\xa2\x7a\x60\x0e\xcc\x86\xfc\xbc\x45\xf7\xd6\x93\x21\xb9\xe6\xd7\x26\x12\x0f\x48\xed\xd2\x7d\x28\xd4\x10\x2d\x9c\x7a\xfe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x78\x0b\xe8\xe8\xd7\x1f\x67\x64\x9a\x47\x42\x7a\xdf\x1f\x8d\x09\x98\xcf\xb8\xcd\xb0\xd3\x6b\x01\x14\x74\xfb\xa0\xf4\xc4\x46\x70": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/EIP712.sol",
            name="EIP712",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x78\x18\x3f\xf4\x8f\x72\xce\x26\x93\x35\x6b\x2a\xa8\x3d\x5d\x0a\x6b\x32\xf0\xad\x30\x2d\x66\x6f\x88\x8b\x3f\x01\xaf\x43\xdd\x22": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x78\x35\x9f\x63\xa0\xa9\xab\x0f\x2c\x5e\xe8\xc5\xc1\x1e\xe2\xde\xb2\xe4\x3c\xea\xb1\xfd\xe6\x25\xcf\x0a\x83\x3f\xf0\x9f\x5a\x4d": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x78\x46\x3b\x86\x7a\x52\x51\x24\x98\xcd\x6a\x32\x35\x52\x31\x84\x0a\xb1\xdd\xe8\x34\x6d\x66\x5b\x84\x19\xcc\x01\x8a\xca\xa6\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.98, v0.0.99
    b"\x78\x5f\x5f\x12\xbb\x00\x4e\x9e\xef\xd4\x44\xcd\x18\xe1\xc0\xff\x15\x80\x30\xa7\x6e\x63\x39\x31\xfe\x93\xae\x36\x0d\xa1\x4f\x14": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217
    b"\x78\x9b\x58\x14\x56\xa6\x77\xa9\xac\x26\x41\x35\x1d\x72\x5d\xd4\x7c\x3a\x05\x75\xd0\x6a\xac\x86\xb3\x52\xa8\xab\xbd\x94\xf0\x39": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.217",
        ),
    ],
    # Hash appears in versions: v4.7.1
    b"\x78\xde\xad\xc2\x56\xa9\x3b\x74\x09\x07\xa4\xfd\xfc\xdc\x01\x58\xf2\x7c\x74\xe6\x34\x46\x21\x48\xdd\x97\xf6\x14\xe4\x71\x19\x79": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.7.1",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\x79\x04\x5d\x21\xec\x43\x75\x1b\xa1\xf8\x88\x1c\x64\x28\x4f\xd3\x15\x4c\x79\xa8\xa2\x1e\x76\x8f\x4a\x09\xe2\xb4\x59\xc2\xd8\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x79\x33\x0a\x4c\x3e\x67\x3e\x13\x36\x8c\xbe\xb6\x9c\xf9\x9b\x9a\xf0\xf8\x62\xb4\x5f\x54\x43\x51\xb5\xce\x8e\x67\x8f\xb6\xda\x05": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239
    b"\x79\x51\x72\x5a\x9e\xa0\x02\xf3\xb7\x7a\xe2\xcd\xab\x97\x2e\x5a\x56\x37\xf6\x79\x56\x4d\xfa\x7b\x31\x1b\x95\x28\xe5\xc8\x4d\xb6": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.239",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x79\x59\xf9\x94\xcc\x17\x79\xbf\x12\xd2\xbb\x2d\x5e\xce\x9e\x06\x44\xfa\x2a\xdc\xbe\x3c\x85\x86\x89\x66\x3e\x45\x12\xb9\x5e\xca": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x79\x5c\xab\x77\xbb\xa6\x7d\xd7\xd9\x11\xcf\x42\xdb\x8c\x01\x00\xa4\x01\x48\x11\xfb\x0d\x2c\x30\x21\x34\xaf\x07\xa1\x15\x02\x56": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x79\x82\x70\x15\x07\x8d\x71\xf1\x55\x97\xe0\x2b\x6d\x3f\x88\x0f\x74\x04\x39\xd6\x15\x74\xca\x93\x01\x95\xf6\xd8\x07\x0c\xcb\x7d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x79\x8d\xa3\x49\xb1\x12\x68\x73\x90\x8c\xb5\xf7\x7a\x42\xc1\x18\xb0\x58\xe0\x75\x73\xd3\xdc\xa7\x91\x49\xc8\x66\x4d\x9c\xa2\x9c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="ContextUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x79\xbc\x72\xcf\x19\x87\x57\x49\xa8\x8a\xbc\x4c\xca\x26\x01\x1d\xfa\x28\x9a\x38\x11\xb8\x95\x65\x08\x63\x5d\x82\xb3\xd2\x60\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275
    b"\x7a\x23\x3f\x35\xbd\x09\x70\x42\xb7\x8c\xb2\xcf\x71\xed\x56\x75\x89\xec\x49\x40\x67\x63\x0f\xd8\xba\x99\x68\x08\xce\x58\xfa\xf7": [
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Pod.sol",
            name="Pod",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x7a\x5a\x5b\x17\x9d\xec\x42\x54\x6f\x8c\x70\xe9\xc9\x59\x1c\xed\x9c\x83\xa6\xb3\xa1\xa0\x58\xef\x89\xa3\x12\x00\x96\xdb\xb8\x5c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x7a\x62\xc5\xa6\xec\x20\x0c\x65\xc4\xd6\x94\x9f\x03\x63\x62\x53\xc6\x87\x76\x42\x29\x83\x62\xf2\x53\x63\xe1\x56\xaf\x48\x2f\x16": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x7a\x65\x52\x9e\x04\xfd\xf3\x90\xf5\xa5\x01\x7f\xc0\x4c\x1c\xbf\xd2\x9e\x83\x96\x17\x36\xe2\x17\xde\xb7\xe1\x25\xd8\x63\x5a\x19": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/StorageSlot.sol",
            name="StorageSlot",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x7a\x73\x2a\xd6\x6f\xf3\xaf\x93\xf6\xeb\xbd\xf7\x02\x44\xa7\xec\x61\xb8\x90\x1a\x76\x2e\x09\xbe\xc8\xd3\x9e\xb9\x12\x4e\x59\x40": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x7a\x84\xf3\xe7\xad\x6b\x2e\xb4\xbb\xc6\x54\x68\xc3\xdc\x97\x03\x50\xd2\x9f\x7c\xe8\x72\xb6\xa8\x2a\xff\x63\xd7\xee\xef\x9b\xc5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.1
    b"\x7a\x87\x88\xaf\xa3\xa9\xfd\xc1\xfd\xe2\xc3\xac\x0c\x02\xe3\xfb\x94\x56\x6a\x28\xd4\x88\x5e\xa0\x93\x04\x34\xf5\xbc\xb1\x47\x17": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x7a\x90\x27\x2a\x2d\x67\x56\xf4\xbf\x81\x98\xee\x3a\x6d\xde\xbb\x5a\x78\x24\xec\x17\xfe\x31\xbb\x5d\xa3\xf4\xa7\x2d\xcf\xae\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x7a\xa0\x1a\x23\xac\xeb\x07\x45\xf6\x3b\x5d\x6a\x6c\xe1\x37\xed\x9f\x64\x91\x41\xd2\x94\x1f\xc1\xa3\xb1\xd8\x5b\xad\x68\x26\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x7a\xab\xd0\x47\xcd\x4e\xdb\x3a\x25\xd4\x24\xb4\x56\x64\xa8\x6d\x60\x23\x73\x4e\x0b\xd8\x8f\xc1\x42\x48\x6b\x77\x03\x4f\x46\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x7a\xab\xe7\xa7\x40\x2b\x7f\xb3\xc5\x27\x40\x0f\x9c\xb2\x9a\xea\xc6\xfb\xd8\xc8\xd2\x99\xc4\xb8\xf2\xae\x45\x4d\x87\x32\xdb\x52": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x7a\xc2\xca\x19\x2f\x75\x28\xa0\xc8\x2f\x2d\x66\xe1\xd9\x63\x5f\xbc\x15\xc8\xe7\x7a\x4c\x27\x01\xf4\x24\x2e\x1a\x9f\x3c\xe6\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x7a\xea\xeb\x29\x39\xf2\x49\xaf\x7f\xdb\x30\x4d\xdf\xe3\x5a\xea\xda\x56\xa3\x3d\x89\x13\x95\xf8\x5d\xd8\xdd\x7c\x18\x82\xbb\x8b": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x7a\xfa\x35\xb6\x32\x77\xb7\x0e\xf2\x9a\x65\xe8\x79\xd1\x5e\x24\x77\x23\x15\x6f\x77\x83\xae\xcc\xfa\xff\xb0\xf4\x70\xa6\x27\x38": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Comparators.sol",
            name="Comparators",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x7b\x15\x45\x03\x47\xe9\xc3\x92\x21\xc4\x2f\xc7\xc7\xc9\x3c\xe1\x4f\x70\xf1\xae\xf9\x67\x36\x73\xce\xf4\x5f\x34\x4a\x0e\x8d\xf1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7b\x2b\x4f\xb9\xd7\x23\x7b\x37\x62\x9d\x0b\xb2\xda\x36\xf0\x2b\x57\xd8\x59\xf1\x6c\xe8\x7b\xac\x4f\xb8\x38\x28\x24\x85\xfa\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\x7b\x31\x81\x3b\x63\xa7\x37\x75\x08\xf9\x7e\x92\x05\x4f\x43\xd8\x6a\x0b\x94\xc0\x75\xa7\x3b\xf8\xd8\x66\x2b\x1c\x5d\xd3\x77\x0f": [
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7b\x3d\x2f\x3c\x79\x79\xd6\x93\x1d\xa2\x57\x8e\x73\x85\x8a\xcf\x46\x25\x57\xdb\x2c\xc8\x59\xaa\xe3\x08\x37\x99\x86\xca\xc5\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x7b\x56\xdd\xe5\x47\xed\xd6\x7f\xf0\xa6\x12\xaf\xc9\x9d\x49\x2d\x15\x9d\x8b\x42\x64\xe0\x16\xce\x54\x1a\xba\xe6\x0b\x91\x67\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7b\x6c\x8f\x63\xe5\x7a\xb1\x8f\x39\xee\x71\x68\x7b\x02\xf1\xa5\xeb\xf4\xb0\x0f\x6c\x01\xce\xfe\xb1\xe7\xfb\xcd\x30\x42\x48\x10": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x7b\x72\x3e\x90\xc3\x13\xa1\x63\x87\x95\x3a\xc1\xb3\x1a\x71\xb1\x75\xff\x76\xf3\x96\xfe\xd1\x31\xdd\xb9\xd8\x25\xca\x30\x65\x54": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x7b\x81\x89\x79\x79\x77\xe6\xf1\xd9\x09\xa5\xe6\xa4\x2a\xa9\xc5\xd2\x0d\xec\x18\x28\x09\x9b\x08\x78\x3e\x91\xb3\x1a\xd4\xbf\x36": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AccessManager.sol",
            name="AccessManager",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.142
    b"\x7b\x82\x5b\x57\x2f\x4f\xfc\xee\xa0\x21\xfe\x38\xfe\x9f\xaa\x34\x47\xc4\x40\x1d\xb6\xab\x0f\x1b\xc1\x78\x99\x47\xa1\xad\xcc\x4f": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.142",
        ),
    ],
    # Hash appears in versions: v0.0.110, v0.0.111, v0.0.112
    b"\x7b\x84\xa8\x10\x51\x4a\xf9\x92\x95\x9a\x1b\x92\x9d\xe4\xae\xa2\x08\x1a\xfc\x24\x72\xa4\x77\x94\x0a\x83\x4b\x10\x4d\x84\x35\xb7": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.112",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\x7b\x96\xdb\xb7\x42\x48\x1f\x91\x23\x48\x86\xe4\xa1\x56\xf1\x79\x22\x43\xc0\x72\xb7\xf7\x52\x3b\x7f\xb8\xaf\x60\xa8\x36\x4c\x3b": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolTicksCounter.sol",
            name="PoolTicksCounter",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolTicksCounter.sol",
            name="PoolTicksCounter",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolTicksCounter.sol",
            name="PoolTicksCounter",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x7b\xa1\x25\x7e\x02\xea\x57\x0c\x0b\x8f\x4a\x11\x8f\x95\x1d\x14\xe2\x3e\x7e\x68\xe2\xa6\xce\xb4\xc6\x99\x6e\xcc\x32\xa6\xfd\xff": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x7b\xa5\x32\x11\x26\x67\xb5\x80\x2c\xad\x59\x09\x7b\x7c\xfc\x1c\x6e\xde\x56\x57\x4a\x9a\x32\x6b\xba\x43\x8c\x47\xda\x38\xb7\xe9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x7b\xaf\x60\x07\x41\xc3\x8c\x39\x05\x77\x54\x71\x9b\x5f\x15\xcc\x0b\x76\x57\x28\xd5\xe5\xbf\xea\x95\x9d\x48\x87\x9d\x65\xcb\x26": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7b\xb3\xb1\xc2\x10\x67\x41\x4d\xe7\xbe\x74\x1d\x58\xf7\xb7\x0a\x07\x11\x1a\x11\x56\xb6\xb1\xca\x1b\x8c\x54\x76\x51\x6e\x45\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x7b\xc5\xe9\x8e\x60\xb6\x23\xbc\x15\x52\x26\x4b\x1e\x7e\x77\xe0\xcf\x7e\xc7\xaf\x09\x6a\xe0\x15\x77\x7b\x1f\xd9\xf7\xc3\xe6\xb4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x7b\xd6\x1b\x7e\xf2\x3b\x32\x31\xad\x5e\x63\x34\xfe\xba\x8c\x46\xd4\xf1\x40\xe3\x65\x9d\x33\x67\x16\xb9\x5e\x06\xf2\xb5\x56\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSetUpgradeable.sol",
            name="EnumerableSetUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x7b\xf1\xae\x94\xc6\x29\xae\x7e\x4a\xfb\xe8\x3a\xe0\x84\xad\x3e\xbb\xb9\xb1\x1e\xe2\xeb\x68\xf7\xab\x83\x6e\xa5\x45\x6c\x40\x27": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2
    b"\x7b\xf4\xbc\x9d\x10\x18\x25\xa5\xab\x1b\x0d\x3b\xea\x9e\xf1\x73\xb0\x27\x8b\xe8\x61\xc5\xd7\x64\x87\x2d\xa1\x6d\x6c\x97\x91\x0b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x7c\x09\x2f\xa9\xae\x18\xc3\x12\xe6\x94\x70\xe7\xaa\xab\x97\x72\x61\x7b\x4e\xfc\xe0\x60\xc5\x3b\x9d\x97\x54\xd8\xc9\x3b\xf0\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x7c\x1f\x83\x37\xe1\x49\x1c\xa1\x8f\x8e\x9a\xeb\x59\xd8\x55\x10\xc8\x03\x9e\x5f\xe8\x19\x26\x3f\x41\x4f\xce\xef\x75\x06\x7b\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167
    b"\x7c\x20\x4f\xb4\x2d\xed\x27\x74\x63\xdc\x22\x2e\x0b\x7a\x03\x13\xcf\x30\x9e\x92\x5f\xcb\x6c\xf2\xd1\x74\x90\xca\xf1\x06\x4b\x4d": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.167",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x7c\x2a\x7d\xe4\x2a\x9c\x0c\x02\x1c\xce\xa6\xe7\x82\x91\x38\x21\x6a\x49\xf2\xe7\x2e\x49\x53\x16\xa0\x0c\x39\x0d\x12\x1b\x3b\xa9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.187, v0.0.188
    b"\x7c\x2b\x3e\x2f\x87\xd0\x23\x2c\x5e\x1e\x3f\x18\xaa\x07\x0f\x18\x88\xb4\x3c\xef\x48\x05\x38\x9e\x9b\xe1\x5b\xd1\xa6\x98\x98\x5c": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.188",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\x7c\x81\x83\xac\x2c\xb8\xdc\xc1\x09\x6a\x1a\x22\xb7\x3e\xd2\xe0\xaf\xbe\x4d\x76\xd2\x4d\x3a\x30\xf2\xf5\x89\xa6\x25\xd0\xdb\xca": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x7c\xa0\x50\x0f\xd5\x24\x45\xc4\xb1\x9f\x17\xaa\xfa\x5e\x8c\xa3\x16\x3c\x5a\xff\x3f\xe0\x00\xb9\xcb\xaf\xed\x33\x08\xe7\xa6\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.1, v4.8.2, v4.8.3
    b"\x7c\xc4\x25\xe3\x8d\x81\x6c\x4e\xcc\x69\xc2\x3f\xa1\x1e\x76\x64\xcd\x26\x9c\x78\xa0\x63\x09\xde\x7b\x5e\x97\x26\xc5\x6e\x06\x2d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.269, v0.0.270, v0.0.271, v0.0.272
    b"\x7c\xd9\x40\x3c\x86\xbd\xde\xbb\xc4\x6a\x31\x02\x72\xfd\xa9\x01\x4f\x9f\x35\x47\xb3\x1f\xda\x53\x17\xcb\x85\xe5\x5f\x69\x1b\x9f": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.272",
        ),
    ],
    # Hash appears in versions: v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x7c\xdc\x84\xa8\x13\x06\x41\x36\xd1\xdf\xa7\x4f\xec\x56\xb6\x83\x6f\xa5\x9c\x63\x22\x2f\x08\xc7\xfa\xf9\xea\xe6\x2f\x03\xdc\xb4": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20Votes.sol",
            name="ERC20Votes",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x7d\x40\x5c\x77\x6b\x56\xb5\x1b\x47\x66\x43\x3a\x03\xec\xaf\x6d\x0e\x12\xda\xb9\xed\x75\xd4\x1a\x8d\x42\xd3\xb1\xa0\xb2\x7e\x2a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x7d\x51\x07\xcb\x66\xe7\xbc\x78\xdd\xa4\x45\xf5\x45\xa5\x76\x06\xc9\xf2\xb7\xef\x53\x04\xf6\x1f\xfa\xec\x79\x32\x18\x83\x8d\x99": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.33
    b"\x7d\x67\x55\x62\x6e\x2a\x8d\xbd\x62\x63\xfb\x69\xe7\x69\x4c\xc0\x23\xbe\x42\x14\x3f\x39\x71\x26\x43\x70\x4e\x92\x95\xab\x0a\x32": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.33",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x7d\x69\x57\xe7\x48\xbc\x98\x91\x40\x56\xc2\x06\x13\xd7\x66\x0f\x9f\x47\x04\x26\x2b\x36\x4f\xe5\xd0\x65\xb9\x0a\x1e\xa7\xa5\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMath.sol",
            name="SafeMath",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7d\x96\x5f\xcb\x78\x0a\x55\x9c\xfc\x43\xe6\xda\x59\x81\x10\x2e\x58\xe0\xdd\x31\x2e\xe8\x42\x2c\x7e\x5b\xf4\x85\x41\x65\x03\x61": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x7d\xb3\x4e\x61\xa8\x66\xeb\x3b\x58\xff\xa0\xc1\x97\xdb\xb1\xd9\xdc\xf2\x61\x57\x09\x63\x22\x5d\xaa\xd2\x75\x43\x9a\xe7\xf2\x54": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x7d\xb8\xaf\x9c\x92\xf3\x74\xe5\x0a\x92\xef\x1e\x4c\xa9\x45\xa0\xdc\x14\x7e\x17\x7e\x87\x41\x47\x28\x02\xf3\x3e\x71\x36\xb4\xc8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x7d\xba\x02\x9b\xa3\x8b\x21\xbd\x62\xf6\x0b\x05\xa6\x16\xab\xb6\xf9\xc0\x98\x1d\xda\xff\xca\xd1\x4c\x47\x7e\x2a\x62\x5b\x76\x9a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184
    b"\x7e\x09\xf3\x1d\x30\xca\x6e\xa0\xab\xf4\xfa\xa0\x8a\xb0\xe5\xcb\xe6\xe4\xd4\x3b\x9a\x3f\x0c\x06\x55\x20\xee\xce\x9e\x67\xd5\xf5": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.184",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x7e\x23\xc3\x6e\xdf\xca\xae\x27\x2d\x8d\x99\x8a\x5c\xbd\x80\x1c\x78\x32\xe8\x80\x29\x19\x25\x9b\x95\xd3\x7f\x07\x64\xc0\xcc\x18": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/polygon/IFxMessageProcessorUpgradeable.sol",
            name="IFxMessageProcessorUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x7e\x35\xbf\x67\x1b\x62\xbb\x02\x1c\x0e\x68\xee\x1e\xaa\x0c\x84\xe8\x4b\x87\x8b\xd8\x0c\x18\x44\x08\x5d\xb9\xcf\xf2\xbf\xa6\xbc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x7e\x49\x07\xdb\xe1\xd8\x53\x6c\x94\x35\x9c\x4a\x42\x6b\x14\x0c\x5c\x41\x30\x34\xa6\x9f\x14\x9b\x75\x04\x25\x54\x07\xb8\x0b\xe4": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/IExttload.sol",
            name="IExttload",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x7e\x8f\x61\xd0\x55\xd9\xdc\x6a\xa5\xea\x5d\x16\x01\x2c\x79\xf2\x95\xc6\xfb\x1e\x1b\xe9\x25\x62\x28\xf4\x9b\x29\xf7\xe5\xb8\xa4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x7e\x97\x37\xf9\xb0\xba\xc7\x49\xd3\x0c\xc2\x7b\x37\xee\xb4\x3b\x2b\xfb\xb2\xdf\x3d\x97\xba\xce\xd2\x6f\xcf\x3a\x3e\xaa\x3c\x6e": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/SwapRouter.sol",
            name="SwapRouter",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/SwapRouter.sol",
            name="SwapRouter",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/SwapRouter.sol",
            name="SwapRouter",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/SwapRouter.sol",
            name="SwapRouter",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/SwapRouter.sol",
            name="SwapRouter",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.229
    b"\x7e\xa3\x72\x10\x2f\x18\x00\x81\xcb\x17\x43\xbd\x09\x5d\x9e\x7f\x9f\x84\x84\xac\x3d\x9e\x6a\x9f\x8c\x8d\x08\xe8\x00\xa5\x29\xe6": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.229",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x7e\xa8\x12\xea\xc6\xdb\x78\xce\x63\xe5\xc3\x2d\x5e\xb9\xc6\x11\xcf\x3d\x58\xac\x94\x3e\xbf\x79\xf7\x1f\x10\xe8\x55\x5c\xfd\x06": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0
    b"\x7e\xb0\xde\xe0\x9e\x99\x85\x9c\xf3\x05\x5f\xf6\x62\x18\xe4\x05\xda\xeb\x23\x08\x21\x50\x69\x21\xd4\x8b\x70\xb2\xfb\x1e\xeb\x24": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2
    b"\x7f\x07\x71\x3a\x07\x84\xaf\x26\xb7\xb9\xb3\x7a\xfa\xbe\x34\x18\xf8\x07\x56\x17\xda\x37\x60\x13\x8f\x66\xf5\x0c\xa2\x1b\x68\xaa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v5.0.0-rc.2",
        ),
    ],
    # Hash appears in versions: v1.3.0
    b"\x7f\x1f\xd4\x52\x07\x62\x7d\x21\xd8\x57\x70\x61\xff\x12\x63\x94\x4e\x14\x19\x90\xdd\x02\x48\xf0\xbf\x60\x03\xe5\xa0\x92\x9a\x75": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/OracleLibrary.sol",
            name="OracleLibrary",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x7f\x23\xb5\x95\x8c\xe6\xbd\x78\x33\xd6\x48\x8f\xe8\x42\xc8\x01\xef\xf6\x02\xf5\xd8\xce\x7a\x2b\x36\x5e\x63\xbd\xfa\xec\x68\xd2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2612Upgradeable.sol",
            name="IERC2612Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x7f\x33\xd5\xcd\xf5\xd4\x18\xc2\xfd\x95\xd6\x77\x6f\x77\xeb\x47\xda\x05\x3b\x0f\xa7\x1c\x88\xfa\xc2\x87\x08\xc6\x64\x14\xa7\xf3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2
    b"\x7f\x6c\xce\x09\x2d\x0f\xda\xaa\x7c\xc2\x5c\xec\xc7\x64\x29\x17\x9a\xf3\xe0\x8b\x44\x08\x13\xd9\xb6\x8c\xd6\xae\x17\xb4\x09\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.8.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118
    b"\x7f\x72\x02\x5c\xb3\xa7\xf8\x70\x38\x4f\xf5\x2b\x03\x40\x95\xa5\xae\x16\x64\x86\x6a\x1a\x68\xfa\xb8\x6f\x16\x32\x80\xbf\x94\x8f": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.118",
        ),
    ],
    # Hash appears in versions: v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\x7f\x81\x0f\x3c\xdf\xfb\x7b\xae\xe0\x4e\xb6\x79\x3e\x70\xd3\x61\x7d\xbd\xcf\xd0\x25\xb7\xd1\x07\xf8\xae\xf9\xd2\x26\xe2\x57\xb7": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibTransient.sol",
            name="LibTransient",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.7.0-rc.0
    b"\x7f\x81\x8b\xbc\x13\x32\xc1\x2e\xef\x06\x84\x12\xe8\x61\x54\x70\x0f\x93\xf3\x82\x7b\x78\x89\xac\x8d\xb9\xb5\xe0\x39\xc5\xf4\x74": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20TokenizedVault.sol",
            name="ERC20TokenizedVault",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.178
    b"\x7f\x93\xfc\x8f\xab\x02\xfa\x7b\xca\xbb\xea\xab\x9e\x7e\x2d\x61\x37\x92\x1b\x83\x76\x20\x2d\x0a\x6e\x9a\x43\xab\xee\xd2\xfd\x92": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.178",
        ),
    ],
    # Hash appears in versions: main
    b"\x7f\xa1\xab\x19\x31\xa1\x4c\x51\x07\x13\xcd\x96\xfc\x49\x80\x28\x87\xcf\xf0\xbc\x9a\x49\x48\xa5\x80\xef\xd0\x87\x97\x81\x9c\xaa": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/SVG.sol",
            name="SVG",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x7f\xe2\xc8\xcc\xdb\x0d\x29\x00\xa5\x7e\x36\xeb\xd5\x8a\x91\x90\x21\x6f\xe6\xbc\x1c\xcf\x81\xbb\x82\xa4\x0c\x4f\xd2\xed\x82\x58": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.9.6, v5.0.2
    b"\x80\x06\x2c\x25\x4e\xbf\x62\x7d\x17\xf5\x2e\x8a\x86\x83\x43\xbf\x19\x41\x9e\xe3\x46\xa0\x0a\xcb\x32\x50\x57\x6d\x50\x3a\x58\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x80\x07\x7b\x34\x04\x18\xd7\x68\x84\x8f\x66\x0c\xff\x04\xe8\x70\xeb\x95\x60\x30\x37\x01\xc5\x09\xe7\xfb\xaa\xdb\x54\x6b\x7e\x3e": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x80\x0c\xfe\xf8\x32\x79\xcb\x64\x50\xf2\xd4\xc0\x49\xca\x54\xa3\x88\xa4\x54\x37\xe8\xff\x7d\xe2\x67\xf7\xef\xf8\x49\x59\x24\x66": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x80\x63\xc4\x35\x45\xd7\x79\xba\x88\x76\xb8\x8a\xfa\x43\x02\xa9\x59\xb2\xb7\xe9\xdf\x68\xbe\x22\xb3\x39\xb1\xec\xed\x92\xa6\xfa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x80\x88\xf7\x36\x8b\x04\x1e\xc9\xa6\x55\x1c\x8f\xbf\x00\xda\xdb\x42\xdc\x9f\xb9\xa2\x78\xa7\x6f\xd6\xd9\xc0\xe7\x1f\xea\xd0\x5b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.37
    b"\x80\x8c\x19\x1c\xc1\xea\x47\x0d\x0d\x6e\x66\x0e\x93\x4e\x4b\x67\xae\xca\x41\xbe\x17\x71\xe8\x27\x3b\x6c\x7d\x13\xbc\x95\x07\x6a": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.37",
        ),
    ],
    # Hash appears in versions: v0.0.87, v0.0.88
    b"\x80\xc7\xc1\xca\x4a\xd8\x95\xf3\xdd\xd8\xb9\xd9\x24\x1c\x49\x59\xe7\x18\x03\xe8\x2b\x21\xab\xb8\x68\x11\x8f\x9f\x81\x9b\x38\xfa": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.88",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x80\xc9\x5f\x0a\xab\xdd\xf4\xe9\x94\xbe\x93\xee\x97\xed\xf7\xf9\x1d\xf1\xea\xfa\x2f\x84\xa0\xb2\x05\x5a\x38\x76\xf9\x08\xaf\x6e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/NoncesUpgradeable.sol",
            name="NoncesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x80\xf4\xcb\xfb\x43\x3c\xbd\x2d\xd7\x70\x70\x00\x2d\x56\xe1\x2e\xa5\xd5\x6d\xb6\xe5\x22\xa4\x72\x96\xed\x78\x7b\x19\x3b\x43\xa8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x80\xf6\x65\x02\xd2\xe6\x77\x72\xad\x7e\xce\x5f\x7c\x9d\x2e\x98\xd6\x0c\x09\x93\x3f\xc0\x96\xa8\x22\xad\x06\x21\xc0\x81\x8e\x8b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x81\x07\x03\xf4\x91\x14\xe5\x36\xa1\x9f\x51\xec\x5a\xf1\x29\xac\x34\xed\x45\x37\xab\x88\xa1\xdc\xd8\xd5\xce\x69\xd2\x07\xdc\xf1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x81\x29\xda\xf5\x55\xc0\xa8\xca\x66\x48\xdb\x7b\xe1\xf4\xb7\x05\x81\xd1\x5d\x75\xcc\x2f\xc5\xc8\xe6\xb5\xd5\x15\x46\x8b\x13\x81": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x81\x2a\xf2\x6f\x55\x84\x57\xad\x20\x86\x0c\xc7\xf8\xd8\x35\xfe\x3a\x19\x37\x9d\x2e\x80\x80\xb6\xe8\x4b\xfd\xd2\xaa\x23\xc7\xe2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x81\x51\x44\xba\xb6\xe1\x75\xd3\x7f\x1a\xb0\x5a\x1b\x88\x97\x31\x32\x27\x4b\x73\xa6\x66\x13\x18\xd7\x33\xc9\x14\xe8\x58\x83\x49": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x81\x6a\x81\x08\xb3\x78\x63\xce\x58\x83\x4d\xbb\xfe\x95\x14\x37\xd0\xa9\x0b\x9e\x30\x1f\x41\x6d\xc0\x8f\xb6\xca\x2f\xf5\xad\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Full.sol",
            name="ERC721Full",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Full.sol",
            name="ERC721Full",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Full.sol",
            name="ERC721Full",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Full.sol",
            name="ERC721Full",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x81\x8e\x1e\x64\xb2\x8d\x2c\x0b\xb4\x81\x65\x3d\xde\xfa\xcd\x06\xa4\x45\x1b\xe5\x08\x5e\x0e\x01\xdc\x6a\x6a\xbc\x5a\x19\xf5\x65": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x81\xc4\x5f\xff\x7b\xfb\xa8\xc1\x07\x62\x54\xe0\xb3\xb0\xbf\x3c\x96\x9c\x0e\xe5\xc7\x39\x5b\xa3\x76\x2e\x8c\x06\xc6\xad\xff\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x82\x8c\x1c\x24\x1e\x53\x91\xa9\x8f\x54\x37\x6d\xd8\xbc\x83\x27\xa2\x68\xd2\x8f\x50\xbb\x96\x7c\x66\x17\x86\x5a\x2a\x78\xd1\x1f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.229
    b"\x82\x92\xf3\x5a\x89\x74\xfc\xda\x84\x51\xc6\x23\x84\xcd\x9f\x03\xb3\x2c\x96\x6e\x95\xa3\x62\x66\x15\x23\xe0\x92\x60\x31\xef\x9f": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.229",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x82\x98\x2c\x6a\x60\x41\xf2\xbf\xa2\xd6\x07\x06\xa3\xb1\x57\xf4\x00\x67\x49\x21\xb8\xbd\x16\x94\xa0\x55\xd6\xdc\x3a\x46\xea\x1a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHub.sol",
            name="IRelayHub",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.1.11, v0.1.12
    b"\x82\xb2\xe0\xec\xcd\x07\xb7\xfb\x60\xfe\xd8\x3e\xb0\x42\xad\x79\x91\x82\x7c\xd7\x41\xff\xcb\x1d\x2f\x79\xf3\xeb\xc1\x54\x49\xa0": [
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.12",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x82\xc9\x46\x58\x86\x05\x84\xaa\x79\x5a\x6a\xb8\xe3\xb4\x28\x2b\xc6\xc3\x6f\xed\xcb\xf8\x79\x20\x47\x44\x76\xb3\xbb\x7a\xa4\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2612.sol",
            name="IERC2612",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/draft-IERC2612.sol",
            name="IERC2612",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.1, v5.0.2
    b"\x82\xd0\x3c\x5a\x1b\xb7\xb6\x91\x22\xe2\x0d\xcb\xb5\x75\xae\x13\x74\x23\x02\xaa\x85\xba\xca\x6d\xb9\x77\x73\xc0\x72\xb8\xe9\xd2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x83\x14\xde\x6f\x60\xf2\xb8\xf5\x97\xc4\x54\x68\x0d\xaa\xbf\x32\x81\x0e\x0b\x57\x54\x95\x36\xb2\xf7\x51\x4f\xb9\xba\x64\x82\xfd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\x83\x35\xc5\xfa\x2b\xf2\x60\x54\x28\xdb\xa4\xc4\xa7\x84\x38\x08\x11\xea\xe8\x4d\x12\x85\x90\x5d\xc8\x31\x16\x1c\x5b\xfa\xec\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x83\x3e\xc6\x62\xb1\x28\x05\xe7\xb9\x2b\xe5\x7d\x68\x81\x1e\xe6\x9f\x70\x76\xd2\x09\xdc\x66\x6b\xc9\x65\xec\x5f\x60\x1f\x60\x5d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x83\x4e\xb8\x1e\xba\x24\x4f\x89\xa8\x7e\xe1\x3d\x5f\x5e\xd1\x0e\xe8\xc9\x99\x0f\x1e\x8c\xd4\xc2\xcc\xbe\x6f\x6e\x69\x70\x77\x58": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x83\x61\xa1\x22\x5d\x6f\xab\xab\xab\xeb\x7a\x8f\x80\xa1\xac\x32\xeb\x74\x6c\xda\x77\x66\xe4\x28\x6d\xf2\xc7\x41\xa9\x41\x3c\x4f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x83\x6c\xae\x9e\xc4\xc7\x5d\x7e\xe2\x29\xa7\xd7\x1e\x99\x1b\x52\xcc\x8d\x3e\x1f\x8c\x23\xff\x4d\xa0\x9e\x36\x38\xa9\x97\xf5\x89": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/LowGasSafeMath.sol",
            name="LowGasSafeMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x83\x6d\x19\x65\x58\xd3\x28\x2e\x3b\x7a\x8e\x7a\x14\x75\x84\x7c\xf3\x61\x7c\x9c\x18\xab\x66\xd3\xda\xca\xfe\xd4\x1a\xa1\x12\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IDelayedMessageProviderUpgradeable.sol",
            name="IDelayedMessageProviderUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.115
    b"\x83\x94\x0f\x29\x1d\x25\x2c\x60\x83\xf5\xa6\x9c\x52\x66\x37\xe9\x4f\x7c\x94\xbe\x90\x1d\x84\x78\x0a\x08\x6a\xf7\xa5\x00\x02\xb9": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.115",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x83\x9a\xeb\x93\x91\xe9\x90\xb4\xf2\x2f\x3a\xfe\xef\xbc\x1e\xa8\x10\x33\x00\x95\x74\xe8\x07\xf8\x3a\x8d\x5f\xe4\xed\xff\xd7\xe4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.92
    b"\x83\xf0\x1a\x2b\xf7\x79\x32\xf9\xdf\xb1\x6a\x51\xda\x20\x25\x41\x59\x7c\xfc\x24\x1c\x81\xe2\xec\xb1\x58\xfe\x31\x1a\xb4\xe9\x1b": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.92",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x84\x16\x4c\xc0\xc5\x32\x55\x8f\x0a\xef\xd6\x33\x48\xa1\xf6\x7f\x68\x2c\xdd\xd5\x15\x5a\x9e\xb8\x8b\x30\x0a\x57\xef\x46\xfe\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0
    b"\x84\x35\x8b\x1a\xc2\x91\x1f\x36\x77\x27\x95\xb3\x83\x7a\x9b\xc2\x32\xb8\x6e\xfe\x97\x36\xf4\x94\x9a\x59\xe2\x0c\xd7\x76\xff\x7f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x84\xa4\xff\x45\xd1\x37\xf8\x09\x42\xd6\xa5\xa3\xca\xee\x54\x38\xb6\x46\x14\xe2\x08\x90\x25\x1d\xfa\xd6\x14\x45\xdd\xf0\x98\x45": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x84\xa6\x26\x47\x0b\x00\x08\x03\x1b\xad\xc5\x72\xf4\x2b\x50\x22\xdc\xc9\x74\x68\xea\xd9\xfa\xa7\x18\x66\x9f\x04\x80\xfc\x69\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/IERC1155MetadataURI.sol",
            name="IERC1155MetadataURI",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x84\xb0\x46\x37\x7c\x45\x82\x04\x23\x3e\xdf\xe4\x1d\x34\xd5\xc9\x77\x1b\x83\x90\x10\xe0\x96\xd4\x5f\x07\x52\x04\x8d\x9b\x30\xb0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x84\xe3\x96\x1d\xbd\x01\x86\x24\xd7\x17\x9e\xfb\x44\x9e\xf8\x43\xa6\xcb\x04\x38\x89\x43\x89\xf7\x96\x5f\xc3\x01\xd2\x38\xf3\xf6": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x84\xf2\x04\xca\xf4\x44\xd2\xa6\x75\x61\xae\x87\x51\x0a\xc0\x99\x82\xd3\x41\xa0\x5d\xd6\xf2\x4c\xb1\x10\x5d\x09\x6e\x39\xb3\xf5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\x84\xf6\x98\x2c\xc2\x61\xba\x21\x44\x02\x30\xdc\x18\x6c\x2c\x13\x4c\x6e\x31\x48\x3e\x86\x71\xe3\xa2\xf2\x72\x6f\xe8\x64\xba\x59": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Packing.sol",
            name="Packing",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x85\x1b\x91\xd9\x17\xfa\x84\xac\x9d\xa2\x1b\x87\x42\xa8\x4d\x46\x72\x6b\x99\xf9\x03\x3d\x4c\x60\xb9\x8e\x4f\x0e\x00\xa1\x40\x8b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x85\x22\xf9\x21\xa0\x57\x41\x19\xcd\x16\x7c\xac\x04\x20\xa7\xfc\x51\x74\x36\x43\xd5\x80\xca\x9e\x64\x5f\x05\x15\x83\xd6\xa6\x08": [
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x85\x44\xf1\x12\x04\x03\xc5\x4e\x63\x72\xa6\x9b\xd9\xd2\xcc\xb5\xc5\x8a\xc6\x82\xa9\x5a\x70\xb7\xab\xa7\xa5\x4c\xff\xa5\x94\x1e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x85\x49\xdf\x6d\x5c\xa4\xb5\x9a\x99\x5f\xb2\xf7\xf9\x57\x55\xf9\x23\xf9\x0f\x26\xad\x5d\xa3\x53\x2d\x1d\x7f\x41\xc6\x5e\xf2\xee": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x85\x56\x5a\xac\x46\xb9\x37\xe3\x5f\xc2\x2c\xe1\xf8\x99\x91\x84\xa0\xb7\x71\x50\xcd\xbc\x30\x69\x27\x8f\xb1\x35\x75\xcc\x0f\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x85\x5d\x50\x12\xb5\x93\x45\x9b\x58\xfb\x32\x25\xca\xdf\x2f\x97\x91\x75\x1e\xeb\xc2\x34\x2f\xd0\x7b\x9a\x63\x38\xe3\x57\xee\x5c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67
    b"\x85\x89\x81\x09\xaa\x61\x5c\x10\xbf\x89\x12\x03\x0d\xf9\x19\x5d\xf0\xa5\xc4\x55\x25\x9c\x8d\xcc\xcd\x05\xdd\x88\x52\xf0\x43\x3f": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.67",
        ),
    ],
    # Hash appears in versions: v0.0.196, v0.0.197
    b"\x85\x91\x57\xa8\x32\xe0\x05\x8a\x61\x68\xb3\xd5\xa3\x87\xee\x89\x50\x08\x62\xcc\xa9\x4f\x18\x61\xdb\xa1\xa8\xae\x54\xff\x7a\x1d": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.197",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\x85\xa2\x5e\xe9\x45\x5a\x9a\x05\x11\x47\x89\xe1\xd3\x6d\x34\x62\x0e\x11\x37\x2f\x54\xb0\xcf\xb9\x35\xf9\x33\x5a\x41\x06\xe7\xcd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x85\xa4\x77\xae\x2c\x7b\x8e\x40\xdd\xf2\xac\x31\xb6\x5d\x8d\x13\x25\xb4\x93\xf2\x4b\xbc\xdd\x34\x17\xf1\x0e\xad\x60\x9f\x10\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x85\xb4\xdc\x6a\xb0\x36\xf7\xfd\x8e\xf4\x3c\x75\xbd\xfb\xb3\x84\xa9\xcd\xec\xf5\xaf\x50\x51\xaf\x55\x96\x79\xd6\x97\x51\x1e\xaa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/Initializable.sol",
            name="Initializable",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x86\x10\x89\x52\x5f\x60\xa5\xaa\x4f\xc1\x3e\x39\x33\x21\xc6\xed\xfb\x9b\x69\xfa\xbc\xcc\x7f\x8c\xbf\x22\xf2\x77\x4a\xd7\x14\xa3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.5, v0.0.6, v0.0.7
    b"\x86\x14\x2f\x1d\xfc\x1b\x2e\x0e\x78\xaa\xe5\xd0\x55\x29\xf3\xc5\x74\x0b\x97\x92\x9b\xac\x06\x53\xb7\x9b\xd5\x2b\x10\xef\x05\xdc": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.7",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x86\x2c\xeb\x78\x1d\x55\x9d\x85\x75\xef\xf2\x17\x7c\x52\x21\x7c\x63\x32\x7a\xd9\x72\x9c\x61\xe6\x67\x3e\x6a\x59\xd4\x1e\x2c\x09": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x86\x3b\x8f\xa2\x09\xa1\x7e\xb1\x65\x30\x52\xa6\x6f\x4d\xa4\x22\x2d\x8f\x2a\x07\x9a\xf2\x7b\x09\x0e\xd5\x55\x0f\xb8\x12\x7d\x82": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x86\x56\x2f\x45\x8f\x97\x64\xf7\x88\x24\xcb\x9d\xbc\xe3\x11\x48\xe6\xf6\x91\xb7\x0e\xf6\x1f\xcb\xe3\xc6\x3d\x3d\xef\xb6\x9e\xa7": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1
    b"\x86\x6b\xfe\x7b\xc0\x7e\x29\xef\xe8\x19\x44\xbe\xf7\xac\xf4\xc1\x78\xee\x73\x30\xe0\xe3\x40\x10\x0e\x9c\x31\x89\x28\x7d\xfa\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.7.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x86\x7c\x97\x6f\xd4\x45\x3b\x0c\x4c\xac\x69\x0c\x3b\x60\x2e\x98\x7d\x6b\xc6\xc0\x18\xd6\x1b\xd7\x26\xc9\xcd\x8d\x30\x07\xea\x11": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.5, v0.0.6, v0.0.7
    b"\x87\x4e\x53\xd1\x3e\xe5\x40\x76\xf6\x79\x1a\x98\xf5\x0d\xd1\xb2\x10\xed\xd8\x98\xd3\x16\x59\x09\xa0\xd3\xea\x0f\x67\x0e\x34\xa1": [
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.7",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x87\x93\x89\x8b\x6b\x7b\x99\x59\x93\x43\x1c\xb2\x0e\x05\x1b\x2b\x89\x9f\xcc\x4d\xb7\x38\x5c\xfc\xf4\xf7\x1b\x59\x59\x3a\xac\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x87\xe2\xc2\x09\x41\xd8\x62\x3e\xa1\x7a\x78\x0c\xa0\x5f\xd7\xe6\x32\x05\x60\x93\x60\x07\xd2\x49\xcf\x32\x35\xd0\x43\xc3\x3e\x57": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57
    b"\x88\x1d\x64\xc3\xc6\xe4\x23\x40\xb9\x60\x1c\x6b\x5f\xcf\x7d\xc5\x94\x9a\xdc\x19\xdf\x9b\xe1\x15\x02\xc3\xe3\x9c\xb9\x9f\x64\x2c": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.57",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x88\x30\xf2\xda\x56\xfa\x69\x38\xa1\x2d\xd0\x35\xec\x2d\x57\x05\xbb\x5b\x37\xee\xa3\xe3\x0b\x85\xdb\x16\x4d\xba\x56\xa0\x98\xd2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x88\x36\xec\xf8\x4e\x79\x6a\xd4\x80\x9a\x5f\x70\x36\x78\xac\xc8\x1a\x75\x1d\xb3\x5c\xc9\x91\x83\x17\x58\xcd\xf0\x23\xc3\x49\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x88\x47\x1b\xe1\x0d\xe1\x25\xdc\xe2\x73\x67\xc3\x9a\x5e\xc6\x4d\x42\x19\xcd\xa6\xa0\x3d\x42\xc3\x50\x6d\xd1\x5c\x14\x5d\x29\x2f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\x88\x48\x8a\x42\xb9\x79\xb9\xc5\x8d\x02\xf1\xf7\x00\xb0\xb5\xf1\xaf\x74\x80\x1a\x46\x7f\x6c\xb9\x35\xa2\xe8\xa7\x99\x34\xd4\xe5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x88\x53\x0a\x08\x23\x32\x5f\x38\x7f\xf1\x32\x0e\x42\x34\xee\x67\x62\x32\xc4\xad\x02\xac\xf9\x47\xfe\xd0\xeb\xd3\xf1\x74\xcc\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSettings.sol",
            name="GovernorSettings",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x88\x53\x8a\x39\xb1\x0c\x23\x89\xfe\x4d\xe3\x09\xf6\x2e\x34\x3a\x1e\x8f\x2e\x87\x5e\xf4\xfb\xf7\x63\xcc\x34\xe0\x67\x13\x1a\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x88\x5e\x1a\x7e\x95\xa6\xb6\xa3\xd4\xa9\x44\x42\xbe\xbf\x99\xbb\x6c\x89\x7d\x0d\xd3\x08\xbc\x33\x55\x15\x89\x6e\x8a\xd5\xf1\x81": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UUPS/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x88\x87\xb5\x16\xad\x87\xfa\x5a\x11\x5f\xdf\x5e\xa6\x1c\xbc\x42\xf2\x75\xb4\xa3\x93\x29\xe3\xf1\x5e\xd4\x86\x65\xaf\xf7\xc5\x04": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/EIP712.sol",
            name="EIP712",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x88\xa3\xc6\x4a\xe7\x1f\x1b\x4c\xb5\xb3\x68\x25\xa7\x9b\xf0\xd7\x0b\x80\x35\x4d\x50\xc0\x20\xec\x36\xc1\x9d\x59\xc2\x6c\xb9\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x88\xb9\xd3\x0a\x87\xcc\xfc\x66\x3c\xe4\x44\x60\x19\x62\x5e\x07\xf9\xb0\x49\xad\x79\xc7\x83\x46\xc5\x7c\x59\x27\x50\xf3\x6c\x5f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721HolderUpgradeable.sol",
            name="ERC721HolderUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0, v5.3.0, v5.3.0-rc.0
    b"\x88\xf2\x44\xad\x2a\x91\x25\x62\xbf\xde\xd8\x8a\x7e\x01\x0c\x59\x5b\xf2\xb4\x67\x09\xb1\x7a\x2b\x06\x70\x2d\x53\x45\xcf\xc7\x0e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x89\x25\x35\x84\xc9\xad\x64\xce\x37\x16\x9b\x3c\xf5\x65\x86\x67\x7f\xd6\xf8\x2d\x36\xa3\xdf\x60\x3c\xd7\xf4\x62\x14\x6f\x6a\x9e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x89\x39\xc0\x37\xe4\x09\x38\xea\x03\xff\xcc\x67\xcd\xb7\x32\xd2\xda\x11\x7d\x73\xd5\x1c\x08\x2f\xbd\xb2\xb4\x4e\x47\x28\xd1\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x89\x3b\xc9\x04\x68\x95\x82\xb3\x62\x4c\xa9\xb2\x84\x85\xea\x3a\xd0\xe1\xb7\x8e\x1d\xdb\x05\x4a\x34\x7d\x01\xf9\x08\x3d\x70\xf8": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x89\x5e\x43\xaa\xc1\x7e\x16\x02\x53\x5e\x98\x92\xc3\xfa\x02\x3c\x9b\x3e\x7e\x88\x1d\x8d\xa3\x96\x21\xbd\x87\x1d\x43\xd2\x46\xbd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.13, v0.0.15
    b"\x89\xb2\x00\x37\x40\x87\x31\xf2\x26\x4c\xa8\xfd\x9f\x4c\x56\x6c\x6d\xba\x69\x26\xe1\x03\x0b\x02\x51\x13\x5e\x29\x4a\x6e\xdc\x87": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.15",
        ),
    ],
    # Hash appears in versions: v0.0.53, v0.0.54, v0.0.55
    b"\x89\xe4\x4a\xc2\x28\x47\xd6\xdb\x93\x03\xa9\x88\x61\x11\x07\xaa\x8f\x69\x33\xf5\x7b\x4d\x9f\xd0\xb2\xfe\x0a\xc4\x3a\xf9\x3e\x10": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x8a\x90\x0f\x22\xe8\xff\x48\xa3\xd2\x44\x2b\x62\x13\x30\xfa\x39\xcc\x3c\x34\x20\x99\x14\x3e\x63\x74\x15\xaa\xf2\x8b\x1e\x13\x39": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x8a\xbb\xc1\xc2\x97\x2b\x8f\x2c\x82\x2e\x86\x70\x99\xb6\x0c\x3d\x90\xe6\x70\x64\x98\xd9\x92\x7b\x19\x11\x42\x80\x05\xa1\xc1\x69": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/UnsafeMath.sol",
            name="UnsafeMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x8a\xc1\xf7\x74\x67\x12\xb2\xc2\x8e\xc6\x4b\x08\x9f\x1f\xe7\x2a\x33\xb3\x93\x87\xf5\x83\xb0\x5b\xc0\x32\x28\x93\xdb\xdb\x63\x96": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x8a\xd3\xf6\x42\x1f\x14\x33\x9a\x04\xe2\xd0\xdb\xdc\xf7\x97\x7d\xe2\x97\xc4\x12\x28\xb5\x5f\xd5\x54\x6d\x73\x63\xc3\xa7\x75\x5d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x8a\xe8\x9f\x14\x9d\xdf\xee\x91\x4d\x6f\xbc\x98\x55\x3e\xf4\x13\xa3\x51\xa6\x2d\xc0\x87\x07\x32\xf4\x60\x73\x75\x76\xa8\x47\xb0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x8a\xff\x2a\xfd\x04\xd0\x71\x2d\x2a\x98\x05\xac\x3e\x5c\x21\x47\x09\x0a\xb1\xf6\x63\x75\xf2\x21\xbd\x68\x93\xc4\xb4\x39\x99\xee": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0-beta.0
    b"\x8b\x07\x9a\xfb\x72\x68\x64\xd4\x7b\xe6\x9b\xf9\x5c\x17\xc1\xf6\x9a\x0f\xd3\x8d\x6c\xa2\xc9\xd6\xf3\xed\x85\xea\x3f\x13\x5b\xc4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x8b\x09\x1d\x18\xd8\xaf\x76\x04\x70\xa7\xe7\x93\x68\x46\x5a\x99\x43\xfa\xcd\xa7\x32\x9f\x98\x46\x1a\x47\x94\x6c\x8e\x9e\x5d\xea": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x8b\x20\x53\x90\xb4\xa9\x1d\x86\xc9\xc0\x86\xde\x5d\xf6\x51\x65\x40\x2f\x8c\x19\x30\x60\x1b\xca\x3b\x75\x0c\x21\xa7\xcb\x7f\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.294
    b"\x8b\x24\x7d\x06\xb7\x91\x95\xce\xee\x8d\x0c\xc2\xbe\x3b\x09\x2f\x25\x0e\x3c\xea\x6f\xaf\x1c\x09\xbd\x55\x49\x66\xac\xe3\xaa\x25": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.294",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8b\x38\x23\xbc\xbd\xd0\xb1\xdf\x4d\x42\x0a\x09\x3a\x93\x5b\xd9\xbb\x96\x6c\xcb\x5a\xd2\x43\x5e\x00\x4f\x4b\x24\x95\x40\xb2\x40": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x8b\x3e\x50\xe7\x04\x49\x54\x74\x30\x60\xe5\x9f\x4c\x67\x0a\xb5\x45\xcb\xa3\x96\x5f\xc4\x35\x20\x84\xa0\x7c\x86\x3b\x8b\xc0\x24": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191
    b"\x8b\x42\x91\x95\x6a\xce\xa9\xb1\x95\xdf\x33\x6e\xc8\xa5\xc4\xf9\x8a\x48\x32\x4f\x7e\x16\x8b\x43\x08\x99\x9f\xc4\x2a\x15\xff\x7e": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.191",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x8b\x45\x4f\x04\x00\xce\xe1\xe9\x52\x83\xb1\x87\x82\x01\xe9\xd5\x92\x95\x52\xde\x59\x38\x83\xf5\x95\x38\x03\xf6\xce\x5d\xf5\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Bytes.sol",
            name="Bytes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Bytes.sol",
            name="Bytes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Bytes.sol",
            name="Bytes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Bytes.sol",
            name="Bytes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Bytes.sol",
            name="Bytes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7
    b"\x8b\x5a\xd7\x06\xcc\x1f\x7e\x59\xed\xb0\x0f\xf4\x08\x68\xbe\x0a\x3d\xf7\x95\x97\x39\x59\x72\x7d\x33\x38\xc6\x9d\xf5\x0a\x77\x3c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\x8b\x75\xa4\xff\x46\x63\x79\x99\x31\x44\x4c\x39\xdc\xe0\xf5\xb6\x1a\xff\x58\x65\x19\x46\x89\xfc\xe3\x70\x0c\xb0\x59\x96\x72\x3c": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\x8b\x83\xa6\x12\x2b\x9d\x96\x62\x1c\xe8\x48\x03\x4d\xe1\x59\xe9\xa4\x4a\xec\xb9\x7c\x5d\xf9\x38\xb0\xda\xe5\x7f\xe5\x6e\x3f\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x8b\x8d\xf0\x63\x23\xdc\x1a\x07\x86\x4e\x15\x44\x14\x92\xfd\xfc\x44\x80\x92\x5a\x0b\x42\x7c\xf6\x5b\x57\x88\xad\x55\x48\xf4\xff": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x8b\x95\xc8\x80\x53\x66\x3f\xc8\x96\xbe\x02\x34\x63\x58\xd9\xbb\x15\xe8\xbb\xbf\x47\x6f\x3a\xd5\x1a\xfd\xff\xf9\x29\x05\xf2\x74": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x8b\xad\x17\x4f\x8d\x86\x19\x01\x5c\x17\xe8\xa8\xb5\xd4\xa5\x43\x64\xbe\xa0\xd9\xe0\x84\x34\xbd\x9a\xbe\x9c\x48\xef\x3b\x34\x5c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.1.5, v0.1.6
    b"\x8b\xb4\x3a\x69\xe8\x47\x6b\x9d\x53\x49\xa9\xbb\x47\x14\x85\x6b\xa1\xdc\xcf\x01\x99\x37\xd6\x76\xba\x3f\xa2\x2c\xe4\xb9\x12\xba": [
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x8b\xe1\x32\x54\x4d\x45\x58\xcc\x9a\x02\x9e\x4d\x4b\xfe\x89\x30\x5a\x1a\xe9\xa4\xd8\x72\x97\x84\xaa\xe1\x0e\xca\xb1\xbd\x05\x43": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="ITransparentUpgradeableProxy",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\x8b\xe1\x3f\x9c\xfb\x8d\x72\xe2\x2b\xe7\x43\x64\x04\xd7\xf9\x95\xd8\xad\x2e\x85\xad\x6f\xda\xf3\x6f\xc1\x06\x6f\xb6\x12\xcb\x98": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/lifecycle/Pausable.sol",
            name="Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/lifecycle/Pausable.sol",
            name="Pausable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x8c\x2a\x5c\xdf\x08\x1c\xd1\xff\x9e\x2d\xb2\x02\x45\xe4\x4e\x3f\x9b\xab\x21\x75\x0d\x82\x54\xec\x2e\x9b\x62\xb8\x90\x9e\xab\xca": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x8c\x43\x73\x8e\xd1\x15\xb4\xba\x51\x7a\xf0\x48\x9a\xdf\xb5\xe3\x08\x91\x8c\x65\x94\x54\x6c\x37\x75\xe7\xdc\x27\xb0\x74\xdc\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x8c\x49\xf8\x33\x54\xb7\x42\xc7\xdd\x36\xe8\x03\xe1\x27\x78\x47\x7c\x72\xd0\xf1\xdf\x53\x5c\xb5\xac\x52\x17\x31\x9f\x82\xa6\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x8c\x4c\xee\x80\x3a\xb0\x05\x24\x8a\x4d\xdd\xfc\x89\x1d\xd7\x40\xe7\x2c\x0a\x33\x3e\x2c\x32\xe4\x81\x2b\x48\xff\x1f\xfe\x1c\x14": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x8c\x5b\xcb\xe8\x9c\x0f\xa9\x83\xb3\xd1\x6d\xc2\x10\x1e\xd0\x51\x80\xc4\x3d\x1c\xaf\x1e\x66\xbb\xe1\x44\xbd\x28\x93\x5c\x77\x05": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x8c\x73\xc9\x38\xf8\xd2\x3a\xd4\x25\xac\x0e\xd4\xfc\xae\x54\x46\xf3\xdc\x1b\x3e\x01\xa0\xbf\x80\x56\x57\xb5\x90\x49\x11\x9a\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="ICompoundTimelockUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x8c\x77\x7e\xf7\xde\x07\x52\xb9\xe2\x94\xf9\x04\x48\x3a\x44\xe5\xc3\x34\x84\x4c\x58\x71\x8a\x70\x5a\x19\x87\xe6\x44\x66\x74\x93": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Utils.sol",
            name="ERC1967Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x8c\xbf\x39\xdd\xa6\xdd\xf9\x7c\x14\x2a\x9f\xe4\x4b\xf8\xbf\xc4\xdf\x66\xf2\xfe\x53\xa4\x10\x4f\x1b\xea\xa8\x19\x85\xe1\x68\xc1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77
    b"\x8c\xc8\x9b\xfc\x08\xff\xe9\xf1\xe8\x6f\xe5\x3d\x90\xa3\xf5\x66\xb3\x6d\x7d\xfd\x51\x1c\x33\x4a\x71\x42\x1f\xe6\xa2\xfe\xf7\x12": [
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EIP712.sol",
            name="EIP712",
            version="v0.0.77",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x8c\xd7\x54\xf2\x30\x55\xba\xf0\x82\x00\xd9\x6e\x42\xdc\x3d\x24\x61\x51\x0b\x0b\xba\xef\xcd\xcd\x05\x9d\xf5\x1d\xe5\x74\x84\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x8c\xe4\x2f\x68\x27\x9a\x5c\x13\xc3\x9d\x65\xa4\xca\xd2\xc7\xd3\x90\xa0\xc0\x9c\x45\x1a\x76\xf1\xee\x1b\xbc\x0f\xf7\x36\x13\xcf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x8d\x0e\x1c\xb4\xa2\xd3\xdc\x04\x7e\x2d\x54\xe2\xe2\x24\x05\x66\xe3\x79\xfe\x8f\x94\xf9\x74\x11\x2c\x17\x29\xef\xf7\x96\x81\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8d\x11\xe3\xc7\xd3\xee\x7b\xdc\xcb\xa5\x9e\xae\x1f\x55\xd9\x1b\x2a\xf1\xb9\x9d\xe8\x4a\xf2\x67\x4c\x4b\xf3\x11\xaf\xfb\x24\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.292
    b"\x8d\x18\xd4\x75\x8e\xae\xcb\x5c\xb8\xd0\xd2\x46\x67\x29\xf9\x74\xf8\xec\x3a\xbc\x60\xb9\xca\x1b\xe6\x90\x2c\x77\x3e\xce\xaa\x01": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263
    b"\x8d\x32\xc2\x4b\xea\x58\x8b\xda\x60\x6d\xfd\x83\x28\x90\xd1\x64\x60\x36\xe3\xc1\x1c\xe7\x08\xb1\x03\xb1\x77\x5e\x7c\x2d\xb3\x2f": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.263",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x8d\x3b\xeb\x5f\x0a\xab\x10\xe4\xc5\xa3\x69\x73\x3e\xaa\x3d\x30\x0a\xf8\x4b\x08\xfe\xf8\x7f\x0d\xf2\x3b\xee\x38\x42\x70\xa5\x95": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.75, v0.0.76, v0.0.77, v0.0.78
    b"\x8d\x3b\xfd\x42\xcd\x00\x63\x15\x87\xd0\x17\xb8\x5f\xd6\xdd\x06\x82\x85\xf5\xea\xe9\xf1\x20\x7c\x77\xa3\x95\x19\x28\x59\x21\x45": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.78",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8d\x4c\x54\xea\xb2\x42\x27\x48\xac\x4e\x2a\x73\x05\x20\xcd\xa0\xb3\xc1\xdf\x98\x4b\x27\x45\x61\xd8\x51\xf2\x12\x20\x72\x30\xa4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x8d\x52\x09\x4d\xd2\xc9\x66\xe6\xbb\x11\x3b\x5c\xeb\xf1\xa2\x78\xc3\x31\x0b\xd2\xb2\x33\x2d\xa4\x53\x10\x69\x48\xbb\xd5\x3e\x47": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x8d\x59\x25\x7d\x87\x02\x0a\xa2\xdb\xd7\x54\xb4\xb3\x44\x02\x00\x6e\x09\x16\x85\x27\xf7\x82\xb1\xef\xf3\x3b\x62\x2c\x17\x87\x28": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.4.1, v3.4.2, v3.4.1
    b"\x8d\x64\x79\xce\x17\x33\x3c\x94\x5b\xc2\xd0\xfe\xa7\x1b\x4c\x1a\x6b\x44\x22\xfa\x69\xe0\x94\xf5\xe2\xdf\x43\xcd\x24\x46\xf6\x1e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x8d\x69\x70\x57\xa9\x6b\xc2\x75\x74\x1d\x3b\xee\x9f\x93\x5d\xa8\xba\x74\x16\x30\x0d\xa2\x35\xa2\x49\xae\x27\x38\xe1\x9c\xab\x44": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x8d\xa3\x0b\xf6\xa0\xa8\x52\xb5\x71\x34\x94\x72\x5e\x39\x50\x2a\x1e\x42\x8d\x35\x31\x98\xfd\x3f\x8f\xb5\xc8\x21\xef\x8a\x2e\x4f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x8d\xe1\x27\x24\x20\xc7\x50\x78\x15\xea\x04\xc1\xb3\x68\x66\x53\x6c\xa8\xac\xee\xb7\x5a\xe3\x2a\xe7\x66\xea\x26\x36\x59\xc2\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.77, v0.0.78
    b"\x8d\xe4\x10\x82\x72\xe4\x9a\x86\x22\x58\xaf\x6e\xe5\x2b\xe4\x1e\xc5\x8b\xb6\x8e\xd8\x82\x2c\xaf\x12\x55\xc3\x3f\x01\xc6\x1d\x44": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.78",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x8d\xf3\x97\x08\x10\x87\x94\xb8\xd5\xc2\xe2\x89\xb7\x96\xb2\x2b\x09\x94\x22\x58\x74\xff\x93\xc8\xcb\x81\x39\xea\x05\x44\xd2\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x8e\x0a\x3e\x03\x55\xab\x1a\x94\x7d\xd6\xf0\xc8\xbb\x7a\x14\x9c\x51\x78\xdc\xe6\xd9\xd2\x3f\xc7\xa2\xb2\xd0\x31\x6c\x7a\x13\xe5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x8e\x25\xff\x0a\x0c\x0a\xf3\x06\x48\x09\x5e\xe1\x21\x76\x43\xc9\x14\xd6\x33\x0c\xd6\x32\x99\xce\x2f\x85\xa1\xeb\xca\x32\xa0\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesSuperQuorumFractionUpgradeable.sol",
            name="GovernorVotesSuperQuorumFractionUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesSuperQuorumFractionUpgradeable.sol",
            name="GovernorVotesSuperQuorumFractionUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.19
    b"\x8e\x2c\xfa\x0a\x1a\x68\x53\x9e\xbd\x76\x40\x6b\x4d\x84\x9f\xa2\xad\x41\x6f\xb9\x1d\x7a\x9c\xa2\xa4\xf6\xfc\x3a\x5f\x25\x87\x50": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\x8e\x4a\x59\x3d\x7f\x34\x5d\xeb\xb4\x93\xf0\xf7\x54\x18\x5a\xf0\xcc\xc9\x3d\xcf\x41\x96\xa6\xaa\x68\xec\x00\xc2\xd4\xff\xff\xa9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0
    b"\x8e\x54\xb6\x96\xc3\x6c\xa9\xe4\x29\xdc\x29\xff\x4e\x71\x37\x55\xfb\x1b\x35\x61\x70\x29\xcb\xd7\x9c\x20\x5a\x72\xd5\x48\x10\x8f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8e\x58\x0c\xa0\x07\xf4\x37\x12\x5b\x85\x98\xb4\x86\x09\xd2\x76\xd8\xe0\xf8\x99\xce\xeb\xc8\x21\x37\x5c\xa8\x7f\xb7\xb6\x4d\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7
    b"\x8e\x65\xbe\xa1\xec\x86\x6a\x07\x98\xbb\xc8\xd7\x7f\x32\x4f\x3d\x62\x14\x04\x01\x36\x01\x5f\x3d\x0f\x16\x0e\xc2\xf8\xe9\xf7\x6b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x8e\x6f\xde\x92\x26\x07\xfc\x32\x29\x9d\xcf\xa3\x72\x58\x56\x33\xa4\xe9\x69\x0d\xfc\x66\x5e\x08\x1f\xa5\xfa\x21\xbf\x60\xf3\x53": [
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/CWIA.sol",
            name="CWIA",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.29, v0.0.30, v0.0.31, v0.0.32
    b"\x8e\x86\xce\xa6\xab\x95\xa8\x2f\x21\x8b\x44\x6f\x30\x05\x41\x66\xb5\x13\x69\xa9\xf1\x10\xe0\x62\xb5\x39\x36\x76\x58\x17\xee\x21": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.32",
        ),
    ],
    # Hash appears in versions: v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275
    b"\x8e\xa0\x43\xfa\x95\xfa\x08\x3d\xe3\xb0\x40\xf6\x13\x81\x59\x95\x5c\x87\x91\x0c\xfa\xaa\xd3\xc5\xa0\xd4\x61\x63\xd8\xa2\x81\x36": [
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x8e\xaa\x7d\x0a\x25\x99\x82\x59\xa9\x93\x6e\x63\xff\xfa\x26\x3d\xcc\x4a\x88\x5e\xf8\xee\xfd\x0b\xb5\x60\x90\xef\xff\x3f\x44\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166
    b"\x8e\xc5\x3d\x98\x0b\x49\xe5\xeb\xb9\x2f\x8d\xc0\x73\xa2\x50\xdb\x04\xb9\x89\x96\x20\x71\xa5\x61\x17\x43\x84\x52\x7d\x98\x4c\xfe": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.166",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x8e\xdb\x0c\xbe\x78\x25\x77\xb9\x5d\xcd\x26\x14\xf1\x48\x07\xd4\x67\x9e\xff\x5d\xcb\x03\x14\xaa\x26\x32\xf0\x53\x01\x5f\x77\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909Metadata.sol",
            name="ERC6909Metadata",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909Metadata.sol",
            name="ERC6909Metadata",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x8e\xed\x33\xe7\xeb\x28\x4a\x7c\x3f\xe4\x86\xef\x02\x77\x7e\xb7\x8c\xb0\xff\xb8\x84\x16\xd3\xd9\xdb\xbb\xf5\x5a\xd0\x8e\xe0\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x8e\xf8\xd1\x42\x80\xd5\x90\xdb\x33\xb6\xbf\x84\x1b\x93\x68\x56\xee\xe9\x84\xa3\x4d\x28\x1c\x4c\x37\x22\x14\xe2\x77\xf8\xd5\x88": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\x8e\xfb\x7a\xe0\xb0\x83\x26\x9f\x2c\x58\xe2\x36\x44\x8c\xb7\x06\xac\x49\xbe\x90\x05\x4e\xe9\x0d\x57\xa7\x2c\x70\x18\xbf\x86\x5d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\x8f\x18\x05\x3f\x20\x69\x0d\x8d\xbf\xb2\x7e\xdf\xec\x5f\xd2\xf2\xa2\x33\xc5\x2b\x3c\x91\xf9\xac\x61\x85\xe4\xd3\x6a\x76\xd4\x3f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\x8f\x20\x0d\xc6\x3b\xff\x5c\xc7\x3e\x8d\x17\x8e\x1f\x6e\x9a\x68\xd1\x23\x9f\x09\x8e\xc0\xbb\x18\x73\xdf\x1d\x89\x10\xad\x82\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v0.1.15, v0.1.16
    b"\x8f\x24\x20\x57\xd5\x5c\xec\xe4\xd0\x79\xbd\x57\xec\xb9\x08\x18\x0d\x76\x9a\xe0\x92\xa1\x38\xb1\xc3\x01\xe4\x71\xa6\xbb\xa1\x60": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x8f\x27\xbb\xe5\xa1\x21\xca\x46\x91\xe7\x28\x9a\x11\x1a\x14\xb4\x7b\x94\xe5\xeb\x86\xf5\xc4\x43\x88\x2b\x6c\x01\xa1\xad\xba\xd6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\x8f\x29\x44\x58\xd2\x36\x41\xc9\xf6\xd4\xc3\xb1\xff\x47\xfe\xac\xc3\x5e\xa1\x78\x4c\xed\x5c\x08\x6e\x55\x94\x65\xee\x81\x3b\xe2": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v6.2.0
    b"\x8f\x40\xa1\x0b\xb4\x99\x27\x41\xe3\xc1\x6a\x3f\xa0\xcd\xf7\x92\x8b\x6d\xdf\x0b\x27\xdc\x3e\x25\x97\x4f\x0a\xf2\xbb\x3d\x5a\x2e": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x8f\x47\x19\x8d\xb1\x26\xbb\x07\x49\xdb\xa1\x67\x38\xdd\xe8\x6e\xab\x69\x92\x56\xa7\x0a\x8d\x16\x4c\xcf\x46\x88\x3b\xe0\xeb\xe4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UtilsUpgradeable.sol",
            name="ERC1967UtilsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8f\x4e\xcb\xb2\x48\x5d\x7f\xec\x80\x1a\x69\x4c\x78\xd2\xb0\xd1\x24\x93\x92\x13\x0e\xcd\x0b\x44\x78\xdf\x83\x62\x36\x7f\xd7\x59": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295
    b"\x8f\x63\x54\xed\x19\x43\x36\xbb\xc1\x4a\x6b\x6c\x80\xe0\x2c\x20\x6a\x8a\xbe\xf7\x44\x9c\x27\xe1\xdd\xe8\xec\x7f\x18\x81\x94\x55": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.295",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\x8f\x6a\x7f\x6b\x49\xf5\x9b\xb8\xb6\x46\xab\x73\x24\x3e\x7c\xa3\x93\x50\xaf\x3d\x13\x60\x6f\x77\x69\xca\x06\x04\x64\x54\x77\x2f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupplyUpgradeable.sol",
            name="ERC777PresetFixedSupplyUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247
    b"\x8f\xa0\xb2\x3b\xc6\xcd\x94\x45\x81\xdf\x4d\xf0\x7d\x9b\x6c\x17\xad\xfd\x11\xa6\xda\x0c\x87\x1f\xdc\x03\x1f\xeb\x4f\x7b\x8a\x17": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.247",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x8f\xc5\xbc\xda\x56\x01\x02\x9b\x85\x5b\xab\x18\xad\x67\x32\xd3\xbc\x67\xba\x44\x85\x0e\x91\xac\xe9\xc1\xd3\x19\xe6\x23\x83\x98": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IInbox.sol",
            name="IInbox",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.226, v0.0.227, v0.0.228
    b"\x8f\xed\x55\x1e\xaa\xc3\x84\x35\xe9\x40\xf3\x6a\x0c\x13\x83\x1a\x5e\xff\x33\x06\xe7\x7f\x76\x50\xfa\x57\x2b\xb8\xb9\x61\xb2\x85": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.228",
        ),
    ],
    # Hash appears in versions: v0.0.123
    b"\x8f\xfd\xf7\x32\xd8\xdc\x8e\x78\x53\x54\xe6\x9b\xe8\x14\x2d\x16\x13\x7a\x24\xaf\x7c\x14\x03\x4a\x1d\xbf\x4e\x11\x77\xa5\xeb\x7d": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.123",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x90\x1b\x7b\x9b\xd6\xbf\x9d\x81\xd7\x99\x62\xc3\xa6\x79\x7a\xc0\x72\xa3\xb0\xc1\xc9\x8d\xe3\x3b\x4f\xd5\x0f\x14\x3e\x58\xeb\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x90\x2e\xfd\xf3\x14\x04\xaf\x41\x7d\xa8\x79\xc8\x25\xda\x60\xa8\x28\xc6\x4e\x20\xcb\x47\x2b\x9a\x31\x2b\xc0\xcf\xb5\xec\xa3\x9c": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IMulticall_v4.sol",
            name="IMulticall_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x90\x37\x23\x2a\x97\x60\xc3\x19\x99\x8a\xae\xe0\x1b\xf8\xd8\x2c\x6e\xf7\x27\xa7\x92\xf1\x04\x21\xd4\xe9\x7c\x70\x26\xe9\x0d\x00": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x90\x4e\x3d\x39\x68\x96\xd7\x3e\x41\x32\x12\x84\x9e\x57\xd6\xfe\xaa\xf7\x29\x62\x8c\x41\x04\xc6\x57\x45\xba\xc2\x85\x6e\x74\x15": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x90\x57\x64\xb3\x55\xa6\xbd\x1d\x0f\x89\x6e\x31\x74\x0a\xdf\x6c\xcd\xd3\x33\xaf\x1d\xcd\x0b\x1c\x27\x40\xc9\x26\xd8\x3b\xf3\x64": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\x90\x73\x8b\xfd\xae\x61\x13\x07\x17\x77\x2a\xeb\xbb\x22\xed\x62\xcb\xc1\xd9\x46\x4e\x98\x67\x3e\xf1\xdf\x1f\xd8\x10\x88\xfe\xae": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209
    b"\x90\x7d\x37\xad\x17\xab\x08\xc5\x89\x11\xe1\x54\x9b\x19\xa6\x5a\xc6\xd2\xfd\xa9\xba\xf0\x91\xc9\x38\x83\xa3\xc9\x63\xed\x61\x17": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.209",
        ),
    ],
    # Hash appears in versions: v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192
    b"\x90\xca\xab\x4f\x37\x17\x43\xf6\x36\x12\xdb\xa8\x3a\x90\x3c\xe7\x34\xad\x8e\xd1\x26\x6b\x05\x76\x55\xa1\xe7\xd8\xc8\xae\x04\x05": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.192",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x90\xd2\x53\x95\x62\xa0\x16\x9d\x65\x70\x88\x7b\x9b\x22\x5c\x3a\xb5\x6e\x67\x71\x78\xef\x18\xc4\x90\x46\x4b\xcb\xf7\xea\xa0\x5f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x90\xed\x61\xfb\x65\xd9\x03\x8f\xff\x47\x3b\x3d\x61\xae\x34\x21\x28\xdf\xc3\x2a\x37\xb9\xe5\x7b\x06\x18\x98\xa0\x6c\xc1\x74\x39": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x91\x08\x6b\xb7\xe8\x1d\x9a\x11\x56\x62\x0d\x9f\xd9\x9b\xcd\x87\x22\x28\x03\x17\xd7\x6c\xa7\x2d\x26\x99\xc7\xdf\xa3\xc3\x4f\x8d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\x91\x35\x96\x18\x07\x65\x38\x04\x7f\x28\xb5\xf0\xc4\xb7\x4c\xb3\xdc\xe2\x52\x6b\x8b\xa5\xac\xea\x37\xea\xbe\x4f\x1c\x72\x21\xfd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x91\x47\x7d\x8b\x73\x6c\x3d\x32\x82\x14\x0b\xa4\x75\x9f\x75\x7f\x87\x31\x70\x2e\x8d\x91\x0f\x3f\x55\xd1\x68\x45\x72\x33\xc9\xef": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.1, v5.0.2
    b"\x91\x78\x17\x05\x37\x02\xdc\x55\x73\x59\xe0\x49\x71\x2b\x0b\xab\xb0\xe2\xe0\xe1\xa3\xeb\x23\x4d\x7b\x30\xe1\x0f\xf5\x59\xf1\x77": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12
    b"\x91\x83\x1b\xcd\x64\x19\x3c\x36\x0e\x31\x27\x8d\xb8\x7a\x92\x2d\x41\xef\xf1\x68\xea\x3d\x33\x8e\xad\xa4\x97\x0f\x83\x21\xd2\xf9": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.12",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x91\xaa\xc2\x65\xf7\x02\x46\xd7\xf2\xb1\x64\x0c\xad\x55\x18\xe9\x46\x5e\x43\x63\xff\xbe\x4b\xb6\x09\x6e\xc1\x6c\x4e\xa7\x2c\xf1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC5805Upgradeable.sol",
            name="IERC5805Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55, v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78
    b"\x91\xb1\xbf\x33\x7d\x87\xea\xd8\x96\x03\x80\x53\x23\x49\x83\xf4\x0d\xd3\x8c\xf3\xab\x0e\xff\xe0\x83\x0e\x25\x34\xc8\xeb\x3d\xf7": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.55",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.78",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x91\xd7\x97\xea\xd9\x0e\x05\x20\xce\x30\x99\x40\xa7\xce\xe6\x27\xef\x61\xf8\x31\x12\xa5\x35\x49\xeb\x64\xbf\x2b\x15\x04\x4f\xf7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Module",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Module",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Module",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Module",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Module",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x91\xe9\x2e\xdc\x34\x85\x01\x03\xfc\xa7\xb1\x73\x4a\x6a\x91\x72\x80\xf6\x31\x6d\x40\x29\x25\xd2\x4e\x37\xfe\x5d\x01\xfe\xf6\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\x92\x06\xb5\xb1\xcf\x8c\xaf\xcf\xed\x82\x39\xa9\x40\xe0\xb1\x81\x73\xdd\xcc\x97\xa4\x6e\x2b\x87\xc3\xf4\xaf\x99\xad\x2a\x8a\xad": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\x92\x0d\xe5\x70\x3b\xc3\x2c\x23\x8b\xea\xe3\x5c\x58\x92\x74\x5d\x76\xb9\xb1\x5f\xdf\xe6\xd1\x73\x8f\x4b\xba\x31\x9f\x21\x5e\x56": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x92\x1d\x48\x23\x99\x31\x6d\x78\xc8\x2c\xfa\x58\x39\x50\xc8\x23\xd2\x14\xe0\x6e\xb9\xb2\x95\x78\xea\xc6\x81\x8b\xd4\xab\x6b\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.0.1, v3.0.2
    b"\x92\x2b\x2a\x03\x4b\xc5\xc7\xa2\x65\x80\x5a\x47\xc5\x14\xd7\xc5\xad\xe5\x2d\x68\x48\xfd\x31\x06\x56\x2c\xff\x4c\x80\xe2\xb1\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\x92\x3e\x98\xe7\x4a\x74\xa3\xad\x30\x4b\xa4\xf3\xe2\x3a\x42\x68\x29\x37\x47\xe9\x04\x4b\xb4\x17\x1c\x86\x86\x30\xf7\x5f\x8e\x26": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x92\x44\xef\x41\xcd\xc7\x96\xb4\x76\x1f\xdf\xb2\x53\x64\x30\xc5\xd2\x23\xf3\xf7\x61\x26\xcf\x1c\x47\xba\x43\x18\x9b\xa9\x42\x76": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20PermitUpgradeable.sol",
            name="IERC20PermitUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x92\x83\x4b\x18\xf4\x72\x51\xd6\x2a\xae\x3f\x80\xcd\xcd\x58\x5e\x02\xcf\x24\xbb\xf3\xcd\xa6\x00\x22\xf4\x13\xfa\x3d\x20\xcf\x63": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.18
    b"\x92\xe2\x07\xd7\x6d\x48\x91\x35\x10\xe2\x9f\xfe\x66\x06\xae\x11\x7a\x8f\xf2\x05\x99\x41\xbb\x83\x7d\x84\x13\x51\x38\x6f\x27\xc5": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.18",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x92\xe2\x72\x3b\xeb\x0c\x28\xbf\x78\x31\x7e\xec\x66\x5c\xae\xe0\x4c\xa4\x72\x95\xac\xd1\x1a\xb9\xf2\x7b\x8d\x33\xca\x48\x81\xd5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable2Step.sol",
            name="Ownable2Step",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x92\xe7\x4c\x8d\xfd\x81\xaf\xb0\xf6\x24\xee\x89\x6d\xdb\x71\x05\xcb\x6f\xcb\x5e\x38\x82\xf7\xb1\xd4\xba\xb0\x6d\x31\x95\xfb\x5e": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/BaseV4Quoter.sol",
            name="BaseV4Quoter",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x92\xf2\xe3\x08\x5e\xfc\x7e\xdb\x12\x0f\xbc\xe9\xfd\x66\x5b\x9b\x0c\x87\x6c\x1b\x06\x4b\xe1\xcb\xd3\xcd\xf1\x6d\xed\x6c\x33\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x93\x02\xf1\x1b\x03\x3a\xf0\x9b\x2d\x21\x59\x16\x01\x49\xf0\xd3\xa1\x3d\x2b\x3c\x38\x22\xfd\x99\x7d\xc0\xcc\xe3\x89\xd6\x30\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\x93\x12\x28\x5f\xab\xe7\x47\x0c\x8f\x49\x7b\x21\x06\x33\xa6\x78\x9f\xd9\x99\x4d\xa0\x8b\xd8\x44\x21\x7b\x20\xc3\xa6\xaf\xe3\x5f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.7.3
    b"\x93\x23\x0e\x33\x19\x1d\xb2\xcc\xcb\x2c\x5f\x31\xf5\xdd\x37\xe1\x64\x17\xa9\x4e\x3e\x89\x50\x14\x30\x74\x20\x09\xcd\x53\xf1\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.34, v0.0.35, v0.0.36
    b"\x93\x39\x87\x0b\xcc\x60\x97\x9f\x63\x63\x45\x1a\x96\x0c\xc7\xa0\x03\x07\x4f\xf1\x00\x9e\x71\xbb\xa6\xd2\x8e\x35\x84\xcd\x10\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.36",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76
    b"\x93\x39\xb6\xc9\x5a\x35\xd1\x34\x02\x9f\xd8\xfb\x9d\x6f\x81\xa0\xa4\xa6\x88\xc3\x91\xa4\x27\xba\x8f\x55\xfa\xe9\x21\xcf\x44\x0d": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.76",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x93\x3f\x6a\x44\xf2\x42\x73\x3c\x9f\xc4\xb8\x6c\x66\xf4\x9b\x0e\x0f\xd1\xac\x9e\xe9\xef\x97\x28\x9f\xe4\x64\x8a\xbc\xf1\xf2\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281
    b"\x93\x62\x88\xf4\xdb\x27\x0b\xc2\x01\x06\xc9\x88\x42\x7c\x5e\xe5\x7f\x78\x3d\x94\x8d\xff\x92\xbe\x20\x02\x94\xc9\x97\x59\x97\xe0": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x93\x65\x77\x77\x92\xed\x08\xde\x35\x97\x5d\x77\x09\xd6\x1f\x93\xfb\x19\x2f\x0c\x34\xd6\xad\xd3\x52\xcd\x3e\x43\xf0\x1d\xd1\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x93\x67\x73\x2d\x90\xcb\x98\x52\x91\xac\x31\xb1\x16\x71\xc0\xa2\x04\x72\xf8\x99\x79\xed\xae\xcd\x44\x1d\x94\x73\x74\x51\x4f\x77": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/BitMath.sol",
            name="BitMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x93\x76\xdb\x9d\xe8\xaa\x9f\x6f\xbe\x89\xb6\xec\x93\xd5\xf3\xbd\x06\x8a\x8d\xf8\xb8\x58\x75\x9f\xa3\x48\xde\x49\x13\x82\x3a\xa9": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/IUniswapV3PoolDeployer.sol",
            name="IUniswapV3PoolDeploye r",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x93\x78\x5e\x87\x50\x1c\xc0\xbf\x18\x53\x26\xd3\x9e\xb6\x34\xa6\xd5\x25\x9e\xb2\xa9\x3d\xf1\xd7\x26\x92\xa1\xe3\x38\x9d\xc8\x87": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34
    b"\x93\x94\x20\xec\x50\xc1\x30\xb1\xff\x35\x86\xa2\x19\x15\xa0\xb7\xba\xf3\xb2\x3c\x91\xf0\xd7\x66\x9b\x83\xc4\x09\xee\xd2\xae\x24": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.34",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x93\xca\x8a\xad\xa2\x54\xa7\x43\xde\x41\xae\x1b\xb4\x1b\xfc\xd6\x86\xb5\xe2\x82\x20\x03\xbb\x97\xa1\x5a\x14\xf7\xdf\xc6\xe1\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x93\xee\x3a\xb6\xc9\xd4\x88\x01\x88\x9a\x22\x98\x4d\x06\x1d\x29\xdb\x53\xcc\xf4\x62\x63\x2f\xa3\xc5\x03\x85\x82\x8c\x83\x92\x1f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManager.sol",
            name="IAccessManager",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.222, v0.0.223, v0.0.224
    b"\x94\x15\x2b\x99\xad\x4d\x27\x73\xe9\xcd\x61\xe4\x3c\x21\x46\xdb\xfe\x51\xb8\xd9\xa9\xc9\x94\x40\xe7\x97\xbe\xc8\xda\xd5\x15\xc5": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.224",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x94\x1b\xba\xfc\xe8\xb4\x6f\x8c\xf9\xb3\xbd\x8d\x9e\xa4\x8f\x23\xdd\x44\x40\xac\x73\x48\xa4\x17\xc9\x15\x88\x3f\xbd\xa7\x2b\x08": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160
    b"\x94\x25\x45\x7d\x0d\xf3\x10\x99\xfb\x83\x8d\x44\x61\x4e\x52\xda\x62\x41\xbc\xfb\x73\x8c\x17\x4f\x32\xa0\xbf\x45\xc8\x2b\xa3\x20": [
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.0.160",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x94\x32\xe7\x36\x50\x26\xe3\xa1\x19\x64\x90\x04\x81\x13\xf1\x2a\x7f\x9d\x93\xbf\xc7\x18\xf0\x94\x1e\x15\x4e\xa9\xcb\xad\x9e\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x94\x3b\x59\xfd\x08\x62\xb2\x36\x6f\x0a\x5e\x7a\x7e\x8c\x0c\x34\x05\x4b\x4b\x1a\xaf\xbd\x9e\x34\x55\xfb\x84\xa2\x36\x09\x59\x79": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x94\x51\xd2\xb7\xa9\x79\xcb\x47\x04\x54\xd0\x4e\xf1\xc0\x73\x95\x3e\xc4\x0f\x90\x64\xcf\x1e\xca\xe4\xb7\x28\x2d\xe4\xa5\xd4\x3c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\x94\x54\xe9\x0b\xd1\x0b\x54\x66\x00\xac\xfa\x2a\x31\x4a\xcb\x21\x59\x9a\x35\x05\xee\x73\x35\x5a\x0b\x75\x11\x74\xe1\xe3\x50\xa2": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/TickBitmap.sol",
            name="TickBitmap",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x94\x6b\x31\x8a\xc0\x24\xbe\xca\xa2\x50\x8f\xa6\xeb\x22\xb8\x15\xb5\x2c\xf8\xca\xfa\x54\x27\x44\x16\x8b\xd5\x37\x9d\x9d\xe6\x51": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/amb/IAMBUpgradeable.sol",
            name="IAMBUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.289
    b"\x94\x7b\xe5\xc8\x9c\x0c\x87\x87\xaf\x92\x0c\xc4\x8c\x09\xae\xe5\xbf\x50\x97\x05\x89\x6e\x9d\x86\xed\x80\x62\x27\x9b\x6d\x7e\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.289",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281
    b"\x94\xb7\xac\x05\xcf\x73\xf6\x58\xbf\x24\xdc\xf0\x6d\x98\x49\x95\x59\xf0\xcf\x79\xb5\x17\x79\xed\x4d\x18\xf7\x82\xb5\x95\x49\xff": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v0.0.122
    b"\x94\xc2\xd2\x89\x28\xa3\x43\x59\x8d\x39\x43\x4e\x26\xad\x54\xc7\x57\x48\xa1\x55\xfb\x2a\xef\xa9\x03\x16\x0d\x94\xac\x52\xdf\x6c": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.122",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x94\xe3\x35\x1f\x34\x06\xa1\x96\xd2\x62\xcf\xf8\x7f\xfb\x36\x7a\x6a\x62\x43\xd8\x66\x73\x26\x2c\x2b\xe3\x41\x46\x53\xf0\xc9\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909TokenSupplyUpgradeable.sol",
            name="ERC6909TokenSupplyUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909TokenSupplyUpgradeable.sol",
            name="ERC6909TokenSupplyUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x94\xe3\xca\x00\x7e\x88\x18\x0d\x12\x47\xba\x9c\x32\xd7\x12\xca\xcd\xdb\x7a\xee\xcd\x38\xb5\x36\xed\xcc\x0f\xc9\x72\x05\x2f\x95": [
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/legacy/LibCWIA.sol",
            name="LibCWIA",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x94\xf8\xae\xb8\x0f\xc7\x27\x3b\xc3\x6e\x0a\xb2\x2f\x1d\x39\x35\x42\x0b\x74\xc5\x18\x4d\xc1\x5f\xc6\xee\xd0\x94\xa8\x53\x24\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146
    b"\x94\xfc\x8f\xc0\x02\x1e\x69\x09\xab\x99\x62\xe8\xf5\x9d\x20\x50\x9a\x02\xcc\x98\xc2\x97\xc8\x6d\xdb\x4b\x87\xa8\xa4\x59\xa5\xe5": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.146",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\x95\x18\xf5\xdb\x72\x9a\x27\xa0\x11\xa7\x07\x09\x2a\xce\xe7\x28\x5f\x5e\x68\xb4\x81\xf1\x38\x9b\x36\x81\x84\xe2\xb1\xfe\x44\x5a": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PoolInitializer.sol",
            name="PoolInitializer",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PoolInitializer.sol",
            name="PoolInitializer",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PoolInitializer.sol",
            name="PoolInitializer",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PoolInitializer.sol",
            name="PoolInitializer",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PoolInitializer.sol",
            name="PoolInitializer",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x95\x47\xa3\xdf\x35\xc3\xbe\x55\xaf\x6a\x8c\x0e\x96\x01\x00\x20\xc2\x97\xc0\x77\xf9\x85\x26\xf4\x5f\x16\xc4\xec\xf6\x4f\xd4\x29": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x95\x62\xc2\xca\xde\x21\x1b\x36\x21\x76\x62\xbe\xdc\xff\x3a\xda\xd0\x6e\x1a\x42\xd4\x49\xc8\xa8\x54\x16\x70\x51\x10\x91\x4c\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221
    b"\x95\x6a\x51\xdd\xf6\x30\xf2\x2c\x61\xba\xc5\xc3\xde\x1a\xd9\x06\xad\x8b\x11\x9c\x1c\x3b\x31\xd9\x27\x7b\xcf\x64\xa3\xf0\x58\xd4": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.221",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\x95\x97\xe1\xfa\xfe\xef\xf3\xef\x17\x5d\xb5\x40\x0e\x6e\xee\x8f\x8c\x46\x75\xbf\xed\x7b\x89\xe0\x78\xe0\xa1\x3e\x8e\xd8\x48\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\x95\xab\xd8\xac\x71\x43\x9a\xdf\x09\x7d\x53\xbb\xb6\x3c\xd1\x12\xcd\x0a\x94\xde\x81\x38\x5f\xf6\x59\xcb\x91\x35\xdd\xa7\xb5\xf4": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/IHooks.sol",
            name="IHooks",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\x95\xc0\x46\x7c\x55\x1a\xb7\xbb\x0e\x65\x01\x46\xf2\x2c\x8e\xd3\x24\x98\x14\x1a\xb6\xb5\xe2\xe6\x38\x38\x8a\xeb\x07\xa7\xd9\x32": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x95\xc4\x6b\xd7\xa0\x0b\x8c\x72\x7b\xb8\xd1\x7f\x1e\x81\x54\x89\xe5\xa2\x54\xac\x34\x4a\x33\x48\x07\xb0\xd9\xb6\x13\x9a\x34\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.1.9
    b"\x95\xce\xb2\xe3\x13\x4d\xbb\x6c\xb3\xfb\x09\x50\x02\x33\x64\x2e\x73\xf9\xa4\x73\xa2\x42\x3d\x7c\x9c\x92\xdf\xe3\xd4\xca\x47\x4b": [
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x95\xd3\x17\x76\x04\x12\x6a\x3b\x8a\xd5\xbb\x40\x74\x80\x30\x52\xf6\x14\xff\xd2\x75\x3e\xe9\x1b\x87\x32\x1a\xfb\x85\xe6\xfb\x71": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x95\xf3\xf0\x63\x3d\xbb\xe5\x72\xa0\x7e\x4c\x91\x08\xf5\x44\xff\xe9\xfa\x9a\x5f\xb7\x39\x3a\x82\x76\xbb\x33\x8d\x29\x67\xfc\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\x95\xfd\x87\x0a\xf4\x85\x1b\x95\xcf\x67\x0f\xa3\x5b\xfe\x5d\xb8\x6a\x19\x6a\x0f\x46\x69\xdc\x52\x39\x03\xec\xfb\xb4\xd7\x41\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x96\x30\xc3\x3c\x8f\x1a\xcc\x98\xbc\x39\x03\x9a\x47\x7a\xae\x2d\xcb\x52\x11\x80\x14\x3c\xd6\x97\xb8\xda\x19\x41\xc7\xde\x9e\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x96\x3b\x3b\xb1\x97\xaf\xd3\x0c\xac\xf2\x58\xa8\x0c\x02\x44\x28\xd2\x2a\xd9\xe9\x8f\x2f\x72\xb3\xd0\xfb\x99\x1e\xf4\x87\x81\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x96\x3b\x47\x02\x63\x2f\xed\x0a\xd9\x1c\x18\xd0\x09\x5c\x2f\x5d\xc6\x37\x3f\xd8\x75\xee\xb7\x76\x6b\x3b\xe5\x1e\xc1\x2f\x1e\x2d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x96\x3b\x5f\xf7\x81\x07\x34\x40\xc6\xd4\x9a\x32\x5a\x02\xa3\x24\x66\x93\x04\x0d\x32\x3c\xe0\x62\x1a\x4e\xfc\xf7\x57\x45\x12\x53": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221
    b"\x96\x5d\x8e\x66\xa9\x45\x82\x61\xeb\xcc\x36\xd1\x65\x92\x39\xa4\xf2\x9e\x27\xdc\x57\xb0\x29\x32\x5a\x50\x47\x31\xf4\xaf\x32\xd8": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.221",
        ),
    ],
    # Hash appears in versions: v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x96\x90\x16\x01\xfb\xb6\x8d\xad\x97\xd2\x85\xe7\xba\x32\x99\xfe\x6a\xb9\x5c\x5b\xe5\x1d\xd0\x6d\x80\xd0\xcc\x93\xea\xc7\x00\x6b": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: main
    b"\x96\xb7\xae\xf4\xe6\x15\x32\x44\xd8\x2e\x02\x75\x5f\x2a\x07\xd5\x57\x93\x79\xea\xd7\xb5\x28\x08\xd7\xfa\x62\x4a\xd0\x5d\x7f\xdc": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/AddressStringUtil.sol",
            name="AddressStringUtil",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68
    b"\x96\xbe\x46\x88\xd0\xf5\x47\x11\x86\x41\xc9\x69\x5d\x6d\xea\xdf\x91\x4e\xe4\xb3\xaa\x4c\xd7\xa2\x5b\xfb\x15\xe2\xb9\x47\x56\xde": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x96\xd3\x3d\x14\x87\x35\x8e\x6d\x96\xd6\xe9\x92\x13\x5a\x29\x1a\x46\xc2\x85\xcb\x3e\x8f\x9d\xd9\x16\x32\xac\x8c\x96\x7c\x37\xd3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x97\x0c\x9b\x31\xf0\x77\xd5\x67\x56\xe8\x8c\xe5\x4f\xd2\x8d\x02\xd4\x5a\x9f\xbc\xbc\x7a\x20\x10\x2b\x0d\xb9\x71\x18\x1d\x91\x56": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x97\x18\x1d\x0f\x3c\x8b\xb7\x3f\x7c\x86\x5e\x98\xd0\x89\xb7\x42\x6d\x0f\x5b\xf2\x96\xa0\x10\x26\xc5\x18\x94\x75\x1e\x1e\x9f\x54": [
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.116
    b"\x97\x37\xff\xa6\x11\x26\x91\x10\xe1\x1a\x20\x23\x0e\x49\x85\xfc\x78\x2d\x4c\xdf\xc0\x01\x1f\xd1\x6e\xf9\xa2\xd0\x53\xd5\x84\x49": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.116",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x97\x44\x42\xdb\xef\x51\x75\xad\x61\x83\x31\xb4\x40\xeb\x7b\x90\x16\x48\x31\xf4\x45\x88\xa3\xbf\x29\x4a\x5d\x61\x4b\xa7\x91\x3e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x97\x5f\x08\x64\xdc\x50\xe2\xdb\xc1\x55\x79\x02\x76\xed\xdc\x78\x1b\xca\xb3\xc0\xc9\xc5\xf0\x34\xf7\xec\x51\xc1\x06\x9a\xf5\xf6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x97\x62\xb7\x71\x59\xd1\x46\x8d\x31\x74\x6c\x7c\x6d\x12\x66\xdf\x89\x99\xeb\xb2\xdf\x7d\x70\x61\xd4\xd6\xb1\xe7\xc5\xdb\x24\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x97\x75\xec\x27\x6b\x4c\x64\x58\x6b\x75\x09\x70\x3f\x7c\xb9\x70\x27\xd4\x92\x3f\x6c\x73\x8a\x5e\xd5\xe4\xd4\xe0\x4c\x88\xc9\x8c": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/Permit2Forwarder.sol",
            name="Permit2Forwarder",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.7.0-rc.0
    b"\x97\x7a\xb9\x95\xed\xa6\x79\x90\xd5\x96\x2e\x25\xfb\xc5\x51\xb8\x75\x44\x25\x18\x37\xb6\x73\xfe\x54\x4b\xc4\xbd\x84\x5e\xb2\xc8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20TokenizedVaultUpgradeable.sol",
            name="ERC20TokenizedVaultUpgradeable",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x97\x89\xe4\xd7\x16\xe4\x92\x5d\x20\x4c\x84\x0b\xd4\x1e\x9d\x08\xfa\xdd\x6d\xaf\xc8\xea\x46\xcf\x68\xb6\x4e\xde\xbe\x8b\x9b\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x97\x8c\x3e\x8f\x51\x06\x53\xe9\x89\xe7\x0c\x57\x6e\xa0\x60\x5c\x1c\x16\x6e\xff\x2f\x48\x44\x1f\xea\x2b\x07\x8f\x3d\x40\xaa\x19": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\x97\xc4\xf9\xeb\x92\x6e\xad\xe6\x2b\x9e\x99\x19\xcd\xcf\xcc\x40\x2a\x25\x3c\x6f\x31\xe9\xa6\xea\x9a\xc6\x3b\x50\xcd\x05\x7d\x1c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\x97\xd7\x19\x99\xce\x99\xf1\x07\xa4\xbc\x74\x15\x55\xc9\xbe\x17\xa5\xce\xde\x78\x76\x8e\x45\x95\xf6\x39\x2f\x24\x3b\x96\x3d\x82": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x97\xe6\x32\x8f\x4d\x29\xa5\x21\x08\xa4\x51\x57\xf8\x2b\x2e\xde\x33\x72\xf2\xba\x94\x5a\x31\x85\x8b\x65\x2a\xe8\x3b\xa7\xc9\xf0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\x98\x0a\xfd\x29\xc8\x1f\xbe\x86\x3b\xf1\x3e\x2a\xa4\x8c\xd1\x86\x8c\xdc\xfd\xe9\xa8\x19\xa3\xca\x44\x15\xff\x25\x5c\xb9\xdc\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.112
    b"\x98\x19\x39\x7c\x2a\xc4\x31\x41\x01\x91\xd7\xf5\x98\x93\x05\x7b\x15\xe6\x7a\x67\x42\xaf\x9b\x5f\x47\x8e\x69\x89\x8c\x4b\xcc\x56": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.112",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\x98\x28\x66\x17\x5d\x68\x62\x9d\x64\x88\xa8\x2a\x2f\x3c\x09\xa4\x83\x2a\xe2\xa1\xd4\xe6\x2c\x32\xaf\x44\xaa\x56\xdf\x0d\xe7\x8d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.207
    b"\x98\x2b\xda\x19\x54\x47\x43\xd3\xff\xe0\xc9\x7e\xd5\xe9\xee\x33\xd1\x47\xa6\x6d\x83\xb6\x20\x3f\xe5\x93\x38\xb0\x59\xee\x93\x85": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.207",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x98\x37\x74\xb0\x24\x9b\x4a\x27\x59\x5c\xbf\xba\xb9\xa8\x5e\x62\xbf\x87\x19\x17\xed\xd6\x12\x1b\xdc\x0a\xf2\xcc\xb6\x9f\xb6\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x98\x3b\x50\xaa\xaf\x20\x14\x13\xaf\x7e\x76\xdf\xa0\x50\xdc\x20\x3a\xe3\x01\x9f\x61\xee\x04\xde\xea\x96\x09\x0a\xfa\x00\x43\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x98\x49\xd2\xf9\x5e\x01\xba\x9c\x95\x6e\xa0\x01\xa8\xb6\x13\xb8\x8e\xfd\xcf\x0b\x3a\xbb\x2f\xf6\x18\x22\x20\x57\x24\x12\xa7\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/AuthorityUtils.sol",
            name="AuthorityUtils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.7.0, v6.8.0
    b"\x98\x8a\xf3\x82\xd4\x39\x1a\x5d\x80\xd1\x67\x41\x2d\x3f\xd1\xd6\x91\xb1\x21\xe3\x0f\xa7\xfa\x59\xc6\x9c\x97\x33\x2b\x9f\x57\xb0": [
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Auth",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168
    b"\x98\x8e\xd0\x06\x45\xfb\xc1\x7c\xc2\xcc\x39\x9c\xdf\xe4\xcc\x5c\x59\x0b\x47\x53\xc5\x65\x7c\xcf\x72\xf5\x52\x4c\xb1\x70\xd7\xb6": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.168",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\x98\xc6\xab\x5b\xea\x2a\x4e\x9d\x39\x30\x79\xed\xd3\x9d\xe4\x53\x89\x7b\x18\x50\x3a\x2e\x17\xdd\x24\xe5\x8f\x68\x90\x38\x11\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: main
    b"\x98\xc8\x9a\x97\xea\xce\xf5\x83\xfc\x04\x83\x1c\xb8\xf6\xf7\x5a\x05\x5d\xbe\xa5\x2b\x40\xd5\xb9\xc8\x07\x21\x59\xc8\xc3\x2f\x4f": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/Actions.sol",
            name="Actions",
            version="main",
        ),
    ],
    # Hash appears in versions: main
    b"\x98\xe7\x06\xd0\xc4\x93\x86\xb7\x08\x61\x68\xe4\x51\x9b\x6e\x12\x24\x32\x31\x32\x67\xd5\xbc\xe1\x60\x4f\xf5\xdf\x6d\x33\x0a\x4e": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/NativeWrapper.sol",
            name="NativeWrapper",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\x98\xf9\xe0\xe4\x95\x31\xe1\x7e\x7f\xef\xd2\x3e\xe8\xcc\xee\x82\x96\xfd\x95\x74\x43\x56\x42\xa3\x41\x38\xbb\xdd\x6f\x2f\xb1\x97": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.2.0-rc.0, v4.3.0-rc.0
    b"\x99\x11\x66\xbd\x35\x22\x0e\x31\x95\x6f\x4d\xeb\x8d\x42\x33\xf1\xb0\x00\xc4\x99\x55\xe9\x46\xef\xb0\x9c\xd8\xea\x0f\xba\xd2\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x99\x21\x09\x90\xb4\x4f\x23\x0b\xe0\x13\xdf\x9b\xb5\x00\x68\xa0\xaa\x65\x74\x70\xa8\xd4\x38\xfa\x7c\x54\x31\x1c\xb5\xe2\xd5\x2b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x99\x2b\x9d\x0d\xd8\xbd\x08\xc4\x32\x5b\xb3\x84\xf8\x3a\xb9\x0b\x2b\x1d\xb2\xc9\x20\x93\xa7\x04\xfe\x54\x72\x45\xf2\x28\x49\xef": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\x99\x2f\x43\x80\xd2\x6e\x1b\x32\xc9\xba\x65\x08\x95\xab\x4c\x88\x87\xaa\x38\x12\x5d\xab\xe2\xc8\x36\xef\xf4\x3e\xe9\xe9\x7a\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107
    b"\x99\x43\xeb\x23\x15\xe5\xa2\x9f\xbe\x71\x8f\xaf\x37\x51\x58\x92\x40\x29\x8a\xaf\x9c\xae\x26\xe4\x21\xca\x23\xbd\xd2\x1b\xa6\xa9": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.107",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x99\x50\x7c\xe5\xee\x27\xe1\x38\x77\xa5\x80\x84\xe3\x8a\x34\xe2\x6e\x3c\xf9\xc8\xc8\xb8\x74\x28\xa8\x36\xcb\xec\x62\xd3\x59\xff": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: main
    b"\x99\x54\x58\xfd\x11\x90\x83\x59\x91\x39\x63\x9c\xbc\x2a\x7f\xc8\xa9\x9c\x9c\xa8\xf6\x1a\x27\x62\x1c\x5e\x0e\x32\xa3\x56\x16\x0b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/utils/HookMiner.sol",
            name="HookMiner",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\x99\x58\x3e\x3c\x33\x3e\xad\xbd\x06\x5b\x7b\x19\x54\x42\x28\xe1\xf0\x3c\x6a\x6e\x61\x39\xe2\xd3\xcb\x3f\x02\x61\x77\x72\xc7\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SignedSafeMath.sol",
            name="SignedSafeMath",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.3.0
    b"\x99\xa7\x26\xa7\x94\x26\xbe\x75\xaa\x37\xad\x68\xf6\x3d\xe7\x0e\x26\x06\x7b\xaf\x45\x79\xd8\x2d\xf1\xbb\x59\x88\xaa\xc6\x62\xaf": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.3.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0
    b"\x99\xcb\xa9\x67\xb1\xf2\xa0\xa7\x28\xfb\xd0\x54\x27\x9c\x09\x33\xc6\x1d\xa1\x3f\x9b\xa2\x9f\xec\x33\x5a\xe4\x50\x09\xbf\x0d\x58": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x9a\x50\x5d\x01\x5b\xbd\xba\x6c\x20\x25\xef\xb1\xa7\x3d\xb1\x7f\xca\x99\x62\x15\x55\x42\xa8\x04\x6b\x5c\xee\x26\x86\x53\xd7\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x9a\x5a\x1e\xca\x09\x77\xf8\xdb\x22\xc6\xa3\x62\xc4\xfe\x45\xff\x08\x83\xbe\x1f\x81\xe3\x51\x14\xb6\xca\x81\x32\xce\x7f\xf5\xca": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9a\x75\x63\x64\x69\x0a\x50\xdd\xa2\xa0\x50\x83\xa3\xef\xd9\x03\x1b\xd6\xdf\x5a\xda\x43\xa0\x22\xb4\x07\x1b\xbc\x34\x61\xb5\x5b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x9a\x82\x48\x4c\xee\x72\xeb\xf0\xe4\xf2\xd3\x3e\xa1\x28\x9d\x38\x98\xc5\xbb\xa4\x2c\x3e\x57\xcd\x07\x60\x82\xc9\x2e\x36\xf7\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9a\x88\x05\xdc\x7e\xf6\x01\xac\x88\x89\x26\x88\x03\x8d\xb8\xae\x96\x42\x25\x29\x26\xdf\x14\x7d\xba\x8e\xfe\x9e\x60\x9d\xc1\x9c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165.sol",
            name="ERC165",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x9a\x89\xea\x65\x94\xea\xe5\x9d\x51\xe1\x8d\xaa\xc3\x53\x80\xaf\x01\x2f\x49\xb8\x60\x89\x5c\x79\x33\xa7\x78\x4a\x0c\xe1\x86\xde": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\x9a\xcd\x2d\x5c\x5e\xa0\xb3\xd5\xf4\x0d\xfb\xad\xe5\x8c\x9f\xf2\x25\xa6\xa1\xd9\x84\xf0\xba\x9c\x80\x13\xa9\x08\x3c\xf4\x5b\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x9a\xd8\x0c\xbd\x39\x79\xa5\x4a\xd0\x19\x09\xee\xc0\x6e\xce\x4b\xc1\x9c\x25\xe4\x4d\x42\xc7\x27\xcf\xd4\xaf\x46\xb9\xe1\xe2\xed": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/draft-ERC6909Upgradeable.sol",
            name="ERC6909Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/draft-ERC6909Upgradeable.sol",
            name="ERC6909Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9a\xe5\x87\x57\x3e\xc9\x92\xde\xeb\x02\x27\x88\x45\xa5\x8b\x09\xa5\x4e\xd2\x73\x33\x1b\x6e\x51\xd6\x0c\xec\xa6\xca\x90\xf1\x78": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1.sol",
            name="CrossChainEnabledArbitrumL1",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.54, v0.0.55
    b"\x9a\xef\x89\xed\xec\xf4\xf9\x54\x4d\x45\x90\xb0\x4f\x84\xe2\x2c\x53\xab\xcd\x81\x32\xaf\x24\x0b\xee\xd3\xcb\x71\x94\xd6\x7c\xac": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\x9a\xf5\x29\x40\x53\x71\xce\x24\xf2\xad\x03\x73\x30\xfd\x97\x25\xcf\x19\x8a\xd6\x90\x21\x4e\xa2\x19\x4f\x8d\xd4\x14\x2d\x46\x8f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\x9a\xfa\x54\x47\x02\x70\x79\xff\x76\x2a\x1e\x2c\x88\xa3\x47\x39\x5e\x32\x74\x13\xce\xd4\xa9\x93\x34\x83\x94\x9f\xab\x42\x09\x39": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThreshold.sol",
            name="GovernorProposalThreshold",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x9b\x05\x56\x5f\x16\x68\xd4\xfe\x93\x59\x0b\x12\x44\x18\x0b\xb2\x8e\x93\x35\x98\x49\x0d\xab\x83\x9a\x45\x5b\xa6\xbc\xe8\xfe\xca": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9b\x0a\x2a\x26\x0c\x26\xa7\x2c\xbe\x7d\x37\xbc\xb0\xb2\x94\x42\xde\x82\x89\x20\xbf\xcd\x23\x64\x64\x78\xea\xb3\x05\x38\x11\xaa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccountExecute",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccountExecute",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccountExecute",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccountExecute",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAccountExecute",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.292
    b"\x9b\x4e\xf6\x5c\x34\x5d\x75\xae\x4e\x62\x9f\x05\x46\x2b\xbf\xdd\xd6\x02\x44\xf9\xf4\x52\x62\xbc\x83\x6b\x7e\x77\xb2\x5a\x0d\x10": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x9b\x55\xaf\xd0\x57\x88\x9b\xd6\x7e\x1f\x4c\x57\x46\x6c\xb1\x50\x15\x53\xc9\x32\x8f\x9b\x79\xee\x41\x22\x60\xb5\x60\xc1\x46\xaa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.1.0
    b"\x9b\x83\x8a\x41\xb6\x53\xc6\x5b\xdb\xd4\x22\xa3\x8e\x87\xd6\xa6\x41\x3d\x6c\xb6\xd1\x24\x23\x97\x0d\x85\x2c\xff\x19\x7d\x36\x90": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.1.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2
    b"\x9b\x9e\xe0\xcb\xe3\x90\x37\x93\xff\x29\x7d\xf6\xc3\x47\xff\xdc\xb3\xd7\x0c\xaa\x00\xb6\xb3\x03\xb6\xd5\x0a\x2a\x85\x29\xf1\x50": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.0-rc.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\x9b\xba\xe1\xd8\x71\x78\x1d\xd3\x10\x5a\x26\x14\x26\xe1\xe3\x2c\x39\x26\x61\xf9\xeb\x81\x0a\x21\x44\x52\xb6\xeb\x34\x34\xdf\x01": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\x9b\xe2\x62\xb1\x3a\xe1\xd4\xca\xa3\xd1\x8d\x2d\xa2\x84\x47\xc6\x30\xc1\x10\x81\x06\xcc\x4a\x77\xd5\xba\x3b\x13\xaa\x2c\x1d\xcf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.67
    b"\x9b\xeb\xf3\x8f\xbd\xdf\xe4\x46\xd2\x6e\x33\x4e\x57\xca\xe0\x2a\x69\x0e\xbf\x2e\xdb\x27\xb4\xbd\x89\xf8\x36\xfe\xe7\x48\x38\x05": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.67",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9b\xf4\xad\x45\x4b\xfd\xf8\x5b\x21\x23\xa8\x48\xbe\x5c\xfb\x45\x81\x66\x8f\x1f\xf1\x80\x80\x1d\x28\xe5\x95\x14\xc6\xcc\x91\xbb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170
    b"\x9b\xfc\x51\xf4\x3e\x7f\x9b\x3c\x21\x40\xa9\xf6\x47\xce\x00\x51\xfe\xbd\x2f\x62\x3d\xfb\xaa\xda\x17\x88\x0f\x44\x40\xea\x2b\x31": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.170",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136
    b"\x9c\x04\xe6\x9d\xca\x47\x3f\x23\xe9\x76\x69\x52\x0b\x8a\x39\x16\xd1\x01\x4d\x12\x42\x8e\x81\xc2\xbe\x62\xd1\x50\x9e\xdd\x88\x74": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.136",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\x9c\x0e\xdd\x50\x26\xac\xf9\x2a\x50\x8c\x05\xf7\x80\xaf\xe0\xeb\x7b\x19\x19\x3d\xdc\x0e\x6d\xfb\xcf\xf2\x38\x57\x71\xc8\x01\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\x9c\x10\xa9\x71\x18\xad\xce\x7b\x0e\xe6\x0b\x79\x93\xf5\x2c\x0e\x1b\xa5\xd4\x8c\x45\x41\x60\x1e\x71\x89\x3c\xba\x6e\x79\x1f\x72": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9c\x2f\x0a\x36\x5d\xa0\xfc\x78\x0c\xef\xd3\xc3\x02\xa6\xbd\x75\x7e\x29\x02\x3f\x43\x6f\x5e\x4f\xac\x2d\x51\xb8\xdc\x7c\x39\x72": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMapsUpgradeable.sol",
            name="BitMapsUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9c\x44\x1c\x4f\x40\xd0\x3d\x09\x7a\x6c\x9e\x20\x65\x38\xe9\xd6\xa2\xe5\x0b\x8d\x48\x7f\x50\x46\x74\x3d\x8e\x22\x23\xb3\xa8\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2
    b"\x9c\x67\xde\xd2\xaf\x08\x45\xab\xdd\x81\x0a\x84\x87\x0e\x08\x7b\x18\x32\x2d\x8e\x88\x42\x64\x54\x2b\x0e\xc4\x5e\x5c\xac\xaa\xf7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\x9c\x6c\x28\xa1\xb8\xaa\xea\x90\x0b\x1f\x23\x07\x48\x3e\xa5\xf6\x94\x0b\xda\x93\x8e\x9b\x25\xcf\x88\x17\xf0\xfd\x19\x14\x38\x83": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\x9c\xa1\x86\xd3\x94\x9b\xe6\x4b\x3d\x93\xdd\xa2\x92\x05\xcd\xa2\xc4\x5e\x67\xb9\x1e\xe5\xb0\x8e\xff\xf4\x25\xfc\x75\xd8\x88\x8b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x9c\xa9\x63\x24\x10\x10\xfa\x59\x46\x05\x6a\xd6\xc8\x33\x88\xca\x08\xd0\xe6\x29\x2e\x33\x11\xc7\x17\x24\xff\xce\x33\x8c\xf2\xb2": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.99
    b"\x9c\xf5\x76\x1b\x32\x27\x28\xe9\x34\x2d\xa0\x90\x85\x9e\x90\x37\x72\xa0\x79\xc3\x69\x4a\xd7\xd6\x8c\x26\xd4\x7e\xbd\x82\xaf\xd7": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\x9c\xf7\x05\x8b\x28\xc1\x38\x9d\x26\x72\x8f\x6f\xaf\x54\x3e\x6b\x03\xa1\xd6\x32\xe6\x04\xf5\x2f\x73\x86\x3d\x51\x38\x5f\x6b\x98": [
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\x9c\xfa\xf6\x93\x68\x97\x1a\xe9\x0c\xbe\xde\x38\x31\xb5\xd2\xe6\x79\x14\xa9\xba\x60\xcd\xe9\x12\xf5\x21\x2c\x4f\x4a\x7c\xc6\x08": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\x9c\xfe\x9a\x0b\x3c\x4b\x41\x68\x6f\xd7\x9f\x25\x61\x69\x80\xaf\xfc\x04\xa8\x24\x8f\x5e\x08\x30\x7e\x51\x05\xe1\xa6\x35\xb3\xf8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\x9d\x20\x2c\x93\x18\x05\x8e\x8f\x2b\x3a\x75\xb9\x51\x78\x12\xea\x72\x21\x11\xd0\x88\xa0\x0b\xc0\x58\xf4\x46\x57\xa6\xa4\xfb\xbe": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/QuoterV2.sol",
            name="QuoterV2",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/QuoterV2.sol",
            name="QuoterV2",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/QuoterV2.sol",
            name="QuoterV2",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9d\x4a\x40\xc7\x71\x9f\x34\x3b\x73\x76\xda\xcf\x72\x4b\xcf\xa1\x94\x53\xcb\xb0\xbf\x2c\xc7\x1f\xfd\xb3\x21\x02\x15\x39\x15\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/Heap.sol",
            name="Heap",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9d\x6a\x98\xc7\xe4\x6b\x2f\x5d\x0d\xbe\xbc\x30\x17\x2f\x35\xbd\x34\x24\xd2\xf8\xe3\x88\x92\x10\xe6\x68\x29\xca\x64\x9c\x6b\x98": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155.sol",
            name="IERC1155",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9d\x6b\x90\x96\xaf\xb2\x31\x89\x1e\xfe\x43\x05\xed\x9d\x4e\xfe\x38\xa5\x8c\x47\x6b\x4c\x02\xbb\xd5\x08\x6c\xea\xa8\x96\x34\x6d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x9d\x83\xb4\x56\x4f\x42\xae\xfa\x17\x73\x9f\x29\x4a\x9d\xa6\x63\x2b\xf4\x64\xa1\xfe\x17\xea\x68\x11\x67\x2a\x70\x21\x42\xe9\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9d\x9b\xf0\x1f\x45\x29\x69\x86\x9f\x2c\xd2\x19\xb8\x4f\x52\xe6\xbd\xa6\xc0\xd2\x44\x45\xe5\x4c\x9f\xec\x5a\x45\xb4\x64\x35\x05": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64.sol",
            name="Base64",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\x9d\xec\x61\xe8\x13\x9c\xf3\xe0\x12\xe6\x05\x40\x72\xe7\x54\xad\xbd\xe7\x50\x8a\x76\x68\x9a\x49\x1a\xa1\x37\x35\x68\xe9\xa8\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\x9d\xf7\x11\x76\x53\x07\x16\x0d\x48\xbe\x53\x73\x4a\x99\xc0\x6e\xb4\x34\xc4\x0e\x07\xb4\x6e\x87\x7d\xda\x80\x56\x37\x80\xa2\xc9": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: main
    b"\x9e\x02\x40\xd0\x6b\xba\x41\x80\xe3\x66\x66\xb0\x90\xc7\x86\xbb\x13\xe5\xae\xe0\x21\x5e\xc4\x49\x83\x72\x02\x98\x73\x41\xe5\x29": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/Notifier.sol",
            name="Notifier",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\x9e\x0c\x26\x75\x57\x9c\x61\xb5\x6c\xcd\xc2\x9c\x0c\x91\x10\x5f\x14\x2d\xaa\xe5\x28\xbd\xac\xfb\xf9\xc8\x66\x70\x78\x18\x1d\x3b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashBorrowerUpgradeable.sol",
            name="IERC3156FlashBorrowerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9e\x0e\x21\x1c\x7f\x1b\x5e\x3c\xac\xa1\x9a\x5a\x33\x3c\x1a\x22\x67\xc8\x2b\x70\x29\x0b\xdc\xd1\x68\x68\xaa\x3c\x78\xa4\x20\x3b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\x9e\x10\x35\x0b\x59\xd6\x15\xf2\x18\x66\xde\x2f\x97\xb4\xc4\x94\x14\x61\x71\x96\x68\x2b\x7c\x0a\x09\x79\x3f\xa7\xa5\xbf\xcf\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x9e\x14\xaa\xb5\x60\xe2\xc9\xac\xc3\x1b\xdf\x97\xc0\x61\x16\x86\x6b\x4b\x18\xd4\x58\x39\x57\x04\xad\xb6\xaf\x4f\xad\x29\xec\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\x9e\x2d\x0c\x81\xb7\xfe\xa4\xc3\x24\x15\x17\x4b\xcc\xa4\xec\x61\xf7\x99\xcb\x29\x1b\xe3\x4c\xe2\x08\x78\xe0\xb5\x24\x64\xf1\xff": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\x9e\x43\x59\xed\xd5\x2c\x41\x78\x62\x8a\x7a\x45\xf0\x87\xa2\x9f\xb7\x97\x18\x95\xa0\x4f\x6d\x49\x8e\xa0\x3c\x64\xfa\x50\xb8\xd7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.1.15, v0.1.16
    b"\x9e\x58\x06\xfe\x0d\x66\xdb\x5b\xcc\xf2\x06\xad\x8b\xb6\x8a\x59\x0e\x43\x44\xf5\x26\x3e\x7f\xd5\xf8\x9e\xe3\x8c\x91\x8a\x5d\x1c": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\x9e\x7b\x42\xb5\x75\xad\x7f\xed\x16\x99\x4d\x7c\xa6\x53\xfb\x29\x06\xf0\xb2\x78\xf2\x2c\x2d\x10\xb4\x7d\xfc\x45\x40\xa0\x7c\x65": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9e\xae\x98\xc3\xc7\xb9\xa4\x1d\x7c\x33\x98\x2a\xf8\x4c\x36\x8c\x4a\x0d\x11\x4f\x4e\x31\x65\x66\x1a\xb3\x43\xcb\x94\x4b\x52\x2b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2Upgradeable.sol",
            name="CrossChainEnabledArbitrumL2Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\x9e\xcd\xb0\x3b\x9d\xd0\x05\xe8\x2e\x46\xc7\x3e\xd8\x5d\x62\xd3\xad\xcc\xbd\x28\x09\x86\xff\x52\xe1\x3b\x94\xf8\x27\xdf\xb5\x51": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\x9e\xd5\x0f\x18\x02\x93\x4c\x1b\x7e\x8d\x2a\x2d\x9d\x0f\x29\x27\x4f\xef\x44\x77\x06\xf6\x1e\x19\x26\xd9\x64\x34\xdc\x89\xc3\x45": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\x9e\xec\xa7\xa4\xee\x77\x0a\x34\x51\x6b\xbf\xea\x5c\x68\xc6\xc3\xac\x18\x7d\xda\xc8\xef\x67\xb1\xcc\xc8\xe1\x60\xae\x4a\x46\x1f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\x9e\xf5\xff\x4a\x2a\x4e\xff\x42\xd1\xd4\x1a\xb0\x27\x01\xcb\xc0\xd2\x5c\xa2\x40\xb4\x6d\x71\xc6\x4d\xdf\x3b\x61\x8f\x15\x97\x13": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20SnapshotUpgradeable.sol",
            name="ERC20SnapshotUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\x9f\x11\x1c\x14\xe1\x9c\xaf\x93\xf8\x63\x95\xb4\xc2\x91\xe6\xbc\xf1\x53\x03\xbe\x2f\x1e\x8e\xfc\xcd\xa3\xc8\x2f\x70\xc6\x71\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9f\x15\x46\x62\xdf\xeb\x00\x13\xe9\xc5\x50\x0a\x1e\xdc\xb0\xa6\x59\x65\xb7\xe5\x75\x3c\x86\x3b\x31\x5f\x35\x0a\x06\x77\x22\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2309.sol",
            name="IERC2309",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\x9f\x29\x54\xcf\xa4\xc6\xa8\xc4\x39\x65\xff\xd3\xb3\x0a\xa8\xd0\xbd\xfc\x6d\xc5\xbd\x17\xfd\x76\x63\x0e\xdc\x7c\x5a\xad\xfa\x2e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\x9f\x90\x40\x46\xdf\x15\xd9\x85\x7c\xa3\x56\x73\x0e\xac\x6d\xfa\x46\x21\x97\x31\xf2\x7d\xf9\xf5\x0e\x32\xdb\x1e\x2d\x0c\x92\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\x9f\xe0\x97\xdb\x34\x98\x4a\x86\x06\x5d\xa8\xbb\x44\x5b\x33\x7f\x94\xdd\x72\xda\x76\x4a\x30\x26\xa5\xfc\xe7\x30\x6f\xb6\x40\xd3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xa0\x27\xe3\x9d\x93\x67\x67\x1f\xed\x6f\xd0\x9a\xbb\x1a\x22\x4d\xd7\x2f\x97\x27\xab\x2d\x9d\x3b\xcf\x94\x49\x18\x21\xef\x9e\xe6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1
    b"\xa0\x28\x42\x1f\x28\x15\x97\x91\x00\xc2\x63\x9f\xc3\xdc\x62\x3d\xb7\x71\xbb\xc2\x7a\x30\xa1\x16\x6e\xcd\xa0\x06\x37\x23\x18\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.1",
        ),
    ],
    # Hash appears in versions: v0.0.122
    b"\xa0\x51\x67\xfd\xfb\x6a\xe1\x58\x99\x18\xc0\xa4\x0c\x7f\xf8\xef\x83\x94\xb5\x1b\x67\x06\x94\xe6\x7d\x43\x84\xa1\x48\xc8\xb9\xa4": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.122",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\xa0\x52\xbf\x87\x0f\x43\x27\x28\xdb\xb6\x9b\x1a\xa6\xc5\x96\xd7\x3c\x99\x0a\x45\x24\x8a\x75\x74\xf6\x93\xd3\xa5\x84\x00\xeb\xe9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xa0\x56\x0a\xed\x4f\x71\xa7\xe1\x1a\x4c\xbf\x95\xd0\xee\xb1\xa4\xff\xab\xe0\x65\xb2\x8d\x58\x27\xef\xcf\x6b\xe9\x58\x20\xe2\x90": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xa0\x72\x6a\xf2\xa5\x29\x48\xf6\xdb\xdf\x99\x05\x30\xd1\x62\xfd\xf5\x3c\x61\x48\xb9\xb2\x67\x4d\xdc\xa1\xf6\x8d\x61\xe2\x01\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2
    b"\xa0\x78\xd8\x9b\xf0\x33\x89\x7c\x1e\x2b\x13\xe7\xef\x97\xeb\x7d\x3e\x57\x76\x57\x5a\x17\x61\x3f\x30\x1a\x32\x50\xd6\xe4\x44\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.7.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xa0\x84\xd2\x22\xac\x31\xed\x13\xc3\x8a\x6d\x7e\x5d\x79\xda\x7d\x0b\x76\x7b\x6d\xa0\x29\x31\x9f\xea\x9f\x0d\x42\xd9\xf5\xbd\x78": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xa0\x86\x7d\x1d\xfc\xec\x19\x49\xad\xe2\x73\x1d\xb9\xda\xa5\xc2\x65\x74\x9f\x6c\x7b\xab\x1f\xbe\x73\xc4\xb3\x29\x68\x39\x3f\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1
    b"\xa0\x8e\x3b\x5c\x00\x35\xeb\x77\xea\xba\xc8\xba\xc0\x4d\x80\x15\xb8\x2d\x7c\x2b\x98\x37\x04\x89\x29\x48\xe6\x62\x9e\x2b\x46\xcc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL2Upgradeable.sol",
            name="LibArbitrumL2Upgradeable",
            version="v4.7.1",
        ),
    ],
    # Hash appears in versions: v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74
    b"\xa0\x95\x45\x19\xb3\x99\x87\x3b\x88\x31\xda\x7d\xcc\x23\x81\xfa\x30\x2c\xc8\xe2\xe1\x24\x33\x6d\x2c\xa2\x79\x97\x3c\x55\x15\x5e": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.74",
        ),
    ],
    # Hash appears in versions: v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xa0\xac\xdb\x17\x29\xb2\xba\x1a\x25\xa8\xd6\x08\xba\x11\x22\xfc\xcb\x64\xc8\x14\x5d\x8d\xb3\x5f\x0a\xab\xe5\xa0\xaf\xc8\x64\xe8": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xa0\xbb\xbb\x36\xb7\x0c\x2d\xb1\xb9\x96\xa4\xf4\x28\x83\x6c\x84\xb2\x1c\xb6\xf9\xba\x65\x2f\x02\xc2\x74\x66\xdf\x44\x69\x76\x4b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipientUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xa0\xf7\x69\x07\xd5\x08\x74\xa3\x0a\x23\x7c\x8d\x94\x59\xac\xec\xe1\x02\xc8\xbe\x06\x2e\xe6\x27\x26\x4e\x23\xa4\x10\x79\x0f\x4c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa1\x01\xd4\x05\xa8\x59\x7e\x70\x5f\x07\xdb\xdb\x7e\x26\x19\x6f\x43\xec\x6d\xb7\x18\x1a\x66\x07\x3d\xa9\x27\x25\x8d\x0f\x78\xa1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xa1\x06\x04\xe5\xaa\xe7\x12\x5a\x10\xbe\x59\x13\x5f\x0a\x81\x79\x69\xcb\x22\xfe\x05\xdf\xc1\x96\xce\x03\x0b\x33\x07\x1c\x7a\x7a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPaymentUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xa1\x22\xf2\x7e\x4a\x40\xf3\xf9\xdc\x63\x9e\xf6\xf3\xc3\x97\x65\x93\xf0\x78\x2b\xe7\xc9\x92\x23\x49\x69\x34\x4d\xe9\xb0\x0a\x98": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVestingUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa1\x23\xb0\x26\x58\x4d\x51\xad\xa1\x22\xe4\x47\xed\x68\x38\x4d\x49\x66\x8a\x45\xb8\x4a\x16\xb2\x35\x18\xb3\x41\x0f\xf2\x11\xfc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7
    b"\xa1\x32\xf6\x44\x31\x74\xc5\xba\x27\xfc\x80\x3c\x9a\xa3\x0d\xf5\x35\x82\x09\xeb\x6f\xf2\x8e\x91\x27\xb7\xb0\x00\x12\xbe\x38\x70": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124
    b"\xa1\x5a\x14\x83\x71\x3b\x38\x84\x48\x7e\xf0\xb5\xd6\x8c\xad\xcd\x62\x1e\x3d\x05\x5b\xc4\x8d\x4e\xd0\xbb\x3c\x05\xd4\x07\xd9\x6e": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xa1\x5e\xb4\xee\x2b\x1b\x82\xab\x0e\x24\x81\xe8\x28\x5e\xcc\x6e\xfa\x00\xc5\xd3\x1e\xf1\xbe\x04\x30\x6c\xf8\xb3\xaa\xc5\x6e\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xa1\x68\x05\x2d\x4d\xc3\x27\xf7\x3c\xaa\x19\x11\xe6\x6d\xe9\x1c\x96\x53\xc2\x53\x40\x2c\x54\x13\x8c\xf1\x5e\x69\xf0\xd8\xbc\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721ReceiverUpgradeable.sol",
            name="IERC721ReceiverUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xa1\x6f\xca\x30\x3d\x09\xc8\x17\x80\xe1\xc7\xb1\x24\xe2\x67\x45\x7d\x51\x37\x43\x7e\xbb\xd2\xdf\x63\x0e\x96\x93\x0e\x26\xb8\x9e": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryImmutableState.sol",
            name="IPeripheryImmutable State",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryImmutableState.sol",
            name="IPeripheryImmutable State",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryImmutableState.sol",
            name="IPeripheryImmutable State",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryImmutableState.sol",
            name="IPeripheryImmutable State",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryImmutableState.sol",
            name="IPeripheryImmutable State",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.2.0-rc.0, v4.3.0-rc.0
    b"\xa1\x79\x9e\xa3\x23\x96\x40\xdd\x08\x81\xfe\xe1\xd4\xe1\x2d\x71\xb6\xe3\xe8\xd6\xd0\x8a\x48\x17\xd7\x37\xdd\xe3\x17\x16\x23\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xa1\xa5\x7e\xeb\xbd\x3e\xe9\x07\x9e\xf1\xbe\xd6\x00\xf2\xee\xb9\xad\x25\x44\xc6\xa6\x7d\x33\x94\xbf\xda\xc5\x8c\x23\xe3\x52\xfb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xa1\xb9\x75\x82\xd4\x44\x29\x84\x2e\x03\x43\x27\x8a\x3a\x6d\x4b\x37\xda\x5e\x23\xa0\x41\xaa\x27\xcd\x87\x08\x5a\xec\x91\xa9\x6d": [
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xa1\xbc\xed\x88\x94\xeb\x13\x0a\x97\xb8\x40\x70\xe6\x4b\x9f\x6e\x1c\x2a\x29\xe8\xca\x1c\x23\xc3\x1c\xcc\xf0\x9b\x98\x7f\xb9\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909Metadata",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909Metadata",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xa1\xc3\xa3\xff\x00\x5b\x9c\x6b\x57\xfb\x04\x44\x24\x4f\x58\xa3\xfb\x69\x91\xc0\x24\x3b\x7b\xbe\x2e\x8d\x81\xda\xe7\x86\xbe\xab": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSequentialProposalIdUpgradeable.sol",
            name="GovernorSequentialProposalIdUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSequentialProposalIdUpgradeable.sol",
            name="GovernorSequentialProposalIdUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xa2\x49\xf6\xbd\x41\xa6\x8c\x95\xc4\xf8\x4b\xa5\x9c\x4b\x61\x38\x18\x51\xee\xa3\xdb\xba\xbe\x24\x3a\xb9\xd1\xbc\x7f\x05\x86\xa8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29
    b"\xa2\x60\x96\xa0\x88\x2d\xa3\x49\xaa\x04\xf9\x25\x55\x89\xf9\x41\xa3\x15\xb4\x54\x8d\x7a\x43\x6c\x13\x14\xfe\x7c\x9b\x23\x54\xe4": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.29",
        ),
    ],
    # Hash appears in versions: v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151
    b"\xa2\x69\x47\x42\xd8\x5f\x5f\x6a\xcc\xfb\x2a\x3f\x4f\x88\x58\x47\xbf\xad\xfb\x50\xca\xd5\x49\xfc\x44\x80\x89\xdb\x31\xf4\x1d\x57": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.151",
        ),
    ],
    # Hash appears in versions: v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133
    b"\xa2\x7f\x24\x65\xa1\x82\x30\xf8\x87\x57\xae\x8f\xca\x09\xdc\x60\x8f\x95\xd7\x38\xbb\x82\xf9\x0f\x74\xa5\x07\x8b\x51\xe0\x2c\x42": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.133",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xa2\x8e\x3c\x2d\x46\x45\x71\x8f\x05\x6f\x69\xe2\x70\x61\x16\x29\xe6\x9e\xc4\xb9\x16\xee\xc7\x44\xd1\x95\xaf\x96\x85\x6c\xb1\xf0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa2\x97\x83\x1b\x97\xdb\x0d\x0c\xff\x43\x14\xde\x4a\x61\xfc\xfe\xed\x91\x06\xa5\x28\x48\x3f\x56\x1b\xed\x10\xcc\x3d\xea\x56\x85": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xa2\x9c\x86\x3b\xd5\x89\x45\xd0\x1d\xb9\xdc\x13\xe7\x24\x61\x8b\x59\x80\xac\x3d\x79\xd6\xea\xa0\x6a\xa2\xfe\xe5\x60\x86\x3a\xbd": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/SwapMath.sol",
            name="SwapMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xa2\xae\x2a\xa7\xd7\xe5\xe6\x85\x23\x99\x77\xdb\xb8\x42\x5b\xf9\xef\x25\xcc\x7c\x41\xf4\xd2\x6e\xaa\x5e\xcd\xef\xb8\x1a\x49\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721Holder",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v1.1.1, v1.2.0, v1.3.0
    b"\xa2\xcc\x14\x29\xe3\x58\x4e\x86\xcd\xf9\x2a\x8a\xa2\xbf\x1f\x75\xdb\x1b\x71\x79\x9e\x06\xea\xde\xd3\xf3\x89\x88\x15\xea\x1b\xd7": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTSVG.sol",
            name="NFTSVG",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTSVG.sol",
            name="NFTSVG",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTSVG.sol",
            name="NFTSVG",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220
    b"\xa2\xd7\x0c\x23\x9d\x37\xd3\x70\x95\x8b\xcd\x36\xb3\x41\xc5\x24\x83\xba\x5f\x63\xe5\x14\x7f\xf7\xa4\x48\xd4\xe7\xff\xfe\x2f\x22": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.220",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xa2\xe9\xeb\x37\x1a\x5c\x10\x4c\xe5\xc1\x2d\x7b\xb5\x60\xc2\x50\xc7\x53\xea\x0b\x2c\xd2\x6c\x1e\xc7\x93\x5e\x38\x29\xff\x99\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xa3\x13\xdd\x32\x75\x41\x00\xeb\x53\x68\x80\xbe\x9a\x1a\x56\x98\xdb\x10\x2c\xd4\xd0\x3b\x27\x64\x69\x0c\x2b\xdf\xf7\xf7\x64\xee": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ITickLens.sol",
            name="ITickLens",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ITickLens.sol",
            name="ITickLens",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ITickLens.sol",
            name="ITickLens",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ITickLens.sol",
            name="ITickLens",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/ITickLens.sol",
            name="ITickLens",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xa3\x13\xfb\x9d\xd1\xa3\x83\xa3\xad\x1e\xd3\xb2\x14\x3b\x52\xf8\x43\xaf\x50\x59\x39\xa6\xaf\x8a\xd0\xc3\x09\x19\x06\x40\x75\xa9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xa3\x1e\xd4\x39\x55\x5c\xb2\xbd\xc3\x9d\xd1\x99\x39\x14\xd9\xdd\x95\x8b\xed\xab\xd2\x4c\xad\x8c\x2c\x8a\xba\x4e\x05\xef\x60\x2b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xa3\x57\xd4\x45\xdb\x37\x75\x61\xac\x79\x83\x93\x5c\x70\x19\xa4\x69\x33\x28\x94\x62\x29\xb2\x5b\xbb\x75\xdc\x90\x94\x8f\xae\x53": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xa3\x7e\xaa\x6a\x3f\x5d\xff\x2a\xdd\x76\xf0\x53\xc1\xa0\xb9\x30\x64\x88\x48\xd9\x8a\xc7\xcb\x4f\x10\x1f\x6d\x47\xb0\x5e\xd5\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363SpenderUpgradeable.sol",
            name="IERC1363SpenderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xa3\xb2\x41\x28\x58\x83\x0b\x6c\x54\x48\x76\x44\x61\x6c\x89\x00\x67\xa4\x1c\xfd\xe8\x27\xda\xfe\x20\x7d\x59\x62\xee\x87\x28\x30": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xa3\xb9\x06\xd9\xf2\x2d\x0b\xd5\xec\xbd\x2f\xf4\x8e\xa6\xda\x6b\x20\xba\x73\xa6\x8f\x61\x82\x40\xbb\x2b\x9d\x87\x02\xba\xbb\x4f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xa3\xcf\xb6\xd5\x3c\x83\xb8\x1e\xd4\x32\xbb\x0f\x03\x2a\xa3\x61\x6c\x3b\x0a\x86\xd8\x72\x72\xbf\x96\x29\xb0\xd4\x9e\x1f\xa7\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\xa3\xef\x3f\x6f\x91\x9b\x74\x51\x8a\x6a\x72\x91\x7d\x6a\x26\xf5\x97\xe4\xee\x40\xba\xd9\x72\xd9\xd8\x0e\xf2\x51\x1a\x51\xde\x8e": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa3\xf7\x1a\x0c\xfa\x98\x35\x08\xd1\xb7\x61\x89\xce\xff\x66\x78\x8c\x78\xba\xaf\x1c\xfe\x94\x38\xc2\xfd\xf1\xb5\xbb\xbc\x99\xb1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesExtendedUpgradeable.sol",
            name="VotesExtendedUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesExtendedUpgradeable.sol",
            name="VotesExtendedUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesExtendedUpgradeable.sol",
            name="VotesExtendedUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesExtendedUpgradeable.sol",
            name="VotesExtendedUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesExtendedUpgradeable.sol",
            name="VotesExtendedUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xa4\x01\xf7\x96\xa8\xd7\x52\x53\xf9\xd5\x61\xf9\x51\x95\xf7\x4e\xd4\x64\x9a\xe8\xce\xac\xaa\x93\x53\x41\x54\xe6\xc3\x84\x79\xa7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xa4\x22\x5d\xb4\x8b\x72\x09\x6a\x29\x4b\xd0\xcc\xc4\x5f\xad\x5d\x7e\xff\x1f\xdc\x33\xa6\xf8\xd2\x03\x73\x63\x83\xce\x33\x05\x38": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xa4\x30\xc2\xe2\x2d\xc5\x37\xbd\x1f\xd8\x43\x81\x96\x13\xa6\xe1\xf7\xae\x0a\x2e\xa2\xa9\x38\x56\x61\xe2\xbe\x13\x00\x09\xa5\x9b": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\xa4\x3a\x4a\xe7\xaa\xa0\x34\x8c\x06\x55\xc9\x95\x57\x6c\x93\x9a\x56\xae\x6f\x83\x41\x32\xb7\x5f\x73\xc8\xb5\x21\x90\x02\x88\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xa4\x3b\xd6\xf0\xa1\x30\x83\x38\x44\x3d\xba\x03\x4c\x7c\x1a\x87\xcd\x5f\xda\x45\xff\x3a\x39\xb5\xc3\x4c\x5b\xed\x98\xac\x58\xd4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xa4\x47\x25\xb2\xcf\x04\x55\x6d\x86\x48\x50\x1f\x82\x45\x0c\xb3\x4d\x34\x09\x1f\x81\xfe\x4e\x89\x70\x48\x49\x71\x46\x06\x43\x05": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xa4\x50\x6c\x15\xbc\xde\xda\x17\x9b\x2e\x05\x3f\xff\xb5\xb4\xc4\x3d\xc1\x7c\x9c\x95\xdd\x12\x6f\x6c\xd1\xe6\xd9\xbd\x1f\xb3\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa4\x54\x14\x82\x4b\xa3\x14\x5e\xc5\xc6\x0a\x12\xce\xca\xfd\x50\xbe\x70\x70\x19\x7c\xef\x6c\x1a\x5f\x30\xd7\x26\xa1\x81\xb4\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xa4\x66\x4f\x44\x83\x42\x9d\x92\xdb\x11\x3a\x8a\xff\x3a\x0b\x25\x17\x00\xb4\x33\xa5\xf3\xfb\x4c\x9e\xd4\x8f\x00\xae\xf4\x0c\xc3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xa4\x71\x18\xaf\xb7\x4d\x30\x32\xb9\xba\x0f\x4a\xf3\x45\x8a\x3a\x6e\xae\x12\x96\xf9\x79\x82\x46\x7f\x9e\xf9\x22\xe3\xa5\xc2\xea": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.6.0, v4.6.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xa4\x7e\x0c\xee\xd6\x1f\x9e\x5e\x5e\xf3\x78\x12\xcc\x44\xb5\x06\x87\x7a\x24\x69\x6e\x73\xbe\x78\x96\x53\xce\x88\x2b\x0f\xbc\xeb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xa4\x91\x11\x3e\x16\xa9\x74\xf9\xcc\x8d\x90\xcc\xd6\x46\xed\x49\xb6\x05\x4a\xf7\xcf\xcf\xec\x0f\xd9\x88\x3a\x44\x85\x56\x54\xc2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.7.0-rc.0
    b"\xa4\xa0\xf0\xac\x15\x6d\xf7\xe2\xda\xfd\x16\xc4\xe1\xb2\x4a\x5c\x23\xd4\x6d\x51\xde\xc6\x0f\x22\xff\xc6\x71\x48\x86\x56\xf7\xee": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xa4\xb7\x8a\xf1\xff\xed\x8d\x9e\xfe\xa6\x15\xe5\x75\x4d\xbf\x8f\x10\x22\xce\xfc\x59\xe4\xc1\xfd\x45\xf3\xdb\xb2\xb2\xd3\x43\xfe": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/Hooks.sol",
            name="Hooks",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa5\x09\x4d\x97\x4c\x42\x60\x3e\xc0\xfc\xdf\x0c\xff\xb9\xea\x0c\x2b\xb8\x36\xcf\xae\xb6\xeb\x65\x3e\x35\xa7\x62\x2a\xd7\x23\xdd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xa5\x16\xd6\xd5\x11\x97\x1e\xed\x71\xd9\xda\x26\xfa\xb0\x3f\x41\x6a\xcf\xef\x4d\x43\xbf\x61\xa8\x12\xac\x05\x6e\xb0\x43\xfb\x0c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.291
    b"\xa5\x4c\x61\xf1\x10\x96\xc5\x06\x5f\x26\xdb\xa4\x2c\xfa\x1c\xe0\x23\x5e\xc7\xff\xfe\xcd\xae\xcb\x1c\x91\x3d\xea\xd8\xcb\x5d\x8e": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.291",
        ),
    ],
    # Hash appears in versions: v6.7.0, v6.8.0
    b"\xa5\x6d\x9a\x3e\x7e\xf9\xa0\x69\xc1\xe7\x58\x4d\x0c\xb7\x50\x86\x9e\x7d\x40\xa2\xec\x3f\x16\x90\x66\x78\xa2\x99\xc2\x19\x3b\xc2": [
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.82, v0.0.83
    b"\xa5\x73\x2f\xc4\xed\x66\x73\xc6\x78\x78\xab\x3f\x50\x78\x87\x49\x33\x33\x2b\x0e\xa6\x45\x37\xb2\x9c\x2d\xbe\x83\xc4\xeb\x7b\x17": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.83",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1
    b"\xa5\xc1\xa3\xde\x9c\x6a\x6f\x7e\x2e\xef\x66\xf5\x83\x5b\xff\xfd\x08\x60\xb3\xa8\x4a\xb9\x1d\xdf\x6a\xe9\x7a\x12\x66\x81\x9d\x27": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.1",
        ),
    ],
    # Hash appears in versions: v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200
    b"\xa5\xc7\x32\xd2\x55\xb2\x31\x8c\xfc\xaa\xb8\x04\x7b\x04\x6e\x4e\x42\xf9\x74\x21\x89\xfb\x44\xab\xee\x0f\x76\xa9\x27\xde\xf2\xa3": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.200",
        ),
    ],
    # Hash appears in versions: v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281
    b"\xa5\xfb\x3f\x23\x5b\xef\x37\xce\x70\x31\x42\x1c\xa2\x6b\x24\x1d\x61\x7b\x59\x61\x19\xb2\x38\x1c\x1a\x3b\x42\xd0\x2c\x86\xf4\x06": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa6\x2f\x81\x11\x13\x93\x42\xe1\x49\x1d\x56\x06\xeb\x9c\xdf\x3a\x8f\x60\x6d\x35\x57\x98\x8f\xb9\xec\x8a\xf8\x2d\xf0\xd6\xcd\xaf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa6\x37\x60\x6c\x10\x09\xc5\xd4\x87\xcf\x30\x5d\xc2\x35\x71\x22\x66\xe4\xef\x71\x33\x24\xc6\x11\x02\x83\xca\xc1\x63\x6b\x51\x93": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xa6\x3b\xb0\x6d\x88\xd0\x76\xd8\x27\xce\xc0\x56\x0d\xbc\x8b\x60\x41\x12\x8e\x5e\x04\xd7\xea\xb7\x7a\xc8\x3f\x2f\x92\xe2\x3a\xfa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.271
    b"\xa6\x3d\x9c\x28\xd0\x7b\x84\x51\x6b\x82\x9c\xe4\x33\x56\x57\x81\x66\x8f\x3b\x36\xae\xc7\x7e\xcf\xfb\x92\xb4\x1f\xda\xfe\xba\xc7": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.271",
        ),
    ],
    # Hash appears in versions: v4.9.5, v4.9.6
    b"\xa6\x3f\x0e\x05\x7b\x40\xef\x5e\x7e\x6a\x8b\x0b\xeb\xee\xfb\xb1\xf5\x3e\x2b\xca\xf3\xbe\x06\x4e\x80\x09\x51\x80\x2b\xf0\x8c\x66": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xa6\x4a\xee\xa6\x49\xe5\xfc\xa9\x88\xc0\x47\xf7\xd3\x35\x79\xc9\xce\x2b\x1c\x26\xd8\x06\xb9\x78\x30\x48\x6e\x3c\xc9\x50\x03\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xa6\x4c\x79\x01\x8b\x13\xab\xea\x4a\x5c\x04\x4e\x60\x66\x7b\xf5\x43\xde\x35\xc3\xcd\xb5\xb5\xd2\x71\xa5\xdc\x66\x95\x91\x8b\xb0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xa6\x78\x24\x76\xca\x47\x6b\x0f\xc1\x01\x45\xa2\x2b\xe4\x01\xa8\x5d\x98\x94\x13\x84\x1e\x8d\xcf\x19\x98\xc1\xdc\x51\xe7\x93\x89": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xa6\x81\xa5\x67\x96\x3d\x33\xd6\xf0\x45\x0a\xf6\x3a\x0e\xb4\x6a\xdc\x34\xe3\x1e\x17\xa4\x4f\xd9\xb4\x16\x99\xc7\x73\x48\x46\xe0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MessageHashUtils.sol",
            name="MessageHashUtils",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xa6\x91\x59\x23\xe9\x7f\xf7\xd2\xd7\xce\x12\x86\x4e\xf9\xd2\x34\xe0\xe7\xb0\x87\x6e\xdc\x03\xa3\x8a\xa5\xc1\x19\x6a\x3d\x61\x33": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xa6\x9a\x2e\x05\x3e\x4e\xea\xa0\x41\x50\xee\x4d\x3f\xf7\x40\x69\xd8\x6f\x21\x3d\x77\x0a\xb7\x92\xe3\x4a\x64\x13\xdd\x28\x77\x19": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xa6\xae\x12\x64\x5d\xc8\x5e\x45\x5d\x03\xc2\xf4\x2f\x80\x9f\xe5\x1e\x06\x6e\xf4\x27\x51\xf3\x17\xe6\x47\xe6\xa4\x91\x84\xc6\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xa6\xb7\x2b\x4a\x23\x53\xf5\x18\x34\x2d\x88\xd3\x27\x58\x77\x07\xcd\xf0\x25\x95\x3f\x6b\x82\xf1\xdb\xc1\x41\xa9\xe1\x8f\x63\xe2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xa6\xc1\x35\xa6\xd5\x50\x5a\x62\x5f\x6a\x8c\x45\x63\x80\x02\xc0\xd4\x39\xbf\x71\xab\x09\x99\xae\xf3\xbd\xb0\xe4\x94\x5a\xa1\x26": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/IProtocolFees.sol",
            name="IProtocolFees",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v5.3.0-rc.0
    b"\xa6\xd6\x56\x8b\x51\xe3\x99\xe3\x7e\x0e\x85\x4a\x27\x1d\xe2\x61\x93\xa4\x76\x84\xd5\x53\x40\xa2\x64\x11\x43\xa9\x49\xc0\x7c\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="ERC4337Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xa6\xd9\xa3\xe0\xf1\xe9\xdb\xa1\x93\x46\xc1\x1e\xfe\x6c\x10\x3d\x79\xce\x45\x67\x07\x3b\x91\x3a\xf4\x38\xff\xfd\x4a\x2d\xb5\x75": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xa6\xdc\x77\x45\xf7\x49\x61\x55\xf6\x28\x65\x51\xd7\xa1\x2d\xda\xd1\x1f\xbf\x7c\xda\x81\xb7\x24\x8a\x1b\x69\x7d\x82\x07\xa4\x40": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa6\xf0\x28\xeb\xb1\xfc\x68\x02\xa0\x4c\xd7\xa3\x35\x2e\x12\xb6\xb0\x05\x8f\x99\x4a\x81\x10\xf1\x13\x92\xd4\x8d\xd8\xa6\xb8\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xa6\xf3\x3e\x90\xf9\xfc\x34\x14\x45\x4e\x5c\x4e\x33\x33\x19\xee\x17\xf7\xb6\x6a\x15\x27\x23\x54\xe8\x03\x38\x87\x8a\x1c\xf3\x48": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xa6\xfb\x28\xc8\x33\xfc\x76\x46\xbb\xa0\x8f\x6b\x94\x14\x8a\x90\x6b\x09\x1e\xc4\x18\xab\xe4\x5c\xd9\xd0\x81\xf3\xbe\x2a\x50\x21": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70
    b"\xa6\xfd\xa8\xb7\x98\x06\x70\x32\x3a\xc8\x85\x5b\x7e\x7a\x62\x81\xc5\xbc\x1c\xdb\x00\x53\xe5\xe7\xcb\x05\x3b\x98\xe2\xe8\x68\xa9": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.70",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xa6\xfe\x05\x04\xaa\x40\x6a\xe7\x4a\xb4\xdb\xfa\xaf\x5d\x2b\x5a\x1a\x36\x50\xb5\xc4\x1b\x79\x8d\x92\x53\x13\xd4\x41\x71\x41\xf7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909ContentURI",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909ContentURI",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0-rc.0, v4.3.0-rc.0
    b"\xa7\x00\x6e\xe8\xfd\xc5\xb0\x66\x9b\x55\x02\xf6\xe9\x75\x49\x93\xd2\x7e\xe0\x2c\xf0\xe1\xbb\x80\xad\xdd\x03\xda\x49\x96\x53\x12": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xa7\x08\x91\xde\x02\xe9\xbd\xc0\x74\x01\xfc\xc1\xe0\xd4\xce\x3c\x65\x2b\xc0\x77\xf0\xaf\xbb\x62\xb5\x9d\x0d\x6b\x75\x07\xeb\xb2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xa7\x1d\x30\x51\xca\xe4\x06\xa3\xbb\x18\xca\xbe\xbe\xa0\xd3\x61\x24\x3e\xfb\xbc\xc6\xaf\x00\x28\x14\xf8\xec\x0a\x4b\x7a\xa2\xc7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Arrays.sol",
            name="Arrays",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xa7\x26\xfd\x83\xf4\x6b\x08\x60\xfa\x8c\x23\x41\x37\x1f\x95\x90\x56\x66\x4f\x50\x47\x9f\x12\x70\x22\x27\x43\xd5\x8d\x73\xd7\xe4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Receiver.sol",
            name="IERC1363Receiver",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa7\x3b\x2c\x0e\x49\x56\xd3\x1b\xd6\xe8\x8e\xf5\x18\xbb\x8e\x8c\x93\xd4\x6f\x36\xdf\x73\x03\xd4\x46\x97\xe3\x42\x4f\x62\xc5\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Nonces.sol",
            name="Nonces",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xa7\x55\xb0\xa2\x84\x98\xfc\xce\x6a\x80\xd5\x75\x07\xb0\x37\xf2\x89\xab\x20\x88\xc2\x6e\x18\xeb\x97\x93\x8a\x4b\xac\x11\xda\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xa7\x69\x99\x4e\xb9\x1d\xe3\x9b\x8b\x37\xd0\x01\x84\x7a\x25\x6c\xa2\x22\x5d\x98\x9b\x1e\x21\x5d\xb2\xef\xb0\xb2\xaa\x0d\x02\xd0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xa7\x76\x72\x46\x09\xcb\x36\x8e\x69\xf2\xec\x15\xc3\xfb\x64\xed\x4d\xf0\x6b\x00\x8d\x31\x32\xa0\xee\xf4\x69\xad\x1c\xf7\xb4\xf6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xa7\xc9\x54\x4b\x7d\x93\x15\xa1\x7b\x51\xca\x99\xd0\x32\x08\x91\x4f\x72\x5f\x43\x82\x6a\x69\xd6\x2a\x49\x68\x31\x28\x3c\xda\xf9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54
    b"\xa7\xe5\x1a\xe1\x05\xa2\x22\xca\x34\x34\x7e\x57\xd5\x61\x92\xb8\x59\x3c\x31\x14\x80\x12\x47\x7b\xc8\xb9\xf1\x63\xf9\x57\x52\x58": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.54",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa8\x1e\x36\x98\x3a\x90\x0c\xc4\x15\x23\x6b\xac\x78\x90\xc0\x01\xf3\x5b\xc5\x35\x66\x1f\xe0\xc9\x4d\xa0\x0f\x64\x8f\x40\x80\x03": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5313.sol",
            name="IERC5313",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xa8\x33\x23\xbf\xfb\x1c\x1d\x6b\xc9\xea\xbd\x3a\xb0\x6d\x1c\x4a\x6b\x1d\x4c\xfb\xb5\x5e\x05\xaf\x5c\x4f\x7f\xa6\x90\xe7\xeb\xec": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolState.sol",
            name="IUniswapV3PoolState",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272
    b"\xa8\x85\x9c\xb9\xd0\xf6\xce\x61\x38\xe4\x25\xe1\xb3\x81\xab\x12\x82\xec\xe9\x1c\xf2\xc2\xba\x55\x2a\x06\xde\x87\x43\x02\xe4\x90": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.272",
        ),
    ],
    # Hash appears in versions: v1.3.0
    b"\xa8\xf0\x61\x8d\x38\x67\xa3\x94\x2f\x2e\xe9\x90\xb0\x5e\x9e\xa6\x6c\xc4\x19\x24\xa0\x94\x1a\x2d\x04\x71\x42\xc0\x04\x5a\x83\x68": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungibleTokenPositionDescriptor.sol",
            name="NonfungibleTokenPosi tionDescriptor",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0, main
    b"\xa8\xff\x73\x0e\xe9\x3c\x74\x8c\x00\x2c\xed\xff\x98\xcd\xcf\xee\x63\x50\xc4\xac\xc9\x92\x67\xa9\x2e\xa0\xf1\x80\x83\x83\x3d\x13": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="v1.3.0",
        ),
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/external/IWETH9.sol",
            name="IWETH9",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xa9\x94\xe4\x8b\xad\x84\x03\x40\x66\xa5\xe2\x7c\x51\x78\x59\x8a\xe4\x5e\x4c\xa7\xe6\xe2\x06\x4e\x9c\xf2\x88\x92\x98\xc3\xf1\x6d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.22, v0.0.23
    b"\xa9\xa8\x36\x3c\xd6\x6d\x02\x4b\x41\xd8\xfe\x51\x1f\x10\x44\x3f\x39\x82\x56\x6e\xcf\xc4\x8e\x9c\x2c\xf3\x07\x92\xb5\x04\xcf\x7a": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.23",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xa9\xaa\x8c\xf4\x69\x21\xfa\xd7\xe2\x22\xa5\x91\x87\x5c\x30\x5d\x6f\x35\x50\xad\x62\x3b\x4f\xd4\x1f\x9f\xc2\xa5\x1f\xc0\xff\x9a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xa9\xea\xe0\x82\x34\xb8\xca\x5f\xd7\x27\x9d\xd3\xed\xb7\x5f\x88\x20\x7a\x75\xc8\xde\xe4\x8b\x72\x6d\x16\xb6\x6e\x16\x42\x13\x74": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xa9\xfe\x9d\x08\xe1\x39\x1f\x34\xcc\xbe\x26\x8d\x24\x11\xf7\x4c\x84\x75\x4b\xc4\xb0\xc7\x87\xed\x45\xa9\x1b\x57\xcd\xf7\x65\xb1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xaa\x44\x94\x9d\x45\xc5\x8d\x8e\xad\x8c\xb2\x80\x89\xc7\xf3\xc7\x14\xf1\x21\xb5\xf6\xdd\x7f\x6a\xea\x33\xf0\x7f\xa7\xbf\xe5\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xaa\x47\x27\xee\x4f\xeb\x57\x98\xb3\x68\x9c\x62\xdc\xb6\x7d\x0d\x4a\xec\x18\x33\x32\x4c\xdc\xea\x39\x5b\x4b\x82\x9c\xc4\x14\x4f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xaa\x49\x2f\xab\x24\x7e\xdb\x81\x00\xd9\xc4\x9c\xaa\x2b\xa1\xdb\x9b\x0d\x75\x29\xf0\x9c\x7a\xa8\x6c\x75\xd4\xf5\x9e\xf1\xd6\xec": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xaa\x54\x15\xec\xad\x53\x02\xfc\x8a\x18\x0c\x83\xe8\x0f\xc4\x7c\xe2\xcd\x5b\x57\x2a\xaa\xae\x5f\x64\x9a\x1f\x56\xe6\x1f\x14\x02": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xaa\x73\x49\x06\xe3\x7d\x5e\x7f\xfa\xf3\x9d\x66\xee\x3d\x90\xa7\x12\xdb\xa3\xc3\x6a\x49\x5a\xe6\xa2\xfa\xab\x8f\x00\x1b\xef\xe4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xaa\x84\xe0\xea\x32\x9f\x5e\xa9\x89\x28\x6c\xc9\x3a\x08\xbd\xc6\x50\x08\x99\x16\x4f\xb7\x9c\x1b\x38\x85\xa1\x0d\xc0\x5b\x3a\x54": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xaa\x8d\xa2\x2a\x1c\x01\x72\x5a\x1e\x64\x12\xef\xce\x67\x56\x73\x15\x3f\x23\xa2\x6a\xb9\x85\x53\x61\xea\xd3\xce\x47\x1a\xb0\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20MetadataUpgradeable.sol",
            name="IERC20MetadataUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xaa\x8d\xc6\x41\x82\x7b\x58\xe7\x9d\x29\xa3\xd4\x6c\x80\xce\xa5\xb8\x16\x5d\xfb\x6a\x61\x65\x0c\xdc\xfe\x47\xf0\xf1\x80\xc4\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xaa\xa2\x60\x2e\x6b\x7b\x87\xa7\x4a\xe2\x5d\x14\x57\xd0\xea\x33\x11\x30\x21\x95\xac\xd9\xce\xe5\x36\xea\x21\x98\xc3\x67\xc3\xaf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApprovalUpgradeable.sol",
            name="ERC20TemporaryApprovalUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7
    b"\xaa\xb4\x35\x21\xdf\xc6\x9b\x08\x0f\x6a\xfd\x0c\x1a\x2f\x7b\x8d\xa3\x31\xbb\x04\xc4\xb7\x0b\x85\x6e\xa2\x89\x64\x8d\x97\x59\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.4.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xaa\xcd\x5d\x73\xd1\x8a\x6a\x1f\x50\xfc\x77\xec\x41\x8a\xbd\xf5\x92\xcf\xbc\x0a\x1b\x75\x3f\xa0\x30\x23\xd1\xfb\xe9\xd0\x13\xfb": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/CurrencyReserves.sol",
            name="CurrencyReserves",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xaa\xd1\x59\x29\xe2\xd5\x2b\x5f\x0d\x1c\x81\x75\xcd\x9b\xd9\xc1\x33\x2e\xc6\xb1\xfa\x6c\x65\x62\x2e\xab\xa9\xbd\x99\xb1\xa9\x8f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155SupplyUpgradeable.sol",
            name="ERC1155SupplyUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xaa\xd3\xb2\x4b\xee\xec\xdd\x3c\xff\x96\x49\x3c\xd0\xce\xa9\x17\xbb\x55\x22\xef\x78\x76\x60\x50\x88\x45\xb1\xe2\x76\xf1\x77\x0d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelockUpgradeable.sol",
            name="IGovernorTimelockUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\xab\x0c\xb4\x11\x6a\x51\xf1\xdf\xc0\x97\xd5\x9a\xe6\x17\x25\xac\x92\x1d\x40\x46\x18\x1c\x17\x18\xc8\xf1\xba\xec\x2e\x30\xdc\x36": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0
    b"\xab\x10\xde\xb7\xf8\xaf\xf7\xfe\x6d\x22\x01\xa1\xef\x51\xdf\xa4\x82\x5f\xce\x9b\x39\x5a\x44\x49\xfa\xf4\x68\x82\x7c\xbd\x62\xbc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xab\x22\x7f\xfc\x80\xb3\x57\x00\xd6\x5f\x35\x99\xe4\x05\x83\xc7\xf7\x85\x1d\xa7\xe0\x6f\x30\x6d\x7c\x7a\x4d\x6b\xd9\x4b\x46\x75": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.137
    b"\xab\x36\x86\x83\xe9\x61\xb1\xa2\xe6\x08\x8f\x23\xf1\x01\x8a\xbe\x8d\x3c\xce\xb2\x0e\xcf\x02\x11\xa1\x05\xec\x8a\x9c\xea\x89\x24": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.137",
        ),
    ],
    # Hash appears in versions: v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xab\x49\x03\x06\xe4\x2d\xef\xcc\x9b\x74\xdf\xa9\xf2\x6f\xdb\xf7\x4f\x66\xdc\xcd\x08\x22\x46\x29\x4f\xdd\x62\x3d\xb5\xfd\x13\xb9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xab\x4d\x08\x03\xe8\xe2\x71\x12\xb8\x83\x1c\x19\x32\x3c\xc0\xff\x0c\x50\x07\xb4\x04\x30\x32\xe2\x5a\xb0\x5f\xdc\xbe\xe7\xdf\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.198
    b"\xab\x72\x68\xee\xcc\x8b\x1b\x9f\x5a\xed\xfd\x86\xbb\x57\x51\x4c\x67\x49\x9b\xb1\xa5\x9d\x0b\x71\xc5\x03\x6e\x11\xd2\xc8\x47\xc5": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.198",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xab\xcb\xa5\x57\xb8\x9e\xbc\x2c\x0a\x01\x84\x6f\x59\x32\x52\x37\xce\xc2\x23\xa8\x93\xb4\x15\x9d\x1a\x26\xff\x03\xf2\xd3\xe4\x94": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909MetadataUpgradeable.sol",
            name="ERC6909MetadataUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC6909/extensions/draft-ERC6909MetadataUpgradeable.sol",
            name="ERC6909MetadataUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xab\xfa\x81\xb3\x0e\x45\xef\xc9\x81\xc1\x67\xc3\xee\x83\x05\x95\xc8\x8c\xe9\x5f\x88\x02\x16\x72\x13\x60\xe1\xb3\x0f\x15\x40\x4f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xab\xfd\x7a\x13\xc1\xf3\x56\xe5\xc3\x5e\xf3\xd2\x55\xe7\x3b\x26\xda\x83\xe6\x52\x82\x09\x05\x7a\xcd\xc9\x0c\xf7\xb5\xab\x12\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1967.sol",
            name="IERC1967",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xac\x05\x15\xbe\x0b\x6a\x33\x77\x95\x64\xe6\x71\x8f\x9f\xd9\x96\xb1\xa2\x28\x09\x2e\x1f\x1a\x84\xec\xbe\xf9\xb5\xce\x3d\xb5\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xac\x34\x61\x15\x17\x4d\xa4\x09\x76\xb1\x97\x44\xab\xd7\x8d\x96\x97\x53\x8c\x61\x83\x0d\x96\xde\xed\x7f\x2e\xf6\x75\xf6\x4c\x9f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xac\x82\x50\x44\xc2\x58\x9e\x56\x2d\x46\x2b\x70\x17\x09\x69\xd3\x4b\x85\x4d\x29\x12\xa6\x46\x27\xe9\xee\x69\xa4\xb6\x00\x34\x3f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Pausable.sol",
            name="ERC1155Pausable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xac\xc2\x4a\x22\x6c\x58\x9f\xae\x16\x6f\x92\x20\xf3\xf3\xf6\x7f\xd5\x0c\x2f\x86\xc9\x48\x50\xc1\x1a\x5e\x5f\x39\x08\x73\x07\x46": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xac\xca\xf2\xff\x7f\x2c\xcb\xa6\xa3\xf3\x2e\x9d\xb8\x12\x80\xb8\xb0\x4a\x3a\x19\x3a\xa4\xb4\x67\xb1\x48\xb4\xb9\xc3\xff\x2a\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xac\xd3\x2e\x3e\xf1\x54\x32\xd8\x2a\x9c\xd8\xb3\xf0\xb6\x94\x66\xf2\x77\xbd\x0e\x41\xc6\x78\x77\x9c\x02\xcb\x0b\x9a\x89\xcd\x3d": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolDerivedState.sol",
            name="IUniswapV3Po olDerivedState",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xac\xe2\xbc\x93\x92\x90\x82\xc4\x02\xde\x58\x68\xf6\x7e\xa4\xa5\x58\xed\xab\x03\xd8\x58\xb7\xca\xcd\xce\x42\x1e\xe1\xb7\x81\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Multicall.sol",
            name="Multicall",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xac\xe7\x8b\x75\x04\x68\xbc\x52\x84\xd6\x02\x42\x6d\x9f\x92\xb4\xea\x9d\x9a\x0c\xe6\xa1\x6d\x53\x8d\x4d\x6d\x06\x1f\x57\x13\xa1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable2StepUpgradeable.sol",
            name="Ownable2StepUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.2, v4.5.0-rc.0
    b"\xac\xe9\xd7\xf0\x55\xf4\x11\x52\xca\xd0\x86\x59\x85\x37\x38\x2f\xcb\xe4\x53\x0e\x2b\xdd\x26\xc2\x02\x00\x35\xd8\x33\x55\xb0\x22": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\xad\xbe\xa9\x8a\x4c\x95\x15\x1b\xde\xe5\x8f\x73\xa6\xd4\xd5\x96\x9d\xbd\xc6\xdf\x70\x4c\x23\xc4\x04\x93\xfa\x24\x0a\xd1\x44\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xad\xc6\x90\x9c\xd1\x37\x57\xc7\xfd\x3c\xbe\x76\x60\x32\x9d\xc1\xc5\xd2\x07\xb2\x42\x1b\x0a\xe8\x4b\x20\x43\x4b\xee\x12\x1e\xb3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xad\xd1\x44\x18\x0e\x57\x68\x37\x22\xbd\x16\x19\xdb\x88\x3e\xcd\x10\xfc\xe1\xd8\x34\x08\x5b\x1b\xd4\xaa\x33\x98\xab\xf2\xad\xab": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\xad\xdf\x99\x91\x16\x66\xf6\xe4\xab\xfd\x78\x9e\xee\x34\x6e\x13\x9b\x79\x38\xc3\xac\xbb\xf1\xcd\x0f\x6d\x65\xd6\xb8\x01\xb0\xa1": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xae\x16\x90\x85\x62\x45\x60\x9c\x4e\x65\x5b\x16\xa9\x27\x30\x20\x04\xd2\x15\x24\x97\x92\xe2\xc5\x78\x9b\xa0\x4c\x84\x44\x6c\x55": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/PoolManager.sol",
            name="PoolManager",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xae\x21\x58\xc5\x85\x1f\x75\x38\x5e\x50\x11\xb8\x98\xa2\xc8\x5c\x20\x49\xc0\xe7\x72\xa3\xc8\xd3\xf1\x34\x66\x7b\x6e\xe0\xa2\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.137
    b"\xae\x31\x96\x73\xad\xb1\xa1\x8e\xd0\xcd\x4f\x74\x38\xc8\x53\x4e\x76\x42\x08\x19\x10\xca\xce\x07\x09\x89\x0e\x13\x51\x15\x0e\x0f": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337Factory.sol",
            name="ERC4337Factory",
            version="v0.0.137",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xae\x45\xce\x0e\x93\xb7\xba\xdb\x53\xe6\xcd\x3b\x31\xe2\x48\x5a\x94\x70\xe6\x35\xea\xdf\x8d\x5a\xdf\x2d\x8d\x09\xe3\x3d\x5a\xca": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721.sol",
            name="IERC721",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.1, v5.3.0-rc.0
    b"\xae\x4b\x8f\x70\x79\xc0\x9f\x1c\x08\xf5\x67\xb5\xcc\x39\xd9\x46\x2f\x06\x3b\x01\xc1\x5c\xb1\xdf\x77\x63\x80\xe6\x03\xe8\x1c\x4a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC7579Utils.sol",
            name="ERC7579Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC7579Utils.sol",
            name="ERC7579Utils",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC7579Utils.sol",
            name="ERC7579Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\xae\x69\x97\x95\x77\xf2\xa2\xb4\x7f\x6c\x31\x91\xea\x89\x24\x4e\x7d\x5d\x9c\xe7\x6f\x5d\x24\x03\xe1\x57\x9e\x94\x86\x45\x74\xc1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235
    b"\xae\x72\x17\x24\x31\x71\x56\x11\x23\xa0\x12\xf3\xbd\xdf\xe8\x4b\xd9\xed\x89\xb8\x3f\x2b\x79\x13\x63\xe6\xdf\x8a\xb9\x59\xcb\x17": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.235",
        ),
    ],
    # Hash appears in versions: v0.0.286, v0.0.287
    b"\xae\x8a\x1e\xb1\x22\xdd\xbc\x5b\x8f\x24\x2c\x95\x32\xf4\x3e\xf4\x73\x26\xd1\x50\x35\x17\x8f\x7b\x35\xbc\x6c\xe7\x7e\xd1\xcf\xa6": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.287",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xae\x98\xc4\xd5\x39\x25\xaa\xb6\x58\x04\xa7\x9f\x20\x43\xfc\x39\x31\x69\xab\xef\x6c\xfe\xfb\xf6\x8a\xfb\x2d\x9e\xe7\x3d\x79\x04": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xae\x9b\xeb\x11\x70\xf2\xec\x29\x0c\x16\x07\x85\xe9\x74\x21\x36\xa3\x98\xbf\x71\x05\x90\x86\x19\x51\x39\x0e\x3c\x98\xd9\x06\x26": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xae\xcf\xb8\x79\x20\x91\x39\x6b\xde\x21\x56\xbe\x4a\x16\xe9\x2f\x57\x30\xe0\x6f\xe0\x8a\x32\xc6\x88\x2a\x01\x8a\xff\x8e\x39\xee": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\xae\xfd\x0e\x48\x1a\x27\xd1\x27\x94\x36\x44\x53\x25\xeb\xbe\xc7\x80\xdb\xac\xc6\x8d\xeb\x51\xcc\xb0\xc4\x24\x2d\x93\x26\xc8\xcc": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/utils/BaseHook.sol",
            name="BaseHook",
            version="main",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xaf\x00\xb7\xc4\xd8\xa3\x29\x9a\x11\x08\x92\x23\x1c\x18\x2b\x75\x8f\xe8\x64\xf5\xf3\xb9\xba\x82\x8f\xfd\xbf\x8e\xd9\x64\xb2\x48": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0
    b"\xaf\x0e\xbf\x5b\x72\xad\x32\x5d\x17\x6b\xe5\x7f\xd8\x89\xd4\x0a\xd1\xa5\xf2\xb8\x2f\x5c\x2f\xe5\xc8\x07\x56\x89\x51\xda\xa2\xf9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="IAccessControlUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xaf\x33\xfd\x31\x26\x13\x71\x55\x7b\x42\xd2\x23\x0f\x31\x6a\x4c\xe6\xee\xba\x3f\x72\xeb\xc0\xd8\xed\x55\x6f\x9c\x8b\xa0\xc9\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ProxyAdmin.sol",
            name="ProxyAdmin",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0
    b"\xaf\x42\x8b\x0c\x8f\xd8\xd3\x6a\xcd\xe1\x80\x08\xdf\x73\x21\x8c\x7e\xc1\x6a\x93\xea\xb4\x75\x77\xbc\x7d\x23\xc2\x4f\xed\x1c\x02": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/Multicall.sol",
            name="Multicall",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/Multicall.sol",
            name="Multicall",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/Multicall.sol",
            name="Multicall",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/Multicall.sol",
            name="Multicall",
            version="v1.2.0",
        ),
    ],
    # Hash appears in versions: v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xaf\x47\xc4\x82\x1f\xa3\x43\x68\x63\xb1\x30\xbf\x76\x1c\xae\xa3\x6a\xfc\xef\xc4\xb5\x06\x4e\x7b\xb7\xe3\x50\x6a\x21\xec\x91\x60": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Context.sol",
            name="ERC2771Context",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.204
    b"\xaf\x52\xed\x60\x5c\xbf\x4e\x3f\xae\xae\x68\x2e\x6a\xbe\x6d\xaf\x79\x95\x1d\x9b\x6c\x34\xcd\x0c\x9a\x17\x59\xb5\x50\x19\x43\xe4": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.204",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xaf\x62\x5a\x8c\x51\x24\x1e\x49\xfe\x23\xd5\x7a\x98\xd0\x1e\x14\x9d\x3b\x5a\x1d\x89\x27\xd5\x36\x8c\x6b\xca\x51\x15\xa1\xb7\x80": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/LibAMB.sol",
            name="LibAMB",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xaf\x68\xbd\x4c\x03\xec\x08\xdd\x8b\x0b\x80\x3d\x31\xd1\xf2\xe9\x26\xb5\x54\x7b\x4b\xd1\xec\x61\xaa\x49\x6b\x4b\xbf\x65\x23\xad": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControl.sol",
            name="IAccessControl",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xaf\x6f\x21\xeb\x6b\x3b\xb9\x39\xa6\xf3\xf0\xa7\xbd\x30\x0c\xb1\xca\xa2\x77\xac\x36\x36\x3b\x6d\x85\x9f\x42\x1b\x51\xb8\xff\xf8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Roles.sol",
            name="Roles",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Roles.sol",
            name="Roles",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xaf\x76\x92\x8b\x3d\xfb\x8a\x30\xab\x42\xa4\x96\x89\x56\xa0\x56\x1a\x3f\xa1\x2d\x38\xc0\xb8\xf2\x58\x76\xb2\xe5\x1f\x0d\xde\x4d": [
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/Bytes32AddressLib.sol",
            name="Bytes32AddressLib",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xaf\x77\xda\x71\xe8\xeb\xb0\x04\x66\x56\x1e\x98\xf9\x33\xeb\x35\xb7\xcc\xd9\x94\xcb\xca\x44\x7d\x3e\xd5\x2b\x8a\xea\x5f\x9a\x6e": [
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Auth.sol",
            name="Authority",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268
    b"\xaf\x7c\x3a\xb1\x7a\x38\x8b\xf0\x95\xad\xee\xa4\xb5\x64\x14\x2e\xce\xc5\x6a\xc7\x32\xcb\xa9\x10\x7d\x55\xca\xa2\x87\xbe\xea\xac": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.268",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xaf\x8d\x0d\x5e\x05\x0b\x40\x5d\x2b\x0f\x78\x69\x34\x5c\xb0\x6a\x14\x49\x75\xda\x63\x50\x24\xb2\xf9\xa1\xa8\xdf\x31\x06\x5f\xa7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAggregator",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAggregator",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAggregator",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAggregator",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IAggregator",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0-beta.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xaf\xae\xce\xf1\x76\x44\x78\x4f\x40\xb1\x5d\x08\x55\x5f\x7d\x78\x48\x16\x61\xd4\xb0\x20\xec\x2a\x61\x5c\x71\xdf\x0b\x09\x1d\x36": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xaf\xb3\x33\x46\xa3\x00\x5d\x22\x21\xb8\x1e\xec\x4e\x32\x1f\xd3\xbb\xf0\x44\x8a\x9b\xaa\xf1\xf4\xf0\xa6\xa1\xaf\x98\x28\xbe\xd5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xaf\xca\x41\x79\x50\x8a\xc3\xff\x70\x51\xe6\xc6\xca\x98\xf5\x8c\x7e\x28\x93\xaa\x0a\xac\xfd\x79\x41\x0c\xc4\xf5\xba\x5d\x78\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignatureUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.3.0
    b"\xaf\xf9\x0e\x13\x1c\x40\x25\x19\x8c\x03\x16\xa4\xe1\x0c\xa8\x5a\xfd\x0c\xe6\xa4\x4f\x9f\xcc\xc3\x8b\x40\x50\x2b\x79\xba\x27\x0a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/EscrowUpgradeable.sol",
            name="EscrowUpgradeable",
            version="v3.3.0",
        ),
    ],
    # Hash appears in versions: v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\xb0\x08\x47\xe6\xf0\xe8\x52\x9e\x33\x79\x38\xcd\xca\x70\x7b\x2d\xe1\x41\x74\xc7\x37\x7c\x62\xeb\xd8\x44\xe5\x1e\x71\xe7\x18\x07": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xb0\x42\xe4\x06\x2d\x84\x20\x33\xf7\x55\xfd\xcd\xda\xb0\xf8\x91\x02\x9f\x45\x4a\x11\x3d\x39\x07\xa7\xd2\x96\x17\x12\xab\x76\x17": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xb0\x57\x2a\x05\x6c\x3e\x25\x95\x9e\xc1\x2a\xf0\x27\x2e\x4d\x24\x8b\x2e\xea\x6c\xc3\x9d\xf4\x9e\x19\xed\x9d\x46\xc3\x6c\xc0\xb1": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/ERC721Permit.sol",
            name="ERC721Permit",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/ERC721Permit.sol",
            name="ERC721Permit",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/ERC721Permit.sol",
            name="ERC721Permit",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/ERC721Permit.sol",
            name="ERC721Permit",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/ERC721Permit.sol",
            name="ERC721Permit",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.292
    b"\xb0\x65\x1b\xed\x82\xd5\x32\x11\x64\xf1\x0a\x23\xe7\xb8\x0e\x9b\xa6\xb3\xdd\x27\xc7\x3f\x92\xc3\x3a\x1c\x04\x3e\x76\x9a\x50\x44": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb0\xad\x4e\x11\x0d\x01\x53\xf8\x9d\x11\x24\xf5\x4a\x4e\x6a\x1d\x25\x76\x96\xa0\x60\x76\xb4\xaa\x81\xd3\x5f\xf2\xa2\xb3\x95\xb3": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableMapLib.sol",
            name="EnumerableMapLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v5.0.0-rc.0
    b"\xb0\xad\x57\xf0\x34\x53\xec\xfd\x33\x89\x99\xf2\xbf\xe8\xfb\x35\x0e\x93\x56\x76\xf1\x74\x08\xae\x69\x69\xb8\xe9\x36\x32\xa6\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb0\xba\x9d\x3e\xd6\x76\x3c\xfe\x49\x8f\x3d\x8d\x45\x39\xfb\x2b\xff\x53\x4b\x3f\x81\x6f\x45\x52\x46\xea\x7d\x29\x8d\x36\x44\x55": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.2.0-rc.0
    b"\xb0\xc0\xbb\xcd\x37\x97\x57\x93\xd1\xc7\x35\xde\x21\xaa\x3c\x2b\x48\xab\x77\xb7\x66\xd8\x4c\xe8\x69\x82\x27\x6c\xe1\x46\xfa\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC7579Utils.sol",
            name="ERC7579Utils",
            version="v5.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xb0\xf2\x67\x97\x4b\xde\xaa\x4f\x65\x1e\x26\x7d\x1a\x62\xac\x4e\x1c\x6e\x65\xdd\x74\x7b\xea\x94\x0e\x5c\xd2\xa6\xf5\x4a\xa8\xcf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xb0\xf9\xf2\x83\x66\x8d\x6d\xbd\x35\x33\x60\x4c\x9f\xb2\x18\x04\xad\x8f\x28\xa8\x40\xa0\xd7\x67\x46\x05\x34\xa2\x62\x67\xa2\x61": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\xb1\x02\xa7\x01\xbf\xfa\x6b\xe2\x98\xc4\x0e\xcb\x10\x59\x86\x15\x4f\xab\xf6\xff\x18\xcd\xfc\xd5\x13\x80\x6b\x2d\x44\x1a\x83\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xb1\x2b\x6a\xd3\xbe\x4e\x00\x33\x47\x25\x46\xb6\x8d\x89\x4b\xdd\x80\x4e\x64\x60\x3e\x0e\x46\xb3\xfc\xee\xf2\x46\x3f\x20\x76\xf4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="EscrowUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb1\x37\x75\x3f\xde\x37\x8f\x8b\x12\x39\xff\x62\xdf\xfa\x9b\xa9\x32\xc1\x36\x4b\xce\x7a\x20\x9e\xda\x0a\x0c\x7f\xa2\xe4\x28\xcb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChildUpgradeable.sol",
            name="CrossChainEnabledPolygonChildUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.0.1, v3.0.2
    b"\xb1\x47\xed\x9e\xf5\xdf\x48\x11\xed\x10\xb0\x2a\x0a\xa6\x95\x73\x34\x48\xff\x96\x45\x69\x34\xf5\xe0\x7c\x9e\xcb\x5c\xa1\xff\xd3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb1\x48\x20\xde\x29\xa2\x43\x78\xef\x69\x1e\xfd\x46\x03\x47\x1d\xb9\x58\x43\x73\x0e\x18\x9e\xd3\xab\xf2\x99\xe5\x83\x11\x84\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWalletCliff.sol",
            name="VestingWalletCliff",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xb1\x48\x2f\xd6\xdc\x23\xc9\x86\x80\x01\x0a\xb3\x3a\x45\xdc\xfb\x70\x32\x8d\x07\x82\x3a\x99\xe8\x50\xb5\xd7\x01\x21\x26\x06\x75": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb1\x54\xc2\x4c\x07\x67\x9a\xa5\xf4\x8f\x89\xdb\xea\xb9\x01\xc4\x51\xf2\xa7\x54\xae\x66\x8d\x82\x58\x67\xb1\xbe\x79\x71\x38\xdd": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibStorage.sol",
            name="LibStorage",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xb1\x5f\x98\x05\x30\x59\xa9\x73\x6e\xef\x97\x0a\x81\xc1\xf5\x0a\xd6\x17\x5d\x61\xf6\xac\x39\x17\x93\x9b\xd8\xf2\xa8\x2a\x57\xab": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/pool/IUniswapV3PoolActions.sol",
            name="IUniswapV3PoolAct ions",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb1\x61\x12\x4c\xbd\x4f\x05\xe4\xa4\x11\x68\xec\xeb\x9f\x77\x31\xb0\x4f\xac\x6b\xaa\xe1\x3b\x17\x37\x47\xcb\xc3\x90\x69\x60\x19": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb1\x65\xbe\x2c\xa0\xa4\x59\xd7\xeb\x19\x40\x7e\x98\x25\x9c\x0f\x17\x5b\x34\x89\x0c\xe0\x2b\xcb\x12\x6c\xb6\xfd\xfd\x93\x7a\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TransientSlot.sol",
            name="TransientSlot",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb1\x71\x51\x7a\xb7\xda\x95\x83\x43\x50\x8f\xa2\x7d\xf4\x5b\x5d\x13\xfa\x86\xf6\x51\x9d\xea\x7a\x6d\x5d\x88\x01\x26\x5e\x88\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20VotesComp.sol",
            name="ERC20VotesComp",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\xb1\x7b\x6f\x8d\xe1\x3f\x17\xf1\x3b\x55\x95\x79\x82\x42\xc6\xf1\xf0\x5a\x97\xe6\xef\xe3\x38\x56\x8e\x91\x99\xcd\x60\x9f\xcb\x10": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xb1\x86\x7f\x5b\x70\x4c\x20\x9d\xe3\xa8\xc5\xbb\x90\x26\xba\x83\xb7\x1b\x8a\x31\x9e\xa8\x3e\xc9\x3e\x6f\xd8\x94\x93\x58\x7c\x40": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.27
    b"\xb1\xb6\xe4\xc9\xbf\x21\xa8\x7b\xc1\x56\x7c\x53\xc9\x13\x4c\x53\xdb\xf9\xfd\x46\x58\xb9\xbc\x0e\x5b\xde\x92\x46\xb4\x3d\x90\x4c": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.27",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\xb1\xe8\xb2\x13\x84\xdd\xcd\x19\x80\xa5\x32\x53\x71\x73\x65\x6a\x2f\x0b\xe4\x8b\x4c\x10\x76\xeb\x90\x08\x61\x26\xd3\x3f\xb8\xaf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xb1\xf5\xf2\xa6\xc4\x8c\xe2\x98\xef\xa1\x70\x71\xe7\xd4\xc2\xd2\xe8\xa4\x90\x98\x76\x04\x33\x5e\xc8\xfb\x56\x4e\xa7\x1e\x63\x8c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/DoubleEndedQueue.sol",
            name="DoubleEndedQueue",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124
    b"\xb2\x0a\xbc\x9a\xb4\xe2\xcb\x91\x12\x17\x29\x6a\xcd\x98\x9b\xdc\x34\xd1\x6c\xd8\x1d\xa2\x8b\x28\x74\x34\xb2\x33\x96\x8c\x39\x26": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v0.0.133
    b"\xb2\x0f\xdc\xc7\x1d\x54\x09\xe0\xca\x49\x67\xb0\x54\x1d\x48\x46\x87\xfa\xde\xd3\x33\xd9\x6d\x39\xbf\x4c\x28\xd7\x07\x87\x2b\xf7": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.133",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb2\x1a\xb0\x62\x55\xa8\x15\xdc\x39\xf3\x4f\x5b\xec\x66\x23\x3f\xb0\x46\xee\xf4\x2b\x06\xa8\xe2\xfa\x0d\x6d\x35\xd6\xeb\x59\x3a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.6, v0.0.7, v0.0.8, v0.0.9
    b"\xb2\x1b\x8f\xd7\x2e\xe6\x72\x5f\x4f\x7e\xaf\x5c\x2d\x9f\xb9\x04\x56\x78\x25\xb2\x1e\xdd\x92\xe8\x8c\x40\x63\x7b\xf6\x2b\xdc\xc6": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\xb2\x21\xf2\x67\x72\xc2\x18\xf6\xfb\x16\x68\xa1\x6b\x3c\x8f\x80\x67\x48\x06\xac\x30\x5b\x5d\x29\x06\xb2\x88\x7d\xae\x9d\xb5\x50": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb2\x2c\x74\xdc\x22\xd7\x14\xbc\xce\x56\xe9\xa8\x9e\x39\x0a\x53\x10\x77\xb9\x75\x22\x3f\x91\x14\xb6\xef\x85\x67\xb9\x7b\x89\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ArraysUpgradeable.sol",
            name="ArraysUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\xb2\x3b\xb5\xb7\xd9\x46\x1f\xfe\xa6\x58\xc2\x9d\x65\xfb\xfe\x28\xb8\xe8\xa6\x9a\xbd\xf7\x3d\x32\xf7\x3e\x3c\xf2\xc2\x71\x86\x9b": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IV4Router.sol",
            name="IV4Router",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xb2\x45\x1f\x6a\xe0\x95\xe4\x8f\xd4\x81\x5a\xf0\x25\xf2\x77\x84\xce\x48\x82\x64\xa6\xd3\x0f\x54\xe1\x60\x7e\x72\x15\x96\x62\x1c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb2\x53\xf1\x49\x52\xe3\x51\x64\x56\x52\xcc\x95\x78\xed\x54\xcd\x81\x01\xe9\xe5\x93\x8d\x5c\x7f\xd6\x6a\xf5\x16\x10\x20\xb4\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockAccessUpgradeable.sol",
            name="GovernorTimelockAccessUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0
    b"\xb2\x6a\x8a\x6a\x02\xfa\xd2\xba\x1d\x21\x52\x4e\x1b\xd0\x56\x70\x85\xe1\xc1\x21\x4a\x83\x70\xf6\x22\xe5\xf3\xba\x9d\x70\x11\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xb2\x9a\x05\xa9\xef\x38\x4a\x9d\xb1\x04\x62\x82\xdd\xd9\x44\xb9\x87\x29\x19\xe6\x55\x2b\x79\x13\x94\x55\x4f\xa4\xa0\xf4\xa8\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0
    b"\xb2\x9a\x1c\xce\x3d\x56\x93\x78\xd3\x3b\x9b\xa9\xd6\x5c\x45\xf3\x2c\xcd\x0d\x55\x6c\x8b\x50\x5d\x22\xd4\x13\x8b\x4e\x67\xd2\x08": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v6.5.0",
        ),
    ],
    # Hash appears in versions: v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.2.1-solc-0.7
    b"\xb2\x9f\x19\x22\x14\x9d\xab\x75\x5d\xa1\xd9\x4b\xe6\x7b\xa6\x90\xe7\x2d\xbb\xfa\x2d\x14\x70\xf0\xb6\xf7\xf4\x7e\x26\x48\x17\xae": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableProxy.sol",
            name="UpgradeableProxy",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xb2\xbe\x93\xf8\x44\x11\x43\xd0\x41\xe6\x4e\x8d\xdb\x9e\x46\x67\x9b\xdd\xe2\x37\x6a\x2b\x67\x16\x7c\x95\x33\xe7\x49\x55\x0c\xad": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb2\xc0\x38\x68\xa9\x4a\xb4\x3c\xea\x71\xb4\x55\x04\x4f\x49\x9e\xb1\xae\xe0\xc7\x62\xc3\x13\x4c\x5f\xb1\x39\x6e\xf4\xd0\xbf\x61": [
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb2\xd2\xf1\xf9\x1f\xfb\xfe\x06\x7d\xaf\xfc\xb1\x41\xf2\x12\xf3\x6b\x17\x39\x25\xf7\xcf\x77\xc7\x6c\x1d\x6b\x33\xba\xbf\x7f\x7f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Burnable.sol",
            name="ERC1155Burnable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xb2\xd6\xf4\x88\xb0\xd9\x75\x4e\x28\x9b\xa5\x85\x0c\x49\x05\xea\xd1\xaf\x69\x8a\x86\x64\x12\x72\x6a\x96\xe9\xa3\xa9\x6c\x92\x8e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1
    b"\xb2\xea\x47\xaf\xe8\x31\xbd\x5d\xac\x0c\x23\x65\xc7\x3a\x4d\x22\x12\xc8\xb5\x4f\x3b\x87\x77\x75\x7b\xbd\x41\xa1\xf1\xfa\x71\xf9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.9.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xb3\x04\xd8\x72\x53\x12\xf3\x3d\xd9\x39\xff\xc7\xd6\x2f\x69\xb5\xea\xd0\x6f\x23\x8a\x59\xe6\x0e\x21\x9b\xb4\x4e\x7f\x18\xb2\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xb3\x23\x6f\x58\x53\x81\x80\xb3\xe3\x1f\x97\x1e\x7f\xf6\xe1\x96\x15\x93\xda\x0d\x67\xbd\xd7\xb2\x14\x5f\xd5\xa3\x3f\xd7\xb4\xc5": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155TokenReceiver",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb3\x5e\x4d\xae\x7c\x51\xac\xf1\xe7\x00\x0b\xd2\xa6\x3b\x49\x13\xcb\x45\xcd\x2c\x2d\x4b\x59\xef\xc9\x5d\x9a\xa3\x3c\xe2\x4b\xd3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.252
    b"\xb3\x62\x8c\x40\x73\x4a\xab\x5f\xdb\x29\xf1\xee\xa8\x1c\x7d\x7a\xd5\xbb\x7d\x7d\x9f\x0e\xc6\xd8\x6c\xaa\x40\x5f\x12\xb6\x0f\x8c": [
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuardTransient.sol",
            name="ReentrancyGuardTransient",
            version="v0.0.252",
        ),
    ],
    # Hash appears in versions: main
    b"\xb3\x68\x80\x69\x13\x59\xfc\x1e\xd6\xbb\xf6\xb6\x20\x5b\x31\x90\xd1\x8a\xa6\xbf\xdb\x50\x3d\x16\x7c\x89\x80\xbe\xcc\xfb\xab\x32": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/hooks/WETHHook.sol",
            name="WETHHook",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.10, v0.0.11, v0.0.12, v0.0.13, v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19
    b"\xb3\x8d\x74\x13\x05\x2c\x32\xb5\x68\xac\xa3\xe7\x63\x90\x18\xba\x32\x39\x59\x16\xb4\xa9\xc2\x6b\x5f\x54\xf6\xd0\x25\x9c\xfd\x6c": [
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Sort.sol",
            name="Sort",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb3\xc6\xa9\x8a\x19\xda\x42\xde\x58\x72\x07\xda\xa4\x8e\xee\xbc\x0a\x4b\x46\x1b\xb3\xf5\x41\xca\xde\x87\xdc\x42\x80\xa5\x68\x78": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb3\xce\x7a\xc7\x28\x8a\x01\x69\x64\x74\xaf\x61\xde\x49\x42\xa3\xe9\xe3\x2b\xd0\xd3\xaa\xf4\xd1\x12\xba\x75\xf1\x30\xc1\x1a\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xb3\xe0\x3a\xd3\xd4\x82\xc1\xbf\xb4\xc3\x99\xb5\xe8\x96\x96\xe2\x22\x01\x74\xd7\xf2\x21\xbd\x40\xa5\x20\x9a\x23\x55\x26\xf0\x89": [
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/RolesAuthority.sol",
            name="RolesAuthority",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.73
    b"\xb3\xe3\xbb\xcf\xa4\xf2\xf7\x0d\x65\x71\x6f\x95\xa5\x58\xe4\xdb\x14\xe2\x7c\x4c\x78\xa7\x27\x94\x3e\x38\xcb\x07\x41\xf9\x5c\x6d": [
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.73",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\xb3\xe6\x2b\x52\xdb\x58\x52\xa3\x28\x95\xde\x1e\xb4\x0e\x00\x59\xb4\x4e\x59\xdd\x6d\x84\x71\xa5\x79\xac\x0b\xbd\xfa\x08\xbf\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xb3\xf3\xf0\x56\x74\x07\xe2\x9d\x16\x78\x74\x27\x0b\x35\x18\x78\x08\xea\x8a\x71\x1e\xce\xf0\xb2\x58\x88\x7c\xed\x7d\x3d\x4f\x21": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Pausable.sol",
            name="ERC20Pausable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xb4\x05\xff\x02\x46\x0e\x07\x6f\x6a\x41\x60\x2a\x06\x21\xfb\x8b\xf8\x44\x89\x80\x65\x20\xb7\x2b\xda\x4e\x8a\xde\xfa\xd7\x1e\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xb4\x11\xad\xee\x19\xb0\x95\xd5\xdd\x64\x3d\x96\x54\x28\x78\xdd\x0a\xae\xc6\xe7\x15\xe3\x24\x2a\xd6\xbd\x59\x40\xe1\xba\x17\xd7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273
    b"\xb4\x45\x94\x6d\x5f\x37\x95\xcc\xf4\x39\x52\x3d\xd1\x5e\xdc\xcd\x15\x6c\x1f\x05\x1b\x04\xa5\xa0\x44\xe7\x12\x47\xe7\x4d\x5a\xa4": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xb4\x4b\x72\x79\x42\x69\x43\xb3\x18\x21\xde\x4b\x1d\x20\x74\xcd\x11\xad\x4c\x33\x21\x94\x93\xc2\xa7\xfb\x1e\x2f\xc3\xf0\x21\xf6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/DoubleEndedQueueUpgradeable.sol",
            name="DoubleEndedQueueUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xb4\x57\xba\xa7\x13\x6a\x47\x10\x6f\x64\xca\x52\xd6\x67\xd4\xc5\x50\x95\xa7\x58\x81\xee\x47\xc0\xe5\x7d\xf2\x7e\x5a\xf0\x6a\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xb4\x8c\xf2\xe3\x6b\x55\x21\xbe\x16\xe2\xe3\x13\xaa\x76\xcf\xcb\x9e\x94\x9d\xed\xfa\xe3\x71\x2d\xeb\xea\x09\xdb\x42\xd0\xb3\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225
    b"\xb4\x92\xbc\x53\x4a\x0f\xf3\xe9\x0f\x93\x08\x7d\x13\x52\xaa\xb7\x24\x6a\xa7\x6a\x8f\x64\xa3\x12\x7a\x6e\xd1\x96\x30\xe8\xf8\x78": [
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/GasBurnerLib.sol",
            name="GasBurnerLib",
            version="v0.0.225",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xb4\xb5\xc0\x8b\x0d\x89\xc2\x16\xa1\x98\xb5\xa0\xa4\xe9\x65\x14\xc9\x46\xe8\x04\x0c\xe7\xfd\xec\xdc\x84\xf2\xf0\x7f\x50\x2f\x39": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\xb4\xc9\x46\x92\x35\x7d\x64\xcd\x94\x48\xbc\x53\x3a\x82\x9a\xa2\x77\xcc\x03\x0b\x17\xfb\xa3\xf9\x69\x87\x8e\x91\xb1\x88\xa6\x63": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb4\xe0\x03\xc3\xb7\xcc\x84\xbe\xbe\x08\x82\x56\xef\xba\xd0\xcd\x01\x93\xb2\x93\x74\xb6\xd4\x7e\x70\x79\x24\xd0\x4e\x97\x01\xf9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/IBeacon.sol",
            name="IBeacon",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/IBeacon.sol",
            name="IBeacon",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186
    b"\xb5\x00\x65\x4f\xd7\x39\x33\xa4\x3e\x6b\x12\x4b\x9f\x60\x88\x4c\xba\x9b\xf0\x2e\x0c\xe5\xf0\xd2\x34\x08\x3e\xe0\x38\xa9\xc4\x63": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.186",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xb5\x54\x1e\x90\x39\x0e\xcc\x54\x4b\x92\xe1\x2d\x1e\xbc\x5e\x85\x79\x69\x0b\xed\x46\xb9\x39\x52\x30\x65\x03\xff\x30\x7c\x27\xfe": [
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/mixins/ERC4626.sol",
            name="ERC4626",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb5\x6d\x48\x72\xea\x78\x2f\xff\x35\x2e\x2f\x36\xd3\x3d\x6e\xc9\x55\x96\x45\x29\x08\xd0\xa3\x52\x90\x26\xde\xc3\xca\x64\xac\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xb5\x89\x80\x26\x0b\x87\x8a\x44\x77\x5c\xb0\xb7\x4d\x33\x2f\xbd\xa6\xb4\x77\x9d\x93\xc4\xde\x00\xa0\xdd\xc6\x7b\x5d\xb7\x32\x6e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293
    b"\xb5\x95\xc5\x4e\x71\xdc\x23\xc5\xc0\x90\xcc\xba\x52\xca\x55\x7a\x85\x2e\xdb\x94\x2b\x38\x82\xdc\xae\x2e\xda\x46\x39\xd9\xd0\xc6": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.293",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb5\x9c\x05\x2e\xe2\xd1\xd0\xea\xed\xbb\xcb\xa7\xb1\xe1\xf9\xa3\x61\xa1\xd8\xd5\x6b\x6a\xe8\x69\x2c\xf4\xe5\xf4\xd1\x1d\xe1\xb3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xb5\xb4\xe7\xb5\xeb\x1b\x5e\x40\x0f\x38\x44\x8e\xe5\x1b\x45\x21\xba\xc2\xb1\xf4\x05\x5b\x77\x31\x9f\xfa\x2a\xb5\x4c\xc2\xd6\x4f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb5\xb8\x41\x97\xd3\x20\xd7\xf8\xb8\x55\x62\x7d\xe8\x00\x47\x10\x70\xa3\xd6\x21\xde\xd2\x95\xc3\x5e\x7d\xbf\x5f\xc6\x54\xc3\x39": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xb5\xbf\x35\xd3\xa0\x23\xdd\x01\xad\xf9\x52\xec\x32\x3e\x84\xf9\x5e\xe4\x01\xb1\x71\x9b\xa1\x08\xe6\x2f\xbe\x94\x4f\xb3\x39\xa2": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.149
    b"\xb5\xc2\x82\x41\x96\xa5\x4e\x3d\xe3\x69\x2f\x7c\xc5\x20\x8c\x4b\x78\xc1\x40\xac\x39\x2d\xae\x19\x0a\x93\x66\x07\x51\x6e\xf8\xa1": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.149",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xb5\xf1\x84\x46\xb7\xba\xee\x40\x0e\x82\x6c\x06\xfe\x35\x13\x10\xb6\x08\x91\xa1\xf9\x9a\x81\x5c\x7c\x20\xdc\x02\x00\xf2\xca\x61": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/Hashes.sol",
            name="Hashes",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.225
    b"\xb5\xf4\x2f\x3c\x1e\x84\x8e\xfb\xb3\xad\x00\xab\xa8\x1e\x8a\xe1\x6a\x55\xce\x92\x99\x5b\x21\xb3\xc3\xbe\x25\xda\x31\xc7\x05\x2a": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.225",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\xb6\x17\x3a\x42\xa3\xf2\x62\x21\x15\x60\x2b\x1a\x5b\x05\x41\x10\x0a\x63\x2f\xa8\x28\xc2\xf3\xf4\x7e\x26\x99\xd5\x0c\x31\xaf\x39": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.4.0, v3.4.2
    b"\xb6\x1f\xd2\xb9\x36\x5c\x65\xe2\x37\x5a\x2d\xe0\x54\x0e\x7b\x68\xdf\x23\x81\x9c\xd1\xee\x6c\x37\xe2\x21\x33\x68\x35\x0f\x3c\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v3.4.2",
        ),
    ],
    # Hash appears in versions: main
    b"\xb6\x28\x05\xc1\x25\x3a\x58\x94\x44\x60\xb1\xe2\x34\xfc\x9e\x84\x44\x26\x7c\x7a\xa8\x95\x58\xb3\x3e\x92\xf8\x0e\xf1\xc1\x35\x51": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IPermit2Forwarder.sol",
            name="IPermit2Forwarder",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb6\x32\x98\x72\xb9\x73\xef\x23\xd9\xe6\x61\x4a\xdb\x82\x91\x03\x7b\x7e\x0d\xaa\x90\xf6\x18\xbb\x30\x31\x17\x04\x73\x7d\x59\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb6\x45\xec\x62\x79\x24\x8a\xbd\x54\x62\xb9\xf1\x28\x14\xe7\x0b\xc3\x03\xa1\x30\xa2\xd0\xa9\xda\x77\x4b\x25\xfe\x7e\x40\x0d\x3a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointStake",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointStake",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointStake",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointStake",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointStake",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xb6\x47\x38\xe1\x8a\xbf\xd2\xa7\x34\xe3\x04\x2f\x2a\xfc\x33\xed\xfe\x8e\x0f\xee\x69\x72\x85\xf0\xc0\x5c\x45\x3d\x62\x61\x01\x33": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb6\x4a\x0c\x91\x56\xdc\x62\x15\xe2\x64\x18\x4b\xe8\xc5\x69\xd3\x62\x82\xca\x48\x97\xeb\xe1\xf0\x4f\x45\xe5\x64\x11\x0f\x31\x33": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/LibOptimism.sol",
            name="LibOptimism",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: main
    b"\xb6\x7c\x44\xd3\x1b\xdd\x7d\xda\x63\x50\xe9\x21\x00\x9c\xe6\x0a\x31\x3f\xcc\xfc\xea\x65\x0d\xc5\x12\xbb\x2d\x62\xbd\xb1\x1a\x7c": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/ERC721PermitHash.sol",
            name="ERC721PermitHash",
            version="main",
        ),
    ],
    # Hash appears in versions: main
    b"\xb6\x81\x7b\x39\xc1\xfa\x9c\x2e\x13\x70\x8a\x1e\xdb\xd2\x5a\x26\x22\xc9\x53\xd7\xa6\x89\x5e\xb3\x55\x71\xcc\x05\xab\x49\x9a\x9a": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/SafeCurrencyMetadata.sol",
            name="SafeCurrencyMetadata",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb6\xe2\xd1\x90\x17\xa4\x90\x84\xd1\xa8\xe3\x9f\xec\x57\xe8\xeb\xc5\x70\xce\x5f\xb7\x0a\x4f\xb7\xc8\x56\xdc\x88\x89\x84\x4a\x19": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\xb6\xef\xff\xf9\xe2\xf9\x0c\x6d\x4c\x01\x9b\xfb\xf4\xb9\x88\x52\xb4\x1d\x73\x22\x11\x5e\x17\x53\xb6\xab\x38\x87\x48\x3e\xec\xe5": [
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb6\xfe\xe1\xa9\xeb\x16\x8b\xe3\xde\x37\x1f\x64\xfc\x60\x2a\x62\x6a\x45\x0b\x54\x34\x5e\x0b\x54\x06\x66\x36\x53\xbd\xa1\xca\xaf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xb7\x88\x89\xaa\x4c\x43\x35\xc2\xfa\x40\xd7\xb9\xb8\x95\x48\xf8\xfe\x81\x10\x61\x81\x0a\xa4\x36\x73\x17\x31\xf2\xb1\x06\x07\xb0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.8.1, v4.8.2, v4.8.3
    b"\xb7\x89\x8d\x67\xcc\x85\x69\xcf\xf6\xb6\x7b\x16\xf5\x87\xbb\x49\xe8\xd6\x28\x1e\x56\x1a\x33\x20\x2c\x4d\x5b\x25\x84\x75\xb1\xf6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xb7\x94\xa9\xb4\x32\x31\xdd\x21\xe0\xa5\xed\x72\xc2\xac\x04\xcb\x10\xbc\xfe\x5a\x84\x3b\x5c\xd2\x3e\xa1\x24\x2d\x6b\xd2\x40\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xb7\x97\xa1\x30\x7c\x20\x85\x99\x44\x1a\x28\xc6\xab\x2f\x9d\x42\x33\x2e\x37\xdc\x30\xe8\x94\xec\x3f\x9d\x15\x70\xd6\xb1\x80\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xb7\xc9\xc4\x50\xa3\x5b\x05\xb8\xa3\xae\x18\x16\x9a\x6c\x4b\x09\xe2\x59\x14\x08\x70\x9e\x49\xf8\x69\x3c\xd1\x32\xe2\xb2\xfe\x10": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/Locker.sol",
            name="Locker",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xb7\xd0\x50\x8e\x67\x9b\x47\xa6\x8b\xb7\x99\x25\xc2\xc2\x69\xcf\x97\x5a\x63\x1f\x92\x1f\x1f\x88\xb0\xda\x4e\xea\x98\x97\x21\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.127
    b"\xb7\xd2\xd4\x17\xab\xa2\xbd\x4e\xca\xe5\x01\xd4\x4b\xf8\x03\x31\x7d\x94\xa2\x0d\x55\x72\xd0\x1b\x90\xf2\x05\x27\x17\x37\x28\xa2": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.0.127",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\xb7\xef\x77\xde\x14\x84\x6f\x6c\x78\x69\xc9\xf1\xf4\xf9\x4e\x73\x80\xc7\x0d\xfa\xa1\x5a\x33\xfb\xfc\x84\x42\xcc\xcf\xcb\x0d\xd6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb7\xf8\xc1\x01\x5b\xf3\x85\x99\x8f\xff\xe1\x8d\xdd\xb3\x19\x24\x79\x86\x85\x3d\xa6\x30\x79\xfc\xdb\x32\x06\xd6\xa0\x25\x55\xcd": [
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2
    b"\xb8\x05\x07\x83\x0d\xb6\xb7\x73\xb1\x75\x5b\xdc\x1a\x8e\x56\x92\x92\xdb\x6d\x96\xad\x93\x7a\xb1\x8f\xec\x08\x55\x84\x5a\x94\x5a": [
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/Owned.sol",
            name="Owned",
            version="v6.6.2",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb8\x15\x79\x2a\x33\xeb\x06\x86\x2f\x3e\x19\xe6\x7d\xd0\x28\x1e\x0a\xee\x22\x59\x9b\x0c\x90\xc1\x33\x92\xf9\x04\xe2\xc5\x78\x19": [
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.39
    b"\xb8\x19\x95\x45\x92\x14\x79\xf5\x67\xef\xfd\xbc\x76\x36\x45\x72\xcf\xe8\xd0\x2d\xde\x6a\x24\xed\x32\x0a\x8a\x0b\x7e\x37\x85\x28": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.39",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xb8\x1f\x75\xe7\x6b\x1b\xb8\x3a\x0d\x6f\xe2\x2e\x5f\x1e\x63\x15\x61\x61\xc9\xc8\x8e\x9b\x10\xe2\x88\x5a\xbc\xe1\xcc\x8a\xa8\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1
    b"\xb8\x2c\xbf\x01\x78\x5f\x8b\x28\x99\x55\x30\x1f\xa0\xf8\x4c\x51\xf9\x39\x19\xa2\x2f\x8d\x3d\x57\x21\x1a\x1e\x62\xf9\xc8\x1a\x7b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.4.1",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.9
    b"\xb8\x32\xff\x85\x55\x3e\x86\x2e\xbc\x7c\xe8\x4f\x3b\x2b\x71\x57\xfc\x73\x8b\xd1\x28\xb2\x95\x16\x70\xb1\x8d\xb4\x07\xf6\xbd\x11": [
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibEIP7702.sol",
            name="LibEIP7702",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xb8\x36\xd8\x8c\x8a\x17\xee\x49\x5c\xee\xe1\x4a\xa1\xcd\x80\x5a\xa0\xc0\x2c\xb6\x0f\x55\x44\x28\x25\xb9\x05\xd8\x8e\x6c\xac\xf1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123
    b"\xb8\x82\xd1\xd8\x91\x9e\x24\x56\x4d\xa9\x65\xa5\x9d\x0e\xcc\xf6\xf2\x1a\x17\x69\xd1\xb0\xd6\x5f\x7f\xbb\x7e\xcb\x2a\x42\xa6\x27": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.123",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb8\x84\x19\x98\xfd\x38\xa2\x0b\x22\xf6\x7f\xc9\xc8\x2e\xf6\x2c\xc6\xb9\x08\x4e\x81\xa9\x38\x0d\xa0\x87\x76\x84\xbe\x52\xfa\x69": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Spender.sol",
            name="IERC1363Spender",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7
    b"\xb8\x8f\x8e\x2c\xf6\xb3\x8a\x47\x74\x7d\x17\x19\x2a\x8c\xb8\x7a\xea\x85\xd5\x8d\xe4\x8a\x4f\x6d\x6e\xda\x88\xaf\x00\x95\x3c\x95": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xb8\xa1\x39\xef\xcb\xfc\x99\xf9\x8d\x05\x32\x44\xd2\x37\xe7\x21\x5f\x82\x74\x14\xcb\xc5\xaf\x19\x8a\x03\x81\xb3\x2d\x66\x9f\x2a": [
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xb8\xa1\xce\x49\xcb\xed\x91\x1e\xfa\xf0\x67\xed\x13\xb9\xad\x6e\xb8\xe8\x2e\xbd\xac\xca\xf5\x6e\x25\xf0\x06\x91\xf7\xb2\xa4\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/transparent/TransparentUpgradeableProxy.sol",
            name="TransparentUpgradeableProxy",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0
    b"\xb8\xbe\x58\x8d\x66\x5d\x90\x08\x67\xac\xd4\x05\x38\x22\x43\x3f\x4f\x19\x32\x16\xa9\x27\x68\xa5\xb1\xfe\xd8\x37\x7b\x4d\x95\x42": [
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v6.7.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xb8\xc5\x55\xea\x18\x7f\x36\xd7\xf5\x06\x9a\xf5\x92\x55\xeb\x3f\xf3\x87\x4d\xf6\x96\x4b\x12\xff\x33\x88\x9f\xc0\xa2\x15\x41\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294
    b"\xb8\xce\x14\xa9\xdb\xfd\x64\xe5\xcd\xd8\x94\xbd\x5a\x7e\x5f\x5d\xf9\x49\xc3\xd9\xd0\x78\x0f\x56\x25\x87\x36\x6d\x13\xd9\x5b\xf4": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.294",
        ),
    ],
    # Hash appears in versions: v0.1.5, v0.1.6
    b"\xb8\xd6\x53\x66\x30\x2e\xef\x3c\xf2\xd6\x0f\x12\xbb\xa1\x3a\xcd\x9c\xd0\xb2\xe8\xbc\x01\xbb\x85\x12\x9f\x43\xc5\x04\xb3\x37\x16": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.1.6",
        ),
    ],
    # Hash appears in versions: v0.0.251, v0.0.252, v0.0.253
    b"\xb9\x10\x2e\x88\x56\xc0\x16\xe8\x20\x9c\x04\x33\xf0\xf5\xd5\x9f\x61\xe7\x2e\xe3\x17\xaf\x25\xa0\x33\x6c\x24\x89\x28\x69\x74\xc7": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.253",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0
    b"\xb9\x11\xa9\x18\x82\x22\xd5\x22\x49\x70\x82\xec\x78\xe7\xf5\xb9\x98\xde\xac\x2a\xbc\x95\x96\xa8\x5d\xb0\x97\x02\xe5\x60\xaf\xf5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v4.8.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xb9\x49\x2b\x0f\x7f\x83\xcc\x1b\x42\x7d\x2d\x3b\xed\x0f\xa5\x4a\x7d\x47\x3a\x2a\xf6\x1e\xf6\x00\x95\x5d\xfd\x20\x3a\x9f\x4d\xd3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xb9\x4f\x8b\x52\xad\x47\xa1\xf2\x55\x6b\x02\x4d\x9a\x37\x68\xfc\x26\x49\x74\xf0\x79\x49\x4f\x70\xf0\x8c\x6d\xd4\x05\x2e\x26\x70": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.3
    b"\xb9\x50\xfa\x4d\xb3\x21\x1a\x9c\x33\x6d\xd4\xb6\xc3\x46\x49\xb2\x0e\x79\x21\x16\x66\x98\xbb\xb4\x6e\x59\xec\x3c\xa9\x78\xee\x3c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xb9\x75\xef\xab\x21\xfc\x4d\x82\x08\xd7\x86\xb8\xd0\x02\x15\x4e\xa8\xe8\x64\x1e\x49\x51\x5f\x78\x9a\x04\xc7\x3c\x67\xcf\x45\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xb9\x88\x84\x0b\xbe\x90\x25\x17\x2b\x2c\x80\xc4\xfc\xca\xca\xf2\x99\x80\x0a\x9a\x54\x05\x62\xc9\xb3\xee\xdf\xad\x38\xb2\x24\x17": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xb9\xc3\x41\x2b\x12\x30\xea\xfd\x21\x28\x8d\xb3\x26\x7a\x9d\xd3\xc3\xe9\xca\xf4\xd2\x91\x4a\x4c\xba\x46\xb6\x73\xf3\xe3\x76\x4c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20.sol",
            name="IERC20",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210
    b"\xb9\xde\xd0\x15\x6b\xa7\x19\xee\x5c\x89\x81\xdd\xd6\xdf\x38\xae\xfd\x92\xd7\x5c\x89\x99\x02\xf9\x66\x91\x07\x06\x9a\x72\x9a\x2b": [
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/WETH.sol",
            name="WETH",
            version="v0.0.210",
        ),
    ],
    # Hash appears in versions: v0.0.61
    b"\xb9\xf2\x17\xc3\xa7\xa0\x82\xbb\x5b\x72\x98\xfe\x76\x0d\x3c\xbf\xa3\x44\x96\x87\xb0\xf3\x6a\xa9\x77\xe9\x4b\x75\x98\xca\x12\x5f": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.61",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xb9\xf8\x7d\x4a\x67\x51\x29\x1d\x45\x05\x64\x7a\xa1\xe3\x9c\x75\x5e\x98\x96\x4c\x73\x3f\x8d\xc2\x43\x73\x43\xa0\xd0\x32\xb4\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156Upgradeable.sol",
            name="IERC3156FlashLenderUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1, v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xba\x04\xf4\x63\xf6\x39\x8c\xd8\xe7\x29\x59\x92\xe5\x57\x61\x14\x86\x8f\x39\x42\x43\x53\x35\xa6\x9e\xb2\xef\xbd\x71\x77\x55\x31": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC20Metadata.sol",
            name="IERC20Metadata",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xba\x0d\xa8\x36\xfb\x29\x9f\xae\x61\x09\x72\x29\x55\x46\xc5\x58\x74\x63\xae\xf5\x1c\xa3\xf0\x0d\x61\x0e\xe2\x37\xab\xce\xc6\x8c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xba\x39\x21\x16\xf2\xcc\xc2\xbd\xef\x75\x85\x55\x6c\x37\xb6\xca\xa4\xb6\xca\x89\xcd\x72\xfa\xeb\xea\x68\x9b\xe6\xff\x8b\xe7\xa6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesQuorumFractionUpgradeable.sol",
            name="GovernorVotesQuorumFractionUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\xba\x52\xba\x71\x89\x5a\xba\x81\x31\xfb\x4e\xdd\xc3\x35\x82\x61\xc5\x20\xcc\x78\x96\x78\x60\x55\x5a\x2c\x9c\x99\xfd\xce\x70\x54": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.281, v0.0.282
    b"\xba\x5a\xe9\xaf\x57\x63\x02\x47\x9d\x7c\xf5\xdc\xca\x5a\x04\x2d\x44\x7d\x3a\x37\xa2\x4c\xa7\xf4\xfe\x7e\x3e\x9d\x51\xf3\x07\x0c": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xba\x64\xd4\xdb\x0b\xc9\xef\x13\xc5\x79\x13\xf2\xbb\x6d\x97\xc9\xb0\x9d\x28\x46\x0e\xf1\x62\xdf\xf2\x54\x84\xf9\x2c\xe8\x9e\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/MathUpgradeable.sol",
            name="MathUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xba\x68\x71\x60\x4c\x37\x9e\xa0\x89\xee\x00\x69\x27\x04\x61\x93\x2e\x57\x8c\x69\xbf\x05\x69\x69\xfb\xaa\x3f\xdd\x43\x85\x50\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0
    b"\xba\x70\xe8\x73\x9f\x11\x3c\x2c\xc9\x93\xc5\x93\x82\x48\xed\xd9\x66\xf2\x5f\x8f\x4a\xee\x32\xc4\x26\x36\xc6\xc0\xed\x53\x58\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="IAccessControlEnumerableUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xba\xa2\x69\x82\xf9\xf5\xe5\x9f\x60\x52\xc7\x13\xe3\x85\x4d\xba\xf2\xc6\x8c\xaf\x4f\x62\xa3\x7a\xb7\x8c\xa9\x60\x58\x32\x84\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\xba\xd7\x18\xc2\x2e\x23\x31\x4f\x7d\x66\x99\x29\xd3\x88\x09\x53\x59\x60\x2e\x21\xf5\x52\x1c\xd6\xee\xb3\x13\xd8\x98\x9d\x86\x02": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xba\xd7\x7a\x5a\xdf\xb5\xdf\xb2\x09\x87\x57\x13\xbf\xcc\xc2\x19\x26\x74\xc8\x19\xac\xb6\xa4\x3a\x08\x0b\x48\xc7\x94\xac\x0d\xdd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xba\xe6\x58\xc7\x1e\xd8\xab\x0d\xec\xb6\xcb\x56\x46\x1a\x88\x67\x66\xa7\xbb\x48\x3e\x3f\xc4\x0e\x47\x3d\x18\x52\x5f\x50\x14\xa9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL2.sol",
            name="CrossChainEnabledArbitrumL2",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xba\xef\xa1\x10\x36\x11\x75\x15\x18\x44\xe6\xbd\xe9\x25\x05\xeb\x22\xfa\x78\xc3\x08\x07\x2b\x1a\x0d\xf0\xbe\x25\xe8\xbd\x08\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xba\xf3\x8f\xc1\x21\xd4\xbb\x14\x15\xd2\xc6\xcf\xbb\x38\x65\x0b\x56\xe8\x7a\x0d\x9e\xd3\x66\x4c\x36\x95\xf7\x7c\x79\xce\x04\x57": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/types/Currency.sol",
            name="CurrencyLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0
    b"\xba\xfa\x0b\x09\x07\xa2\x0d\xeb\xfe\xe7\xb7\x2b\x91\xb4\x9e\x3d\xaa\xed\x30\x2b\x40\x53\x92\x38\x28\x95\x35\xa0\x7e\x56\x09\x67": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.4.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234
    b"\xbb\x05\xe9\xa5\xcc\x10\x51\x16\x5a\xaa\x9d\xa8\xcb\x0c\x56\x0a\xe7\x8c\x90\xb2\xf7\x11\xf4\x7c\xd8\x74\x85\x68\xb9\xe5\x73\x79": [
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.234",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xbb\x15\xa1\x41\x81\xae\x29\xf6\xfe\x07\x81\xa5\xf4\x41\x0e\x54\xd0\x36\xfa\x9f\x98\xef\x4f\xc2\x7d\xb8\xc9\x64\xd4\xa8\xf3\x4f": [
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Multicallable.sol",
            name="Multicallable",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xbb\x15\xc9\x2d\x9c\xa1\x2b\x07\xd8\x5d\x67\x0e\xd1\x26\x16\x9e\x57\x4b\xc7\xcd\x6b\x0f\xe5\xf7\xc5\xf2\x40\x87\x20\x47\xaa\x1b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xbb\x15\xef\x79\x10\x4b\x51\x00\xa1\x0c\x9c\x7b\xa1\x94\x82\x90\x19\xee\xc5\xbb\x8e\x15\x63\x9a\xa3\x1e\x3e\x07\x0a\x14\xff\x2f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xbb\x5d\x00\x43\x71\x46\x61\x6a\xeb\x2e\x75\x30\xd6\xae\x17\x1f\x50\x63\xec\x25\xd7\xba\x04\xb8\xda\x7b\x37\xae\x78\x69\xf1\x8a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.193
    b"\xbb\x70\x97\x8a\xdb\x26\xae\xd7\x01\x01\x7e\xff\x9c\xc8\x93\x59\xf5\x14\x32\x22\x69\x30\x76\xc3\xfb\xd1\x75\x30\x6b\xa3\xd3\xbb": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.193",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xbb\x8d\xbd\xd8\xc5\x78\x4e\x9a\x05\xca\xa8\xf6\x85\xb4\xb7\x72\xa6\xa5\x4a\x9d\x5f\x32\x6f\xd9\x21\x38\x4e\x5e\x5d\xdc\x7f\xa3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: main
    b"\xbb\xa2\x2f\x58\x23\x72\x5f\xc9\x9e\x8a\xe7\x40\xa5\x86\x06\x19\xb8\x31\x23\xae\x37\x9c\x27\x94\x72\x79\x8b\x2e\xc4\xf2\xaf\x01": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/external/IWstETH.sol",
            name="IWstETH",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.232
    b"\xbb\xae\xb1\x10\xaa\xce\x60\x36\x39\xa6\x56\x01\xf3\xec\x17\x14\x02\x6d\x6f\xec\xb5\xb8\x41\x22\xc4\xf3\x33\x68\x6e\xfb\x85\xc3": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.232",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xbb\xb8\xc9\x0f\x6c\x46\xf3\x0e\xa9\x8d\xb7\x07\x4c\xf5\xe7\xb1\x9e\x10\x1b\xd2\x17\xcb\x24\xdc\x14\xe2\x4d\x1e\x86\xd9\x6c\x57": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xbb\xba\x1a\x6a\x34\xff\xd2\x18\x60\xf0\x30\xfa\x30\x47\x62\x61\x01\x7d\xf8\x4e\x10\x1b\xa0\x50\x33\xfc\x77\xf1\x98\xfc\x40\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777Sender.sol",
            name="IERC777Sender",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Sender.sol",
            name="IERC777Sender",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xbb\xe9\x75\x5a\xe3\x45\xeb\x5d\x54\x4a\x83\x11\xfe\xd6\x49\xf1\x3c\xe3\x8b\x98\x95\x6b\xb0\xbc\x2f\x47\xbc\x56\xec\xd3\xf5\x8c": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC721Permit.sol",
            name="IERC721Permit",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC721Permit.sol",
            name="IERC721Permit",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC721Permit.sol",
            name="IERC721Permit",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC721Permit.sol",
            name="IERC721Permit",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IERC721Permit.sol",
            name="IERC721Permit",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xbc\x39\x87\x60\xe9\x15\xbc\x84\xce\x65\x45\x8f\x24\x35\x5a\x36\x1f\x58\x32\xae\xb7\x37\xfb\x48\xa4\xa2\x29\x64\xb1\x10\x57\xfa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/presets/ERC777PresetFixedSupply.sol",
            name="ERC777PresetFixedSupply",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79
    b"\xbc\x4e\x19\x61\x56\xff\x65\x87\x72\x3b\xc6\x58\x78\x85\xec\x4a\x60\x7a\xa0\x52\xa5\xe4\xfd\xad\xd6\x5a\x24\x6c\xfc\x10\xde\x89": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.79",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.0-rc.0, v3.4.1
    b"\xbc\x51\x1a\xc4\x56\x55\xdb\x32\x9d\xae\xec\x82\x73\x57\x13\x4b\xbc\x04\xf9\xc4\x99\x21\x01\xbb\x08\x28\x98\x5b\x54\xc4\xc9\x1f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xbc\x5b\xe7\xb6\x12\x87\x2a\x19\x85\xd8\x61\xfe\x81\xd2\x19\x52\xa2\xc2\xbe\xed\xc9\x03\x72\x36\x6a\xba\x80\xcf\x06\xae\x37\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.1.0-solc-0.7
    b"\xbc\x7a\xfc\xaf\x60\xc3\x41\xd2\x8e\x93\xe9\x33\xcc\x54\xe7\x00\xd3\x77\x94\x0a\xca\x18\x20\x71\x23\xc8\x7c\x06\x7b\xaa\x09\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.1.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298
    b"\xbc\x94\x38\x58\xc6\x0e\x55\x3e\x00\x35\x97\xcf\x1a\xc4\x72\xa2\xa2\x92\xb9\x58\xac\x7b\x2d\x59\x8b\x09\x73\x7c\xd3\x58\x46\xd3": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.298",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xbc\xb6\x08\xcb\x40\x8b\x72\xc6\x90\x94\xc6\xf4\x75\x20\x3e\xda\x82\x61\x43\xb4\xdf\xc0\xc3\x1b\x27\xf4\x08\x11\x56\x3c\x7f\x0c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesCompUpgradeable.sol",
            name="GovernorVotesCompUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.3
    b"\xbc\xdf\x74\xc9\x20\x94\x40\xf8\x93\x16\x37\xe9\xc2\x85\x8d\x4f\xac\x4b\x30\xbf\x3e\x4b\x2c\x88\xac\x7d\x19\xfa\x5c\x27\x0b\xef": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: main
    b"\xbc\xe4\x25\x2c\xa0\x84\xe3\xa6\x2b\x72\x0e\x49\xde\xc8\x13\x00\x0a\x15\x01\xff\x5f\x11\x38\xfc\x3e\x3a\xde\x65\x82\x8d\xa7\x3f": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/PositionInfoLibrary.sol",
            name="PositionInfoLibrary",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.0.0, v3.0.1, v3.0.2
    b"\xbc\xed\x0d\x16\x68\x42\xb9\xa2\xf3\x3a\x14\xf7\xe8\x44\x20\x62\xc4\x34\xa6\x01\xff\xb2\x32\xd7\x10\x47\x8f\x8c\xaa\x92\x23\x19": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xbc\xf5\xf9\x4d\xf8\x74\x35\x54\xd1\x50\xd5\xd5\x0d\x88\x95\x83\xd2\x33\x1f\xb8\x63\x13\x89\x36\x2f\x5e\x18\x2c\x2b\x59\x7f\x4a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0
    b"\xbd\x05\xc7\xd6\x13\x2d\x7b\x7d\xb9\xe0\x52\x9b\x2e\xf0\x4a\x58\xb7\x32\xa6\xe7\x3f\xe9\x62\xcd\x81\x73\xb8\x9c\x23\xc2\x2b\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v4.8.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.1.13
    b"\xbd\x19\x53\xfe\x81\x30\x59\xc0\x08\x3d\xe4\xc1\x4f\x67\x82\x74\x8d\x6d\x4d\x4b\x7f\x7c\x53\x40\x92\x04\xb4\xc4\x6f\x52\x2c\xd9": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.13",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xbd\x1a\xdf\x34\xb5\x3b\xfb\xca\xae\x6d\x8f\x49\xde\x89\x29\x01\x44\xa0\xe1\x94\xb0\xeb\x0b\x3e\x62\x00\x46\xe2\xed\x8a\xf7\xb9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC1363Upgradeable.sol",
            name="ERC1363Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.19
    b"\xbd\x20\x69\x74\x0c\x15\x14\x08\xb6\x8c\xf2\x0e\xe2\x8e\x0a\x9d\x13\x2e\xae\x32\xfc\x10\xa0\xdf\xb1\xca\x7a\x1c\x4d\x27\x83\x9c": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.19",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xbd\x2c\x06\xdc\xad\x59\xea\x3c\x7a\xff\x1c\x89\x23\xb4\xd5\x0a\x3b\xc2\xb3\x67\xdb\x54\x74\x60\xf7\x20\x08\xa1\x86\x4f\x88\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPaymentUpgradeable.sol",
            name="PullPaymentUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xbd\x43\xcb\x25\xbb\x0a\x5d\x63\x42\x8e\xc3\x5a\x57\x37\x64\x6a\x6a\xf6\x68\x0a\xf1\x5c\xdb\xd3\xe3\x95\x4d\x3b\x03\x8e\xb3\xbe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xbd\x47\x4c\xc9\x13\xef\x9a\xcd\x52\x80\x68\xad\x16\x04\x15\x71\x36\x84\x6a\x69\x14\x7d\x40\xcf\x6c\x17\xd6\xeb\xa1\xbb\xa1\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.276
    b"\xbd\x7b\x32\x7f\xf4\x3c\xbe\xb6\x3f\xed\x81\x17\xb0\x15\xb6\x20\x1b\xd3\x94\xfd\xb4\xaa\x57\xa8\x0d\x10\x24\xeb\x28\x01\xfa\x34": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.276",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xbd\x84\xd4\xb5\x54\x6c\xe5\xa4\xb7\x00\xf4\x4b\xaf\x0b\xac\xcb\xa2\xba\x40\xc3\x10\x71\x5e\x0b\xac\x12\x77\xdd\x1a\xf9\x3e\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Calldata.sol",
            name="Calldata",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Calldata.sol",
            name="Calldata",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xbd\x9e\x19\xc0\x2c\xf6\x64\xb5\x2b\xe5\x3c\xc2\x2f\x79\x14\x2d\x2d\x25\x7f\x69\xf7\xfc\xef\x01\xa2\x2c\xc8\xea\x81\x1a\x16\xd0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xbd\xaa\x74\xcc\xd1\xda\x15\xa9\x8e\xa0\x5f\x93\xce\xae\x22\xd0\xc4\xed\x19\x35\x09\xeb\x73\xdc\x41\xc0\x3d\x2f\xaa\x3d\xa1\x87": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20BurnableUpgradeable.sol",
            name="ERC20BurnableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7
    b"\xbd\xaf\x26\xe1\x17\x96\xa0\x93\xb6\xb2\xb3\x9f\x58\x8d\xd9\x29\x10\x06\xc8\xc6\x58\x54\xa2\x76\x96\x13\xab\x2e\x9f\x64\x54\x0b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableSet.sol",
            name="EnumerableSet",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291
    b"\xbd\xb7\xab\x32\x58\xce\x74\x28\x6f\x29\x86\x1c\x16\x45\xe8\x99\x84\x17\x9c\xc6\x7c\xa7\x92\xa5\x16\xe2\xa6\xe2\x5e\x67\x2b\xe9": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.291",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xbd\xe9\x72\x78\x6d\x95\xe5\xe0\x98\x1f\xe3\x73\xf6\x9b\xc9\x14\x86\x02\x51\xbb\x64\x32\xa7\x74\xb8\xac\x97\x75\x7a\xed\x56\xb7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Votes.sol",
            name="ERC20Votes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xbd\xe9\x89\xc5\xfb\x4f\xd9\xe0\x82\x99\xad\x74\xf9\xcc\xb0\x97\x76\x8c\xb0\xb1\x0e\xc2\xc1\xfa\x84\x9e\x48\x9f\xf3\xfb\xc1\x6b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingOverridableUpgradeable.sol",
            name="GovernorCountingOverridableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingOverridableUpgradeable.sol",
            name="GovernorCountingOverridableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xbd\xfb\xcb\x40\x6b\xcb\x89\xce\xee\x3b\xb2\xf7\xff\xe9\x99\xef\xa0\xef\x7a\x64\x00\xfb\xe7\x1e\xc2\xdd\xca\x8d\x7c\x78\x2c\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.3.0
    b"\xbe\x16\x18\x68\x07\xed\x29\x89\x8f\xba\xfd\x44\x48\x5d\xaf\xdf\x97\x2a\xaf\x21\x10\x85\xb8\x98\x3a\x7e\x59\x18\xf1\x67\xca\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC4337Utils.sol",
            name="IEntryPointExtra",
            version="v5.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xbe\x21\x42\xfc\x8e\xed\x1c\x36\x78\x0a\xe7\x38\xe7\x03\xdc\xc1\x8a\x9a\xb8\x4f\xe0\xfc\x7f\x20\x13\xaf\x24\x36\xb5\xdf\x60\x33": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xbe\x30\x06\x91\x26\x2f\xb5\x3f\xb2\xb5\x18\xb6\x17\x89\xd7\xab\x63\x2b\xd9\x04\x6d\xb0\xe9\x98\x50\x83\xc2\xaa\xde\xcd\x52\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xbe\x57\x4e\x35\x9b\xd8\xdc\x3e\x17\x41\x6e\x8b\xa4\xb2\x75\x2b\xa1\xc3\xe4\xa7\x32\x0c\x35\x41\x3b\xcc\x39\x67\x85\x25\xab\x23": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/base/ImmutableState.sol",
            name="ImmutableState",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\xbe\x71\xc7\x84\x6c\x7e\x89\xa4\x8d\x5e\x00\xa5\x6f\x4d\xe5\x1f\xe4\x31\x52\x7d\x13\x46\x4f\xd7\x05\xf3\xfd\x35\xa3\x88\xad\x5d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xbe\x73\x57\x70\x77\xb6\xa7\x89\x96\x59\xc2\x9b\x74\xac\x49\xaa\x1a\x25\x4c\x63\xc5\x7d\xab\x40\x0f\xff\xa6\x08\xd8\x00\xef\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xbe\xbc\x59\x0a\x60\xab\xd9\x61\xd3\x51\xfe\xf3\xb1\xe1\x85\x19\x51\x26\x54\xae\x70\x26\x8c\x19\xf1\xc9\x6c\xaf\x79\x0d\xbe\x89": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xbe\xd6\x80\xb9\xbb\x37\x97\x47\xc0\x18\x41\xd9\x0e\xec\x3f\x47\x3d\xca\x0f\x7f\x6f\xd3\xff\x44\x62\x84\x45\x19\x71\x1c\x63\x61": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/amb/CrossChainEnabledAMB.sol",
            name="CrossChainEnabledAMB",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.1, v4.7.2, v4.7.3
    b"\xbe\xdd\x9e\xcb\xee\x80\x11\x53\x0f\xc5\x1b\x03\xd1\x60\xeb\x25\x80\xf9\x17\xa3\x2e\xdd\x3b\x46\xb1\x81\x34\x02\x3e\x3a\x9a\x03": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xbe\xec\xef\xc7\x0a\x94\x37\x93\x91\x6b\x52\xd7\x14\x41\xe7\xff\x4e\x24\x3d\x3b\xfd\xa2\x17\xd5\x33\x59\xe3\xe3\x59\x15\x68\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4906Upgradeable.sol",
            name="IERC4906Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xbe\xf2\x46\xe1\x2e\x30\x1c\xbb\x6c\x38\x76\xe0\x19\x17\x5c\x5f\x91\x45\xbf\xf2\x0b\x10\x00\x8e\x8f\xfc\x7f\x01\x85\x50\x91\x99": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IInboxUpgradeable.sol",
            name="IInboxUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\xbe\xf3\x14\x94\x04\xe9\xb3\x56\x02\x22\x8c\x95\x91\x5e\xb7\xef\xa5\x2e\x30\xe4\x21\x5a\x42\x5f\x51\x7d\xc9\x70\xdb\x21\xea\x3f": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xbf\x02\x40\x0d\x5d\x4c\x4e\x2a\x2e\x4f\x96\x3e\xac\x70\xb2\xd6\x3e\x27\xdc\xb2\x4c\x4a\x54\x3a\x21\xf5\x6a\x6f\xa2\xfa\xe8\xeb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xbf\x17\x16\x38\x9b\x74\x4a\x8e\x34\x1b\xd4\x34\x18\x46\x17\x2f\xbb\x2d\x19\x4a\x50\xb6\x06\x1b\x5a\xc3\xa0\x50\x83\xe3\x64\x94": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xbf\x1e\x4d\x37\xe3\x8f\x7e\xaf\xa5\xa9\xd5\x51\x48\x95\xd8\x50\x7d\x84\x3c\xae\x90\xdf\xab\x7b\xc1\x2c\xe6\x11\xd8\xda\xdc\xe0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.15, v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20
    b"\xbf\x24\x81\x7d\x84\x95\x5c\xeb\x99\x64\xc6\x5e\xb4\x18\xcf\x3a\x3e\x65\xfc\x29\x54\xa3\x10\xda\x7d\xce\xb9\x73\x89\x2f\xb4\x7a": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.20",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xbf\x25\x59\xfc\x28\x2d\x51\x05\xee\x03\x5c\x61\x72\xf4\x3e\x19\x4b\x04\x8c\x77\xfa\x02\xd5\x51\x1d\x3e\xf2\x3c\x9f\x4b\x53\x39": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/ERC1967FactoryConstants.sol",
            name="ERC1967FactoryConstants",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xbf\x37\x53\xdf\xc8\x04\xb2\x22\x3b\x12\x42\xcc\xd4\x05\xc8\xe2\x30\x59\xa9\x27\xf7\xb6\xcb\xd1\x32\xbc\x4a\x68\x1b\xb9\x52\x00": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73
    b"\xbf\x54\xae\xdd\x9a\xab\x8a\x50\x5d\x92\x3d\x3d\x49\xdc\x7d\xed\x18\xdd\x21\x50\x93\xb1\x18\xd2\x8d\x23\xa8\x81\x44\x1d\x1e\xec": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.73",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xbf\x75\x96\x45\x89\x16\xb3\x65\xfe\x6a\x9c\x82\x17\x1a\x8b\x82\x05\x54\xa0\xd6\x79\x89\x8b\x48\x4c\x0a\xa0\x0c\xc3\xc7\xb0\xd2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/StringsUpgradeable.sol",
            name="StringsUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xbf\x81\xde\x73\x94\x3a\x43\xb9\xf2\x69\x6e\x7f\x76\x35\x57\x5b\xf9\xd0\x1c\x9d\x09\x89\x91\x25\x14\x0e\xa7\xd9\x1c\x30\x02\xd9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xbf\x95\xf0\x64\xf0\xaf\x49\x61\x70\xa6\x3f\x28\xee\xb2\x80\x49\xf6\x85\x21\xfd\x22\x21\xaf\x19\xc9\x0a\x3d\xd1\xb8\xc5\xe6\x09": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xbf\xa5\xe8\xef\x46\x74\x1a\x5b\x2e\x9c\x59\x58\x68\x60\xa8\x82\x0b\xbb\x55\xdc\x29\x1c\x1f\x04\x4c\x11\xbd\x54\xff\x36\x6d\x0a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721BurnableUpgradeable.sol",
            name="ERC721BurnableUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xbf\xaa\x2f\x8e\xfd\xa0\x0d\xbb\x8c\x79\xb0\x72\xb1\x5c\xa9\x1c\x52\xf0\x0d\xf4\xbb\x60\xbd\x3b\x43\xc7\xbe\x73\xab\x2e\x3a\xe9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xbf\xaa\x88\x44\x54\x3e\xa7\x97\x13\x03\x6d\x42\xa4\xcc\xc3\x5c\x6b\x63\xb5\xc1\xb9\x74\xa9\xf9\xb9\xf6\xbf\x26\x45\x5c\x5b\x0b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Receiver.sol",
            name="IERC721Receiver",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1
    b"\xbf\xd8\x0f\x6c\x77\xd9\xb9\x1e\x0b\x9d\xc1\x64\x07\x11\xb3\x3b\x47\x98\x14\x40\x88\x01\x51\xa2\x61\x8b\x65\x6c\xb7\x3a\x0e\x9d": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTSVG.sol",
            name="NFTSVG",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/NFTSVG.sol",
            name="NFTSVG",
            version="v1.0.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xc0\x01\xfa\x4a\xb7\xc7\xf4\xe9\x90\x44\x66\x10\xfc\x6d\x36\xe9\x04\x8c\x67\x74\xb0\x63\x8c\xcb\xfc\x33\xa3\x00\xf9\xb8\x1e\x93": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xc0\x6d\x16\x1e\x41\x78\xd1\x3f\x67\x4e\xf8\x12\x44\x7b\x28\x60\xbc\x76\x7e\x3a\x7c\x4d\x3f\x2a\x7b\x19\x57\xbd\x25\x6c\x67\xb6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xc0\x77\xa4\xb1\x57\xc4\x80\xb3\xc0\x96\x02\xa7\xdc\x24\x2a\xef\x70\x05\x88\x15\x53\xef\x50\xd9\x53\x1a\x38\x00\xa9\x84\x4f\x69": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/IPoolManager.sol",
            name="IPoolManager",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xc0\x78\x13\x40\xd2\x23\xf1\x28\x48\x1f\xdd\x05\x91\xe1\xf4\x8e\xa6\xdd\x71\x1b\x96\xff\x91\x83\xe0\x17\xf2\x2c\x5f\xae\xfd\x63": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="__unstable__ERC20OwnedUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xc0\x7f\x9f\x1a\x77\xa5\x3a\x55\x09\x87\xaa\x4c\x98\x2b\x2e\x13\x4c\x05\x34\x54\xef\x68\x0b\xe9\x12\x46\xa7\x64\x92\x69\x00\x8b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xc0\x97\x2e\x81\x90\xf8\x19\x85\x65\xe4\x6f\xc6\xff\x67\x8b\x53\x75\x47\x99\x42\x70\x2a\x1e\x9f\x99\x3f\x89\xc4\xa7\x6b\xca\xde": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xc0\xb7\xdb\xd4\xd7\xfb\x72\xf3\x10\x26\xb7\x0f\x8b\x4f\xb1\x15\x36\x9c\x0d\xc9\x18\x41\xf4\x48\x30\x77\xf3\x1c\x6b\x64\xda\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.239, v0.0.240
    b"\xc0\xd3\x1d\x4f\xde\xd3\xb2\xde\x96\x22\xcb\xfc\xce\xbf\x47\xd2\x7b\xaa\x73\x68\x9e\xb9\x3f\x01\x38\x00\x38\x09\xc0\xcd\x98\xf4": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.240",
        ),
    ],
    # Hash appears in versions: v6.8.0
    b"\xc1\x01\x0b\xc2\x22\x26\xda\xb8\xd5\xa9\x19\x6f\x3a\x93\x2f\x4a\x7a\x9d\x5d\x1f\x44\xb9\x03\x34\x0e\x97\x35\x14\xda\x09\x4c\xda": [
        KnownContract(
            project="solmate",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.118
    b"\xc1\x0a\x29\x6a\x04\x3c\x18\xf7\x87\xb3\x58\x34\xd3\xa4\xa2\xd9\x80\x4b\x67\xc0\xea\x25\x13\x2a\xd1\xac\xba\x41\xf5\x8b\x68\x74": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.118",
        ),
    ],
    # Hash appears in versions: v0.0.68
    b"\xc1\x0d\x29\xe4\x60\xb4\x48\xe4\x2a\xbf\x60\xe3\x60\x7b\xc0\xe1\x03\xf7\x33\xc2\x29\x9f\x9c\xa1\x58\x3a\x9a\xe1\x14\x08\x54\xb6": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xc1\x0e\x08\x63\x71\xeb\x6a\x02\xc7\x18\x4c\x8d\x91\x87\x35\xdd\xfe\x56\xa3\x1e\x4f\x69\xb4\xce\xa8\x62\xab\x42\x8f\x39\x1c\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc1\x15\x0d\x9b\x1f\x30\x46\x83\x03\xd6\x16\xb9\xa2\xf0\xa3\x9c\x26\xad\xa3\xbc\xa7\x48\xc8\x03\xee\x15\x5d\xec\x8c\xd3\xa5\x5b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721RoyaltyUpgradeable.sol",
            name="ERC721RoyaltyUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc1\x47\xd2\x2c\x06\x83\x1e\xe4\xa0\xfd\x0f\xeb\x08\x4b\x6b\xe5\xb7\x11\x3c\x99\xf3\xcc\xf7\xbf\x82\xb4\x45\x33\x7d\x7b\xf4\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xc1\x53\x31\x93\xba\x15\x35\x27\xc2\x3b\xac\x56\x33\xaa\xf9\xf3\xb7\x91\x1d\x2a\x04\x56\x93\xfb\xf0\x07\x33\xd4\x7f\x2a\x3f\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorPreventLateQuorum.sol",
            name="GovernorPreventLateQuorum",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7
    b"\xc1\x5a\x85\xd0\x83\xa2\xe0\xe0\xfe\xcb\x7d\x1f\x2d\x1a\x34\x0a\x6e\x59\xd7\xa5\xf6\xa5\x38\xa0\x90\xa4\xce\xfc\x82\xe7\xf7\xa1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xc1\x65\x2a\x26\x07\xca\xd6\x4d\xdf\x0c\xcf\xe8\x4f\x49\x63\x49\x20\xcb\xb7\x72\xac\xdf\x4e\x8f\xfd\xe2\xe6\x87\xc5\x35\x1d\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc1\x77\xdb\xba\x82\x50\x99\x6a\xec\xa9\xe6\x98\x8b\x2c\xbe\xe1\x6d\xdb\x00\xf3\xb7\x3c\xe3\x6c\x9a\x29\xd9\x5c\x4c\x6b\xeb\xf0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/types/Time.sol",
            name="Time",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.12
    b"\xc1\x8e\x55\xd1\x0d\x65\x6c\xa9\x89\x02\x04\x86\x4f\xbd\x1f\x66\x33\x9d\x72\x70\xdb\x39\x33\x1e\xb4\xf3\x6e\x57\x02\xc8\x0d\xf1": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.12",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xc1\x93\x53\xe1\xdc\xfd\xfe\x5a\xb5\x94\xe8\xb0\x65\x8c\x08\x3e\x02\x22\x61\x2b\xf8\xde\xe8\xd5\xe1\xf7\x18\x43\x64\x3d\xba\xc3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155Upgradeable.sol",
            name="IERC1155Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xc1\xa9\xf0\xd3\xb3\xce\xa0\xfa\xa7\xf0\x5e\xba\x41\x16\x5f\x03\x0a\xdd\x09\x35\x22\xb7\xcc\xe5\x32\x02\xab\x06\x72\x71\x55\x54": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xc1\xd7\x01\x0f\x40\xce\xbe\xc4\xa3\xda\x99\x4b\xd1\xad\x2d\x5a\x9f\x9d\xd1\x1c\x4f\x7f\x01\x23\x30\x9a\xca\x94\x3f\xdc\x0d\xbf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.217
    b"\xc1\xe3\x48\x19\xf7\x25\x34\xbe\x09\xb5\x23\x3a\xc1\xdf\xc0\x27\xdb\x32\x0e\xff\x63\xcf\x88\xe2\x20\x1b\xc5\xdb\x6d\x2e\x1b\xba": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.217",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc1\xf3\xb0\x5f\x32\x9c\x0d\x84\xe9\x0c\x23\x67\x4f\x05\x9c\xbc\x49\x51\xe9\x55\xb7\x96\x12\x9e\x23\x77\x8c\x19\xc0\x0b\x9b\xd6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xc2\x14\x12\x3e\x0c\xc0\x3b\x92\x89\x86\xa7\x51\x9d\x55\x39\xc2\xa7\xac\xd9\xc0\x83\xdd\x6b\x74\xc5\xdb\x7e\xc3\xc8\x10\x53\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xc2\x22\x09\xb7\x3c\x37\x13\x21\xc8\x51\x65\x77\x5d\x35\xa7\xb2\xe3\xc7\x5e\x1f\x2d\x96\x95\xed\x72\x2f\x7e\x45\x55\x4c\x4b\x7c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\xc2\x3a\x29\xb2\x77\xaf\xa4\x7c\x08\x3c\x70\x5a\x52\xae\x68\x4b\x54\x1c\xe3\xab\x52\x19\xf9\xca\x60\xdf\x4f\xac\xd9\xc8\xd7\x24": [
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xc2\x44\x33\x68\xde\x52\xf3\x56\x8a\xc9\x4a\x0a\xb7\x50\xff\x25\x96\xc1\xec\x20\x3c\x8f\xec\xcb\xf3\x85\x02\x4e\xbf\x18\xa4\x01": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.5
    b"\xc2\x59\xfe\xf8\xee\x4e\xf9\xdb\x27\xb3\x55\x3a\xe9\x6f\xfb\x65\xe3\x0e\x20\x81\xb8\xbd\x2d\x41\xf0\x6d\x3e\x88\x25\xf9\x03\xd9": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.5",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xc2\x7c\xb1\xca\x89\x3a\x57\x1b\xe3\x35\x88\x36\x6a\x4a\x2c\x1f\x83\xa2\x5f\x6b\xfc\x41\x91\xe9\x2d\xd4\x27\xc1\xb5\x2b\xd9\x51": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc2\x7d\xc9\x96\x94\xab\xfe\xa1\xf8\x28\xde\x96\xb7\x26\x67\x8f\x2b\x94\x0f\x17\xeb\xad\x4f\x54\xbd\x1e\x05\x3e\xb8\x20\xae\x62": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xc2\x7e\xb5\x7a\xc2\x16\xae\xc9\xba\xd8\xb6\x76\xa3\xca\xf8\xc7\xf4\xad\x3b\xe8\x9c\x55\x42\x34\x17\x37\x57\x71\x75\x27\xd9\x3a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xc2\x83\x41\xdc\x9c\x93\xb4\x85\xb2\x50\xf5\x36\x21\x2c\x97\x46\x7f\xf1\x41\xbf\x2c\x3a\xde\x80\xd7\xee\x7e\x09\xb3\x43\xf3\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/IERC1155ReceiverUpgradeable.sol",
            name="IERC1155ReceiverUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xc2\x97\x99\x77\xfa\xb6\xf2\x36\xec\xbf\x04\xee\xb5\x79\x05\xcd\xd2\xab\x0c\x96\x54\x83\x97\x3f\x9f\xb1\x32\xce\x20\xe4\x97\xd3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.163
    b"\xc2\xa9\xcf\xad\xa0\xff\xee\xd5\x37\x3e\x6e\xd1\xa1\xf8\xbc\x4a\x41\xc8\x04\x32\x7f\xc0\x3f\xc1\xa2\x5f\x83\x7a\x65\xd5\xb8\xaa": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.163",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xc2\xb2\x0a\xed\x66\xa2\x9b\xad\xcd\x2d\x2d\x58\x8b\x00\xcb\xed\xe1\x81\x75\xe6\x6f\xf5\x74\xd5\x43\xe6\x24\xb8\xba\x89\x51\xa4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909TokenSupply.sol",
            name="ERC6909TokenSupply",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909TokenSupply.sol",
            name="ERC6909TokenSupply",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc3\x36\xb8\xbd\xcb\x72\xff\x7e\x80\x77\xf8\xc8\x59\x80\xc4\xb9\xa9\x7d\x51\xa0\x66\xba\x5b\x12\xb0\x8e\xec\x78\xac\x56\x9f\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xc3\x3c\x8a\x4b\x58\x28\xc1\x1e\x2d\xbb\x38\x74\x5e\x74\xf0\x69\x77\x27\xa8\x58\xac\x14\xd1\x55\xd2\x24\x7a\x0e\xab\x08\xb2\x83": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Detailed.sol",
            name="ERC20Detailed",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Detailed.sol",
            name="ERC20Detailed",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Detailed.sol",
            name="ERC20Detailed",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Detailed.sol",
            name="ERC20Detailed",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xc3\x49\x0e\x7a\x50\x6c\x0b\x8a\xa9\x63\x04\xc3\x36\x02\x99\x79\x2c\x04\x21\x05\x6e\x37\x70\x06\xfe\xd8\x9c\xbc\x1e\x86\x2d\x17": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xc3\x58\xc2\x39\xff\x33\xca\xe6\xc3\xfa\x13\xb0\x8b\x65\x0d\xea\xd5\xce\x9a\xa0\xbf\x4e\x44\xa4\x53\xbe\xfb\x51\x85\x41\x0b\x49": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/Context.sol",
            name="Context",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xc3\x69\x01\x89\x1c\x88\xe5\xbe\xdd\x76\xda\x08\x01\xf7\x30\x52\x7c\x05\x69\x46\xe4\xd5\xba\xf0\x35\xf3\x43\x91\x4d\x61\x5d\x92": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xc3\x6f\x9d\xcd\x8a\x71\x05\xb8\x02\x3b\xc1\x94\xed\x1a\x08\x20\x06\x3d\x23\x3c\x8c\x20\x4c\x95\x8e\xcf\xf6\xe6\x31\x6c\xc7\xef": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xc3\xa8\xd0\xed\x7d\x85\x9d\x2a\xdf\xbe\x6c\x23\xbc\x23\xf3\x7b\x11\xf8\xe4\x1b\xef\xaa\x09\xa4\xc6\x25\x72\xf5\x42\x2a\xb4\x4e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721UpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xc3\xaf\x70\x49\xad\x9c\x97\xc8\x61\x92\xc2\xba\x0d\xad\x6b\x8c\x06\xe1\x9b\x50\x8b\x09\x9c\xc7\xfd\x5e\x5b\x40\xf1\xa6\x24\xd0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc3\xc2\x43\x51\xa4\xe2\xc7\xc5\x18\x51\x95\x41\x99\xf0\x3f\xbe\x3b\x20\xd2\x07\x1b\x0e\x8e\x9b\x82\x16\xd1\x2c\x9b\x85\x8d\xe0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xc3\xe3\x15\x29\x81\xa8\x33\x38\xe6\x20\xf4\x6a\x63\x36\x93\x00\x1b\xfb\xb1\xfd\xdd\xe1\x92\xd2\xfb\x82\x4e\x9f\xbb\x41\xd0\x07": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IMulticall.sol",
            name="IMulticall",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IMulticall.sol",
            name="IMulticall",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IMulticall.sol",
            name="IMulticall",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IMulticall.sol",
            name="IMulticall",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IMulticall.sol",
            name="IMulticall",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xc3\xe8\x70\xae\x4f\x6f\x94\xe2\x62\x37\x80\xdd\xd2\x31\xf2\x93\x48\x62\xd7\xfa\x66\xad\x2c\x9b\x4d\x97\xbb\x19\x5e\xd9\xc3\x3e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc3\xf5\xcf\x37\x5e\x71\xfd\x0a\x01\x5e\xcb\x58\x6a\x7d\x33\x0c\x06\xb0\xbd\x47\x74\xd4\x5e\x2e\xbe\xcf\x3e\x28\x64\x26\xe8\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.9.6
    b"\xc4\x0c\x0d\x08\x06\xef\x87\x23\xf0\x18\x0a\x59\xc6\xa4\xe3\x31\x55\x13\x72\x8f\x76\xfd\x11\x2b\x8d\x1a\x94\xc1\x42\xc0\x1b\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Base64Upgradeable.sol",
            name="Base64Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xc4\x0e\xfc\x07\x1e\x44\x4a\x7f\xac\x83\x0e\xca\x52\x28\x65\x92\x76\x36\x2f\x18\xdf\x19\x10\xf8\x82\x98\x4b\xa7\xd0\xa5\x74\x56": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/interfaces/IExtsload.sol",
            name="IExtsload",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc4\x11\x4e\xe9\xeb\xf5\x14\x87\x56\x90\x3d\x1b\x90\xdd\xa7\x54\xc7\xf0\x69\xf1\x26\x70\x07\xda\xe4\x14\x84\xf6\x23\x5c\xd0\xd1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xc4\x1c\xd8\x8d\xa7\x22\x71\x80\xfb\xfd\xa9\x81\x26\xec\xcd\xd9\x30\x49\x64\x46\x19\xf2\xd5\xd4\xc4\x6d\x04\x86\x8d\x6e\xb6\xbb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xc4\x26\x2c\x32\xe8\x56\x2c\x90\x40\x11\x58\xe3\x58\x22\xd7\x2f\x8b\x35\x41\x7c\xd6\xde\x1c\x7b\x50\x96\xa3\xa9\xcc\x37\xce\xbb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc4\x2c\x68\x6e\xdb\xd3\xeb\xac\x97\xb1\xd8\x11\xb7\xdf\xaa\xad\xfa\xe9\xce\x98\x33\x7c\x4f\x06\x1a\x6f\x31\x5b\xe4\x0a\x19\x8a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201
    b"\xc4\x72\xdc\xd6\xe3\x71\x8b\x78\xab\x48\x67\x21\x9a\xc3\xdb\x56\x79\xb5\x3b\xa6\x38\x8c\x58\x0a\x20\x5f\x04\xf0\x4e\xaa\xe6\x52": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.201",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281
    b"\xc4\x79\x3a\x0c\x6f\xaa\x2e\x7b\x8c\x0b\x52\x11\x2c\x3a\x4c\xd8\x19\x65\xd9\x58\x2f\xcb\xa8\xf7\xff\x19\x7f\xa7\xa7\x5f\x9d\x34": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xc4\x79\x5c\x7d\xeb\xec\x29\xd5\x89\x77\x3a\xa2\x17\x58\x6d\x28\xf1\xd0\x79\x59\x4b\x56\xcf\x5e\x79\xdd\x23\xd6\x81\x8c\x6c\x35": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xc4\x80\xa7\x0c\xdc\x19\xcc\xe3\x44\xf9\x8f\x5a\xd2\xab\x06\xf8\xcb\x6b\x8b\x3d\xed\x35\xaf\x7c\x87\x61\x53\xd5\x2f\xa2\xfd\x8f": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xc4\x81\x44\x89\xdc\xcf\xb6\x8c\xe8\x82\xb0\x2f\x59\x4e\x2d\xec\x30\x60\xbf\x1b\x2b\xbe\xf1\x60\xb1\xd2\x00\x3e\xd1\x36\x14\xfd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xc4\xc3\x2c\x50\x1d\xf8\x74\x72\x7a\xc9\x51\x44\x18\x3a\x65\x37\xee\xad\xf3\x79\x5b\x9f\x68\xe4\x7f\xdc\xe5\xee\xe3\x86\x4f\x95": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xc4\xd0\x7d\x8d\x20\x52\x77\x21\x8c\x24\x4b\x7c\x28\x67\x69\xe7\x11\x5c\x9f\xa3\x01\xbd\xf5\x8b\x47\x8b\x13\xaa\xab\x90\xcb\x76": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/BipsLibrary.sol",
            name="BipsLibrary",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xc4\xd6\x59\xc7\xaa\x50\xb9\xc1\xb6\x25\x71\xf5\xd7\x96\x8e\xc7\x3a\x49\x40\xc8\x0d\x97\x9a\x14\xa1\x0a\x8e\xd1\x89\xd0\x2a\x37": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xc4\xfb\xe6\xea\x97\x83\x28\x63\x83\xf6\x02\x79\xcd\xc9\xc5\x0c\xde\x51\x4c\xc9\x63\x65\xf2\xb5\x1f\xc6\x25\x34\x4f\xb9\xfd\x1c": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/SlippageCheck.sol",
            name="SlippageCheck",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232
    b"\xc5\x04\xd8\xa0\xfa\xe4\x5c\x3d\x74\x03\x52\x7f\x0d\x65\x9c\x05\xc0\x5e\x6e\x01\xe6\x5f\x4e\x42\x0c\x8b\x6d\x1f\x8c\xde\x2a\xec": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.232",
        ),
    ],
    # Hash appears in versions: v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273
    b"\xc5\x17\x37\xb3\x8e\x3a\x97\x0b\xfe\x67\xf8\x8a\x04\x9a\xba\xb1\xc7\x34\x28\x26\xf1\x9c\x4c\x1f\xbe\xeb\x39\x94\xf6\x8d\x86\x17": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xc5\x1d\x20\xc1\x6c\xea\x86\xad\x25\x07\x3f\x82\x5c\x14\x04\x66\xcd\xb7\x3c\x47\x1c\x95\xed\x06\xaa\x9b\xea\xad\xac\x68\xb6\x10": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.216
    b"\xc5\x20\x8e\x33\x55\x0b\x73\x20\xeb\x75\xcc\x10\x56\xd3\xe6\xd2\x8f\x32\x32\x75\x5a\xfb\x5f\x1d\x88\xd8\x7e\xc8\x67\x5e\xac\xe4": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.216",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7
    b"\xc5\x30\x76\x8c\xb7\x55\x52\xa4\xfa\xc5\xfe\x70\x7c\xa7\x72\x1c\xfa\x08\x09\x21\x47\xde\x9c\x2d\x67\xed\xf6\xe5\x35\xec\x05\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Initializable.sol",
            name="Initializable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xc5\x47\x61\x4a\x14\x61\xff\x2f\xf5\xb4\x67\xe2\xa9\x40\x26\x2b\xfe\xd4\xb4\x86\x48\x10\xe9\xaa\xf6\x02\x69\xc0\xc1\x6f\x47\x7b": [
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibCall.sol",
            name="LibCall",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xc5\x72\x85\x8a\xa8\xa1\xf6\xfe\x8b\xc4\x97\x0b\x71\xf6\x11\x97\x42\xa1\x60\x39\x35\x26\x7d\x1f\x1a\xdc\xb3\x64\xf0\xf5\x93\xae": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xc5\x7e\xed\xda\x60\x54\xaa\xc2\x14\x52\x42\x0b\x79\xac\x6d\xbf\x24\xb9\x76\x25\x23\x44\x8f\x90\x56\x2e\x2e\xd3\xcc\xf3\x67\x26": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetFixedSupplyUpgradeable.sol",
            name="ERC20PresetFixedSupplyUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xc5\x96\x4a\x35\x14\x2d\xb8\x12\xa2\x4c\xd7\x0f\x26\x44\x69\xe0\xb8\x5e\xb7\x55\x7b\x95\x3b\x8d\x17\x1c\x12\x0a\x20\xd2\x7f\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909ContentURI.sol",
            name="ERC6909ContentURI",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/extensions/draft-ERC6909ContentURI.sol",
            name="ERC6909ContentURI",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc5\x9e\x1e\x38\x41\x5d\xdd\xc9\xd7\x8d\x5f\xdd\x63\x78\xfb\x11\xe4\xb1\x22\x49\x8d\xd5\xe8\x81\x26\x6a\xbf\xa5\xca\x80\x75\x4f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Consecutive.sol",
            name="ERC721Consecutive",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xc5\x9e\x78\x2d\x14\x13\x65\x95\x07\xf5\xb7\xe1\x79\xf8\xa4\x98\x10\x73\x28\x06\x4a\x81\x5a\x2f\x65\xb1\x6f\x43\xbc\x5c\xb5\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/TimelockController.sol",
            name="TimelockController",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xc5\xd0\x17\x67\xb0\xb6\xd7\xb8\x68\xaf\x59\xa1\x7f\x8b\x15\x42\x4e\x0a\xef\xe5\xfa\x57\x70\xab\xd0\xcd\x59\x31\x9b\x89\x09\x76": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/CrossChainEnabledUpgradeable.sol",
            name="CrossChainEnabledUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0
    b"\xc5\xeb\x66\x2b\x6b\x0d\x11\xab\x73\x78\x9f\x61\xe1\x0a\x23\xbc\xd5\x70\x6c\x3e\x41\x04\x6c\x3e\xe5\x0a\x33\xf6\x6d\xaa\x4e\x0e": [
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.2.0",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xc6\x1f\x7e\xdb\xe3\x2d\x48\x75\xfc\x59\xb2\x26\x1e\xd3\x9a\x6b\xc0\xac\xaf\x8a\x96\xbd\x12\x89\x4a\x38\x1a\x00\x6e\x6a\xcc\xf5": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721TokenReceiver",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xc6\x26\x11\x9d\x58\x7f\xec\x8b\xe6\xc9\xd5\xcc\x97\x4a\xa4\xf2\xb1\x38\xb3\x61\xfb\xc2\x59\x3a\x94\x24\x90\x72\x16\xdd\x0f\x17": [
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xc6\x32\x8a\x58\xa5\x9c\x7c\xb6\x6a\x82\x25\x68\xbd\x68\xd4\x13\x42\x4c\x80\xce\xd7\xf2\xda\x30\x89\x18\x08\xa4\x3d\x86\x83\x61": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/IAccessManagedUpgradeable.sol",
            name="IAccessManagedUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\xc6\x44\xe8\x41\xcd\x36\x39\xe7\x05\x4b\xe1\x16\xcf\xff\x5e\x5f\x33\x72\x22\xa8\x2c\x1c\x26\x43\x5d\x49\x75\x9d\x66\x4e\x39\xd1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xc6\x49\x26\x97\x62\xbb\x69\xbe\xf1\xa1\x30\x54\x23\xb6\xa8\xfb\xad\x56\x38\xc0\x02\xf6\x47\x2c\xc2\x92\x6a\xa1\x3e\xf6\x07\x12": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xc6\x82\x3a\xe7\x0b\x68\xa2\xf4\x6a\x0a\x1d\x50\x08\xd7\x9e\xf0\x2e\x36\x59\x06\x1f\x90\x0a\xbc\x61\x3a\x20\x43\x02\x55\xa9\xf3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingOverridableUpgradeable.sol",
            name="GovernorCountingOverridableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingOverridableUpgradeable.sol",
            name="GovernorCountingOverridableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingOverridableUpgradeable.sol",
            name="GovernorCountingOverridableUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xc6\xc7\x32\x20\x26\x90\xaf\xb9\x9c\x4d\x40\xe3\x9e\x35\xda\x58\x0d\x2c\xbe\xc8\x38\x7a\x09\x45\x2e\x5f\x2d\x1d\x82\x41\x4d\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xc6\xce\x8f\x70\xdc\x5e\x1d\x1f\xd0\x0f\xf4\x39\x3a\xd3\xe8\x03\xc9\xf8\x14\x57\xed\xc6\xd2\x90\xad\xdd\x30\x23\x90\x8a\xd4\xb4": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryImmutableState.sol",
            name="PeripheryImmutableState",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryImmutableState.sol",
            name="PeripheryImmutableState",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryImmutableState.sol",
            name="PeripheryImmutableState",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryImmutableState.sol",
            name="PeripheryImmutableState",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryImmutableState.sol",
            name="PeripheryImmutableState",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xc7\x00\x44\x4c\x3b\xc2\xe2\x24\x8c\xd1\xeb\x1d\xf9\xaa\x9c\x25\x5a\x6a\x51\x0d\x72\xf6\xb8\xfd\x78\xe1\x38\x9a\x83\x77\xf0\x22": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20PrimaryAdmin",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20PrimaryAdmin",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181
    b"\xc7\x06\x6f\x15\xa9\x8b\x79\xae\x44\xc0\xad\xad\x41\x75\x77\x26\x41\x0c\x62\x76\x15\xfa\x36\xcc\xd5\x2f\x6d\xf5\x86\xcc\x57\x85": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.181",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc7\x45\x06\x3e\x73\xb5\x24\xd3\xa5\x80\xc7\x47\x84\x73\xc9\xf8\x00\x58\x14\x0b\xe1\x52\x44\xc8\x4e\xaf\x5f\xa9\x2d\xc6\x62\x59": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Utils.sol",
            name="ERC721Utils",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xc7\x4a\x45\x75\xf6\xbe\x13\x00\xca\xf1\xf8\xa3\x5b\x57\x00\x53\x9f\x5c\x16\xbe\x87\xa7\x76\xaf\xf2\xb1\x9c\xa0\x51\x8a\x18\x7e": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xc7\x6a\xbb\x5c\xd8\x1b\x41\xed\x3c\x4d\x5e\xeb\xf1\xc6\xbd\xa6\x15\xce\x6e\x91\xb5\x69\xa8\x5e\xfa\x7f\x03\xd0\xb8\x5c\xf6\x0b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xc7\x71\xdf\x7d\x37\x7c\xc0\x9f\xda\xd1\x5a\x11\xec\xc9\xd2\x04\x7f\x0c\x25\xf1\x0e\x0a\xfd\x5f\xdf\x48\xec\x4a\x75\x17\xb6\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xc7\x85\x19\x25\xb3\x97\x1f\x2e\xb8\xe7\x77\xc5\x2d\xe1\x3b\x9e\xc7\x18\xa7\x97\x81\x19\xa4\x1a\x2b\x68\x4d\xd6\x49\x4a\x7c\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155Holder.sol",
            name="ERC1155Holder",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xc7\x89\x8f\x3b\x50\x65\xce\xea\xc5\x75\xda\xc3\x88\xd7\xbf\xd3\x7f\x4b\xc6\x9b\x9c\x2c\x84\x65\x13\x97\x9d\xc7\x73\xce\x4e\x7f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardUpgradeable.sol",
            name="ReentrancyGuardUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.159
    b"\xc7\x8c\xf3\x5d\xc5\x86\xd5\xbe\xba\x2d\x1b\x78\xb2\xea\x3f\xcc\x5a\xe1\xe7\x3b\xb5\x47\x19\x26\x70\xc4\x6a\xf5\xae\x39\x0a\x21": [
        KnownContract(
            project="solady",
            path="src/utils/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.159",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xc7\x8d\x79\xb7\x24\xe0\x16\x90\x3a\xbf\xeb\xd7\x42\xcb\x35\x2c\x40\x5c\xec\xc3\x0a\x6c\x2c\xd3\xf9\x5d\x33\x89\x68\xa2\x12\xfb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.276, v0.0.277, v0.0.278, v0.0.279
    b"\xc7\x9f\xe9\xe9\xbb\x62\x8d\x3a\x4b\xcb\xbd\xbe\xbe\x66\x3c\x3f\xe4\x8b\x15\xe8\x75\x7d\xaf\x51\x8d\xb1\x2e\xce\xfc\x41\x52\xb0": [
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/auth/EnumerableRoles.sol",
            name="EnumerableRoles",
            version="v0.0.279",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xc7\xbe\xca\xfa\x15\x80\xee\xca\xc7\x60\x60\xb2\xcf\x9e\xda\xb2\xa6\x10\xfc\xdb\x55\xc8\xae\x87\x26\x5b\xce\x23\x7a\xb2\x80\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingOverridable.sol",
            name="GovernorCountingOverridable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingOverridable.sol",
            name="GovernorCountingOverridable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc7\xc4\x88\x94\xa8\xc3\x3d\x4d\x1a\x9c\x1d\xf6\x09\x27\x92\xbb\xe5\xb1\xf8\xfc\xa5\xb8\x0c\x18\x27\xcf\x5d\x3f\x2e\x0a\xb3\x34": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xc7\xc9\xb2\x82\xfc\xc9\x1e\x19\xa0\x2c\xf9\x62\xc4\xc6\xec\xe6\x2e\xf1\x6e\x73\xc0\xe0\xb3\x60\x7f\xe3\xec\xd2\xdd\x3e\xd3\xbd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xc7\xf6\xdf\x72\x46\xc8\xcc\x28\x8d\x28\x98\x34\x8a\xd9\x5d\x0c\x73\xb9\x53\x58\x6f\x81\xc9\x7a\xf7\x47\xc5\xa3\x9d\x2c\x7b\xdb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CountersUpgradeable.sol",
            name="CountersUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xc8\x08\x65\x3a\x9d\x14\x03\x35\x50\x54\xb6\x63\x3c\x92\x9e\x16\x2e\xa5\x61\x02\x7e\x18\x0c\x36\xc7\xb3\xd2\xec\xf0\x8c\xaa\x27": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc8\x33\x3b\x46\x09\x00\xa4\x7e\x4a\x4e\xe8\x6c\x79\xcd\xc0\x7d\xcb\x1c\x76\x10\x58\xca\x80\x28\xdd\x3c\x8f\xb9\x2b\x50\xad\xb1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IBridge.sol",
            name="IBridge",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xc8\x37\x06\x8c\xf0\x2b\x49\xdf\x62\xcc\xd0\x55\xea\x5b\x81\x08\x15\x66\x29\x1e\xee\x31\x1c\x3d\x49\x78\x2a\xf1\x57\x82\x3b\x5d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xc8\x3d\x98\xc9\x7b\x2c\xf5\x4d\x7c\x7d\xfd\x1d\x68\xa0\xee\x2b\x7f\x81\x5f\x29\xec\xaf\x61\x84\xef\x7f\x10\x58\xab\x2e\x93\x88": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-beta.1
    b"\xc8\x45\x8d\x4f\x40\x22\x18\x87\x6c\x58\xbc\xe1\xee\x7c\xb0\xf3\x93\x50\x66\x66\x27\xc6\x4f\xec\x50\x1a\x10\xda\x17\xd1\x5e\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xc8\x92\x4b\x00\x8b\x5f\xc5\x39\x1d\x1f\xce\x75\xf6\x73\x15\x83\x58\x70\xea\xa1\xef\xca\xdd\xbc\xad\x4c\xac\x82\x0f\x50\x78\xf5": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xc8\xc0\x16\xf8\x46\x76\x9d\xac\x52\x4e\xf9\x78\x7e\x4f\xb8\x16\xa1\xac\xc0\x31\xec\xa9\x62\x32\xec\x48\xed\x70\xd4\x95\x5e\xcc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xc8\xc4\x73\xba\x8a\x15\xca\xd0\x4e\x98\xfd\xf0\xe1\x8d\x97\x15\x18\xb7\x9c\x32\xf1\x0f\x60\xca\x85\xe1\x89\x65\x11\x6d\xdb\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xc8\xee\xc8\x11\x83\x26\x51\x4e\x51\x20\xc3\x34\x64\x6c\x9a\x4b\xfc\x7a\x84\x36\x34\x32\xf2\x7f\xad\x3c\x22\xcd\xf5\xce\x64\x07": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20CappedUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xc8\xf3\x10\x95\xca\x40\x21\x21\xc2\x7d\x74\xfd\xd2\xb3\xbf\x84\x5d\x4e\x6e\xab\x80\x3d\x2f\x03\xc6\x94\x9c\xd6\x29\xd9\x73\xff": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/MerkleTree.sol",
            name="MerkleTree",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xc8\xfd\x88\x89\x72\xd9\x93\x56\xa3\x69\xf8\xe7\x32\x09\x0e\x3d\x8a\x38\x18\xa4\x8c\x28\xf3\x86\x0e\xec\x5d\xfa\xef\xfe\x3f\x0d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.2.1-solc-0.7, v3.2.1-solc-0.7
    b"\xc9\x16\xb1\xae\xad\x41\xb9\xb7\xb8\xc8\x55\x28\x0b\x0c\x33\x3d\x29\xa1\xad\x07\xe2\xaa\x19\x43\x18\x2f\x46\x72\x95\xbd\x9a\xfa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.1.0-rc.0
    b"\xc9\x43\x72\x60\x06\x8d\xb7\x62\x65\x9e\x89\x96\x5b\xc1\x2b\x94\x21\xd7\xd7\x65\x04\x5f\x06\x2e\x11\x4b\x9b\x77\xa3\x99\xa8\x5f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xc9\x55\x92\x6d\xdd\xbb\x0b\xaa\x05\x73\x12\x59\x66\x1d\xf9\xe3\x76\x3b\x89\x8d\xfc\x4b\xff\xca\x66\xc7\xe6\x6b\x6c\xaa\x11\xdd": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingOverridable.sol",
            name="GovernorCountingOverridable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingOverridable.sol",
            name="GovernorCountingOverridable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingOverridable.sol",
            name="GovernorCountingOverridable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xc9\x5a\x00\x69\xba\x0c\x55\xcf\x91\xf8\xf0\x29\xaf\x5d\x5f\x85\xbb\x04\x1e\x82\xd0\xbc\x36\x96\x69\x8a\x87\xdc\x69\xb0\xff\xed": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Burnable.sol",
            name="ERC721Burnable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xc9\x5a\x66\x50\xd5\x72\xb6\xe1\xdf\x1c\xa2\xd5\x5c\x34\xa9\xa7\x8c\xda\xa8\x51\x38\x61\x97\x7c\x79\x8a\x2d\xbe\x81\x96\x0c\x5a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xc9\xaa\x07\x02\x55\x74\x0b\xba\x4d\x67\x39\xda\x52\xb0\xd9\x2d\x49\x1c\x37\x85\x0e\x49\x82\x3f\x0a\x43\x3d\x10\x8c\xde\xe1\x95": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorage.sol",
            name="ERC1155URIStorage",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xc9\xe1\x23\x85\xb3\x52\x18\xef\x20\xa2\x67\xb0\x10\x5c\x65\x5c\x96\x73\x1c\x6d\xcb\x75\xdd\x68\xe4\x49\xed\x6c\xf6\x69\xc4\xad": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/types/BalanceDelta.sol",
            name="BalanceDeltaLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xc9\xf1\xb8\x99\xa4\x83\x63\x49\x3a\xed\x50\x36\x4a\x78\xc1\xa0\xae\xa7\x58\x54\xc3\x63\xae\x6f\x96\x74\xb5\x0f\xf2\x6f\x7d\xaa": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.2, v4.7.3
    b"\xc9\xfa\x75\xfd\xae\x7b\xbe\xaf\x8b\x02\xd1\x05\xc8\xeb\x2e\x82\xbe\x04\xd5\xe7\xe6\x5e\x1b\x90\xfe\x2b\xe3\x55\xc1\x92\xff\xc2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.179
    b"\xca\x0b\x02\x24\x5a\x9a\x7d\xe7\x25\x45\x16\x41\xc8\x52\x84\x19\xf1\xd4\x95\x18\x42\xff\x17\x88\x2c\x0b\x67\x11\x5b\xce\x59\x2c": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.179",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xca\x1b\x7b\x20\x73\x1f\xd6\x7e\x84\xad\xbe\xff\x43\x71\xba\xa8\x1a\x38\x37\x75\xa5\x2c\xfa\x4b\xc1\x81\x78\x18\xa8\xd0\x0f\x0e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xca\x1b\x8a\x60\xc4\x96\xe1\xdf\x02\x5a\xf9\xd0\x79\xd1\x55\xb2\xc2\x01\x7f\xd7\xe5\x9d\x12\x0c\xe6\xc7\x6b\xa8\xbb\x3d\x24\x63": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xca\x21\x43\x72\x89\x37\x73\xf9\xc7\xdc\xf9\x00\xa8\xa8\x9a\xf9\x5e\x95\x8f\x63\x4a\xc3\x59\x1b\x43\xd0\xa9\xef\xb1\xb8\xcc\xa9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC777RecipientUpgradeable.sol",
            name="IERC777RecipientUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xca\x2c\x1a\x4c\xfc\x82\x73\x36\xf4\x31\x9f\x48\x84\x70\x80\xc4\x41\x36\x0b\xd5\xb9\xa1\x79\xa7\x58\x55\xce\x24\x52\x3c\xc8\xf9": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v0.0.244
    b"\xca\x2c\xa9\x6e\xf5\x0b\xd7\x2f\x9f\x6d\xea\xb8\xce\xd8\x7d\x17\x7d\x2d\x6d\x30\xc9\x3f\xa8\x4d\xc4\x5d\x12\xf9\xf5\x52\x44\x17": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.244",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xca\x45\xb2\x35\x4f\x30\x09\x88\x55\x20\x9f\x41\x4e\x76\xf6\xb5\x70\xe2\x84\x25\x25\xd5\xc7\x77\x1e\xb6\x7b\xa2\xb6\x19\x89\xa8": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TokenRatioSortOrder.sol",
            name="TokenRatioSortOrder",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TokenRatioSortOrder.sol",
            name="TokenRatioSortOrder",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TokenRatioSortOrder.sol",
            name="TokenRatioSortOrder",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TokenRatioSortOrder.sol",
            name="TokenRatioSortOrder",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/TokenRatioSortOrder.sol",
            name="TokenRatioSortOrder",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xca\x6c\xc9\x18\xc7\xb7\xf9\xed\x52\x9d\xad\xa6\xa3\x59\xb4\x2f\xb2\x18\x53\x8d\xc4\xb9\xc0\xd7\xaf\xdf\x06\x12\x72\x15\xab\xb5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC5805.sol",
            name="IERC5805",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xca\x6d\xcc\xc5\x23\x08\x06\x50\x56\x62\xd7\x66\x63\x86\xbc\x82\xee\x65\xb7\x0c\xe8\x4c\x28\x7f\x5b\x40\x97\xaf\x6d\x9a\xc6\x66": [
        KnownContract(
            project="openzeppelin",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xca\x7f\x2b\x81\xfa\x7c\x04\x76\x45\xdf\xbb\xc8\x87\x8a\xa7\x4e\xc6\x26\xbd\x4b\x5d\x6c\xee\x93\xa2\x3b\xc6\x89\xe0\xa4\x6f\xeb": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IEIP712_v4.sol",
            name="IEIP712_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xca\xb8\x1d\x77\x59\x13\x5b\xf5\x11\xb0\x36\x45\x7e\x9f\xab\x80\xfc\x15\x6f\xe5\x56\x42\xaa\x74\x0d\x30\x4e\x4e\x6f\x2e\x49\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\xca\xc9\x01\x7f\x4c\x99\x0b\x3b\x0c\xc0\x4b\x12\x9b\x93\x6b\x04\x19\x83\xe6\x52\x64\x62\x13\x06\x12\x7c\x22\xc2\x9d\x37\x53\x35": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xca\xf0\x50\x43\xbf\x9a\x5a\xb9\x05\x06\x4a\x54\x71\xde\x5d\x70\x6b\xbb\xda\x86\x63\x63\x2e\x97\x69\x09\x85\x9d\x6b\x0a\xb1\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/draft-ERC6909.sol",
            name="ERC6909",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC6909/draft-ERC6909.sol",
            name="ERC6909",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.181, v0.0.182
    b"\xca\xf8\x34\x81\x26\xd9\x1c\x38\xfb\x90\xb4\xc0\xf1\x6f\x82\xf6\xfd\x34\xd3\x13\xde\x4c\xdc\x19\xf8\x59\x7b\xcf\xa9\x2a\x2a\xea": [
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.182",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.1, v4.4.2
    b"\xcb\x5e\xb6\xb9\xf9\x0c\x62\x81\xc0\xb2\xcb\x04\xd4\x93\xad\x51\x4c\x4c\x43\x0d\x41\xb2\x5f\x08\x20\x9f\x69\xba\x10\xb5\xf8\x30": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xcb\x7d\x3b\x03\x94\xa4\x7d\x0f\x28\xd4\x71\x59\xe2\xb9\x0e\x3d\x75\x1b\x87\xa1\x73\x30\xe7\x88\x20\x5b\xf4\x6b\x9e\x40\x37\x8a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignature.sol",
            name="GSNRecipientSignature",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xcb\x83\x83\x9a\x05\xbb\x9e\x61\xa1\xfa\x96\xee\x0a\x90\xa2\xf0\x03\x54\x1f\x66\xcf\x9d\x4b\xcb\x4e\x64\x29\xd3\x0e\x8c\xe6\x7d": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/IERC20Minimal.sol",
            name="IERC20Minimal",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xcb\x97\x58\x1b\xf3\x0c\xb0\xf8\x36\x48\x1e\xfb\x38\x71\x4f\xde\xcd\x26\xe7\xf2\x42\xc9\x14\xbd\x40\x68\x50\x21\x47\xde\xd7\x86": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xcb\x9a\x7f\x30\x55\xac\x51\xc6\x21\x3c\x02\xe0\xab\xb4\x8e\x47\xef\x51\x1f\x08\xb0\x06\xee\x98\x13\xc7\x55\xe2\x57\x4c\x65\xdf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xcb\xa1\xf8\x3b\x7f\xa0\xad\x8f\x74\x1c\x3f\x15\x93\x66\x84\xda\xf5\x45\x25\x46\xbc\x9d\xf8\xb0\x85\x7c\xe8\x63\x7a\x52\xcd\x08": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IMessageProvider.sol",
            name="IMessageProvider",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.3.0
    b"\xcb\xad\xa7\x4d\xcd\x0f\xff\xfe\x3a\x43\xde\x68\x21\xbd\x33\x07\xa0\xf7\xc6\x29\x03\x9b\x87\xa3\xc3\xa6\x3e\xae\xcc\x52\x18\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/account/utils/draft-ERC7579Utils.sol",
            name="ERC7579Utils",
            version="v5.3.0",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\xcb\xaf\x7a\xbe\x78\xb2\x9c\x25\x24\xd8\xbe\xf9\xb4\x8e\xd6\x4d\xb8\xf5\x9c\xb5\xeb\x1b\xc6\x36\xdf\xa4\xef\x8b\x65\xa8\x97\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xcb\xb0\x04\x54\x65\xfa\x6d\xd9\xdd\x2c\x99\xd7\x8a\xa3\x65\x86\x7b\xbd\x44\xa2\x27\xf8\x85\xe5\xdb\xe0\xf7\xcd\xf7\x8c\xd9\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Royalty.sol",
            name="ERC721Royalty",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277
    b"\xcb\xbb\xee\x94\xad\x9d\xb2\x82\x16\x03\x0a\x1b\x9b\xc4\xa8\xb9\x66\xeb\x34\x34\x20\xdd\x53\x29\xf4\x8a\x01\xe8\xc2\x8a\xe3\xd9": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.277",
        ),
    ],
    # Hash appears in versions: v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xcb\xc3\x90\xa8\x11\xd8\xcf\xff\x8c\x3a\x49\x0b\x5c\x8b\xdf\x50\xa7\x9a\xe7\x72\x7b\x5d\x4c\xed\x75\xc4\xf8\xb5\x22\xd9\x6b\x6f": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC2981.sol",
            name="ERC2981",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.2.0-rc.0, v5.2.0-rc.1
    b"\xcb\xc7\x33\xb9\x26\x4c\xf0\x2b\xef\xc1\x75\x28\xe8\x7f\x47\xfa\xd1\x6e\xbf\x0e\x8c\x35\xfa\xc6\xea\x66\x41\x53\x60\x16\x3b\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Execution",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Execution",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43
    b"\xcc\x27\x39\x66\x67\x6b\x46\xff\x49\x29\xb8\x15\x1a\xaa\x2b\x6d\xa0\xd5\x18\x1b\x93\x2b\x4c\xf3\xbf\xab\xe5\xd2\x6a\x67\x12\x66": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.43",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xcc\x38\x7b\xef\x8a\x77\x7c\xcc\x93\xa5\xd3\x83\x9c\x28\x28\x88\x00\xe0\x26\xb2\xe1\xf5\xed\xfc\xee\xb7\x05\x8c\x16\x66\xd0\xee": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC777SenderUpgradeable.sol",
            name="IERC777SenderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.0.0, v6.1.0, v6.2.0, v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xcc\x45\x4d\x2c\xc6\x45\x14\x09\xcf\x88\x5b\xdd\xca\x7a\x7a\xd2\x1a\x16\xb9\xf1\x5d\xd7\x2c\x81\x3b\x7e\xfc\xc3\x57\x4a\xbd\x69": [
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.0.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.1.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.2.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/auth/authorities/MultiRolesAuthority.sol",
            name="MultiRolesAuthority",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\xcc\x67\x0d\xa1\x87\x22\xf4\xc3\xde\xb4\x4f\xf7\xd8\x6a\xd2\xa2\x62\x40\x93\x3d\x98\x8b\xa3\xba\xe8\x9e\x0e\xe1\x6f\x40\x94\xfb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xcc\x6e\x40\x28\x47\x4f\x7d\xce\x63\x83\x89\x88\x85\x52\x2a\xd1\xab\xea\x1d\x67\x0e\x06\xa6\x9b\xb7\x52\xb2\x46\xba\xc9\xba\x91": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\xcc\x76\x35\x35\x85\xe7\x0a\xf0\xa5\xce\xcf\x3d\x85\xde\xd2\x05\xfb\x8b\x15\xe2\x0c\x15\x29\x4a\xbd\xe3\xa2\x7e\xcd\xe5\xf1\xf7": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xcc\x82\x52\xaa\xd0\xfd\xa3\x8b\x46\x72\x90\x0e\xd6\xff\xce\x48\xdd\x8b\x87\x38\x14\xa1\x73\xd5\xa4\x9c\x9d\x3c\xbb\x6e\x02\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC1820Implementer.sol",
            name="ERC1820Implementer",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xcc\xb9\x62\x0b\x1c\x4c\x4d\x83\x04\x3b\xfa\x74\xf3\x4c\x1b\x97\x2d\xc7\x5b\x1b\x12\x9f\x17\x02\x35\xe2\x88\x45\x11\x19\xcc\x18": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0
    b"\xcc\xe1\x40\x0e\x51\x3f\x59\xe6\x42\x77\xa6\xb3\xea\xd8\x26\x60\x1e\x9f\xa4\x27\x16\xb6\x0d\xac\x7b\xfa\x31\xa5\xe6\x56\x39\x18": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xcc\xe7\x4d\xad\xe9\x99\xcd\xa8\x00\xa5\x0c\xb6\xf4\x75\x80\xb1\x62\xa5\x5d\x8c\x79\x4f\x80\x6e\xe5\xc7\x86\x1b\x40\x38\xe5\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Counters.sol",
            name="Counters",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xcc\xfb\xda\xb3\x9c\xee\x0d\x99\x95\x11\x74\x47\xa3\x5d\xe7\x22\x04\x2d\xae\x74\x05\x4e\x8e\x9f\x40\xa1\x62\xd8\xfc\x28\xb1\xe3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ClonesUpgradeable.sol",
            name="ClonesUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1
    b"\xcd\x1a\xc0\x4e\x40\xbf\x20\x19\x89\xa2\x4d\xb8\xa8\x58\xec\xcc\xb9\xf7\xc2\xbe\xbf\x50\xc3\xaf\x5e\x76\x45\xe5\x41\xad\xe5\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL2.sol",
            name="LibArbitrumL2",
            version="v4.7.1",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124
    b"\xcd\x29\xdc\x7a\x12\x87\xd8\xdc\xe9\xc3\xb5\xea\xe0\x7d\x52\xca\xb4\x2d\xeb\x3f\x3b\x7d\x9f\xcd\xa0\x67\x0c\xa3\x3d\xcc\xc4\x25": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xcd\x5c\x77\x70\x96\x32\x41\x67\xad\x83\x0c\x2a\xe0\x61\x05\x0a\x94\x57\x0b\xe7\x53\x70\x6e\xf0\x47\xc7\xdc\xff\x96\x84\xe4\xb2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xcd\x6d\x25\x6d\x9c\xf8\x6c\x7c\xeb\x0d\x4f\x27\xf1\x97\x31\x48\x12\xee\xf3\x50\x3e\x7e\xb8\xbd\x77\x15\x0c\x87\x7a\x35\x1d\x7b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363.sol",
            name="IERC1363",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\xcd\x95\x9e\x42\x60\x44\x3c\xf0\x1d\xe3\x84\x89\x87\x51\xcd\xa4\x8c\x78\x39\x29\x23\x27\xbf\x0b\x9d\x32\x5e\x71\x60\x23\x46\xba": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xcd\xaa\x5a\xbc\xfd\x05\xa4\xf5\x8c\x8a\x73\x02\xee\x4f\xaa\xa4\x07\xa0\xfc\x43\x5c\xc8\xbc\xbe\xb6\xbc\xce\x35\xa4\xf2\x8e\x9d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xcd\xaa\x6a\x55\xc5\x9e\xc8\x52\x27\xc1\xa9\x8d\x6e\x5a\x89\xc2\x7c\x9c\x75\xb1\xbd\x2a\xf6\x97\xdb\x6d\x83\xb9\x44\x8e\xa6\x23": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/OwnableUpgradeable.sol",
            name="OwnableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xcd\xb4\x7d\x3c\xe1\x4e\x5c\x8b\xfb\x96\xbc\xa1\x69\x1f\x0d\x4b\xd8\x95\x63\x31\xe7\x03\x8f\x45\x8b\x70\x49\x3d\x59\xc4\xe8\xac": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungibleTokenPositionDescriptor.sol",
            name="INonfung ibleTokenPositionDescriptor",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungibleTokenPositionDescriptor.sol",
            name="INonfung ibleTokenPositionDescriptor",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungibleTokenPositionDescriptor.sol",
            name="INonfung ibleTokenPositionDescriptor",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungibleTokenPositionDescriptor.sol",
            name="INonfung ibleTokenPositionDescriptor",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\xcd\xd7\xc1\x1f\x96\x71\x86\x90\x0e\xca\x54\x56\x3b\xc4\x4d\xa4\x90\x2b\x4b\xc1\xda\xd3\xbc\x6b\x61\x24\x3a\x66\xd4\xe4\x8c\x1a": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xcd\xf4\x98\xc4\x3b\xc5\x73\xf1\xac\x9c\xc1\x3c\xb3\x4c\xe5\x7e\x48\x87\x7c\xbf\x78\xae\x8f\xe1\x50\x6d\x99\x5c\xd1\x09\xe0\xec": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xce\x0c\x41\x02\xbd\xf7\x13\xee\x80\xe2\x56\x33\xa8\xf3\x63\xf4\x57\xea\x40\x29\xaf\xd8\x6c\x50\x0a\xe8\xc6\x01\x24\x39\xfb\xdc": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoter.sol",
            name="IQuoter",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoter.sol",
            name="IQuoter",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoter.sol",
            name="IQuoter",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoter.sol",
            name="IQuoter",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IQuoter.sol",
            name="IQuoter",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xce\x0d\xae\x02\xa4\xa9\x82\x3b\x5e\x7c\xc4\xa3\x46\xbd\x97\x43\x62\xf2\x58\x97\xef\x7d\x55\xe9\x25\xce\xfc\xb0\x04\x85\x03\x75": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xce\x13\xf9\xff\xb4\x7a\x73\xe9\x9a\x7a\xb4\xa3\xd1\x79\x9f\x22\xfd\x0e\xce\x8b\x55\x04\x78\x76\x0b\xb7\xcc\xff\xac\x13\x79\x20": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68, v0.0.69, v0.0.70, v0.0.71
    b"\xce\x51\x31\x6e\x09\x8f\x87\x11\x80\x28\x03\x76\x53\x3c\x0e\x1a\xad\xa4\x71\x75\x4d\x1b\x24\xee\xf4\x44\x7a\xa7\x9b\xeb\x50\x3b": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.68",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.71",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xce\x7e\xef\xf2\xef\x84\xbd\x1f\xef\x75\xb9\x1d\xd2\x70\xd3\x50\xd4\x6b\x85\x08\xad\x83\x7e\x87\x55\x26\x79\xe3\xed\x8a\x4c\xc7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xce\xba\x06\xa8\x28\x88\x4c\xc1\x18\x20\x30\xad\x37\x95\xa5\x51\x11\xcf\x92\x98\x89\x4b\x72\x9c\x16\x55\x5b\x7a\x97\xea\xd3\x0e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xce\xbc\x42\x61\xdf\x0d\x59\xdd\x01\x3c\x7a\xef\x78\xea\xaf\xf9\xb3\xc0\x94\x4b\xa8\x67\x76\x6e\xc7\x4d\x5d\x90\xaa\xea\x31\xb6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC4626Upgradeable.sol",
            name="IERC4626Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xce\xbe\x3a\x9d\xb1\x8e\x8d\x9c\x97\x04\x5b\xae\xab\xf0\xd5\x37\x6d\xc4\xd7\xe9\x92\x7e\xab\x87\x65\x3e\xee\x7d\x65\xc7\x0b\xb3": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/TickLens.sol",
            name="TickLens",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/TickLens.sol",
            name="TickLens",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/TickLens.sol",
            name="TickLens",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/TickLens.sol",
            name="TickLens",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/lens/TickLens.sol",
            name="TickLens",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xcf\x20\x6b\xd3\x6c\x34\xf8\x73\xe5\x73\x32\x56\xce\xb6\xf6\xb5\xdd\x29\x04\x6b\x8b\xe0\x12\xae\x23\x0b\xab\x8c\xa8\x23\xe0\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuardUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xcf\x23\xde\xca\x6f\x53\x55\xa7\xdc\x82\x72\x78\x7c\xc5\x45\x68\xc3\x88\xf0\xa4\xca\xd9\x66\x23\x16\x5c\xcf\x3c\x05\x7a\xef\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xcf\xa2\x79\x27\xf2\x32\xe3\x07\x6e\x03\x96\x57\x07\xeb\x12\xc3\x1b\x04\x86\x76\xd2\x27\x13\xf3\x6c\x33\x1d\xbc\xf5\x2a\x80\x35": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xcf\xfb\x63\xb5\x7d\xd2\x29\x63\x61\xd9\x9d\x2c\x5c\x4a\x7d\xf3\xc8\x72\x32\x2c\xa7\x09\xe8\x48\xa9\xfc\x5e\x83\xea\xb0\x0f\xcb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IOutbox.sol",
            name="IOutbox",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xcf\xfd\x9b\x62\x99\xa2\x3b\x46\xc9\x6a\x64\xc8\x90\xe2\xc5\x8d\xa1\xd8\x1d\x21\xc6\xe0\x08\x46\x90\xe3\x15\x77\x1d\x3f\x76\xd7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721WrapperUpgradeable.sol",
            name="ERC721WrapperUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xd0\x38\x55\x90\x9a\x4a\xfc\xa5\xd1\x23\x26\x0c\x93\xbb\x36\x1f\x5d\x00\x57\xf3\x67\x17\x8e\xa0\x87\x29\xf9\x73\xec\x25\x84\xc1": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd0\x41\x62\x54\xd3\xa3\x89\xab\x97\x11\x90\x08\xcb\xc7\xb7\x10\x2c\xa0\xd0\x6f\xab\xa5\x0a\x7b\xd0\x18\x6c\xc0\x29\x6d\xe4\xf2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/CircularBuffer.sol",
            name="CircularBuffer",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xd0\x4d\x58\x73\x35\x79\x89\x51\xc6\x49\x06\x9d\x7b\x89\xd7\xb3\xb8\xa0\x15\x90\x59\x34\x8b\x4f\xeb\xe1\x9a\xa5\x50\x90\x90\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd0\x66\x9d\x50\xac\x0a\x1d\xee\xea\x4c\x00\x5e\xe6\x74\x60\x1d\x9f\x4d\x58\xb7\x9d\x0e\x5a\x89\x85\xf0\xe5\xe5\xee\x27\xde\x61": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.63, v0.0.64
    b"\xd0\x68\x3b\x34\x12\x17\xb0\x96\x87\xfc\xe2\x8a\x02\xa8\x8c\xca\xfb\xb8\xf5\x28\xed\x4a\xaf\xa7\x9c\x10\x2f\xb9\x3e\xd1\x63\x41": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.64",
        ),
    ],
    # Hash appears in versions: v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63
    b"\xd0\x78\x42\x11\x41\xda\x00\xbe\x36\x56\x13\x9b\x01\xa3\x3d\x51\x00\x36\x9a\xe9\x56\x85\x87\x89\xfe\x3a\xf6\x25\xf2\xda\xc2\xf0": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.63",
        ),
    ],
    # Hash appears in versions: v0.0.16, v0.0.17, v0.0.18, v0.0.19, v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xd0\x7a\x60\x75\x82\x9e\x78\x20\x2d\x40\xe1\xf6\x8a\xaa\xb9\xb5\x1e\x26\xbc\x51\x40\xce\x04\x63\xd4\x5c\xd0\x53\x7b\x16\xc1\x1a": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.17",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.18",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.19",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd0\x97\xf6\x83\x47\x2e\x0a\x5b\xf1\x38\xb3\xd3\xad\xb3\xe5\xe4\xcd\x0a\x8d\xdd\xd8\x29\x95\xdc\xb6\x0c\xf8\x54\x6c\xaf\x6a\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagerUpgradeable.sol",
            name="AccessManagerUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xd0\xaf\x53\xaa\x1a\xf1\x85\x82\x72\x9d\x61\x2a\xe3\x4a\xd8\xfb\x6e\x6e\x26\xaf\x3b\x88\x9b\xf1\x00\xfe\x05\x85\x97\x0a\x39\x95": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/IAccessManagerUpgradeable.sol",
            name="IAccessManagerUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xd0\xbd\x7b\xfe\xbd\xc7\xdd\x95\xc9\x76\x3f\xdd\xfb\x25\xda\x38\xd3\xce\x93\xb5\x88\x25\x3c\x48\x2d\xf4\xf1\x8f\x58\x0b\x01\xf4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerUpgradeable.sol",
            name="TimelockControllerUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd0\xd1\x4e\x64\x69\xb0\xda\x73\x78\x39\xf7\x33\xd3\x05\x3a\x4b\x79\xc2\x75\x81\x14\xf2\xff\xbd\x6f\x20\xcd\xa8\xd0\xfc\x2a\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/Votes.sol",
            name="Votes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\xd1\x29\xc5\xd1\x25\x77\x60\xe1\x28\x72\x3a\x86\xa1\x13\x5c\x3e\xf9\xc9\xe9\x0f\x8e\xd0\x97\xb6\x88\x71\xee\x26\x98\xeb\x2a\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xd1\x29\xd7\xa8\x24\x10\x44\xd7\x6f\xa5\xf6\x10\xdc\x84\x15\xa3\x62\x3c\x17\x57\xb5\x93\xd7\xf3\x05\x70\xf8\xd9\xc7\xbf\x96\x6c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177
    b"\xd1\x37\x48\x4b\x84\xae\x30\xa6\xfd\x0f\x74\x29\xfc\xaa\xa3\xd3\x99\x60\x68\x57\xb4\xf9\x23\xdb\x1d\xde\x33\x1d\x80\x0d\x97\x61": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.177",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xd1\x55\x31\xaf\xa5\x8d\xad\x5d\x4c\xe8\x92\x97\x8c\xbf\x68\xda\x4c\x23\xbe\xd6\x08\xe1\x73\x69\xaf\x50\x4c\xd9\x8f\x96\xce\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712Upgradeable.sol",
            name="EIP712Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xd1\x6a\xb4\xa7\xab\xb3\xa9\x0c\xfa\x17\xf7\xd3\xb6\x35\xd2\x6c\x0a\xc5\xb1\x68\x49\x45\x78\xbd\x3f\xa0\x20\xa5\xda\x4d\x0a\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155PausableUpgradeable.sol",
            name="ERC1155PausableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.3.0, v5.3.0-rc.0
    b"\xd1\x8a\xed\x39\x85\x38\x9e\x21\xd2\xdc\xfb\xda\x86\x8a\x60\x34\x22\xc1\x7a\x83\x7e\x7a\xde\x0e\xb2\xfe\x25\x01\xdd\xc3\xdd\xc6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xd1\x8e\x60\x56\x8c\x25\x52\xa6\x1e\xf5\x50\xb4\x53\xd7\x41\x1c\x94\x99\x51\x49\xf3\x82\xe2\x09\xb6\x1f\x58\xa1\x8b\xf8\x76\xc8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231
    b"\xd1\x9d\x0b\x5f\x18\xf1\x3d\x2e\x0f\x0d\x2e\x8b\xc6\x15\xba\xa0\x68\xb6\x4a\xc9\xe4\x12\x45\xfd\x1b\x83\x58\x09\x3f\x01\x6b\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.231",
        ),
    ],
    # Hash appears in versions: v0.0.174
    b"\xd1\xe3\x42\x37\xca\x86\x94\x0c\x21\xd3\x50\x72\x9b\xab\xf8\xca\xa8\xac\x67\x47\xe4\x37\x4b\x57\xb8\x83\x53\x62\x92\xf0\xb5\xa1": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.174",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124
    b"\xd1\xe9\x87\x81\x04\x7d\xf5\x65\x93\x4f\x21\x94\x51\xbb\x01\xbe\x23\xcf\xc3\x47\x45\x46\x00\x6b\x72\x45\xf7\xb7\xb8\x06\x03\xae": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.124",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.0
    b"\xd2\x09\xb0\x11\x96\xb6\xfb\x34\x50\xe8\xed\x3b\x29\x3d\x05\x3f\x47\x4d\x31\xd9\xe7\x91\x22\xbc\x1e\xb5\xf1\x96\x14\x23\x88\xb1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xd2\x0a\x03\x4d\x92\x88\x07\xbc\xde\x7f\xdc\x87\x23\xbe\x82\xdf\x2e\x4e\x8c\x4c\xd3\x82\xaa\xe6\xf1\x74\xc9\xe7\x31\x75\xed\x63": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.1.15, v0.1.16
    b"\xd2\x2f\xaa\x6d\x1c\xef\x26\x0a\x6f\xb6\x9d\x33\x05\xb4\xd0\x22\x31\xb3\x05\x1b\x3f\x8d\xc6\xff\x30\x6f\x20\xe5\x76\x47\x37\x49": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xd2\x32\x09\x52\xbd\xc7\x43\xbd\x29\xdd\x09\x13\x69\x53\x53\xec\x4e\x09\x5a\x8c\x87\x15\x0c\x0c\xa4\x05\xdc\x70\x55\x88\x15\x25": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xd2\x34\x79\xe5\x59\x59\xd4\x88\xfb\x77\x5a\xe3\x97\x60\x9b\x48\xb6\xc6\x8d\xb5\x13\x7b\xf8\x3a\x36\xbc\x5e\x01\x9a\x47\x8d\xbf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xd2\x49\x0f\xea\xb1\x49\xe8\xe0\xa3\x10\xbb\x97\xb7\x31\x82\x92\xcd\x0d\xa6\xe5\xb1\x42\x9f\x69\x59\x38\x0b\x97\x7e\xf6\x46\x98": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278
    b"\xd2\x67\x38\x20\xf8\xd6\x34\xb6\x0b\xce\xcf\xa9\x3e\x4a\xbb\x8d\x3f\x8e\x25\x43\xc8\x0e\xf3\xa7\xb0\x6b\xab\x59\xbd\xdc\xc5\x7d": [
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.278",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58
    b"\xd2\x7e\x6f\x77\xe4\x0d\x5d\x2d\xed\x17\x2d\xfc\x83\x16\x6e\x70\x9a\x3b\xc9\x1d\xb6\x66\xc4\xac\xd1\x32\xd2\xf6\x5d\x2a\x48\xa7": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.58",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xd2\x8b\x02\xbb\xa7\xbb\xfd\x70\xff\x2f\x30\x20\x1f\x6b\x4e\xc7\x33\xb0\x0f\xf5\x5e\xe7\x4e\x27\xd7\x2e\x63\xbb\x0b\x9a\x6d\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xd2\xab\x31\x3d\x5a\x8e\x16\x9c\xe9\x3f\xf6\xad\x4a\xef\xaf\xe7\xfe\xdf\xd2\x85\x69\x59\x31\x8b\x54\x3e\xb2\xae\x46\x8a\xb4\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xd2\xae\xf9\x94\xf5\x01\x1c\x7d\xbb\x88\xd6\x68\xed\x12\x18\x37\xb3\x18\xc0\xc2\xf9\x08\x5e\x59\xa9\xa8\x26\x13\x8c\xa7\x09\xb4": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd2\xbe\xf6\xed\x64\xfc\x63\x1f\x70\x98\x6e\x13\x34\x9f\xe1\xc2\x1e\x80\xf0\x31\xd8\xe2\x66\x50\x19\x11\xe7\x81\x10\x60\x56\xc5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Upgradeable.sol",
            name="IERC721Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd2\xc2\x96\x86\x0b\x77\x77\x42\xb8\x1a\x37\x97\x59\xc5\xdc\xef\x95\xbc\x77\x3f\xab\xd8\x47\xc7\xaa\x0b\xb5\x06\x75\xf5\x4f\x9f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.170
    b"\xd3\x0c\x27\x38\x94\x3c\xe3\x02\xb6\xc9\x06\x2f\x73\x99\xf0\x0a\xf0\xf7\x7d\xcd\x97\x88\x8b\x9f\xdd\x70\x73\x48\x0f\x85\x4e\x02": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.170",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xd3\x12\x77\xbd\x3a\x06\xc1\xf8\x63\xc6\xb8\xbb\x06\x6e\x1c\x48\x8a\x9a\xda\x0f\x96\xe8\x59\x2a\x74\x7e\xc6\x7c\xc9\xb5\x82\x2a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xd3\x21\x5b\x13\xa8\x6d\x89\x90\x46\xe7\x30\x5a\x8f\x27\xe2\x9f\xbc\x21\xbd\x32\xf0\xf9\x73\xa3\x88\xd3\x51\xe2\xc0\x93\xf6\x13": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xd3\x23\xff\x82\x72\xe0\x7f\xa2\x6d\x33\x43\xd1\xf8\x49\x21\x64\x05\xe7\x67\xfb\xf0\x97\x16\xdc\xdb\x2a\xc8\xd0\x51\x87\x5e\x8a": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibPRNG.sol",
            name="LibPRNG",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xd3\x36\x23\xe4\x71\xe0\xa4\xca\x85\x94\x4b\xf3\xe3\x34\xb1\xfe\x10\xa0\x3d\x3c\x16\x50\xa4\x79\xf1\x92\xf2\x29\x13\x13\x11\x67": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd3\x62\x71\xbf\xa2\xaf\x93\xd9\xda\x12\x4d\x49\x14\x7e\x35\x18\x14\xec\x61\x42\x6c\xa8\xf7\xcc\xa0\x9c\x3c\xb3\xd3\x0f\xa0\xb6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121
    b"\xd3\xb3\x00\xe1\xc9\x21\xaa\xcb\xb3\x1c\x33\xc2\x2b\x31\x93\x12\xad\x26\xcb\x04\xed\xc5\x3d\xaa\xbf\x38\x84\x58\x6d\x26\x12\xdb": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xd3\xb8\xea\xfc\xf5\xa0\x51\x79\xea\xea\x45\xf7\xe8\x48\x76\x76\xea\xd3\xb8\xb2\xc6\xc2\x6d\xa5\x4e\xd4\x11\x62\x6d\xa7\xe7\x09": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165Upgradeable.sol",
            name="ERC165Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd3\xe5\x52\x0e\xe8\x0b\x56\x4a\x65\x0f\x42\x10\x7a\x3d\xab\x5f\xbc\xa0\xc6\xde\x10\x70\x01\x0b\x96\xcd\xd3\xfc\x55\x57\xd9\xe7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xd3\xf4\x1c\x26\x8a\x5e\xe9\x5a\x0c\xfc\x96\x89\x57\x90\x72\xe3\xd2\x91\x4e\xd1\xcc\xb4\x33\xd6\x41\x57\x39\x9b\xa9\x96\x59\xcd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0
    b"\xd4\x21\x6d\xdd\xc3\xed\x7b\x4d\x98\xec\xbd\xcd\x8c\xcb\x1c\xe9\xeb\xba\xa4\xfe\xb2\x11\xcb\x95\xc7\x2b\xcb\xec\x46\x96\xfb\xc0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xd4\x23\xed\xb6\x1a\x87\x29\xb5\x55\xf4\x6c\xf5\x2d\xd0\xce\xc5\x9f\x9a\xd1\x05\xb9\xde\x29\x73\xfb\x9a\xee\x6a\x31\x01\xec\x46": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.8.0
    b"\xd4\x2b\xe0\xce\xaa\xbe\x99\x63\xf9\x3b\xa8\xe2\x20\xfa\x3c\xc9\x8d\x3d\xde\xd2\xf9\x5c\x5f\x0a\x3c\x34\x85\x47\xfb\x38\xd4\xed": [
        KnownContract(
            project="solmate",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xd4\x35\x8a\xd8\x8e\x1f\x84\x7a\x64\x4b\x60\xc3\x2f\x02\xf2\x80\x12\xa2\x21\x38\x56\xce\x4d\xdc\x17\x6b\x0a\x84\xba\xed\xb0\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xd4\x63\x3f\x23\x91\x4a\x87\x22\x19\x54\x1b\x9f\xb2\x95\xca\xe2\x8c\xd6\x06\xc9\x8d\x1f\x6c\x42\x5d\xb2\xa8\xd1\x1e\x6d\x8a\x36": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155BurnableUpgradeable.sol",
            name="ERC1155BurnableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1
    b"\xd4\x65\x6c\x83\x07\xd4\x04\xe3\x89\x21\x3d\x62\x14\x92\x4e\xaf\x57\x85\x85\x73\x18\x78\x34\xeb\x89\xd4\x8b\x80\x59\x6d\x90\x4f": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.1.1",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xd4\x75\x63\xb6\x0a\x80\xb9\xb6\xe0\x14\xbc\xa8\x28\x52\x8f\xd9\xcf\x0d\x71\x06\x32\x7f\x1a\x46\xd8\x28\xed\xef\x92\x61\x55\xde": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v3.2.0
    b"\xd4\x89\x02\x84\x74\x88\x01\xa7\x5a\xaf\xca\x35\x9a\xe6\x2e\x6c\x8e\x89\x6e\xbd\x43\x32\x0a\x73\x66\xe3\x8c\xc7\x40\xf3\x9a\x35": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.2.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xd4\x93\xba\x20\xed\x24\x01\xcd\xcd\xca\xe5\x81\x7d\x22\xd9\x31\xc2\x15\xa6\x03\x1d\x89\x0a\x55\x75\x22\xf4\xc1\x7c\xb1\xf4\xe7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xd4\x98\xa8\xc3\xf4\xa6\xd1\x6c\x9b\x5a\x7d\x61\x2e\x8e\xa2\xa0\x0e\xf0\x5c\x80\x09\xa5\x24\xce\x02\xe4\x3c\x96\x2b\x48\xca\xb8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/CrossChainEnabledArbitrumL1Upgradeable.sol",
            name="CrossChainEnabledArbitrumL1Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191
    b"\xd4\xc3\x21\x87\x73\x72\x53\xb6\x32\xb6\x63\x26\x04\x34\xa3\xf3\x01\x11\x89\xeb\x03\x68\xff\xb4\xe8\x26\x48\xba\xfd\x87\xc7\x12": [
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.191",
        ),
    ],
    # Hash appears in versions: v3.3.0
    b"\xd4\xc3\xf0\xe5\x6b\x2f\x55\x6c\x80\x96\x45\x44\xcd\x53\x39\x24\x4a\xe6\x06\x40\x9c\xd5\x71\x0f\xb0\x11\xa2\x3e\xf0\xbc\xf0\x02": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v3.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.29
    b"\xd4\xe3\x7f\x77\xbf\x95\xed\x67\xbc\x1a\x70\xe7\x9b\xdb\x8b\xe0\x93\x4a\xb4\xf3\x9a\x65\x37\xca\x5c\x4b\x49\xd5\x77\xe5\x16\xeb": [
        KnownContract(
            project="solady",
            path="src/utils/LibBit.sol",
            name="LibBit",
            version="v0.0.29",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0, v6.6.0, v6.6.1, v6.6.2, v6.7.0, v6.8.0
    b"\xd4\xe6\x0d\x8b\xf3\x26\x72\x74\xd4\x96\xb0\x45\x68\x31\x88\x08\x03\x12\xbf\x97\x8c\xf7\xeb\x39\x5e\x8b\x98\x1d\xb6\x1d\xb1\x18": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.5.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.6.2",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.7.0",
        ),
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.8.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xd5\x19\x60\x74\xa8\xed\x4c\xd1\xa1\x7c\x48\x0f\x09\x9c\xea\x8c\xce\x74\xcb\x5d\x34\xc3\xef\x9e\xff\x5e\x06\x16\x9d\x91\x64\xc2": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/V4Router.sol",
            name="V4Router",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xd5\x46\x80\x3c\xf2\xf4\x06\x27\x64\xcc\x73\xf6\xfb\x2a\x92\x1e\x6f\x24\x5d\xba\xa3\x50\xa6\xfa\xb7\x4d\x70\xdf\x75\x10\xea\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/SafeERC20.sol",
            name="SafeERC20",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xd5\x4a\x46\x5d\x0a\x16\xd6\xca\xd2\x70\xf1\x00\x30\x18\xc3\x67\x0d\x3f\x67\xcc\x8d\x15\x96\x12\xbc\x8b\x46\x96\xb7\x2a\x67\xa8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd5\x60\xb0\x07\x56\x06\xdd\x3a\x08\xd5\xee\xab\xcc\x0e\x61\x54\x83\x04\x6a\x1b\xd1\xb7\x9d\x3a\x3a\x3d\x05\xab\x69\x1f\x9a\x5a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xd5\x82\x64\x1a\xb6\x47\xd1\xfa\xf1\xd9\xf5\x0e\xe1\xf4\xb5\x69\x6c\xc7\x08\x11\x8e\x4b\x0b\x8c\x99\x4a\x63\x74\xe0\x20\x2e\x57": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Metadata.sol",
            name="IERC721Metadata",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xd5\xa3\x2e\x4d\x02\xa8\x13\xc9\xdb\x71\xc9\xf9\xb9\xed\xde\x8c\x68\xc6\xfa\x77\xec\x0c\xa0\xf0\xe2\xa0\x7b\x2e\xee\x9b\x7a\x66": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/SafeCast.sol",
            name="SafeCast",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xd5\xcb\x77\x19\x4e\x6f\x8b\xf5\x54\x79\x21\x36\x56\xab\x56\xfe\xe2\x23\x79\x22\x4b\x7a\xb2\xcd\xe8\x09\x46\x49\xdb\xfe\xa3\xb6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimpleUpgradeable.sol",
            name="GovernorCountingSimpleUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd5\xcf\x45\x05\x66\x85\xb3\x56\xc5\xf8\x44\x6f\x49\x79\xc8\x9b\x9f\xda\x0f\x52\x36\x0d\xc7\x7c\xc3\x2d\xfb\x14\xbb\x08\xc8\xdc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xd5\xde\x5e\x21\x3f\xfa\x25\x87\x1e\x3e\x74\xf4\x48\x1a\x54\x12\xcf\x90\x6f\x4c\x4b\x89\x11\xc9\x4e\xd2\x99\x03\x56\xf7\x98\x05": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\xd6\x06\x18\x92\xe1\x07\xd2\x9b\x29\x47\x4c\x71\xcf\xb3\x83\x8c\x3b\x02\x8b\x59\xa1\x77\xf6\xa9\x3f\xc6\x6d\x26\xfd\x30\x5d\x92": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/SignatureCheckerUpgradeable.sol",
            name="SignatureCheckerUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd6\x36\x76\x8f\x61\xd3\xed\x71\x3e\xe9\x86\xe6\xf9\xba\xf3\x3a\x78\x25\x66\x61\x00\xa6\x11\x1a\xf0\xcd\x93\x84\xee\xe2\xd6\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xd6\x3a\x80\xb7\xe3\x57\x75\xa8\xf4\x7e\x23\xba\x60\x90\x47\xd3\x1a\x73\x41\x22\xc3\x9d\x37\x4a\x46\xfa\x3e\xfc\xe3\x6e\x85\xd7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.150
    b"\xd6\x41\xe0\x7c\x4f\x21\xc6\x5e\xc4\x53\xa8\x05\x84\xdd\xb8\xae\x1f\xd9\xff\x77\xc7\x65\x6f\xaf\x17\xf8\x0f\x98\x4f\xbf\xd5\xea": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.150",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd6\x53\x2a\x68\x96\x62\x14\x12\xef\x6b\x0e\x41\x70\xc9\xb0\xee\x28\x45\xa0\x33\x81\x32\x5d\xd2\x76\x0a\x81\x94\xb9\x37\xf0\x3a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.0.1, v3.0.2
    b"\xd6\x5a\x1f\xbd\x2a\xa1\xeb\xf1\x3b\xaf\x97\xb4\xb4\x29\xea\x09\xbc\x0f\xfd\x3f\xf8\x52\xbf\x6c\xbe\x09\xf0\x84\x35\x04\x8b\x00": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.77
    b"\xd6\x6f\x52\xb8\x39\x69\x5c\x9e\x6c\x60\xe2\xd1\x84\xd7\xc2\x24\xf6\x7d\x09\x8e\x02\x2f\x71\xbb\x8a\x0a\xbf\xae\xd7\x5f\xa9\x77": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.77",
        ),
    ],
    # Hash appears in versions: v0.0.130, v0.0.131
    b"\xd6\x6f\x72\xcb\x1f\x6f\x8f\x1d\x31\x5a\x93\xb4\x70\x59\xe7\x53\x12\xfd\x22\x83\x5a\x25\xb9\xde\x9e\xcb\x92\x26\x4a\x4d\xb2\xb3": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.131",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xd6\x7d\xf3\xaa\x9e\xb1\x72\xa8\xf1\xc5\xf0\x2a\x8a\xc7\x17\x7d\xe9\xfa\x87\xb5\xac\x26\x46\xc6\x98\xff\xa0\x89\xf3\x44\x4e\xca": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xd6\x81\x63\x08\xcb\x13\x51\x33\xc0\x13\x8d\x65\x30\x20\xe8\x9f\x37\x5d\x5a\xa9\xa7\xfa\xf1\x94\x76\xb0\x01\x0d\x27\xad\xa8\x29": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xd6\xb0\x8a\xc2\xb3\x5b\x38\x8d\x72\x60\x27\x89\x4c\x9a\x76\x46\x3c\x28\xee\xf6\x59\xb3\xa5\x79\x92\xf8\xb1\xf0\xfd\xc4\xdd\x4a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xd6\xc5\x7a\x6f\xae\xe2\x83\x8e\xdb\x45\x21\x2a\xd5\xaa\xeb\xc5\xef\xfc\x53\x5a\xd9\x09\x53\x3a\xdc\xa9\xa5\xbf\x24\xc0\xf3\xb4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777UpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xd6\xdb\x54\xb1\x18\xfb\xde\x70\xe2\xf7\xe4\x59\xce\x52\xb0\xe8\xfd\x1d\x7f\x6b\xe5\x7f\x59\xbd\xfe\x73\x13\x8e\xca\x12\xfd\x0a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: main
    b"\xd7\x13\x7e\x97\x5f\xf7\x9d\x94\x9e\x83\x27\xf0\x01\x49\xf1\xbe\xc4\xcc\x2c\x9f\xb0\x97\x8b\x01\x13\x0d\x15\x98\xd2\xcf\x35\xc2": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IERC721Permit_v4.sol",
            name="IERC721Permit_v4",
            version="main",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\xd7\x22\x4e\xf1\x4e\xaa\xc8\x3c\x48\x98\x19\xf0\x30\xab\xdb\xc8\x35\x07\x68\x40\xfd\xad\xfa\xe8\x92\x70\xe8\x00\xd1\xd7\x1a\x20": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd7\x4b\xeb\x29\x66\x7a\xf7\x48\x3b\xdb\x50\xf3\x77\xf7\xac\x09\x69\x2a\xad\xca\xb5\xf7\x65\x3a\x21\x70\xda\xf6\x02\x07\x57\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.273
    b"\xd7\x4d\xbf\xbe\xc3\xf4\xf0\xa6\x54\x06\x08\x9f\x67\xad\x37\xf6\xa2\xd2\x40\x7c\x52\xef\xe3\x39\xa9\x92\xe1\xac\xd5\x6a\xc5\x14": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xd7\x5e\xef\x74\x6a\xdd\x8a\x8d\x8a\x0d\x8f\x60\xb2\x68\x00\x76\xa0\x19\x0e\x93\xcb\xb8\xc0\x62\x22\xf2\x42\xf9\xe0\xf0\xcc\x15": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd7\x69\x8f\x3c\x6a\x37\x29\x41\x79\xf6\x14\x3a\x9f\xd7\x05\x89\xef\x68\x97\x62\x20\xbc\x1c\x57\x2e\xc1\x21\x8e\xd3\xd0\x54\xf2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorSettingsUpgradeable.sol",
            name="GovernorSettingsUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd7\x6e\x5c\x7e\x7b\x53\x8c\x5e\x0f\x9a\xc3\x9f\x3f\x02\x33\xee\x82\xb8\xfd\x00\xb1\xcf\xb1\xca\x83\x41\xae\xed\xce\x88\x30\x19": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorPreventLateQuorumUpgradeable.sol",
            name="GovernorPreventLateQuorumUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\xd7\x78\xfe\x27\x57\x14\xc6\xa5\xf0\xa6\xdd\xe1\xa5\xcb\x40\x86\x44\xc6\x69\xa6\x5d\x53\x57\x30\x8b\x73\x2c\x11\x5d\xf1\xb7\x6c": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\xd8\x0f\xc9\x4c\xea\xc1\xb6\x00\x64\x0b\x73\xd9\x12\x24\x7b\x8c\x93\x61\x00\x1d\x89\x2a\x6f\xa6\xf5\x7e\x44\x78\xad\xd5\xc9\xf1": [
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xd8\x11\x73\xf2\xb1\x63\x97\xcd\x61\x03\xc1\xa5\xa3\xcb\xf7\x29\x5a\x27\xe5\x35\xaf\xb5\x76\xb0\x9a\x76\x6c\x79\x39\x63\x49\x66": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/Math.sol",
            name="Math",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd8\x3d\x99\x2e\x87\xe4\x25\x01\xda\xda\x95\xe0\x46\x9d\xd6\x9a\x27\x24\x65\x44\x10\xcb\x95\xec\x69\x5d\xc2\x79\x11\x50\x96\x42": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363ReceiverUpgradeable.sol",
            name="IERC1363ReceiverUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78
    b"\xd8\x75\x4c\x49\x38\x6e\x3c\xd0\x4f\xe8\x83\xab\x2d\xe1\x37\xfe\x68\xfd\xe3\x53\xcc\x8d\xec\x44\x6c\xb6\x68\x5a\xa8\xd9\x2a\x51": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.78",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xd8\xad\x83\x30\x68\x61\x9d\x3a\x74\xb1\x20\xdf\x73\x65\xc8\x0f\x3b\x83\xb9\x81\xfa\x58\x6e\x7d\x0b\x74\x9c\xa1\x60\xe1\xdf\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/IGovernorTimelock.sol",
            name="IGovernorTimelock",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xd8\xcd\x1f\x96\xa2\x09\x65\xfe\x1b\x50\x83\x74\x06\x76\xd3\x83\xb3\x79\x7b\xa0\x8e\xb4\x6f\x4c\x21\x3b\x0e\x27\x85\x78\x74\xd3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xd8\xd0\x46\x04\x47\x3d\x43\xae\xba\x0d\x74\xbf\x62\xd7\x6c\x9d\x66\x59\x03\x6e\xef\x41\x54\x44\x15\x03\xb0\x2d\x34\x7b\x56\xae": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xd8\xe3\x2b\xf9\x01\xfb\x04\x00\x2c\x33\xd5\xa7\x6a\x5a\xc3\x44\x29\x28\xdf\x98\x81\x39\x89\xed\x84\x1a\x55\xbc\xc8\x90\x4d\xd4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravoUpgradeable.sol",
            name="IGovernorCompatibilityBravoUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xd8\xf4\x4b\xa8\x42\x13\x51\xaa\x7d\xfc\x17\xde\xe6\x4b\x42\xf3\xa8\x90\xcb\x26\xa6\xdb\xe0\x7c\xaa\xd6\x7c\x37\x24\x11\xf9\x19": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xd8\xfd\xe7\x1f\x3e\x4e\x8e\x82\xd4\x40\xe8\x07\x74\x2d\x55\x1a\xe6\x8e\xf0\xa2\x1d\x88\xc8\xe3\x23\x9c\xf1\x07\xc1\x46\xab\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorStorage.sol",
            name="GovernorStorage",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.129, v0.0.130, v0.0.131, v0.0.132
    b"\xd9\x07\x31\x45\x39\x5a\x4e\x70\xa3\x55\x74\xf0\xb8\xca\xff\xef\x0f\xe3\x48\x97\x8f\x8d\x17\xe9\x60\xe0\xdb\xf2\xdd\x52\xf0\x00": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.132",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0
    b"\xd9\x1b\xe6\x6f\x48\xf2\x56\x32\x33\xff\x0f\x18\xa6\x2b\x6c\x75\xab\x7b\xdc\xb5\x2f\x52\x22\xd3\xf4\x2c\x42\x6c\x3d\x60\x9d\x62": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\xd9\x1c\xe5\x39\xec\xa0\x32\x22\xec\x9e\x9d\x80\x2e\x80\x8d\x99\x77\x80\xaf\x67\xd5\x3a\x5e\x86\x3f\x50\x97\xb7\xec\xc4\xdb\x28": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xd9\x24\x55\x35\x69\x34\x54\x22\x12\xdd\xb7\xe9\xea\xda\x1a\xc3\xd7\x4d\x42\x44\x0e\x4a\xcd\x9c\x51\x12\xbc\x7a\x40\x04\x6c\x57": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryValidation.sol",
            name="PeripheryValidation",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryValidation.sol",
            name="PeripheryValidation",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryValidation.sol",
            name="PeripheryValidation",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryValidation.sol",
            name="PeripheryValidation",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/PeripheryValidation.sol",
            name="PeripheryValidation",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.88, v0.0.89, v0.0.90, v0.0.92
    b"\xd9\x43\xa0\xf1\x6b\x01\xbc\x2b\x3f\x56\x19\x0a\x08\x18\x1b\xe3\x55\xe1\x33\x54\x0b\x68\xbe\xd5\x64\x41\xf5\xd2\xea\x14\x85\x74": [
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Clone.sol",
            name="Clone",
            version="v0.0.92",
        ),
    ],
    # Hash appears in versions: v0.1.14, v0.1.15, v0.1.16
    b"\xd9\x59\x6d\x29\x00\x26\x42\x38\x4f\xe1\x57\xee\x6d\x6f\x40\xfc\x73\xe1\x6a\x7f\xef\x6b\x14\x7c\x6a\x7d\x67\x94\x75\x89\xfd\xfc": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xd9\x71\x80\xbd\xf8\x63\x78\x4d\xf1\x74\xcc\xb0\x22\xd4\x9c\x5c\x11\x6a\x60\xb9\x9f\xfd\xb4\x49\x53\x2a\x3f\x18\xf5\xca\x00\xdf": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/CustomRevert.sol",
            name="CustomRevert",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xd9\xa2\x14\xb3\x9f\x92\x70\x5d\x66\x61\x8a\xa4\xec\x03\x0c\x03\x2f\xb9\x3e\x54\x2d\x1d\x1d\xd8\x57\x6c\xb9\x86\x4e\x14\xf5\x33": [
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibStorage.sol",
            name="LibStorage",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xd9\xb6\xf8\x0f\xc3\x95\x1e\xa7\x66\xf9\x53\x05\x03\xc1\x19\x66\x98\x26\x78\xad\x85\xf3\xed\xf2\x92\x73\xaa\x42\x21\xa4\x31\x52": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xd9\xc8\x75\xb4\xd4\xb1\x33\x08\xa7\x66\x43\x5e\x90\xbe\xc3\x1d\xc8\x80\x27\xd8\x79\xa1\x92\xe3\x48\x03\x3f\x5a\x26\xf8\x5c\xcc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xd9\xd5\x89\x12\x23\x15\x09\x68\xd7\x6e\xcf\x17\xe8\x75\x77\xb0\x4e\x75\x23\x47\x0c\xa9\xe1\xe2\x74\xc2\xd1\xf9\x6c\xac\xda\xea": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2.sol",
            name="Create2",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xd9\xf0\xd5\x3c\xfb\x76\x3b\xce\x26\x76\x79\x66\x6c\xdf\x6d\x01\x49\x27\x84\x43\x16\x1c\xfc\x39\x99\x15\xa9\x46\x7e\xc7\x16\x6b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientSignatureUpgradeable.sol",
            name="GSNRecipientSignatureUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xda\x2d\xb7\xe3\x41\xeb\x66\x81\x2e\x42\x4b\xd7\x43\xc1\x25\xb5\xbb\xd1\x13\x40\x66\xc2\xe5\x02\x4d\x94\x3c\x4e\x63\xf7\xa6\xb3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotes.sol",
            name="GovernorVotes",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xda\x4b\xd4\xa5\xef\x83\xa8\x11\x0c\xe6\x16\x25\xd6\xde\xfa\x00\x91\xf4\xde\xe7\x7b\x4a\xc7\xe0\x85\x6b\x6f\x40\xa9\xbd\xbb\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarderUpgradeable.sol",
            name="MinimalForwarderUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.112
    b"\xda\xa9\xc5\x66\xef\x44\xb6\x4c\x1f\x25\x8e\x2f\x8f\x6c\xbd\x71\xb1\x68\x51\x67\x30\xef\x1f\x41\xa2\x3b\xe8\x21\x53\xad\x30\xa6": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.112",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\xda\xb3\xea\x25\x26\xfe\x33\x37\xad\x99\xa8\x53\xf3\x0e\xcc\x97\xeb\x41\xbc\xbe\xbd\x5f\x10\xdc\x1f\x7b\xb4\xdf\x3a\xcd\x7b\xf4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xda\xb7\x17\xde\x7b\x46\xd8\xb6\x6a\x67\xe8\x9b\xe4\xc5\x99\x90\x0e\x37\x9a\x18\x75\x19\xf0\xd7\x27\x43\xb6\x90\xde\x57\xf2\x69": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC6551.sol",
            name="LibERC6551",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xda\xbe\x24\x7d\xac\x3a\x2a\xed\xe6\xa4\xa8\x64\x4d\xfb\x4a\x41\xb2\x2e\x23\x98\xe6\x5d\x49\xcb\x04\x1e\x62\x83\xfd\xe1\x18\x7c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xda\xf5\x60\x9d\xb4\xe2\x15\x06\xda\x0b\x26\x4c\x04\xd7\x63\x69\x42\xa6\xcb\x0a\xe2\x4d\x52\xe2\x03\x17\xec\x8b\x5d\xe9\x8a\x50": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC4626.sol",
            name="ERC4626",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xdb\x0e\xbd\xf9\xd3\x46\x61\x7c\x6c\x9d\xa0\x9c\x9d\xb3\x0a\x5d\xe1\x77\x6c\xba\x01\xb6\xd2\xb4\x4e\xc0\xce\x54\xa1\xf0\xa0\x29": [
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Lifebuoy.sol",
            name="Lifebuoy",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xdb\x11\x52\x4f\xfe\x48\x0b\xb6\xbd\x12\x04\xa7\x2a\x9c\x2a\x5e\xc0\xe3\x8b\x8f\x4a\x69\x7f\xe0\x1d\xd4\xba\xf4\x8a\x81\x34\x05": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xdb\x40\x3c\x68\x0d\x25\x1b\xd8\x68\x3b\xaf\x2b\xcc\x88\x41\x8d\xd0\xe3\x7a\x3e\xda\xc9\x1c\xb7\x7c\x2b\x0f\xaf\x37\xf5\x14\x46": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/amb/IAMB.sol",
            name="IAMB",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xdb\x45\x23\x04\x51\xc9\xa3\xfc\x5a\x88\x43\x17\xde\x4f\xb1\x9b\x78\xe5\x97\x24\x1d\x75\xaa\xed\xf4\x9d\x15\xa7\x06\x66\xf7\xcc": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155HolderUpgradeable.sol",
            name="ERC1155HolderUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xdb\x5f\xeb\x11\x54\x06\x1b\x3d\x3a\x0b\xda\xdd\x56\xc5\x13\xf0\x73\xe1\xe2\x2b\x0a\x5a\x9b\x8d\xec\x28\x1e\x51\x14\x60\x98\x8f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6093.sol",
            name="IERC20Errors",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92
    b"\xdb\x64\xd9\xc4\x43\x38\x9f\xbd\xc2\x1c\x18\x06\x4c\x68\x80\x15\x71\x64\xd4\xd3\x6e\xd1\xff\xc7\xde\x4b\x72\x31\xd9\xd4\x7f\xbb": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.92",
        ),
    ],
    # Hash appears in versions: v0.0.119, v0.0.120, v0.0.121
    b"\xdb\x8c\x51\x7d\x67\x93\x0e\x10\x70\x86\x70\x30\xf0\x7e\x9b\x8d\x5d\xed\x6d\x3f\x02\x4f\xcf\xb0\x98\x66\xf2\x28\xb8\xd3\xe6\x21": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC20.sol",
            name="ERC20",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\xdb\x9c\xfd\x00\x67\x1a\x35\xbc\x24\x99\x15\xbe\xfc\x79\xa7\xd9\x19\x81\x78\x77\xa6\xde\x52\x7e\x85\x6c\xed\x6e\x98\xcb\xb1\xfe": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xdb\xa0\x47\x9e\x2b\xeb\x25\xa7\x52\xc8\xdb\x14\xf8\x17\x2e\xcc\xf5\x2e\xa7\xe0\x5a\x98\x81\x21\xb4\x3e\x47\x49\x5a\x24\x8c\x37": [
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Base64.sol",
            name="Base64",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xdb\xa2\xc3\x01\x6d\x05\xe8\xfb\xe6\xdc\xf0\xed\xe4\x0c\xe0\x6d\xc2\x4d\xa4\xb8\xe7\xb5\xc5\x89\x7b\xb5\xd1\x1a\x9e\x0f\xad\x33": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlCrossChainUpgradeable.sol",
            name="AccessControlCrossChainUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.229, v0.0.230, v0.0.231, v0.0.232
    b"\xdb\xad\x7a\x59\x0f\xaa\xc3\x9a\xc5\x35\xc1\xcd\xe5\x5e\x12\x37\x1c\x85\x2b\x64\x4c\xfa\x9c\xde\xbc\xeb\xbd\x13\x75\xf9\x74\x1d": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.232",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xdb\xf0\xc1\x6b\x8b\x72\x7d\xbe\x7a\xc3\x83\x91\xb9\x0c\x63\x08\x61\xd1\xe3\x32\x9b\xb0\x09\xa7\x17\x9f\x0d\xc2\xbd\x22\xd0\x57": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981.sol",
            name="IERC2981",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.53
    b"\xdc\x40\xcd\x7e\xb3\xcb\xf4\x5f\xfe\x99\xff\x51\xb2\x22\x04\x3f\x93\x7d\xa7\x31\xa1\x67\x3b\x3c\xc1\x75\x8c\x9d\xba\xfc\xaa\xc2": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.53",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xdc\x45\x48\x6f\x56\xc4\xed\x78\x4e\x06\xab\xc5\xeb\x03\x06\x2f\x66\x56\x29\x9e\x59\xa0\xc2\xdc\x07\xf3\x20\xbf\x56\xe0\xd0\x80": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2981Upgradeable.sol",
            name="IERC2981Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.238, v0.0.239, v0.0.240
    b"\xdc\x4c\xf8\xce\x01\xb6\x1a\x71\xc8\x1f\xa2\x80\x31\x9b\xeb\xe0\x95\xd7\x17\xe6\xd8\x37\x31\x82\xf8\xca\x6b\x42\x1c\xf5\x9f\x77": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.240",
        ),
    ],
    # Hash appears in versions: v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208
    b"\xdc\x5d\x0d\x4a\x8b\x13\x19\xce\xe8\x47\x18\xfc\xd1\x90\x1e\xe1\x55\x43\x8f\x4c\xdb\x34\xa3\x14\xe2\xa2\x0c\xd2\xfd\x84\xa3\x5d": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.208",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xdc\x74\xfa\xa3\xfc\x8d\x67\xa1\x53\xc4\x45\xbd\x87\x3a\x30\xdc\x08\x46\x8f\x0d\x3a\x63\x9b\x62\xc5\x2a\xe2\x57\x59\x21\x63\x56": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/IGovernor.sol",
            name="IGovernor",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xdc\x7d\x57\x6c\xbc\x84\x25\xd1\x59\x95\x4d\x3c\x8e\xa0\xd7\x44\x9f\x9f\x9f\x16\x99\x10\x6e\x6a\xb7\x6c\x24\x5f\x37\xdd\x75\xf1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Hook",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Hook",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Hook",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Hook",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579Hook",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xdc\x85\x2e\x4d\xdd\xeb\xf4\x07\x09\x22\x52\x87\xef\x29\xdb\x6d\x88\x0d\x07\x7f\x95\x9b\x06\x5e\x2e\x63\x84\x7c\x43\x7c\x7d\x49": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP10.sol",
            name="CAIP10",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP10.sol",
            name="CAIP10",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP10.sol",
            name="CAIP10",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP10.sol",
            name="CAIP10",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CAIP10.sol",
            name="CAIP10",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xdc\x97\x09\x2d\x56\x56\x27\x3e\xfe\xa3\x5d\x63\x3d\xb3\x6c\xf4\xbf\x87\x44\xe7\xc3\x93\xd7\xdb\xa0\x55\x85\x36\x8b\x41\x69\xec": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ShortStringsUpgradeable.sol",
            name="ShortStringsUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xdc\x9c\x19\xda\xd9\x6c\x1c\x9f\x43\xf0\xa2\x86\x01\xec\x04\x35\x87\x8a\xc9\xb6\xb1\x7b\x38\xbb\x44\xf3\x24\xfd\x76\xd6\x86\x7a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Capped.sol",
            name="ERC20Capped",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238
    b"\xdc\xad\x5e\x7e\xee\xce\xe1\xa0\xb1\x23\xe7\x73\x63\x2a\x44\x8a\x92\xca\x56\x78\x07\x90\x4e\x22\x3c\x2c\xc5\x2a\x53\x95\x5e\x77": [
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibRLP.sol",
            name="LibRLP",
            version="v0.0.238",
        ),
    ],
    # Hash appears in versions: v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89
    b"\xdc\xc5\xc2\xed\xdf\x06\xda\x24\x01\xc1\xe2\xe2\xa3\x8f\x9e\x41\xf8\x52\x05\xf7\x4b\xfa\x42\x3a\x26\x02\x87\x40\x56\x3a\x44\x11": [
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibMap.sol",
            name="LibMap",
            version="v0.0.89",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3
    b"\xdc\xe0\x1c\x1a\x75\x7a\x64\xa7\xd8\x42\x9c\xbf\x02\x0e\x35\xd1\xf3\x25\xe9\xdd\x31\xff\x62\xe4\x4f\xd3\xc2\xc1\x0e\xbd\x8f\x79": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.3.3",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xdc\xee\x76\x1f\x62\x68\x85\xf8\xf0\x6b\x66\xa2\x43\x5f\xd5\x1e\x2a\x21\xe7\x5c\xa1\xe1\x5a\x14\xd3\x62\x46\x3a\x53\x22\x69\x36": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipient.sol",
            name="GSNRecipient",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xdd\x0c\x25\x12\x6c\xec\x32\x1b\xab\xc1\x83\x61\xc1\xef\xf9\xcf\xe9\x93\x6a\x5d\x94\x37\x04\x18\x7f\xde\x5b\x12\xa4\xd5\x31\x22": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xdd\x0e\x92\x1c\xa1\x0c\x73\xa7\x7c\x09\xc9\xee\xdb\x04\xeb\xbf\xef\x84\xed\x70\x1e\x69\xce\x61\x89\x78\xd2\xe9\xd5\x69\x9d\xef": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xdd\x4f\x25\x8d\xc9\x78\x90\xa1\x48\x7e\xc6\xef\x35\x95\xda\x96\xf2\x8a\xad\x23\xd3\xc4\xba\x93\xc3\xf6\x69\x0d\xd0\xef\x62\x09": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\xdd\x53\xbd\x80\xd9\xc3\x9d\x79\xc2\x05\xa7\xf3\xbb\x13\x62\x87\x64\x3f\x7b\x07\x55\xea\xa7\x2b\x3f\x7b\xe8\x9f\xfe\xa0\x8c\x83": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273
    b"\xdd\x60\x4f\x4b\xd7\x6f\x52\x8d\x9c\xcb\x24\x3e\x6f\x5f\x28\x6a\xb0\xa1\x8b\x72\x50\x9a\xb4\x02\xa6\x38\x12\x48\xa4\xc9\xaf\x04": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.273",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\xdd\x60\xed\xc6\xd9\xaf\xf2\xc6\xd9\x66\x34\xcc\xf0\xd0\xfe\x6c\x6d\xd7\x76\x7f\xd6\x96\xe8\xd9\x19\xcc\x77\xeb\x87\x49\x48\x90": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xdd\x9e\x35\xbc\x33\xe2\x43\x2c\x49\xda\x72\xf5\x9b\x46\xcb\xff\x92\x1a\x8a\x24\x4d\xe0\x3d\x6f\x46\x8d\xad\xb8\xa2\x70\xe5\x79": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/IERC20Upgradeable.sol",
            name="IERC20Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xdd\xeb\xf5\x02\x41\x7e\x2c\x8a\xf6\x61\xf3\xd4\xd9\x8d\xaf\x8c\xbb\x7a\xbc\x17\x7b\xd7\xb4\xcb\x22\xfe\x4d\x41\x0f\x27\xdf\xa4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravoUpgradeable.sol",
            name="GovernorCompatibilityBravoUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xde\x3e\xd2\x79\x2e\x2e\x3e\xba\x5e\xac\x55\x41\x3a\xee\xe9\x72\x0b\xee\xb0\x52\x1f\x8d\xf2\xe6\x71\x65\x32\x75\x70\xa7\x95\xca": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xde\x52\x6b\x6a\x1d\x21\x52\x61\x49\x0c\xf1\xa0\x48\x22\x17\xe5\xc3\x48\xfb\x2a\xc3\x1e\x20\x67\x8f\xa2\x19\x73\xb3\x1f\x37\xbf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/Governor.sol",
            name="Governor",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\xde\x61\x80\x07\x56\x27\x1d\x54\x4b\xde\x86\x84\x34\x68\xd0\x43\x4a\xa7\xef\xfe\xca\x80\xae\x7e\x88\x37\xb9\x39\x49\x95\x8f\xf8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144
    b"\xde\x91\x03\xf9\xe4\x11\x7a\x86\x61\x20\xeb\xe6\x61\x5d\x69\xb6\x1a\xc4\xe1\x37\xb8\x0f\x10\xe0\x73\x82\x13\xbe\x99\xd0\x28\xbc": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC6909.sol",
            name="ERC6909",
            version="v0.0.144",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xde\xcc\x0e\xf5\xec\x59\xf5\xc7\xed\x93\xc8\x5e\xce\xc3\x93\x3b\x14\x95\xf0\x1a\xa7\x04\xbc\x4e\xa9\x84\xe8\xa8\x4c\xf5\x21\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.226, v0.0.227, v0.0.228
    b"\xde\xdb\xbc\x4d\xf0\xb3\xd7\x75\x09\xf7\x98\x8c\xeb\xbe\x4e\xdc\x73\x38\xb8\xd8\x09\xf4\x31\x2d\xe4\x44\xc9\x8b\x87\x99\xbe\x44": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.228",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xdf\x1d\x68\xa9\xd5\xcc\x48\xa9\x24\xec\xfe\x95\xd4\x2e\x8b\x03\xb3\xec\x1e\x61\xe0\x1c\x54\xa2\x31\xb7\xbf\x26\x99\x16\xcf\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xdf\x26\x16\xeb\x71\x96\x4a\x6d\x99\xb8\x44\x40\xd8\xab\xac\xd4\xb5\x0c\xd8\x48\xde\xd6\x0f\x35\xa7\xb6\x4f\x99\x28\xec\xf1\x32": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPayments.sol",
            name="IPeripheryPayments",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPayments.sol",
            name="IPeripheryPayments",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPayments.sol",
            name="IPeripheryPayments",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPayments.sol",
            name="IPeripheryPayments",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPeripheryPayments.sol",
            name="IPeripheryPayments",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xdf\x5f\x91\xb3\xb4\x8f\xbd\x78\x8c\x45\xa4\x5e\x92\x2d\x40\xdd\xca\x4f\x71\xd4\xa0\x9a\xe0\x8c\xd2\xdd\x17\x35\xb4\xc2\xa7\x7e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/BeaconProxy.sol",
            name="BeaconProxy",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xdf\x85\xc0\xf5\x0d\xfe\x1a\x79\xcc\xfe\x8e\xc6\x72\xd4\xeb\x35\xbf\x71\xac\x64\x8e\x6d\x75\x41\x7a\xa7\xe7\x07\xb6\x8d\x9e\xce": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xdf\x88\xce\x89\x3e\x06\x9e\xa8\x51\x56\x25\xc1\x04\xf9\xf3\xee\x29\x06\x9f\x69\xf7\x76\xba\xcc\x44\x32\xd2\xa5\xf5\x72\x5f\x5b": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.3.2, v4.3.3, v4.4.0
    b"\xdf\xb2\xc8\x5a\x4c\x46\x65\x28\xc3\xab\x9e\x6d\x6d\x70\xa7\x50\x2c\x21\x41\xb6\x32\x4c\xb6\x27\x62\x2f\xbd\xfb\xf0\x74\xde\x7a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.223
    b"\xdf\xc2\x6f\x64\xcc\x63\xbd\x90\xfa\x20\x92\x26\x9b\x33\x50\x88\x0f\x19\xca\x3f\xb2\x68\x7f\x32\x26\xf6\xf6\x6e\xb9\x92\x2c\x33": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.223",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe0\x0e\x98\xa7\x34\x1e\x7a\xaa\x1c\x57\x69\x52\x0f\xbb\x09\x1a\x25\xc5\x98\xec\xc4\xf2\x27\xee\x14\xa3\x0b\xd3\x9f\x81\xcd\x13": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xe0\x16\x76\x1d\x89\xfb\x9e\x8a\x13\x4b\xe0\xa7\x0e\x41\x6a\xbb\x79\x56\xe4\xa2\x62\x35\x13\x08\x0e\x06\x2f\xfa\x77\x96\x40\xec": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe0\x45\x94\x79\xd4\xbd\x23\xcb\x7d\xc5\x38\xcf\x1b\x10\xaa\x6f\xd2\x2a\x37\xd1\x16\xd1\x32\xcc\x90\xb3\x38\xae\x6e\x0c\xe7\x54": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashLender",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashLender",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156.sol",
            name="IERC3156FlashLender",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC3156FlashLender.sol",
            name="IERC3156FlashLender",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe0\x46\x27\x7d\x97\x04\xa6\xae\xa5\xcb\x8d\xb4\x77\x1d\x14\x44\x5d\x93\x10\x48\xd4\x54\xc6\x02\xb4\x02\x1c\x24\x37\x79\x8b\x0a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xe0\x50\x3d\x84\x3c\xe1\x83\x1a\xd7\xb7\xac\x17\x6c\x1a\x6e\xe1\x2c\xf0\x26\x14\xa3\xff\x4c\x8a\x5c\x29\xae\xf9\x88\x58\xf1\x31": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155URIStorageUpgradeable.sol",
            name="ERC1155URIStorageUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xe0\x5c\x7a\x7d\xdb\x2f\xf4\x16\xcb\x0e\xb0\xc4\xab\xe0\xa8\xfd\x45\xed\xf5\xa8\x01\xa7\x22\x5f\x75\xce\x03\xb6\x32\xc1\x03\xdd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xe0\x65\xef\x1b\xf8\xe7\x4e\x39\x3a\xfe\x7c\x56\x46\xd1\xd3\x26\x1e\xde\xec\x25\x41\xfd\xff\x6a\x03\x81\x28\xc5\x87\x54\x9c\x38": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.157, v0.0.158, v0.0.159, v0.0.160, v0.0.161, v0.0.162, v0.0.163, v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218
    b"\xe0\x67\xcb\x9e\x4f\x36\xb6\x14\xd5\xee\xa8\xa0\x02\x45\x21\x03\xc0\x6c\xe5\xd7\x84\xe8\xbd\x51\xa2\x2f\x15\xad\x07\x2c\x52\xba": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.157",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.158",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.159",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.160",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.161",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.162",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.163",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.218",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe0\x7f\x12\xdf\xa0\x1d\xb3\x04\x50\xde\x97\x2f\x8b\x5c\x44\x74\xc6\xa5\x24\x41\x54\xe9\xd6\xc8\xd9\x41\xd8\xb6\x48\x22\xa8\x92": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Proxy.sol",
            name="Proxy",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.3.0
    b"\xe0\x94\x72\xb4\x73\x9e\x4e\x79\x71\x4b\x79\x62\x17\x03\xdd\xb3\x35\x59\xc9\xf4\x3e\x54\x08\xe8\x57\xee\x49\xcd\x7a\x12\x7d\x6c": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/base/Multicall.sol",
            name="Multicall",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2
    b"\xe0\x9f\xf8\x7f\xaf\x9c\xdf\x5c\x04\xe1\xcb\xde\x54\xe0\xca\x9c\x7d\x8c\x7d\x61\x68\x5b\xa8\xdc\xf9\x9d\x49\x70\xb3\xac\x0e\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/ERC1967/ERC1967Upgrade.sol",
            name="ERC1967Upgrade",
            version="v4.8.2",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xe0\xe6\x41\x83\x11\x89\x82\xf8\xee\xfc\x90\x79\x90\xc2\x03\x5f\x12\xe8\xb1\xe2\xf9\x24\x9b\x82\x06\x86\x5d\xf1\x07\x60\xb2\x6e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20PausableUpgradeable.sol",
            name="ERC20PausableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe0\xf0\x67\xfa\x3a\x73\xc0\xf1\x52\xe3\x63\x2a\xa0\x31\x21\x6b\xb6\xb8\xda\xac\x8a\xb2\xd1\xf7\x57\x3b\xb5\x0c\x06\xcd\x35\xaf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe0\xf4\x85\x4b\x32\x4c\x6d\xb4\xb7\x22\x24\x0e\x7a\x4a\xc9\x3c\x3c\x1a\xff\x5d\xc8\x64\x7a\xdd\xfd\x79\x93\xc2\x2b\xf1\xed\x15": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xe0\xfa\x6a\xb8\x87\x2e\x63\x5e\xd6\xe6\x32\xa7\xa6\x7d\xe9\x5c\xdf\x07\x7d\x5b\xde\x0d\x95\xc7\xc4\x27\xef\x21\x9f\x09\x24\x25": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitterUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.1, v4.4.2
    b"\xe1\x38\xd0\xf3\xb1\xa3\x20\xcc\x94\xd2\x96\xb8\x7f\xbe\x04\xc4\xf8\x1a\xee\x75\x9f\x51\xc9\x7e\xfb\x9e\xb3\xfc\x20\x2e\x4b\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xe1\x3e\x10\x4d\xf1\x55\x5c\xa3\x7b\x3b\x58\x61\x75\xcd\x0c\x71\xd7\x5e\xe7\x55\x5d\xab\x1a\x49\x51\x0b\x86\x61\x93\x82\x41\xf5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ForwarderUpgradeable.sol",
            name="ERC2771ForwarderUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe1\x41\xab\x95\x33\xea\x0b\xa2\xe0\xbc\x23\xbb\xfa\x30\x02\x96\x52\x71\x29\x85\xac\x3e\x4f\x4e\x04\x70\x5e\x85\x9c\x3d\x35\xc5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/IERC20Permit.sol",
            name="IERC20Permit",
            version="v5.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\xe1\x54\xae\x59\x80\xe5\x6c\xea\xb2\x35\x3e\x83\x69\xba\x23\x43\xdb\x4c\x92\x8d\x70\x50\x09\xc4\xd9\x11\x76\x58\xdf\xd7\xac\x00": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe1\x56\x35\x48\xaf\x2e\x5a\xae\x52\xc1\x27\x54\x2f\x53\x6f\x84\x14\xfc\x23\x34\x5d\x68\xbc\xd1\x9c\x9e\x6e\x61\x23\xa1\xc1\x0d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PullPayment.sol",
            name="PullPayment",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe1\x5a\xe1\x48\x4d\xe4\xaf\xb5\xff\xa3\x3e\xd6\x94\x40\xbf\x0a\xdb\x77\xfb\x8f\x70\x7a\xb6\xc9\x1b\xca\x4b\x02\xcb\x0e\x4f\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xe1\x5f\xd2\xc9\xee\x69\x7e\xe7\xad\x87\x8b\x13\x6e\x46\x54\x89\xca\xde\x74\x1c\xcb\xbf\xef\xdf\xd5\x8d\x9e\x49\x8f\xe2\xa3\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC165.sol",
            name="ERC165",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.119, v0.0.120, v0.0.121
    b"\xe1\x6c\xac\x29\x9b\xb3\x14\x1f\x3d\x45\x65\x96\x3b\x19\xa5\x1d\x4a\xf4\x8e\x46\xd0\x99\x7d\x1c\xa5\x80\x22\x6a\xbc\xd9\xc0\xad": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe1\x87\xf1\x3f\x7f\x62\xa7\xf7\x89\xb4\x2e\x4a\x0d\xb1\x2e\x94\x2e\xd3\x44\xd0\x56\x0b\x4c\xef\x82\x3b\x13\x64\xd9\xd9\x6d\xa2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/TimersUpgradeable.sol",
            name="TimersUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257
    b"\xe1\x97\xa1\x87\xac\x0c\x90\x64\x21\x5d\x05\x4d\x05\xe8\x3a\xa3\x58\xb5\x74\x44\x2f\xd5\xb2\x15\x4f\xfc\x04\x89\xda\x2f\x5e\x7d": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.257",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7
    b"\xe1\xc7\xa6\x42\xf6\xc5\xe9\x26\xb0\x03\xe2\x79\x72\x36\xb3\xde\xc2\xfd\x1a\x93\x99\x68\x1b\x8f\xed\x0c\xf2\x39\x9b\xf5\x54\x36": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v3.2.1-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xe2\x16\x6b\x22\xa2\x7a\x0c\x75\x0e\x65\x16\xd1\xa9\xa2\x8a\x87\xe4\x79\x0b\x1e\xcf\x4b\x94\xa2\x55\xce\xd5\x4c\x6f\x92\x57\x60": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xe2\x20\x85\x06\x89\xf7\x0b\x7b\x3d\x39\x3f\x07\xbf\x1b\x16\x62\xd7\x03\x48\xa1\xe2\x11\x4c\xaf\x65\x6c\x4f\x9c\xda\xef\xef\xd4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe2\x2e\x07\x5f\x79\x14\x6f\x73\xf7\xc7\x71\x7f\xb3\x59\xc7\x2a\xd5\x1f\x88\x78\x8b\xaa\x83\xbe\xd9\x15\x77\x80\x39\x47\x93\x12": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/IERC777.sol",
            name="IERC777",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC777.sol",
            name="IERC777",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe2\x39\xe2\x7e\x93\x3d\x54\x3a\xf2\x62\xe4\xed\xfc\xb1\x9b\x25\xef\x4e\xc4\x90\x15\x10\x9a\xa0\xc4\x37\x07\xa2\xa8\x6c\x37\x04": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/MinimalForwarder.sol",
            name="MinimalForwarder",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.3.0-solc-0.7
    b"\xe2\x40\xe8\xcd\x0b\x17\x1a\x8d\x46\xe4\xc4\x5c\x4d\xb5\x74\x0e\x52\x1e\x94\x39\x5d\xf7\xac\xcc\x26\xbb\xb6\x46\xeb\xa5\x4a\x0c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20.sol",
            name="ERC20",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xe2\x5a\xc1\x55\x08\xc7\xb3\xbd\x8b\xae\xe1\xee\xd7\x78\x18\x50\x73\x91\x0b\xd8\x4b\xda\x09\x4e\xf5\x6c\xfd\xd8\xd0\xaa\x48\xb6": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/HexStrings.sol",
            name="HexStrings",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/HexStrings.sol",
            name="HexStrings",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/HexStrings.sol",
            name="HexStrings",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/HexStrings.sol",
            name="HexStrings",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/HexStrings.sol",
            name="HexStrings",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.278, v0.0.279, v0.0.280, v0.0.281
    b"\xe2\x69\x79\x1d\xdf\x21\x42\xe7\x5b\xf7\x25\xe7\x8f\x2d\x37\x07\x44\x39\x5b\x9d\x31\xa6\x52\x2d\xc2\xd4\x26\x8c\x29\xb1\x59\xeb": [
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/LibERC7579.sol",
            name="LibERC7579",
            version="v0.0.281",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xe2\x71\x48\x69\x26\x7b\x29\x62\x59\x5a\xe4\x63\x5f\x79\x2f\x13\xb9\x60\x6e\xe9\x5b\x57\x6f\x4a\xbc\x4d\x78\x2f\x78\x10\xbf\xc1": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/types/Slot0.sol",
            name="Slot0Library",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe2\x73\xcb\xdf\x24\x68\xa3\xa3\x80\xd7\x87\x6c\x3f\xc4\xc9\x77\x6a\xed\xcd\x04\xf1\x74\x3c\x62\xae\x52\x41\x9e\x81\xf6\x27\x08": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0
    b"\xe2\x75\x5e\x95\x1e\x52\xa7\x65\xad\x19\x5b\x8c\xec\x56\xf2\xbd\x36\x2c\x19\xf1\x44\xd5\x4b\xd8\x51\xcd\x26\xf2\x30\x90\x44\x1b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.2.0",
        ),
    ],
    # Hash appears in versions: v4.4.0-rc.0, v4.4.0-rc.0
    b"\xe2\xa9\x60\x33\xab\xf7\x6e\x9e\x04\xee\x3b\xb9\x23\x97\x6d\xd0\xdc\x8e\x4b\x5e\xc8\x73\x31\x99\x1e\x6d\xa1\x25\x08\xbf\xb2\xce": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/GovernorCompatibilityBravo.sol",
            name="GovernorCompatibilityBravo",
            version="v4.4.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe2\xb4\x08\xfb\xfa\x8f\xdb\x28\xdc\x2f\x30\x50\x26\x22\x1a\x12\x53\x33\x44\x20\xa9\xd6\xaa\x99\x98\x04\x4b\x97\x5e\x11\x2c\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ShortStrings.sol",
            name="ShortStrings",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2
    b"\xe2\xb5\xee\xfb\x28\xa0\xb0\xa2\x3a\x6d\x77\xaf\x53\xda\xc1\x42\x27\xb3\x38\x8a\xb2\x14\xc0\xc5\x99\x37\x50\x4e\xd1\xb2\x09\xd6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe2\xcf\xf9\x77\x81\xc6\x58\x1f\x03\x4e\x52\xdc\xe2\xd5\xc0\x6a\xa9\xef\x96\xcc\x71\x3d\x08\xc2\x72\x9f\x4a\x8c\x07\x6e\x6c\xc4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.3.0, v6.4.0, v6.5.0
    b"\xe2\xcf\xfa\xed\x35\x14\x5a\x0e\xb9\x08\x4f\x39\x85\xb6\xfc\x6e\x0e\x42\xa4\xa4\x0b\x81\xd8\xf1\x92\x17\x18\x12\x88\x08\x15\xbd": [
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.3.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.4.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.5.0",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe2\xd6\x8a\x51\x5f\xfe\x6f\x6f\x5d\xf4\x32\xe2\x1d\x48\xda\x58\x93\xb5\x0c\x76\x51\xb0\x4d\xa7\x77\x6d\xcf\x36\xcf\xbc\xed\xc7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/compatibility/IGovernorCompatibilityBravo.sol",
            name="IGovernorCompatibilityBravo",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe2\xe4\x02\x89\x5b\x46\xe5\xaa\xa9\x7d\x2f\xa5\x26\xd9\x02\xb1\x4f\xd0\xa7\x0f\xc1\x4e\xf9\xb9\x5a\xb8\x37\x1a\xf5\xbd\x27\xb9": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/IAccessControlDefaultAdminRulesUpgradeable.sol",
            name="IAccessControlDefaultAdminRulesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xe2\xea\x98\xf8\xda\x76\x5f\xd2\xd5\xd8\xf2\x33\x44\xe2\x48\x3e\xeb\xb2\xed\x59\xa4\xe8\x2f\xa0\x21\x7b\x95\x65\xd0\x61\xc3\xe6": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibBytes.sol",
            name="LibBytes",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xe3\x08\x9f\x64\xbe\x24\x44\xc3\x52\xbf\x3a\x29\x60\x73\x7c\x31\x3d\x3e\x95\xbc\x3f\x29\xe0\xfc\xbc\x86\x96\x68\x5a\xa9\xd2\xa1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/TokenTimelock.sol",
            name="TokenTimelock",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xe3\x2a\x91\x01\xf5\xa0\x58\x5f\x90\x5e\x15\x42\xb6\x29\xb8\x8b\x87\xb4\x6b\xba\x63\x94\x76\x7c\x1f\xf3\xa9\x70\x00\x5c\x7a\x8b": [
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe3\x5e\x56\xdc\x65\x59\xfb\x08\x47\x77\x38\xb3\x24\xa9\xf5\x56\x4c\x46\x31\x48\xf2\x51\xc0\xf3\x3e\x58\x9d\xf1\x1e\x27\xcd\x3e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockControllerWith46MigrationUpgradeable.sol",
            name="TimelockControllerWith46MigrationUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.69, v0.0.70, v0.0.71
    b"\xe3\x68\x85\x4c\xcb\x35\x23\x25\xcf\x95\x47\x90\x02\x17\x2c\x85\x17\x85\x6e\x27\x6c\xde\xc1\x25\xd8\x69\x7e\x6b\xd1\xf8\x93\x56": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.71",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xe3\x7e\x65\x32\xb6\x2e\x63\xa3\x95\x71\xd2\xaf\x2a\x28\xab\x9d\x85\xf8\x01\x02\xee\xa7\x77\x58\xc1\xdf\x20\xad\xdf\xbc\x16\x0c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165StorageUpgradeable.sol",
            name="ERC165StorageUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xe3\xee\xc3\xfa\xfd\x8f\x0d\x48\x20\x8f\x75\xc7\xc3\x1c\x25\x2e\x3c\x58\xc6\x0b\x98\x89\x2d\x94\xec\xb6\xf0\xb5\x92\x4a\x14\x21": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe4\x15\x5e\xa7\x89\x2a\xf7\x42\x1c\xd5\x02\xb5\x66\xad\xc6\x7c\x93\x75\x09\x53\xe7\x53\x10\x18\xc1\xf1\xc5\x4d\x19\x35\xfe\x81": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\xe4\x5c\x21\xe0\xab\xd3\x76\x7f\x97\xf1\x9d\x30\xb0\x81\x04\x61\x27\x53\xea\x37\x24\xc4\x6f\xa3\x5c\x7d\x75\xcb\xda\x40\x12\xf3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompound.sol",
            name="GovernorTimelockCompound",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe4\x82\x4b\xac\xce\xb8\xf9\x1e\xa8\xff\x7c\x31\x5b\xc5\xd8\x95\xd9\x87\xf1\x6e\xe6\x2f\x75\xd4\x85\xb8\x6f\x69\xd3\x18\x03\xda": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2
    b"\xe4\x8f\xe3\x41\xdc\xab\xcf\xe9\x6d\xc4\xc3\x80\x31\x25\x4e\xee\xab\x17\x12\xd5\xb0\xd3\xf0\xb8\x01\xf7\x72\x3d\x3e\xd1\xd3\x16": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/TokenVesting.sol",
            name="TokenVesting",
            version="v3.0.2",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe4\x92\xc5\x68\xcb\x5e\xc3\x9d\xfd\x11\xbf\xba\x2e\xe4\x40\x55\x70\x11\x5e\x29\xf4\x42\x20\xaa\x07\xb8\x9d\xca\x7f\x53\xcc\xbf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.192
    b"\xe4\xa2\xb7\x19\x27\x43\xff\x21\xbf\x11\x3b\xf4\x1d\x99\xf0\x73\xd0\x2b\xa6\x0f\xb4\x15\x36\x0c\xe8\xed\x6d\x86\x91\xb5\x1e\x39": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.192",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\xe5\x10\x0e\xdf\xd5\xc6\x5a\xcb\x86\xcb\x3b\x37\x68\xb9\xdf\x87\x2f\x49\xcf\x5a\xb7\xb8\xd7\xea\x14\x20\xf5\x9e\x31\xba\xb4\xcd": [
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/Initializable.sol",
            name="Initializable",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v0.0.196
    b"\xe5\x1c\x90\x7e\x72\xfb\x5a\xfc\xa2\xd4\xc4\x32\xc6\x4c\x6d\x93\xac\xe4\xc8\xfd\x1c\x7e\x2a\x06\xc4\x9a\xb1\x65\x3b\xd7\x77\x08": [
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v0.0.196",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xe5\x28\x24\xa1\xd2\x54\xe2\xa8\xad\xff\x54\x2a\x24\x67\x45\xca\x16\x01\xc8\x9d\x36\x3d\x86\x86\xbb\xe3\xdc\x13\xf2\x57\x42\xc1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/polygon/CrossChainEnabledPolygonChild.sol",
            name="CrossChainEnabledPolygonChild",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0-beta.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xe5\x28\x29\x5b\xf5\xd6\x4f\xe9\xde\x3b\xdf\x2d\x28\xe4\x77\xa5\x9c\x4f\x55\xf7\xf8\x7a\xcf\x0c\x6f\x20\x21\x2d\xc8\x82\xa5\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/Clones.sol",
            name="Clones",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xe5\x3d\x2c\xc6\x68\xb9\x0e\x1c\x9a\x06\x11\x1c\x55\xdb\x12\x11\xae\x9b\x3b\x69\x66\x1c\x00\x82\xa9\xd3\xcc\x02\x80\x36\x92\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xe5\x40\xe4\x34\x37\xd4\x00\x74\x3e\x00\x92\x60\x92\x2d\x42\xce\xa8\xaa\x97\x83\x3e\x54\x8f\x6f\x02\x67\x58\x77\x68\xbe\x9e\xbf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe5\x41\x99\xbe\x57\x0c\xb7\x01\x35\x9a\x59\xde\x8c\xac\x8d\xf9\xed\x91\x17\x94\x4a\x4f\x8d\xb7\x2a\xcd\xd1\x30\xc2\x18\x35\x09": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/manager/AccessManagedUpgradeable.sol",
            name="AccessManagedUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xe5\x62\x47\xdd\x4b\xc3\xa0\xe3\x14\x69\x98\xff\x83\x2f\xa6\xdd\x6c\x3d\x9b\x1b\xb4\x83\x08\x97\xd5\xb9\xc7\x9a\xff\x43\xc1\xc2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Holder.sol",
            name="ERC721HolderUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1
    b"\xe5\x67\x0d\xff\x0e\x47\x0d\xb5\x67\xa4\x17\x59\xc6\x71\xc8\x7c\xec\x8d\x95\x5f\xfd\x33\xbf\xbf\x8f\x5e\xe6\x0a\x6d\x54\xc2\x19": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.8.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xe5\xbf\x99\xc1\x43\x02\x37\x59\x71\x14\x78\x7c\xcb\x4c\xc2\xa4\xd6\xb7\x8a\x59\xba\x97\xe0\x78\x36\xdb\x6a\xea\xd2\x30\x08\xf4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0-rc.0, v4.8.0-rc.1
    b"\xe5\xc7\x4e\x7c\xde\x65\xf0\xea\x1a\x38\x63\xba\xae\x7a\x95\xed\x45\x4f\x2b\x79\x06\x12\xec\xff\x2b\x2c\xc4\xf1\x44\x99\x7e\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/arbitrum/IArbSys.sol",
            name="IArbSys",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe5\xcb\xb7\x7b\xb3\x0a\x64\x47\x2f\x36\xfc\xe5\x29\x10\x66\x0d\x88\x21\xe0\x18\xa5\x64\xa4\x7e\xad\x4b\x2d\x1c\xbc\x62\x1c\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.3.0
    b"\xe5\xd0\xb1\x83\x3f\x3e\x02\xb0\x8b\x24\x81\x18\x8a\xf3\xe5\x5e\x10\xea\x3f\x21\x84\x5b\xad\xae\xb5\x1e\x13\xf2\xe5\x06\xb4\x36": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrowUpgradeable.sol",
            name="RefundEscrowUpgradeable",
            version="v3.3.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xe5\xd8\x29\x41\xe9\x15\x51\xbe\x43\xde\xbe\xa5\xba\x2a\x2c\x0f\x90\x4f\x50\xb0\x6f\x64\x35\x9d\xd7\xc1\x7b\xab\xcb\x0b\x12\x37": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xe6\x0f\xba\xfd\x29\x51\xe4\x26\x9f\xb0\x49\xff\xd5\x83\x69\x6f\xd3\x01\xc5\xa3\xc2\xc7\x53\x1a\xb1\xd7\x1e\x63\x17\x18\x94\x94": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xe6\x1e\x9d\x54\xdb\xb0\x60\xcc\x86\x6c\xe2\x4f\xeb\x18\x5e\x5d\xee\xa4\xe4\x5c\x5f\x90\xef\xa2\xb9\x6b\xcf\x50\xf6\xda\x15\x43": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrow",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe6\xaa\xe5\xab\xa8\x81\xa7\xa4\xea\x00\x02\x83\x02\x21\x32\x67\x26\x76\x4b\x16\xf1\x30\x66\xc4\xaa\x55\x55\xbe\x4b\x5f\xef\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/arbitrum/LibArbitrumL1.sol",
            name="LibArbitrumL1",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.172
    b"\xe6\xad\x75\x18\xca\xc9\xee\x37\xe0\x8f\x70\x77\x3b\xc3\x53\xdb\x9a\x95\x20\x52\x2f\x65\x11\xdb\x1c\x86\x93\x50\xae\xd4\x40\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.172",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218
    b"\xe6\xf6\x0a\x07\x37\xc5\xe0\x0d\xe0\xea\x29\x6b\xc4\x02\xb5\x90\xb8\x48\xfb\xc3\x41\x46\xf9\xba\x3e\x2e\xe7\xc6\x21\x1a\xad\x1f": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.218",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\xe7\x2d\x86\x46\x80\xf5\xc2\x59\x47\x6c\x67\x90\x42\x6e\x1e\xae\x77\x0a\x6f\x52\xd0\x46\xc0\x61\x18\xb8\xe5\x43\x6b\x7b\x64\x95": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/presets/ERC1155PresetMinterPauserUpgradeable.sol",
            name="ERC1155PresetMinterPauserUpgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: main
    b"\xe7\x31\xd4\xe1\xcb\xfc\xbe\x83\xed\xb8\x00\xe2\x85\x5d\x5e\x39\xd1\xed\x4c\x5c\x9b\x78\x3c\x07\x04\xa7\x47\xc2\x7a\xa8\x2e\x73": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/lens/V4Quoter.sol",
            name="V4Quoter",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.9.4
    b"\xe7\x3c\xd8\xcc\x02\xa6\xee\xa1\x77\xb1\x51\xf9\x8f\xc7\x08\xea\x6e\x07\x91\xa2\x52\xe8\x4b\x44\x59\x69\xe2\x2d\x76\x10\x7f\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/MulticallUpgradeable.sol",
            name="MulticallUpgradeable",
            version="v4.9.4",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\xe7\x79\x78\x59\x94\x6c\x39\xb9\xd2\x72\x6b\x51\xac\x66\x80\x9e\x4c\x69\x48\x55\xc7\x07\x70\x11\x69\x3b\x4d\x04\x16\xce\x8a\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/deploy-ready/ERC721MinterPauser.sol",
            name="ERC721MinterPauser",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/deploy-ready/ERC721MinterPauser.sol",
            name="ERC721MinterPauser",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.202
    b"\xe7\x81\x79\x72\x1d\x0b\x3b\x62\x0b\x0e\xf6\x5c\xdd\xdb\xa5\x8b\xd0\xc2\x90\x1d\x50\x03\x72\x1b\xe9\x41\x28\xa2\x4e\xd9\xb7\x1a": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.202",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe7\xc4\xb5\x86\x8b\xe8\xee\x46\x90\xad\xa8\xd0\xc4\x38\xe2\xbb\xc6\x0f\x21\x53\x4a\x86\xfb\x94\x70\xc2\xdf\xab\xe2\xb5\xaf\x87": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xe7\xce\x90\x8e\xcd\x6f\xd2\x8b\x59\x3f\x5a\x6d\x6c\xf5\x2b\x10\x12\xe1\x2d\xa3\x9e\x27\x61\x49\x98\xec\x8c\xa9\x02\xc2\xb5\x05": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/RefundEscrow.sol",
            name="RefundEscrowUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe7\xd1\xba\x28\x3e\x54\xa3\x17\x32\xba\x60\xa2\x0c\x40\xc5\xbc\xc3\x1f\xce\x55\xe7\x8a\x91\x98\x97\x47\xcf\x17\x77\xe4\xde\x9e": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.45
    b"\xe7\xd7\xbc\x6e\xa1\xf7\x16\xa9\x4b\x5b\xe7\xd2\xf2\x5f\x52\xf4\x08\xf8\xa8\xef\xf4\x50\xe7\xc5\x7b\x2d\xc3\x94\xd9\x59\x8c\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.45",
        ),
    ],
    # Hash appears in versions: v0.1.5
    b"\xe8\x1b\xda\x08\x09\x4a\x96\xa2\xc5\xc8\x1e\x60\xa6\xf3\x5f\x0d\x61\x62\x9a\x06\x81\x1a\xf3\x89\x24\x66\x45\xd9\x83\xa7\xf0\x8b": [
        KnownContract(
            project="solady",
            path="src/utils/CallContextChecker.sol",
            name="CallContextChecker",
            version="v0.1.5",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.0.0-beta.1
    b"\xe8\x24\x22\x68\x24\x02\x87\xe3\x38\xba\x81\x0c\xa6\x90\x46\x6c\x3c\x15\x33\x88\x5d\xa5\xa8\xcc\xca\xff\xb6\x05\x18\x23\xba\x3c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.0.286, v0.0.287
    b"\xe8\x2d\x53\x19\x6c\xe0\xe2\x1b\x39\xf4\x9b\xee\x24\xbc\x5f\xd6\xc1\xd3\x4f\xca\xb7\x43\x36\x70\x0c\xdc\x4c\x01\x59\xbc\x66\x2b": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.287",
        ),
    ],
    # Hash appears in versions: v0.0.53, v0.0.54, v0.0.55
    b"\xe8\x56\xcb\x20\xd8\x3d\x8b\xf7\xb9\x55\xff\x8c\x34\x1b\x56\x22\x6e\x9c\x94\x21\x53\x02\x7a\xe0\x81\xd9\xb9\x33\x39\xf8\x9a\x5c": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xe8\x74\xae\xdf\x31\x93\x3d\x5d\x02\x00\xbe\x4c\xf3\x1e\x5e\x59\x80\xaa\xe3\x5c\xa5\x43\x9a\xd8\x8e\x69\xd9\xf5\x5b\x1a\x6d\x40": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/IERC777Upgradeable.sol",
            name="IERC777Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe8\x8e\xaf\xbd\x70\x38\x23\xb6\x9c\xeb\x08\xbd\x65\xe9\x24\xd5\xea\x33\xfb\x23\xf2\xde\xb8\x3d\x29\x40\x99\x52\x5c\xdf\x50\xa3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xe8\x8f\xa7\xa5\x29\x67\xd4\xd9\x37\xaa\x97\x66\x70\x02\x89\xe9\xf0\xd0\x75\xa9\x2d\xf6\xf8\x9f\x93\xbd\x68\x60\x0b\x22\x94\xc9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Burnable.sol",
            name="ERC20Burnable",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xe8\x96\xa6\xfc\x59\x93\x02\xc1\xbc\x8f\x00\x85\x74\x73\x8f\xeb\x84\x00\x37\x28\xa0\xf5\x07\x76\x8f\x24\x68\xc7\x04\x09\xbf\xbe": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibRLP.sol",
            name="LibRLP",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe8\xbf\x36\x37\x57\xfc\x91\x1d\x1f\x3e\x1f\x03\x15\xa2\x0d\x57\xd5\x52\x7f\x24\x29\x11\xe5\x29\x72\xfe\x61\x05\x74\x9b\x97\x11": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Upgradeable.sol",
            name="ERC20Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.247, v0.0.248
    b"\xe8\xc9\x04\x6c\xee\x6f\x42\xf6\x1c\x46\xff\x2c\xc9\xf2\xd9\xac\x5b\x49\x5d\x63\x96\x0b\x37\x59\x2e\x44\x19\xc4\xde\xff\xeb\x5f": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.248",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe8\xf4\xaf\xad\xfd\x70\xe0\x7f\x18\xcd\xd5\x5a\xfe\xb6\xed\x6e\x31\x4d\x7c\x8c\x44\x0d\x37\x80\xda\xe8\xc1\xc5\xd7\x31\xa7\xdc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4906.sol",
            name="IERC4906",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xe9\x0f\x5d\xb9\xd0\x4d\xf9\x47\x4a\xf9\xa8\xc2\x3e\x5d\x43\xef\xd6\x31\x6f\x76\x21\x6e\x92\x1e\xbf\x0d\x4c\x14\x60\xc6\x89\xcc": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/v2-core/contracts/interfaces/IUniswapV2Pair.sol",
            name="IUniswapV2Pa ir",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/v2-core/contracts/interfaces/IUniswapV2Pair.sol",
            name="IUniswapV2Pa ir",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/v2-core/contracts/interfaces/IUniswapV2Pair.sol",
            name="IUniswapV2Pa ir",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/v2-core/contracts/interfaces/IUniswapV2Pair.sol",
            name="IUniswapV2Pa ir",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/v2-core/contracts/interfaces/IUniswapV2Pair.sol",
            name="IUniswapV2Pa ir",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.7
    b"\xe9\x10\x67\xca\x35\xab\x8b\x32\x99\x9a\x91\x76\x97\x5c\x12\x8f\xdf\x07\xcc\x45\x5f\x54\x23\x8b\x6f\xaa\x23\x3a\x51\xba\xec\xfe": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.7",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xe9\x14\x32\xee\x15\x3c\x51\x5e\xd4\x19\x2c\x0d\xd6\x83\xba\xf7\x5f\xf3\x83\x2a\xee\x74\x55\xd4\x6e\x9e\xd6\xa4\x64\xa3\x6f\xc8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/CheckpointsUpgradeable.sol",
            name="CheckpointsUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xe9\x1f\x5d\x61\x19\xbd\x07\x0a\x2d\x19\x15\x51\xb9\x90\x07\xc8\xa7\xbc\x7e\x28\x5e\x3b\x7d\x8b\x53\x81\xdd\x01\xc8\x3a\x34\xca": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SignedMathUpgradeable.sol",
            name="SignedMathUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xe9\x23\x37\x22\x00\x9d\xb3\xd1\xca\x81\x24\x0f\x95\x77\x3a\x5c\x2f\x97\x35\xa3\x4e\x47\xd5\xaa\xd5\x63\x73\xda\x7f\x80\xd3\xb1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.8, v0.0.9
    b"\xe9\x2b\x89\x71\xf5\xcc\x19\xde\x08\x5e\xa3\xe5\xd4\x39\x3a\x77\xee\x9b\x47\x60\x55\x09\xc7\x20\x92\x48\xd0\xe5\x70\x84\xb5\xa1": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xe9\x2f\xe1\xb9\x92\xac\xdb\x35\xf3\xe2\x3a\xaa\xb8\xf7\x21\x74\x2e\x92\xbe\x4b\x69\x5b\x2f\x6c\x42\x6d\x5f\xf0\x14\x5a\x86\x45": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWalletUpgradeable.sol",
            name="VestingWalletUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xe9\x3b\x51\x60\x4a\x93\xb9\x66\x8e\xa0\x2d\xbf\x8b\xfa\x65\x98\x94\x5b\x64\x4b\x0a\xd5\xe7\x68\xc6\x50\xa1\x7b\xdf\xe5\x9a\x74": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SignedMath.sol",
            name="SignedMath",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xe9\x62\xe1\xcf\x4e\x00\xf0\x61\x45\xe5\x0e\x2c\x08\xf1\xba\x4b\xfd\xa2\x60\x97\xf3\xb7\xb8\x2f\xe4\x3c\x44\xec\x90\xa0\x1a\xc6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/security/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2, v4.5.0-rc.0
    b"\xe9\x9c\x63\xbd\x2f\xdd\x6a\xa1\xd2\x2f\x08\x9e\x4b\x24\x29\xee\xa2\xa2\x3f\x39\xc8\xdc\x5a\xbb\x1a\xcd\x8b\xf5\x0a\xcf\xdc\x77": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721PausableUpgradeable.sol",
            name="ERC721PausableUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xe9\xa0\x46\xa1\x57\x93\x36\xd2\x5d\xa6\x38\x45\x82\xa4\xec\xbd\xcd\xe1\xe2\x03\x8f\xcc\x47\x2b\x13\x76\xfd\x5c\x07\x6a\x39\x97": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xe9\xae\x77\x51\xa1\x57\x25\x38\x89\x68\x89\x73\xf0\xc2\x03\x1c\x84\x78\xf1\x9d\xaf\xd8\xb9\x8c\xb8\x81\xaa\x88\xde\xe2\x98\x1c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.3.0-rc.0
    b"\xe9\xeb\x42\x87\xe1\x77\x29\xac\xb6\xa4\xa0\x10\x76\xb6\x35\xd9\x1e\xd8\xdd\x2a\x5a\x35\x55\x71\x85\x5c\x0c\xb8\xaf\xf4\x83\x96": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xe9\xfa\x66\x0a\x08\xd3\xb5\xb2\xf8\xea\xee\x74\x2c\xd5\x07\x71\x9a\xed\x96\x84\x20\xa2\xfc\xe1\xa7\x31\x8f\x6f\x17\x6e\x75\x0f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Pausable.sol",
            name="Pausable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xea\x38\x62\x58\xd3\x13\xed\xbf\x75\xe8\xc9\xe1\xe3\x27\x62\x5d\x65\x53\xf3\x9f\x5e\x76\xae\x96\x51\xbb\x67\x6a\xb2\x9a\x36\x37": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721Enumerable.sol",
            name="IERC721Enumerable",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xea\x73\xcb\xf8\xfe\xd6\xc6\x06\x30\x2b\xae\x49\x0b\x4a\x16\x5f\x0f\x66\x71\x35\xeb\xd8\xd5\x95\x18\x40\x2a\x3a\xb1\xf3\x07\xba": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SingleUseETHVault.sol",
            name="SingleUseETHVault",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.140
    b"\xea\x7c\x90\xf7\x3f\x6a\xb6\xc3\x3f\x3d\x5a\x23\x3b\xb7\x26\x28\x38\x88\x9f\xe4\x6b\x03\xb8\xaf\xd2\x25\x44\x35\x0b\x6e\x81\x1d": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.140",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xea\x99\x07\x1a\x0e\x45\x50\x22\xbd\x5c\x22\xd0\xa8\xc7\xae\x9c\x27\xb9\xb4\xdd\x96\xa9\x8b\x63\xed\xcf\xde\x6e\xbc\x8f\x1a\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xea\xa2\xd2\x5b\x4c\x48\x67\x05\x51\x08\x42\xae\x9c\x0d\xf8\x75\x8f\x1c\xcc\x11\xfc\x67\x75\x10\x32\xd9\x36\x3f\x16\x82\x48\xb5": [
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ERC1967Factory.sol",
            name="ERC1967Factory",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xea\xa4\xa0\x16\xd3\x15\x32\x7f\x1f\x3a\x48\x95\x01\x47\x2e\x0a\xc5\x0b\xc0\x94\xd6\xf1\x6c\x9a\xe8\x19\x2c\xc5\x77\x62\x13\x3a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/EnumerableMapUpgradeable.sol",
            name="EnumerableMapUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xea\xab\xba\x6d\x13\x48\xd3\x63\x76\xe5\xb4\x9c\x98\x2e\xdb\xf2\x50\x3f\xb5\x54\x89\xfa\x87\x2b\x47\x92\xd4\xad\xca\x6f\xa7\xb4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/metatx/ERC2771Forwarder.sol",
            name="ERC2771Forwarder",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xeb\x0a\xc4\x90\x96\x63\x58\xff\x02\x13\xfb\x32\x2c\x26\x01\x0d\x14\x38\x67\x16\x72\x23\xdf\xb8\x3f\x5f\x1b\x66\x53\xef\x22\x65": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20Upgradeable.sol",
            name="SafeERC20Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xeb\x19\x4e\x49\x01\x03\x05\x08\xe3\xbc\xcf\x6b\x62\x39\x42\xa7\x99\x59\x76\xf3\x8d\xe1\x3d\x9f\xb1\xa5\xf5\x0b\x40\xda\x9c\x3f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauserUpgradeable.sol",
            name="ERC20PresetMinterPauserUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xeb\x1e\xbd\x84\x4b\x1b\x82\x31\xf6\x95\x46\x26\x26\x29\xd0\x91\x28\x1e\x8d\xeb\x68\x2b\xf7\x53\xc8\xc1\x0f\xef\x6e\x8c\xa4\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v4.3.3, v4.4.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.4.0-rc.1
    b"\xeb\x35\x1c\xb5\x12\x2c\xc6\x2b\x6c\xdc\xbb\x4e\xb2\x08\x31\x8e\x1d\xe1\xb5\x88\x89\xce\x54\x42\xf4\x60\x60\xfc\xc8\xf2\xe7\xa8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xeb\x48\x87\x3d\x35\xf4\x64\x87\xa9\x16\xf3\x96\x2f\x55\x12\x2c\xdf\xe0\x3a\x9a\x77\x69\xea\x7c\x88\x4b\xcc\x96\xef\x90\xf3\x0a": [
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/UUPSUpgradeable.sol",
            name="UUPSUpgradeable",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xeb\x58\xac\x34\xd4\xdb\x39\xd2\xa1\x69\x57\xee\x41\x8c\xf7\x8c\x46\x10\xf6\x97\xfc\xa8\x4e\xd1\xd2\xae\x5e\x06\x9d\x13\x74\xe9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721URIStorage.sol",
            name="ERC721URIStorage",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\xeb\x66\x7f\xac\x4a\x89\xb9\xbe\x0d\x6a\xa2\xf3\xc5\xbd\xb8\xd8\x8a\x9c\xd4\xa2\xe8\x87\x31\xa9\xca\x85\xae\x96\xcb\x3e\x35\xcf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/draft-ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Votes.sol",
            name="ERC721Votes",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xeb\x6c\xa7\x09\xf2\x2c\xd6\xe2\x86\x4b\x74\x40\x25\x61\x47\x3e\xfb\xb3\xf3\x48\xb8\x18\xd2\xd4\xe2\xcb\xd3\x32\xd1\xce\xe6\x75": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungibleTokenPositionDescriptor.sol",
            name="NonfungibleTokenPosi tionDescriptor",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207
    b"\xeb\x8e\x20\xfa\x1f\xcf\x3c\x37\x62\x4d\x93\x3c\x73\x41\xbe\x05\x7d\x42\x2d\x8c\xbe\xca\xeb\x87\x4c\xef\xa4\x22\x12\xd4\x97\xbe": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.207",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xeb\x9a\x6e\x15\xaf\x84\x27\x12\x2b\x56\x7f\xe3\xc1\xfe\x3b\x65\x47\x3b\x63\x78\xe9\xff\xd0\xa4\x1c\xb1\x79\x42\xe7\x44\xbf\x42": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientUpgradeable.sol",
            name="GSNRecipientUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7
    b"\xeb\x9a\xbe\x8c\x12\x7a\x32\xd8\x37\xdc\x3a\x1d\x2d\x45\xa3\xac\x9c\xee\x13\xf8\x07\x94\xa2\x8b\x80\xc1\xe4\x1e\x61\x1c\xe2\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.56, v0.0.57, v0.0.58, v0.0.59, v0.0.60, v0.0.61, v0.0.62, v0.0.63, v0.0.64, v0.0.65, v0.0.66, v0.0.67
    b"\xeb\xa2\xc1\xaf\x4f\x57\x97\xe4\x5e\x2d\xbe\x0e\x59\x2d\x1d\x21\x5b\x27\x4f\xf7\xc6\x59\x92\xfb\xde\x48\xb0\xcd\xce\x19\x74\xbb": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.56",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.57",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.58",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.59",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.60",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.61",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.62",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.63",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.67",
        ),
    ],
    # Hash appears in versions: v0.0.236
    b"\xeb\xa3\xdc\x10\x6d\xf7\x30\x53\xf6\x7c\x2e\x79\x6b\x42\xdb\xd6\xe0\x1e\xc9\xea\x1c\xa5\xb8\xef\xfe\x7e\xdf\xdc\x0c\xa1\x9a\xf2": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.236",
        ),
    ],
    # Hash appears in versions: main
    b"\xeb\xb7\xde\x53\xbb\x24\xdc\x66\xc4\xa7\x98\xcf\x2d\xb1\x07\x90\x59\xab\x85\x54\xb5\xe7\x70\x2b\x2f\x81\xc8\x15\x28\x87\xe1\x89": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/PositionConfigId.sol",
            name="PositionConfigIdLibrary",
            version="main",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xeb\xd0\xe8\x1c\x39\x1f\xef\xb8\x42\x4e\xda\xeb\xcb\xc8\x37\xd8\x80\xad\x3f\xd4\xc7\xc2\x5a\x1e\x3d\x3a\xcb\x71\xa8\xc7\x44\x7f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xeb\xdb\x5d\x64\xd6\xaf\xc6\x1e\x0a\x70\x86\x42\x65\xa5\x43\x28\x8a\xa8\xf0\xf5\xee\xf9\xc2\xd8\x0f\xcb\x74\x72\xef\xe7\xd1\x51": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.1.13, v0.1.14, v0.1.15, v0.1.16
    b"\xec\x0b\xbf\xc4\xf5\x09\x8e\x0e\x04\x7b\xee\x1a\x6c\xd8\xfa\x1e\xe3\xc9\xcb\xb5\x4b\x6f\x3f\x30\xc7\x70\x64\xbd\xf2\x33\x9e\xb0": [
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/EIP7702Proxy.sol",
            name="EIP7702Proxy",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xec\x11\x7c\x11\xb5\x76\x6a\x74\x35\x21\xf4\x49\xe8\x0e\x13\xb9\x03\xd7\xf2\x20\xdc\x05\x7d\x7c\x7d\x2d\x64\x47\x42\x7e\x29\x21": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableMap.sol",
            name="EnumerableMap",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xec\x20\x7b\xfb\x49\x15\x71\xe3\x5f\xb0\x70\x33\xed\x5d\x17\xf5\x4c\x59\xb8\xc9\xc0\xa5\x4f\x27\x32\x10\x98\xcf\x59\xc2\xbf\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0-rc.0
    b"\xec\x2b\x0c\xb6\xb3\xc8\x17\xc9\xff\xe7\xe8\xfd\x93\x23\x2c\xd5\xfb\x3a\xee\x82\x0b\xbd\xa1\x2b\xce\x6a\x1e\x1a\x78\x3a\xfb\x7b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ReentrancyGuardTransientUpgradeable.sol",
            name="ReentrancyGuardTransientUpgradeable",
            version="v5.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xec\x3a\x8a\xa9\x14\xcf\xf2\xaa\xde\xc0\x1c\x1a\x71\x8f\x60\xc2\xdb\x32\xf4\x99\x57\x7f\x08\xb4\x08\xe3\xce\x1f\x09\x49\x2d\x19": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/extensions/AccessControlDefaultAdminRulesUpgradeable.sol",
            name="AccessControlDefaultAdminRulesUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xec\x48\xb9\xa2\x3a\x7b\x22\x49\xc6\xce\xf5\x5a\xf2\x02\x27\x93\xea\xdf\x56\xe7\x17\x65\xb3\x24\x07\xa5\x12\x9b\x49\x0c\x74\x60": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/MerkleProofUpgradeable.sol",
            name="MerkleProofUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109
    b"\xec\x51\xd0\x9e\xab\x41\x46\x87\xdf\x19\x52\x2a\x6d\x6d\xd3\x79\x6c\xe6\xa3\x38\xf6\x39\x21\x9b\x01\x81\x4c\xd2\xc6\xd8\x2c\x3a": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.109",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0
    b"\xec\x6d\xba\x77\xd6\x2a\x50\x44\x96\x55\x2a\x6a\xbe\x42\xfa\xb8\x7b\xe1\x31\x62\xce\xd3\xbc\xa3\xf7\x16\xa7\x9f\xa7\xc2\xfc\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xec\x8d\x0c\xec\xf3\xf0\xcf\x80\xac\x1d\xee\xee\xe4\xce\x64\x42\xf8\x90\x7a\xbb\xf1\x52\xf9\xc7\x2a\xd8\xbe\xe2\xaf\x30\x5b\x38": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/ERC1820ImplementerUpgradeable.sol",
            name="ERC1820ImplementerUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: main
    b"\xec\x96\x73\x50\x7f\xda\x34\x33\x6b\x78\x3e\x21\x4b\xec\xbb\x08\x0b\xa2\x27\xf5\x6a\x53\xf4\x5b\x24\x6c\x6a\x53\x7f\xce\x8c\x4e": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/libraries/CalldataDecoder.sol",
            name="CalldataDecoder",
            version="main",
        ),
    ],
    # Hash appears in versions: v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xec\xb9\x13\x76\x0b\xfe\x3e\x5d\x0b\x37\xa5\x43\x49\xab\x37\x63\xb0\x88\xef\x0f\xbc\xba\x4e\xce\x04\xdd\x18\x84\xc5\xbb\xf5\xbc": [
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/RedBlackTreeLib.sol",
            name="RedBlackTreeLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.64, v0.0.65, v0.0.66, v0.0.67, v0.0.68
    b"\xec\xc7\xab\xa7\x15\x45\x64\x07\x22\xfc\x08\x8a\x6f\x8a\x17\x9d\x3a\x5c\xf2\x6e\x8d\xe5\x6a\x38\xc1\xa8\x29\x29\x25\x2c\x84\x3a": [
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.64",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.65",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.66",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.67",
        ),
        KnownContract(
            project="solady",
            path="src/auth/Ownable.sol",
            name="Ownable",
            version="v0.0.68",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xec\xea\x6a\xa9\x46\xeb\xad\x70\x67\x0e\x6d\xae\x04\xe9\xba\xbf\x80\x9b\x73\x98\x64\xc2\xe9\x68\x10\xce\x33\x1c\x74\x7d\xe0\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointNonces",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointNonces",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointNonces",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointNonces",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC4337.sol",
            name="IEntryPointNonces",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92
    b"\xec\xf4\x79\x9b\x79\x73\x4b\x04\x98\x27\x9d\xc4\x4b\x20\xae\x30\xf0\x93\xbc\x2b\x0c\xb3\x6c\x6f\x4a\x06\x74\x6b\x6f\x7d\xf3\x41": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.92",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xec\xfa\x45\x83\x2c\xd9\x82\x7e\xf8\x4d\x47\xc4\x1e\xb3\x98\x67\x46\x2d\x70\x03\x9a\xa2\x89\x27\x27\xa3\x80\xad\x59\xe7\x85\x95": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/SafeERC20Namer.sol",
            name="SafeERC20Namer",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/SafeERC20Namer.sol",
            name="SafeERC20Namer",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/SafeERC20Namer.sol",
            name="SafeERC20Namer",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/SafeERC20Namer.sol",
            name="SafeERC20Namer",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="@uniswap/lib/contracts/libraries/SafeERC20Namer.sol",
            name="SafeERC20Namer",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198
    b"\xed\x0e\xab\x62\x44\x36\xeb\x9a\x05\x6d\xcd\xa3\xb4\xfb\xde\x8d\x92\x10\x7b\x4e\x48\x8e\x31\x63\x38\x26\xd5\x76\x02\x12\x45\x3d": [
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibClone.sol",
            name="LibClone",
            version="v0.0.198",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xed\x1a\x03\x31\xde\xaa\x5e\xbd\x67\x1b\xb7\x5d\xa0\xa5\xfb\x8d\x8a\x5c\x5a\xf8\xd8\x5c\x69\x5a\x78\xa2\x90\x53\xae\xc2\xde\x01": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/arbitrum/LibArbitrumL1Upgradeable.sol",
            name="LibArbitrumL1Upgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xed\x28\xbf\xa8\xd9\x07\x67\x88\xc1\xeb\xbc\xc9\xa9\x0d\x37\xb1\xb9\xe2\x5d\x58\x36\xdf\xe5\x0f\x0e\xaf\xbd\x97\x64\x6c\x8f\x30": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeCast.sol",
            name="SafeCast",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xed\x36\x7e\xaf\xf9\x9c\x62\x54\xd0\x51\x7b\x13\x25\x38\x7e\xc0\xa2\x80\x45\x85\xe9\x21\x26\x4a\x88\x36\x5e\xe4\x32\x57\x1e\x5d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorVotesUpgradeable.sol",
            name="GovernorVotesUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xed\x49\x9f\x53\x35\x45\xb0\xea\x6a\x1d\xe5\x68\xe8\x7f\x6c\xab\xa3\xf1\xf1\xb6\xfa\x77\xb6\xfb\x33\x2c\xe4\xbf\x1f\x11\x62\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorVotesComp.sol",
            name="GovernorVotesComp",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.248
    b"\xed\x6b\x9c\x11\x83\xec\xb8\xfe\x2b\x54\x39\xc2\x01\x96\x3a\x90\xa6\x4d\x4b\x46\x62\x54\xdf\xca\x41\xbb\x86\x7e\x08\xcb\x4d\xf8": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.248",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xed\xa9\x6a\xf3\xf4\xe1\x7b\xe9\xa9\xc0\x6e\xf3\x98\x58\xe4\xac\xfc\xce\x75\x9a\x53\xee\xd8\xa2\x12\x72\x93\x4a\x65\x72\xbc\x03": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/libraries/UnsafeMath.sol",
            name="UnsafeMath",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xed\xa9\x94\xa9\x4b\xac\x1a\x0f\x05\xc1\x01\xaa\xc9\xa1\x12\xa5\x4d\xc8\x7e\x3e\xe8\xca\xc2\xdb\xe6\x91\x80\x86\xfe\x52\xfb\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/math/SafeMathUpgradeable.sol",
            name="SafeMathUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xed\xbb\x22\xfa\x9d\x3f\x0a\xd3\x2f\x68\x0e\xdf\x59\x6a\x9d\xc3\xdb\xe9\x91\x02\x90\xd6\xe7\x42\x01\xf2\x48\x91\xf1\x26\x38\x33": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/metatx/ERC2771ContextUpgradeable.sol",
            name="ERC2771ContextUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xed\xbc\xe2\xd6\xe7\x77\xe0\x5f\xd4\x5b\xba\x73\x1a\x33\x2e\xa5\x66\x88\x5b\x47\xe9\xa0\x52\x90\x88\x3b\xd6\x15\xce\xa9\xc6\x2c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xed\xc7\x53\xea\x7f\xb5\x83\xa9\x18\x94\xbf\xd6\x99\xaf\x83\x6e\xf5\xf9\x6a\x23\x35\x76\x05\xe4\x7a\x95\xbe\x29\x9c\x5b\xac\x7e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.211
    b"\xee\x06\x00\xc8\x28\xd4\xba\xfd\x8c\x1a\xfc\x66\x9b\xdc\x51\xd6\xdc\x67\x06\xd0\xe4\x7c\xa4\xe6\x54\x7a\x27\xfa\x77\x57\xb9\x09": [
        KnownContract(
            project="solady",
            path="src/utils/EfficientHashLib.sol",
            name="EfficientHashLib",
            version="v0.0.211",
        ),
    ],
    # Hash appears in versions: v6.0.0
    b"\xee\x16\xf1\x9f\xaa\xc0\x12\x62\xd5\x30\x07\xcc\xc8\x00\x6c\x28\x66\xbe\xff\x01\x37\xb3\x59\x05\x99\x06\xdb\x71\x08\x52\x6e\x4a": [
        KnownContract(
            project="solmate",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v6.0.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2
    b"\xee\x22\x82\x3d\x93\x1c\x93\xe7\x28\xf9\xa0\x9e\x44\xe4\x8c\xa1\xa5\xe8\x05\x2c\x4c\x1f\xb7\xf3\x96\x7e\xce\xfd\x20\xd8\x22\xb3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/IERC721EnumerableUpgradeable.sol",
            name="IERC721EnumerableUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v0.0.164, v0.0.165, v0.0.166, v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xee\x46\x3d\xb0\xc1\x8d\xa9\xc7\x23\xea\xe1\x92\xfe\xc8\x20\xb3\x37\xbe\xce\x99\xc5\x28\x20\x66\xad\x7f\x23\xb2\x2f\xe4\x33\x08": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.164",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.165",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.166",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-rc.0
    b"\xee\x95\xcd\x9a\x5c\xff\x17\x32\x3a\xae\x1f\xae\xd1\xad\xd8\x81\x59\x1f\x86\x52\x28\xc1\x59\x07\x34\x28\xb2\x73\x4e\x19\x6e\xf7": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xee\x9e\x53\x94\x91\xf5\xd5\xaa\xb5\x20\x95\x6d\xc6\xe9\x0c\x3b\x06\xc7\xf0\x4a\x48\x0f\xf7\x24\xbe\xc6\xe1\x78\x8e\xe6\xd9\x6d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesUpgradeable.sol",
            name="ERC20VotesUpgradeable",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xee\xb5\xa5\x24\x0f\x03\x83\x97\x58\xfa\x5f\xfd\xc5\x67\xdc\x18\x1b\xba\xb6\xf1\xd9\x4d\x81\xd1\xf8\x11\x8b\x5a\x6c\x35\xd5\x2b": [
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/DynamicArrayLib.sol",
            name="DynamicArrayLib",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xee\xba\x94\x24\x66\x7b\x99\x10\xb3\x81\xc7\x27\x69\x0e\x31\xe4\xc6\x88\x87\x98\x4d\xf9\xfe\xf4\xa7\xd7\x42\x5c\xff\x6b\x8b\x65": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Create2Upgradeable.sol",
            name="Create2Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xee\xd4\x4d\x6c\xa3\xf7\xed\x47\x23\x8b\xce\xb9\x76\xf4\x58\xd9\x5a\xec\xbe\xee\x27\x2c\x82\xdd\xd0\xbc\x30\xa7\xb1\xa4\x3f\x03": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xee\xeb\xd6\xb5\x50\xdc\xef\x40\x29\x07\x20\x70\x61\xdb\xd3\x2c\x07\x55\x07\xc5\xf5\x02\x65\xe0\xb1\x5c\xa4\x9d\x26\x24\x70\xdb": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xee\xf3\xb3\x3b\xd9\xde\xe4\xe5\xb5\x97\xd5\x2c\x2b\x03\xf8\xbc\xd0\x91\x76\xa2\xaf\x23\xed\xbd\xae\x66\xc4\x23\x75\x49\x79\x14": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.1, v3.0.0-rc.1
    b"\xee\xfe\x93\x61\x85\x62\x7f\xef\x96\x74\xd5\xf9\x8b\x0f\xdf\x8e\xee\x07\x1a\x01\x69\xe9\x76\xb0\x1b\x75\x67\x75\xd9\xca\xb2\xf8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="AccessControl",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0
    b"\xef\x05\xac\x84\x40\xce\x81\x43\xf8\xf0\x64\x38\xc9\xfd\x6e\x19\xd3\xa9\x19\xe1\x75\xfc\xdb\x32\x55\xbf\x71\xcc\xf3\x51\x8a\x8f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PullPayment.sol",
            name="PullPayment",
            version="v3.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2
    b"\xef\x23\x09\xcb\xa3\xd0\x21\xcc\xbf\x7d\xcd\xf1\x22\x24\xc3\x0f\x80\x26\x14\x32\xe0\xf5\x97\xc3\x5f\x88\x78\x2d\x24\x58\x3a\x86": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.2",
        ),
    ],
    # Hash appears in versions: v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275
    b"\xef\x2a\xc6\xfd\x81\x1e\xba\x1c\xe8\xc8\x69\xaa\x93\x0c\x59\xf1\xc7\x5c\x40\x3f\xae\xe6\xc2\x8a\x0b\x8c\xbe\xee\xae\xc5\x62\x12": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC4337.sol",
            name="ERC4337",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v0.0.188, v0.0.189
    b"\xef\x39\x2e\x0e\x0b\x22\x70\xcc\x35\x79\xff\x8f\x72\xe9\x9c\x7d\xf6\x7a\xcb\xd4\x3a\xed\x30\x75\x42\x6a\xcc\x9d\x81\x38\xee\x01": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC1271.sol",
            name="ERC1271",
            version="v0.0.189",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xef\x45\x96\xe8\x65\xa8\x5d\x13\xbe\x1c\xd2\x49\x19\x3d\xf9\x5b\x19\x76\x7b\x93\x82\x01\x61\xb8\x49\xaf\x02\xe3\x7b\xf5\xf2\x74": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/BeaconProxy.sol",
            name="BeaconProxy",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: main
    b"\xef\x5c\x47\x99\x69\xad\xde\xae\x49\x46\x41\x84\x61\x49\x1c\xe0\x94\x35\x9b\xb6\x71\xff\x24\xe2\x7e\x0c\x38\x8e\x19\xe1\x47\xed": [
        KnownContract(
            project="uniswap-v4-periphery",
            path="src/interfaces/IPositionDescriptor.sol",
            name="IPositionDescriptor",
            version="main",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xef\x7c\x03\xcd\xcc\xe7\xa8\xdf\x4e\x3f\x8c\x8e\xcb\x6e\xf4\x8b\xab\x62\x55\x3d\x1f\x5d\x9b\xd0\x21\x92\xf2\x6f\x60\x7f\xac\x24": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/utils/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xef\x81\x9a\xa4\xee\x9a\x85\x88\xae\x63\x17\x40\xa8\xaa\xa8\xfc\x15\xdf\xe6\x8b\x5d\xe5\xfb\xb3\xdd\x4d\xc2\x51\x34\x49\xa0\x14": [
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ECDSA.sol",
            name="ECDSA",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.4.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0
    b"\xef\x95\x00\x4e\x92\x6c\x57\x95\x76\xfd\x9d\xf6\x8a\x23\x01\x6a\x36\x09\xdf\xcd\x5d\x4e\xb5\xe4\x2f\xe5\x64\xf1\x7d\x16\x93\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.4.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.2.0
    b"\xef\x9a\x81\x3c\x5e\x87\xc4\x7f\x01\xe6\x41\x98\xb2\x60\x1b\x49\x63\xd0\x97\x25\xb7\x28\xf8\x8d\x0d\x75\xd7\x7c\x15\x5d\x0f\x46": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/WeightedOracleLibrary.sol",
            name="WeightedOracleLibrary",
            version="v1.2.0",
        ),
    ],
    # Hash appears in versions: v0.0.20, v0.0.21, v0.0.22, v0.0.23, v0.0.24, v0.0.25, v0.0.26, v0.0.27, v0.0.29, v0.0.30, v0.0.31, v0.0.32, v0.0.33, v0.0.34, v0.0.35, v0.0.36, v0.0.37, v0.0.38, v0.0.39, v0.0.40, v0.0.41, v0.0.42, v0.0.43, v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xef\xac\xd7\xf5\xa5\xcf\x35\x51\x66\x77\x4b\xd2\x98\xfb\x58\x41\x84\x96\xb9\x83\x21\xb5\x8d\xe3\xa1\x5b\xb1\xa4\x42\xaf\xf2\xf1": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.20",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.21",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.22",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.23",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.24",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.25",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.26",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.27",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.29",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.30",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.31",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.32",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.33",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.34",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.35",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.36",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.37",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.38",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.39",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.40",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.41",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.42",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.43",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xef\xb5\x22\x23\x64\x94\x1a\x8b\x41\x67\x0b\xf1\x7d\x2f\x59\xd1\x50\x04\x35\xca\xd2\xbe\x77\xbc\x18\x65\x7d\xb7\x82\x09\x93\x30": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/utils/Initializable.sol",
            name="Initializable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.196
    b"\xef\xdd\xd4\xa0\x99\x64\x05\xf5\x1d\x3b\x79\x10\xb2\xf0\x21\xf7\x6f\xa0\xa0\x11\xbe\x71\x13\x1f\xc3\x34\xd8\x95\x93\x6a\xdb\xf9": [
        KnownContract(
            project="solady",
            path="src/utils/UpgradeableBeacon.sol",
            name="IUpgradeableBeacon",
            version="v0.0.196",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf0\x04\x7f\x88\x23\xe7\xfa\x0d\x5d\x09\x2e\x49\x1f\x13\xf8\x14\x24\x67\xdb\xd5\xd1\x82\xb4\xf4\xce\x28\xe6\xfe\x50\x41\x48\x3d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20TemporaryApproval.sol",
            name="ERC20TemporaryApproval",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76
    b"\xf0\x0c\xa6\x83\xc7\x30\xa6\xb9\x05\xbd\xb8\x16\xcf\xb9\x82\x46\xf3\xd1\x65\x50\x9d\x57\x2a\x71\xcd\xc4\xc3\xa4\x4b\x45\xfa\x2b": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.76",
        ),
    ],
    # Hash appears in versions: v0.0.262, v0.0.263
    b"\xf0\x16\xa7\xf5\x55\x1e\x5b\x86\xca\x86\x8e\x83\x00\x94\x79\x99\x79\x8a\xb3\xb3\x4c\x88\x7a\xcf\x6e\x44\xf1\xff\xcd\x4f\x55\x47": [
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/g/LibString.sol",
            name="LibString",
            version="v0.0.263",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf0\x1b\x45\x56\xb6\x6a\x13\x58\x69\xf3\xa9\xfb\xf9\xeb\xb0\xcf\x30\xd1\xe5\xb0\x19\xa7\x4f\x4b\x69\x35\x29\xf4\xb8\x5c\x63\x73": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/EnumerableSet.sol",
            name="EnumerableSet",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xf0\x39\x02\x24\x98\x64\x30\xef\x27\x3c\xde\xa3\xf3\x05\xd9\x27\xe9\xc0\x6c\xbf\x9a\x3e\xf2\x86\x70\x33\x32\xc4\xf1\xdf\x68\xc4": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoIdUpgradeable.sol",
            name="ERC721PresetMinterPauserAutoIdUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\xf0\x46\xc7\x9b\x73\xc7\x93\x13\x3f\x3d\x48\xf9\x84\x05\xe5\xd3\x15\xa4\x27\x6f\xd5\xc0\x56\xe1\xb4\x62\xa8\x5d\xd9\xa3\xfa\xd1": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockCompoundUpgradeable.sol",
            name="GovernorTimelockCompoundUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xf0\x60\xbf\xcc\x74\x99\x51\x90\x04\x5c\xe2\xcf\xee\x4b\xc8\x49\x99\x9c\x1f\x6e\x95\x99\x07\x8c\xa1\x3f\x17\xc6\x55\x62\xf0\xf5": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/TickMath.sol",
            name="TickMath",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v0.0.296, v0.0.297, v0.0.298
    b"\xf0\x92\x3a\xf4\xd1\xb2\x36\x35\x7a\x19\x82\xdd\x38\x54\x89\x89\x2a\xea\x7f\x92\x9b\xbe\xd9\xf1\xc7\xa9\x93\xcc\x9d\xd1\xa2\x00": [
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibSort.sol",
            name="LibSort",
            version="v0.0.298",
        ),
    ],
    # Hash appears in versions: v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xf0\xa6\x10\xbd\x33\xc8\x10\x4b\x1d\x95\xe9\xcf\x64\xc4\xfe\xfa\x42\x74\xb4\x65\x3d\xbb\xb2\x78\x72\xf9\x86\xe6\xf8\xcc\xbf\x69": [
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ReentrancyGuard.sol",
            name="ReentrancyGuard",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xf0\xb1\x17\xd3\xf3\x0b\x1e\xf3\xc2\x2b\x03\x23\xc2\x90\x7e\x1d\xb9\xef\x9a\x36\x3a\xb8\xd5\x02\xc6\x91\x8e\x8e\xdc\x55\x52\xfd": [
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/WebAuthn.sol",
            name="WebAuthn",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xf0\xde\xa8\x58\xa1\xae\xb1\x6f\x2e\x17\x15\x8d\x35\x56\xbc\x2f\x86\xa9\xb3\x5e\xa5\x85\xd0\x46\xef\xf5\x0c\x57\x13\x92\xf9\x66": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Upgradeable.sol",
            name="ERC721Upgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xf0\xfe\xa5\xec\x9e\xd4\x68\xf8\xdd\x1f\x42\x08\xa2\xd8\xaa\x18\x09\x61\x68\xe0\x02\x44\xdd\x38\xeb\x12\xab\xa5\xf9\xd9\x92\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20FeeUpgradeable.sol",
            name="GSNRecipientERC20FeeUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.2
    b"\xf1\x1b\x94\x64\x03\xc3\xe4\x61\xf9\xe4\x37\xcc\xba\x50\xf7\xa1\xd5\x41\x1d\xe9\x26\x5a\xc5\x38\x8c\xf3\xf8\xff\xcd\xc7\xf1\xed": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v3.4.2",
        ),
    ],
    # Hash appears in versions: v4.8.2, v4.8.3
    b"\xf1\x28\x4c\x3e\x14\x27\x93\xf5\x8b\x53\x03\x22\x51\x74\x6a\x46\x22\x50\xc7\x1a\x06\x02\x4f\xc1\x21\xfc\x46\x5e\x85\xbe\x19\x28": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721.sol",
            name="ERC721",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xf1\x45\xf2\x5e\x06\x05\x3a\x55\xee\x02\xd9\x2a\x2a\x74\xd3\x97\x4d\x38\xbe\xdf\x80\x25\xdc\x58\x4f\xda\x99\x44\x3a\xfc\x24\x5d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/vendor/arbitrum/IOutboxUpgradeable.sol",
            name="IOutboxUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xf1\x4e\xc4\x4f\xe1\xab\xbe\x3a\x66\x5d\x6c\x41\xcb\xe3\x64\x2d\x7e\x74\xea\x9c\x06\xac\x26\xf5\x2c\x4e\x31\x9a\xe7\x3b\xdd\x32": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/escrow/Escrow.sol",
            name="Escrow",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v0.0.148, v0.0.149, v0.0.150
    b"\xf1\x5f\xa8\x4a\xe6\x04\x2b\x5b\xbb\xc5\x1d\x4c\xf6\x02\xfe\xe9\x79\x8b\x55\x7d\x4d\x34\x03\xd5\x35\x74\xe3\xdc\xd2\xee\xdf\x98": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.150",
        ),
    ],
    # Hash appears in versions: v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282
    b"\xf1\x61\x0a\x9e\xc7\xfe\x10\x50\x2b\xf8\xb8\x12\xdc\xc0\xf8\x45\x3d\x16\xa0\xbe\xa9\x4a\x48\x60\x9f\xa1\x1d\x61\xe0\x06\xe0\xe9": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.282",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf1\x9a\x09\x49\xc9\xff\x18\x06\x8a\x1b\xc9\xe4\x70\xd0\x24\x61\xf8\x72\x42\xd8\xf5\xb3\x18\x94\x0c\x79\x98\x2c\x53\x88\x6c\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Wrapper.sol",
            name="ERC20Wrapper",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.283, v0.0.284, v0.0.285
    b"\xf1\x9f\x0c\x8c\x13\x63\x10\x6d\x8f\x62\xe3\x5e\xb5\xb5\x76\xa5\x93\x58\x9d\x53\x65\x95\x23\xdf\x19\x7d\xb1\x19\xf0\xca\x5b\x85": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBytes.sol",
            name="LibBytes",
            version="v0.0.285",
        ),
    ],
    # Hash appears in versions: v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0
    b"\xf1\xa3\xe8\x17\x80\x20\x38\xfb\x53\xe5\x34\xeb\x70\x75\xb1\x4c\x6e\xf3\x15\x75\x4a\x4b\x4f\x56\xcf\x58\x82\x45\xda\x4a\x88\xd4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/MerkleProof.sol",
            name="MerkleProof",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xf1\xd3\x4b\x5b\x33\x0e\x93\x96\xe9\x6f\x82\x17\x4c\xed\x3e\x75\x18\xa5\x56\xcf\xd3\x6b\x33\xb4\x83\x99\xb9\x63\x55\x1f\x5b\xc3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20Permit.sol",
            name="ERC20Permit",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.15
    b"\xf1\xd6\xef\x34\x52\x4f\xec\x1b\xeb\xa2\xba\x7d\xed\xec\x1a\xfb\xf7\x45\x73\x23\xcb\x71\x04\xa9\xaf\x0f\x08\xa9\x9a\xf4\x04\x30": [
        KnownContract(
            project="solady",
            path="src/utils/LibBytemap.sol",
            name="LibBytemap",
            version="v0.0.15",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xf1\xe1\xf6\x08\x3a\x96\xc2\xd0\x15\xc7\x9b\xe8\x4c\x66\x6a\x67\x6d\xb1\xd6\xd7\xae\x00\xfe\x04\xc7\x14\x4a\x54\x11\xc7\x19\xa2": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf1\xf3\x5b\xa3\x15\x8b\xc0\x25\xc3\x07\x8d\x70\x81\x56\x7e\x0f\x11\x7a\xc6\x58\x34\x89\x15\xda\x9b\xeb\x1f\xf5\x30\x96\xe2\x51": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/extensions/ERC1155Supply.sol",
            name="ERC1155Supply",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xf2\x1b\xc3\x3a\xa2\x56\x72\x20\x99\x6c\x52\xd0\x64\x11\x24\x35\xf2\x53\x47\xd3\xff\xb8\x29\x00\x35\xda\x5c\xfa\xaf\x5f\x2a\x80": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolAddress.sol",
            name="PoolAddress",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolAddress.sol",
            name="PoolAddress",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolAddress.sol",
            name="PoolAddress",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolAddress.sol",
            name="PoolAddress",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/libraries/PoolAddress.sol",
            name="PoolAddress",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xf2\x41\x9e\x85\x12\x7a\xbf\x68\xe6\x7b\x4a\xb2\xc5\xd7\xb6\xbd\x88\x07\x4e\xe2\xe2\x89\x31\xe3\xf4\x99\x17\x04\xb2\xd8\x8c\x5d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-ERC20PermitUpgradeable.sol",
            name="ERC20PermitUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xf2\x42\x1d\xd7\x5d\x62\x57\xa2\xe7\x3c\x3d\xba\xe3\x26\x4f\xc9\xee\x33\x70\x3e\x9c\xda\xfb\x10\x81\x71\xb8\xbe\xa0\x80\x66\xe5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.4.0-rc.0, v4.4.0-rc.1, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf2\x4a\x7f\x37\x5c\x52\x2e\xab\x32\xe5\x97\xdb\x00\xf7\xc9\x50\x72\x09\xa2\xb3\x4f\x49\x85\xc0\x86\x3a\x9a\x5a\xa6\x9c\xf7\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/VestingWallet.sol",
            name="VestingWallet",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0-solc-0.7
    b"\xf2\x51\xb8\xfb\x01\xc2\x5a\xad\x09\xfb\xf1\x8e\x5a\x4f\xe8\x65\x1d\x6a\xab\x8b\x09\xc2\x8a\x9b\x65\x22\xbb\xe3\x66\x70\xe0\x9b": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xf2\x5e\x16\x1c\x89\x5f\x2a\x13\x8c\xe4\xd5\x69\xdb\x29\x72\x1d\xe2\x40\x61\x15\x4c\xd4\x13\x4f\x8e\xf2\x2c\xbd\x12\x89\x63\xe5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.1, v4.7.2, v4.7.3
    b"\xf2\x7f\x33\x02\x17\x3e\x03\xb1\x4b\xc5\xc6\xa9\x82\x21\x79\xf2\x27\x73\xff\x9c\x2a\x7d\x5a\x59\xe3\xb5\x05\x20\xe3\xf7\x98\xe6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.171
    b"\xf2\x80\xc9\xad\x6c\x64\xb3\x3c\x01\x49\x38\x4b\xc4\x6c\x31\xb7\x79\x8c\x28\x1c\xdd\xb9\xcd\xaa\x21\x72\xc1\x8e\x25\x1b\x33\xa6": [
        KnownContract(
            project="solady",
            path="src/utils/MinHeapLib.sol",
            name="MinHeapLib",
            version="v0.0.171",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xf2\xc0\x7b\xed\x65\x12\x57\x30\x2c\x1b\xa1\xa5\x99\x6a\x1c\x4e\x66\x84\x35\x9a\x85\x50\xc8\x69\x7d\x54\xfd\xbf\x6c\x2d\x12\x20": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/Extsload.sol",
            name="Extsload",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf2\xca\xe2\x6e\xb8\x3a\xd0\xc1\xd4\xb6\x02\xab\x73\xfa\x67\x80\x2e\xe1\x6b\x10\x2a\x25\x2f\x37\x9b\x1a\x2d\x4f\x19\x49\x60\x00": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721EnumerableUpgradeable.sol",
            name="ERC721EnumerableUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf2\xf9\xb7\xc1\xfa\x16\xf5\xb0\x4e\x4a\xc4\x2e\xd1\x6d\x58\xe8\xb9\x12\xe7\x3b\x12\x08\xa9\xc4\x37\x29\x54\x44\xc7\x9e\xa9\x13": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorStorageUpgradeable.sol",
            name="GovernorStorageUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf3\x07\xc7\x61\xc4\x29\x09\xb7\x13\xc3\x7a\x32\xdf\x6e\xad\xca\x18\x56\xe4\xa4\xb8\xb2\xdc\xf7\x7f\x8a\x65\x0a\xf2\x0f\x4f\xfc": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165Storage.sol",
            name="ERC165Storage",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xf3\x30\x2c\x62\xf2\x15\xab\xb3\x75\x1e\x4f\xd3\xcc\x66\xc3\x44\x3e\x55\xf6\x1f\x7e\x21\x1c\x68\x53\x6a\xdf\x38\x01\xdf\x01\xa0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820Registry.sol",
            name="IERC1820Registry",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xf3\x41\xc0\xe2\x09\x62\x69\x38\xfc\xd8\x4e\x5c\x91\x38\x01\x19\x19\xef\xd1\x46\xb3\xba\xbc\x32\x06\x86\xc1\xdd\x18\x1d\x24\x84": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xf3\x62\x2a\x49\x85\x0a\xe2\x27\xee\xc7\xe6\x24\xb5\xaa\xce\xd1\x13\xa9\xe2\x66\x47\x72\xef\xe9\x30\x50\xcb\x60\xc4\x59\x2f\xc2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControl.sol",
            name="IAccessControl",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf3\x67\xd1\x98\x8f\x2f\xb6\x46\xf8\x2a\xb3\x54\x6a\xf2\x45\xa7\xed\xa5\x28\xe0\x97\x1d\xb3\x8d\xb2\x20\xa9\x58\x9a\x4e\x44\x92": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/SafeERC20.sol",
            name="SafeERC20",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xf3\x72\x2c\x76\x4e\x48\x45\xc0\x1b\xf0\xa8\x86\x13\x85\xc0\xfe\xa5\x8b\x10\x22\x5e\xfc\x5e\x2f\x97\x0d\x3d\x76\x95\x92\xaa\x64": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.0.0
    b"\xf3\x76\x84\xd5\x34\x40\x4e\x02\xc7\xda\xfd\x38\x53\xbe\x14\xe2\x97\x44\x09\x31\x10\x37\x33\x0c\x13\x6b\xe6\xc8\x0d\x54\x37\x0f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20Pausable.sol",
            name="ERC20PausableUpgradeSafe",
            version="v3.0.0",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xf3\xb5\x93\x7c\x47\x21\x9d\x63\x39\x7c\x53\x7c\xca\xdf\xb4\x9d\xdf\xbe\x0b\xcd\x09\x0b\x54\x2c\x98\x50\xd8\x42\x64\x85\x50\xad": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlEnumerableUpgradeable.sol",
            name="AccessControlEnumerableUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-beta.0
    b"\xf3\xb9\xdc\x2f\xe9\x92\xb9\x81\x80\x34\xa7\xf1\xab\x2a\x62\x86\xc4\x3f\x4a\xee\x7d\x1d\x0d\x94\x74\xc1\x04\xed\x75\x22\x97\x28": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Metadata.sol",
            name="ERC721Metadata",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Metadata.sol",
            name="ERC721Metadata",
            version="v3.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xf3\xbd\xe8\x48\x9c\xe8\xc5\xbb\x23\x15\xe0\xb1\x75\x13\xe2\xf8\xd0\x79\x27\xf0\xcb\xf1\x23\xeb\x81\xe0\xb0\x50\x6f\x5e\x3a\xc8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/ERC20CappedUpgradeable.sol",
            name="ERC20CappedUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xf3\xcd\x8e\xc6\xbc\xf9\x6c\x7c\x64\xf6\x92\x30\x14\x18\x4a\x23\xea\x6b\xd6\xdd\xfa\x62\x39\x51\xce\x66\xdc\x5f\x49\xc7\x7c\x1d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC4626.sol",
            name="IERC4626",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.136
    b"\xf3\xde\x9d\x1d\x2e\x43\x2b\xc6\xf8\x55\x75\xe9\xe0\xf7\x37\x17\xca\xfe\x29\x1b\xf5\xaf\x61\xc0\xe7\xad\x25\x2d\xeb\x0c\xd4\xd9": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.136",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0
    b"\xf3\xe6\x34\xa8\xa3\x04\x0d\x45\xf8\xd9\x8c\xd3\x01\xd6\x5f\x8a\x36\xe9\x8a\x95\xf6\xc4\x8e\x83\xce\xad\xd0\x60\x3b\xd5\x18\x3a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.0.0-beta.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf3\xec\xde\xd2\x64\xf2\xa1\xfd\x03\x49\x18\x81\xc5\x4d\xf9\xff\xde\xa5\x4b\x40\xc7\xfd\x10\x3e\x4c\x7c\x93\x9a\x05\x08\xa1\xb6": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/extensions/AccessControlDefaultAdminRules.sol",
            name="AccessControlDefaultAdminRules",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-beta.1
    b"\xf3\xf5\xe5\x78\xe9\xaa\xe4\xa6\x3d\x1b\x96\xea\x92\x9d\xa8\xeb\x44\x2f\x32\xab\x09\x42\xb4\x80\x1f\xb3\x3a\x20\x12\x5a\x08\xdf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155.sol",
            name="ERC1155",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf4\x22\x4c\x54\x11\x1a\x13\xad\x5d\x72\x49\x9c\x77\x37\x5f\x1d\xa3\xf2\x67\x7c\xf3\xad\x14\x25\x21\xf4\x9c\xc4\x35\xb5\x8b\x45": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/IVotesUpgradeable.sol",
            name="IVotesUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf4\x2d\x85\x46\xf5\x77\x6d\xdd\x78\xc3\x0a\xb7\x48\xf0\xdc\xce\x47\xeb\x68\x3c\xfa\x09\xa9\xde\x45\xc4\x50\x52\xcd\x37\x47\x6e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xf4\x39\xd1\x37\x8e\x9a\x96\xdf\x2d\x2d\x73\xaf\xaf\x70\xc8\x2c\xef\x03\x66\x81\xa3\xdf\xc0\x65\x57\xb1\xec\xac\xd7\xb5\x8e\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.184
    b"\xf4\x5f\x84\x51\xac\x87\xe7\x5d\x68\x7f\x33\x0c\xea\x2e\x5e\x91\x98\x78\xae\x3a\xad\xfb\x60\x37\x79\x59\xa4\xde\xc8\xbd\x93\x2e": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.184",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xf4\x68\xe4\xc3\x78\xc3\x59\x28\x9a\xf5\xbe\x94\xba\x25\x3f\xca\x72\xcf\xa6\xa3\xff\x17\x21\x5f\xb1\x3b\xd3\xe4\x55\xc0\x25\x56": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC2309Upgradeable.sol",
            name="IERC2309Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201
    b"\xf4\x81\x35\x7b\x68\xe1\x63\x97\xab\x09\xf6\x4b\x1e\x46\xc5\x20\x71\x23\x34\x08\x0d\x5d\x28\xf8\xef\x5e\x97\xb7\x03\x17\xb9\x1f": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.201",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf4\x87\xc2\xa5\x1d\xa5\x7c\xb9\x19\x0f\x99\x47\xbf\x50\x94\x01\xf4\xbf\x02\xa9\xf3\x66\x6a\x5e\xa0\x13\x74\x07\xe6\xdf\x91\x25": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579AccountConfig",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579AccountConfig",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579AccountConfig",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579AccountConfig",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579AccountConfig",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xf4\xc4\xcb\xe2\xbe\x2a\xef\x6a\x07\xb1\xcb\xf5\x44\xf0\x65\x6d\xbd\x7d\xbb\x05\x9d\xaa\x3c\x83\xe9\xd5\x38\x91\x0e\xc7\x67\x15": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/utils/VotesUpgradeable.sol",
            name="VotesUpgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xf4\xfe\x5a\xe1\xa4\x45\x29\xd2\xed\xbd\x29\x6a\x65\x7b\xb8\xf3\xed\xb8\x6d\x4f\x7b\x44\x93\xa8\x3d\x4a\x98\xe1\x8b\x68\xbe\xe1": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xf5\x1f\xb6\xf3\x47\x90\xc3\xa5\xc5\xf5\x30\xd3\x81\xff\x00\x9d\x73\x88\xb6\xde\x99\x7e\x20\x8e\x6d\x82\x6f\xa5\x31\xd7\x0b\x8d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSuperQuorum.sol",
            name="GovernorSuperQuorum",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorSuperQuorum.sol",
            name="GovernorSuperQuorum",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3
    b"\xf5\x29\xc8\x3c\xbf\x07\x81\xc5\xe7\x5e\x28\xa2\x5e\x85\x49\x70\x97\x4d\xc7\xd8\xdd\xe8\x07\x71\x80\xac\x02\xc8\xbc\xeb\x94\xf0": [
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ext/zksync/ERC721.sol",
            name="ERC721",
            version="v0.1.3",
        ),
    ],
    # Hash appears in versions: v4.5.0-rc.0
    b"\xf5\x88\xd3\xeb\x62\x28\x3b\x01\x8a\xba\x4a\x89\xf6\x96\x82\x8a\x3c\xf2\xf3\x92\x7e\x56\xc3\x8e\xfe\x03\x75\xa0\x67\x17\x49\xd3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777.sol",
            name="ERC777",
            version="v4.5.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf5\x94\x4f\xc0\xb6\x8e\x5d\x50\x16\x52\x43\x4d\xb4\xc0\xd2\x21\x0f\x13\x99\xc8\x43\x4d\xed\x5e\xfc\x65\xfa\x2f\x3a\x03\xde\x6b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20FlashMint.sol",
            name="ERC20FlashMint",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.262
    b"\xf5\xa3\xb9\xee\x33\x0d\x7d\xd3\x9a\xbc\x3f\xd2\x72\x09\xbe\x77\x0b\x47\x7c\x28\xe5\x31\x88\xf0\x34\x83\x40\x5d\x8d\xf8\x09\xf8": [
        KnownContract(
            project="solady",
            path="src/utils/LibTransient.sol",
            name="LibTransient",
            version="v0.0.262",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xf5\xac\xfc\x42\xe3\x92\x1a\xcf\xf0\x4b\x54\x9b\x76\x45\xd2\xa9\x79\xf1\x31\x9c\x80\x51\x25\x65\xd9\x0b\xf6\xbf\xa4\xdc\x3c\xec": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeCastUpgradeable.sol",
            name="SafeCastUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xf5\xb4\x9c\xfd\x4c\xd7\xdc\x97\x29\x9f\x47\xd9\xf9\x98\x2b\x3d\x65\xdf\x34\xe4\x52\x24\x82\xe9\x37\x1c\x45\xb0\x87\x7d\x47\xc1": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/V3Migrator.sol",
            name="V3Migrator",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/V3Migrator.sol",
            name="V3Migrator",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/V3Migrator.sol",
            name="V3Migrator",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/V3Migrator.sol",
            name="V3Migrator",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/V3Migrator.sol",
            name="V3Migrator",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xf5\xd0\x9a\xc7\x5e\xc1\xad\x86\x34\x10\x28\xa4\x22\xbc\x00\x11\x24\x5c\xdc\x18\x14\xc4\x12\x61\xf9\xe8\x7d\x96\xf4\xb9\x92\xda": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/IGovernorUpgradeable.sol",
            name="IGovernorUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf5\xf2\x73\x83\x81\x25\x11\x6c\x1e\x18\xf0\x82\xe2\xf8\x30\xc2\x2f\x12\xe6\x80\xbd\x53\x72\xc4\x57\x63\xd7\x51\xc1\x1f\x11\x9c": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/manager/IAccessManaged.sol",
            name="IAccessManaged",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf5\xf2\xb2\x61\x74\xf4\x6a\xd1\x29\xe0\x97\xe2\x38\x66\x03\x2e\xc9\x02\x69\x42\x14\x6d\xe1\xb4\xcf\x66\xbf\x06\xf9\x75\x4a\x9d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/security/Pausable.sol",
            name="Pausable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1
    b"\xf6\x14\x66\x5c\xd1\x36\xb3\xc6\x47\xad\x04\xbf\x3c\x6a\x3d\xff\x75\x22\xa5\xe2\xd2\x88\x0f\xd1\xb2\x80\x07\x44\xdf\x6a\xb4\x8b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Strings.sol",
            name="Strings",
            version="v5.1.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xf6\x1d\x57\x8b\xb4\x8b\xd1\x24\xbe\xde\x6c\xdd\x32\xb2\xaa\xa4\x6f\x43\x79\x91\xb7\xc0\x8a\x18\xb2\x9c\x49\x32\x4f\xdc\x04\x5e": [
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/CREATE3.sol",
            name="CREATE3",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf6\x2c\x1d\x01\x1c\xbe\x17\xcf\x96\xfb\x36\x9e\xd1\xa5\x37\x29\xf5\x8b\x1f\xce\x19\x46\xe7\xf3\x58\xb8\xa6\x64\x8e\x79\xbd\xc2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579ModuleConfig",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579ModuleConfig",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579ModuleConfig",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579ModuleConfig",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC7579.sol",
            name="IERC7579ModuleConfig",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.4.0
    b"\xf6\x53\xa1\xf2\xaa\xac\x48\x6d\x47\x16\x9d\x76\x0a\x4b\x47\x81\x91\x1d\x69\x1f\xdc\xf9\xcc\x71\x2a\x2f\x34\xae\xac\x76\x70\xdd": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/finance/PaymentSplitterUpgradeable.sol",
            name="PaymentSplitterUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xf6\x8b\x2b\xbd\xea\x3b\xef\xa9\x9b\x75\xf9\x7a\xb1\x85\xf4\x1e\x38\xea\xfc\x26\x84\xe6\x5a\xa5\xdc\x75\x6a\x13\x1f\x92\xb6\xcb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v0.0.17
    b"\xf6\x9c\x92\x5d\xdd\xcb\xc8\xf9\xc7\x8d\xcb\x94\xe9\xef\xc5\x6f\x19\x25\x17\xb1\xf7\x79\xb2\x86\xe1\x2e\x59\xb3\xbf\xbd\x04\xc4": [
        KnownContract(
            project="solady",
            path="src/auth/OwnableRoles.sol",
            name="OwnableRoles",
            version="v0.0.17",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf6\x9c\xee\xee\xd4\x9e\xcd\x73\x04\x0a\x00\x9a\xac\x66\x1e\x97\xd1\xf2\xf8\x47\x97\x02\x6c\x74\xb0\x3f\xe8\xf0\x1d\xdd\x48\x9b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Timers.sol",
            name="Timers",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.9
    b"\xf6\xbb\x73\x8f\x08\x75\x3a\x6d\x11\x7a\xf4\xe0\xe4\xd0\xf1\x77\x56\x89\xa7\x6f\x01\xd8\x9f\xac\x08\x5e\xea\x21\x2a\x90\x75\x90": [
        KnownContract(
            project="solady",
            path="src/utils/MerkleProofLib.sol",
            name="MerkleProofLib",
            version="v0.0.9",
        ),
    ],
    # Hash appears in versions: v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292
    b"\xf6\xc7\x34\x65\x3b\xa9\xdc\x63\xb3\x9e\x22\x81\xca\x0a\x58\x90\xce\xa2\x2a\x9b\xc5\xda\xdf\x7c\x55\xa7\x05\x0e\xc1\x10\x84\x5b": [
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/Timelock.sol",
            name="Timelock",
            version="v0.0.292",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf7\x04\xcf\xc9\xf4\xbb\xac\x6f\x54\x79\x9e\x19\x6c\x7d\x14\x23\x4d\xea\x98\xa9\xc1\xd5\xdf\x49\xc7\x68\x10\x8b\x4f\xbd\xec\x8f": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/utils/ERC721Holder.sol",
            name="ERC721Holder",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0
    b"\xf7\x13\x2e\x9c\xcc\xc0\xbd\x23\xe8\x75\xff\xac\xf7\xb4\x62\xc3\xeb\xcf\x9d\x83\x1a\xa5\x6a\x88\x68\x34\xb5\x96\xc0\x98\x74\x64": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/utils/TokenTimelockUpgradeable.sol",
            name="TokenTimelockUpgradeable",
            version="v4.4.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0
    b"\xf7\x20\x64\xc7\x3e\xc7\x75\xa0\xdf\xbf\x50\xad\xc8\x28\x91\xc3\x41\x99\x8b\xf7\xbe\x17\x81\x27\x71\x4b\x4f\xe3\xb0\x82\xd6\x07": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Upgradeable.sol",
            name="ERC1155Upgradeable",
            version="v4.6.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xf7\x2a\x0c\xa6\x7c\x2b\x74\x61\xf5\xac\x87\xb7\x47\xc6\x67\xba\x08\x75\xca\xa3\x6b\xe2\x28\x1a\xae\x80\xb3\xfb\x2f\x22\x77\xe5": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/AddressUpgradeable.sol",
            name="AddressUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf7\x31\x16\x7c\xad\x7b\xd2\xc1\x5a\xb1\x16\xce\x85\x9a\x61\x39\x10\x3a\xfc\x8a\x9d\x65\x19\x08\x6c\x55\x63\xa6\xc3\xd7\x9b\x24": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/math/SafeMath.sol",
            name="SafeMath",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf7\x4a\xb7\x9e\x49\x7d\xeb\xb7\x61\xae\xd8\x89\xaa\x65\x70\xa7\x75\x83\xb9\x7d\x17\xb1\x6c\x59\x78\xf8\xb3\x4c\x07\x4f\x98\x21": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20WrapperUpgradeable.sol",
            name="ERC20WrapperUpgradeable",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253
    b"\xf7\x4a\xb9\xe4\x9a\x88\x13\x90\xe7\x17\x97\xef\xfc\x86\xdd\xb3\x94\x5e\x5f\x84\x67\x43\x0f\x7c\xee\xfa\xd3\xee\x14\x65\xf4\x0f": [
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/EnumerableSetLib.sol",
            name="EnumerableSetLib",
            version="v0.0.253",
        ),
    ],
    # Hash appears in versions: v4.8.3
    b"\xf7\x4d\x90\xad\x19\x58\x86\xbc\x5c\x8c\x1b\x9d\x7d\x3b\x20\xb3\x7e\x08\x9a\x4f\xe3\x7d\x56\xf9\x0b\x2d\x44\xd5\xe3\x40\x4c\xe8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/ERC1967/ERC1967UpgradeUpgradeable.sol",
            name="ERC1967UpgradeUpgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2
    b"\xf7\x76\x33\xdc\x70\x67\x87\x4b\xef\xc5\xe9\x47\x95\xf7\x47\x1b\xae\x76\x7d\x60\x8c\x87\x83\xa6\x60\xfc\x4d\x63\x60\xa5\x1e\x71": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.5.2",
        ),
    ],
    # Hash appears in versions: v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-solc-0.7, v3.0.0, v3.0.0-rc.1, v3.0.1, v3.0.2, v3.1.0, v3.1.0-rc.0, v3.1.0-solc-0.7, v3.2.0-rc.0, v3.2.1-solc-0.7, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xf7\x83\x09\x38\xed\x81\x8e\xe5\xc6\x07\x37\x46\x95\x97\x83\x7d\xd7\x21\x58\xf5\xbb\x90\x05\xef\xc9\x50\x9f\x4a\x63\xc8\xad\xaf": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/EnumerableMap.sol",
            name="EnumerableMap",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf7\x8b\x16\xdd\x1e\x88\x20\x2d\xdc\x78\xa5\x29\x14\x4a\x46\xea\xd8\x2f\x0a\xc6\x68\x6a\x00\x77\x60\xb5\x71\xc1\xc1\x3b\x91\xd0": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorProposalThresholdUpgradeable.sol",
            name="GovernorProposalThresholdUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.272
    b"\xf7\x9a\x31\x3c\x59\x2c\x27\x34\x52\x3b\x9e\x4e\x76\xe3\xc8\xb6\xe1\xd4\x39\xc8\x73\x39\xc0\xce\x97\xee\x12\x5c\xbe\x68\x92\xc0": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC7821.sol",
            name="ERC7821",
            version="v0.0.272",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xf7\xa1\x25\xef\xf8\x1d\x93\x25\xdd\x43\x75\x59\x5d\x86\x33\x41\xe8\x72\xbe\x44\x08\x3d\x51\x95\x75\x8b\x3d\x65\x25\xe8\x80\x3b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/payment/PaymentSplitter.sol",
            name="PaymentSplitter",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1
    b"\xf7\xb0\x4a\x38\x6e\x40\xf3\xb4\xca\x2c\xbe\x6b\x66\x89\x31\xc6\x25\xe2\x81\xef\x8f\x3c\x8d\xf6\xe6\x23\x9b\xf2\x2f\x59\x59\xe3": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/PausableUpgradeable.sol",
            name="PausableUpgradeable",
            version="v5.2.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.1.14, v0.1.15, v0.1.16
    b"\xf7\xd9\xea\xaa\xbf\xeb\x74\xb8\x4e\x30\xf0\xa3\xfb\x5c\x47\x44\xf8\x37\xfc\xe2\xf8\x9f\xf6\xfc\xe7\x2b\x80\xf1\xa8\xb9\xc9\x71": [
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/P256.sol",
            name="P256",
            version="v0.1.16",
        ),
    ],
    # Hash appears in versions: v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf8\x02\x17\xe6\xb0\x42\x8b\x8a\xaf\x9f\x34\x39\x70\xaa\x1a\x75\x32\xd5\x19\x12\x7e\x36\x4a\xfc\x43\xdc\x8c\x09\x14\x2b\x2f\xc7": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/Ownable.sol",
            name="Ownable",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7
    b"\xf8\x0d\xbf\xf1\x07\x26\x9a\xb6\x29\x0a\xf5\x63\xad\xd2\xde\x84\x3b\xe6\x4f\x2c\x0c\x04\xc0\xfd\xf1\x15\x0b\xfe\x05\x45\x1d\x13": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf8\x24\x05\xa0\x8c\xaf\x17\xcf\x66\x90\xcb\xe7\x7e\x19\xbc\x2b\x9a\xdb\x27\xe5\xff\xa0\xd5\xd3\xb7\x0f\xa0\x44\x5c\xc1\x78\x68": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/crosschain/amb/CrossChainEnabledAMBUpgradeable.sol",
            name="CrossChainEnabledAMBUpgradeable",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0
    b"\xf8\x43\xbd\x51\x54\x6a\x61\xf1\x3c\xbc\x13\x70\x05\xc7\x39\xb0\x91\x79\xea\xde\xa4\xb4\x68\xe6\xe0\x0b\x1e\x29\xe1\x28\xf9\x78": [
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/access/AccessControlEnumerable.sol",
            name="AccessControlEnumerable",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf8\x46\x80\x47\xe1\x64\x40\x10\x1e\x40\x3a\x64\x3b\x29\x02\x5e\xef\x1f\xbb\x3b\xb1\xb8\x7b\xfa\x3f\x40\x92\x89\x5a\x33\x1b\xde": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xf8\x4f\x86\x43\x9a\x52\xf0\xc1\x83\x32\x01\xce\x34\xe3\x5d\xf2\xca\xc8\x40\x76\x2b\x5d\xa2\xce\x1a\x7b\x1c\x16\xbc\xd3\x76\x37": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungiblePositionManager.sol",
            name="NonfungiblePositionManager",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungiblePositionManager.sol",
            name="NonfungiblePositionManager",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungiblePositionManager.sol",
            name="NonfungiblePositionManager",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungiblePositionManager.sol",
            name="NonfungiblePositionManager",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/NonfungiblePositionManager.sol",
            name="NonfungiblePositionManager",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v0.0.127, v0.0.129, v0.0.130
    b"\xf8\x55\xb4\xc8\x84\x76\xa9\x11\x8d\x26\xde\x15\x05\xb1\x35\xbd\xa3\x47\x43\x6e\x65\x49\xa8\xcf\x05\x84\x02\x20\x16\xe3\xc3\x67": [
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibString.sol",
            name="LibString",
            version="v0.0.130",
        ),
    ],
    # Hash appears in versions: v3.0.0-rc.0, v3.0.0-rc.1, v3.0.0-rc.0, v3.0.0-rc.1
    b"\xf8\x59\x06\xb3\x1e\x7a\x6c\x66\x2d\x24\x31\x4e\x2f\x14\x1c\x42\x82\xb6\x21\x2b\x1b\x50\x13\xd8\x73\x64\x3c\x9d\x6d\x31\x2a\x7d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Address.sol",
            name="Address",
            version="v3.0.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xf8\xc0\x3d\x19\xbc\x68\xb5\x72\xf2\x86\x77\x4c\x96\x07\x6b\xa1\x15\x7f\x8e\x77\xde\x1b\x07\x73\x6a\xbc\x12\x4c\x96\xd5\xd5\xa8": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721URIStorageUpgradeable.sol",
            name="ERC721URIStorageUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.167, v0.0.168, v0.0.169, v0.0.170, v0.0.171, v0.0.172, v0.0.173, v0.0.174, v0.0.175, v0.0.176, v0.0.177, v0.0.178, v0.0.179, v0.0.180, v0.0.181, v0.0.182, v0.0.183, v0.0.184, v0.0.185, v0.0.186, v0.0.187, v0.0.188, v0.0.189, v0.0.190, v0.0.191, v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288
    b"\xf8\xd1\xe7\xd2\xc7\xcf\xbc\x2d\x16\x5b\x7a\x02\x2f\x27\xe2\x19\x01\x37\x9d\xcf\x00\x86\xc3\x30\xb8\xeb\x73\x17\x35\x25\x55\x19": [
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.167",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.168",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.169",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.170",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.171",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.172",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.173",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.174",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.175",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.176",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.177",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.178",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.179",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.180",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.181",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.182",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.183",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.184",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.185",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.186",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.187",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.188",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.189",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.190",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.191",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibZip.sol",
            name="LibZip",
            version="v0.0.288",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf8\xee\x23\x5e\x0f\x43\x97\xb9\xa4\x2f\xd0\xac\x3c\xbe\xf4\xa2\x52\xd0\xd3\x96\x32\x53\xc8\xd9\x4d\xd3\xf9\x41\xa5\x8d\xfb\xa4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/polygon/IFxMessageProcessor.sol",
            name="IFxMessageProcessor",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xf9\x05\xb1\x34\xd6\x89\x38\xc5\x5d\x4b\xfe\xab\x0b\x02\xaa\x05\x96\xd0\x65\xcd\xb3\xac\xd4\x2f\x81\x85\x81\x62\x4e\x07\x13\xa8": [
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/proxy/beacon/UpgradeableBeacon.sol",
            name="UpgradeableBeacon",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v0.0.49
    b"\xf9\x0f\xb5\x74\x0e\xe1\x94\xc3\x92\xc0\x9c\x33\x1d\xf0\x86\x69\xc9\x17\xf9\x89\x2f\x23\xd3\xd0\xe6\x07\x89\xfc\xd1\xa1\xc6\x15": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.49",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xf9\x13\xcf\xdf\xfb\x35\xfb\x87\x2e\x4c\x50\x67\x67\xd8\x28\xb3\x88\x8f\x5d\x09\xc4\x55\x61\x79\x7a\x80\x47\xf6\x3f\xb2\xcf\xa6": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1820RegistryUpgradeable.sol",
            name="IERC1820RegistryUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xf9\x2d\x85\x42\xae\x1f\x5c\xa4\x81\x36\xb7\x15\xbc\xc5\x4a\x61\x93\x88\x05\xcf\x70\x78\x2f\xf5\x0d\x8b\xd5\x53\x70\x08\x41\xc5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165Checker.sol",
            name="ERC165Checker",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.258
    b"\xf9\x48\x1d\xf4\xa6\xe3\xba\xf0\xd7\xc0\xe2\x73\x20\x8a\xbf\x6a\x6f\x41\xaa\x53\x3b\x14\x41\xc5\x37\x54\x24\x9c\xad\x8b\x08\x68": [
        KnownContract(
            project="solady",
            path="src/utils/SafeTransferLib.sol",
            name="SafeTransferLib",
            version="v0.0.258",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xf9\x52\xbd\xa3\xc6\x39\x58\x52\x39\xa1\x78\xb4\x49\x37\x14\xdf\xe7\x78\x18\x7c\x87\x65\xeb\x18\x13\x49\x70\xef\x94\xc9\x8e\x3a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20VotesCompUpgradeable.sol",
            name="ERC20VotesCompUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.0-solc-0.7, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2, v3.4.2-solc-0.7, v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v3.4.0-rc.0, v3.4.1, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v4.0.0-beta.1
    b"\xf9\x7e\x9d\x3e\xf7\xf1\x1b\xe0\x93\x5e\x4d\xd4\xe3\x05\xb5\x20\x6f\xa1\x20\xb5\xbc\xf7\x2d\xd1\x8a\xd0\x77\x57\xe6\xff\xb6\xd2": [
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/drafts/IERC20Permit.sol",
            name="IERC20Permit",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/draft-IERC20Permit.sol",
            name="IERC20Permit",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8
    b"\xf9\x7e\xcf\xd5\x84\xa2\x9d\xbd\x76\x5b\x08\x14\x3c\xb7\xcc\xd2\xf7\x64\xb2\x1e\xe0\x76\xe6\x15\x76\x78\x90\x2a\x4b\x7a\x48\xf5": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.1.8",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xf9\xd9\x05\xe0\xee\x2d\xf3\xed\x8d\xe9\x01\x3c\x2e\x89\x96\x6d\x54\x48\x4a\x93\xa6\x37\xd0\xe4\x8d\xb6\xf6\x17\x8b\xdc\x99\x36": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungiblePositionManager.sol",
            name="INonfungiblePosi tionManager",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungiblePositionManager.sol",
            name="INonfungiblePosi tionManager",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungiblePositionManager.sol",
            name="INonfungiblePosi tionManager",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungiblePositionManager.sol",
            name="INonfungiblePosi tionManager",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/INonfungiblePositionManager.sol",
            name="INonfungiblePosi tionManager",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xf9\xe9\xb8\xac\x62\xac\xf5\xff\xee\x57\x85\xd7\x18\xf7\xc2\xc2\x1c\x40\x86\x8c\xb0\x82\x89\x86\x7b\x7a\x1a\x74\x2b\xe5\xaf\x35": [
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/crosschain/optimism/CrossChainEnabledOptimism.sol",
            name="CrossChainEnabledOptimism",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.2, v0.1.3
    b"\xfa\x00\x37\xcb\x96\x05\x6a\x7f\xef\x02\x44\x48\xa1\x0d\xef\xcf\x8f\x81\xcd\xe3\xe4\x2c\xdc\x4a\x2b\x7e\x7b\x6f\x4f\x8f\x31\x0a": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.1.3",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.2.0-rc.0
    b"\xfa\x04\x73\xbe\x0b\xd0\x24\xaf\x70\x8a\xe1\x61\x8f\xbf\xd9\x8a\x07\xd8\xcd\x57\x7f\xd0\x7b\x61\xea\x5f\xdb\x46\x4c\xfc\x6c\x79": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/structs/BitMaps.sol",
            name="BitMaps",
            version="v4.2.0-rc.0",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xfa\x0e\x55\x12\x5f\x0b\x74\x9f\x99\x98\x99\x94\x76\xf0\x71\x96\x02\x74\xe1\x2b\x15\x53\x0f\xa3\x5b\x41\x89\xdd\x86\xfb\x61\xa3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC4626.sol",
            name="ERC4626",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v6.6.0, v6.6.1, v6.6.2
    b"\xfa\x1f\x05\x86\x4c\x4a\x6b\x78\x2e\x76\x09\x13\xc2\xb7\x89\x31\x4b\x1f\xf3\x59\x69\xc4\x4a\x8a\xe5\x18\x1e\x9a\xff\xec\xf7\x97": [
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.6.0",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.6.1",
        ),
        KnownContract(
            project="solmate",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v6.6.2",
        ),
    ],
    # Hash appears in versions: v1.0.0, v1.0.1, v1.1.1, v1.2.0, v1.3.0
    b"\xfa\x28\xce\xca\x76\xc4\x1d\xe6\x50\xc9\x35\x51\x78\xde\x8c\xd7\x23\xdb\x4e\x06\x6f\x7f\x00\x90\xb8\x45\x09\x9e\x91\xac\xd4\x0d": [
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPoolInitializer.sol",
            name="IPoolInitializer",
            version="v1.0.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPoolInitializer.sol",
            name="IPoolInitializer",
            version="v1.0.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPoolInitializer.sol",
            name="IPoolInitializer",
            version="v1.1.1",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPoolInitializer.sol",
            name="IPoolInitializer",
            version="v1.2.0",
        ),
        KnownContract(
            project="uniswap-v3-periphery",
            path="contracts/interfaces/IPoolInitializer.sol",
            name="IPoolInitializer",
            version="v1.3.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xfa\x30\x93\x08\x8b\x96\x61\x98\x9f\x77\x69\x8b\x3a\xf6\xba\x7c\xe4\x78\x99\x34\xe7\x4f\x6c\xba\x13\x06\x56\x6d\x79\xeb\xb3\xc6": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/ERC6909Claims.sol",
            name="ERC6909Claims",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.3.0-rc.1, v3.3.0-rc.2
    b"\xfa\x3f\xe2\x69\x74\x64\x4f\xb7\xa5\xb7\x3f\x5f\x83\x44\x61\x02\x7b\x28\x1b\xa6\x08\xa3\x72\x85\xb8\x01\x20\xe6\x29\x87\x82\xfa": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155Receiver.sol",
            name="ERC1155Receiver",
            version="v3.3.0-rc.2",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xfa\x54\x6c\xa0\x90\x16\x53\x8b\xeb\x4a\x16\x02\x18\x65\x1c\xff\x5f\x55\x97\x44\x39\x53\x91\x73\xdb\x60\x57\xa6\x2a\x03\x0b\x07": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/extensions/ERC721Wrapper.sol",
            name="ERC721Wrapper",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xfa\x58\x03\x9a\x84\xcd\x27\x4e\xd2\x34\xbf\x68\x92\xeb\x00\x4b\x5d\x0c\xe7\xec\xe9\x2c\x26\x71\xed\x0a\x99\xa3\xea\x2c\xf3\x2a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/ECDSA.sol",
            name="ECDSA",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xfa\x77\xe6\xc1\xe4\x9b\x98\x3a\x31\x24\x88\x7f\x01\xa6\x67\x5f\xf3\xb7\x65\x91\xb5\x8d\x78\xdb\x37\x52\xcd\x52\x47\xc4\x5f\x6a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorCountingSimple.sol",
            name="GovernorCountingSimple",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6
    b"\xfa\xcf\x4b\x90\xbe\x7e\x49\xa7\xc5\x12\xc4\x9f\x0e\x36\x81\x2a\x18\x37\xff\x59\x31\xf1\xeb\x11\x90\x76\x9d\x56\x83\xe5\x32\x76": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v4.9.6",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0
    b"\xfa\xe2\xc9\x96\x47\xab\xef\x3a\xdf\xae\xdd\x45\x9e\x56\x95\x9c\x19\xcb\x45\xa4\x60\xe9\x1f\x10\x52\x27\xe2\x33\x30\x30\xe6\x5b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/introspection/ERC165CheckerUpgradeable.sol",
            name="ERC165CheckerUpgradeable",
            version="v4.7.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.100, v0.0.101, v0.0.102, v0.0.103, v0.0.104, v0.0.105, v0.0.106, v0.0.107, v0.0.108, v0.0.109, v0.0.110, v0.0.111, v0.0.112, v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121, v0.0.122, v0.0.123, v0.0.124, v0.0.127, v0.0.129, v0.0.130, v0.0.131, v0.0.132, v0.0.133, v0.0.134, v0.0.135, v0.0.136, v0.0.137, v0.0.138, v0.0.139, v0.0.140, v0.0.141, v0.0.142, v0.0.143, v0.0.144, v0.0.145, v0.0.146, v0.0.147, v0.0.148, v0.0.149, v0.0.150, v0.0.151, v0.0.152, v0.0.153, v0.0.154, v0.0.155, v0.0.156, v0.0.69, v0.0.70, v0.0.71, v0.0.72, v0.0.73, v0.0.74, v0.0.75, v0.0.76, v0.0.77, v0.0.78, v0.0.79, v0.0.80, v0.0.81, v0.0.82, v0.0.83, v0.0.84, v0.0.85, v0.0.86, v0.0.87, v0.0.88, v0.0.89, v0.0.90, v0.0.92, v0.0.93, v0.0.94, v0.0.95, v0.0.96, v0.0.97, v0.0.98, v0.0.99
    b"\xfa\xf3\x27\x25\x1e\xf9\x71\xdc\x5a\x3f\xe1\xac\x7f\xf6\xa1\x1e\x1b\xd2\xe8\x4a\xf2\x3a\x29\xab\x66\x19\xeb\xa7\xdf\xe1\x9f\x90": [
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.100",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.101",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.102",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.103",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.104",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.105",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.106",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.107",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.108",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.109",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.110",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.111",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.112",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.121",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.122",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.123",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.124",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.127",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.129",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.130",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.131",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.132",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.133",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.134",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.135",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.136",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.137",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.138",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.139",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.140",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.141",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.142",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.143",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.144",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.145",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.146",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.147",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.148",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.149",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.150",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.151",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.152",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.153",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.154",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.155",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.156",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.69",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.70",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.71",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.72",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.73",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.74",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.75",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.76",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.77",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.78",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.79",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.80",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.81",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.82",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.83",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.84",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.85",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.86",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.90",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.92",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.93",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.94",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.95",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.96",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.97",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.98",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DateTimeLib.sol",
            name="DateTimeLib",
            version="v0.0.99",
        ),
    ],
    # Hash appears in versions: v0.0.289, v0.0.290
    b"\xfa\xfd\xe4\x59\xb1\x7c\xdd\xea\xef\xcb\x7d\xf7\xa8\x20\x3b\x2d\x2d\xdd\x34\xfa\xde\xab\x48\xf1\xda\xe5\x7a\x0e\xd2\x15\x17\x67": [
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/ext/zksync/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.290",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xfb\x05\xf1\x83\x3b\xc1\xca\x17\x82\xc3\x81\x64\x1a\x54\xa4\x46\x33\x3e\x02\xaa\x9e\x8b\x83\x69\xff\xc8\x0b\x48\xd3\x57\x04\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Capped.sol",
            name="ERC20Capped",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.141
    b"\xfb\x1f\x43\x9a\x6a\x77\x4a\x06\x31\x60\x50\xda\x4e\xb5\x56\xb0\x0c\xcf\xc2\xd0\x36\xa8\x7b\xb9\x48\x75\xca\xb8\xeb\x61\x93\xd2": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.141",
        ),
    ],
    # Hash appears in versions: v0.0.117
    b"\xfb\x3c\xe0\xee\x72\x4f\xa0\x89\x9d\x5a\x73\xb5\xe6\x5a\x10\x0c\x77\x02\xab\xe4\x9a\xcc\xdc\xb3\x71\xe0\x5c\xf6\x36\xdd\x34\xe3": [
        KnownContract(
            project="solady",
            path="src/utils/JSONParserLib.sol",
            name="JSONParserLib",
            version="v0.0.117",
        ),
    ],
    # Hash appears in versions: v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238
    b"\xfb\x73\x00\xe5\x55\x21\x14\x90\x60\xf1\xa5\x96\xb0\xf2\x68\xb3\x7c\x3a\x65\xff\xba\x34\x7c\xb3\x91\x7c\xb1\xe2\x33\x8d\xcf\x63": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.238",
        ),
    ],
    # Hash appears in versions: v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249
    b"\xfb\x7a\xae\xbd\x96\x54\x35\x49\xf6\xee\x51\xbe\x24\x46\x0c\xfa\x4c\xae\x8e\x67\xe7\xca\xec\x8c\x69\xe5\x07\x53\x7a\x6a\xa9\x24": [
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SSTORE2.sol",
            name="SSTORE2",
            version="v0.0.249",
        ),
    ],
    # Hash appears in versions: v0.0.79
    b"\xfb\x8a\xb4\x6d\xea\xe6\x6e\x8b\xb0\x9f\x6b\x79\x2e\x45\xe3\x95\xe1\x72\xcc\xb4\xd0\xc2\x8d\x7a\x4d\xb8\x7d\xc7\xb3\x7a\x9d\x7c": [
        KnownContract(
            project="solady",
            path="src/utils/FixedPointMathLib.sol",
            name="FixedPointMathLib",
            version="v0.0.79",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xfb\x93\xa5\xcb\x5d\xcf\xdc\x93\xc0\x14\xf1\x01\xcb\xea\x3f\x85\x4c\x20\x3e\xdd\x8e\x2f\x71\x92\x1c\xc6\x55\x61\xa8\x2b\xb3\xfe": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/Context.sol",
            name="Context",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xfb\x94\x76\x53\x3f\x0f\xc9\xff\x87\xb3\xfb\xed\xdc\x70\x37\x7e\x75\xd8\xc7\x1b\xe0\xf0\x1e\xfd\x32\x92\x93\x60\x65\x43\xe0\xf0": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/common/ERC2981Upgradeable.sol",
            name="ERC2981Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121
    b"\xfb\x95\x41\xf8\x97\x32\xe3\x85\x63\x16\xf7\xa4\x81\x3c\x1b\x8d\x9e\xd5\xf1\x2a\x25\x70\x5d\xc2\xda\x8e\xf0\x18\x27\x48\x9c\xd8": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-beta.1
    b"\xfb\xbc\x81\x76\xe7\xe1\x0d\xc5\xc1\xc1\x86\x9d\x85\xe7\x9a\xf0\x80\x14\x63\x32\xeb\x5b\x45\x2f\x0e\x0c\x56\xe8\x1a\xff\x5b\x0a": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/presets/ERC721PresetMinterPauserAutoId.sol",
            name="ERC721PresetMinterPauserAutoId",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.4.0-rc.0, v3.4.1
    b"\xfb\xcc\xd2\x36\xba\x1a\xc3\x6e\xa9\xee\xc4\x52\x9f\x81\xe0\x35\x49\xbd\x77\x57\xaf\x9d\xa3\xae\xf2\xde\x68\xe0\xdd\x41\xa5\x51": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="GSNRecipientERC20Fee",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3, v4.9.4, v4.9.5, v4.9.6, v5.0.0-rc.0
    b"\xfb\xe9\xba\xa3\xf6\x32\x80\x6c\x8b\xd4\xd5\xc5\x49\x6d\x8a\xea\x54\xc1\xa5\xf4\x49\xd8\xfa\x90\xf7\x2f\x63\x01\xe7\x7f\x95\x3c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v4.9.6",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC6372Upgradeable.sol",
            name="IERC6372Upgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xfc\x18\xba\xc3\xce\xb0\x54\x0e\x28\x39\xa6\x71\x42\xbb\xf0\xd0\x6d\x94\x8a\x5b\xdf\x98\xcf\x9c\x74\x9a\xa7\x4d\xf0\xde\x51\xac": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayHubUpgradeable.sol",
            name="IRelayHubUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v0.0.290, v0.0.291
    b"\xfc\x30\xa2\x6a\x53\xd4\x40\x77\x9e\x2c\x5e\x1d\x8a\x87\x3e\xa9\x31\xa4\x6c\x57\x54\x30\x91\xee\xbe\x9e\x76\x31\xd9\x3c\x12\x9f": [
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/SignatureCheckerLib.sol",
            name="SignatureCheckerLib",
            version="v0.0.291",
        ),
    ],
    # Hash appears in versions: v4.1.0, v4.1.0-rc.0
    b"\xfc\x3e\x64\x17\x68\x9f\xaa\x48\xa2\x46\x14\xce\xa5\x7a\x54\x89\x0b\x7a\x82\x60\xb5\x1f\xe9\x24\xb5\x7a\x8d\x86\x8f\xba\xc7\x27": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/TimelockController.sol",
            name="TimelockController",
            version="v4.1.0-rc.0",
        ),
    ],
    # Hash appears in versions: v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xfc\x3e\xbe\xcc\x16\xf0\x53\x40\x42\xc3\x0a\x20\xe0\x1d\x46\xd7\x42\xa8\x8c\x98\x26\x88\x4f\xe3\xce\xf0\xc5\xdb\xe0\xd3\xe9\x2b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/GSNRecipientERC20Fee.sol",
            name="__unstable__ERC20Owned",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v0.0.44, v0.0.45, v0.0.46, v0.0.47, v0.0.48, v0.0.49, v0.0.50, v0.0.51, v0.0.52, v0.0.53, v0.0.54, v0.0.55
    b"\xfc\x57\x04\x89\xab\x20\x4e\x6c\x31\x82\xf1\xf5\x7c\xa3\x8f\x51\x2c\x5a\x3d\x39\x29\xde\x87\x58\xb1\xe7\x5e\x56\x7d\xf4\xa1\xc3": [
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.44",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.45",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.46",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.47",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.48",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.49",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.50",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.51",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.52",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.53",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.54",
        ),
        KnownContract(
            project="solady",
            path="src/utils/LibBitmap.sol",
            name="LibBitmap",
            version="v0.0.55",
        ),
    ],
    # Hash appears in versions: v4.8.0-rc.0, v4.8.0-rc.1
    b"\xfc\x95\xe1\x90\x91\x21\x47\x79\x4c\x7d\x35\xb9\x96\x7d\xf3\x06\xe1\x0b\xf8\x1b\x0c\xbf\x4d\x20\xed\x01\x59\xea\x2f\x10\xe1\x6f": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/ERC721ConsecutiveUpgradeable.sol",
            name="ERC721ConsecutiveUpgradeable",
            version="v4.8.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.5.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.9.0, v4.9.0-rc.0, v4.9.0-rc.1, v4.9.1, v4.9.2, v4.9.3
    b"\xfc\xb5\x27\xd9\x1a\x0b\x37\x87\x1e\x84\x9d\x92\xc8\xc8\x75\x70\x82\x0c\xf0\xf8\xaa\x10\xb2\x37\xeb\x38\xc5\x6c\x1d\xd9\x8b\x52": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/ContextUpgradeable.sol",
            name="ContextUpgradeable",
            version="v4.9.3",
        ),
    ],
    # Hash appears in versions: v0.0.200
    b"\xfc\xea\x4a\xba\xa2\x94\x19\x4b\x88\x6f\xca\x4a\xcd\xfc\x42\x99\x8d\x59\xb7\x66\x85\xae\xf7\x84\xae\x69\x7e\x04\x5f\x75\x14\x23": [
        KnownContract(
            project="solady",
            path="src/utils/SafeCastLib.sol",
            name="SafeCastLib",
            version="v0.0.200",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xfc\xf5\x47\x16\x01\x55\xfd\x5b\x5a\xae\xe9\xf0\xa6\x00\xcd\x62\x4b\x91\x07\x29\xc8\x58\x9f\xac\x65\xea\x0b\x1b\x2e\xc0\x1f\x09": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/common/ERC2981.sol",
            name="ERC2981",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xfd\x0c\xb1\x4e\x11\xd0\x4c\xfe\x5d\xf9\x08\x88\x7a\xb1\xd9\x4e\xb5\x98\x41\x53\x94\x7f\x9d\x4c\x9a\xee\x57\xbd\x26\xcc\x62\x40": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/utils/cryptography/draft-EIP712.sol",
            name="EIP712",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-rc.0
    b"\xfd\x1b\x58\x1a\x06\x11\xaa\x99\x3e\xf3\xf3\x74\x68\xe4\x60\x96\x1a\xfe\x28\x94\xf2\x06\xad\x8d\x89\x2d\x7f\x5d\xd0\x8a\xbd\x4a": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v1.0.0
    b"\xfd\x2f\xc8\x72\xc2\x19\x1a\x80\x81\xb2\x75\x78\x71\x8e\x28\x31\xb4\x2b\x07\xb1\xea\xc7\x24\xb8\x12\xa4\x08\x9d\x81\x29\x2e\x9f": [
        KnownContract(
            project="uniswap-v3-core",
            path="contracts/interfaces/callback/IUniswapV3SwapCallback.sol",
            name="IUniswapV3Sw apCallback",
            version="v1.0.0",
        ),
    ],
    # Hash appears in versions: v4.0.0
    b"\xfd\x3a\xba\x2c\x1e\x6b\x76\xb9\x83\xd6\x1d\x92\x9f\x3b\x0e\xb5\x24\x62\xd3\x1f\x57\xae\x02\x55\xa2\xc6\x9c\xbe\x2d\x9f\x69\x6f": [
        KnownContract(
            project="uniswap-v4-core",
            path="src/libraries/LPFeeLibrary.sol",
            name="LPFeeLibrary",
            version="v4.0.0",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xfd\x45\x34\xad\x43\x85\x46\x9f\xf6\xf9\x09\xf3\x90\x97\xe0\xe9\x1c\x5f\xd6\xda\xe6\x38\x61\x5a\x51\x78\x39\x03\x95\x38\xa1\xfb": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/access/AccessControlUpgradeable.sol",
            name="AccessControlUpgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xfd\x68\x6c\x71\x49\xc4\x69\x0f\xef\x99\x5a\xe3\xdd\x78\xf9\xb0\xb0\x14\x46\x4a\x02\xfb\x3d\x95\x1d\x4e\x5c\xb1\x32\x95\xd9\x23": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorCountingFractional.sol",
            name="GovernorCountingFractional",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275
    b"\xfd\x73\xc0\x7a\xdf\xea\xc1\x7f\x18\x28\x38\xb0\x22\x37\x38\x26\xe0\x33\xc0\xd7\xef\x44\x44\x95\x59\x5e\xf6\xbe\x59\xb3\x9b\xe0": [
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/accounts/ERC6551.sol",
            name="ERC6551",
            version="v0.0.275",
        ),
    ],
    # Hash appears in versions: v0.0.192, v0.0.193, v0.0.194, v0.0.195, v0.0.196, v0.0.197, v0.0.198, v0.0.199, v0.0.200, v0.0.201, v0.0.202, v0.0.203, v0.0.204, v0.0.205, v0.0.206, v0.0.207, v0.0.208, v0.0.209, v0.0.210, v0.0.211, v0.0.212, v0.0.213, v0.0.214, v0.0.215, v0.0.216, v0.0.217, v0.0.218, v0.0.219, v0.0.220, v0.0.221, v0.0.222, v0.0.223, v0.0.224, v0.0.225, v0.0.226, v0.0.227, v0.0.228, v0.0.229, v0.0.230, v0.0.231, v0.0.232, v0.0.233, v0.0.234, v0.0.235, v0.0.236, v0.0.237, v0.0.238, v0.0.239, v0.0.240, v0.0.241, v0.0.242, v0.0.243, v0.0.244, v0.0.245, v0.0.246, v0.0.247, v0.0.248, v0.0.249, v0.0.250, v0.0.251, v0.0.252, v0.0.253, v0.0.254, v0.0.255, v0.0.256, v0.0.257, v0.0.258, v0.0.259, v0.0.260, v0.0.261, v0.0.262, v0.0.263, v0.0.264, v0.0.265, v0.0.266, v0.0.267, v0.0.268, v0.0.269, v0.0.270, v0.0.271, v0.0.272, v0.0.273, v0.0.274, v0.0.275, v0.0.276, v0.0.277, v0.0.278, v0.0.279, v0.0.280, v0.0.281, v0.0.282, v0.0.283, v0.0.284, v0.0.285, v0.0.286, v0.0.287, v0.0.288, v0.0.289, v0.0.290, v0.0.291, v0.0.292, v0.0.293, v0.0.294, v0.0.295, v0.0.296, v0.0.297, v0.0.298, v0.1.0, v0.1.1, v0.1.10, v0.1.11, v0.1.12, v0.1.13, v0.1.14, v0.1.15, v0.1.16, v0.1.2, v0.1.3, v0.1.4, v0.1.5, v0.1.6, v0.1.7, v0.1.8, v0.1.9
    b"\xfd\x8e\x2c\xc0\xf6\x46\x0a\xa8\xae\xff\x29\xcc\x20\x8f\x34\x40\xe5\x26\x9c\x78\x49\x73\x22\x00\xbd\x58\x9e\xdb\x8b\x6c\xc5\xe0": [
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.192",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.193",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.194",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.195",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.196",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.197",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.198",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.199",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.200",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.201",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.202",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.203",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.204",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.205",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.206",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.207",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.208",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.209",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.210",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.211",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.212",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.213",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.214",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.215",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.216",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.217",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.218",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.219",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.220",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.221",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.222",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.223",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.224",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.225",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.226",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.227",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.228",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.229",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.230",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.231",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.232",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.233",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.234",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.235",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.236",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.237",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.238",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.239",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.240",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.241",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.242",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.243",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.244",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.245",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.246",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.247",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.248",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.249",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.250",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.251",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.252",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.253",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.254",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.255",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.256",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.257",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.258",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.259",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.260",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.261",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.262",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.263",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.264",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.265",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.266",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.267",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.268",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.269",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.270",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.271",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.272",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.273",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.274",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.275",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.276",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.277",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.278",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.279",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.280",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.281",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.282",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.283",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.284",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.285",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.286",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.287",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.288",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.289",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.290",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.291",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.292",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.293",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.294",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.295",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.296",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.297",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.0.298",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.0",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.1",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.10",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.11",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.12",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.13",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.14",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.15",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.16",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.2",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.3",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.4",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.5",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.6",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.7",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.8",
        ),
        KnownContract(
            project="solady",
            path="src/utils/DeploylessPredeployQueryer.sol",
            name="DeploylessPredeployQueryer",
            version="v0.1.9",
        ),
    ],
    # Hash appears in versions: v0.0.40
    b"\xfd\x9a\x63\x80\xfb\xe8\x51\xb4\x75\x61\x9a\x9a\x80\xfa\x47\x61\x5b\x9c\x0d\x4f\x0c\x52\x47\x2b\xac\x34\x6c\x76\x9d\xa0\x67\x33": [
        KnownContract(
            project="solady",
            path="src/utils/DynamicBufferLib.sol",
            name="DynamicBufferLib",
            version="v0.0.40",
        ),
    ],
    # Hash appears in versions: v3.0.0-beta.0, v3.0.0-rc.0, v3.0.0-beta.0, v3.0.0-rc.0
    b"\xfd\xbc\x72\x44\x0a\x78\xc3\xa4\xca\x23\xdb\x2a\xa9\xd3\x87\xdb\xd6\x64\x11\xbf\xdf\xdd\x1b\xd1\x46\x10\xd2\x5d\xe0\x7d\x23\xac": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC721/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v3.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v3.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/ERC721Enumerable.sol",
            name="ERC721Enumerable",
            version="v3.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.5.1, v4.5.2, v4.3.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.5.1, v4.5.2, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xfd\xd9\x1a\x54\x04\xa2\x26\x9f\x0c\xa1\x03\x59\x6c\xac\xa4\xc7\xd1\x9c\x33\x5d\x41\x41\x27\x8f\x00\x72\xb5\x2c\xf9\x82\x33\x3b": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.5.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/interfaces/IERC1363Upgradeable.sol",
            name="IERC1363Upgradeable",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v5.0.0-rc.0
    b"\xfd\xeb\x42\xe0\xb2\x3d\x44\xe5\x5e\x12\xac\x39\xfc\x40\x19\x1b\xa4\x65\x61\x29\xd9\xc7\xa1\x30\xc7\x26\x77\x0b\xd9\xfe\x1e\x30": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/extensions/GovernorTimelockControlUpgradeable.sol",
            name="GovernorTimelockControlUpgradeable",
            version="v5.0.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3
    b"\xfd\xee\x88\xe0\x0a\xa7\xb7\xca\x8e\xe9\xe1\xd9\xa5\x5a\x4b\xb2\x53\x35\xc2\x75\x5c\xa2\x98\x9b\x0f\x93\xb7\x1c\x76\xa0\x8f\xbe": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC721/extensions/draft-ERC721VotesUpgradeable.sol",
            name="ERC721VotesUpgradeable",
            version="v4.7.3",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xfe\x1b\x5b\x15\xfc\xed\x58\xf7\x6d\xb4\x4e\xdf\xbc\x39\x60\x44\x43\x78\x5c\x54\x67\x69\x5b\x00\xf3\x0e\x34\xbb\x8b\x20\x6d\xdf": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC4626Upgradeable.sol",
            name="ERC4626Upgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.113, v0.0.114, v0.0.115, v0.0.116, v0.0.117, v0.0.118, v0.0.119, v0.0.120, v0.0.121
    b"\xfe\x5b\xcd\xaf\xa5\x18\xdc\x45\x45\xed\x3a\x95\x7b\x20\x79\xc0\xe0\x4c\xb1\x66\xdc\x24\x0b\x3c\xc7\x23\x35\x2f\x6a\x0e\x42\x4a": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.113",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.114",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.115",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.116",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.117",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.118",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.119",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.120",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC1155.sol",
            name="ERC1155",
            version="v0.0.121",
        ),
    ],
    # Hash appears in versions: v3.1.0, v3.1.0-rc.0, v3.2.0, v3.2.0-rc.0, v3.3.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0, v3.4.0-rc.0, v3.4.1, v3.4.2, v3.1.0, v3.1.0-rc.0, v3.2.0-rc.0, v3.3.0-rc.0, v3.3.0-rc.1, v3.3.0-rc.2, v3.4.0-rc.0, v3.4.1
    b"\xfe\x65\x9f\xb8\x18\x13\xa8\xc6\x24\xe7\xa1\x15\xb1\x0b\x38\x04\xf7\x7f\xe6\xe5\x0f\x28\x58\x9b\x9e\xba\x66\x57\xde\x15\x62\xdb": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.3.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC1155PresetMinterPauser.sol",
            name="ERC1155PresetMinterPauser",
            version="v3.4.1",
        ),
    ],
    # Hash appears in versions: v3.3.0, v3.3.0-solc-0.7
    b"\xfe\xa9\x3b\x2e\xda\x59\xa5\x7f\x66\xb2\xb4\xd3\xde\x6a\xdb\xb0\xcf\xba\xb5\x5b\x32\x2b\xa7\xf9\xc3\x13\x92\x9a\x9a\x34\x39\x9e": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/cryptography/ECDSAUpgradeable.sol",
            name="ECDSAUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
    ],
    # Hash appears in versions: v4.4.1, v4.4.2
    b"\xfe\xbb\x09\xe9\x80\x24\xb2\x03\x2f\x95\x70\x97\x64\xc5\xd6\xdf\xa6\x3d\xa3\x2d\x33\x99\xaa\xe1\x18\xbd\x8f\xd8\x8f\x0b\x48\x85": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC777/ERC777Upgradeable.sol",
            name="ERC777Upgradeable",
            version="v4.4.2",
        ),
    ],
    # Hash appears in versions: v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xfe\xe2\x91\x96\xbc\x90\x55\xf7\xad\xaa\xc2\x63\xba\x4e\x2b\xe0\x39\x57\xcb\xa2\xcb\xb9\x67\x92\x11\x5a\xc8\xb2\xfc\xac\x83\xc4": [
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/utils/cryptography/SignatureChecker.sol",
            name="SignatureChecker",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.2.2-solc-0.7, v3.3.0-solc-0.7, v3.4.0-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2, v3.4.2-solc-0.7, v3.1.0-solc-0.7, v3.2.1-solc-0.7, v3.4.1-solc-0.7, v3.4.1-solc-0.7-2
    b"\xfe\xea\x9a\xb4\x4d\xfb\x2f\x2b\x9f\xa4\x7c\xc8\x9e\xbf\x32\x33\x38\x91\x3f\x62\xca\xcd\xcd\x0f\x69\x45\x8e\x17\xa4\x37\x62\x4d": [
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1-solc-0.7-2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.1.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.2.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v3.4.1-solc-0.7-2",
        ),
    ],
    # Hash appears in versions: v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xff\x2c\x86\x8c\x50\xc5\x20\x92\x0a\xff\x60\x16\x20\x85\x69\xa4\xfb\x3a\xee\x1d\x27\x94\x5f\x26\xb1\x1f\x49\xdc\x96\x82\x80\x3c": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC1155/ERC1155ReceiverUpgradeable.sol",
            name="ERC1155ReceiverUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v3.2.0, v3.2.2-solc-0.7, v3.3.0, v3.3.0-solc-0.7, v3.4.0, v3.4.0-solc-0.7, v3.4.2, v3.4.2-solc-0.7
    b"\xff\x4c\x97\xf4\xee\x6b\xd7\xb7\xe8\x0d\xba\x69\x0b\xe8\xe2\xe7\xe2\xb1\x57\xd1\x9c\xca\xef\xf6\xc5\x4e\xf0\xe8\x59\xee\x10\x61": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.2.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.2.2-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.3.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.3.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.4.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.4.0-solc-0.7",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.4.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/GSN/IRelayRecipientUpgradeable.sol",
            name="IRelayRecipientUpgradeable",
            version="v3.4.2-solc-0.7",
        ),
    ],
    # Hash appears in versions: v5.3.0, v5.3.0-rc.0
    b"\xff\x63\x97\x12\xe2\x6a\xbb\xa2\x8a\x4c\x41\x4b\xee\x02\x11\x05\x89\x1e\xb8\xae\xdf\x38\x8b\x1d\x40\x62\xc2\xaf\xb4\xaa\x22\xe5": [
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/interfaces/draft-IERC6909.sol",
            name="IERC6909",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.0.0, v4.0.0-beta.0, v4.0.0-beta.1, v4.0.0-rc.0, v4.1.0, v4.1.0-rc.0, v4.0.0-beta.1
    b"\xff\x71\x36\x95\x19\x7e\x07\xfe\xff\x3e\x7a\x3c\x68\x80\x32\x5f\x15\x45\x0c\xab\x94\x67\x73\xb9\x90\x61\x29\xab\xf0\x38\x8f\x09": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.0.0-beta.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.0.0-beta.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/presets/ERC20PresetMinterPauser.sol",
            name="ERC20PresetMinterPauser",
            version="v4.0.0-beta.1",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.2, v5.0.1, v5.0.2
    b"\xff\x82\x1b\xa3\x1a\x35\xd7\x74\x36\x72\xa7\x6f\x35\xc7\xe3\x1b\xc4\x34\x6f\x50\x27\x92\xf9\xde\xc6\x19\xbe\xcb\xed\xf1\x92\x55": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/governance/GovernorUpgradeable.sol",
            name="GovernorUpgradeable",
            version="v5.0.2",
        ),
    ],
    # Hash appears in versions: v0.0.46
    b"\xff\xad\xa3\x19\xaf\x46\xbd\x22\x38\x96\xce\x28\xfc\x9f\xe1\xee\xab\xe2\x3b\x93\x0a\x39\xc2\x10\xa5\x0d\x00\xd1\x7f\xad\x7b\x05": [
        KnownContract(
            project="solady",
            path="src/utils/LibPRNG.sol",
            name="LibPRNG",
            version="v0.0.46",
        ),
    ],
    # Hash appears in versions: v5.0.0, v5.0.0-rc.0, v5.0.0-rc.1, v5.0.0-rc.2, v5.0.1, v5.0.2, v5.1.0, v5.1.0-rc.0, v5.1.0-rc.1, v5.2.0, v5.2.0-rc.0, v5.2.0-rc.1, v5.3.0, v5.3.0-rc.0
    b"\xff\xb7\x39\x46\xb4\xb4\x56\xe6\xcd\xcf\x3d\x48\x4c\x17\xe2\x4f\xaa\xfc\xfb\xf4\x3a\x9c\xb5\x7a\xed\x02\xc8\xcd\x34\x17\x2f\x86": [
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.0.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.1.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.1.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.1.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.2.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/governance/extensions/GovernorTimelockControl.sol",
            name="GovernorTimelockControl",
            version="v5.3.0-rc.0",
        ),
    ],
    # Hash appears in versions: v4.2.0, v4.2.0-rc.0, v4.3.0, v4.3.0-rc.0, v4.3.1, v4.3.2, v4.3.3, v4.4.0, v4.4.0-rc.0, v4.4.0-rc.1, v4.4.1, v4.4.2, v4.5.0, v4.5.0-rc.0, v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3, v4.2.0-rc.0, v4.3.0-rc.0, v4.4.0-rc.0, v4.4.0-rc.1
    b"\xff\xc2\x10\x70\xd1\xde\x1c\x21\x79\x13\x39\xa4\xa0\xa4\x93\x8a\xbb\x0f\x0d\x5b\x96\xb1\x79\x0a\xb3\xf1\x55\x66\xe7\x76\x9b\xb9": [
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.2.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.5.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.5.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.8.3",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.2.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.3.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20Snapshot.sol",
            name="ERC20Snapshot",
            version="v4.4.0-rc.1",
        ),
    ],
    # Hash appears in versions: v4.6.0, v4.6.0-rc.0, v4.7.0, v4.7.0-rc.0, v4.7.1, v4.7.2, v4.7.3, v4.8.0, v4.8.0-rc.0, v4.8.0-rc.1, v4.8.0-rc.2, v4.8.1, v4.8.2, v4.8.3
    b"\xff\xd4\xd8\xaa\x99\x2f\xa1\xd1\x86\xfc\xf6\x9d\x5c\x51\x75\x72\xf6\x99\x96\xf6\x2e\x31\xcc\xda\xe6\xb3\x45\x59\xb3\x73\xdb\xb3": [
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.6.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.6.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.7.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.7.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.7.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.7.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.7.3",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.0-rc.0",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.0-rc.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.0-rc.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.1",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.2",
        ),
        KnownContract(
            project="openzeppelin",
            path="contracts/vendor/optimism/ICrossDomainMessenger.sol",
            name="ICrossDomainMessenger",
            version="v4.8.3",
        ),
    ],
    # Hash appears in versions: v0.0.87, v0.0.88, v0.0.89, v0.0.90
    b"\xff\xf4\x7a\xdc\x9e\xee\x72\x1a\xe0\x08\xf4\x36\x99\x0c\xbe\xac\x9c\xdc\x8c\xd7\xa8\xa4\x60\x4e\x77\x15\xb7\xc7\x80\x8b\x7c\xba": [
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.87",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.88",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.89",
        ),
        KnownContract(
            project="solady",
            path="src/tokens/ERC721.sol",
            name="ERC721",
            version="v0.0.90",
        ),
    ],
    # Hash appears in versions: v4.9.4, v4.9.5, v4.9.6
    b"\xff\xf6\x71\x11\x5e\x38\xa9\x46\x3f\x83\xe5\x5f\xd5\xe0\x22\xee\xb4\x75\x30\x8b\x2e\xa3\x0a\x19\x05\xdf\x5c\x8a\x31\x29\x40\x3d": [
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.4",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.5",
        ),
        KnownContract(
            project="openzeppelin-upgradeable",
            path="contracts/token/ERC20/extensions/ERC20FlashMintUpgradeable.sol",
            name="ERC20FlashMintUpgradeable",
            version="v4.9.6",
        ),
    ],
}